/*
 * jQuery treeTable Plugin unreleased
 * http://ludo.cubicphuse.nl/jquery-plugins/treeTable/doc/
 *
 * Copyright 2011, Ludo van den Boom
 * Dual licensed under the MIT or GPL Version 2 licenses.
 */

(function($) {
  // Helps to make options available to all functions
  // TODO: This gives problems when there are both expandable and non-expandable
  // trees on a page. The options shouldn't be global to all these instances!
  var options;
  var defaultPaddingLeft;

  $.fn.treeTable = function(opts) {
    options = $.extend({}, $.fn.treeTable.defaults, opts);

    return this.each(function() {
      $(this).addClass("treeTable").find("tbody tr").each(function() {
        // Initialize root nodes only if possible
        if(!options.expandable || $(this)[0].className.search(options.childPrefix) == -1) {
          // To optimize performance of indentation, I retrieve the padding-left
          // value of the first root node. This way I only have to call +css+
          // once.
          if (isNaN(defaultPaddingLeft)) {
            defaultPaddingLeft = parseInt($($(this).children("td")[options.treeColumn]).css('padding-left'), 10);
          }

          initialize($(this));
        } else if(options.initialState == "collapsed") {
          this.style.display = "none"; // Performance! $(this).hide() is slow...
        }
      });
    });
  };

  $.fn.treeTable.defaults = {
    childPrefix: "child-of-",
    clickableNodeNames: false,
    expandable: true,
    indent: 19,
    initialState: "collapsed",
    onNodeShow: null,
    treeColumn: 0
  };

  // Recursively hide all node's children in a tree
  $.fn.collapse = function() {
    $(this).addClass("collapsed");

    childrenOf($(this)).each(function() {
      if(!$(this).hasClass("collapsed")) {
        $(this).collapse();
      }

      this.style.display = "none"; // Performance! $(this).hide() is slow...
    });

    return this;
  };

  // Recursively show all node's children in a tree
  $.fn.expand = function() {
    $(this).removeClass("collapsed").addClass("expanded");

    childrenOf($(this)).each(function() {
      initialize($(this));

      if($(this).is(".expanded.parent")) {
        $(this).expand();
      }

      // this.style.display = "table-row"; // Unfortunately this is not possible with IE :-(
      $(this).show();

      if($.isFunction(options.onNodeShow)) {
        options.onNodeShow.call();
      }
    });

    return this;
  };

  // Reveal a node by expanding all ancestors
  $.fn.reveal = function() {
    $(ancestorsOf($(this)).reverse()).each(function() {
      initialize($(this));
      $(this).expand().show();
    });

    return this;
  };

  // Add an entire branch to +destination+
  $.fn.appendBranchTo = function(destination) {
    var node = $(this);
    var parent = parentOf(node);

    var ancestorNames = $.map(ancestorsOf($(destination)), function(a) { return a.id; });

    // Conditions:
    // 1: +node+ should not be inserted in a location in a branch if this would
    //    result in +node+ being an ancestor of itself.
    // 2: +node+ should not have a parent OR the destination should not be the
    //    same as +node+'s current parent (this last condition prevents +node+
    //    from being moved to the same location where it already is).
    // 3: +node+ should not be inserted as a child of +node+ itself.
    if($.inArray(node[0].id, ancestorNames) == -1 && (!parent || (destination.id != parent[0].id)) && destination.id != node[0].id) {
      indent(node, ancestorsOf(node).length * options.indent * -1); // Remove indentation

      if(parent) { node.removeClass(options.childPrefix + parent[0].id); }

      node.addClass(options.childPrefix + destination.id);
      move(node, destination); // Recursively move nodes to new location
      indent(node, ancestorsOf(node).length * options.indent);
    }

    return this;
  };

  // Add reverse() function from JS Arrays
  $.fn.reverse = function() {
    return this.pushStack(this.get().reverse(), arguments);
  };

  // Toggle an entire branch
  $.fn.toggleBranch = function() {
    if($(this).hasClass("collapsed")) {
      $(this).expand();
    } else {
      $(this).removeClass("expanded").collapse();
    }

    return this;
  };

  // === Private functions

  function ancestorsOf(node) {
    var ancestors = [];
    while(node = parentOf(node)) {
      ancestors[ancestors.length] = node[0];
    }
    return ancestors;
  };

  function childrenOf(node) {
    return $("table.treeTable tbody tr." + options.childPrefix + node[0].id);
  };

  function getPaddingLeft(node) {
    var paddingLeft = parseInt(node[0].style.paddingLeft, 10);
    return (isNaN(paddingLeft)) ? defaultPaddingLeft : paddingLeft;
  }

  function indent(node, value) {
    var cell = $(node.children("td")[options.treeColumn]);
    cell[0].style.paddingLeft = getPaddingLeft(cell) + value + "px";

    childrenOf(node).each(function() {
      indent($(this), value);
    });
  };

  function initialize(node) {
    if(!node.hasClass("initialized")) {
      node.addClass("initialized");

      var childNodes = childrenOf(node);

      if(!node.hasClass("parent") && childNodes.length > 0) {
        node.addClass("parent");
      }

      if(node.hasClass("parent")) {
        var cell = $(node.children("td")[options.treeColumn]);
        var padding = getPaddingLeft(cell) + options.indent;

        childNodes.each(function() {
          $(this).children("td")[options.treeColumn].style.paddingLeft = padding + "px";
        });

        if(options.expandable) {
          cell.prepend('<span style="margin-left: -' + options.indent + 'px; padding-left: ' + options.indent + 'px" class="expander"></span>');
          $(cell[0].firstChild).click(function() { node.toggleBranch(); });

          if(options.clickableNodeNames) {
            cell[0].style.cursor = "pointer";
            $(cell).click(function(e) {
              // Don't double-toggle if the click is on the existing expander icon
              if (e.target.className != 'expander') {
                node.toggleBranch();
              }
            });
          }

          // Check for a class set explicitly by the user, otherwise set the default class
          if(!(node.hasClass("expanded") || node.hasClass("collapsed"))) {
            node.addClass(options.initialState);
          }

          if(node.hasClass("expanded")) {
            node.expand();
          }
        }
      }
    }
  };

  function move(node, destination) {
    node.insertAfter(destination);
    childrenOf(node).reverse().each(function() { move($(this), node[0]); });
  };

  function parentOf(node) {
    var classNames = node[0].className.split(' ');

    for(var key=0; key<classNames.length; key++) {
      if(classNames[key].match(options.childPrefix)) {
        return $("#" + classNames[key].substring(9));
      }
    }

    return null;
  };
})(jQuery);
/*
 * Katello JavaScript Menu File
 * Author: jrist
 * Date: 10/25/11
 * Time: 10:19 AM
 */

/**
 * Document Ready function
 */

$(document).ready(function (){
    $("#tabs").tabs();
});
$(document).ready(function() {
    var spinner = '<i class="fa fa-spinner fa-spin"></i>';

    $("#content_tabs").tabs({
        cache: true,
        load: function(event, ui) {
            KT.redhat_provider_page.init_tab($(ui.panel));
        }
    });

    $("#content_tabs > .ui-tabs-panel").append(spinner);

    $("#content_tabs").delegate('.repo_enable', 'change', function() {
        KT.redhat_provider_page.repoChange($(this), false);
    });

    $("#content_tabs").delegate('.repo_set_refresh', 'click', function(){
        var icon = $(this),
            row = icon.parents('.repo_set').first(),
            expanderArea;

        if(!icon.hasClass('disabled')){
            expanderArea = row.find('.expander_area');
            console.log(expanderArea);
            KT.redhat_provider_page.repoSetRefresh(expanderArea);
        }
    });
});


KT.redhat_provider_page = (function($) {
    var repoChange = function(checkbox) {

        var name = checkbox.attr("name"),
            options = {},
            url = checkbox.attr("data-url"),
            id = checkbox.attr("value"),
            set_checkbox = checkbox.parents(".repo_set").find('.repo_set_enable');

        if (checkbox.is(":checked")) {
            options['repo'] = "1";
        } else {
            options['repo'] = "0";
        }
        options['pulp_id'] = checkbox.attr("data-pulp-id");
        options['content_id'] = checkbox.attr("data-content-id");
        options['releasever'] = checkbox.attr("data-releasever");
        options['basearch'] = checkbox.attr("data-basearch");
        options['registry_name'] = checkbox.attr("data-registry-name");

        $(checkbox).hide();
        $('#spinner_'+id).removeClass('hidden').show();
        checkbox.parent().parent().siblings().first().find('.errors').remove();
        $.ajax({
            type: "PUT",
            url: url,
            data: options,
            cache: false,
            success: function(data, textStatus, jqXHR){
              KT.redhat_provider_page.checkboxHighlightRow(checkbox.attr("data-pulp-id"));
              if(data['can_disable_repo_set']){
                set_checkbox.removeAttr('disabled');
              }
              else {
                set_checkbox.attr('disabled','disabled');
              }
            },
            error: function(data) {
                checkbox.parent().parent().siblings().first().append(data.responseText);
                $(checkbox).removeClass('hidden').show().attr('checked', false);
                $('#spinner_'+id).addClass('hidden').hide();
            }
        });
        return false;
    },
    repoSetRefresh = function(expanderArea) {
        var url = expanderArea.data('url'),
            content_id = expanderArea.data('content-id');
        refresh_repo_set(url, content_id);
    },
    disable_repo_set = function(url, content_id){
        var row = $('#repo_set_' + content_id);
        hide_repos(content_id);
        row.find('.repo_set_enable').hide();
        row.find('.repo_set_spinner').show();
        row.addClass("disable");
        row.find('.repo_set_refresh').addClass('disabled');
        $.ajax({
            type: "PUT",
            url: url,
            data: {content_id:content_id},
            cache: false,
            success: function(data){
                var checkbox = row.find('.repo_set_enable');
                row.find('table').replaceWith('<table style="display: none;"> </table>');
                checkbox.show();
                if (checkbox.data('orphaned')) {
                    checkbox.attr("disabled", "disabled");
                }
                row.find('.repo_set_spinner').hide();
                row.find('.expander').removeClass('disabled').hide();
                row.find('.repo_set_refresh').removeClass('disabled').hide();
            },
            error: function(){
                row.removeClass("disable");
                row.find('.repo_set_enable').show();
                row.find('.repo_set_spinner').hide();
                row.find('.repo_set_refresh').removeClass('disabled');
            }
        });
    },
    refresh_repo_set = function(url, content_id){
        var row = $('#repo_set_' + content_id);
        row.addClass("disable");
        row.find('.expander').hide();
        row.find('.repo_set_spinner').show();
        $.ajax({
            type: "GET",
            url: url,
            data: {content_id:content_id},
            cache: false,
            success: function(data){
                row.find('table').replaceWith(data);
                row.removeClass("disable");
                row.find('.repo_set_spinner').hide();
                row.find('.expander').show();
                row.find('table').show();
            },
            error: function(){
                row.removeClass("disable");
                row.find('.expander').show();
                row.find('.repo_set_spinner').hide();
            }
        });
    },
    checkboxHighlightRow = function(id){
      $('#repo-'+id).effect('highlight', {'color':'#390'}, 1000);
      $('#spinner_'+id).hide().addClass('hidden');
      $('#input_repo_'+id).show();
    },
    on_node_show = function(a, b, c){
    },
    hide_repos = function(content_id){
        var row = $('#repo_set_' + content_id);
        if(row.hasClass('expanded')){
            row.find('.expander_area').click();
        }
    },
    show_repos = function(content_id){
        var row = $('#repo_set_' + content_id);
        if(row.hasClass('collapsed')){
            row.find('.expander_area').click();
        }
    },
    init_tab = function(panel) {
        panel.find(".tree_table").treeTable({
            expandable: true,
            initialState: "collapsed",
            clickableNodeNames: true,
            onNodeShow: KT.redhat_provider_page.on_node_show
        });

        panel.find(".content_table").delegate('.expander_area', 'click', function(){
            var area = $(this),
                row = area.parents('tr').first();
            if(row.hasClass("collapsed")){
                KT.redhat_provider_page.repoSetRefresh(area);
                row.removeClass("collapsed").addClass('expanded');
            } else {
                area.parent().find('table').hide();
                row.addClass("collapsed").removeClass('expanded');
            }
        });

    };

    return {
        repoChange: repoChange,
        checkboxHighlightRow: checkboxHighlightRow,
        on_node_show: on_node_show,
        repoSetRefresh: repoSetRefresh,
        hide_repos: hide_repos,
        init_tab: init_tab
    };
}(jQuery));



