/**
 * @ngdoc module
 * @name  Bastion.host-collections
 *
 * @description
 *   Module for host collections related functionality.
 */

angular.module('Bastion.host-collections', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.common',
    'Bastion.utils'
]);
/**
 * @ngdoc object
 * @name  Bastion.host-collections.controller:HostCollectionAddHostsController
 *
 * @requires $scope
 * @requires $location
 * @requires translate
 * @requires Nutupane
 * @requires CurrentOrganization
 * @requires Host
 * @requires HostCollection
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the host collection add content hosts pane.
 */

angular.module('Bastion.host-collections').controller('HostCollectionAddHostsController',
    ['$scope', '$state', '$location', 'translate', 'Nutupane', 'CurrentOrganization', 'Host', 'HostCollection', 'Notification',
    function ($scope, $state, $location, translate, Nutupane, CurrentOrganization, Host, HostCollection, Notification) {
        var contentNutupane, params, nutupaneParams;

        params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'page': 1,
            'sort_by': 'name',
            'sort_order': 'ASC',
            'paged': true
        };

        nutupaneParams = {
            'disableAutoLoad': true
        };
        contentNutupane = new Nutupane(Host, params, undefined, nutupaneParams);
        $scope.controllerName = 'hosts';
        contentNutupane.searchTransform = function (term) {
            var addition = "-host_collection_id=" + $scope.$stateParams.hostCollectionId;
            if (term === "" || angular.isUndefined(term)) {
                return addition;
            }

            return term + " and " + addition;
        };

        contentNutupane.refresh();
        $scope.table = contentNutupane.table;
        $scope.isAdding = false;
        $scope.table.closeItem = function () {};

        $scope.disableAddButton = function () {
            return $scope.table.numSelected === 0 || $scope.isAdding;
        };

        $scope.addSelected = function () {
            var selected;
            selected = _.map($scope.table.getSelected(), 'id');

            $scope.isAdding = true;
            HostCollection.addHosts({id: $scope.hostCollection.id, 'host_ids': selected}, function (data) {
                angular.forEach(data.displayMessages.success, function (success) {
                    Notification.setSuccessMessage(success);
                });

                angular.forEach(data.displayMessages.error, function (error) {
                    Notification.setErrorMessage(error);
                });

                $scope.isAdding = false;
                contentNutupane.refresh();
                $scope.refreshHostCollection();
            }, function (response) {
                Notification.setErrorMessage(response.data.displayMessage);
                $scope.isAdding = false;
            });
        };

    }]
);
(function () {
    'use strict';

    /**
     * @ngdoc controller
     * @name  Bastion.host-collections.controller:HostCollectionCopyController
     *
     * @description
     *   For copying a host collection.
     */
    function HostCollectionCopyController($scope, Notification, HostCollection) {
        $scope.copy = function (newName) {
            HostCollection.copy({id: $scope.$stateParams.hostCollectionId, 'host_collection': {name: newName}}, function (response) {
                $scope.transitionTo('host-collection.info', {hostCollectionId: response.id});
            }, function (response) {
                Notification.setErrorMessage(response.data.displayMessage);
            });
        };
    }

    angular
        .module('Bastion.host-collections')
        .controller('HostCollectionCopyController', HostCollectionCopyController);

    HostCollectionCopyController.$inject = ['$scope', 'Notification', 'HostCollection'];

})();
/**
 * @ngdoc object
 * @name  Bastion.host-collections.controller:HostCollectionDetailsController
 *
 * @requires $scope
 * @requires $state
 * @requires $q
 * @requires translate
 * @requires Host
 * @requires ContentHostsModalHelper
 * @requires HostCollection
 * @requires Notification
 * @requires ApiErrorHandler
 *
 * @description
 *   Provides the functionality for the host collection details action pane.
 */

angular.module('Bastion.host-collections').controller('HostCollectionDetailsController',
    ['$scope', '$state', '$q', 'translate', 'Host', 'ContentHostsModalHelper', 'HostCollection', 'Notification', 'ApiErrorHandler',
    function ($scope, $state, $q, translate, Host, ContentHostsModalHelper, HostCollection, Notification, ApiErrorHandler) {
        $scope.panel = {
            error: false,
            loading: true
        };

        if ($scope.hostCollection) {
            $scope.panel.loading = false;
        }

        $scope.hostCollection = HostCollection.get({id: $scope.$stateParams.hostCollectionId}, function (hostCollection) {
            $scope.$broadcast('hostCollection.loaded', hostCollection);
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        $scope.getHostIds = function() {
            return {
                included: {
                    search: 'host_collection_id = %s'.replace('%s', $scope.$stateParams.hostCollectionId)
                }
            };
        };

        ContentHostsModalHelper.resolveFunc = $scope.getHostIds;

        $scope.refreshHostCollection = function () {
            $scope.hostCollection.$get().then(function (hostCollection) {
                $scope.$emit("updateContentHostCollection", hostCollection);
            });
        };

        $scope.openPackagesModal = function () {
            ContentHostsModalHelper.openPackagesModal();
        };

        $scope.openErrataModal = function () {
            ContentHostsModalHelper.openErrataModal();
        };

        $scope.openHostCollectionsModal = function () {
            ContentHostsModalHelper.openHostCollectionsModal();
        };


        $scope.openEnvironmentModal = function () {
            ContentHostsModalHelper.openEnvironmentModal();
        };


        $scope.openSubscriptionsModal = function () {
            ContentHostsModalHelper.openSubscriptionsModal();
        };

        $scope.save = function (hostCollection) {
            var deferred = $q.defer();

            hostCollection.$update(function (response) {
                deferred.resolve(response);
                Notification.setSuccessMessage(translate('Host Collection updated'));
                $scope.table.replaceRow(response);
            }, function (response) {
                deferred.reject(response);
                Notification.setErrorMessage(translate("An error occurred saving the Host Collection: ") + response.data.displayMessage);
            });
            return deferred.promise;
        };

        $scope.removeHostCollection = function (hostCollection) {
            hostCollection.$delete(function () {
                $scope.transitionTo('host-collections');
                Notification.setSuccessMessage(translate('Host Collection removed.'));
            }, function (response) {
                Notification.setErrorMessage(translate("An error occurred removing the Host Collection: ") + response.data.displayMessage);
            });
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.host-collections.controller:HostCollectionHostsController
 *
 * @requires $scope
 * @requires $location
 * @requires Notification
 * @requires translate
 * @requires Nutupane
 * @requires CurrentOrganization
 * @requires Host
 * @requires HostCollection
 *
 * @description
 *   Provides the functionality for the host collection details action pane.
 */

angular.module('Bastion.host-collections').controller('HostCollectionHostsController',
    ['$scope', '$location', 'Notification', 'translate', 'Nutupane', 'CurrentOrganization', 'Host', 'HostCollection',
    function ($scope, $location, Notification, translate, Nutupane, CurrentOrganization, Host, HostCollection) {
        var params, nutupaneParams;

        params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'page': 1,
            'sort_by': 'name',
            'sort_order': 'ASC',
            'paged': true
        };

        nutupaneParams = {
            'disableAutoLoad': true
        };

        $scope.contentNutupane = new Nutupane(Host, params, undefined, nutupaneParams);
        $scope.controllerName = 'hosts';
        $scope.contentNutupane.searchTransform = function (term) {
            var addition = "host_collection_id=" + $scope.$stateParams.hostCollectionId;
            if (term === "" || angular.isUndefined(term)) {
                return addition;
            }

            return term + " and " + addition;
        };

        $scope.table = $scope.contentNutupane.table;
        $scope.table.closeItem = function () {};
        $scope.isRemoving = false;
        $scope.contentNutupane.setSearchKey('contentHostSearch');
        $scope.contentNutupane.refresh();

        $scope.removeSelected = function () {
            var selected = _.map($scope.table.getSelected(), 'id');

            $scope.isRemoving = true;
            HostCollection.removeHosts({id: $scope.hostCollection.id, 'host_ids': selected}, function (data) {
                $scope.contentNutupane.table.selectAll(false);
                $scope.contentNutupane.refresh();

                angular.forEach(data.displayMessages.success, function (success) {
                    Notification.setSuccessMessage(success);
                });

                angular.forEach(data.displayMessages.error, function (error) {
                    Notification.setErrorMessage(error);
                });
                $scope.refreshHostCollection();

                $scope.isRemoving = false;
            }, function (response) {
                $scope.isRemoving = false;
                Notification.setErrorMessage(translate("An error occurred removing the content hosts.") + response.data.displayMessage);
            });
        };

    }]
);
/**
 * @ngdoc factory
 * @name  Bastion.host-collections.factory:HostCollection
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for host collections.
 */

angular.module('Bastion.host-collections').factory('HostCollection',
    ['BastionResource', function (BastionResource) {

        return BastionResource('katello/api/v2/host_collections/:id/:action', {id: '@id'}, {
            get: {method: 'GET', params: {fields: 'full'}},
            update: {method: 'PUT'},
            copy: {method: 'POST', params: {action: 'copy'}},
            removeHosts: {method: 'PUT', params: {action: 'remove_hosts'}},
            addHosts: {method: 'PUT', params: {action: 'add_hosts'}},
            autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}}
        });

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.host-collections.controller:HostCollectionsController
 *
 * @requires $scope
 * @requires $location
 * @requires translate
 * @requires Nutupane
 * @requires HostCollection
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality specific to host collections for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.host-collections').controller('HostCollectionsController',
    ['$scope', '$location', 'translate', 'Nutupane', 'HostCollection', 'CurrentOrganization',
    function ($scope, $location, translate, Nutupane, HostCollection, CurrentOrganization) {
        var params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'sort_by': 'name',
            'sort_order': 'ASC',
            'paged': true
        };

        var nutupane = new Nutupane(HostCollection, params);
        $scope.controllerName = 'katello_host_collections';
        nutupane.masterOnly = true;

        $scope.table = nutupane.table;
        $scope.removeRow = nutupane.removeRow;

        $scope.$on("updateContentHostCollection", function (event, hostCollectionRow) {
            $scope.table.replaceRow(hostCollectionRow);
        });

    }]
);
angular.module('Bastion.host-collections').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('host-collections', {
        url: '/host_collections',
        permission: 'view_host_collections',
        template: '<div ui-view></div>',
        views: {
            '@': {
                controller: 'HostCollectionsController',
                templateUrl: 'host-collections/views/host-collections.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{ 'Host Collections' | translate }}"
        }
    })
    .state('host-collections.new', {
        url: '/new',
        permission: 'create_host_collections',
        views: {
            '@': {
                controller: 'NewHostCollectionController',
                templateUrl: 'host-collections/new/views/new-host-collection.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{ 'New Host Collection' | translate }}"
        }
    });

    $stateProvider.state("host-collection", {
        abstract: true,
        url: '/host_collections/:hostCollectionId',
        permission: 'view_host_collections',
        controller: 'HostCollectionDetailsController',
        templateUrl: 'host-collections/details/views/host-collection-details.html'
    })
    .state('host-collection.info', {
        url: '',
        permission: 'view_host_collections',
        templateUrl: 'host-collections/details/views/host-collection-info.html',
        ncyBreadcrumb: {
            label: "{{ hostCollection.name }}",
            parent: 'host-collections'
        }
    })
    .state('host-collection.hosts', {
        abstract: true,
        templateUrl: 'host-collections/details/views/host-collection-hosts.html'
    })
    .state('host-collection.hosts.list', {
        url: '/hosts',
        permission: 'view_host_collections',
        controller: 'HostCollectionHostsController',
        templateUrl: 'host-collections/details/views/host-collection-hosts-list.html',
        ncyBreadcrumb: {
            label: "{{ 'List Hosts' | translate }}",
            parent: 'host-collection.info'
        }
    })
    .state('host-collection.hosts.add', {
        url: '/add-hosts',
        permission: 'edit_host_collections',
        controller: 'HostCollectionAddHostsController',
        templateUrl: 'host-collections/details/views/host-collection-add-hosts.html',
        ncyBreadcrumb: {
            label: "{{ 'List Hosts' | translate }}",
            parent: 'host-collection.info'
        }
    })
    .state('host-collection.copy', {
        url: '/copy',
        permission: 'create_host_collections',
        controller: 'HostCollectionCopyController',
        templateUrl: 'host-collections/details/views/host-collection-copy.html',
        ncyBreadcrumb: {
            label: "{{ 'Create Copy' | translate }}",
            parent: 'host-collection.info'
        }
    });
}]);
/**
 * @ngdoc object
 * @name  Bastion.host-collections.controller:HostCollectionFormController
 *
 * @requires $scope
 * @requires $q
 * @requires HostCollection
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality specific to HostCollections for creating a new host collection
 */

angular.module('Bastion.host-collections').controller('HostCollectionFormController',
    ['$scope', '$q', 'HostCollection', 'CurrentOrganization',
    function ($scope, $q, HostCollection, CurrentOrganization) {

        function success(response) {
            $scope.table.addRow(response);
            $scope.transitionTo('host-collection.info', {hostCollectionId: $scope.hostCollection.id});
        }

        function error(response) {
            $scope.working = false;
            angular.forEach(response.data.errors, function (errors, field) {
                $scope.hostCollectionForm[field].$setValidity('', false);
                $scope.hostCollectionForm[field].$error.messages = errors;
            });
        }

        $scope.hostCollection = $scope.hostCollection || new HostCollection();
        $scope.hostCollection['unlimited_hosts'] = true;

        $scope.save = function (hostCollection) {
            hostCollection['organization_id'] = CurrentOrganization;
            hostCollection.$save(success, error);
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.host-collections.controller:NewHostCollectionController
 *
 * @requires $scope
 * @requires HostCollection
 * @requires CurrentOrganization
 *
 * @description
 *   Controls the creation of an empty HostCollection object for use by sub-controllers.
 */

angular.module('Bastion.host-collections').controller('NewHostCollectionController',
    ['$scope', 'HostCollection', 'CurrentOrganization',
    function ($scope, HostCollection, CurrentOrganization) {

        $scope.hostCollection = new HostCollection();
        $scope.panel = {loading: false};

        function success() {
            $scope.transitionTo('host-collection.info', {hostCollectionId: $scope.hostCollection.id});
        }

        function error(response) {
            $scope.working = false;
            angular.forEach(response.data.errors, function (errors, field) {
                $scope.hostCollectionForm[field].$setValidity('', false);
                $scope.hostCollectionForm[field].$error.messages = errors;
            });
        }

        $scope.hostCollection = $scope.hostCollection || new HostCollection();
        $scope.hostCollection['unlimited_hosts'] = true;

        $scope.save = function (hostCollection) {
            hostCollection['organization_id'] = CurrentOrganization;
            hostCollection.$save(success, error);
        };
    }]
);
/**
 * @ngdoc module
 * @name  Bastion.common
 *
 * @description
 *   Module for common functionality.
 */

angular.module('Bastion.common', [
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.components'
]);
(function () {
    'use strict';

    /**
     * @ngdoc service
     * @name  Bastion.common.service:ApiErrorHandler
     *
     * @description
     *   Provides common functionality in handling Katello/Foreman API Errors.
     */
    function ApiErrorHandler(translate, Notification) {
        function handleError(response, $scope, defaultErrorMessage) {
            if (response.hasOwnProperty('data') && response.data.hasOwnProperty('errors')) {
                angular.forEach(response.data.errors, function (error) {
                    Notification.setErrorMessage(error);
                });
            } else {
                Notification.setErrorMessage(defaultErrorMessage);
            }

            if ($scope && $scope.hasOwnProperty('panel')) {
                $scope.panel.error = true;
            }
        }

        this.handleGETRequestErrors = function (response, $scope) {
            var defaultErrorMessage = translate('Something went wrong when retrieving the resource.');
            handleError(response, $scope, defaultErrorMessage);
        };

        this.handlePUTRequestErrors = function (response, $scope) {
            var defaultErrorMessage = translate('Something went wrong when saving the resource.');
            handleError(response, $scope, defaultErrorMessage);
        };
    }

    angular.module('Bastion.common').service('ApiErrorHandler', ApiErrorHandler);
    ApiErrorHandler.$inject = ['translate', 'Notification'];
})();
(function () {
    'use strict';

    /**
     * @ngdoc directive
     * @name Bastion.common.directive:selectActionDropdown
     *
     * @description
     *   Used to create a dropdown menu with a list of actions in it and the words "Select Action" on the button.
     *
     *   Necessary because of https://github.com/angular-ui/bootstrap/issues/5841
     *
     * @example
     *  <span select-action-dropdown>
     *    <ul>...</ul>
     *  </span>
     */
    function selectActionDropdown() {
        return {
            restrict: 'AE',
            transclude: true,
            scope: true,
            templateUrl: 'common/views/select-action-dropdown.html',
            compile: function (tElement, tAttribute, transclude) {
                return function (scope, element) {
                    scope.status = {
                        isOpen: false
                    };

                    scope.toggleDropdown = function (event) {
                        event.preventDefault();
                        event.stopPropagation();
                        scope.status.isOpen = !scope.status.isOpen;
                    };

                    transclude(scope, function(clone) {
                        element.find('.btn-group').append(clone);
                    });
                };
            }
        };
    }

    angular.module('Bastion.common').directive('selectActionDropdown', selectActionDropdown);
})();
/**
 * @ngdoc module
 * @name  Bastion.content-hosts
 *
 * @description
 *   Module for content hosts related functionality.
 */

angular.module('Bastion.content-hosts', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.common',
    'Bastion.components',
    'Bastion.components.formatters',
    'Bastion.organizations',
    'Bastion.subscriptions',
    'Bastion.capsules',
    'Bastion.hosts',
    'Bastion.errata',
    'Bastion.host-collections',
    'Bastion.repository-sets'
]);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostsBulkEnvironmentModalController
 *
 * @requires $scope
 * @requires $state
 * @requires $uibModalInstance
 * @requires HostBulkAction
 * @requires Organization
 * @requires CurrentOrganization
 * @requires ContentView
 * @requires Notification
 * @requires hostIds
 *
 * @description
 *   A controller for providing bulk action functionality for setting content view and environment
 */

angular.module('Bastion.content-hosts').controller('ContentHostsBulkEnvironmentModalController',
    ['$scope', '$state', '$uibModalInstance', 'HostBulkAction', 'Organization', 'CurrentOrganization', 'ContentView', 'Notification', 'hostIds',
    function ($scope, $state, $uibModalInstance, HostBulkAction, Organization, CurrentOrganization, ContentView, Notification, hostIds) {

        function actionParams() {
            var params = hostIds;
            params['organization_id'] = CurrentOrganization;
            params['environment_id'] = $scope.selected.environment.id;
            params['content_view_id'] = $scope.selected.contentView.id;
            return params;
        }

        $scope.selected = {
            environment: undefined,
            contentView: undefined
        };

        $scope.environments = Organization.readableEnvironments({id: CurrentOrganization});

        $scope.disableAssignButton = function (confirm) {
            return confirm || hostIds === 0 ||
                angular.isUndefined($scope.selected.environment) || angular.isUndefined($scope.selected.contentView);
        };

        $scope.$watch('selected.environment', function (environment) {
            if (environment) {
                $scope.fetchViews();
            }
        });

        $scope.fetchViews = function () {
            $scope.fetchingContentViews = true;

            ContentView.queryUnpaged({ 'environment_id': $scope.selected.environment.id }, function (response) {
                $scope.contentViews = response.results;
                $scope.fetchingContentViews = false;
            });
        };

        $scope.performAction = function () {
            HostBulkAction.environmentContentView(actionParams(), function (task) {
                $scope.ok();
                $state.go('content-hosts.bulk-task', {taskId: task.id});
            }, function (response) {
                angular.forEach(response.data.errors, function (error) {
                    Notification.setErrorMessage(error);
                });
            });
        };

        $scope.ok = function () {
            $uibModalInstance.close();
        };

        $scope.cancel = function () {
            $uibModalInstance.dismiss('cancel');
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostsBulkErrataModalController
 *
 * @requires $scope
 * @requires $http
 * @requires $location
 * @requires $window
 * @requires $uibModalInstance
 * @requires HostBulkAction
 * @requires HostCollection
 * @requires Nutupane
 * @requires CurrentOrganization
 * @requires Erratum
 * @requires Notification
 * @requires BastionConfig
 * @requires hostIds
 *
 * @description
 *   A controller for providing bulk action functionality to the content hosts page.
 */

angular.module('Bastion.content-hosts').controller('ContentHostsBulkErrataModalController',
    ['$scope', '$http', '$location', '$window', '$uibModalInstance', 'HostBulkAction', 'HostCollection', 'Nutupane', 'CurrentOrganization', 'Erratum', 'Notification', 'BastionConfig', 'hostIds',
    function ($scope, $http, $location, $window, $uibModalInstance, HostBulkAction, HostCollection, Nutupane, CurrentOrganization, Erratum, Notification, BastionConfig, hostIds) {
        var nutupane;

        function installParams() {
            var params = hostIds;
            params['content_type'] = 'errata';
            params.content = _.map($scope.table.getSelected(), 'errata_id');
            params['organization_id'] = CurrentOrganization;
            return params;
        }

        function fetchErratum(errataId) {
            $scope.erratum = Erratum.get({id: errataId, 'organization_id': CurrentOrganization});
        }

        nutupane = new Nutupane(HostBulkAction, hostIds, 'installableErrata');
        $scope.controllerName = 'katello_errata';
        nutupane.masterOnly = true;

        $scope.table = nutupane.table;
        $scope.table.errataFilterTerm = "";
        $scope.table.initialLoad = false;
        $scope.initialLoad = true;
        $scope.remoteExecutionPresent = BastionConfig.remoteExecutionPresent;
        $scope.remoteExecutionByDefault = BastionConfig.remoteExecutionByDefault;

        $scope.fetchErrata = function () {
            var params = hostIds;
            params['organization_id'] = CurrentOrganization;
            nutupane.setParams(params);
            $scope.table.working = true;
            if ($scope.table.numSelected > 0) {
                nutupane.refresh().then(function () {
                    $scope.table.working = false;
                });
            } else {
                $scope.table.working = false;
            }
        };

        $scope.transitionToErrata = function (erratum) {
            fetchErratum(erratum['errata_id']);
            $scope.transitionTo('content-hosts.bulk-actions.errata.details', {errataId: erratum['errata_id']});
        };

        $scope.transitionToErrataContentHosts = function (erratum) {
            $scope.erratum = erratum;
            $scope.transitionTo('content-hosts.bulk-actions.errata.content-hosts', {errataId: erratum['errata_id']});
        };

        $scope.installErrata = function () {
            if ($scope.remoteExecutionByDefault) {
                $scope.installErrataViaRemoteExecution();
            } else {
                $scope.installErrataViaKatelloAgent(false);
            }
        };

        $scope.installErrataViaKatelloAgent = function () {
            var params = installParams();
            HostBulkAction.installContent(params,
                function (data) {
                    nutupane.invalidate();
                    $scope.ok();
                    $scope.transitionTo('content-hosts.bulk-task', {taskId: data.id});
                },
                function (response) {
                    angular.forEach(response.data.errors, function (error) {
                        Notification.setErrorMessage(error);
                    });
                });
        };

        $scope.installErrataViaRemoteExecution = function(customize) {
            var formData = {},
                errataIds = _.map($scope.table.getSelected(), 'errata_id'),
                selectedHosts = hostIds;

            formData.authenticityToken = $window.AUTH_TOKEN.replace(/&quot;/g, '');
            formData.remoteAction = 'errata_install';
            formData.errata = errataIds.join(',');
            if (selectedHosts.included.ids) {
                formData.hostIds = selectedHosts.included.ids.join(',');
            }
            formData.search = selectedHosts.included.search;
            formData.customize = customize;

            $http.post('katello/remote_execution', formData, {
                headers: {'Content-Type': 'application/x-www-form-urlencoded'
            }});
        };

        $scope.ok = function () {
            $uibModalInstance.close();
        };

        $scope.cancel = function () {
            $uibModalInstance.dismiss('cancel');
        };

        $scope.fetchErrata();
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostsBulkHostCollectionsModalController
 *
 * @requires $scope
 * @requires $location
 * @requires $uibModalInstance
 * @requires translate
 * @requires Nutupane
 * @requires HostBulkAction
 * @requires HostCollection
 * @requires CurrentOrganization
 * @requires Notification
 * @requires hostIds
 *
 * @description
 *   A controller for providing bulk action functionality to the content hosts page.
 */

angular.module('Bastion.content-hosts').controller('ContentHostsBulkHostCollectionsModalController',
    ['$scope', '$location', '$uibModalInstance', 'translate', 'Nutupane', 'HostBulkAction', 'HostCollection', 'CurrentOrganization', 'Notification', 'hostIds',
    function ($scope, $location, $uibModalInstance, translate, Nutupane, HostBulkAction, HostCollection, CurrentOrganization, Notification, hostIds) {
        var nutupane, nutupaneParams;

        $scope.hostCollections = {
            action: null
        };

        nutupaneParams = {
            'organization_id': CurrentOrganization,
            'offset': 0,
            'sort_by': 'name',
            'sort_order': 'ASC',
            'paged': true
        };

        nutupane = new Nutupane(HostCollection, nutupaneParams, 'queryPaged');
        $scope.controllerName = 'katello_host_collections';
        nutupane.masterOnly = true;

        $scope.table = nutupane.table;

        $scope.confirmHostCollectionAction = function (action) {
            $scope.hostCollections.confirm = true;
            $scope.hostCollections.action = action;
        };

        $scope.performHostCollectionAction = function () {
            var params, action, success, error;

            action = $scope.hostCollections.action;
            params = hostIds;
            params['organization_id'] = CurrentOrganization;
            params['host_collection_ids'] = nutupane.getAllSelectedResults('id').included.ids;

            $scope.hostCollections.action = null;

            success = function (data) {
                angular.forEach(data.displayMessages, function (message) {
                    Notification.setSuccessMessage(message);
                });
                nutupane.invalidate();
            };

            error = function (response) {
                angular.forEach(response.data.errors, function (responseError) {
                    Notification.setErrorMessage(responseError);
                });
                $scope.editMode = true;
            };

            if (action === 'add') {
                HostBulkAction.addHostCollections(params, success, error);
            } else if (action === 'remove') {
                HostBulkAction.removeHostCollections(params, success, error);
            }
        };

        $scope.ok = function () {
            $uibModalInstance.close();
        };

        $scope.cancel = function () {
            $uibModalInstance.dismiss('cancel');
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostsBulkPackagesModalController
 *
 * @requires $scope
 * @resource $location
 * @resource $timeout
 * @resource $window
 * @requires $uibModalInstance
 * @requires HostBulkAction
 * @requires CurrentOrganization
 * @requires translate
 * @requires Notification
 * @requires BastionConfig
 * @requires hostIds
 *
 * @description
 *   A controller for providing bulk action functionality to the content hosts page.
 */

angular.module('Bastion.content-hosts').controller('ContentHostsBulkPackagesModalController',
    ['$scope', '$location', '$timeout', '$window', '$uibModalInstance', 'HostBulkAction', 'CurrentOrganization', 'translate', 'Notification', 'BastionConfig', 'hostIds',
    function ($scope, $location, $timeout, $window, $uibModalInstance, HostBulkAction, CurrentOrganization, translate, Notification, BastionConfig, hostIds) {

        function successMessage(type) {
            var messages = {
                install: translate("Successfully scheduled package installation"),
                update: translate("Successfully scheduled package update"),
                remove: translate("Successfully scheduled package removal"),
                "update all": translate("Successfully scheduled an update of all packages")
            };
            return messages[type];
        }

        function installParams() {
            var params = hostIds;
            params['content_type'] = $scope.content.contentType;
            if ($scope.content.action === "update all") {
                params['update_all'] = true;
                params.content = null;
            } else {
                params.content = $scope.content.content.split(/ *, */);
            }
            params['organization_id'] = CurrentOrganization;
            return params;
        }

        $scope.remoteExecutionPresent = BastionConfig.remoteExecutionPresent;
        $scope.remoteExecutionByDefault = BastionConfig.remoteExecutionByDefault;

        $scope.packageActionFormValues = {
            authenticityToken: $window.AUTH_TOKEN.replace(/&quot;/g, '')
        };

        $scope.content = {
            confirm: false,
            placeholder: translate('Enter Package Name(s)...'),
            contentType: 'package'
        };

        $scope.updatePlaceholder = function (contentType) {
            if (contentType === "package") {
                $scope.content.placeholder = translate('Enter Package Name(s)...');
            } else if (contentType === "package_group") {
                $scope.content.placeholder = translate('Enter Package Group Name(s)...');
            }
        };

        $scope.confirmContentAction = function (action, actionInput) {
            $scope.content.confirm = true;
            $scope.content.action = action;
            $scope.content.actionInput = actionInput;
        };

        $scope.performContentAction = function () {
            if ($scope.remoteExecutionByDefault) {
                $scope.performViaRemoteExecution();
            } else {
                $scope.performViaKatelloAgent();
            }
        };

        $scope.performViaKatelloAgent = function (action, actionInput) {
            var success, error, params;

            if (action) {
                $scope.content.action = action;
            }

            if (actionInput) {
                $scope.content.actionInput = actionInput;
            }

            $scope.content.confirm = false;

            success = function () {
                Notification.setSuccessMessage(successMessage($scope.content.action));
            };

            error = function (response) {
                angular.forEach(response.data.errors, function (responseError) {
                    Notification.setErrorMessage(responseError);
                });
            };

            params = installParams();
            if ($scope.content.action === "install") {
                HostBulkAction.installContent(params, success, error);
            } else if ($scope.content.action === "update") {
                HostBulkAction.updateContent(params, success, error);
            } else if ($scope.content.action === "remove") {
                HostBulkAction.removeContent(params, success, error);
            } else if ($scope.content.action === "update all") {
                HostBulkAction.updateContent(params, success, error);
            }
        };

        $scope.performViaRemoteExecution = function(action, customize) {
            var selectedHosts = hostIds;

            $scope.content.confirm = false;
            $scope.packageActionFormValues.customize = customize;

            if (!action) {
                action = $scope.content.action;
            }

            if (action === "update all") {
                action = "update";
            }

            if ($scope.content.contentType === 'package_group') {
                $scope.packageActionFormValues.remoteAction = 'group_' + action;
            } else if ($scope.content.contentType === 'package') {
                $scope.packageActionFormValues.remoteAction = 'package_' + action;
            }

            $scope.packageActionFormValues.hostIds = selectedHosts.included.ids.join(',');
            $scope.packageActionFormValues.search = selectedHosts.included.search;

            $timeout(function () {
                angular.element('#packageActionForm').submit();
            }, 0);
        };

        $scope.ok = function () {
            $uibModalInstance.close();
        };

        $scope.cancel = function () {
            $uibModalInstance.dismiss('cancel');
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostsBulkReleaseVersionModalController
 *
 * @requires $scope
 * @requires $state
 * @requires $uibModalInstance
 * @requires HostBulkAction
 * @requires Organization
 * @requires CurrentOrganization
 * @requires Notification
 * @requires hostIds
 *
 * @description
 *   A controller for providing bulk action functionality for setting content view and environment
 */

angular.module('Bastion.content-hosts').controller('ContentHostsBulkReleaseVersionModalController',
    ['$scope', '$state', '$uibModalInstance', 'HostBulkAction', 'Organization', 'CurrentOrganization', 'Notification', 'hostIds',
    function ($scope, $state, $uibModalInstance, HostBulkAction, Organization, CurrentOrganization, Notification, hostIds) {

        function actionParams() {
            var params = hostIds;
            params['organization_id'] = CurrentOrganization;
            params['release_version'] = $scope.selected.release;

            return params;
        }

        $scope.selected = {
            release: undefined
        };
        $scope.fetchingReleases = true;

        Organization.releaseVersions({id: CurrentOrganization}, function (response) {
            $scope.releases = response.results;
            $scope.fetchingReleases = false;
        });

        $scope.disableAssignButton = function (confirm) {
            return confirm || hostIds === 0 || angular.isUndefined($scope.selected.release);
        };

        $scope.performAction = function () {
            HostBulkAction.releaseVersion(actionParams(), function (task) {
                $scope.ok();
                $state.go('content-hosts.bulk-task', {taskId: task.id});
            }, function (response) {
                angular.forEach(response.data.errors, function (error) {
                    Notification.setErrorMessage(error);
                });
            });
        };

        $scope.ok = function () {
            $uibModalInstance.close();
        };

        $scope.cancel = function () {
            $uibModalInstance.dismiss('cancel');
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostsBulkRepositorySetsModalController
 *
 * @requires $scope
 * @requires $location
 * @requires $uibModalInstance
 * @requires translate
 * @requires Nutupane
 * @requires HostBulkAction
 * @requires RepositorySet
 * @requires CurrentOrganization
 * @requires Notification
 * @requires hostIds
 *
 * @description
 *   A controller for providing bulk action functionality to the content hosts page.
 */

angular.module('Bastion.content-hosts').controller('ContentHostsBulkRepositorySetsModalController',
    ['$scope', '$location', '$uibModalInstance', 'translate', 'Nutupane', 'HostBulkAction', 'RepositorySet', 'CurrentOrganization', 'Notification', 'hostIds',
    function ($scope, $location, $uibModalInstance, translate, Nutupane, HostBulkAction, RepositorySet, CurrentOrganization, Notification, hostIds) {
        var nutupane, nutupaneParams;

        $scope.repositorySets = {
            action: null
        };

        nutupaneParams = {
            'organization_id': CurrentOrganization,
            'offset': 0,
            'sort_by': 'name',
            'sort_order': 'ASC',
            'paged': true
        };

        nutupane = new Nutupane(RepositorySet, nutupaneParams, 'queryPaged');
        $scope.controllerName = 'katello_repository_sets';
        nutupane.masterOnly = true;

        $scope.table = nutupane.table;

        $scope.confirmRepositorySetAction = function (action) {
            $scope.repositorySets.confirm = true;
            $scope.repositorySets.action = action;
        };

        $scope.performRepositorySetAction = function () {
            var params, action, success, error, contentOverrides;

            action = $scope.repositorySets.action;
            params = hostIds;
            params['organization_id'] = CurrentOrganization;
            contentOverrides = [];
            angular.forEach(nutupane.getAllSelectedResults('id').included.resources, function (repositorySet) {
                var value, remove;

                if (action === 'enable') {
                    value = true;
                    remove = false;
                } else if (action === 'disable') {
                    value = false;
                    remove = false;
                } else if (action === 'reset') {
                    value = true;
                    remove = true;
                }
                contentOverrides.push({
                    'content_label': repositorySet.label,
                    name: 'enabled',
                    value: value,
                    remove: remove
                });
            });
            params['content_overrides'] = contentOverrides;

            $scope.repositorySets.action = null;

            success = function (response) {
                nutupane.invalidate();
                $scope.ok();
                $scope.transitionTo('content-hosts.bulk-task', {taskId: response.id});
            };

            error = function (response) {
                angular.forEach(response.data.errors, function (responseError) {
                    Notification.setErrorMessage(responseError);
                });
                $scope.editMode = true;
            };

            HostBulkAction.updateRepositorySets(params, success, error);
        };

        $scope.ok = function () {
            $uibModalInstance.close();
        };

        $scope.cancel = function () {
            $uibModalInstance.dismiss('cancel');
        };

        $scope.repositorySetsAction = function (action) {
            $scope.repositorySets.action = action;
            $scope.repositorySets.working = true;
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostsBulkSubscriptionsModalController
 *
 * @requires $scope
 * @requires $location
 * @requires $uibModalInstance
 * @requires CurrentOrganization
 * @requires HostBulkAction
 * @requires Subscription
 * @requires ContentHost
 * @requires SubscriptionsHelper
 * @requires Notification
 * @requires hostIds
 *
 * @description
 *   A controller for providing bulk action functionality to the content hosts page.
 */

angular.module('Bastion.content-hosts').controller('ContentHostsBulkSubscriptionsModalController',
    ['$scope', '$location', '$uibModalInstance', 'Nutupane', 'CurrentOrganization', 'HostBulkAction', 'Subscription', 'SubscriptionsHelper', 'Notification', 'hostIds',
        function ($scope, $location, $uibModalInstance, Nutupane, CurrentOrganization, HostBulkAction, Subscription, SubscriptionsHelper, Notification, hostIds) {
            var success, error, params = {
                'organization_id': CurrentOrganization,
                'sort_order': 'ASC',
                'available_for': 'host',
                'host_id': $scope.$stateParams.hostId
            };

            function getBulkSubscriptionParams() {
                var bulkParams = hostIds;
                bulkParams.subscriptions = SubscriptionsHelper.getSelectedSubscriptionAmounts($scope.table);
                return bulkParams;
            }

            success = function (response) {
                $scope.contentNutupane.invalidate();
                $scope.ok();
                $scope.transitionTo('content-hosts.bulk-task', {taskId: response.id});
            };

            error = function (response) {
                angular.forEach(response.data.errors, function (responseError) {
                    Notification.setErrorMessage(responseError);
                });
            };

            $scope.contentNutupane = new Nutupane(Subscription, params);
            $scope.controllerName = 'katello_subscriptions';
            $scope.table = $scope.contentNutupane.table;
            $scope.contentNutupane.setSearchKey('subscriptionSearch');
            $scope.contentNutupane.masterOnly = true;
            $scope.groupedSubscriptions = {};

            $scope.$watch('table.rows', function (rows) {
                $scope.groupedSubscriptions = SubscriptionsHelper.groupByProductName(rows);
            });

            $scope.getAmountSelectorValues = SubscriptionsHelper.getAmountSelectorValues;

            $scope.addSelected = function () {
                var bulkParams = getBulkSubscriptionParams();
                HostBulkAction.addSubscriptions(bulkParams, success, error);
            };

            $scope.removeSelected = function () {
                var bulkParams = getBulkSubscriptionParams();
                HostBulkAction.removeSubscriptions(bulkParams, success, error);
            };

            $scope.autoAttach = function () {
                var bulkParams = getBulkSubscriptionParams();
                HostBulkAction.autoAttach(bulkParams, success, error);
            };

            $scope.ok = function () {
                $uibModalInstance.close();
            };

            $scope.cancel = function () {
                $uibModalInstance.dismiss('cancel');
            };
        }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostEventsController
 *
 * @requires $scope
 * @requires translate
 * @requires ContentHost
 * @requires Nutupane
 */

angular.module('Bastion.content-hosts').controller('ContentHostEventsController',
    ['$scope', 'translate', 'HostSubscription', 'Nutupane',
    function ($scope, translate, HostSubscription, Nutupane) {
        var params = {id: $scope.$stateParams.hostId},
            nutupane = new Nutupane(HostSubscription, params, 'events');
        $scope.eventTable = nutupane.table;
    }]
);
/**
 * @ngdoc service
 * @name  Bastion.content-hosts.service:ContentHostsModalHelper
 *
 * @requires $uibModal
 *
 * @description
 *  Helper service that contains functionality common amongst content hosts action modals.
*/


angular.module('Bastion.content-hosts').service('ContentHostsModalHelper', ['$uibModal',
    function ($uibModal) {

        this.resolveFunc = null;

        this.openHostCollectionsModal = function() {
        $uibModal.open({
            templateUrl: 'content-hosts/bulk/views/content-hosts-bulk-host-collections-modal.html',
            controller: 'ContentHostsBulkHostCollectionsModalController',
            size: 'lg',
            resolve: {
                hostIds: this.resolveFunc()
            }
        });
    };

        this.openPackagesModal = function() {
            $uibModal.open({
            templateUrl: 'content-hosts/bulk/views/content-hosts-bulk-packages-modal.html',
            controller: 'ContentHostsBulkPackagesModalController',
            size: 'lg',
            resolve: {
                hostIds: this.resolveFunc()
            }
        });
        };

        this.openErrataModal = function() {
            $uibModal.open({
                templateUrl: 'content-hosts/bulk/views/content-hosts-bulk-errata-modal.html',
                controller: 'ContentHostsBulkErrataModalController',
                size: 'lg',
                resolve: {
                    hostIds: this.resolveFunc()
                }
            });
        };

        this.openEnvironmentModal = function() {
            $uibModal.open({
                templateUrl: 'content-hosts/bulk/views/content-hosts-bulk-environment-modal.html',
                controller: 'ContentHostsBulkEnvironmentModalController',
                size: 'lg',
                resolve: {
                    hostIds: this.resolveFunc()
                }
            });
        };

        this.openReleaseVersionModal = function() {
            $uibModal.open({
                templateUrl: 'content-hosts/bulk/views/content-hosts-bulk-release-version-modal.html',
                controller: 'ContentHostsBulkReleaseVersionModalController',
                size: 'lg',
                resolve: {
                    hostIds: this.resolveFunc()
                }
            });
        };

        this.openSubscriptionsModal = function() {
            $uibModal.open({
                templateUrl: 'content-hosts/bulk/views/content-hosts-bulk-subscriptions-modal.html',
                controller: 'ContentHostsBulkSubscriptionsModalController',
                size: 'lg',
                resolve: {
                    hostIds: this.resolveFunc()
                }
            });
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostRegisterController
 *
 * @requires $scope
 * @requires $location
 * @requires Capsule
 * @requires Organization
 * @requires CurrentOrganization
 * @requires BastionConfig
 *
 * @description
 *     Provides values to populate the code commands for registering a content host.
 */

angular.module('Bastion.content-hosts').controller('ContentHostRegisterController',
    ['$scope', '$location', 'Capsule', 'Organization', 'CurrentOrganization', 'BastionConfig',
    function ($scope, $location, Capsule, Organization, CurrentOrganization, BastionConfig) {

        $scope.organization = Organization.get({id: CurrentOrganization});
        $scope.consumerCertRPM = BastionConfig.consumerCertRPM;
        $scope.katelloHostname = $location.host();
        $scope.noCapsulesFound = true;

        $scope.capsules = Capsule.queryUnpaged(function (data) {
            var defaultCapsule = _.filter(data.results, function (result) {
                var featureNames = _.map(result.features, 'name');
                return _.includes(featureNames, 'Pulp');
            });

            $scope.noCapsulesFound = _.isEmpty(data.results);
            $scope.selectedCapsule = _.isEmpty(defaultCapsule) ? data.results[0] : defaultCapsule[0];
        });


        $scope.hostname = function (url) {
            if (url) {
                url = url.split('://')[1].split(':')[0];
            }

            return url;
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostStatusController
 *
 * @requires $scope
 * @requires translate
 * @requires ContentHost
 */

angular.module('Bastion.content-hosts').controller('ContentHostStatusController',
    ['$scope', 'translate',
    function ($scope, translate) {

        $scope.statusReason = translate("Loading...");
        $scope.statusRetrieved = false;

    }]
);
/**
 * @ngdoc service
 * @name  Bastion.content-hosts.service:ContentHostsHelper
 *
 * @description
 *   Helper service that contains functionality common amongst content hosts.
 */

angular.module('Bastion.content-hosts').service('ContentHostsHelper',
    function () {
        this.convertMemToGB = function (memoryValue) {
            if (angular.isString(memoryValue)) {
                memoryValue = memoryValue.toLowerCase();
                if (_.includes(memoryValue, "gb")) {
                    memoryValue = memoryValue.replace("gb", "").trim();
                    return memoryValue;
                }
                memoryValue = parseInt(memoryValue);
            }
            memoryValue = (memoryValue / 1048576).toFixed(2);
            return memoryValue;
        };

        this.getHostStatusIcon = function (globalStatus) {
            var icons;
            var colors = {
                // we can remove 'valid', 'partial', and 'invalid' when http://projects.theforeman.org/issues/15347 is fixed
                'valid': 'green',
                'partial': 'yellow',
                'invalid': 'red',
                0: 'green',
                1: 'yellow',
                2: 'red'
            };

            globalStatus = colors[globalStatus] || "red";
            icons = {
                'green': globalStatus + ' host-status pficon pficon-ok status-ok',
                'yellow': globalStatus + ' host-status pficon pficon-info status-warn',
                'red': globalStatus + ' host-status pficon pficon-error-circle-o status-error'
            };

            return icons[globalStatus];
        };
    }
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostsController
 *
 * @requires $scope
 * @requires $q
 * @requires $location
 * @requires $uibModal
 * @requires translate
 * @requires Nutupane
 * @requires Host
 * @requires HostBulkAction
 * @requires Notification
 * @requires CurrentOrganization
 * @requires ContentHostsHelper
 *
 * @description
 *   Provides the functionality specific to Content Hosts for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.content-hosts').controller('ContentHostsController',
    ['$scope', '$q', '$state', '$location', '$uibModal', 'translate', 'Nutupane', 'Host', 'HostBulkAction', 'Notification', 'CurrentOrganization', 'ContentHostsHelper', 'ContentHostsModalHelper', '$httpParamSerializer',
    function ($scope, $q, $state, $location, $uibModal, translate, Nutupane, Host, HostBulkAction, Notification, CurrentOrganization, ContentHostsHelper, ContentHostsModalHelper, $httpParamSerializer) {
        var nutupane, params, query;

        if ($location.search().search) {
            query = '"' + $location.search().search + '"';
        } else {
            query = "";
        }

        params = {
            'organization_id': CurrentOrganization,
            'search': query,
            'sort_by': 'name',
            'sort_order': 'ASC'
        };

        $scope.csvQuery = $httpParamSerializer(params);
        nutupane = new Nutupane(Host, params);
        $scope.controllerName = 'hosts';
        nutupane.masterOnly = true;

        $scope.table = nutupane.table;
        $scope.nutupane = nutupane;

        // @TODO begin hack necessary because of foreman API bug http://projects.theforeman.org/issues/13877
        $scope.table.sortBy = function (column) {
            var sort = $scope.table.resource.sort,
                sortOrder;
            if (!column) {
                return;
            }

            params.sort = column.id;
            if (column.id === sort.by) {
                sortOrder = (sort.order === 'ASC') ? 'DESC' : 'ASC';
            } else {
                sortOrder = 'ASC';
            }

            params.order = [column.id, sortOrder].join(' ');

            column.sortOrder = sortOrder;
            column.active = true;
            $scope.table.rows = [];
            $scope.nutupane.query();
        };
        // @TODO end hack

        nutupane.enableSelectAllResults();

        if ($location.search()['select_all']) {
            nutupane.table.initialSelectAll = true;
        }

        $scope.table.getHostStatusIcon = ContentHostsHelper.getHostStatusIcon;

        $scope.table = $scope.table;

        $scope.performDestroyHosts = function () {
            var destroyParams, success, error, deferred = $q.defer();

            destroyParams = $scope.nutupane.getAllSelectedResults();
            destroyParams['organization_id'] = CurrentOrganization;

            success = function (data) {
                deferred.resolve(data);
                $location.path("/foreman_tasks/tasks/" + data.id);
            };

            error = function (response) {
                deferred.reject(response.data.errors);
                angular.forEach(response.data.errors, function (responseError) {
                    Notification.setErrorMessage(responseError);
                });
            };

            HostBulkAction.destroyHosts(destroyParams, success, error);
            return deferred.promise;
        };
        $scope.getHostIds = function() {
            return $scope.nutupane.getAllSelectedResults('id');
        };

        ContentHostsModalHelper.resolveFunc = $scope.getHostIds;

        $scope.openHostCollectionsModal = function() {
            nutupane.invalidate();
            ContentHostsModalHelper.openHostCollectionsModal();
        };

        $scope.openRepositorySetsModal = function () {
            nutupane.invalidate();
            $uibModal.open({
                templateUrl: 'content-hosts/bulk/views/content-hosts-bulk-repository-sets-modal.html',
                controller: 'ContentHostsBulkRepositorySetsModalController',
                size: 'lg',
                resolve: {
                    hostIds: function () {
                        return $scope.nutupane.getAllSelectedResults('id');
                    }
                }
            });
        };

        $scope.openPackagesModal = function () {
            nutupane.invalidate();
            ContentHostsModalHelper.openPackagesModal();
        };

        $scope.openErrataModal = function () {
            nutupane.invalidate();
            ContentHostsModalHelper.openErrataModal();
        };

        $scope.openEnvironmentModal = function () {
            nutupane.invalidate();
            ContentHostsModalHelper.openEnvironmentModal();
        };

        $scope.openSetReleaseVersionModal = function () {
            nutupane.invalidate();
            ContentHostsModalHelper.openReleaseVersionModal();
        };

        $scope.openSubscriptionsModal = function () {
            nutupane.invalidate();
            ContentHostsModalHelper.openSubscriptionsModal();
        };
    }]
);
/**
 * @ngdoc object
 * @name Bastion.content-hosts.config
 *
 * @requires $stateProvider
 *
 * @description
 *   Used for content hosts level configuration such as setting up the ui state machine.
 */


angular.module('Bastion.content-hosts').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('content-hosts', {
        url: '/content_hosts',
        permission: 'view_hosts',
        views: {
            '@': {
                controller: 'ContentHostsController',
                templateUrl: 'content-hosts/views/content-hosts.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{ 'Content Hosts' | translate }}"
        }
    });

    $stateProvider.state('content-hosts.register', {
        url: '/register',
        permission: 'view_hosts',
        views: {
            '@': {
                controller: 'ContentHostRegisterController',
                templateUrl: 'content-hosts/views/register.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{ 'Register a Host' | translate }}"
        }
    });

    $stateProvider.state('content-hosts.bulk-task', {
        url: '/bulk-tasks/:taskId',
        permission: ['view_hosts'],
        views: {
            '@': {
                controller: 'TaskDetailsController',
                templateUrl: 'tasks/views/task-details.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{ 'Bulk Task' | translate }}",
            parent: 'content-hosts'
        }
    });

    $stateProvider.state("content-host", {
        abstract: true,
        url: '/content_hosts/:hostId',
        permission: 'view_hosts',
        controller: 'ContentHostDetailsController',
        templateUrl: 'content-hosts/details/views/content-host-details.html'
    })
    .state('content-host.info', {
        url: '',
        permission: 'view_hosts',
        controller: 'ContentHostDetailsInfoController',
        templateUrl: 'content-hosts/details/views/content-host-info.html',
        ncyBreadcrumb: {
            label: "{{ host.name }}",
            parent: 'content-hosts'
        }
    })
    .state('content-host.provisioning', {
        url: '/provisioning',
        permission: 'view_hosts',
        templateUrl: 'content-hosts/details/views/content-host-provisioning-info.html',
        ncyBreadcrumb: {
            label: "{{ 'Provisioning' | translate }}",
            parent: 'content-host.info'
        }
    })
    .state('content-host.products', {
        url: '/products',
        permission: 'view_products',
        controller: 'ContentHostRepositorySetsController',
        templateUrl: 'content-hosts/details/views/content-host-repository-sets.html',
        ncyBreadcrumb: {
            label: "{{ 'Products' | translate }}",
            parent: 'content-host.info'
        }
    });

    $stateProvider.state('content-host.events', {
        url: '/events',
        permission: 'view_hosts',
        controller: 'ContentHostEventsController',
        templateUrl: 'content-hosts/details/views/content-host-events.html',
        ncyBreadcrumb: {
            label: "{{ 'Events' | translate }}",
            parent: 'content-host.info'
        }
    });

    $stateProvider.state('content-host.tasks', {
        abstract: true,
        template: '<div ui-view></div>'
    })
    .state('content-host.tasks.index', {
        url: '/tasks',
        permission: 'view_hosts',
        templateUrl: 'content-hosts/details/views/content-host-tasks.html',
        ncyBreadcrumb: {
            label: "{{ 'Tasks' | translate }}",
            parent: 'content-host.info'
        }
    })
    .state('content-host.tasks.details', {
        url: '/tasks/:taskId',
        permission: 'view_hosts',
        controller: 'TaskDetailsController',
        templateUrl: 'tasks/views/task-details.html',
        ncyBreadcrumb: {
            label: "{{ task.id }}",
            parent: 'content-host.info'
        }
    });

    $stateProvider.state('content-host.subscriptions', {
        abstract: true,
        controller: 'ContentHostBaseSubscriptionsController',
        templateUrl: 'content-hosts/details/views/content-host-subscriptions.html'
    })
    .state('content-host.subscriptions.list', {
        url: '/subscriptions',
        permission: 'view_hosts',
        controller: 'ContentHostSubscriptionsController',
        templateUrl: 'content-hosts/details/views/content-host-subscriptions-list.html',
        ncyBreadcrumb: {
            label: "{{ 'Subscriptions' | translate }}",
            parent: 'content-host.info'
        }
    })
    .state('content-host.subscriptions.add', {
        url: '/add-subscriptions',
        permission: 'attach_subscriptions',
        controller: 'ContentHostAddSubscriptionsController',
        templateUrl: 'content-hosts/details/views/content-host-add-subscriptions.html',
        ncyBreadcrumb: {
            label: "{{ 'Add Subscriptipons' | translate }}",
            parent: 'content-host.info'
        }
    });

    $stateProvider.state('content-host.host-collections', {
        abstract: true,
        templateUrl: 'content-hosts/details/views/content-host-host-collections.html'
    })
    .state('content-host.host-collections.list', {
        url: '/host-collections',
        permission: 'view_hosts',
        controller: 'ContentHostHostCollectionsController',
        templateUrl: 'content-hosts/details/views/content-host-host-collections-table.html',
        ncyBreadcrumb: {
            label: "{{ 'Host Collections' | translate }}",
            parent: 'content-host.info'
        }
    })
    .state('content-host.host-collections.add', {
        url: '/host-collections/add',
        permission: 'edit_hosts',
        controller: 'ContentHostAddHostCollectionsController',
        templateUrl: 'content-hosts/details/views/content-host-host-collections-table.html',
        ncyBreadcrumb: {
            label: "{{ 'Add Host Collections' | translate }}",
            parent: 'content-host.info'
        }
    });

    $stateProvider.state('content-host.packages', {
        controller: 'ContentHostPackagesController',
        abstract: true,
        templateUrl: 'content-hosts/content/views/content-host-packages.html'
    })
    .state('content-host.packages.actions', {
        url: '/packages/actions',
        permission: 'edit_hosts',
        controller: 'ContentHostPackagesActionsController',
        templateUrl: 'content-hosts/content/views/content-host-packages-actions.html',
        ncyBreadcrumb: {
            label: "{{ 'Package Actions' | translate }}",
            parent: 'content-host.info'
        }
    })
    .state('content-host.packages.installed', {
        url: '/packages/installed',
        permission: 'view_hosts',
        controller: 'ContentHostPackagesInstalledController',
        templateUrl: 'content-hosts/content/views/content-host-packages-installed.html',
        ncyBreadcrumb: {
            label: "{{ 'Installed Packages' | translate }}",
            parent: 'content-host.info'
        }
    })
    .state('content-host.packages.applicable', {
        url: '/packages/applicable',
        permission: 'view_hosts',
        controller: 'ContentHostPackagesApplicableController',
        templateUrl: 'content-hosts/content/views/content-host-packages-applicable.html',
        ncyBreadcrumb: {
            label: "{{ 'Applicable Packages' | translate }}",
            parent: 'content-host.info'
        }
    })
    .state('content-host.errata', {
        abstract: true,
        controller: 'ContentHostErrataController',
        template: '<div ui-view></div>'
    })
    .state('content-host.errata.index', {
        url: '/errata?getSearch',
        permission: 'view_hosts',
        templateUrl: 'content-hosts/content/views/content-host-errata.html',
        ncyBreadcrumb: {
            label: "{{ 'Errata' | translate }}",
            parent: 'content-host.info'
        }
    })
    .state('content-host.errata.details', {
        url: '/errata/:errataId',
        permission: 'view_hosts',
        templateUrl: 'content-hosts/content/views/errata-details.html',
        ncyBreadcrumb: {
            label: "{{ erratum.errata_id  }}",
            parent: 'content-host.errata.index'
        }
    })
    .state('content-host.traces', {
        abstract: true,
        collapsed: true,
        controller: 'ContentHostTracesController',
        template: '<div ui-view></div>'
    })
    .state('content-host.traces.index', {
        url: '/traces',
        permission: 'view_hosts',
        templateUrl: 'content-hosts/content/views/content-host-traces.html',
        ncyBreadcrumb: {
            label: "{{ 'Traces' | translate }}",
            parent: 'content-host.info'
        }
    });
}]);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostErrataController
 *
 * @requires $scope
 * @resource $timeout
 * @resource $window
 * @requires HostErratum
 * @requires Nutupane
 * @requires BastionConfig
 *
 * @description
 *   Provides the functionality for the content host package list and actions.
 */

angular.module('Bastion.content-hosts').controller('ContentHostErrataController',
    ['$scope', '$timeout', '$window', 'translate', 'HostErratum', 'Nutupane', 'Organization', 'Environment', 'BastionConfig',
    function ($scope, $timeout, $window, translate, HostErratum, Nutupane, Organization, Environment, BastionConfig) {
        var errataNutupane, params = {
            'sort_by': 'updated',
            'sort_order': 'DESC',
            'paged': true,
            'errata_restrict_applicable': true,
            'id': $scope.$stateParams.hostId
        };

        function loadErratum(errataId) {
            $scope.erratum = HostErratum.get({'id': $scope.host.id,
                'errata_id': errataId});
        }

        errataNutupane = new Nutupane(HostErratum, params, 'get');
        $scope.controllerName = 'katello_errata';
        errataNutupane.masterOnly = true;
        $scope.table = errataNutupane.table;
        $scope.table.errataFilterTerm = "";
        $scope.table.errataCompare = function (item) {
            var searchText = $scope.table.errataFilterTerm;
            return item.type.indexOf(searchText) >= 0 ||
                item['errata_id'].indexOf(searchText) >= 0 ||
                item.title.indexOf(searchText) >= 0;
        };

        $scope.remoteExecutionPresent = BastionConfig.remoteExecutionPresent;
        $scope.remoteExecutionByDefault = BastionConfig.remoteExecutionByDefault;
        $scope.errataActionFormValues = {
            authenticityToken: $window.AUTH_TOKEN.replace(/&quot;/g, '')
        };

        $scope.selectedErrataOption = 'current';
        $scope.errataOptions = [{name: "Current Environment", label: 'current'}, {name: 'foo', label: 'bar'}];

        $scope.table.initialLoad = false;

        $scope.setupErrataOptions = function (host) {
            var libraryString = translate("Library Synced Content"),
                currentEnv,
                previousEnv;

            if (host.hasContent()) {
                currentEnv = translate("Current Environment (%e/%cv)").replace("%e", host.content_facet_attributes.lifecycle_environment.name).replace("%cv", host.content_facet_attributes.content_view_name);
                $scope.errataOptions = [{name: currentEnv, label: 'current', order: 3}];

                if (!host['content_facet_attributes']['lifecycle_environment_library?']) {
                    Environment.get({id: host['content_facet_attributes'].lifecycle_environment.id}).$promise.then(function (env) {
                        previousEnv = translate("Previous Environment (%e/%cv)").replace('%e', env.prior.name).replace("%cv", host.content_facet_attributes.content_view_name);
                        $scope.errataOptions.push({name: previousEnv,
                                                   label: 'prior', order: 2, 'content_view_id': host.content_facet_attributes.content_view_id, 'environment_id': env.prior.id});

                    });
                }
                if (!host.content_facet_attributes['content_view_default?']) {
                    Organization.get({id: host['organization_id']}).$promise.then(function (org) {
                        $scope.errataOptions.push({name: libraryString, label: 'library', order: 1,
                                                   'content_view_id': org['default_content_view_id'], 'environment_id': org.library_id});
                    });
                }
            }
        };

        $scope.host.$promise.then(function() {
            $scope.setupErrataOptions($scope.host);
            if ($scope.host['content_facet_attributes'] && $scope.host.id) {
                errataNutupane.setParams({id: $scope.host.id});
                errataNutupane.load();
            }
        });

        $scope.calculateApplicability = function () {
            $scope.calculatingApplicability = true;
            HostErratum.regenerateApplicability({id: $scope.host.id},
                function (task) {
                    $scope.transitionTo('content-host.tasks.details', {taskId: task.id});
                }, function() {
                    $scope.calculatingApplicability = false;
                });
        };

        $scope.refreshErrata = function (selected) {
            var option, errataParams;
            errataParams = {'id': $scope.host.id};
            $scope.selectedErrataOption = selected;

            if (selected === 'library' || selected === 'prior') {
                option = _.find($scope.errataOptions, function (opt) {
                    return opt.label === selected;
                });
                errataParams['content_view_id'] = option['content_view_id'];
                errataParams['environment_id'] = option['environment_id'];
            }

            errataNutupane.setParams(errataParams);
            errataNutupane.refresh();
        };

        $scope.transitionToErratum = function (erratum) {
            loadErratum(erratum['errata_id']);
            $scope.transitionTo('content-host.errata.details', {errataId: erratum.errata_id});
        };

        $scope.selectedErrataIds = function () {
            var errataIds = [], selected = $scope.table.getSelected();
            angular.forEach(selected, function (value) {
                errataIds.push(value['errata_id']);
            });
            return errataIds;
        };

        $scope.performViaKatelloAgent = function () {
            var errataIds = $scope.selectedErrataIds();
            if (errataIds.length > 0) {
                HostErratum.apply({id: $scope.host.id, 'errata_ids': errataIds},
                                   function (task) {
                                        $scope.table.selectAll(false);
                                        $scope.transitionTo('content-host.tasks.details', {taskId: task.id});
                                    });
            }
        };

        $scope.performViaRemoteExecution = function(customize) {
            var errataIds = $scope.selectedErrataIds();
            $scope.errataActionFormValues.errata = errataIds.join(',');
            $scope.errataActionFormValues.hostIds = $scope.host.id;
            $scope.errataActionFormValues.customize = customize;

            $timeout(function () {
                angular.element('#errataActionForm').submit();
            }, 0);
        };

        $scope.applySelected = function () {
            if ($scope.remoteExecutionByDefault) {
                $scope.performViaRemoteExecution(false);
            } else {
                $scope.performViaKatelloAgent();
            }
        };

        if ($scope.$stateParams.errataId) {
            loadErratum($scope.$stateParams.errataId);
        }
    }
]);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostPackagesActionsController
 *
 * @requires $scope
 *
 * @description
 *   Provides the functionality for the content host package actions.
 */

angular.module('Bastion.content-hosts').controller('ContentHostPackagesActionsController',
    ['$scope', function ($scope) {
        $scope.packageAction = {actionType: 'packageInstall'}; //default to packageInstall
    }
]);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostPackagesApplicableController
 *
 * @requires $scope
 * @resource $timeout
 * @resource $window
 * @requires Package
 * @requires HostPackage
 * @requires translate
 * @requires Nutupane
 *
 * @description
 *   Provides the functionality for the content host packages list and actions.
 */

angular.module('Bastion.content-hosts').controller('ContentHostPackagesApplicableController',
    ['$scope', '$timeout', '$window', 'Package', 'HostPackage', 'translate', 'Nutupane',
    function ($scope, $timeout, $window, Package, HostPackage, translate, Nutupane) {
        var packagesNutupane, openEventInfo;

        $scope.getSelectedPackages = function () {
            var selected = $scope.table.getSelected();
            selected = _.map(selected, function(pkg) {
                return pkg.name + '-' + pkg.version + '-' + pkg.release + '.' + pkg.arch;
            });
            return selected;
        };

        $scope.getRemoteExecutionCommand = function() {
            return $scope.getSelectedPackages().join(' ');
        };

        $scope.getKatelloAgentCommand = function () {
            return $scope.getSelectedPackages().join(',');
        };

        $scope.performDefaultUpdateAction = function () {
            if ($scope.remoteExecutionByDefault) {
                $scope.performViaRemoteExecution('packageUpdate', $scope.getRemoteExecutionCommand(), false);
            } else {
                $scope.performViaKatelloAgent('packageUpdate', $scope.getKatelloAgentCommand());
            }
        };

        packagesNutupane = new Nutupane(Package, {'host_id': $scope.$stateParams.hostId, 'packages_restrict_upgradable': true});
        $scope.controllerName = 'katello_erratum_packages';
        packagesNutupane.masterOnly = true;
        $scope.table = packagesNutupane.table;
        $scope.table.openEventInfo = openEventInfo;
        $scope.table.contentHost = $scope.contentHost;
    }
]);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostPackagesInstalledController
 *
 * @requires $scope
 * @resource $timeout
 * @resource $window
 * @requires HostPackage
 * @requires translate
 * @requires Nutupane
 *
 * @description
 *   Provides the functionality for the content host packages list and actions.
 */

angular.module('Bastion.content-hosts').controller('ContentHostPackagesInstalledController',
    ['$scope', '$timeout', '$window', 'HostPackage', 'translate', 'Nutupane',
    function ($scope, $timeout, $window, HostPackage, translate, Nutupane) {
        var packagesNutupane;

        $scope.removeSelectedPackages = function () {
            var selected = $scope.table.getSelected();

            if (!$scope.working) {
                $scope.working = true;
                HostPackage.remove({
                    id: $scope.host.id,
                    packages: selected
                }, $scope.openEventInfo, $scope.errorHandler);
            }
        };

        packagesNutupane = new Nutupane(HostPackage, {id: $scope.$stateParams.hostId});
        packagesNutupane.masterOnly = true;
        $scope.table = packagesNutupane.table;
        $scope.table.contentHost = $scope.contentHost;
    }
]);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostPackagesController
 *
 * @requires $scope
 * @resource $timeout
 * @resource $window
 * @requires ContentHostPackage
 * @requires translate
 * @requires Nutupane
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the content host packages list and actions.
 */

angular.module('Bastion.content-hosts').controller('ContentHostPackagesController',
    ['$scope', '$timeout', '$window', 'HostPackage', 'translate', 'Nutupane', 'BastionConfig', 'Notification',
    function ($scope, $timeout, $window, HostPackage, translate, Nutupane, BastionConfig, Notification) {
        var packageActions;

        $scope.openEventInfo = function (event) {
            // when the event has label defined, it means it comes
            // from foreman-tasks
            if (event.label) {
                $scope.transitionTo('content-host.tasks.details', {taskId: event.id});
            } else {
                $scope.transitionTo('content-host.events.details', {eventId: event.id});
            }
            $scope.working = false;
        };

        $scope.errorHandler = function (response) {
            angular.forEach(response.data.errors, function (responseError) {
                Notification.setErrorMessage(responseError);
            });
            $scope.working = false;
        };

        $scope.working = false;
        $scope.remoteExecutionPresent = BastionConfig.remoteExecutionPresent;
        $scope.remoteExecutionByDefault = BastionConfig.remoteExecutionByDefault;
        $scope.packageActionFormValues = {
            authenticityToken: $window.AUTH_TOKEN.replace(/&quot;/g, '')
        };

        $scope.updateAll = function () {
            $scope.working = true;
            HostPackage.updateAll({id: $scope.host.id}, $scope.openEventInfo, $scope.errorHandler);
        };

        $scope.performPackageAction = function (actionType, term) {
            if ($scope.remoteExecutionByDefault) {
                $scope.performViaRemoteExecution(actionType, term, false);
            } else {
                $scope.performViaKatelloAgent(actionType, term);
            }
        };

        $scope.performViaKatelloAgent = function (actionType, term) {
            var terms = term.split(/ *, */);
            $scope.working = true;
            packageActions[actionType](terms);
        };

        $scope.performViaRemoteExecution = function(actionType, term, customize) {
            $scope.packageActionFormValues.package = term;
            $scope.packageActionFormValues.remoteAction = actionType;
            $scope.packageActionFormValues.hostIds = $scope.host.id;
            $scope.packageActionFormValues.customize = customize;

            $timeout(function () {
                angular.element('#packageActionForm').submit();
            }, 0);
        };

        packageActions = {
            packageInstall: function (termList) {
                HostPackage.install({id: $scope.host.id, packages: termList}, $scope.openEventInfo, $scope.errorHandler);
            },
            packageUpdate: function (termList) {
                HostPackage.update({id: $scope.host.id, packages: termList}, $scope.openEventInfo, $scope.errorHandler);
            },
            packageRemove: function (termList) {
                HostPackage.remove({id: $scope.host.id, packages: termList}, $scope.openEventInfo, $scope.errorHandler);
            },
            groupInstall: function (termList) {
                HostPackage.install({id: $scope.host.id, groups: termList}, $scope.openEventInfo, $scope.errorHandler);
            },
            groupRemove: function (termList) {
                HostPackage.remove({id: $scope.host.id, groups: termList}, $scope.openEventInfo, $scope.errorHandler);
            }
        };
    }
]);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostTracesController
 *
 * @requires $scope
 * @resource $timeout
 * @resource $window
 * @requires HostTraces
 * @requires Nutupane
 * @requires BastionConfig
 *
 * @description
 *   Provides the functionality for the content host package list and actions.
 */
/*jshint camelcase:false*/

angular.module('Bastion.content-hosts').controller('ContentHostTracesController',
    ['$scope', '$timeout', '$window', 'translate', 'HostTraces', 'Nutupane', 'BastionConfig',
    function ($scope, $timeout, $window, translate, HostTraces, Nutupane, BastionConfig) {
        var tracesNutupane, params = {
            'paged': true
        };

        tracesNutupane = new Nutupane(HostTraces, params, 'get');
        tracesNutupane.masterOnly = true;
        $scope.table = tracesNutupane.table;
        $scope.table.tracesFilterTerm = "";
        $scope.table.tracesCompare = function (item) {
            var searchText = $scope.table.tracesFilterTerm;
            return item.app_type.indexOf(searchText) >= 0 ||
                item.application.indexOf(searchText) >= 0 ||
                item.helper.indexOf(searchText) >= 0;
        };

        $scope.remoteExecutionPresent = BastionConfig.remoteExecutionPresent;

        $scope.host.$promise.then(function() {
            if ($scope.host.id) {
                tracesNutupane.setParams({id: $scope.host.id});
                tracesNutupane.load();
            }
        });

        $scope.selectedTraceHelpers = function() {
            var traceHelpers = [],
                selected = $scope.table.getSelected();

            var reboot = selected.some(function(value) {
                return value.app_type === "static";
            });

            if (reboot) {
                return ["reboot"];
            }

            selected.forEach(function(value) {
                if (value.app_type !== "session") {
                    if (traceHelpers.indexOf(value.helper) === -1) {
                        traceHelpers.push(value.helper);
                    }
                }
            });
            return traceHelpers;
        };

        $scope.traceActionFormValues = {
            authenticityToken: $window.AUTH_TOKEN.replace(/&quot;/g, '')
        };

        $scope.performViaRemoteExecution = function(customize) {
            var traceHelpers = $scope.selectedTraceHelpers();
            $scope.traceActionFormValues.helper = $.grep(traceHelpers, Boolean).join(',');
            $scope.traceActionFormValues.hostIds = $scope.host.id;
            $scope.traceActionFormValues.customize = customize;

            $timeout(function () {
                angular.element('#traceActionForm').submit();
            }, 0);
        };

    }
]);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostAddHostCollectionsController
 *
 * @requires $scope
 * @requires $q
 * @requires $location
 * @requires translate
 * @requires HostCollection
 * @requires Host
 * @requires CurrentOrganization
 * @requires Nutupane
 * @requires Notification
 *
 * @description
 *   Provides the functionality for adding host collections to a content host.
 */

angular.module('Bastion.content-hosts').controller('ContentHostAddHostCollectionsController',
    ['$scope', '$q', '$location', 'translate', 'HostCollection', 'Host', 'CurrentOrganization', 'Nutupane', 'Notification',
    function ($scope, $q, $location, translate, HostCollection, Host, CurrentOrganization, Nutupane, Notification) {
        var params, nutupane;

        params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'sort_by': 'name',
            'sort_order': 'ASC',
            'paged': true,
            'available_for': 'host',
            'host_id': $scope.$stateParams.hostId
        };

        nutupane = new Nutupane(HostCollection, params);
        $scope.controllerName = 'katello_host_collections';
        nutupane.masterOnly = true;

        $scope.table = nutupane.table;

        $scope.addHostCollections = function (host) {
            var deferred = $q.defer(),
                success,
                error,
                data,
                hostCollections,
                hostCollectionsToAdd;

            success = function (response) {
                Notification.setSuccessMessage(translate('Added %x host collections to content host "%y".')
                    .replace('%x', $scope.table.numSelected).replace('%y', host.name));
                $scope.table.working = false;
                $scope.table.selectAll(false);
                nutupane.refresh();
                $scope.host.$get();
                deferred.resolve(response);
            };

            error = function (response) {
                deferred.reject(response.data.errors);
                Notification.setErrataIds(response.data.errors.base);
                $scope.table.working = false;
            };

            $scope.table.working = true;

            hostCollections = _.map(host['host_collections'], 'id');
            hostCollectionsToAdd = _.map($scope.table.getSelected(), 'id');
            data = {"host_collection_ids": _.union(hostCollections, hostCollectionsToAdd)};

            Host.updateHostCollections({id: host.id}, data, success, error);

            return deferred.promise;
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostAddSubscriptionsController
 *
 * @requires $scope
 * @requires $location
 * @requires translate
 * @requires CurrentOrganization
 * @requires Subscription
 * @requires ContentHost
 * @requires SubscriptionsHelper
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the content host details action pane.
 */

angular.module('Bastion.content-hosts').controller('ContentHostAddSubscriptionsController',
    ['$scope', '$location', 'translate', 'Nutupane', 'CurrentOrganization', 'Host', 'HostSubscription', 'Subscription', 'SubscriptionsHelper', 'Notification',
    function ($scope, $location, translate, Nutupane, CurrentOrganization, Host, HostSubscription, Subscription, SubscriptionsHelper, Notification) {

        var params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'sort_order': 'ASC',
            'available_for': 'host',
            'host_id': $scope.$stateParams.hostId
        };

        $scope.nutupane = new Nutupane(Subscription, params);
        $scope.controllerName = 'katello_subscriptions';
        $scope.nutupane.setSearchKey('subscriptionSearch');
        $scope.nutupane.masterOnly = true;
        $scope.table = $scope.nutupane.table;

        $scope.isAdding = false;
        $scope.groupedSubscriptions = {};

        $scope.$watch('table.rows', function (rows) {
            $scope.groupedSubscriptions = SubscriptionsHelper.groupByProductName(rows);
        });

        $scope.getAmountSelectorValues = SubscriptionsHelper.getAmountSelectorValues;

        $scope.showMatchHost = false;
        $scope.showMatchInstalled = false;
        $scope.showNoOverlap = false;

        $scope.toggleFilters = function () {
            $scope.nutupane.table.params['match_host'] = $scope.showMatchHost;
            $scope.nutupane.table.params['match_installed'] = $scope.showMatchInstalled;
            $scope.nutupane.table.params['no_overlap'] = $scope.showNoOverlap;
            $scope.nutupane.refresh();
        };

        $scope.disableAddButton = function () {
            return $scope.table.numSelected === 0 || $scope.isAdding;
        };

        $scope.addSelected = function () {
            var selected;
            selected = SubscriptionsHelper.getSelectedSubscriptionAmounts($scope.table);

            $scope.isAdding = true;
            HostSubscription.addSubscriptions({id: $scope.$stateParams.hostId, 'subscriptions': selected}, function () {
                Host.get({id: $scope.$stateParams.hostId}, function (host) {
                    $scope.$parent.host = host;
                    Notification.setSuccessMessage(translate("Successfully added %s subscriptions.").replace('%s', selected.length));
                    $scope.isAdding = false;
                    $scope.nutupane.refresh();
                });
            }, function (response) {
                Notification.setErrorMessage(response.data.displayMessage);
                $scope.isAdding = false;
                $scope.nutupane.refresh();
            });
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostBaseSubscriptionsController
 *
 * @requires $scope
 * @requires translate
 * @requires CurrentOrganization
 * @requires Subscription
 * @requires Nutupane
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the content host details action pane.
 */

angular.module('Bastion.content-hosts').controller('ContentHostBaseSubscriptionsController',
    ['$scope', '$location', 'translate', 'CurrentOrganization', 'Subscription', 'HostSubscription', 'Notification',
    function ($scope, $location, translate, CurrentOrganization, Subscription, HostSubscription, Notification) {

        function success() {
            $scope.subscription.workingMode = false;
            Notification.setSuccessMessage(translate('Successfully updated subscriptions.'));
            $scope.host.$get();
        }

        function failure() {
            $scope.subscription.workingMode = false;
            Notification.setErrorMessage(translate('An error occurred trying to auto-attach subscriptions.  Please check your log for further information.'));
        }

        $scope.subscription = {
            workingMode: false
        };

        $scope.autoAttachSubscriptions = function () {
            $scope.subscription.workingMode = true;
            HostSubscription.autoAttach({id: $scope.host.id}, success, failure);
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostDetailsController
 *
 * @requires $scope
 * @requires $q
 * @requires translate
 * @requires ContentHost
 * @requires ContentView
 * @requires Organization
 * @requires CurrentOrganization
 * @requires CurrentHostsHelper
 *
 * @description
 *   Provides the functionality for the content host details action pane.
 */

angular.module('Bastion.content-hosts').controller('ContentHostDetailsInfoController',
    ['$scope', '$q', 'translate', 'HostSubscription', 'ContentView', 'Organization', 'CurrentOrganization', 'ContentHostsHelper',
    function ($scope, $q, translate, HostSubscription, ContentView, Organization, CurrentOrganization, ContentHostsHelper) {
        function doubleColonNotationToObject(dotString) {
            var doubleColonObject = {}, tempObject, parts, part, key, property;
            for (property in dotString) {
                if (dotString.hasOwnProperty(property)) {
                    tempObject = doubleColonObject;
                    parts = property.split('::');
                    key = parts.pop();
                    while (parts.length) {
                        part = parts.shift();
                        tempObject = tempObject[part] = tempObject[part] || {};
                    }
                    tempObject[key] = dotString[property];
                }
            }
            return doubleColonObject;
        }

        $scope.host.$promise.then(function (host) {
            $scope.hostFactsAsObject = doubleColonNotationToObject(host.facts);
            if (host.hasContent()) {
                $scope.originalEnvironment = host.content_facet_attributes.lifecycle_environment;
            }
        });

        $scope.showVersionAlert = false;
        $scope.editContentView = false;
        $scope.disableEnvironmentSelection = false;
        $scope.environments = [];

        $scope.environments = Organization.readableEnvironments({id: CurrentOrganization});

        $scope.$watch('host.content_facet_attributes.lifecycle_environment', function (environment) {
            if (environment && $scope.originalEnvironment) {
                if (environment.id !== $scope.originalEnvironment.id) {
                    $scope.editContentView = true;
                    $scope.disableEnvironmentSelection = true;
                }
            }
        });

        $scope.cancelReleaseVersionUpdate = function () {
            $scope.showVersionAlert = false;
        };

        $scope.cancelContentViewUpdate = function () {
            if ($scope.editContentView) {
                $scope.editContentView = false;
                $scope.host.content_facet_attributes['lifecycle_environment'] = $scope.originalEnvironment;
                $scope.disableEnvironmentSelection = false;
            }
        };

        $scope.saveContentView = function (host) {
            $scope.editContentView = false;

            $scope.saveContentFacet(host).then(function (response) {
                $scope.originalEnvironment = response.content_facet_attributes.lifecycle_environment;
            });
            $scope.disableEnvironmentSelection = false;
        };

        $scope.releaseVersions = function () {
            var deferred = $q.defer();

            HostSubscription.releaseVersions({ id: $scope.host.id }, function (response) {
                if (response.total === 0) {
                    $scope.showVersionAlert = true;
                }
                deferred.resolve(response.results);
            });

            return deferred.promise;
        };

        $scope.clearReleaseVersion = function () {
            $scope.host.subscription_facet_attributes['release_version'] = '';
            $scope.saveSubscriptionFacet($scope.host);
        };

        $scope.clearServiceLevel = function () {
            $scope.host.subscription_facet_attributes['service_level'] = '';
            $scope.saveSubscriptionFacet($scope.host);
        };

        $scope.contentViews = function () {
            var deferred = $q.defer();

            ContentView.queryUnpaged({ 'environment_id': $scope.host.content_facet_attributes.lifecycle_environment.id}, function (response) {
                deferred.resolve(response.results);
                $scope.contentViews = response.results;
            });

            return deferred.promise;
        };

        $scope.getActivationKeyLink = function (activationKey) {
            return '/activation_keys!=&panel=activation_key_%s&panelpage=edit'.replace('%s', activationKey.id);
        };

        $scope.virtualGuestIds = function (host) {
            var ids = [];
            if (host && host.subscription_facet_attributes) {
                angular.forEach(host.subscription_facet_attributes['virtual_guests'], function (guest) {
                    ids.push('name = %s'.replace('%s', guest.name));
                });
            }

            return ids.join(" or ");
        };

        $scope.convertMemToGB = ContentHostsHelper.convertMemToGB;

        $scope.virtual = function (virt) {
            return (virt === true || virt === 'true');
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostDetailsController
 *
 * @requires $scope
 * @requires $state
 * @requires $q
 * @requires translate
 * @requires Organization
 * @requires CurrentOrganization
 * @requires MenuExpander
 * @requires ApiErrorHandler
 * @requires deleteHostOnUnregister
 *
 * @description
 *   Provides the functionality for the content host details action pane.
 */

angular.module('Bastion.content-hosts').controller('ContentHostDetailsController',
    ['$scope', '$state', '$q', 'translate', 'Host', 'HostSubscription', 'Organization', 'CurrentOrganization', 'Notification', 'MenuExpander', 'ApiErrorHandler', 'deleteHostOnUnregister',
    function ($scope, $state, $q, translate, Host, HostSubscription, Organization, CurrentOrganization, Notification, MenuExpander, ApiErrorHandler, deleteHostOnUnregister) {
        $scope.menuExpander = MenuExpander;

        $scope.panel = {
            error: false,
            loading: true
        };

        $scope.host = Host.get({id: $scope.$stateParams.hostId}, function (host) {
            host.unregisterDelete = !host.hasSubscription() || deleteHostOnUnregister;
            host.deleteHostOnUnregister = deleteHostOnUnregister;
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        // @TODO begin hack for content and subscript facets
        // see http://projects.theforeman.org/issues/13763
        $scope.saveContentFacet = function (host) {
            var newHost = {id: host.id};
            newHost['content_facet_attributes'] = {
                id: host.content_facet_attributes.id,
                'content_view_id': host.content_facet_attributes.content_view.id,
                'lifecycle_environment_id': host.content_facet_attributes.lifecycle_environment.id
            };
            return $scope.save(newHost, true);
        };

        $scope.saveSubscriptionFacet = function (host) {
            var newHost = {id: host.id};
            newHost['subscription_facet_attributes'] = {
                id: host.subscription_facet_attributes.id,
                autoheal: host.subscription_facet_attributes.autoheal,
                'service_level': host.subscription_facet_attributes.service_level,
                'release_version': host.subscription_facet_attributes.release_version
            };
            return $scope.save(newHost, true);
        };
        // @TODO end hack

        $scope.save = function (host, saveFacets) {
            var deferred = $q.defer();

            // @TODO begin hack needed to use the foreman host API, see the following bugs:
            // http://projects.theforeman.org/issues/13622
            // http://projects.theforeman.org/issues/13669
            // http://projects.theforeman.org/issues/13670
            // http://projects.theforeman.org/issues/13672
            // http://projects.theforeman.org/issues/13759

            var whitelistedHostObject = {},
                whitelist = [
                    "name",
                    "comment"
                ];

            if (saveFacets) {
                whitelist.push("content_facet_attributes");
                whitelist.push("subscription_facet_attributes");
            }

            angular.forEach(whitelist, function (key) {
                whitelistedHostObject[key] = host[key];
            });

            Host.update({id: host.id, host: whitelistedHostObject}, function (response) {
                deferred.resolve(response);
                $scope.host = response;
                Notification.setSuccessMessage(translate('Save Successful.'));
            }, function (response) {
                deferred.reject(response);
                _.each(response.data.error.full_messages, function (errorMessage) {
                    Notification.setErrorMessage(translate("An error occurred saving the Content Host: ") + errorMessage);
                });
            });
            // @TODO end hack

            return deferred.promise;
        };

        $scope.transitionTo = function (state, params) {
            var hostId = $scope.$stateParams.hostId;

            if ($scope.host && $scope.host.id) {
                hostId = $scope.host.id;
            }

            if (hostId) {
                params = params ? params : {};
                params.hostId = hostId;
                $state.transitionTo(state, params);
                return true;
            }
            return false;
        };

        $scope.serviceLevels = function () {
            var deferred = $q.defer();

            Organization.get({id: CurrentOrganization}, function (organization) {
                deferred.resolve(organization['service_levels']);
            });

            return deferred.promise;
        };

        $scope.unregisterContentHost = function (host) {
            var errorHandler = function (response) {
                host.deleting = false;
                Notification.setErrorMessage(translate('An error occured: %s').replace('%s', response.data.displayMessage));
            };
            host.deleting = true;

            if (host.unregisterDelete) {
                host.$delete(function () {
                    host.deleting = false;
                    Notification.setSuccessMessage(translate('Host %s has been deleted.').replace('%s', host.name));
                    $scope.transitionTo('content-hosts');
                }, errorHandler);
            } else {
                HostSubscription.delete({id: host.id}, function () {
                    host.deleting = false;
                    Notification.setSuccessMessage(translate('Host %s has been unregistered.').replace('%s', host.name));
                    $scope.transitionTo('content-hosts');
                }, errorHandler);
            }
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostHostCollectionsController
 *
 * @requires $scope
 * @requires $q
 * @requires $location
 * @requires translate
 * @requires HostCollection
 * @requires Host
 * @requires CurrentOrganization
 * @requires Nutupane
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the list host collections details action pane.
 */

angular.module('Bastion.content-hosts').controller('ContentHostHostCollectionsController',
    ['$scope', '$q', '$location', 'translate', 'HostCollection', 'Host', 'CurrentOrganization', 'Nutupane', 'Notification',
    function ($scope, $q, $location, translate, HostCollection, Host, CurrentOrganization, Nutupane, Notification) {
        var nutupane, params;

        params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'sort_by': 'name',
            'sort_order': 'ASC',
            'paged': true,
            'host_id': $scope.$stateParams.hostId
        };

        nutupane = new Nutupane(HostCollection, params);
        $scope.controllerName = 'katello_host_collections';
        nutupane.masterOnly = true;

        $scope.table = nutupane.table;

        $scope.removeHostCollections = function (host) {
            var deferred = $q.defer(),
                success,
                error,
                data,
                hostCollections,
                hostCollectionsToRemove;

            success = function (response) {
                var message = translate('Removed %x host collections from content host "%y".')
                    .replace('%x', $scope.table.numSelected).replace('%y', host.name);

                Notification.setSuccessMessage(message);
                $scope.table.working = false;
                $scope.table.selectAll(false);
                nutupane.refresh();
                $scope.host.$get();
                deferred.resolve(response);
            };

            error = function (response) {
                deferred.reject(response.data.errors);
                angular.forEach(response.data.errors, function (responseError) {
                    Notification.setErrorMessage(responseError);
                });
                $scope.table.working = false;
            };

            $scope.table.working = true;
            hostCollections = _.map(host['host_collections'], 'id');
            hostCollectionsToRemove = _.map($scope.table.getSelected(), 'id');

            data = {"host_collection_ids": _.difference(hostCollections, hostCollectionsToRemove)};
            Host.updateHostCollections({id: host.id}, data, success, error);

            return deferred.promise;
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostRepositorySetsController
 *
 * @requires $scope
 * @requires translate
 * @requires Nutupane
 * @requires HostSubscription
 * @requires ContentOverrideHelper
 * @requires Notification
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality for the content-host products action pane.
 */

angular.module('Bastion.content-hosts').controller('ContentHostRepositorySetsController',
    ['$scope', 'translate', 'Nutupane', 'HostSubscription', 'ContentOverrideHelper', 'Notification', 'CurrentOrganization',
    function ($scope, translate, Nutupane, HostSubscription, ContentOverrideHelper, Notification, CurrentOrganization) {
        var params, saveContentOverride, success, error;

        params = {
            id: $scope.$stateParams.hostId,
            'organization_id': CurrentOrganization,
            enabled: true,
            'full_result': true,
            'include_available_content': true
        };

        $scope.controllerName = 'katello_products';
        $scope.nutupane = new Nutupane(HostSubscription, params, 'repositorySets');
        $scope.table = $scope.nutupane.table;

        $scope.contentAccessModes = {
            contentAccessModeAll: false,
            contentAccessModeEnv: false
        };
        $scope.toggleFilters = function () {
            $scope.nutupane.table.params['content_access_mode_all'] = $scope.contentAccessModes.contentAccessModeAll;
            $scope.nutupane.table.params['content_access_mode_env'] = $scope.contentAccessModes.contentAccessModeEnv;
            $scope.nutupane.refresh();
        };

        success = function () {
            $scope.table.working = false;
            Notification.setSuccessMessage(translate('Repository Sets settings saved successfully.'));
            $scope.nutupane.refresh();
        };

        error = function (response) {
            $scope.table.working = false;
            Notification.setErrorMessage(response.data.errors);
        };

        saveContentOverride = function (contentOverrides) {
            $scope.table.working = true;
            HostSubscription.contentOverride({id: $scope.$stateParams.hostId}, contentOverrides, success, error);
        };

        $scope.overrideToEnabled = function () {
            var contentOverrides = ContentOverrideHelper.getEnabledContentOverrides($scope.table.getSelected());
            saveContentOverride(contentOverrides);
        };

        $scope.overrideToDisabled = function () {
            var contentOverrides = ContentOverrideHelper.getDisabledContentOverrides($scope.table.getSelected());
            saveContentOverride(contentOverrides);
        };

        $scope.resetToDefault = function () {
            var contentOverrides = ContentOverrideHelper.getDefaultContentOverrides($scope.table.getSelected());
            saveContentOverride(contentOverrides);
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-hosts.controller:ContentHostSubscriptionsController
 *
 * @requires $scope
 * @requires $location
 * @requires translate
 * @requires Subscription
 * @requires SubscriptionsHelper
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the content host details action pane.
 */

angular.module('Bastion.content-hosts').controller('ContentHostSubscriptionsController',
    ['$scope', '$location', 'translate', 'Nutupane', 'CurrentOrganization', 'Subscription', 'Host', 'HostSubscription', 'SubscriptionsHelper', 'Notification',
    function ($scope, $location, translate, Nutupane, CurrentOrganization, Subscription, Host, HostSubscription, SubscriptionsHelper, Notification) {

        var params = {
            'organization_id': CurrentOrganization,
            'id': $scope.$stateParams.hostId,
            'search': $location.search().search || "",
            'sort_order': 'ASC'
        };

        $scope.nutupane = new Nutupane(HostSubscription, params);
        $scope.controllerName = 'katello_subscriptions';
        $scope.nutupane.table.initialLoad = false;
        $scope.table = $scope.nutupane.table;
        $scope.nutupane.masterOnly = true;
        $scope.isRemoving = false;

        $scope.groupedSubscriptions = {};
        $scope.$watch('table.rows', function (rows) {
            $scope.groupedSubscriptions = SubscriptionsHelper.groupByProductName(rows);
        });

        $scope.nutupane.setParams(params);
        $scope.nutupane.load(true);

        $scope.disableRemoveButton = function () {
            return $scope.table.numSelected === 0 || $scope.isRemoving;
        };

        $scope.removeSelected = function () {
            var selected;
            selected = SubscriptionsHelper.getSelectedSubscriptions($scope.table);

            $scope.isRemoving = true;
            HostSubscription.removeSubscriptions({id: $scope.$stateParams.hostId, 'subscriptions': selected}, function () {
                Host.get({id: $scope.$stateParams.hostId}, function (host) {
                    $scope.$parent.host = host;
                    $scope.nutupane.table.selectAll(false);
                    $scope.nutupane.refresh();
                    Notification.setSuccessMessage(translate("Successfully removed %s subscriptions.").replace('%s', selected.length));
                    $scope.isRemoving = false;
                });
            }, function (response) {
                $scope.isRemoving = false;
                Notification.setErrorMessage(translate("An error occurred removing the subscriptions.") + response.data.displayMessage);
            });
        };
    }
]);
/**
 * @ngdoc module
 * @name  Bastion.hosts
 *
 * @description
 *    Module for hosts
 */

angular.module('Bastion.hosts', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.components'
]);
/**
 * @ngdoc service
 * @name  Bastion.hosts.factory:HostBulkAction
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for bulk actions on hosts.
 */

angular.module('Bastion.hosts').factory('HostBulkAction',
    ['BastionResource', function (BastionResource) {

        return BastionResource('api/v2/hosts/bulk/:action', {}, {
            addHostCollections: {method: 'PUT', params: {action: 'add_host_collections'}},
            removeHostCollections: {method: 'PUT', params: {action: 'remove_host_collections'}},
            updateRepositorySets: {method: 'PUT', params: {action: 'content_overrides'}},
            addSubscriptions: {method: 'PUT', params: {action: 'add_subscriptions'}},
            removeSubscriptions: {method: 'PUT', params: {action: 'remove_subscriptions'}},
            installableErrata: {method: 'POST', params: {action: 'installable_errata'}},
            installContent: {method: 'PUT', params: {action: 'install_content'}},
            updateContent: {method: 'PUT', params: {action: 'update_content'}},
            removeContent: {method: 'PUT', params: {action: 'remove_content'}},
            autoAttach: {method: 'PUT', params: {action: 'auto_attach'}},
            destroyHosts: {method: 'PUT', params: {action: 'destroy'}},
            environmentContentView: {method: 'PUT', params: {action: 'environment_content_view'}},
            releaseVersion: {method: 'PUT', params: {action: 'release_version'}},
            availableIncrementalUpdates: {method: 'POST', isArray: true, params: {action: 'available_incremental_updates'}}
        });

    }]
);
/**
 * @ngdoc service
 * @name  Bastion.content-hosts.factory:HostErratum
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for the errata of a single content host
 */

angular.module('Bastion.hosts').factory('HostErratum',
    ['BastionResource', function (BastionResource) {

        return BastionResource('api/v2/hosts/:id/errata/:errata_id/:action', {id: '@id'}, {
            get: {method: 'GET', isArray: false, transformResponse: function (data) {
                data = angular.fromJson(data);
                angular.forEach(data.results, function (errata) {
                    errata.unselectable = !errata.installable;
                });
                return data;
            }},
            regenerateApplicability: {method: 'PUT', isArray: false, params: {action: 'applicability'}},
            autocomplete: {method: 'GET', isArray: true, params: {action: 'auto_complete_search'}},
            apply: {method: 'PUT', params: {action: 'apply'}}
        });

    }]
);
/**
 * @ngdoc service
 * @name  Bastion.hosts.factory:HostPackage
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for the packages of a single content host
 */

angular.module('Bastion.hosts').factory('HostPackage',
    ['BastionResource', function (BastionResource) {

        return BastionResource('api/v2/hosts/:id/packages/:action', {id: '@id'}, {
            get: {method: 'GET', isArray: false},
            remove: {method: 'PUT', params: {action: 'remove'}},
            install: {method: 'PUT', params: {action: 'install'}},
            update: {method: 'PUT', params: {action: 'upgrade'}},
            updateAll: {method: 'PUT', params: {action: 'upgrade_all'}},
            autocomplete: {method: 'GET', isArray: true, params: {action: 'auto_complete_search'}},
            applicable: {method: 'GET', params: {action: 'applicable'}}
        });

    }]
);
/**
 * @ngdoc service
 * @name  Bastion.content-hosts.factory:ContentHost
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for one or more content hosts.
 */

angular.module('Bastion.hosts').factory('HostSubscription',
    ['BastionResource', function (BastionResource) {

        return BastionResource('api/v2/hosts/:id/subscriptions/:action/', {id: '@id'}, {
            events: {method: 'GET', params: {action: 'events'}},
            autoAttach: {method: 'PUT', params: {action: 'auto_attach'}},
            removeSubscriptions: {method: 'put', isArray: false, params: {action: 'remove_subscriptions'}},
            addSubscriptions: {method: 'put', isArray: false, params: {action: 'add_subscriptions'}},
            repositorySets: {method: 'get', isArray: false, params: {action: 'product_content'}},
            contentOverride: {method: 'put', isArray: false, params: {action: 'content_override'}},
            releaseVersions: {method: 'get', isArray: false, params: {action: 'available_release_versions'}}
        });

    }]
);
/**
 * @ngdoc service
 * @name  Bastion.hosts.factory:HostTraces
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for the packages of a single content host
 */

angular.module('Bastion.hosts').factory('HostTraces',
    ['BastionResource', function (BastionResource) {

        return BastionResource('api/v2/hosts/:id/traces/:action', {id: '@id'}, {
            get: {method: 'GET', isArray: false}
        });

    }]
);
/**
 * @ngdoc service
 * @name  Bastion.hosts.factory:Host
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for one or more hosts.
 */

angular.module('Bastion.hosts').factory('Host',
    ['BastionResource', function (BastionResource) {
        var resource = BastionResource('api/v2/hosts/:id/:action', {id: '@id'}, {
            postIndex: {method: 'POST', params: {action: 'post_index'}},
            update: {method: 'PUT'},
            updateHostCollections: {method: 'PUT', params: {action: 'host_collections'}},
            autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}}
        });
        resource.prototype.hasContent = function () {
            return angular.isDefined(this.content_facet_attributes) && angular.isDefined(this.content_facet_attributes.uuid);
        };
        resource.prototype.hasSubscription = function () {
            return angular.isDefined(this.subscription_facet_attributes) && angular.isDefined(this.subscription_facet_attributes.uuid);
        };
        return resource;
    }]
);
/**
 * @ngdoc module
 * @name  Bastion.subscriptions
 *
 * @description
 *   Module for subscriptions
 */

angular.module('Bastion.subscriptions', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.organizations',
    'Bastion.common',
    'Bastion.components',
    'Bastion.components.formatters'
]);
/**
 * @ngdoc directive
 * @name Bastion.subscriptions:contentAccessModeBanner
 *
 * @requires contentAccessMode
 *
 * @description
 *   Component for showing information about content access mode (whether content is
 *   allowed with or without a subscription)
 */

angular.module('Bastion.subscriptions').directive('contentAccessModeBanner',
    ['contentAccessMode',
    function (contentAccessMode) {
        return {
            restrict: 'AE',
            controller: ['$scope', function ($scope) {
                $scope.contentAccessMode = contentAccessMode;
            }],
            templateUrl: 'subscriptions/views/content-access-mode-banner.html'
        };
    }
]);
/**
 * @ngdoc object
 * @name  Bastion.subscriptions.controller:SubscriptionDetailsController
 *
 * @requires $scope
 * @requires translate
 * @requires Subscription
 * @requires ApiErrorHandler
 *
 * @description
 *   Provides the functionality for the subscription details action pane.
 */

angular.module('Bastion.subscriptions').controller('SubscriptionDetailsController',
    ['$scope', 'translate', 'Subscription', 'ApiErrorHandler',
    function ($scope, translate, Subscription, ApiErrorHandler) {
        $scope.panel = {
            error: false,
            loading: true
        };

        if ($scope.subscription) {
            $scope.panel.loading = false;
        }

        $scope.searchQuery = function (subscription) {
            return 'subscription_id="%s"'.replace('%s', subscription.id);
        };

        $scope.subscription = Subscription.get({id: $scope.$stateParams.subscriptionId}, function () {
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        $scope.subscriptionLimits = function (subscription) {
            var limits = [];

            if (subscription.sockets) {
                limits.push(translate("Sockets: %s").replace("%s", subscription.sockets));
            }
            if (subscription.cores) {
                limits.push(translate("Cores: %s").replace("%s", subscription.cores));
            }
            if (subscription.ram) {
                limits.push(translate("RAM: %s GB").replace("%s", subscription.ram));
            }

            if (limits.length > 0) {
                return limits.join(", ");
            }

            return "";
        };

        $scope.virtWhoToolTip = translate("If the virt-who field is Yes then the subscription requires the use of virt-who. Learn how to configure and use this tool in the <a href=\"https://access.redhat.com/documentation/en/red-hat-satellite/6.2/single/virtual-instances-guide/\">Virtual Instances Guide</a>.");
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.subscriptions.controller:SubscriptionProductDetailsController
 *
 * @requires $scope
 */

angular.module('Bastion.subscriptions').controller('SubscriptionProductDetailsController',
    ['$scope',
    function ($scope) {
        $scope.expanded = true;
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.subscriptions.controller:SubscriptionProductsController
 *
 * @requires $scope
 * @requires $state
 * @requires Subscription
 * @requires Product
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality for the subscription products action pane.
 */

angular.module('Bastion.subscriptions').controller('SubscriptionProductsController',
    ['$scope', '$state', 'Subscription', 'Product', 'CurrentOrganization', function ($scope, $state, Subscription, Product, CurrentOrganization) {

        $scope.displayArea = { working: true };

        Product.queryUnpaged({'organization_id': CurrentOrganization,
                              'subscription_id': $scope.$stateParams.subscriptionId,
                              enabled: true,
                              'full_result': true,
                              'include_available_content': true
                             }, function (response) {
            $scope.products = response.results;
            $scope.displayArea.working = false;
        });

    }]
);
/**
 * @ngdoc filter
 * @name  Bastion.subscriptions.filter:groupedFilter.filter.js
 *
 */

angular.module('Bastion.subscriptions').filter('groupedFilter', function () {
    return function (grouped, search) {
        var result = {};
        angular.forEach(grouped, function (value, key) {
            if (angular.isUndefined(search)) {
                search = "";
            } else {
                search = search.toLowerCase();
            }
            if (key.toLowerCase().indexOf(search) > -1 || search === "") {
                result[key] = value;
            }
        });
        return result;
    };
});
/**
 * @ngdoc object
 * @name  Bastion.subscriptions.controller:ManifestDetailsController
 *
 * @requires $scope
 * @requires $q
 * @requires CurrentOrganization
 * @requires Organization
 *
 * @description
 *   Controls the import of a manifest.
 */

angular.module('Bastion.subscriptions').controller('ManifestDetailsController',
    ['$scope', '$q', 'CurrentOrganization', 'Organization',
    function ($scope, $q, CurrentOrganization, Organization) {

        $scope.organization = Organization.get({id: CurrentOrganization});
        $scope.redhatProvider = Organization.redhatProvider();

        $q.all([$scope.organization.$promise]).then(function () {
            $scope.details = $scope.organization['owner_details'];
            $scope.upstream = $scope.details.upstreamConsumer;

            angular.forEach($scope.redhatProvider['owner_imports'], function (value) {
                if (value.upstreamId === $scope.upstream.uuid) {
                    $scope.manifestImport = value;
                }
            });

            $scope.panel.loading = false;
        });

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.subscriptions.controller:ManifestHistoryController
 *
 * @requires $scope
 *
 * @description
 *   Controls the import of a manifest.
 */

angular.module('Bastion.subscriptions').controller('ManifestHistoryController',
    ['$scope', 'Subscription', function ($scope, Subscription) {
        $scope.histories = Subscription.manifestHistory();
        $scope.histories.$promise.then(function (result) {
            $scope.statuses = result;
            $scope.panel.loading = false;
        });
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.subscriptions.controller:ManifestImportController
 *
 * @requires $scope
 * @requires $q
 * @requires translate
 * @requires CurrentOrganization
 * @requires Organization
 * @requires Subscription
 * @requires Task
 *
 * @description
 *   Controls the import of a manifest.
 */

angular.module('Bastion.subscriptions').controller('ManifestImportController',
    ['$scope', '$q', 'translate', 'CurrentOrganization', 'Organization', 'Task', 'Subscription', 'Notification',
    function ($scope, $q, translate, CurrentOrganization, Organization, Task, Subscription, Notification) {

        function buildManifestLink(upstream) {
            var url = upstream.webUrl,
                upstreamId = upstream.uuid;

            if (!url.match(/^http/)) {
                url = "https://" + url;
            }
            if (!url.match(/\/$/)) {
                url = url + "/";
            }

            url += upstreamId;

            return url;
        }

        function initializeManifestDetails(organization) {
            $scope.details = organization['owner_details'];
            $scope.upstream = $scope.details.upstreamConsumer;

            if (!_.isNull($scope.upstream)) {
                $scope.manifestLink = buildManifestLink($scope.upstream);
                $scope.manifestName = $scope.upstream.name || $scope.upstream.uuid;
            }
        }

        function getInitialTask() {
            return {pending: true};

        }

        $scope.uploadErrorMessages = [];
        $scope.progress = {uploading: false};
        $scope.uploadURL = 'katello/api/v2/organizations/' + CurrentOrganization + '/subscriptions/upload';
        $scope.organization = Organization.get({id: CurrentOrganization});

        $q.all([$scope.organization.$promise]).then(function () {
            $scope.panel.loading = false;
            initializeManifestDetails($scope.organization);
        });

        $scope.$on('$destroy', function () {
            $scope.unregisterSearch();
        });

        $scope.isTaskPending = function () {
            return $scope.task && $scope.task.pending;
        };

        $scope.unregisterSearch = function () {
            Task.unregisterSearch($scope.searchId);
            $scope.searchId = undefined;
        };

        $scope.handleTaskErrors = function (task, errorMessage) {
            var errorMessageWithDetails = errorMessage;
            if (task.result === 'error' || task.result === 'warning') {
                if (task.humanized.output && task.humanized.output.length > 0) {
                    errorMessageWithDetails += ' ' + task.humanized.output;
                }
                if (task.humanized.errors.length > 0) {
                    errorMessageWithDetails += ' ' + task.humanized.errors.join(' ');
                }
                Notification.setErrorMessage(errorMessageWithDetails);
                $scope.histories = Subscription.manifestHistory();
            }
        };

        $scope.updateTask = function (task) {
            $scope.task = task;

            if (!$scope.task.pending) {
                $scope.unregisterSearch();
                if ($scope.task.result === 'success') {
                    $scope.refreshOrganizationInfo();
                    Notification.setSuccessMessage(translate("Manifest successfully imported."));
                } else {
                    $scope.handleTaskErrors(task, translate("Error importing manifest."));
                }
            }
        };

        $scope.deleteManifest = function () {
            $scope.task = getInitialTask();
            $scope.taskStatusText = translate('Removing Manifest');
            Subscription.deleteManifest({}, function (returnData) {
                $scope.task = returnData;
                $scope.searchId = Task.registerSearch({'type': 'task', 'task_id': $scope.task.id}, $scope.deleteManifestTask);
            }, function (response) {
                $scope.saveError = true;
                $scope.errors = response.data.errors;
            });
        };

        $scope.deleteManifestTask = function (task) {
            $scope.task = task;
            if (!$scope.task.pending) {
                $scope.unregisterSearch();
                if ($scope.task.result === 'success') {
                    $scope.saveSuccess = true;
                    Notification.setSuccessMessage(translate("Manifest successfully deleted."));
                    $scope.refreshOrganizationInfo();
                } else {
                    $scope.handleTaskErrors(task, translate("Error deleting manifest."));
                }
            }
        };

        $scope.refreshOrganizationInfo = function () {
            $scope.organization = Organization.get({id: CurrentOrganization});
            $q.all([$scope.organization.$promise]).then(function () {
                initializeManifestDetails($scope.organization);
            });
            $scope.histories = Subscription.manifestHistory();
        };

        $scope.refreshManifest = function () {
            $scope.task = getInitialTask();
            $scope.taskStatusText = translate('Refreshing Manifest');
            Subscription.refreshManifest({}, function (returnData) {
                $scope.task = returnData;
                $scope.searchId = Task.registerSearch({'type': 'task', 'task_id': $scope.task.id}, $scope.refreshManifestTask);
            }, function (response) {
                $scope.saveError = true;
                $scope.errors = response.data.errors;
            });
        };

        $scope.refreshManifestTask = function (task) {
            $scope.task = task;
            if (!$scope.task.pending) {
                $scope.unregisterSearch();
                if ($scope.task.result === 'success') {
                    $scope.saveSuccess = true;
                    Notification.setSuccessMessage(translate("Manifest successfully refreshed."));
                    $scope.refreshOrganizationInfo();
                } else {
                    $scope.handleTaskErrors(task, translate("Error refreshing manifest."));
                }
            }
        };

        $scope.saveCdnUrl = function (organization) {
            var deferred;

            // @TODO hack needed to prevent upload of fields users, parent_name, and parent_id
            // http://projects.theforeman.org/issues/12894
            var whitelistedOrganizationObject = {},
                whitelist = [
                    "id",
                    "redhat_repository_url"
                ];

            angular.forEach(whitelist, function (key) {
                whitelistedOrganizationObject[key] = organization[key];
            });

            deferred = Organization.update(whitelistedOrganizationObject, function () {
                Notification.setSuccessMessage(translate('Repository URL updated'));
                $scope.refreshOrganizationInfo();
            }, function (response) {
                angular.forEach(response.data.error['full_messages'], function (message) {
                    Notification.setErrorMessage(translate("An error occurred saving the URL: ") + message);
                });
            });

            return deferred.$promise;
        };

        $scope.uploadManifest = function (content) {
            var returnData;
            if (content) {
                $scope.task = getInitialTask();
                $scope.taskStatusText = translate('Uploading Manifest');

                try {
                    returnData = angular.fromJson(angular.element(content).html());
                } catch (err) {
                    returnData = content;
                }

                if (!returnData) {
                    returnData = content;
                }

                if (returnData !== null && angular.isUndefined(returnData.errors)) {
                    $scope.task = returnData;
                    $scope.searchId = Task.registerSearch({'type': 'task', 'task_id': $scope.task.id}, $scope.updateTask);
                } else {
                    $scope.uploadErrorMessages = [translate('Error during upload: ') + returnData.displayMessage];
                }

                $scope.progress.uploading = false;
            }
        };

        $scope.uploadError = function (error, content) {
            if (angular.isString(content) && content.indexOf("Request Entity Too Large")) {
                error = translate('File too large.');
            } else {
                error = content;
            }
            $scope.uploadErrorMessages = [translate('Error during upload: ') + error];
            $scope.progress.uploading = false;
        };

        $scope.histories = Subscription.manifestHistory();

        $scope.showHistoryMoreLink = false;

        $scope.truncateHistories = function (histories) {
            var numToDisplay = 4;
            var result = [];
            angular.forEach(histories, function (history, index) {
                if (index < numToDisplay) {
                    result.push(history);
                }
            });
            return result;
        };

        $scope.isTruncated = function (subset, set) {
            return subset.length < set.length;
        };

        $scope.$watch('histories', function (changes) {
            changes.$promise.then(function (results) {
                $scope.statuses = $scope.truncateHistories(results);
                $scope.showHistoryMoreLink = $scope.isTruncated($scope.statuses, results);
            });
        });
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.subscriptions.controller:ManifestController
 *
 * @requires $scope
 * @requires translate
 * @requires ManifestHistoryService
 *
 * @description
 *   Controls the managment of manifests for use by sub-controllers.
 */

angular.module('Bastion.subscriptions').controller('ManifestController',
    ['$scope', function ($scope) {
        $scope.panel = {loading: true};
    }]
);
/**
 * @ngdoc directive
 * @name Bastion.subscriptions:subscriptionStartDate
 *
 * @description
 *   Directive for the subscription start date
 */

angular.module('Bastion.subscriptions').directive('subscriptionStartDate', function () {
    return {
        restrict: 'AE',
        scope: {
            subscription: '=subscriptionStartDate'
        },
        templateUrl: 'subscriptions/views/subscription-start-date.html',
        controller: ['$scope', 'translate', function ($scope, translate) {
            $scope.checkFutureDate = function (date) {
                return (Date.parse(date) > Date.parse(Date()) ? translate(' (future)') : '');
            };
        }]
    };
});
/**
 * @ngdoc directive
 * @name Bastion.subscriptions:subscriptionType
 *
 * @description
 *   Directive for the subscription type
 */

angular.module('Bastion.subscriptions').directive('subscriptionType', function () {
    return {
        restrict: 'AE',
        scope: {
            subscription: '=subscriptionType'
        },
        templateUrl: 'subscriptions/views/subscription-type.html'
    };
});
/**
 * @ngdoc filter
 * @name  Bastion.subscriptions.filter:subscriptionAttachAmountFilter.filter.js
 *
 * @requires translate
 *
 */

angular.module('Bastion.subscriptions').filter('subscriptionAttachAmountFilter',
    ['translate',
    function (translate) {
        return function (subscription) {
            var amount = subscription["quantity_attached"];
            return (!amount || amount < 1) ? translate("Automatic") : amount;
        };
    }]
);
/**
 * @ngdoc filter
 * @name  Bastion.subscriptions.filter:subscriptionConsumedFilter
 *
 * @description
 *
 *
 * @example
 *
 */

angular.module('Bastion.subscriptions').filter('subscriptionConsumedFilter',
    ['$filter', 'translate',
    function ($filter, translate) {
        return function (subscription) {
            var quantity = $filter('unlimitedFilter')(subscription.quantity);
            return translate('%(consumed)s out of %(quantity)s')
                .replace('%(consumed)s', subscription.consumed)
                .replace('%(quantity)s', quantity);
        };
    }]
);
/**
 * @ngdoc service
 * @name  Bastion.subscriptions.service:SubscriptionsHelper
 *
 * @description
 *   Helper service that contains functionality common amongst subscriptions.
 */

angular.module('Bastion.subscriptions').service('SubscriptionsHelper',
    function () {

        this.groupByProductName = function (rows) {
            var grouped,
                offset,
                subscription;

            grouped = {};
            for (offset = 0; offset < rows.length; offset += 1) {
                subscription = rows[offset];
                if (angular.isUndefined(grouped[subscription.name])) {
                    grouped[subscription.name] = [];
                }
                grouped[subscription.name].push(subscription);
            }

            return grouped;
        };

        this.getSelectedSubscriptionAmounts = function (table) {
            var selected,
                amount;

            selected = [];
            angular.forEach(table.getSelected(), function (subscription) {
                if (subscription['multi_entitlement']) {
                    amount = subscription.amount;
                    if (!amount) {
                        amount = 0;
                    }
                } else {
                    amount = 1;
                }
                selected.push({"id": subscription.id, "quantity": amount});
            });
            return selected;
        };

        this.getAmountSelectorValues = function (subscription) {
            var step, value, values;

            step = subscription['instance_multiplier'];
            if (!step || step < 1) {
                step = 1;
            }
            values = [];
            for (value = step; value < subscription.quantity && values.length < 5; value += step) {
                values.push(value);
            }
            values.push(subscription.quantity);
            return values;
        };

        this.getSelectedSubscriptions = function (table) {
            var selected;

            selected = [];
            angular.forEach(table.getSelected(), function (subscription) {
                selected.push({"id": subscription.id, "quantity": subscription.quantity_consumed});
            });
            return selected;
        };
    }
);
/**
 * @ngdoc object
 * @name  Bastion.subscriptions.controller:SubscriptionsController
 *
 * @requires $scope
 * @requires $q
 * @requires $location
 * @requires translate
 * @requires Nutupane
 * @requires Subscription
 * @requires Organization
 * @requires CurrentOrganization
 * @requires unlimitedFilterFilter
 *
 * @description
 *   Provides the functionality specific to Subscriptions for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.subscriptions').controller('SubscriptionsController',
    ['$scope', '$filter', '$location', 'translate', 'Nutupane', 'Subscription', 'Organization', 'CurrentOrganization', 'SubscriptionsHelper',
    function ($scope, $filter, $location, translate, Nutupane, Subscription, Organization, CurrentOrganization, SubscriptionsHelper) {
        var params, nutupane;

        params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'sort_order': 'ASC',
            'enabled': true,
            'paged': true
        };

        nutupane = new Nutupane(Subscription, params);
        $scope.controllerName = 'katello_subscriptions';
        nutupane.masterOnly = true;
        $scope.table = nutupane.table;
        $scope.refreshTable = nutupane.refresh;

        $scope.groupedSubscriptions = {};
        $scope.$watch('table.rows', function (rows) {
            $scope.groupedSubscriptions = SubscriptionsHelper.groupByProductName(rows);
        });

        $scope.hasManifest = false;
        Organization.get({id: CurrentOrganization}, function (organization) {
            $scope.hasManifest = organization['owner_details'] && organization['owner_details'].upstreamConsumer;
        });

        $scope.formatConsumed = function (subscription) {
            var quantity = $filter('unlimitedFilter')(subscription.quantity);

            return translate('%(consumed)s out of %(quantity)s').replace('%(consumed)s', subscription.consumed).replace('%(quantity)s', quantity);
        };

        $scope.formatInstanceBased = function (subscription) {
            if (angular.isUndefined(subscription['instance_multiplier']) || subscription['instance_multiplier'] === "" || subscription['instance_multiplier'] === 0) {
                return translate("No");
            }
            return translate("Yes");
        };

        $scope.redhatProvider = Organization.redhatProvider();
        $scope.subscriptions = Subscription.queryPaged();

        $scope.$on('$stateChangeSuccess', function () {
            $scope.subscriptions.$promise.then(function () {
                if ($scope.subscriptions.results.length < 1 && $scope.permitted('import_manifest')) {
                    $scope.transitionTo('subscriptions-manifest.import');
                }
            });
        });
    }]
);
/**
 * @ngdoc service
 * @name  Bastion.subscriptions.factory:Subscription
 *
 * @requires BastionResource
 * @requires CurrentOrganization
 *
 * @description
 *   Provides a BastionResource for a subscription or list of subscriptions
 */

angular.module('Bastion.subscriptions').factory('Subscription', ['BastionResource', 'CurrentOrganization',

    function (BastionResource, CurrentOrganization) {
        return BastionResource('katello/api/v2/organizations/:org/subscriptions/:id/:action',
            {org: CurrentOrganization, id: '@id'},
            {
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}},
                deleteManifest: {
                    method: 'POST',
                    url: 'katello/api/v2/organizations/:org/subscriptions/delete_manifest',
                    params: {'org': CurrentOrganization}
                },

                refreshManifest: {
                    method: 'PUT',
                    url: 'katello/api/v2/organizations/:org/subscriptions/refresh_manifest',
                    params: {'org': CurrentOrganization}
                },

                manifestHistory: {
                    method: 'GET',
                    url: 'katello/api/v2/organizations/:org/subscriptions/:action',
                    params: {action: 'manifest_history'},
                    isArray: true
                }
            });
    }]
);
/**
 * @ngdoc object
 * @name Bastion.subscriptions.config
 *
 * @requires $stateProvider
 *
 * @description
 *   Used for subscriptions level configuration such as setting up the ui state machine.
 */

angular.module('Bastion.subscriptions').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('subscriptions', {
        url: '/subscriptions',
        permission: 'view_subscriptions',
        template: '<div ui-view></div>',
        views: {
            '@': {
                controller: 'SubscriptionsController',
                templateUrl: 'subscriptions/views/subscriptions.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{ 'Subscriptions' | translate }}"
        }
    });

    $stateProvider.state('subscription', {
        abstract: true,
        url: '/subscriptions/:subscriptionId',
        permission: 'view_subscriptions',
        controller: 'SubscriptionDetailsController',
        templateUrl: 'subscriptions/details/views/subscription-details.html'
    })
    .state('subscription.info', {
        url: '',
        permission: 'view_subscriptions',
        templateUrl: 'subscriptions/details/views/subscription-info.html',
        ncyBreadcrumb: {
            label: "{{ subscription.name }}",
            parent: 'subscriptions'
        }
    })
    .state('subscription.products', {
        url: '/products',
        permission: 'view_subscriptions',
        controller: 'SubscriptionProductsController',
        templateUrl: 'subscriptions/details/views/subscription-products.html',
        ncyBreadcrumb: {
            label: "{{ 'Product Content' | translate }}",
            parent: 'subscription.info'
        }
    })
    .state('subscription.activation-keys', {
        url: '/activation-keys',
        permission: 'view_subscriptions',
        controller: 'SubscriptionActivationKeysController',
        templateUrl: 'subscriptions/details/views/subscription-activation-keys.html',
        ncyBreadcrumb: {
            label: "{{ 'Activation Keys' | translate }}",
            parent: 'subscription.info'
        }
    })
    .state('subscription.content-hosts', {
        url: '/content-hosts',
        permission: 'view_subscriptions',
        controller: 'SubscriptionContentHostsController',
        templateUrl: 'subscriptions/details/views/subscription-content-hosts.html',
        ncyBreadcrumb: {
            label: "{{ 'Content Hosts' | translate }}",
            parent: 'subscription.info'
        }
    });

    $stateProvider.state('subscriptions-manifest', {
        abstract: true,
        url: '/subscriptions/manifest',
        permission: 'import_manifest',
        controller: 'ManifestController',
        templateUrl: 'subscriptions/manifest/views/manifest.html'
    })
    .state('subscriptions-manifest.details', {
        url: '/details',
        permission: 'import_manifest',
        controller: 'ManifestDetailsController',
        templateUrl: 'subscriptions/manifest/views/manifest-details.html',
        ncyBreadcrumb: {
            label: "{{ 'Manifest Details' | translate }}",
            parent: 'subscriptions'
        }
    })
    .state('subscriptions-manifest.import', {
        url: '/import',
        permission: 'import_manifest',
        controller: 'ManifestImportController',
        templateUrl: 'subscriptions/manifest/views/manifest-import.html',
        ncyBreadcrumb: {
            label: "{{ 'Import Manifest' | translate }}",
            parent: 'subscriptions'
        }
    })
    .state('subscriptions-manifest.history', {
        url: '/history',
        permission: 'import_manifest',
        controller: 'ManifestHistoryController',
        templateUrl: 'subscriptions/manifest/views/manifest-import-history.html',
        ncyBreadcrumb: {
            label: "{{ 'Manifest History' | translate }}",
            parent: 'subscriptions'
        }
    });
}]);
/**
 * @ngdoc module
 * @name  Bastion.content-views.versions
 *
 * @description
 *   Module for content view version related functionality.
 */

angular.module('Bastion.content-views.versions', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.components',
    'Bastion.repositories',
    'Bastion.packages',
    'Bastion.errata',
    'Bastion.package-groups',
    'Bastion.puppet-modules',
    'Bastion.ostree-branches',
    'Bastion.files',
    'Bastion.debs'
]);
(function () {
    'use strict';

    /**
     * @ngdoc controller
     * @name  Bastion.content-views.versions.controller:ContentViewVersionContent
     *
     * @description
     *   Handles fetching content view version content and populating Nutupane based on the current
     *   ui-router state.
     */
    function ContentViewVersionContentController($scope, translate, Nutupane, Package, Erratum,
                                                 PackageGroup, PuppetModule, OstreeBranch, Deb, Repository,
                                                 ContentViewVersion) {
        var nutupane, contentTypes, currentState, params;

        currentState = $scope.$state.current.name.split('.').pop();

        contentTypes = {
            'docker': {
                type: Repository,
                params: {
                    'content_type': "docker",
                    'content_view_version_id': $scope.$stateParams.versionId

                }
            },
            'yum': {
                type: Repository,
                params: {
                    'content_type': "yum",
                    'content_view_version_id': $scope.$stateParams.versionId,
                    library: true
                }
            },
            'packages': {
                type: Package
            },
            'package-groups': {
                type: PackageGroup,
                params: {
                    'sort_by': 'name',
                    'sort_order': 'ASC',
                    'content_view_version_id': $scope.$stateParams.versionId
                }
            },
            'errata': {
                type: Erratum
            },
            'puppet-modules': {
                type: PuppetModule
            },
            'ostree-branches': {
                type: OstreeBranch,
                params: {
                    'content_type': "ostree",
                    'content_view_version_id': $scope.$stateParams.versionId,
                    'sort_by': 'version_date',
                    'sort_order': 'DESC'
                }
            },
            'components': {
                type: ContentViewVersion,
                params: {
                    'composite_version_id': $scope.$stateParams.versionId
                }
            },
            'file': {
                type: Repository,
                params: {
                    'content_type': 'file',
                    'content_view_version_id': $scope.$stateParams.versionId,
                    library: true
                }
            },
            'apt': {
                type: Repository,
                params: {
                    'content_type': 'deb',
                    'content_view_version_id': $scope.$stateParams.versionId,
                    library: true
                }
            },
            'deb': {
                type: Deb
            }
        };

        params = contentTypes[currentState].params || {'content_view_version_id': $scope.$stateParams.versionId};
        nutupane = new Nutupane(contentTypes[currentState].type, params, 'queryPaged');
        nutupane.masterOnly = true;

        $scope.nutupane = nutupane;
        $scope.table = nutupane.table;

        $scope.repository = {name: translate('All Repositories'), id: 'all'};
        $scope.repositories = [];

        $scope.version.$promise.then(function (version) {
            $scope.repositories = version.repositories;
            $scope.repositories.unshift($scope.repository);
        });

        $scope.$watch('repository', function (repository) {
            var nutupaneParams = nutupane.getParams();

            if (repository.id === 'all') {
                nutupaneParams['repository_id'] = null;
                nutupane.setParams(nutupaneParams);
            } else {
                nutupaneParams['repository_id'] = repository.id;
                nutupane.setParams(nutupaneParams);
            }

            nutupane.refresh();
        });
    }

    angular
        .module('Bastion.content-views.versions')
        .controller('ContentViewVersionContentController', ContentViewVersionContentController);

    ContentViewVersionContentController.$inject = ['$scope', 'translate', 'Nutupane', 'Package', 'Erratum',
                                                   'PackageGroup', 'PuppetModule', 'OstreeBranch', 'Deb', 'Repository', 'ContentViewVersion'];

})();
(function () {
    'use strict';

    /**
     * @ngdoc controller
     * @name  Bastion.content-views.versions.controller:ContentViewVersion
     *
     * @description
     *   Handles fetching of a content view version based on the route ID and putting it
     *   on the scope.
     */
    function ContentViewVersionController($scope, ContentViewVersion) {

        $scope.version = ContentViewVersion.get({id: $scope.$stateParams.versionId});

        $scope.hasRepositories = function (version, type) {
            var found;

            found = _.find(version.repositories, function (repository) {
                return repository['content_type'] === type;
            });

            return found;
        };

        $scope.hasErrata = function (version) {
            var found = false;

            if (version['errata_counts'] &&
                version['errata_counts'].total &&
                version['errata_counts'].total !== 0) {
                return true;
            }
            return found;
        };

    }

    angular
        .module('Bastion.content-views.versions')
        .controller('ContentViewVersionController', ContentViewVersionController);

    ContentViewVersionController.$inject = ['$scope', 'ContentViewVersion'];

})();
/**
 * @ngdoc service
 * @name  Bastion.content-views.factory:ContentViewVersion
 *
 * @requires BastionResource
 * @requires CurrentOrganization
 *
 * @description
 *   Provides a BastionResource for interacting with Content View Versions.
 */

angular.module('Bastion.content-views.versions').factory('ContentViewVersion',
    ['BastionResource', function (BastionResource) {

        return BastionResource('katello/api/v2/content_view_versions/:id/:action',
            {id: '@id'},
            {
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}},
                update: {method: 'PUT'},
                incrementalUpdate: {method: 'POST', params: {action: 'incremental_update'}},
                promote: {method: 'POST', params: {action: 'promote'}},
                republishRepositories: {method: 'PUT', params: {action: 'republish_repositories'}}
            }
        );

    }]
);
/**
 * @ngdoc module
 * @name  Bastion.content-views
 *
 * @description
 *   Module for content view related functionality.
 */

angular.module('Bastion.content-views', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.utils',
    'Bastion.content-views.versions',
    'Bastion.components',
    'Bastion.packages'
]);
/**
 * @ngdoc service
 * @name  Bastion.content-views.factory:ContentView
 *
 * @requires BastionResource
 * @requires translate
 * @requires CurrentOrganization
 *
 * @description
 *   Provides a BastionResource for interacting with environments.
 */

angular.module('Bastion.content-views').factory('ContentView',
    ['BastionResource', 'translate', 'CurrentOrganization',
    function (BastionResource, translate, CurrentOrganization) {

        return BastionResource('katello/api/v2/content_views/:id/:action',
            {id: '@id', 'organization_id': CurrentOrganization},
            {
                copy: {method: 'POST', params: {action: 'copy'}},
                update: {method: 'PUT'},
                publish: {method: 'POST', params: {action: 'publish'}},
                removeAssociations: {method: 'PUT', params: {action: 'remove'}},
                versions: {method: 'GET', isArray: false, params: {action: 'content_view_versions'}},
                conflictingVersions: {method: 'GET', isArray: true, params: {action: 'content_view_versions'},
                    transformResponse: function (data) {
                        var response = angular.fromJson(data);
                        return _.reject(response.results, function (version) {
                            return version.environments.length === 0;
                        });
                    }
                },
                contentViewComponents: {method: 'GET', transformResponse: function (data) {
                    var contentView = angular.fromJson(data);
                    return {results: contentView.content_view_components};
                }},
                availablePuppetModules: {method: 'GET', params: {action: 'available_puppet_modules'},
                    transformResponse: function (data) {
                        return angular.fromJson(data);
                    }
                },
                availablePuppetModuleNames: {method: 'GET', params: {action: 'available_puppet_module_names'}},
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}}
            }
        );

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewsController
 *
 * @requires $scope
 * @requires Nutupane
 * @requires ContentView
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality specific to ContentViews for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.content-views').controller('ContentViewsController',
    ['$scope', 'Nutupane', 'ContentView', 'CurrentOrganization',
    function ($scope, Nutupane, ContentView, CurrentOrganization) {

        var nutupane = new Nutupane(ContentView, {
            'nondefault': true,
            'organization_id': CurrentOrganization,
            'sort_by': 'name',
            'sort_order': 'ASC'
        });
        nutupane.masterOnly = true;
        $scope.controllerName = 'katello_content_views';

        $scope.table = nutupane.table;
    }]
);
/**
 * @ngdoc object
 * @name Bastion.content-views.config
 *
 * @requires $stateProvider
 *
 * @description
 *   State routes defined for the content views module.
 */

angular.module('Bastion.content-views').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('content-views', {
        url: '/content_views',
        permission: 'view_content_views',
        views: {
            '@': {
                controller: 'ContentViewsController',
                templateUrl: 'content-views/views/content-views.html'
            }
        },
        ncyBreadcrumb: {
            label: '{{ "Content Views" | translate }}'
        }
    })
    .state('content-views.new', {
        url: '/new',
        permission: 'create_content_views',
        views: {
            '@': {
                controller: 'NewContentViewController',
                templateUrl: 'content-views/new/views/content-view-new.html'
            }
        },
        ncyBreadcrumb: {
            label: '{{ "New Content View" | translate }}',
            parent: 'content-views'
        }
    });

    $stateProvider.state('content-view', {
        abstract: true,
        url: '/content_views/:contentViewId',
        permission: 'view_content_views',
        controller: 'ContentViewDetailsController',
        templateUrl: 'content-views/details/views/content-view-details.html'
    })

    .state('content-view.versions', {
        url: '/versions',
        permission: 'view_content_views',
        controller: 'ContentViewVersionsController',
        templateUrl: 'content-views/details/views/content-view-versions.html',
        ncyBreadcrumb: {
            label: '{{ "Versions" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.version', {
        abstract: true,
        url: '/versions/:versionId',
        controller: 'ContentViewVersionController',
        templateUrl: 'content-views/versions/views/content-view-version.html'
    })
    .state('content-view.version.details', {
        url: '',
        permission: 'view_content_views',
        templateUrl: 'content-views/versions/views/content-view-version-details.html',
        ncyBreadcrumb: {
            label: '{{ version.name }}',
            parent: 'content-view.versions'
        }
    })
    .state('content-view.version.components', {
        url: '/components',
        permission: 'view_content_views',
        controller: 'ContentViewVersionContentController',
        templateUrl: 'content-views/versions/views/content-view-version-components.html',
        ncyBreadcrumb: {
            label: '{{ "Components" | translate }}',
            parent: 'content-view.version.details'
        }
    })
    .state('content-view.version.yum', {
        url: '/yum',
        permission: 'view_content_views',
        controller: 'ContentViewVersionContentController',
        templateUrl: 'content-views/versions/views/content-view-version-yum.html',
        ncyBreadcrumb: {
            label: '{{ "Yum Content" | translate }}',
            parent: 'content-view.version.details'
        }
    })
    .state('content-view.version.docker', {
        url: '/docker',
        permission: 'view_content_views',
        controller: 'ContentViewVersionContentController',
        templateUrl: 'content-views/versions/views/content-view-version-docker.html',
        ncyBreadcrumb: {
            label: '{{ "Docker" | translate }}',
            parent: 'content-view.version.details'
        }
    })
    .state('content-view.version.packages', {
        url: '/packages',
        permission: 'view_content_views',
        controller: 'ContentViewVersionContentController',
        templateUrl: 'content-views/versions/views/content-view-version-packages.html',
        ncyBreadcrumb: {
            label: '{{ "Packages" | translate }}',
            parent: 'content-view.version.details'
        }
    })
    .state('content-view.version.package-groups', {
        url: '/package_groups',
        permission: 'view_content_views',
        controller: 'ContentViewVersionContentController',
        templateUrl: 'content-views/versions/views/content-view-version-package-groups.html',
        ncyBreadcrumb: {
            label: '{{ "Package Groups" | translate }}',
            parent: 'content-view.version.details'
        }
    })
    .state('content-view.version.errata', {
        url: '/errata',
        permission: 'view_content_views',
        controller: 'ContentViewVersionContentController',
        templateUrl: 'content-views/versions/views/content-view-version-errata.html',
        ncyBreadcrumb: {
            label: '{{ "Errata" | translate }}',
            parent: 'content-view.version.details'
        }
    })
    .state('content-view.version.puppet-modules', {
        url: '/puppet_modules',
        permission: 'view_content_views',
        controller: 'ContentViewVersionContentController',
        templateUrl: 'content-views/versions/views/content-view-version-puppet-modules.html',
        ncyBreadcrumb: {
            label: '{{ "Puppet Modules" | translate }}',
            parent: 'content-view.version.details'
        }
    })
    .state('content-view.version.ostree-branches', {
        url: '/ostree_branches',
        permission: 'view_content_views',
        controller: 'ContentViewVersionContentController',
        templateUrl: 'content-views/versions/views/content-view-version-ostree-branches.html',
        ncyBreadcrumb: {
            label: '{{ "OSTree Branches" | translate }}',
            parent: 'content-view.version.details'
        }
    })
    .state('content-view.version.file', {
        url: '/file',
        permission: 'view_content_views',
        controller: 'ContentViewVersionContentController',
        templateUrl: 'content-views/versions/views/content-view-version-file.html',
        ncyBreadcrumb: {
            label: '{{ "File" | translate }}',
            parent: 'content-view.version.details'
        }
    })
    .state('content-view.version.apt', {
        url: '/apt',
        permission: 'view_content_views',
        controller: 'ContentViewVersionContentController',
        templateUrl: 'content-views/versions/views/content-view-version-apt.html',
        ncyBreadcrumb: {
            label: '{{ "Apt Repositories" | translate }}',
            parent: 'content-view.version.details'
        }
    })
    .state('content-view.version.deb', {
        url: '/deb',
        permission: 'view_content_views',
        controller: 'ContentViewVersionContentController',
        templateUrl: 'content-views/versions/views/content-view-version-deb.html',
        ncyBreadcrumb: {
            label: '{{ "Deb" | translate }}',
            parent: 'content-view.version.details'
        }
    });

    $stateProvider.state('content-view.promotion', {
        url: '/versions/:versionId/promotion',
        permission: 'promote_or_remove_content_views',
        controller: 'ContentViewPromotionController',
        templateUrl: 'content-views/details/views/content-view-promotion.html',
        ncyBreadcrumb: {
            label: '{{ "Promotion" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.version-deletion', {
        abstract: true,
        controller: 'ContentViewVersionDeletionController',
        template: '<div ui-view></div>'
    })
    .state('content-view.version-deletion.environments', {
        url: '/versions/:versionId/delete/environments',
        permission: 'promote_or_remove_content_views',
        controller: 'ContentViewVersionDeletionEnvironmentsController',
        templateUrl: 'content-views/deletion/views/version-deletion-environments.html',
        ncyBreadcrumb: {
            label: '{{ "Deletion" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.version-deletion.content-hosts', {
        url: '/versions/:versionId/delete/content-hosts',
        permission: 'promote_or_remove_content_views',
        controller: 'ContentViewVersionDeletionContentHostsController',
        templateUrl: 'content-views/deletion/views/version-deletion-content-hosts.html',
        ncyBreadcrumb: {
            label: '{{ "Deletion" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.version-deletion.activation-keys', {
        url: '/versions/:versionId/delete/activation-keys',
        permission: 'promote_or_remove_content_views',
        controller: 'ContentViewVersionDeletionActivationKeysController',
        templateUrl: 'content-views/deletion/views/version-deletion-activation-keys.html',
        ncyBreadcrumb: {
            label: '{{ "Deletion" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.version-deletion.confirm', {
        url: '/versions/:versionId/delete/confirm',
        permission: 'promote_or_remove_content_views',
        controller: 'ContentViewVersionDeletionConfirmController',
        templateUrl: 'content-views/deletion/views/version-deletion-confirm.html',
        ncyBreadcrumb: {
            label: '{{ "Deletion" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.deletion', {
        url: '/delete',
        permission: 'promote_or_remove_content_views',
        controller: 'ContentViewDeletionController',
        templateUrl: 'content-views/deletion/views/content-view-deletion.html',
        ncyBreadcrumb: {
            label: '{{ "Deletion" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.repositories', {
        abstract: true,
        template: '<div ui-view></div>'
    })
    .state('content-view.repositories.yum', {
        abstract: true,
        template: '<div ui-view></div>'
    })
    .state('content-view.repositories.yum.list', {
        url: '/repositories/yum',
        permission: 'view_content_views',
        controller: 'ContentViewRepositoriesListController',
        templateUrl: 'content-views/details/views/content-view-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "Yum Repositories" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.repositories.yum.available', {
        url: '/repositories/yum/available',
        permission: 'view_content_views',
        controller: 'ContentViewAvailableRepositoriesController',
        templateUrl: 'content-views/details/views/content-view-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "Add Yum Repositories" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.repositories.file', {
        abstract: true,
        template: '<div ui-view></div>'
    })
    .state('content-view.repositories.file.list', {
        url: '/repositories/file',
        permission: 'view_content_views',
        controller: 'ContentViewFileRepositoriesListController',
        templateUrl: 'content-views/details/views/content-view-file-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "File Repositories" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.repositories.file.available', {
        url: '/repositories/file/available',
        permission: 'view_content_views',
        controller: 'ContentViewAvailableFileRepositoriesController',
        templateUrl: 'content-views/details/views/content-view-file-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "Add File Repositories" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.repositories.deb', {
        abstract: true,
        template: '<div ui-view></div>'
    })
    .state('content-view.repositories.deb.list', {
        url: '/repositories/deb',
        permission: 'view_content_views',
        controller: 'ContentViewDebRepositoriesListController',
        templateUrl: 'content-views/details/views/content-view-deb-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "Deb Repositories" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.repositories.deb.available', {
        url: '/repositories/deb/available',
        permission: 'view_content_views',
        controller: 'ContentViewAvailableDebRepositoriesController',
        templateUrl: 'content-views/details/views/content-view-deb-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "Add Deb Repositories" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.repositories.docker', {
        abstract: true,
        template: '<div ui-view></div>'
    })
    .state('content-view.repositories.docker.list', {
        url: '/repositories/docker',
        permission: 'view_content_views',
        controller: 'ContentViewDockerRepositoriesListController',
        templateUrl: 'content-views/details/views/content-view-docker-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "Docker Repositories" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.repositories.docker.available', {
        url: '/repositories/docker/available',
        permission: 'view_content_views',
        controller: 'ContentViewAvailableDockerRepositoriesController',
        templateUrl: 'content-views/details/views/content-view-docker-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "Add Docker Repositories" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.repositories.ostree', {
        abstract: true,
        template: '<div ui-view></div>'
    })
    .state('content-view.repositories.ostree.list', {
        url: '/repositories/ostree',
        permission: 'view_content_views',
        controller: 'ContentViewOstreeRepositoriesListController',
        templateUrl: 'content-views/details/views/content-view-ostree-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "OSTree Repositories" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.repositories.ostree.available', {
        url: '/repositories/ostree/available',
        permission: 'view_content_views',
        controller: 'ContentViewAvailableOstreeRepositoriesController',
        templateUrl: 'content-views/details/views/content-view-ostree-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "Add OSTree Repositories" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.components', {
        abstract: true,
        template: '<div ui-view></div>'
    })
    .state('content-view.components.composite-content-views', {
        abstract: true,
        templateUrl: 'content-views/details/components/views/content-view-composite.html'
    })
    .state('content-view.components.composite-content-views.list', {
        url: '/content-views',
        permission: 'view_content_views',
        controller: 'ContentViewCompositeContentViewsListController',
        templateUrl: 'content-views/details/components/views/content-view-composite-content-views-list.html',
        ncyBreadcrumb: {
            label: '{{ "Content Views" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.components.composite-content-views.available', {
        url: '/content-views/available',
        permission: 'edit_content_views',
        controller: 'ContentViewCompositeAvailableContentViewsController',
        templateUrl: 'content-views/details/components/views/content-view-composite-available-content-views.html',
        ncyBreadcrumb: {
            label: '{{ "Add Content Views" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.puppet-modules', {
        abstract: true,
        template: '<div ui-view></div>'
    })
    .state('content-view.puppet-modules.list', {
        url: '/puppet_modules',
        permission: 'view_content_views',
        controller: 'ContentViewPuppetModulesController',
        templateUrl: 'content-views/details/puppet-modules/views/content-view-puppet-modules.html',
        ncyBreadcrumb: {
            label: '{{ "Puppet Modules" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.puppet-modules.names', {
        url: '/puppet_modules/names',
        permission: 'edit_content_views',
        controller: 'ContentViewPuppetModuleNamesController',
        templateUrl: 'content-views/details/puppet-modules/views/content-view-puppet-module-names.html',
        ncyBreadcrumb: {
            label: '{{ "Add Puppet Module" | translate }}',
            parent: 'content-view.puppet-modules.list'
        }
    })
    .state('content-view.puppet-modules.versions', {
        url: '/puppet_modules/:moduleName/versions',
        permission: 'edit_content_views',
        controller: 'ContentViewPuppetModuleVersionsController',
        templateUrl: 'content-views/details/puppet-modules/views/content-view-puppet-module-versions.html',
        ncyBreadcrumb: {
            label: '{{ "Version for Module:" | translate }} {{ }}',
            parent: 'content-view.puppet-modules.names'
        }
    })
    // Necessary until ui-router supports optional parameters, see https://github.com/angular-ui/ui-router/issues/108
    .state('content-view.puppet-modules.versionsForModule', {
        collapsed: true,
        url: '/puppet_modules/:moduleName/versions/:moduleId',
        permission: 'edit_content_views',
        controller: 'ContentViewPuppetModuleVersionsController',
        templateUrl: 'content-views/details/puppet-modules/views/content-view-puppet-module-versions.html',
        ncyBreadcrumb: {
            label: '{{ "Versions for Module" | translate }}',
            parent: 'content-view.puppet-modules.versions'
        }
    })
    .state('content-view.history', {
        url: '/history',
        permission: 'view_content_views',
        controller: 'ContentViewHistoryController',
        templateUrl: 'content-views/details/histories/views/content-view-history.html',
        ncyBreadcrumb: {
            label: '{{ "History" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.info', {
        url: '',
        permission: 'view_content_views',
        templateUrl: 'content-views/details/views/content-view-info.html',
        ncyBreadcrumb: {
            label: '{{ contentView.name }}',
            parent: 'content-views'
        }
    })
    .state('content-view.copy', {
        url: '/copy',
        permission: 'create_content_views',
        templateUrl: 'content-views/details/views/content-view-copy.html',
        ncyBreadcrumb: {
            label: '{{ "Copy" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.publish', {
        url: '/publish',
        permission: 'publish_content_views',
        controller: 'ContentViewPublishController',
        templateUrl: 'content-views/details/views/content-view-publish.html',
        ncyBreadcrumb: {
            label: '{{ "Publish" | translate }}',
            parent: 'content-view.info'
        }
    })

    .state('content-view.yum', {
        url: '/repositories/yum',
        abstract: true,
        template: '<div ui-view></div>'
    })

    .state('content-view.yum.filters', {
        url: '/filters',
        permission: 'view_content_views',
        controller: 'FiltersController',
        templateUrl: 'content-views/details/filters/views/filters.html',
        ncyBreadcrumb: {
            label: '{{ "Yum Filters" | translate }}',
            parent: 'content-view.info'
        }

    })
    .state('content-view.yum.filters.new', {
        url: '/new',
        permission: 'edit_content_views',
        views: {
            '@content-view': {
                controller: 'NewFilterController',
                templateUrl: 'content-views/details/filters/views/new-filter.html'
            }
        },
        ncyBreadcrumb: {
            label: '{{ "Create Yum Filter" | translate }}',
            parent: 'content-view.yum.filters'
        }
    });

    $stateProvider.state('content-view.yum.filter', {
        abstract: true,
        controller: 'FilterDetailsController',
        templateUrl: 'content-views/details/filters/views/filter-details.html'
    })

    .state('content-view.yum.filter.rpm', {
        abstract: true,
        url: '/filters/:filterId/package',
        permission: 'view_content_views',
        controller: 'PackageFilterController',
        templateUrl: 'content-views/details/filters/views/package-filter.html'
    })
    .state('content-view.yum.filter.rpm.details', {
        url: '/details',
        permission: 'view_content_views',
        controller: 'PackageFilterController',
        templateUrl: 'content-views/details/filters/views/package-filter-details.html',
        ncyBreadcrumb: {
            label: '{{ filter.name }}',
            parent: 'content-view.yum.filters'
        }
    })
    .state('content-view.yum.filter.rpm.edit', {
        url: '/edit',
        controller: 'FilterEditController',
        permission: 'edit_content_views',
        templateUrl: 'content-views/details/filters/views/edit-filter.html',
        ncyBreadcrumb: {
            label: '{{ "Edit" | translate}} {{ filter.name }}',
            parent: 'content-view.yum.filters'
        }
    })
    .state('content-view.yum.filter.rpm.repositories', {
        url: '/repositories',
        permission: 'view_content_views',
        controller: 'FilterRepositoriesController',
        templateUrl: 'content-views/details/filters/views/filter-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "Repositories" | translate }}',
            parent: 'content-view.yum.filter.rpm.edit'
        }
    })

    .state('content-view.yum.filter.package_group', {
        abstract: true,
        url: '/filters/:filterId/package-group',
        templateUrl: 'content-views/details/filters/views/package-group-filter.html'
    })
    .state('content-view.yum.filter.package_group.edit', {
        url: '/edit',
        controller: 'FilterEditController',
        permission: 'edit_content_views',
        templateUrl: 'content-views/details/filters/views/edit-filter.html',
        ncyBreadcrumb: {
            label: '{{ "Edit" | translate }} {{ filter.name }}',
            parent: 'content-view.yum.filters'
        }
    })
    .state('content-view.yum.filter.package_group.list', {
        url: '/list',
        permission: 'view_content_views',
        controller: 'PackageGroupFilterListController',
        templateUrl: 'content-views/details/filters/views/package-group-filter-details.html',
        ncyBreadcrumb: {
            label: '{{ "Package Groups" | translate }}',
            parent: 'content-view.yum.filter.package_group.edit'
        }
    })
    .state('content-view.yum.filter.package_group.available', {
        url: '/available',
        permission: 'edit_content_views',
        controller: 'AvailablePackageGroupFilterController',
        templateUrl: 'content-views/details/filters/views/package-group-filter-details.html',
        ncyBreadcrumb: {
            label: '{{ "Add Package Groups" | translate }}',
            parent: 'content-view.yum.filter.package_group.edit'
        }
    })
    .state('content-view.yum.filter.package_group.repositories', {
        url: '/repositories',
        permission: 'view_content_views',
        controller: 'FilterRepositoriesController',
        templateUrl: 'content-views/details/filters/views/filter-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "Repositories | translate }}',
            parent: 'content-view.yum.filter.package_group.edit'
        }
    })

    .state('content-view.yum.filter.erratum', {
        abstract: true,
        url: '/filters/:filterId/errata',
        controller: 'ErrataFilterController',
        templateUrl: 'content-views/details/filters/views/errata-filter.html'
    })
    .state('content-view.yum.filter.erratum.edit', {
        url: '/edit',
        controller: 'FilterEditController',
        permission: 'edit_content_views',
        templateUrl: 'content-views/details/filters/views/edit-filter.html',
        ncyBreadcrumb: {
            label: '{{ "Edit" | translate }} {{ filter.name }}',
            parent: 'content-view.yum.filters'
        }
    })
    .state('content-view.yum.filter.erratum.list', {
        url: '/list',
        permission: 'view_content_views',
        controller: 'ErrataFilterListController',
        templateUrl: 'content-views/details/filters/views/errata-filter-details.html',
        ncyBreadcrumb: {
            label: '{{ "Erratum" | translate }}',
            parent: 'content-view.yum.filter.erratum.edit'
        }
    })
    .state('content-view.yum.filter.erratum.available', {
        url: '/available',
        permission: 'edit_content_views',
        controller: 'AvailableErrataFilterController',
        templateUrl: 'content-views/details/filters/views/errata-filter-details.html',
        ncyBreadcrumb: {
            label: '{{ "Add Erratum" | translate }}',
            parent: 'content-view.yum.filter.erratum.edit'
        }
    })
    .state('content-view.yum.filter.erratum.dateType', {
        url: '/date_type',
        permission: 'view_content_views',
        controller: 'DateTypeErrataFilterController',
        templateUrl: 'content-views/details/filters/views/date-type-errata-filter.html',
        ncyBreadcrumb: {
            label: '{{ filter.name }}',
            parent: 'content-view.yum.filters'
        }

    })
    .state('content-view.yum.filter.erratum.repositories', {
        url: '/repositories',
        permission: 'view_content_views',
        controller: 'FilterRepositoriesController',
        templateUrl: 'content-views/details/filters/views/filter-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "Repositories" | translate }}',
            parent: 'content-view.yum.filter.erratum.edit'
        }
    })

    .state('content-view.docker', {
        url: '/repositories/docker',
        abstract: true,
        template: '<div ui-view></div>'
    })

    .state('content-view.docker.filters', {
        url: '/filters',
        permission: 'view_content_views',
        controller: 'FiltersController',
        templateUrl: 'content-views/details/filters/views/filters.html',
        ncyBreadcrumb: {
            label: '{{ "Docker Filters" | translate }}',
            parent: 'content-view.info'
        }

    })
    .state('content-view.docker.filters.new', {
        url: '/new',
        permission: 'edit_content_views',
        views: {
            '@content-view': {
                controller: 'NewFilterController',
                templateUrl: 'content-views/details/filters/views/new-filter.html'
            }
        },
        ncyBreadcrumb: {
            label: '{{ "Create Docker Filter" | translate }}',
            parent: 'content-view.docker.filters'
        }
    });

    $stateProvider.state('content-view.docker.filter', {
        abstract: true,
        controller: 'FilterDetailsController',
        templateUrl: 'content-views/details/filters/views/filter-details.html'
    })

    .state('content-view.docker.filter.tag', {
        abstract: true,
        url: '/filters/:filterId/docker',
        permission: 'view_content_views',
        controller: 'DockerTagFilterController',
        templateUrl: 'content-views/details/filters/views/docker-filter.html'
    })
    .state('content-view.docker.filter.tag.details', {
        url: '/details',
        permission: 'view_content_views',
        controller: 'DockerTagFilterController',
        templateUrl: 'content-views/details/filters/views/docker-tag-filter-details.html',
        ncyBreadcrumb: {
            label: '{{ filter.name }}',
            parent: 'content-view.docker.filters'
        }
    })
    .state('content-view.docker.filter.tag.edit', {
        url: '/edit',
        controller: 'FilterEditController',
        permission: 'edit_content_views',
        templateUrl: 'content-views/details/filters/views/edit-filter.html',
        ncyBreadcrumb: {
            label: '{{ "Edit" | translate}} {{ filter.name }}',
            parent: 'content-view.docker.filters'
        }
    })
    .state('content-view.docker.filter.tag.repositories', {
        url: '/repositories',
        permission: 'view_content_views',
        controller: 'FilterRepositoriesController',
        templateUrl: 'content-views/details/filters/views/filter-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "Repositories" | translate }}',
            parent: 'content-view.docker.filter.tag.edit'
        }
    })

    .state('content-view.tasks', {
        abstract: true,
        template: '<div ui-view></div>'
    })
    .state('content-view.tasks.index', {
        url: '/tasks',
        permission: 'view_content_views',
        templateUrl: 'content-views/details/views/content-view-details-tasks.html',
        ncyBreadcrumb: {
            label: '{{ "Tasks" | translate }}',
            parent: 'content-view.info'
        }
    })
    .state('content-view.task', {
        url: '/tasks/:taskId',
        permission: 'view_content_views',
        controller: 'TaskDetailsController',
        templateUrl: 'tasks/views/task-details.html',
        ncyBreadcrumb: {
            label: '{{ task.id }}',
            parent: 'content-view.tasks'
        }
    });
}]);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewDeletionController
 *
 * @requires $scope
 * @requires ContentView
 * @requires Notification
 *
 * @description
 *   Provides the functionality for deleting Content Views
 */

angular.module('Bastion.content-views').controller('ContentViewDeletionController',
    ['$scope', 'ContentView', 'Notification', function ($scope, ContentView, Notification) {

        function success() {
            $scope.transitionTo('content-views');
            $scope.working = false;
        }

        function failure(response) {
            Notification.setErrorMessage(response.data.displayMessage);
            $scope.working = false;
        }

        $scope.delete = function () {
            $scope.working = true;
            ContentView.remove({id: $scope.contentView.id}, success, failure);
        };

        $scope.environmentNames = function (version) {
            return _.map($scope.readableEnvironments(version), 'name');
        };

        $scope.readableEnvironments = function (version) {
            return _.reject(version.environments, function (env) {
                return !env.permissions.readable;
            });
        };

        $scope.conflictingVersions = ContentView.conflictingVersions({id: $scope.$stateParams.contentViewId});
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewVersionDeletionActivationKeys
 *
 * @requires $scope
 * @requires $location
 * @requires Organization
 * @requires CurrentOrganization
 * @requires Nutupane
 * @requires ActivationKey
 *
 * @description
 *   Provides the functionality for selecting an environment and content view to move activation
 *   keys to.
 */

angular.module('Bastion.content-views').controller('ContentViewVersionDeletionActivationKeysController',
    ['$scope', '$location', 'Organization', 'CurrentOrganization', 'Nutupane', 'ActivationKey',
    function ($scope, $location, Organization, CurrentOrganization, Nutupane, ActivationKey) {
        var params, nutupane;

        $scope.validateEnvironmentSelection();
        params = {
            'organization_id': CurrentOrganization,
            'content_view_id': $scope.contentView.id,
            'sort_by': 'name',
            'sort_order': 'ASC'
        };
        nutupane = new Nutupane(ActivationKey, params);
        $scope.controllerName = 'katello_activation_keys';

        nutupane.searchTransform = function (term) {
            var addition,
                envClausses = [];

            angular.forEach($scope.selectedEnvironmentNames(), function(env) {
                envClausses.push("environment = " + env);
            });
            addition = '(' + envClausses.join(" OR ") + ')';
            addition = addition + " AND content_view_id = " + $scope.contentView.id;
            if (term === "" || angular.isUndefined(term)) {
                return addition;
            }

            return term + " AND " + addition;
        };

        $scope.table = nutupane.table;
        $scope.table.closeItem = function () {};

        $scope.environments = Organization.readableEnvironments({id: CurrentOrganization});
        $scope.initEnvironmentWatch($scope);

        if ($scope.deleteOptions.activationKeys.contentView) {
            $scope.selectedContentViewId = $scope.deleteOptions.activationKeys.contentView.id;
        }
        $scope.selectedEnvironment = $scope.deleteOptions.activationKeys.environment;

        $scope.processSelection = function () {
            $scope.deleteOptions.activationKeys.environment = $scope.selectedEnvironment;
            $scope.deleteOptions.activationKeys.contentView = _.find($scope.contentViewsForEnvironment,
                {id: $scope.selectedContentViewId});
            $scope.selectedEnvironment = undefined;
            $scope.selectedContentViewId = undefined;
            $scope.transitionToNext();
        };

        $scope.activationKeyLink = function () {
            var search = $scope.searchString($scope.contentView, $scope.deleteOptions.environments);
            return $scope.$state.href('activation-keys') + '?search=' + search;
        };

        $scope.toggleKeys = function () {
            $scope.showKeys = !$scope.showKeys;
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewVersionDeletionConfirm
 *
 * @requires $scope
 * @requires ContentView
 * @requires translate
 * @requires Notification
 *
 * @description
 *   Provides the confirmation and submit code for the content view version deletion
 *   workflow
 */

angular.module('Bastion.content-views').controller('ContentViewVersionDeletionConfirmController',
    ['$scope', 'ContentView', 'translate', 'Notification',
    function ($scope, ContentView, translate, Notification) {

        function success() {
            var message = translate('Successfully initiated removal of %cv version %ver.');

            if ($scope.deleteOptions.deleteArchive) {
                message = translate('Successfully initiated deletion of %cv version %ver.');
            }

            message = message.replace('%cv', $scope.contentView.name).replace('%ver', $scope.version.version);
            Notification.setSuccessMessage(message);
            $scope.transitionTo('content-view.versions', {contentViewId: $scope.contentView.id});
        }

        function error(response) {
            $scope.deleting = false;
            angular.forEach(response.data.errors, function (responseError) {
                Notification.setErrorMessage(responseError);
            });
        }

        $scope.validateEnvironmentSelection();

        $scope.performDeletion = function () {
            var params = {
                id: $scope.contentView.id,
                'environment_ids': $scope.selectedEnvironmentIds()
            };
            $scope.deleting = true;

            if ($scope.deleteOptions.deleteArchive) {
                params['content_view_version_ids'] = [$scope.version.id];
            }
            if ($scope.deleteOptions.activationKeys.contentView) {
                params['key_content_view_id'] = $scope.deleteOptions.activationKeys.contentView.id;
            }
            if ($scope.deleteOptions.activationKeys.environment) {
                params['key_environment_id'] = $scope.deleteOptions.activationKeys.environment.id;
            }
            if ($scope.deleteOptions.contentHosts.contentView) {
                params['system_content_view_id'] = $scope.deleteOptions.contentHosts.contentView.id;
            }
            if ($scope.deleteOptions.contentHosts.environment) {
                params['system_environment_id'] = $scope.deleteOptions.contentHosts.environment.id;
            }

            ContentView.removeAssociations(params, success, error);
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewVersionDeletionContentHosts
 *
 * @requires $scope
 * @requires $location
 * @requires Organization
 * @requires CurrentOrganization
 * @requires Nutupane
 * @requires ContentHost
 *
 * @description
 *   Provides functionality for picking which content view and environment to move Content Hosts to
 *   as part of content view version deletion.
 */

angular.module('Bastion.content-views').controller('ContentViewVersionDeletionContentHostsController',
    ['$scope', '$location', 'Organization', 'CurrentOrganization', 'Nutupane', 'Host',
    function ($scope, $location, Organization, CurrentOrganization, Nutupane, Host) {

        var params, nutupane;

        $scope.validateEnvironmentSelection();
        params = {
            'organization_id': CurrentOrganization,
            'content_view_id': $scope.contentView.id,
            'sort_by': 'name',
            'sort_order': 'ASC'
        };
        nutupane = new Nutupane(Host, params);
        $scope.controllerName = 'hosts';

        nutupane.searchTransform = function (term) {
            var addition,
                envIdClausses = [];

            angular.forEach($scope.selectedEnvironmentIds(), function(envId) {
                envIdClausses.push("lifecycle_environment_id = " + envId);
            });
            addition = '(' + envIdClausses.join(" OR ") + ')';
            addition = addition + " AND content_view_id = " + $scope.contentView.id;
            if (term === "" || angular.isUndefined(term)) {
                return addition;
            }

            return term + " AND " + addition;
        };
        $scope.table = nutupane.table;
        $scope.table.closeItem = function () {};

        $scope.environments = Organization.readableEnvironments({id: CurrentOrganization});
        $scope.initEnvironmentWatch($scope);

        $scope.selectedEnvironment = $scope.deleteOptions.contentHosts.environment;
        if ($scope.deleteOptions.contentHosts.contentView) {
            $scope.selectedContentViewId = $scope.deleteOptions.contentHosts.contentView.id;
        }

        $scope.processSelection = function () {
            $scope.deleteOptions.contentHosts.environment = $scope.selectedEnvironment;
            $scope.deleteOptions.contentHosts.contentView = _.find($scope.contentViewsForEnvironment,
                {id: $scope.selectedContentViewId});
            $scope.selectedEnvironment = undefined;
            $scope.selectedContentViewId = undefined;
            $scope.transitionToNext();
        };

        $scope.contentHostsLink = function () {
            var search = $scope.searchString($scope.contentView, $scope.deleteOptions.environments);
            return $scope.$state.href('content-hosts') + '?search=' + search;
        };

        $scope.toggleHosts = function () {
            $scope.showHosts = !$scope.showHosts;
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewVersionDeletionEnvironments
 *
 * @requires $scope
 *
 * @description
 *   Provides the functionality for selecting which environments a user wants to remove
 *   a specific content view from.
 */

angular.module('Bastion.content-views').controller('ContentViewVersionDeletionEnvironmentsController',
    ['$scope',
    function ($scope) {
        $scope.environmentsTable = {rows: {}};
        $scope.version.$promise.then(function () {
            var numSelections;
            $scope.environmentsTable.rows = $scope.version.environments;
            if ($scope.version.environments.length === 0) {
                $scope.deleteOptions.deleteArchive = true;
            } else {
                angular.forEach($scope.environmentsTable.rows, function (row) {
                    row.unselectable = !row.permissions['promotable_or_removable'] ||
                                         !row.permissions['all_hosts_editable'] ||
                                         !row.permissions['all_keys_editable'];
                });

                if ($scope.deleteOptions.environments.length === 0) {
                    //select all by default
                    angular.forEach($scope.environmentsTable.rows, function (row) {
                        row.selected = !row.unselectable;
                    });
                } else {
                    //set existing selections
                    angular.forEach($scope.environmentsTable.rows, function (row) {
                        row.selected = angular.isDefined(_.find($scope.deleteOptions.environments, {unselectable: false, id: row.id}));

                    });
                }

                numSelections = _.countBy($scope.environmentsTable.rows, function (row) {
                    return row.selected ? 'selected' : 'unselected';
                });

                $scope.environmentsTable.numSelected = numSelections.selected;
            }
        });

        $scope.canDeleteArchive = function () {
            return $scope.environmentsTable.numSelected === $scope.environmentsTable.rows.length;
        };

        $scope.selectionEmpty = function () {
            return !$scope.deleteOptions.deleteArchive && $scope.environmentsTable.numSelected === 0;
        };

        $scope.anySelectable = function () {
            var anySelectable;
            if ($scope.environmentsTable.rows.length === 0) {
                anySelectable = true;
            } else {
                anySelectable = angular.isDefined(_.find($scope.environmentsTable.rows, {unselectable: false}));
            }
            return anySelectable;
        };

        $scope.allSelectable = function () {
            return angular.isUndefined(_.find($scope.environmentsTable.rows, {unselectable: true}));
        };

        $scope.processSelection = function () {
            $scope.deleteOptions.environments = $scope.environmentsTable.getSelected();
            $scope.transitionToNext();
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewVersionDeletion
 *
 * @requires $scope
 * @requires ContentViewVersion
 * @requires ContentView
 *
 * @description
 *   Provides the base controller functionality for content view deletion including
 *   keeping track of selections through the workflow.
 *   Also provides logic to control the content view & environment selector for the two
 *   child controllers that use it.
 */

angular.module('Bastion.content-views').controller('ContentViewVersionDeletionController',
    ['$scope', '$state', 'ContentViewVersion', 'ContentView',
    function ($scope, $state, ContentViewVersion, ContentView) {

        $scope.version = ContentViewVersion.get({id: $scope.$stateParams.versionId});

        $scope.stepStates = {
            environments: 'content-view.version-deletion.environments',
            activationKeys: 'content-view.version-deletion.activation-keys',
            contentHosts: 'content-view.version-deletion.content-hosts',
            confirm: 'content-view.version-deletion.confirm'
        };

        if (angular.isUndefined($scope.deleteOptions)) {
            $scope.deleteOptions = {
                deleteArchive: false,
                environments: [],
                contentHosts: {},
                activationKeys: {}
            };
        }

        $scope.isStep = function (step) {
            return $state.is($scope.stepStates[step]);
        };

        $scope.transitionToNext = function (currentState) {
            var stepStates = $scope.stepStates;
            if (angular.isUndefined(currentState)) {
                currentState = $state.current.name;
            }

            if (currentState === stepStates.environments) {
                if ($scope.needHosts()) {
                    $scope.transitionTo(stepStates.contentHosts,
                                        {contentViewId: $scope.contentView.id, versionId: $scope.version.id});
                } else {
                    $scope.transitionToNext(stepStates.contentHosts);
                }
            } else if (currentState === stepStates.contentHosts) {
                if ($scope.needActivationKeys()) {
                    $scope.transitionTo(stepStates.activationKeys,
                                        {contentViewId: $scope.contentView.id, versionId: $scope.version.id});
                } else {
                    $scope.transitionToNext(stepStates.activationKeys);
                }
            } else {
                $scope.transitionTo(stepStates.confirm,
                                    {contentViewId: $scope.contentView.id, versionId: $scope.version.id});
            }
        };

        $scope.transitionBack = function (currentState) {
            var stepStates = $scope.stepStates;
            if (angular.isUndefined(currentState)) {
                currentState = $state.current.name;
            }

            if (currentState === stepStates.confirm) {
                if ($scope.needActivationKeys()) {
                    $scope.transitionTo(stepStates.activationKeys,
                                        {contentViewId: $scope.contentView.id, versionId: $scope.version.id});
                } else {
                    $scope.transitionBack(stepStates.activationKeys);
                }
            } else if (currentState === stepStates.activationKeys) {
                if ($scope.needHosts()) {
                    $scope.transitionTo(stepStates.contentHosts,
                                        {contentViewId: $scope.contentView.id, versionId: $scope.version.id});
                } else {
                    $scope.transitionBack(stepStates.contentHosts);
                }
            } else if (currentState === stepStates.contentHosts) {
                $scope.transitionTo(stepStates.environments,
                                    {contentViewId: $scope.contentView.id, versionId: $scope.version.id});
            }
        };

        $scope.validateEnvironmentSelection = function () {
            if (!$scope.deleteOptions.deleteArchive && $scope.deleteOptions.environments.length === 0) {
                $scope.transitionTo("content-view.version-deletion.environments",
                        {contentViewId: $scope.$stateParams.contentViewId, versionId: $scope.$stateParams.versionId});
            }
        };

        $scope.needHosts = function () {
            return $scope.totalHostCount() > 0;
        };

        $scope.needActivationKeys = function () {
            return $scope.totalActivationKeyCount() > 0;
        };

        $scope.selectedEnvironmentIds = function () {
            return _.map($scope.deleteOptions.environments, 'id');
        };

        $scope.selectedEnvironmentNames = function () {
            return _.map($scope.deleteOptions.environments, 'name');
        };

        $scope.totalHostCount = function () {
            return _.reduce($scope.deleteOptions.environments, function (sum, env) {
                return sum + env['host_count'];
            }, 0);
        };

        $scope.totalActivationKeyCount = function () {
            return _.reduce($scope.deleteOptions.environments, function (sum, env) {
                return sum + env['activation_key_count'];
            }, 0);
        };

        $scope.searchString = function (contentView, environments) {
            var envStrings = [],
                string = 'content_view:"%s"'.replace('%s', contentView.name);

            angular.forEach(environments, function (environment) {
                envStrings.push('environment:"%s"'.replace('%s', environment.name));
            });
            string = string + " AND (" + envStrings.join(" OR ") + ")";
            return string.replace(/"/g, "%22");
        };

        $scope.initEnvironmentWatch = function (childScope) {
            var removingEnvironment;
            childScope.$watch('selectedEnvironment', function () {
                if (angular.isUndefined(childScope.selectedEnvironment)) {
                    childScope.contentViewsForEnvironment = [];
                } else {
                    removingEnvironment = angular.isDefined(_.find(childScope.deleteOptions.environments, {id: childScope.selectedEnvironment.id}));
                    $scope.fetchingViews = true;
                    ContentView.queryUnpaged({ 'environment_id': childScope.selectedEnvironment.id },
                        function (response) {
                            $scope.fetchingViews = false;
                            childScope.contentViewsForEnvironment = _.reject(response.results, function (view) {
                                return (view.id === childScope.version['content_view_id']) && removingEnvironment;
                            });
                        }
                    );
                }
            });
        };

    }]
);
/**
 * @ngdoc service
 * @name  Bastion.content-views.factory:ContentViewComponent
 *
 * @requires BastionResource
 * @requires CurrentOrganization
 *
 * @description
 *   Provides a BastionResource for interacting with content view components.
 */

angular.module('Bastion.content-views').factory('ContentViewComponent',
    ['BastionResource', 'CurrentOrganization',
    function (BastionResource, CurrentOrganization) {

        return BastionResource('katello/api/v2/content_views/:compositeContentViewId/content_view_components/:id/:action',
            {id: '@id', compositeContentViewId: '@compositeContentViewId', 'organization_id': CurrentOrganization},
            {
                update: {method: 'PUT'},
                removeComponents: {
                    method: 'PUT',
                    isArray: false,
                    url: 'katello/api/v2/content_views/:compositeContentViewId/content_view_components/remove'
                },
                addComponents: {
                    method: 'PUT',
                    isArray: false,
                    url: 'katello/api/v2/content_views/:compositeContentViewId/content_view_components/add'
                }
            }
        );
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewCompositeAvailableContentViewsController
 *
 * @requires $scope
 * @requires Nutupane
 * @requires CurrentOrganization
 * @requires ContentView
 * @requires ContentViewComponent
 *
 * @description
 *  Provides a nutupane for eligible content views to be included in the composite.
 */

angular.module('Bastion.content-views').controller('ContentViewCompositeAvailableContentViewsController',
    ['$scope', 'Nutupane', 'CurrentOrganization', 'ContentView', 'ContentViewComponent',
        function ($scope, Nutupane, CurrentOrganization, ContentView, ContentViewComponent) {
            var nutupane, params;

            params = {
                'full_result': true,
                nondefault: true,
                noncomposite: true,
                'organization_id': CurrentOrganization
            };

            nutupane = new Nutupane(ContentView, params);
            $scope.controllerName = 'katello_content_views';
            nutupane.masterOnly = true;
            nutupane.table.initialLoad = false;
            $scope.table = nutupane.table;

            $scope.contentView.$promise.then(function (contentView) {
                var filterIds = [];
                if (contentView["content_view_components"]) {
                    filterIds = _.map(contentView["content_view_components"], function (component) {
                        return component['content_view'].id;
                    });
                }
                filterIds.push(contentView.id);

                params['without[]'] = filterIds;

                nutupane.setParams(params);
                nutupane.load(true);
            });

            $scope.addContentViews = function () {
                var selectedRows = nutupane.getAllSelectedResults().included.resources,
                    components;
                components = _.map(selectedRows, function (view) {
                    var component = {};
                    if ((!view.versionId) || view.versionId === "latest") {
                        component.latest = true;
                        component["content_view_id"] = view.id;
                    } else {
                        component["content_view_version_id"] = view.versionId;
                    }
                    return component;
                });

                ContentViewComponent.addComponents({compositeContentViewId: $scope.contentView.id,
                                                    components: components}, function () {
                    var newContentIds = _.map(selectedRows, 'id');
                    $scope.saveSuccess();
                    params['without[]'] = params["without[]"].concat(newContentIds);
                    nutupane.setParams(params);
                    nutupane.refresh();
                }, $scope.saveError);
            };
        }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewCompositeContentViewsListController
 *
 * @requires $scope
 * @requires translate
 * @requires Nutupane
 * @requires CurrentOrganization
 * @requires ContentView
 * @requires ContentViewComponent
 *
 * @description
 *  Provides a nutupane for existing content views that are included in the composite.
 */

angular.module('Bastion.content-views').controller('ContentViewCompositeContentViewsListController',
    ['$scope', 'translate', 'Nutupane', 'CurrentOrganization', 'ContentView', 'ContentViewComponent',
    function ($scope, translate, Nutupane, CurrentOrganization, ContentView, ContentViewComponent) {
        var nutupane = new Nutupane(ContentView, {
            'organization_id': CurrentOrganization,
            'id': $scope.$stateParams.contentViewId,
            'full_result': true
        }, 'contentViewComponents');
        $scope.controllerName = 'katello_content_views';

        nutupane.masterOnly = true;
        $scope.table = nutupane.table;

        $scope.saveContentViewComponent = function (contentViewComponent) {
            var component = {
                id: contentViewComponent.id,
                latest: contentViewComponent.versionId === "latest",
                compositeContentViewId: $scope.contentView.id
            };

            if (contentViewComponent.versionId !== "latest") {
                component["content_view_version_id"] = contentViewComponent.versionId;
            }
            ContentViewComponent.update(component, function () {
                nutupane.refresh();
                $scope.saveSuccess();
            }, $scope.saveError);
        };

        $scope.removeContentViewComponents = function () {
            var selected = nutupane.getAllSelectedResults().included.ids,
                params = {compositeContentViewId: $scope.contentView.id,
                          'component_ids': selected};
            ContentViewComponent.removeComponents(params, function () {
                nutupane.refresh();
                $scope.contentView.$get($scope.saveSuccess, $scope.saveError);
            }, $scope.saveError);
        };

        $scope.getVersionString = function (contentViewComponent) {
            if (contentViewComponent.latest) {
                return translate('Latest (Currently %s)').replace('%s', contentViewComponent['content_view_version'].version.toString());
            }
            return contentViewComponent['content_view_version'].version.toString();
        };

        $scope.getVersionId = function (contentViewComponent) {
            if (contentViewComponent.latest) {
                return "latest";
            }
            return contentViewComponent.content_view_version.id;
        };
    }]
);
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.content-views.controller:ContentViewAvailableDebRepositoriesController
     *
     * @requires $scope
     * @requires Repository
     * @requires Nutupane
     * @requires CurrentOrganization
     * @requires ContentViewRepositoriesUtil
     *
     * @description
     *    Provides UI functionality add deb repositories to a content view
     */
    function ContentViewAvailableDebRepositoriesController($scope, Repository, Nutupane, CurrentOrganization, ContentViewRepositoriesUtil) {
        var nutupane;

        ContentViewRepositoriesUtil($scope);

        nutupane = new Nutupane(Repository, {
            'organization_id': CurrentOrganization,
            'library': true,
            'content_type': 'deb',
            'content_view_id': $scope.$stateParams.contentViewId,
            'available_for': 'content_view'
        },
        'queryUnpaged');
        $scope.controllerName = 'katello_repositories';

        nutupane.load();

        $scope.table = nutupane.table;

        $scope.addRepositories = function (contentView) {
            $scope.addSelectedRepositoriesToContentView(nutupane, contentView);
        };
    }

    angular
        .module('Bastion.content-views')
        .controller('ContentViewAvailableDebRepositoriesController', ContentViewAvailableDebRepositoriesController);

    ContentViewAvailableDebRepositoriesController.$inject = [
        '$scope', 'Repository', 'Nutupane', 'CurrentOrganization', 'ContentViewRepositoriesUtil'
    ];

})();
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewAvailableDockerRepositoriesController
 *
 * @requires $scope
 * @requires Repository
 * @requires Nutupane
 * @requires CurrentOrganization
 * @requires ContentViewRepositoriesUtil
 *
 * @description
 *    Provides UI functionality add docker repositories to a content view
 */

angular.module('Bastion.content-views').controller('ContentViewAvailableDockerRepositoriesController',
    ['$scope', 'Repository', 'Nutupane', 'CurrentOrganization', 'ContentViewRepositoriesUtil',
    function ($scope, Repository, Nutupane, CurrentOrganization, ContentViewRepositoriesUtil) {

        var nutupane;

        ContentViewRepositoriesUtil($scope);

        nutupane = new Nutupane(Repository, {
            'organization_id': CurrentOrganization,
            'library': true,
            'content_type': 'docker',
            'content_view_id': $scope.$stateParams.contentViewId,
            'available_for': 'content_view'
        },
        'queryUnpaged');
        $scope.controllerName = 'katello_repositories';

        nutupane.masterOnly = true;

        $scope.table = nutupane.table;

        $scope.addRepositories = function (contentView) {
            $scope.addSelectedRepositoriesToContentView(nutupane, contentView);
        };
    }]
);
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.content-views.controller:ContentViewAvailableFileRepositoriesController
     *
     * @requires $scope
     * @requires Repository
     * @requires Nutupane
     * @requires CurrentOrganization
     * @requires ContentViewRepositoriesUtil
     *
     * @description
     *    Provides UI functionality add file repositories to a content view
     */
    function ContentViewAvailableFileRepositoriesController($scope, Repository, Nutupane, CurrentOrganization, ContentViewRepositoriesUtil) {
        var nutupane;

        ContentViewRepositoriesUtil($scope);

        nutupane = new Nutupane(Repository, {
            'organization_id': CurrentOrganization,
            'library': true,
            'content_type': 'file',
            'content_view_id': $scope.$stateParams.contentViewId,
            'available_for': 'content_view'
        },
        'queryUnpaged');
        $scope.controllerName = 'katello_repositories';

        nutupane.load();

        $scope.table = nutupane.table;

        $scope.addRepositories = function (contentView) {
            $scope.addSelectedRepositoriesToContentView(nutupane, contentView);
        };
    }

    angular
        .module('Bastion.content-views')
        .controller('ContentViewAvailableFileRepositoriesController', ContentViewAvailableFileRepositoriesController);

    ContentViewAvailableFileRepositoriesController.$inject = [
        '$scope', 'Repository', 'Nutupane', 'CurrentOrganization', 'ContentViewRepositoriesUtil'
    ];

})();
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewAvailableOstreeRepositoriesController
 *
 * @requires $scope
 * @requires Repository
 * @requires Nutupane
 * @requires CurrentOrganization
 * @requires ContentViewRepositoriesUtil
 *
 * @description
 *    Provides UI functionality add ostree repositories to a content view
 */

angular.module('Bastion.content-views').controller('ContentViewAvailableOstreeRepositoriesController',
    ['$scope', 'Repository', 'Nutupane', 'CurrentOrganization', 'ContentViewRepositoriesUtil',
    function ($scope, Repository, Nutupane, CurrentOrganization, ContentViewRepositoriesUtil) {

        var nutupane;

        ContentViewRepositoriesUtil($scope);

        nutupane = new Nutupane(Repository, {
            'organization_id': CurrentOrganization,
            'library': true,
            'content_type': 'ostree',
            'content_view_id': $scope.$stateParams.contentViewId,
            'available_for': 'content_view'
        },
        'queryUnpaged');
        $scope.controllerName = 'katello_repositories';

        nutupane.masterOnly = true;

        $scope.table = nutupane.table;

        $scope.addRepositories = function (contentView) {
            $scope.addSelectedRepositoriesToContentView(nutupane, contentView);
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:AvailablePuppetModulesController
 *
 * @description
 */

angular.module('Bastion.content-views').controller('AvailablePuppetModulesController',
    [function () {}]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewAvailableRepositoriesController
 *
 * @requires $scope
 * @requires Repository
 * @requires Nutupane
 * @requires CurrentOrganization
 * @requires ContentViewRepositoriesUtil
 *
 * @description
 *   Provides the functionality specific to ContentViews for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.content-views').controller('ContentViewAvailableRepositoriesController',
    ['$scope', 'Repository', 'Nutupane', 'CurrentOrganization', 'ContentViewRepositoriesUtil',
    function ($scope, Repository, Nutupane, CurrentOrganization, ContentViewRepositoriesUtil) {

        var nutupane;

        ContentViewRepositoriesUtil($scope);

        nutupane = new Nutupane(Repository, {
            'organization_id': CurrentOrganization,
            'library': true,
            'content_type': 'yum',
            'content_view_id': $scope.$stateParams.contentViewId,
            'available_for': 'content_view'
        },
        'queryUnpaged');
        $scope.controllerName = 'katello_repositories';

        nutupane.load();

        $scope.table = nutupane.table;

        $scope.addRepositories = function (contentView) {
            $scope.addSelectedRepositoriesToContentView(nutupane, contentView);
        };

    }]
);
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.content-views.controller:ContentViewDebRepositoriesListController
     *
     * @requires $scope
     * @requires Repository
     * @requires Nutupane
     * @requires CurrentOrganization
     * @requires ContentViewRepositoriesUtil
     *
     * @description
     *    Provides UI functionality list/remove deb repositories from a content view
     */
    function ContentViewDebRepositoriesListController($scope, Repository, Nutupane, CurrentOrganization, ContentViewRepositoriesUtil) {
        var nutupane;

        ContentViewRepositoriesUtil($scope);

        nutupane = new Nutupane(Repository, {
            'organization_id': CurrentOrganization,
            'content_view_id': $scope.$stateParams.contentViewId,
            'content_type': 'deb'
        },
        'queryUnpaged');
        $scope.controllerName = 'katello_repositories';

        nutupane.load();

        $scope.table = nutupane.table;

        $scope.removeRepositories = function () {
            $scope.removeSelectedRepositoriesFromContentView(nutupane, $scope.contentView);
        };
    }

    angular
        .module('Bastion.content-views')
        .controller('ContentViewDebRepositoriesListController', ContentViewDebRepositoriesListController);

    ContentViewDebRepositoriesListController.$inject = [
        '$scope', 'Repository', 'Nutupane', 'CurrentOrganization', 'ContentViewRepositoriesUtil'
    ];

})();
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewDetailsController
 *
 * @requires $scope
 * @requires $q
 * @requires ContentView
 * @requires translate
 * @requires ApiErrorHandler
 * @requires Notification
 *
 * @description
 *   Provides the functionality specific to the Content View Details page.
 */

angular.module('Bastion.content-views').controller('ContentViewDetailsController',
    ['$scope', '$q', 'ContentView', 'translate', 'ApiErrorHandler', 'Notification',
    function ($scope, $q, ContentView, translate, ApiErrorHandler, Notification) {
        $scope.saveSuccess = function () {
            Notification.setSuccessMessage(translate('Content View updated.'));
        };

        $scope.saveError = function (response) {
            angular.forEach(response.data.errors, function (errorMessage) {
                Notification.setErrorMessage(translate("An error occurred updating the Content View: ") + errorMessage);
            });
        };

        $scope.panel = {
            error: false,
            loading: true
        };

        $scope.taskTypes = {
            publish: "Actions::Katello::ContentView::Publish",
            promotion: "Actions::Katello::ContentView::Promote",
            deletion: "Actions::Katello::ContentView::Remove",
            incrementalUpdate: "Actions::Katello::ContentView::IncrementalUpdates",
            export: "Actions::Katello::ContentViewVersion::Export"
        };

        $scope.copy = function (newName) {
            ContentView.copy({id: $scope.contentView.id, 'content_view': {name: newName}}, function (response) {
                $scope.transitionTo('content-view.info', {contentViewId: response.id});
            }, function (response) {
                Notification.setErrorMessage(response.data.displayMessage);
            });
        };

        $scope.save = function (contentView) {
            return contentView.$update($scope.saveSuccess, $scope.saveError);
        };

        $scope.fetchContentView = function () {
            $scope.contentView = ContentView.get({id: $scope.$stateParams.contentViewId}, function () {
                $scope.panel.loading = false;
            }, function (response) {
                $scope.panel.loading = false;
                ApiErrorHandler.handleGETRequestErrors(response, $scope);
            });
        };

        $scope.getAvailableVersions = function (paramContentView) {
            var deferred, latestVersion, latest;

            if (paramContentView.latest_version) {
                latestVersion = translate('Always Use Latest (Currently %s)').replace('%s', paramContentView.latest_version.toString());
                latest = [{id: "latest", version: latestVersion}];
            } else {
                return [];
            }

            if (angular.isUndefined(paramContentView.versions)) {
                deferred = $q.defer();

                ContentView.get({id: paramContentView.id}, function (response) {
                    deferred.resolve(latest.concat(response.versions.reverse()));
                });

                return deferred.promise;
            }
            return latest.concat(paramContentView.versions.reverse());
        };

        $scope.fetchContentView();
    }]
);
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.content-views.controller:ContentViewDockerRepositoriesListController
     *
     * @requires $scope
     * @requires Repository
     * @requires Nutupane
     * @requires CurrentOrganization
     * @requires ContentViewRepositoriesUtil
     *
     * @description
     *    Provides UI functionality list/remove docker repositories from a content view
     */
    function ContentViewDockerRepositoriesListController($scope, Repository, Nutupane, CurrentOrganization, ContentViewRepositoriesUtil) {
        var nutupane;

        ContentViewRepositoriesUtil($scope);

        nutupane = new Nutupane(Repository, {
            'organization_id': CurrentOrganization,
            'content_view_id': $scope.$stateParams.contentViewId,
            'content_type': 'docker'
        },
        'queryUnpaged');
        $scope.controllerName = 'katello_repositories';

        nutupane.masterOnly = true;
        nutupane.load();

        $scope.table = nutupane.table;

        $scope.removeRepositories = function () {
            $scope.removeSelectedRepositoriesFromContentView(nutupane, $scope.contentView);
        };
    }

    angular
        .module('Bastion.content-views')
        .controller('ContentViewDockerRepositoriesListController', ContentViewDockerRepositoriesListController);

    ContentViewDockerRepositoriesListController.$inject = [
        '$scope', 'Repository', 'Nutupane', 'CurrentOrganization', 'ContentViewRepositoriesUtil'
    ];

})();
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.content-views.controller:ContentViewFileRepositoriesListController
     *
     * @requires $scope
     * @requires Repository
     * @requires Nutupane
     * @requires CurrentOrganization
     * @requires ContentViewRepositoriesUtil
     *
     * @description
     *    Provides UI functionality list/remove file repositories from a content view
     */
    function ContentViewFileRepositoriesListController($scope, Repository, Nutupane, CurrentOrganization, ContentViewRepositoriesUtil) {
        var nutupane;

        ContentViewRepositoriesUtil($scope);

        nutupane = new Nutupane(Repository, {
            'organization_id': CurrentOrganization,
            'content_view_id': $scope.$stateParams.contentViewId,
            'content_type': 'file'
        },
        'queryUnpaged');
        $scope.controllerName = 'katello_repositories';

        nutupane.load();

        $scope.table = nutupane.table;

        $scope.removeRepositories = function () {
            $scope.removeSelectedRepositoriesFromContentView(nutupane, $scope.contentView);
        };
    }

    angular
        .module('Bastion.content-views')
        .controller('ContentViewFileRepositoriesListController', ContentViewFileRepositoriesListController);

    ContentViewFileRepositoriesListController.$inject = [
        '$scope', 'Repository', 'Nutupane', 'CurrentOrganization', 'ContentViewRepositoriesUtil'
    ];

})();
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.content-views.controller:ContentViewOstreeRepositoriesListController
     *
     * @requires $scope
     * @requires Repository
     * @requires Nutupane
     * @requires CurrentOrganization
     * @requires ContentViewRepositoriesUtil
     *
     * @description
     *    Provides UI functionality list/remove ostree repositories from a content view
     */
    function ContentViewOstreeRepositoriesListController($scope, Repository, Nutupane, CurrentOrganization, ContentViewRepositoriesUtil) {
        var nutupane;

        ContentViewRepositoriesUtil($scope);

        nutupane = new Nutupane(Repository, {
            'organization_id': CurrentOrganization,
            'content_view_id': $scope.$stateParams.contentViewId,
            'content_type': 'ostree'
        },
        'queryUnpaged');
        $scope.controllerName = 'katello_repositories';

        nutupane.masterOnly = true;
        nutupane.load();

        $scope.table = nutupane.table;

        $scope.removeRepositories = function () {
            $scope.removeSelectedRepositoriesFromContentView(nutupane, $scope.contentView);
        };
    }

    angular
        .module('Bastion.content-views')
        .controller('ContentViewOstreeRepositoriesListController', ContentViewOstreeRepositoriesListController);

    ContentViewOstreeRepositoriesListController.$inject = [
        '$scope', 'Repository', 'Nutupane', 'CurrentOrganization', 'ContentViewRepositoriesUtil'
    ];

})();
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewPromotionController
 *
 * @requires $scope
 * @requires $q
 * @requires translate
 * @requires ContentViewVersion
 * @requires Organization
 * @requires CurrentOrganization
 * @requires Notification
 *
 * @description
 *   Provides the functionality specific to ContentViews for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.content-views').controller('ContentViewPromotionController',
    ['$scope', '$q', 'translate', 'ContentViewVersion', 'Organization', 'CurrentOrganization', 'Notification',
    function ($scope, $q, translate, ContentViewVersion, Organization, CurrentOrganization, Notification) {

        function success() {
            var message = translate('Successfully initiated promotion of %cv version %ver to %env.');
            message = message.replace('%cv', $scope.contentView.name).replace('%env', $scope.selectedEnvironment.name);
            message = message.replace('%ver', $scope.version.version);
            $scope.promoting = false;
            Notification.setSuccessMessage(message);
            $scope.transitionTo('content-view.versions', {contentViewId: $scope.contentView.id});
        }

        function failure(response) {
            $scope.promoting = false;
            Notification.setErrorMessage(response.data.displayMessage);
        }

        $scope.promotion = {};
        $scope.working = false;

        $scope.version = ContentViewVersion.get({id: $scope.$stateParams.versionId});
        $scope.currentOrganization = CurrentOrganization;
        $scope.availableEnvironments = Organization.paths({id: CurrentOrganization, 'permission_type': 'promotable'});
        $scope.suggestedEnvironments = [];

        $q.all([$scope.availableEnvironments.$promise, $scope.version.$promise]).then(function (args) {
            var environments = args[0],
                version = args[1];

            angular.forEach(environments, function (path) {
                angular.forEach(path.environments, function (environment) {
                    environment.disabled = $scope.checkDisabled(environment, version.environments);
                    if ($scope.checkSuggested(environment, version.environments)) {
                        environment.iconClass = 'fa fa-star';
                    }
                });
            });

            $scope.working = true;
        });

        $scope.checkDisabled = function (env, environments) {
            var enabled = false,
                envIds = _.map(environments, 'id');

            if (!env.prior) {
                env.prior = {};
            }

            if (envIds.indexOf(env.id) !== -1) {
                // if version is already promoted to the environment
                enabled = false;
            } else {
                enabled = true;
            }

            return !enabled;
        };

        $scope.checkSuggested = function (env, environments) {
            var suggest = false,
                envIds = _.map(environments, 'id');

            if (envIds.indexOf(env.id) === -1) {
                // if version is not promoted to the environment
                if (environments.length === 0 && env.library) {
                    // if version is not yet promoted to any environment & environment is library
                    suggest = true;
                    $scope.suggestedEnvironments.push(env);
                } else if (envIds.indexOf(env.prior.id) !== -1) {
                    // if environment is a successor an existing environment
                    suggest = true;
                    $scope.suggestedEnvironments.push(env);
                }
            }
            return suggest;
        };

        $scope.promote = function () {
            $scope.promoting = true;
            ContentViewVersion.promote({id: $scope.version.id,
                                        'environment_id': $scope.selectedEnvironment.id,
                                        'description': $scope.description,
                                        'force_yum_metadata_regeneration': $scope.forceMetadataRegeneration,
                                        force: true},
                success, failure);
        };

        $scope.verifySelection = function () {
            if ($scope.suggestedEnvironments.indexOf($scope.selectedEnvironment) !== -1) {
                $scope.promote();
            } else {
                $scope.openModal();
            }
        };

        $scope.suggestedEnvironmentMessage = function () {
            var envs = _.uniq(_.map($scope.suggestedEnvironments, 'name')),
                message = "";

            if (envs.length === 0) {
                message = "There are no environments suggested in the promotion path.";
            } else if (envs.length === 1) {
                message = "Suggested environment is: ".concat(envs[0]);
            } else {
                message = "Suggested environments are: ".concat(envs.join(', '));
            }

            return message;
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewPublishController
 *
 * @requires $scope
 * @requires translate
 * @requires ContentView
 * @requires Notification
 *
 * @description
 *   Provides the functionality specific to ContentViews for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.content-views').controller('ContentViewPublishController',
    ['$scope', 'translate', 'ContentView', 'Notification', function ($scope, translate, ContentView, Notification) {

        function success() {
            $scope.transitionTo('content-view.versions',
                                {contentViewId: $scope.contentView.id});

            //get the latest version number from the server
            $scope.$parent.contentView = ContentView.get({id: $scope.$stateParams.contentViewId});

            $scope.working = false;
        }

        function failure(response) {
            Notification.setErrorMessage(response.data.displayMessage);
            $scope.working = false;
        }

        $scope.version = {};

        $scope.publish = function (contentView) {
            var description = $scope.version.description,
                forceMetadataRegeneration = $scope.version.forceMetadataRegeneration,
                data = {'id': contentView.id, 'description': description, 'force_yum_metadata_regeneration': forceMetadataRegeneration};
            $scope.working = true;
            ContentView.publish(data, success, failure);
        };

        //Refetch the content view so that the contentView is updated for latest components
        $scope.fetchContentView();

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewRepositoriesListController
 *
 * @requires $scope
 * @requires Repository
 * @requires Nutupane
 * @requires CurrentOrganization
 * @requires ContentViewRepositoriesUtil
 *
 * @description
 *   Provides the functionality specific to ContentViews for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.content-views').controller('ContentViewRepositoriesListController',
    ['$scope', 'Repository', 'Nutupane', 'CurrentOrganization', 'ContentViewRepositoriesUtil',
    function ($scope, Repository, Nutupane, CurrentOrganization, ContentViewRepositoriesUtil) {
        var nutupane;

        ContentViewRepositoriesUtil($scope);

        nutupane = new Nutupane(Repository, {
            'organization_id': CurrentOrganization,
            'content_view_id': $scope.$stateParams.contentViewId,
            'content_type': 'yum'
        },
        'queryUnpaged');
        $scope.controllerName = 'katello_repositories';

        nutupane.load();

        $scope.table = nutupane.table;

        $scope.removeRepositories = function () {
            $scope.removeSelectedRepositoriesFromContentView(nutupane, $scope.contentView);
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.service:ContentViewRepositoriesService
 *
 * @requires translate
 *
 * @description
 *   Provides the functionality specific to ContentViews for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.content-views').service('ContentViewRepositoriesUtil',
    ['translate', function (translate) {

        return function (scope) {

            function extractProducts(repositories) {
                var products = {};

                scope.product = {name: translate('All Products'), id: 'all'};

                angular.forEach(repositories, function (repository) {
                    products[repository.product.id] = repository.product;
                });

                products[scope.product.id] = scope.product;

                return products;
            }

            scope.product = {id: 'all'};
            scope.products = {};
            scope.filteredItems = [];

            scope.$watch('table.rows', function (repositories) {
                scope.products = extractProducts(repositories);
            });

            scope.repositoryFilter = function (repository) {
                var include = true;

                if (scope.product && scope.product.id !== 'all') {
                    if (repository.product.id !== scope.product.id) {
                        include = false;
                    }
                }

                return include;
            };

            scope.getSelected = function (nutupane) {
                var selected = nutupane.getAllSelectedResults().included.ids,
                    filtered = _.map(scope.filteredItems, 'id');

                selected = _.reject(selected, function (id) {
                    return !_.includes(filtered, id);
                });

                return selected;
            };

            scope.removeSelectedRepositoriesFromContentView = function (nutupane, contentView) {
                var ids = [],
                    selected = scope.getSelected(nutupane);

                angular.forEach(contentView['repository_ids'], function (id) {
                    if (selected.indexOf(id) === -1) {
                        ids.push(id);
                    }
                });

                contentView['repository_ids'] = ids;

                scope.save(contentView).then(function () {
                    nutupane.refresh();
                });
            };

            scope.addSelectedRepositoriesToContentView = function (nutupane, contentView) {
                var selected = scope.getSelected(nutupane);

                contentView['repository_ids'] = contentView['repository_ids'].concat(selected);

                scope.save(contentView).then(function () {
                    nutupane.refresh();
                });
            };

        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewVersionsController
 *
 * @requires $scope
 * @requires translate
 * @requires Nutupane
 * @requires ContentViewVersion
 * @requires AggregateTask
 * @requires ApiErrorHandler
 * @requires Notification
 *
 * @description
 *   Provides the functionality specific to ContentViews for use with the table view UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.content-views').controller('ContentViewVersionsController',
    ['$scope', 'translate', 'Nutupane', 'ContentViewVersion', 'AggregateTask', 'ApiErrorHandler', 'Notification',
    function ($scope, translate, Nutupane, ContentViewVersion, AggregateTask, ApiErrorHandler, Notification) {
        var nutupane;

        function pluralSafe(count, strings) {
            if (count === 1) {
                return strings[0];
            }

            return strings[1];
        }

        function publishingMessage(count) {
            var messages = [translate("Publishing and promoting to 1 environment."),
                translate("Publishing and promoting to %count environments.").replace(
                    '%count', count)];
            return pluralSafe(count, messages);
        }

        function promotingMessage(count) {
            var messages = [translate('Promoting to 1 environment.'),
                translate("Promoting to %count environments.").replace('%count', count)];
            return pluralSafe(count, messages);
        }

        function deletingMessage(count) {
            var messages = [translate('Deleting from 1 environment.'),
                translate("Deleting from %count environments.").replace('%count', count)];
            return pluralSafe(count, messages);
        }

        function findTaskTypes(activeHistory, taskType) {
            return _.filter(activeHistory, function (history) {
                return history.task.label === taskType;
            });
        }

        function publishCompleteMessage(version) {
            return translate("Successfully published %cv version %ver and promoted to Library")
                .replace('%cv', version['content_view'].name)
                .replace('%ver', version.version);
        }

        function promotionCompleteMessage(version, task) {
            return translate("Successfully promoted %cv version %ver to %env")
                .replace('%cv', version['content_view'].name)
                .replace('%ver', version.version)
                .replace('%env', task.input.environments.join(", "));
        }

        function deletionCompleteMessage(version, task) {
            var message;

            if (task.input['content_view_ids'] && task.input['content_view_ids'].length > 0) {
                message = translate("Successfully deleted %cv version %ver.")
                    .replace('%cv', version['content_view'].name)
                    .replace('%ver', version.version);
            } else {
                message = translate("Successfully removed %cv version %ver from environments: %env")
                    .replace('%cv', version['content_view'].name)
                    .replace('%ver', version.version)
                    .replace('%env', task.input['environment_names'].join(', '));
            }
            return message;
        }

        function updateVersion(version) {
            var versionIds = _.map($scope.table.rows, 'id'),
                versionIndex = versionIds.indexOf(version.id);

            ContentViewVersion.get({'id': version.id}).$promise.then(function (newVersion) {
                $scope.panel.loading = false;
                $scope.contentView.versions[versionIndex] = newVersion;
                $scope.table.rows[versionIndex] = newVersion;
            }, function (response) {
                $scope.panel.loading = false;
                ApiErrorHandler.handleGETRequestErrors(response, $scope);
            });
        }

        function taskUpdated(version, task) {
            var taskTypes = $scope.taskTypes;

            if (!task.pending) {
                $scope.pendingVersionTask = false;

                if (task.result === 'success') {
                    if (task.label === taskTypes.promotion) {
                        Notification.setSuccessMessage(promotionCompleteMessage(version, task));
                    } else if (task.label === taskTypes.publish) {
                        Notification.setSuccessMessage(publishCompleteMessage(version));
                    } else if (task.label === taskTypes.deletion) {
                        Notification.setSuccessMessage(deletionCompleteMessage(version, task));
                        $scope.reloadVersions();
                    }
                }
            } else {
                $scope.pendingVersionTask = true;
            }
        }

        function processTasks(versions) {
            _.each(versions, function (version) {
                var taskIds = _.map(version['active_history'], function (history) {
                    return history.task.id;
                });

                if (taskIds.length > 0) {
                    version.task = AggregateTask.new(taskIds, function (task) {
                        taskUpdated(version, task);
                        if (task.label === $scope.taskTypes.publish && !task.pending && task.result === 'success') {
                            updateVersion(version);
                        }
                    });
                }
            });
        }

        $scope.regenerateRepositories = function(version) {
            ContentViewVersion.republishRepositories({id: version.id}, function(task) {
                $scope.transitionTo('content-view.task', {taskId: task.id, contentViewId: version.content_view_id});
            });
        };

        $scope.hideProgress = function (version) {
            return version['active_history'].length === 0 || (version.task.state === 'stopped' &&
                version.task.progressbar.type === 'success');
        };

        $scope.historyText = function (version) {
            var taskTypes = $scope.taskTypes,
                taskType = version['last_event'].task.label,
                message = "";

            if (taskType === taskTypes.deletion) {
                if (version['last_event'].environment) {
                    message = translate("Deletion from %s").replace('%s', version['last_event'].environment.name);
                } else {
                    message = translate("Version Deletion");
                }
            } else if (taskType === taskTypes.promotion) {
                message = translate("Promoted to %s").replace('%s', version['last_event'].environment.name);
            } else if (taskType === taskTypes.publish) {
                message = translate("Published");
            } else if (taskType === taskTypes.incrementalUpdate) {
                message = translate("Incremental Update");
            }
            return message;
        };

        $scope.status = function (version) {
            var taskTypes = $scope.taskTypes,
                deletionEvents = findTaskTypes(version['active_history'], taskTypes.deletion),
                promotionEvents = findTaskTypes(version['active_history'], taskTypes.promotion),
                publishEvents = findTaskTypes(version['active_history'], taskTypes.publish),
                messages = "";

            if (deletionEvents.length > 0) {
                messages = deletingMessage(deletionEvents.length);
            } else if (promotionEvents.length > 0) {
                messages = promotingMessage(promotionEvents.length);
            } else if (publishEvents.length > 0) {
                messages = publishingMessage(publishEvents.length);
            }

            return messages;
        };

        $scope.taskInProgress = function (version) {
            var inProgress = false;
            if (version.task && (version.task.state === 'pending' || version.task.state === 'running')) {
                inProgress = true;
            }
            return inProgress;
        };

        $scope.taskFailed = function (version) {
            var failed = false;
            if (version.task && (version.task.result === 'error')) {
                failed = true;
            }
            return failed;
        };

        $scope.reloadVersions = function () {
            $scope.table.rows = [];
            nutupane.refresh();
        };

        $scope.$watch('table.rows', function () {
            if ($scope.table && $scope.table.rows.length > 0) {
                processTasks($scope.table.rows);
            }
        });

        $scope.$on('$destroy', function () {
            _.each($scope.versions, function (version) {
                if (version.task) {
                    version.task.unregisterAll();
                }
            });
        });

        $scope.panel = {
            error: false,
            loading: true
        };

        nutupane = new Nutupane(ContentViewVersion, {'content_view_id': $scope.$stateParams.contentViewId});
        $scope.controllerName = 'katello_content_views';
        nutupane.setSearchKey('contentViewVersionSearch');
        nutupane.masterOnly = true;
        $scope.table = nutupane.table;

        $scope.pendingVersionTask = false;
        $scope.reloadVersions();
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:AvailableErrataFilterController
 *
 * @requires $scope
 * @requires translate
 * @requires Nutupane
 * @requires Filter
 * @requires Rule
 * @requires Notification
 *
 * @description
 *   Handles loading of errata that is available to be added to a filter and provides
 *   functionality to create filter rules based off selected errata.
 */

angular.module('Bastion.content-views').controller('AvailableErrataFilterController',
    ['$scope', 'translate', 'Nutupane', 'Erratum', 'Rule', 'Notification',
    function ($scope, translate, Nutupane, Erratum, Rule, Notification) {

        var nutupane, filterByDate;

        function success(data) {
            var rules;

            if (data.results) {
                rules = data.results;
            } else {
                rules = [data];
            }

            $scope.filter.rules = _.union($scope.filter.rules, rules);
            Notification.setSuccessMessage(translate('Errata successfully added.'));
            nutupane.table.selectAllResults(false);
            nutupane.refresh();
        }

        function failure(response) {
            Notification.setErrorMessage(response.data.displayMessage);
        }

        function saveRules(rules, filter) {
            var params = {filterId: filter.id};

            return rules.$save(params, success, failure);
        }

        $scope.addErrata = function (filter) {
            var errataIds,
                rules,
                results = nutupane.getAllSelectedResults('errata_id');

            if (nutupane.table.allResultsSelected) {
                rules = new Rule({'errata_ids': results});
            } else {
                errataIds = results.included.ids;
                rules = new Rule({'errata_ids': errataIds});
            }

            nutupane.table.working = true;
            saveRules(rules, filter);
        };

        $scope.updateTypes = function (errataTypes) {
            var types = [];

            angular.forEach(errataTypes, function (value, key) {
                if (value) {
                    types.push(key);
                }
            });

            nutupane.addParam('types[]', types);
            nutupane.refresh();
        };

        $scope.updateDateType = function () {
            nutupane.addParam("sort_by", $scope.rule["date_type"] );
            nutupane.addParam("date_type", $scope.rule["date_type"]);
            nutupane.refresh();
        };

        $scope.$watch('rule.start_date', function (start) {
            if (start) {
                filterByDate(start, 'start_date');
            }
        });

        $scope.$watch('rule.end_date', function (end) {
            if (end) {
                filterByDate(end, 'end_date');
            }
        });

        $scope.nutupane = nutupane = new Nutupane(Erratum, {
                filterId: $scope.$stateParams.filterId,
                'sort_order': 'DESC',
                'sort_by': 'issued',
                'available_for': 'content_view_filter'
            },
            'queryUnpaged'
        );
        $scope.controllerName = 'katello_erratum_packages';
        nutupane.masterOnly = true;

        filterByDate = function (date, type) {
            date = date.toISOString().split('T')[0];
            nutupane.addParam(type, date);
            nutupane.refresh();
        };

        $scope.updateTypes($scope.types);
        $scope.table = nutupane.table;

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:AvailablePackageGroupFilterController
 *
 * @requires $scope
 * @requires translate
 * @requires Filter
 * @requires Rule
 * @requires Nutupane
 * @requires Notification
 *
 * @description
 *   Handles fetching package groups that are available to add to a filter and saving
 *   each selected package group as a filter rule.
 */

angular.module('Bastion.content-views').controller('AvailablePackageGroupFilterController',
    ['$scope', 'translate', 'PackageGroup', 'Rule', 'Nutupane', 'Notification',
    function ($scope, translate, PackageGroup, Rule, Nutupane, Notification) {
        var nutupane;

        nutupane = new Nutupane(
            PackageGroup,
            {
              filterId: $scope.$stateParams.filterId,
              'available_for': 'content_view_filter',
              'sort_order': 'ASC',
              'sort_by': 'name'

            },
            'queryUnpaged'
        );
        $scope.controllerName = 'katello_package_groups';

        function success(rule) {
            nutupane.removeRow(rule.uuid, 'uuid');
            $scope.filter.rules.push(rule);
            Notification.setSuccessMessage(translate('Package Group successfully added.'));
        }

        function failure(response) {
            angular.forEach(response.data.displayMessage, function (error) {
                Notification.setErrorMessage(error);
            });
        }

        function saveRule(rule, filter) {
            var params = {filterId: filter.id};

            rule.$save(params, success, failure);
        }

        $scope.table = nutupane.table;
        nutupane.masterOnly = true;
        nutupane.table.closeItem = function () {};

        $scope.addPackageGroups = function (filter) {
            var packageGroups = nutupane.getAllSelectedResults().included.resources;

            angular.forEach(packageGroups, function (group) {
                var rule = new Rule({uuid: group.uuid, name: group.name});
                saveRule(rule, filter);
            });
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:DateTypeErrataFilterController
 *
 * @requires $scope
 * @requires translate
 * @requires Rule
 * @requires Notification
 *
 * @description
 *   Handles creating an errata filter that allows specification of a start date, end date and/or
 *   set of errata types by which to dynamically filter.
 */

angular.module('Bastion.content-views').controller('DateTypeErrataFilterController',
    ['$scope', 'translate', 'Rule', 'Notification', function ($scope, translate, Rule, Notification) {

        function convertRuleStringsToDates() {
            $scope.rule['start_date'] = new Date($scope.rule['start_date']);
            $scope.rule['end_date'] = new Date($scope.rule['end_date']);
        }

        function success() {
            convertRuleStringsToDates();
            Notification.setSuccessMessage(translate('Updated errata filter - ' + $scope.filter.name));
        }

        function failure(response) {
            $scope.rule.working = false;
            angular.forEach(response.data.errors, function (error) {
                Notification.setErrorMessage(error);
            });
        }

        $scope.filter.$promise.then(function (filter) {
            $scope.types = {
                enhancement: false,
                bugfix: false,
                security: false
            };
            $scope.rule = new Rule(filter.rules[0]);

            convertRuleStringsToDates();

            angular.forEach($scope.types, function (value, type) {
                if ($scope.rule.types.indexOf(type) > -1) {
                    $scope.types[type] = true;
                }
            });

            if (angular.isUndefined($scope.rule['date_type'])) {
                $scope.rule['date_type'] = "updated";
            }
        });

        $scope.updateTypes = function (types) {
            $scope.rule.types = [];

            angular.forEach(types, function (chosen, type) {
                if (chosen) {
                    $scope.rule.types.push(type);
                }
            });
        };

        $scope.save = function (rule, filter) {
            var params;

            // Remove any hours from the dates
            rule['start_date'].setHours(0, 0, 0, 0);
            rule['end_date'].setHours(0, 0, 0, 0);

            params = {filterId: filter.id, ruleId: rule.id};
            rule.$update(params, success, failure);
            $scope.filter.rules[0] = rule;
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:DockerTagFilterController
 *
 * @requires $scope
 * @requires $location
 * @requires translate
 * @requires Nutupane
 * @requires CurrentOrganization
 * @requires Filter
 * @requires Rule
 * @requires DockerTag
 * @requires Notification
 *
 * @description
 *   Handles docker tag filter rules for a content view.
 */

angular.module('Bastion.content-views').controller('DockerTagFilterController',
    ['$scope', '$location', 'translate', 'Nutupane', 'CurrentOrganization', 'Filter', 'Rule', 'DockerTag', 'Notification', function ($scope, $location, translate, Nutupane, CurrentOrganization, Filter, Rule, DockerTag, Notification) {
        var nutupane, params;

        function failure(response) {
            Notification.setErrorMessage(response.data.displayMessage);
        }

        function createRule(rule) {
            var success;

            success = function (result) {
                rule.id = result.id;
                rule.editMode = false;
                rule.working = false;

                Notification.setSuccessMessage(translate('Package successfully added.'));
            };

            if ($scope.valid(rule)) {
                Rule.save({filterId: $scope.filter.id}, rule, success, failure);
            }
        }

        function updateRule(rule) {
            var updateParams = {filterId: $scope.filter.id, ruleId: rule.id},
                success, error;

            // Need access to the original rule
            success = function () {
                rule.previous = {};
                rule.editMode = false;
                rule.working = false;
                Notification.setSuccessMessage(translate('Filter rule successfully updated.'));
            };

            error = function () {
                rule.working = false;
            };

            Rule.update(updateParams, rule, success, error);
        }

        function removeRule(rule) {
            var success, ruleId = rule.id;

            success = function () {
                nutupane.removeRow(ruleId);
                Notification.setSuccessMessage(translate('Filter rule successfully removed.'));
            };

            Rule.delete({filterId: $scope.$stateParams.filterId, ruleId: ruleId}, success, failure);
        }

        params = {
            filterId: $scope.$stateParams.filterId,
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'sort_by': 'name',
            'sort_order': 'ASC',
            'paged': true
        };

        nutupane = new Nutupane(Filter, params, 'rules');
        $scope.table = nutupane.table;

        $scope.addRule = function () {
            var rule = new Rule();
            rule.editMode = true;
            $scope.table.addRow(rule);
        };

        $scope.saveRule = function (rule) {
            if (rule.id) {
                updateRule(rule);
            } else {
                createRule(rule);
            }
        };

        $scope.valid = function (rule) {
            return rule.name ? true : false;
        };

        $scope.backupPrevious = function (rule) {
            rule.previous = {};
            angular.copy(rule, rule.previous);
        };

        $scope.restorePrevious = function (rule) {
            if (rule.id) {
                angular.copy(rule.previous, rule);
            } else {
                $scope.table.rows.shift();
            }
            rule.previous = {};
        };

        $scope.removeRules = function () {
            angular.forEach($scope.table.getSelected(), function (rule) {
                removeRule(rule);
            });
        };

        $scope.fetchAutocompleteName = function (term) {
            var repositoryIds = $scope.contentView['repository_ids'],
                promise;

            promise = DockerTag.autocompleteName({'repoids[]': repositoryIds, term: term}).$promise;

            return promise.then(function (data) {
                return data.results;
            });
        };

        $scope.filterRepositoriesByType = function () {
            return (_.filter($scope.filter.repositories, ["content_type", "docker"]));
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:FilterEditController
 *
 * @requires $scope
 * @requires $q
 * @requires translate
 * @requires Notification
 *
 * @description
 *   Provides functionality for editing name and description of content view filters.
 */

angular.module('Bastion.content-views').controller('FilterEditController',
    ['$scope', '$q', 'translate', 'Notification', function ($scope, $q, translate, Notification) {
    $scope.save = function (filter) {
        var deferred = $q.defer();
        var success;
        var failure = function (response) {
            deferred.reject(response);
            angular.forEach(response.data.errors, function (errorMessage) {
                Notification.setErrorMessage(translate("An error occurred saving the Filter: ") + errorMessage);
            });
            $scope.working = false;
        };

        success = function (response) {
            deferred.resolve(response);
            Notification.setSuccessMessage(translate('Filter Saved'));
            $scope.working = false;
            $scope.$emit('filter.updated');
        };

        filter.$update(success, failure);
        return deferred.promise;
    };
}]);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ErrataFilterListController
 *
 * @requires $scope
 * @requires translate
 * @requires Nutupane
 * @requires Filter
 * @requires Rule
 * @requires Notification
 *
 * @description
 *   Handles displaying a list of errata currently added to the filter and the ability
 *   to remove errata from the filter.
 */

angular.module('Bastion.content-views').controller('ErrataFilterListController',
    ['$scope', 'translate', 'Nutupane', 'Erratum', 'Rule', 'Notification',
    function ($scope, translate, Nutupane, Erratum, Rule, Notification) {
        var nutupane;

        function findRules(errataIds) {
            var rules = [];

            angular.forEach(errataIds, function (id) {
                var found;

                found = _.find($scope.filter.rules, function (rule) {
                    return (rule['errata_id'] === id);
                });

                if (found) {
                    rules.push(new Rule(found));
                }
            });

            return rules;
        }

        function success(rule) {
            nutupane.removeRow(rule['errata_id'], 'errata_id');
            $scope.filter.rules = _.reject($scope.filter.rules, function (filterRule) {
                return rule.id === filterRule.id;
            });
            Notification.setSuccessMessage(translate('Errata successfully removed.'));
        }

        function failure(response) {
            Notification.setErrorMessage(response.data.displayMessage);
        }

        $scope.nutupane = nutupane = new Nutupane(Erratum, {
                filterId: $scope.$stateParams.filterId,
                'sort_order': 'DESC',
                'sort_by': 'issued'
            },
            'queryUnpaged'
        );
        $scope.controllerName = 'katello_errata';

        $scope.table = nutupane.table;

        $scope.removeErrata = function () {
            var errataIds = nutupane.getAllSelectedResults('errata_id').included.ids,
                rules;

            rules = findRules(errataIds);

            angular.forEach(rules, function (rule) {
                rule.$delete(success, failure);
            });
        };

        $scope.errataFilter = function () {
            return true;
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ErrataFilterController
 *
 * @requires $scope
 *
 * @description
 *   Provides common functionality on the $scope for Errata filters.
 */

angular.module('Bastion.content-views').controller('ErrataFilterController',
    ['$scope', function ($scope) {
        $scope.rule = {
            errataType: 'all',
            'start_date': null,
            'end_date': null,
            'date_type': "updated"
        };

        $scope.date = {
            startOpen: false,
            endOpen: false
        };

        $scope.types = {
            enhancement: true,
            bugfix: true,
            security: true
        };

        $scope.errataFilter = function (errata) {
            var include = false,
                issued = new Date(errata.issued);

            if ($scope.types[errata.type]) {
                include = true;
            }

            if ($scope.rule['start_date']) {
                include = include && (issued.getTime() >= $scope.rule['start_date'].getTime());
            }

            if ($scope.rule['end_date']) {
                include = include && (issued.getTime() <= $scope.rule['end_date'].getTime());
            }

            return include;
        };

        $scope.openEndDate = function ($event) {
            $event.preventDefault();
            $event.stopPropagation();

            $scope.date.endOpen = true;
            $scope.date.startOpen = false;
        };

        $scope.openStartDate = function ($event) {
            $event.preventDefault();
            $event.stopPropagation();

            $scope.date.startOpen = true;
            $scope.date.endOpen = false;
        };

        $scope.onlySelected = function (object, type) {
            var selected = _.filter(object, function (value) {
                return value;
            });

            return object[type] && (selected.length === 1);
        };

    }]
);
/**
 * @ngdoc filter
 * @name  Bastion.content-views.filter:FilterContentType
 *
 * @requires FilterHelper
 *
 * @description
 *   A filter to turn a filter type into an eaiser to read string.
 *
 * @example
 *   {{ 'rpm' | filterContentType }} will produce the string "Packages".
 */

angular.module('Bastion.content-views').filter('filterContentType',
    ['FilterHelper', function (FilterHelper) {

        return function (type) {
            return FilterHelper.contentType(type);
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:FilterDetailsController
 *
 * @requires $scope
 * @requires Filter
 *
 * @description
 *   Handles fetching a filter.
 */

angular.module('Bastion.content-views').controller('FilterDetailsController',
    ['$scope', 'Filter', function ($scope, Filter) {

        $scope.filter = Filter.get({'content_view_id': $scope.$stateParams.contentViewId, filterId: $scope.$stateParams.filterId});

        $scope.updateFilter = function (filter) {
            filter.$update();
        };

    }]
);
/**
 * @ngdoc service
 * @name  Bastion.content-views.service:FilterHelper
 *
 * @requires translate
 *
 * @description
 *   Provides helper methods for common Filter operations.
 */

angular.module('Bastion.content-views').service('FilterHelper',
    ['translate', function (translate) {

        this.contentTypes = {
            'rpm': translate('RPM'),
            'erratum': translate('Errata'),
            'package_group': translate('Package Groups'),
            'docker': translate('Docker Tags')
        };

        this.contentType = function (type) {
            return this.contentTypes[type];
        };

        this.type = function (type) {
            return type ? translate('Include') : translate('Exclude');
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:FilterRepositoriesController
 *
 * @requires $scope
 * @requires translate
 * @requires Filter
 * @requires ContentViewRepositoriesUtl
 * @requires Notification
 *
 * @description
 *   Provides a way for users to select which repositories the filter applies to.
 */

angular.module('Bastion.content-views').controller('FilterRepositoriesController',
    ['$scope', 'translate', 'Filter', 'ContentViewRepositoriesUtil', 'Notification',
    function ($scope, translate, Filter, ContentViewRepositoriesUtil, Notification) {
        var refreshTable, success, error;

        ContentViewRepositoriesUtil($scope);

        refreshTable = function (filter) {
            var displayedRepositories = filter['content_view'].repositories,
                repositoryType,
                filterRepositories = filter.repositories;

            if ($scope.stateIncludes('content-view.yum')) {
                repositoryType = 'yum';
            } else {
                repositoryType = 'docker';
            }
            $scope.repositoryType = repositoryType;
            displayedRepositories = _.filter(displayedRepositories, ["content_type", repositoryType]);

            $scope.$parent.filter = filter;

            if (filterRepositories.length === 0) {
                displayedRepositories = _.map(displayedRepositories, function (repository) {
                    repository.selected = true;
                    return repository;
                });
            } else {
                displayedRepositories = _.map(displayedRepositories, function (repository) {
                    repository.selected = _.map(filterRepositories, 'id').indexOf(repository.id) >= 0;
                    return repository;
                });
            }

            $scope.table.rows = displayedRepositories;
            $scope.table.numSelected = displayedRepositories.length;
            $scope.table.resource = {
                subtotal: displayedRepositories.length
            };
            $scope.showRepos = filterRepositories.length !== 0;
        };

        success = function (filter) {
            refreshTable(filter);
            Notification.setSuccessMessage(translate('Affected repositories have been updated.'));
        };

        error = function (response) {
            angular.forEach(response.errors, function (responseError) {
                Notification.setErrorMessage(responseError);
            });
        };

        $scope.showRepos = false;
        $scope.table = {};

        $scope.filter.$promise.then(refreshTable);

        $scope.updateRepositories = function () {
            var repositoryIds = _.map($scope.table.getSelected(), 'id');

            if (repositoryIds.length === 0) {
                Notification.setErrorMessage(translate('You must select at least one repository.'));
            } else {
                Filter.update({id: $scope.filter.id, 'repository_ids': repositoryIds}, success, error);
            }
        };

        $scope.selectAllRepositories = function () {
            Filter.update({id: $scope.filter.id, 'repository_ids': []}, success, error);
        };
    }]
);
/**
 * @ngdoc filter
 * @name  Bastion.content-views.filter:FilterType
 *
 * @requires FilterHelper
 *
 * @description
 *   A filter to turn a filter type into an eaiser to read string.
 *
 * @example
 *   {{ filter.type | filterType }} will produce the string "Include" if true and "Exclude" if false;
 */

angular.module('Bastion.content-views').filter('filterType',
    ['FilterHelper', function (FilterHelper) {

        return function (type) {
            return FilterHelper.type(type);
        };

    }]
);
/**
 * @ngdoc service
 * @name  Bastion.content-views.factory:Filter
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for interacting with content view filters.
 */

angular.module('Bastion.content-views').factory('Filter',
    ['BastionResource', function (BastionResource) {

        return BastionResource('katello/api/v2/content_view_filters/:filterId/:action',
            {filterId: '@id', 'content_view_id': '@content_view.id'},
            {
                autocomplete: {method: 'GET', isArray: true, params: {filterId: 'auto_complete_search'}},
                update: {method: 'PUT'},
                availableErrata: {
                    method: 'GET',
                    params: {action: 'errata', 'available_for': 'content_view_filter'}
                },
                rules: {method: 'GET', transformResponse: function (data) {
                    var response = angular.fromJson(data),
                        rules = response.rules || [];

                    function getType(rule) {
                        var typeId;

                        if (rule.version) {
                            typeId = 'equal';
                        } else if (rule['min_version'] && !rule['max_version']) {
                            typeId = 'greater';
                        } else if (!rule['min_version'] && rule['max_version']) {
                            typeId = 'less';
                        } else if (rule['min_version'] && rule['max_version']) {
                            typeId = 'range';
                        } else {
                            typeId = 'all';
                        }

                        return typeId;
                    }

                    angular.forEach(rules, function (rule) {
                        rule.type = getType(rule);
                    });

                    return {results: rules, subtotal: rules.length, total: rules.length};
                }},
                errata: {
                    method: 'GET',
                    params: {action: 'errata'}
                },
                availablePackageGroups: {
                    method: 'GET',
                    params: {action: 'package_groups', 'available_for': 'content_view_filter'}
                },
                packageGroups: {
                    method: 'GET',
                    params: {action: 'package_groups'}
                }
            }
        );

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:FiltersController
 *
 * @requires $scope
 * @requires translate
 * @requires Filter
 * @requires Nutupane
 * @requires Notification
 *
 * @description
 *   Handles loading all filters for a content view.
 */

angular.module('Bastion.content-views').controller('FiltersController',
    ['$scope', 'translate', 'Filter', 'Nutupane', 'Notification', function ($scope, translate, Filter, Nutupane, Notification) {
        var nutupane, filterTypes;

        function removeFilter(id) {
            var success, failure;

            success = function () {
                nutupane.removeRow(id);
                Notification.setSuccessMessage(translate('Filters successfully removed.'));
            };

            failure = function (response) {
                Notification.setErrorMessage(response.data.displayMessage);
            };

            Filter.delete({filterId: id}, success, failure);
        }

        if ($scope.stateIncludes('content-view.yum')) {
            filterTypes = ['rpm', 'package_group', 'erratum'];
        } else {
            filterTypes = ['docker'];
        }
        nutupane = new Nutupane(Filter, {
            'content_view_id': $scope.$stateParams.contentViewId,
            'types[]': filterTypes
        });

        $scope.table = nutupane.table;

        $scope.$on('filter.created', function () {
            nutupane.refresh();
        });

        $scope.$on('filter.updated', function () {
            nutupane.refresh();
        });

        $scope.removeFilters = function () {
            var filterIds = nutupane.getAllSelectedResults().included.ids;

            angular.forEach(filterIds, removeFilter);
        };

        nutupane.table.closeItem = function () {};

        $scope.getFilterState = function (filter) {
            var state;

            switch (filter.type) {
            case "erratum":
                state = "content-view.yum.filter.erratum.list({filterId: filter.id})";
                if (filter.rules[0].types) {
                    state = "content-view.yum.filter.erratum.dateType({filterId: filter.id})";
                }
                break;
            case "rpm":
                state = "content-view.yum.filter.rpm({filterId: filter.id})";
                break;
            case "package_group":
                state = "content-view.yum.filter.package_group.list({filterId: filter.id})";
                break;
            }

            return state;
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:NewFilterController
 *
 * @requires $scope
 * @requires translate
 * @requires Filter
 * @requires Rule
 * @requires Notification
 *
 * @description
 */

angular.module('Bastion.content-views').controller('NewFilterController',
    ['$scope', 'translate', 'Filter', 'Rule', 'Notification', function ($scope, translate, Filter, Rule, Notification) {
        var filterType;

        function transitionToDetails(filter) {
            var state = '';

            if (filterType === 'erratumId') {
                state = 'content-view.yum.filter.erratum.available';
            } else if (filterType === 'erratumDateType') {
                state = 'content-view.yum.filter.erratum.dateType';
            } else if (filterType === 'rpm') {
                state = 'content-view.yum.filter.rpm.details';
            } else if (filterType === 'package_group') {
                state = 'content-view.yum.filter.package_group.available';
            } else if (filterType === 'docker') {
                state = 'content-view.docker.filter.tag.details';
            }

            $scope.$emit('filter.created');
            $scope.transitionTo(state, {filterId: filter.id, contentViewId: filter['content_view'].id});
        }

        function addErrataDateTypeRule(filter) {
            var rule = new Rule({
                    types: ['security', 'enhancement', 'bugfix']
                }),
                addSuccess, error;

            addSuccess = function () {
                transitionToDetails(filter);
            };

            error = function (response) {
                Notification.setErrorMessage(response.data.displayMessage);
            };

            rule.$save({filterId: filter.id}, addSuccess, error);
        }

        function success(filter) {
            if (filterType === 'erratumDateType') {
                addErrataDateTypeRule(filter);
            } else {
                transitionToDetails(filter);
            }
        }

        function failure(response) {
            $scope.working = false;
            angular.forEach(response.data.errors, function (errors, field) {
                $scope.filterForm[field].$setValidity('server', false);
                $scope.filterForm[field].$error.messages = errors;
            });
        }

        $scope.filter = new Filter();
        $scope.working = false;

        if ($scope.stateIncludes('content-view.yum')) {
            $scope.filterChoices = [
                {id: 'rpm', name: translate('Package')},
                {id: 'package_group', name: translate('Package Group')},
                {id: 'erratumId', name: translate('Erratum - by ID')},
                {id: 'erratumDateType', name: translate('Erratum - Date and Type')}
            ];
        } else {
            $scope.filter.type = "docker";
            $scope.filterChoices = [
                {id: 'docker', name: translate('Docker Tag')}
            ];
        }

        $scope.save = function (filter, contentView) {
            filterType = filter.type;

            if (filter.type === 'erratumId' || filter.type === 'erratumDateType') {
                filter.type = 'erratum';
            }

            filter.$save({'content_view_id': contentView.id}, success, failure);
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:PackageFilterController
 *
 * @requires $scope
 * @requires $location
 * @requires translate
 * @requires Nutupane
 * @requires CurrentOrganization
 * @requires Filter
 * @requires Rule
 * @requires Package
 * @requires Notification
 *
 * @description
 *   Handles package filter rules for a content view.
 */

angular.module('Bastion.content-views').controller('PackageFilterController',
    ['$scope', '$location', 'translate', 'Nutupane', 'CurrentOrganization', 'Filter', 'Rule', 'Package', 'Notification', function ($scope, $location, translate, Nutupane, CurrentOrganization, Filter, Rule, Package, Notification) {
        var nutupane, params;

        function failure(response) {
            Notification.setErrorMessage(response.data.displayMessage);
        }

        function createRule(rule) {
            var success;

            success = function (result) {
                rule.id = result.id;
                rule.editMode = false;
                rule.working = false;

                Notification.setSuccessMessage(translate('Package successfully added.'));
            };

            if ($scope.valid(rule)) {
                Rule.save({filterId: $scope.filter.id}, rule, success, failure);
            }
        }

        function updateRule(rule) {
            var updateParams = {filterId: $scope.filter.id, ruleId: rule.id},
                success, error;

            // Need access to the original rule
            success = function () {
                rule.previous = {};
                rule.editMode = false;
                rule.working = false;
                Notification.setSuccessMessage(translate('Package successfully updated.'));
            };

            error = function () {
                rule.working = false;
            };

            Rule.update(updateParams, rule, success, error);
        }

        function removeRule(rule) {
            var success, ruleId = rule.id;

            success = function () {
                nutupane.removeRow(ruleId);
                Notification.setSuccessMessage(translate('Package successfully removed.'));
            };

            Rule.delete({filterId: $scope.$stateParams.filterId, ruleId: ruleId}, success, failure);
        }

        params = {
            filterId: $scope.$stateParams.filterId,
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'sort_by': 'name',
            'sort_order': 'ASC',
            'paged': true
        };

        nutupane = new Nutupane(Filter, params, 'rules');
        $scope.table = nutupane.table;

        $scope.addRule = function () {
            var rule = new Rule();
            rule.editMode = true;
            $scope.table.addRow(rule);
        };

        $scope.saveRule = function (rule) {
            if (rule.id) {
                updateRule(rule);
            } else {
                createRule(rule);
            }
        };

        $scope.valid = function (rule) {
            var valid = rule.name ? true : false;

            if (rule.type === 'equal') {
                valid = valid && rule.version;
            } else if (rule.type === 'less') {
                valid = valid && rule['max_version'];
            } else if (rule.type === 'greater') {
                valid = valid && rule['min_version'];
            } else if (rule.type === 'range') {
                valid = valid && rule['min_version'] && rule['max_version'];
            }

            return valid;
        };

        $scope.clearValues = function (rule) {
            rule.version = undefined;
            rule['min_version'] = undefined;
            rule['max_version'] = undefined;
        };

        $scope.backupPrevious = function (rule) {
            rule.previous = {};
            angular.copy(rule, rule.previous);
        };

        $scope.restorePrevious = function (rule) {
            if (rule.id) {
                angular.copy(rule.previous, rule);
            } else {
                $scope.table.rows.shift();
            }
            rule.previous = {};
        };

        $scope.removeRules = function () {
            angular.forEach($scope.table.getSelected(), function (rule) {
                removeRule(rule);
            });
        };

        $scope.fetchAutocompleteName = function (term) {
            var repositoryIds = $scope.contentView['repository_ids'],
                promise;

            promise = Package.autocompleteName({'repoids[]': repositoryIds, term: term}).$promise;

            return promise.then(function (data) {
                return data.results;
            });
        };

        $scope.fetchAutocompleteArch = function (term) {
            var repositoryIds = $scope.contentView['repository_ids'],
                promise;

            promise = Package.autocompleteArch({'repoids[]': repositoryIds, term: term}).$promise;

            return promise.then(function (data) {
                return data.results;
            });
        };

        $scope.filterRepositoriesByType = function () {
            return (_.filter($scope.filter.repositories, ["content_type", "yum"]));
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:PackageFilterListController
 *
 * @requires $scope
 * @requires translate
 * @requires Filter
 * @requires Rule
 * @requires Nutupane
 * @requires Notification
 *
 * @description
 *   Handles loading package groups that have been added to the filter via filter rules
 *   and provides a method to remove them.
 */

angular.module('Bastion.content-views').controller('PackageGroupFilterListController',
    ['$scope', 'translate', 'PackageGroup', 'Rule', 'Nutupane', 'Notification',
    function ($scope, translate, PackageGroup, Rule, Nutupane, Notification) {
        var nutupane;

        function success(rule) {
            nutupane.removeRow(rule.uuid, 'uuid');
            $scope.filter.rules = _.reject($scope.filter.rules, function (filterRule) {
                return rule.id === filterRule.id;
            });
            Notification.setSuccessMessage(translate('Package Group successfully removed.'));
        }

        function failure(response) {
            Notification.setErrorMessage(response.data.displayMessage);
        }

        function findRules(packageGroupIds) {
            var rules = [];

            angular.forEach(packageGroupIds, function (id) {
                var found;

                found = _.find($scope.filter.rules, function (rule) {
                    return (rule.uuid === id);
                });

                if (found) {
                    rules.push(new Rule(found));
                }
            });

            return rules;
        }

        nutupane = new Nutupane(PackageGroup, {
            filterId: $scope.$stateParams.filterId,
            'sort_order': 'ASC',
            'sort_by': 'name'
            },
            'queryUnpaged'
        );
        $scope.controllerName = 'katello_package_groups';

        $scope.table = nutupane.table;
        nutupane.table.closeItem = function () {};

        $scope.removePackageGroups = function () {
            var packageGroupIds = nutupane.getAllSelectedResults('uuid').included.ids,
                rules;

            rules = findRules(packageGroupIds);

            angular.forEach(rules, function (rule) {
                rule.$delete(success, failure);
            });
        };

    }]
);
/**
 * @ngdoc service
 * @name  Bastion.content-views.filters.factory:Filter
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for interacting with content view filter rules.
 */

angular.module('Bastion.content-views').factory('Rule',
    ['BastionResource', function (BastionResource) {

        return BastionResource('katello/api/v2/content_view_filters/:filterId/rules/:ruleId',
            {ruleId: '@id', filterId: '@content_view_filter_id'},
            {
                update: {method: 'PUT'}
            }
        );

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewHistoryController
 *
 * @requires $scope
 * @requires ContentView
 * @requires Nutupane
 *
 * @description
 *   A controller for showing the history of a content view
 */

angular.module('Bastion.content-views').controller('ContentViewHistoryController',
    ['$scope', 'translate', 'ContentViewHistory', 'Nutupane',
    function ($scope, translate, ContentViewHistory, Nutupane) {
        var nutupane;

        nutupane = new Nutupane(ContentViewHistory, {
            contentViewId: $scope.$stateParams.contentViewId,
            'sort_by': 'created_at',
            'sort_order': 'DESC'
        });
        $scope.controllerName = 'katello_content_views';

        nutupane.table.closeItem = function () {};
        $scope.table = nutupane.table;

        $scope.actionText = function (history) {
            var message,
                taskType = history.task.label,
                taskTypes = $scope.taskTypes;

            if (taskType === taskTypes.deletion) {
                message = translate("Deleted from %s").replace('%s', history.environment.name);
            } else if (taskType === taskTypes.promotion) {
                message = translate("Promoted to %s").replace('%s', history.environment.name);
            } else if (taskType === taskTypes.publish) {
                message = translate("Published new version");
            } else if (taskType === taskTypes.export) {
                message = translate("Exported content view");
            } else if (taskType === taskTypes.incrementalUpdate) {
                message = translate("Incremental update");
            }

            return message;
        };
    }]
);
/**
 * @ngdoc service
 * @name  Bastion.content-views.factory:ContentViewHistory
 *
 * @requires BastionResource
 * @requires CurrentOrganization
 *
 * @description
 *   Provides a BastionResource for interacting with content view histories.
 */

angular.module('Bastion.content-views').factory('ContentViewHistory',
    ['BastionResource', 'CurrentOrganization',
    function (BastionResource, CurrentOrganization) {

        return BastionResource('katello/api/v2/content_views/:contentViewId/history/:action',
            {id: '@id', contentViewId: '@contentViewId', 'organization_id': CurrentOrganization},
            {
                autocomplete: {method: 'GET', isArray: true, params: {action: 'auto_complete_search'}}
            }
        );
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewPuppetModulesController
 *
 * @requires $scope
 * @requires Nutupane
 * @requires ContentView
 *
 * @description
 *   Provides functionality to the puppet modules name list.
 */

angular.module('Bastion.content-views').controller('ContentViewPuppetModuleNamesController',
    ['$scope', 'Nutupane', 'ContentView', 'CurrentOrganization', 'PuppetModule',
    function ($scope, Nutupane, ContentView, CurrentOrganization, PuppetModule) {

        var nutupane = new Nutupane(
            ContentView,
            {id: $scope.$stateParams.contentViewId},
            'availablePuppetModuleNames'
        );
        $scope.controllerName = 'katello_content_views';
        nutupane.masterOnly = true;
        $scope.table = nutupane.table;

        $scope.table.fetchAutocomplete = function (term) {
            var promise;

            promise = PuppetModule.autocomplete({'organization_id': CurrentOrganization, search: term}).$promise;

            return promise.then(function (data) {
                return data;
            });
        };

        $scope.selectVersion = function (moduleName) {
            $scope.transitionTo('content-view.puppet-modules.versions',
                {
                    contentViewId: $scope.$stateParams.contentViewId,
                    moduleName: moduleName
                }
            );
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewPuppetModulesController
 *
 * @requires $scope
 * @requires translate
 * @requires Nutupane
 * @requires ContentView
 * @requires ContentViewPuppetModule
 * @requires Notification
 *
 * @description
 *   Provides the ability to select a version of a Puppet Module for a Content View.
 */

angular.module('Bastion.content-views').controller('ContentViewPuppetModuleVersionsController',
    ['$scope', 'translate', 'Nutupane', 'ContentView', 'ContentViewPuppetModule', 'Notification',
    function ($scope, translate, Nutupane, ContentView, ContentViewPuppetModule, Notification) {
        var success, error, nutupane, params;

        params = {
            name: $scope.$stateParams.moduleName,
            id: $scope.$stateParams.contentViewId,
            'full_result': true
        };

        nutupane = new Nutupane(ContentView, params, 'availablePuppetModules');
        $scope.controllerName = 'katello_content_views';
        nutupane.masterOnly = true;
        $scope.table = nutupane.table;

        $scope.selectVersion = function (module) {
            var contentViewPuppetModule, contentViewPuppetModuleData = {
                contentViewId: $scope.$stateParams.contentViewId,
                uuid: module.uuid,
                author: module.author,
                name: module.name
            };

            contentViewPuppetModule = new ContentViewPuppetModule(contentViewPuppetModuleData);

            if ($scope.$stateParams.moduleId) {
                contentViewPuppetModule.id = $scope.$stateParams.moduleId;
                contentViewPuppetModule.$update(success, error);
            } else {
                contentViewPuppetModule.$save(success, error);
            }
        };

        success = function () {
            $scope.transitionTo('content-view.puppet-modules.list',
                {contentViewId: $scope.$stateParams.contentViewId});
            Notification.setSuccessMessage(translate('Puppet module added to Content View'));
        };

        error = function (response) {
            angular.forEach(response.data.errors, function (errorMessage) {
                Notification.setErrorMessage(translate("An error occurred updating the Content View: ") + errorMessage);
            });
        };
    }]
);
/**
 * @ngdoc service
 * @name  Bastion.content-views.factory:ContentViewPuppetModule
 *
 * @requires BastionResource
 * @requires CurrentOrganization
 *
 * @description
 *   Provides a BastionResource for interacting with content view puppet modules.
 */

angular.module('Bastion.content-views').factory('ContentViewPuppetModule',
    ['BastionResource', 'CurrentOrganization',
    function (BastionResource, CurrentOrganization) {

        return BastionResource('katello/api/v2/content_views/:contentViewId/content_view_puppet_modules/:id/:action',
            {id: '@id', contentViewId: '@contentViewId', 'organization_id': CurrentOrganization},
            {
                update: {method: 'PUT'},
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}}
            }
        );
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:ContentViewPuppetModulesController
 *
 * @requires $scope
 * @requires translate
 * @requires Nutupane
 * @requires ContentViewPuppetModule
 * @requires Notification
 *
 * @description
 *   Provides functionality to the Content View existing Puppet Modules list.
 */

angular.module('Bastion.content-views').controller('ContentViewPuppetModulesController',
    ['$scope', 'translate', 'Nutupane', 'ContentViewPuppetModule', 'Notification',
    function ($scope, translate, Nutupane, ContentViewPuppetModule, Notification) {
        var nutupane = new Nutupane(ContentViewPuppetModule, {
            contentViewId: $scope.$stateParams.contentViewId
        });

        nutupane.masterOnly = true;

        $scope.table = nutupane.table;

        $scope.versionText = function (module) {
            var version;
            if (module['computed_version']) {
                version = translate("Latest (Currently %s)").replace('%s', module['computed_version']);
            } else {
                version = translate("Unable to determine version");
            }
            if (module['puppet_module']) {
                version = module['puppet_module'].version;
            }
            return version;
        };

        $scope.selectNewVersion = function (module) {
            $scope.transitionTo('content-view.puppet-modules.versionsForModule',
                {
                    contentViewId: $scope.$stateParams.contentViewId,
                    moduleName: module.name,
                    moduleId: module.id
                }
            );
        };

        $scope.removeModule = function (module) {
            var success, error;

            success = function () {
                Notification.setSuccessMessage(translate('Module %s removed from Content View.')
                    .replace('%s', module.name));
                nutupane.removeRow(module.id);
            };

            error = function (response) {
                angular.forEach(response.data.errors, function (errorMessage) {
                    Notification.setErrorMessage(translate("An error occurred updating the Content View: ") + errorMessage);
                });
            };

            ContentViewPuppetModule.remove({
                contentViewId: $scope.$stateParams.contentViewId,
                id: module.id
            }, success, error);
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.content-views.controller:NewContentViewController
 *
 * @requires $scope
 * @requires ContentView
 * @requires FormUtils
 * @requires CurrentOrganization
 *
 * @description
 */

angular.module('Bastion.content-views').controller('NewContentViewController',
    ['$scope', 'ContentView', 'FormUtils', 'CurrentOrganization',
    function ($scope, ContentView, FormUtils, CurrentOrganization) {

        function success(response) {
            var successState = 'content-view.repositories.yum.available';

            if (response.composite) {
                successState = 'content-view.components.composite-content-views.available';
            }

            $scope.transitionTo(successState, {contentViewId: response.id});
        }

        function error(response) {
            $scope.working = false;
            angular.forEach(response.data.errors, function (errors, field) {
                $scope.contentViewForm[field].$setValidity('server', false);
                $scope.contentViewForm[field].$error.messages = errors;
            });
        }

        $scope.contentView = new ContentView({'organization_id': CurrentOrganization});
        $scope.createOption = 'new';
        $scope.table = {};

        $scope.save = function (contentView) {
            contentView.$save(success, error);
        };

        $scope.$watch('contentView.name', function () {
            if ($scope.contentViewForm.name) {
                $scope.contentViewForm.name.$setValidity('server', true);
                FormUtils.labelize($scope.contentView);
            }
        });

    }]
);
(function () {
    'use strict';

    /**
     * @ngdoc module
     * @name  Bastion.docker-manifests
     *
     * @description
     *   Module for Docker Manifest related functionality.
     */
    angular
        .module('Bastion.docker-manifests', [
            'ui.router',
            'Bastion',
            'Bastion.i18n',
            'Bastion.components'
        ]);

})();
(function () {
    'use strict';

    /**
     * @ngdoc factory
     * @name  Bastion.docker-manifests.factory:DockerManifest
     *
     * @description
     *   Provides a BastionResource for interacting with Docker Manifests
     */
    function DockerManifest(BastionResource) {
        return BastionResource('katello/api/v2/docker_manifests/:id',
            {'id': '@id'},
            {
                'autocomplete': {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}}
            }
        );
    }

    angular
        .module('Bastion.docker-manifests')
        .factory('DockerManifest', DockerManifest);

    DockerManifest.$inject = ['BastionResource'];

})();
(function () {
    'use strict';

    /**
     * @ngdoc module
     * @name  Bastion.docker-manifest-lists
     *
     * @description
     *   Module for Docker Manifest List related functionality.
     */
    angular
        .module('Bastion.docker-manifest-lists', [
            'ui.router',
            'Bastion',
            'Bastion.i18n',
            'Bastion.components'
        ]);

})();
(function () {
    'use strict';

    /**
     * @ngdoc factory
     * @name  Bastion.docker-manifest-lists.factory:DockerManifestList
     *
     * @description
     *   Provides a BastionResource for interacting with Docker Manifest Lists
     */
    function DockerManifestList(BastionResource) {
        return BastionResource('katello/api/v2/docker_manifest_lists/:id',
            {'id': '@id'},
            {
                'autocomplete': {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}}
            }
        );
    }

    angular
        .module('Bastion.docker-manifest-lists')
        .factory('DockerManifestList', DockerManifestList);

    DockerManifestList.$inject = ['BastionResource'];

})();
/**
 * @ngdoc module
 * @name  Bastion.docker-tags
 *
 * @description
 *   Module for Docker Tags related functionality.
 */

angular.module('Bastion.docker-tags', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.components',
    'Bastion.components.formatters'
]);
/**
 * @ngdoc object
 * @name  Bastion.docker-tags.controller:DockerTagDetailsController
 *
 * @requires $scope
 * @requires $location
 * @requires DockerTag
 * @requires CurrentOrganization
 * @requires ApiErrorHandler
 * @requires translate
 *
 * @description
 *   Provides the functionality for the docker tags details action pane.
 */

angular.module('Bastion.docker-tags').controller('DockerTagDetailsController',
    ['$scope', '$location', 'DockerTag', 'CurrentOrganization', 'ApiErrorHandler', 'translate',
    function ($scope, $location, DockerTag, CurrentOrganization, ApiErrorHandler, translate) {
        $scope.panel = {
            error: false,
            loading: true
        };

        if ($scope.tag) {
            $scope.panel.loading = false;
        }

        $scope.tag = DockerTag.get({id: $scope.$stateParams.tagId}, function (data) {
            if (data.manifest_schema1) {
                data.manifest_schema1["manifest_type_display"] = $scope.getManifestDisplayType(data.manifest_schema1);
            }
            if (data.manifest_schema2) {
                data.manifest_schema2["manifest_type_display"] = $scope.getManifestDisplayType(data.manifest_schema2);
            }
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        $scope.getManifestDisplayType = function (schema) {
            if (schema['manifest_type'] === 'image') {
                return translate("Image");
            }
            return translate("List");
        };
    }
]);
/**
 * @ngdoc object
 * @name  Bastion.docker-tags.controller:DockerTagDetailsController
 *
 * @requires $scope
 * @requires $location
 * @requires DockerTag
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality for the docker tags details environments list.
 */

angular.module('Bastion.docker-tags').controller('DockerTagEnvironmentsController',
    ['$scope', '$location', 'Nutupane', 'DockerTag', 'CurrentOrganization',
    function ($scope, $location, Nutupane, DockerTag, CurrentOrganization) {
        var params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'sort_by': 'name',
            'sort_order': 'ASC',
            'paged': false
        };
        var nutupane = new Nutupane(DockerTag, params, null, {disableAutoLoad: true});

        var renderTable = function () {
            var ids = _.map($scope.tag.related_tags, 'id');
            var newParams = {
                'organization_id': CurrentOrganization,
                'search': $location.search().search || "",
                'sort_by': 'name',
                'sort_order': 'ASC',
                'paged': false,
                'ids[]': ids
            };
            $scope.table = nutupane.table;
            nutupane.setParams(newParams);
            $scope.panel.loading = false;
            if (!_.isEmpty(ids)) {
                nutupane.refresh();
            }
        };

        $scope.controllerName = 'katello_docker_tags';

        if ($scope.tag) {
            $scope.panel.loading = false;
        }

        if ($scope.tag && $scope.tag.related_tags) {
            renderTable();
        } else {
            $scope.tag.$promise.then(renderTable);
        }
    }
]);
/**
 * @ngdoc object
 * @name  Bastion.docker-tags.controller:DockerTagsController
 *
 * @requires $scope
 * @requires $location
 * @requires Nutupane
 * @requires DockerTag
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality specific to docker tags for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.docker-tags').controller('DockerTagsController',
    ['$scope', '$location', 'Nutupane', 'DockerTag', 'CurrentOrganization',
    function ($scope, $location, Nutupane, DockerTag, CurrentOrganization) {

        var params = {
            'organization_id': CurrentOrganization,
            'sort_by': 'name',
            'sort_order': 'ASC',
            'grouped': true
        };

        var nutupane = new Nutupane(DockerTag, params);
        $scope.controllerName = 'katello_docker_tags';
        $scope.table = nutupane.table;

        $scope.table.closeItem = function () {
            $scope.transitionTo('docker-tags');
        };

        $scope.availableSchemaVersions = function (tag) {
            var versions = [];
            if (tag.manifest_schema1) {
                versions.push(1);
            }

            if (tag.manifest_schema2) {
                versions.push(2);
            }
            return versions.join(", ");
        };
    }]
);
/**
 * @ngdoc service
 * @name  Bastion.docker-tags.factory:DockerTag
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for Docker Tags
 */

angular.module('Bastion.docker-tags').factory('DockerTag',
    ['BastionResource', function (BastionResource) {

        return BastionResource('katello/api/v2/docker_tags/:id/',
            {id: '@id'},
            {
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}},
                'autocompleteName': {method: 'GET', isArray: false, params: {id: 'auto_complete_name'},
                    transformResponse: function (data) {
                        data = angular.fromJson(data);
                        return {results: data};
                    }
                }
            }
        );

    }]
);
/**
 * @ngdoc object
 * @name Bastion.docker-tags.config
 *
 * @requires $stateProvider
 *
 * @description
 *   State routes defined for the docker tags module.
 */

angular.module('Bastion.docker-tags').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('docker-tags', {
        url: '/docker_tags',
        permission: ['view_products', 'view_content_views'],
        template: '<div ui-view></div>',
        views: {
            '@': {
                controller: 'DockerTagsController',
                templateUrl: 'docker-tags/views/docker-tags.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{ 'Docker Tags' | translate }}"
        }
    })
    .state('docker-tag', {
        abstract: true,
        url: '/docker_tags/:tagId',
        permission: 'view_products',
        controller: 'DockerTagDetailsController',
        templateUrl: 'docker-tags/details/views/docker-tag-details.html',
        ncyBreadcrumb: {
            label: "{{ 'Docker Tags' | translate }}",
            parent: 'docker-tags'
        }
    })
    .state('docker-tag.info', {
        url: '',
        permission: 'view_products',
        templateUrl: 'docker-tags/details/views/docker-tag-info.html',
        ncyBreadcrumb: {
            label: "{{ tag.name }}",
            parent: 'docker-tag'
        }
    })
    .state('docker-tag.environments', {
        url: '/environments',
        permission: 'view_environments',
        templateUrl: 'docker-tags/details/views/docker-tag-environments.html',
        controller: 'DockerTagEnvironmentsController',
        ncyBreadcrumb: {
            label: "{{ 'Lifecycle Environments' | translate }}",
            parent: 'docker-tag.info'
        }
    });
}]);
(function () {
    'use strict';

    /**
     * @ngdoc module
     * @name  Bastion.files
     *
     * @description
     *   Module for File related functionality.
     */
    angular
        .module('Bastion.files', [
            'ngResource',
            'ui.router',
            'Bastion',
            'Bastion.common',
            'Bastion.i18n'
        ]);

})();
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.files.controller:FileContentViewsController
     *
     * @description
     *   Provides the functionality for the files details contentViews page.
     */
    function FileContentViewsController($scope, Nutupane, ContentViewVersion, CurrentOrganization) {
        var contentViewsNutupane,
            params = {
                'file_id': $scope.$stateParams.fileId,
                'organization_id': CurrentOrganization
            };

        contentViewsNutupane = new Nutupane(ContentViewVersion, params);
        $scope.controllerName = 'katello_content_views';
        contentViewsNutupane.masterOnly = true;
        contentViewsNutupane.setSearchKey('contentViewsSearch');

        $scope.table = contentViewsNutupane.table;

        $scope.environmentNames = function (environments) {
            var names = _.map(environments, function (environment) {
                return environment.name;
            });

            return names.join(',');
        };
    }

    angular
        .module('Bastion.files')
        .controller('FileContentViewsController', FileContentViewsController);

    FileContentViewsController.$inject = [
        '$scope',
        'Nutupane',
        'ContentViewVersion',
        'CurrentOrganization'
    ];

})();
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.files.controller:FileRepositoriesController
     *
     * @description
     *   Provides the functionality for the files details repositories page.
     */
    function FileRepositoriesController($scope, Nutupane, Repository, CurrentOrganization) {
        var repositoriesNutupane,
            params = {
                'file_id': $scope.$stateParams.fileId,
                'organization_id': CurrentOrganization
            };

        repositoriesNutupane = new Nutupane(Repository, params);
        $scope.controllerName = 'katello_repositories';
        repositoriesNutupane.masterOnly = true;
        repositoriesNutupane.setSearchKey('repositoriesSearch');

        $scope.table = repositoriesNutupane.table;
    }

    angular
        .module('Bastion.files')
        .controller('FileRepositoriesController', FileRepositoriesController);

    FileRepositoriesController.$inject = [
        '$scope',
        'Nutupane',
        'Repository',
        'CurrentOrganization'
    ];

})();
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.files.controller:FileController
     *
     * @description
     *   Provides the functionality for the files details action pane.
     */
    function FileController($scope, File, ApiErrorHandler) {
        $scope.panel = {
            error: false,
            loading: true
        };

        if ($scope.file) {
            $scope.panel.loading = false;
        }

        $scope.file = File.get({id: $scope.$stateParams.fileId}, function () {
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });
    }

    angular
        .module('Bastion.files')
        .controller('FileController', FileController);

    FileController.$inject = ['$scope', 'File', 'ApiErrorHandler'];

})();
(function () {
    'use strict';

    /**
     * @ngdoc factory
     * @name  Bastion.files.factory:File
     *
     * @description
     *   Provides a BastionResource for interacting with Files
     */
    function File(BastionResource) {

        return BastionResource('katello/api/v2/files/:id',
            {'id': '@id'},
            {
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}}
            }
        );

    }

    angular
        .module('Bastion.files')
        .factory('File', File);

    File.$inject = ['BastionResource'];

})();
(function () {
    'use strict';

    /**
     * @ngdoc controller
     * @name  Bastion.files.controller:FilesController
     *
     * @description
     *   Handles fetching files and populating Nutupane based on the current
     *   ui-router state.
     */
    function FilesController($scope, Nutupane, File, CurrentOrganization) {
        var nutupane;

        var params = {
            'organization_id': CurrentOrganization,
            'sort_by': 'name',
            'sort_order': 'ASC'
        };

        nutupane = new Nutupane(File, params);
        $scope.controllerName = 'katello_files';
        nutupane.masterOnly = true;

        $scope.table = nutupane.table;
        $scope.controllerName = 'katello_files';
    }

    angular
        .module('Bastion.files')
        .controller('FilesController', FilesController);

    FilesController.$inject = ['$scope', 'Nutupane', 'File', 'CurrentOrganization'];

})();
(function () {
    'use strict';

    /**
     * @ngdoc object
     * @name Bastion.files.config
     *
     * @requires $stateProvider
     *
     * @description
     *   State routes defined for the files module
     */
    function FilesConfig($stateProvider) {
        $stateProvider.state('files', {
            url: '/files',
            permission: ['view_products', 'view_content_views'],
            views: {
                '@': {
                    controller: 'FilesController',
                    templateUrl: 'files/views/files.html'
                }
            },
            ncyBreadcrumb: {
                label: "{{ 'Files' | translate }}"
            }
        })
        .state('file', {
            abstract: true,
            url: '/files/:fileId',
            permission: ['view_products', 'view_content_views'],
            controller: 'FileController',
            templateUrl: 'files/details/views/file.html'
        })
        .state('file.info', {
            url: '',
            permission: ['view_products', 'view_content_views'],
            templateUrl: 'files/details/views/file-info.html',
            ncyBreadcrumb: {
                label: "{{ file.name }}",
                parent: 'files'
            }
        })
        .state('file.repositories', {
            url: '/repositories',
            permission: ['view_products'],
            controller: 'FileRepositoriesController',
            templateUrl: 'files/details/views/file-repositories.html',
            ncyBreadcrumb: {
                label: "{{ 'Repositories' | translate }}",
                parent: 'file.info'
            }
        })
        .state('file.content-views', {
            url: '/content_views',
            collapsed: true,
            permission: ['view_content_views'],
            controller: 'FileContentViewsController',
            templateUrl: 'files/details/views/file-content-views.html',
            ncyBreadcrumb: {
                label: "{{ 'Content Views' | translate }}",
                parent: 'file.info'
            }
        });
    }

    angular
        .module('Bastion.files')
        .config(FilesConfig);

    FilesConfig.$inject = ['$stateProvider'];
})();
/**
 * @ngdoc module
 * @name  Bastion.ostree-branches
 *
 * @description
 *   Module for Ostree Branch related functionality.
 */

angular.module('Bastion.ostree-branches', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.common',
    'Bastion.components'
]);
/**
 * @ngdoc object
 * @name  Bastion.ostree-branches.controller:OstreeBranchRepositoriesController
 *
 * @requires $scope
 * @requires $q
 * @requires Nutupane
 * @requires Repository
 * @requires Environment
 * @requires ContentView
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality for the OSTree branch repositories page.
 */

angular.module('Bastion.ostree-branches').controller('OstreeBranchRepositoriesController',
    ['$scope', '$q', 'Nutupane', 'Repository', 'Environment', 'ContentView', 'CurrentOrganization',
    function ($scope, $q, Nutupane, Repository, Environment, ContentView, CurrentOrganization) {
        var repositoriesNutupane, environment, contentView, params;
        params = {
            'ostree_branch_id': $scope.$stateParams.branchId,
            'organization_id': CurrentOrganization,
            'paged': false
        };

        repositoriesNutupane = new Nutupane(Repository, params);
        $scope.controllerName = 'katello_repositories';
        $scope.table = repositoriesNutupane.table;
        $scope.table.initialLoad = false;
        repositoriesNutupane.masterOnly = true;
        repositoriesNutupane.setSearchKey('repositoriesSearch');

        environment = Environment.queryUnpaged(function (response) {
            $scope.environments = response.results;
            $scope.environmentFilter = _.find($scope.environments, {library: true}).id;
        });

        contentView = ContentView.queryUnpaged(function (response) {
            $scope.contentViews = response.results;
            $scope.contentViewFilter = _.find($scope.contentViews, {'default': true});
        });

        $scope.table.working = true;
        $q.all([contentView.$promise, environment.$promise]).then(function () {
            $scope.filterBranches();
            $scope.table.working = false;
        });

        $scope.filterBranches = function () {
            params['environment_id'] = $scope.environmentFilter;
            params['content_view_version_id'] = $scope.contentViewFilter;

            if ($scope.contentViewFilter) {
                params['content_view_version_id'] = _.map($scope.contentViewFilter.versions, 'id');
            }

            repositoriesNutupane.setParams(params);
            repositoriesNutupane.refresh();
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.ostree-branches.controller:OstreeBranchController
 *
 * @requires $scope
 * @requires OstreeBranch
 * @requires ApiErrorHandler
 *
 * @description
 *   Provides the functionality for the Ostree Branch action pane.
 */

angular.module('Bastion.ostree-branches').controller('OstreeBranchController', ['$scope', 'OstreeBranch', 'ApiErrorHandler',
    function ($scope, OstreeBranch, ApiErrorHandler) {
        $scope.panel = {
            error: false,
            loading: true
        };

        if ($scope.branch) {
            $scope.panel.loading = false;
        }

        $scope.branch = OstreeBranch.get({id: $scope.$stateParams.branchId}, function (branch) {
            $scope.$broadcast('branch.loaded', branch);
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });
    }
]);
(function () {
    'use strict';

    /**
     * @ngdoc factory
     * @name  Bastion.ostree-branches.factory:OstreeBranch
     *
     * @description
     *   Provides a BastionResource for interacting with Ostree Branches
     */
    function OstreeBranch(BastionResource) {
        return BastionResource('katello/api/v2/ostree_branches/:id',
            {'id': '@id'},
            {
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}}
            }

        );
    }

    angular
        .module('Bastion.ostree-branches')
        .factory('OstreeBranch', OstreeBranch);

    OstreeBranch.$inject = ['BastionResource'];

})();
/**
 * @ngdoc object
 * @name  Bastion.ostree-branches.controller:OstreeBranchesController
 *
 * @requires $scope
 * @requires $location
 * @requires translate
 * @requires Nutupane
 * @requires OstreeBranch
 * @requires Repository
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality specific to ostree branches for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */


angular.module('Bastion.ostree-branches').controller('OstreeBranchesController',
    ['$scope', '$location', 'translate', 'Nutupane', 'OstreeBranch', 'Repository', 'CurrentOrganization',
    function ($scope, $location, translate, Nutupane, OstreeBranch, Repository, CurrentOrganization) {
        var nutupane, params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'paged': true
        };

        nutupane = $scope.nutupane = new Nutupane(OstreeBranch, params);
        $scope.controllerName = 'katello_ostree_branches';
        $scope.table = nutupane.table;
        $scope.removeRow = nutupane.removeRow;

        $scope.table.closeItem = function () {
            $scope.transitionTo('ostree-branches.index');
        };

        $scope.repository = {name: translate('All Repositories'), id: 'all'};

        Repository.queryUnpaged({'organization_id': CurrentOrganization, 'content_type': 'ostree'}, function (response) {
            $scope.repositories = [$scope.repository];
            $scope.repositories = $scope.repositories.concat(response.results);

            if ($location.search().repositoryId) {
                $scope.repository = _.find($scope.repositories, function (repository) {
                    return repository.id === parseInt($location.search().repositoryId, 10);
                });
            }
        });

        $scope.$watch('repository', function (repository) {
            var nutupaneParams = nutupane.getParams();

            if (repository.id === 'all') {
                nutupaneParams['repository_id'] = null;
                nutupane.setParams(nutupaneParams);
            } else {
                $location.search('repositoryId', repository.id);
                nutupaneParams['repository_id'] = repository.id;
                nutupane.setParams(nutupaneParams);
            }

            if (!nutupane.table.initialLoad) {
                nutupane.refresh();
            }
        });
    }]
);
/**
 * @ngdoc object
 * @name Bastion.ostree-branches.config
 *
 * @requires $stateProvider
 *
 * @description
 *   State routes defined for the ostree branches module.
 */

angular.module('Bastion.ostree-branches').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('ostree-branches', {
        url: '/ostree_branches',
        permission: ['view_products', 'view_content_views'],
        views: {
            '@': {
                controller: 'OstreeBranchesController',
                templateUrl: 'ostree-branches/views/ostree-branches.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{ 'OSTree Branches' | translate }}"
        }
    })
    .state('ostree-branch', {
        abstract: true,
        url: '/ostree_branches/:branchId',
        permission: ['view_products', 'view_content_views'],
        controller: 'OstreeBranchController',
        templateUrl: 'ostree-branches/details/views/ostree-branch.html'
    })
    .state('ostree-branch.info', {
        url: '',
        permission: ['view_products', 'view_content_views'],
        templateUrl: 'ostree-branches/details/views/ostree-branch-info.html',
        ncyBreadcrumb: {
            label: "{{ branch.name }}",
            parent: 'ostree-branches'
        }
    })
    .state('ostree-branch.repositories', {
        url: '/repositories',
        permission: ['view_products', 'view_content_views'],
        controller: 'OstreeBranchRepositoriesController',
        templateUrl: 'ostree-branches/details/views/ostree-branch-repositories.html',
        ncyBreadcrumb: {
            label: "{{ 'Repositories' | translate }}",
            parent: 'ostree-branch.info'
        }
    });
}]);
(function () {
    'use strict';

    /**
     * @ngdoc module
     * @name  Bastion.debs
     *
     * @description
     *   Module for Deb Packages related functionality.
     */
    angular
        .module('Bastion.debs', [
            'ngResource',
            'ui.router',
            'Bastion',
            'Bastion.common',
            'Bastion.i18n'
        ]);

})();
(function () {
    'use strict';

    /**
     * @ngdoc factory
     * @name  Bastion.debs.factory:Deb
     *
     * @description
     *   Provides a BastionResource for interacting with deb Packages
     */
    function Deb(BastionResource) {

        return BastionResource('katello/api/v2/debs/:id',
            {'id': '@id'},
            {
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}}
            }
        );

    }

    angular
        .module('Bastion.debs')
        .factory('Deb', Deb);

    Deb.$inject = ['BastionResource'];

})();
(function () {
    'use strict';

    /**
     * @ngdoc controller
     * @name  Bastion.debs.controller:DebsController
     *
     * @description
     *   Handles fetching deb packages and populating Nutupane based on the current
     *   ui-router state.
     */
    function DebsController($scope, Nutupane, Deb, CurrentOrganization) {
        var nutupane;

        var params = {
            'organization_id': CurrentOrganization,
            'sort_by': 'name',
            'sort_order': 'ASC'
        };

        nutupane = new Nutupane(Deb, params);
        nutupane.masterOnly = true;

        $scope.table = nutupane.table;
        $scope.controllerName = 'katello_debs';
    }

    angular
        .module('Bastion.debs')
        .controller('DebsController', DebsController);

    DebsController.$inject = ['$scope', 'Nutupane', 'Deb', 'CurrentOrganization'];

})();
(function () {
    'use strict';

    /**
     * @ngdoc object
     * @name Bastion.debs.config
     *
     * @requires $stateProvider
     *
     * @description
     *   State routes defined for the debs module
     */
    function DebsConfig($stateProvider) {
        $stateProvider.state('debs', {
            url: '/debs',
            permission: ['view_products', 'view_content_views'],
            views: {
                '@': {
                    controller: 'DebsController',
                    templateUrl: 'debs/views/debs.html'
                }
            },
            ncyBreadcrumb: {
                label: "{{ 'Debs' | translate }}"
            }
        })
        .state('deb', {
            abstract: true,
            url: '/debs/:debId',
            permission: ['view_products', 'view_content_views'],
            controller: 'DebController',
            templateUrl: 'debs/details/views/deb.html'
        })
        .state('deb.info', {
            url: '',
            permission: ['view_products', 'view_content_views'],
            templateUrl: 'debs/details/views/deb-info.html',
            ncyBreadcrumb: {
                label: "{{ deb.name }}",
                parent: 'debs'
            }
        })
        .state('deb.repositories', {
            url: '/repositories',
            permission: ['view_products'],
            controller: 'DebRepositoriesController',
            templateUrl: 'debs/details/views/deb-repositories.html',
            ncyBreadcrumb: {
                label: "{{ 'Repositories' | translate }}",
                parent: 'deb.info'
            }
        })
        .state('deb.content-views', {
            url: '/content_views',
            collapsed: true,
            permission: ['view_content_views'],
            controller: 'DebContentViewsController',
            templateUrl: 'debs/details/views/deb-content-views.html',
            ncyBreadcrumb: {
                label: "{{ 'Content Views' | translate }}",
                parent: 'deb.info'
            }
        });
    }

    angular
        .module('Bastion.debs')
        .config(DebsConfig);

    DebsConfig.$inject = ['$stateProvider'];
})();
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.debs.controller:DebContentViewsController
     *
     * @description
     *   Provides the functionality for the debs details contentViews page.
     */
    function DebContentViewsController($scope, Nutupane, ContentViewVersion, CurrentOrganization) {
        var contentViewsNutupane,
            params = {
                'deb_id': $scope.$stateParams.debId,
                'organization_id': CurrentOrganization
            };

        contentViewsNutupane = new Nutupane(ContentViewVersion, params);
        $scope.controllerName = 'katello_content_views';
        contentViewsNutupane.masterOnly = true;
        contentViewsNutupane.setSearchKey('contentViewsSearch');

        $scope.table = contentViewsNutupane.table;

        $scope.environmentNames = function (environments) {
            var names = _.map(environments, function (environment) {
                return environment.name;
            });

            return names.join(',');
        };
    }

    angular
        .module('Bastion.debs')
        .controller('DebContentViewsController', DebContentViewsController);

    DebContentViewsController.$inject = [
        '$scope',
        'Nutupane',
        'ContentViewVersion',
        'CurrentOrganization'
    ];

})();
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.debs.controller:DebRepositoriesController
     *
     * @description
     *   Provides the functionality for the debs details repositories page.
     */
    function DebRepositoriesController($scope, Nutupane, Repository, CurrentOrganization) {
        var repositoriesNutupane,
            params = {
                'deb_id': $scope.$stateParams.debId,
                'organization_id': CurrentOrganization
            };

        repositoriesNutupane = new Nutupane(Repository, params);
        $scope.controllerName = 'katello_repositories';
        repositoriesNutupane.masterOnly = true;
        repositoriesNutupane.setSearchKey('repositoriesSearch');

        $scope.table = repositoriesNutupane.table;
    }

    angular
        .module('Bastion.debs')
        .controller('DebRepositoriesController', DebRepositoriesController);

    DebRepositoriesController.$inject = [
        '$scope',
        'Nutupane',
        'Repository',
        'CurrentOrganization'
    ];

})();
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.debs.controller:DebController
     *
     * @description
     *   Provides the functionality for the debs details action pane.
     */
    function DebController($scope, Deb, ApiErrorHandler) {
        $scope.panel = {
            error: false,
            loading: true
        };

        if ($scope.deb) {
            $scope.panel.loading = false;
        }

        $scope.deb = Deb.get({id: $scope.$stateParams.debId}, function () {
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });
    }

    angular
        .module('Bastion.debs')
        .controller('DebController', DebController);

    DebController.$inject = ['$scope', 'Deb', 'ApiErrorHandler'];

})();
/**
 * @ngdoc module
 * @name  Bastion.errata
 *
 * @description
 *   Module for Errata related functionality.
 */

angular.module('Bastion.errata', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.common',
    'Bastion.i18n',
    'Bastion.components.formatters',
    'Bastion.tasks'
]);
/**
 * @ngdoc object
 * @name  Bastion.errata.controller:IncrementalUpdateController
 *
 * @requires $scope
 * @requires translate
 * @requires IncrementalUpdate
 * @requires ContentHostBulkAction
 * @requires ContentViewVersion
 * @requires CurrentOrganization
 * @requires Notification
 *
 * @description
 *   Display confirmation screen and apply Errata.
 */

angular.module('Bastion.errata').controller('ApplyErrataController',
    ['$scope', 'translate', 'IncrementalUpdate', 'HostBulkAction', 'ContentViewVersion', 'CurrentOrganization', 'Notification',
        function ($scope, translate, IncrementalUpdate, HostBulkAction, ContentViewVersion, CurrentOrganization, Notification) {
            var applyErrata, incrementalUpdate;

            function transitionToTask(task) {
                if ($scope.$stateParams.hasOwnProperty('errataId')) {
                    $scope.transitionTo('erratum.task', {errataId: $scope.$stateParams.errataId,
                        taskId: task.id});
                } else {
                    $scope.transitionTo('errata.tasks.task', {taskId: task.id});
                }
                $scope.applyingErrata = false;
            }

            $scope.applyingErrata = false;

            $scope.hasComposites = function (updates) {
                var composite;

                if (updates) {
                    composite = _.find(updates, function (update) {
                        return update.components;
                    });
                    return composite;
                }
                return false;
            };

            $scope.toggleComponents = function (update) {
                update.componentsVisible = !update.componentsVisible;
            };

            incrementalUpdate = function () {
                var error, params = {}, cvIdEnvIds = {};

                $scope.applyingErrata = true;

                params['add_content'] = {
                    'errata_ids': IncrementalUpdate.getErrataIds()
                };

                params['content_view_version_environments'] = [];
                params['resolve_dependencies'] = true;

                //get a list of unique content view version ids with their environments
                angular.forEach($scope.updates, function (update) {
                    var versionId = update['content_view_version'].id;

                    if (update.components) {
                        angular.forEach(_.map(update.components, 'id'), function (componentId) {
                            if (angular.isUndefined(cvIdEnvIds[componentId])) {
                                cvIdEnvIds[componentId] = [];
                            }
                        });
                    }
                    if (angular.isUndefined(cvIdEnvIds[versionId])) {
                        cvIdEnvIds[versionId] = [];
                    }
                    cvIdEnvIds[versionId] = _.uniq(cvIdEnvIds[versionId].concat(_.map(update.environments, 'id')));
                });

                angular.forEach(cvIdEnvIds, function (envIds, cvId) {
                    params['content_view_version_environments'].push({
                        'content_view_version_id': parseInt(cvId, 10),
                        'environment_ids': envIds
                    });
                });

                if ($scope.applyErrata) {
                    params['update_hosts'] = IncrementalUpdate.getBulkContentHosts();
                }

                error = function (response) {
                    angular.forEach(response.data.errors, function (responseError) {
                        Notification.setErrorMessage(responseError);
                    });
                    $scope.applyingErrata = false;
                };

                ContentViewVersion.incrementalUpdate(params, transitionToTask, error);
            };

            applyErrata = function () {
                var params = IncrementalUpdate.getBulkContentHosts(), error;

                $scope.applyingErrata = true;

                params['content_type'] = 'errata';
                params.content = IncrementalUpdate.getErrataIds();
                params['organization_id'] = CurrentOrganization;

                error = function (response) {
                    angular.forEach(response.data.errors, function (responseError) {
                        Notification.setErrorMessage(responseError);
                    });
                    $scope.applyingErrata = false;
                };

                HostBulkAction.installContent(params, transitionToTask, error);
            };

            $scope.selectedContentHosts = IncrementalUpdate.getBulkContentHosts();
            $scope.selectedContentHosts['errata_ids'] = IncrementalUpdate.getErrataIds();
            $scope.selectedContentHosts['organization_id'] = CurrentOrganization;
            HostBulkAction.availableIncrementalUpdates($scope.selectedContentHosts, function (updates) {
                $scope.updates = updates;
            });

            $scope.confirmApply = function() {
                $scope.applyingErrata = true;
                if ($scope.updates.length === 0) {
                    applyErrata();
                } else {
                    incrementalUpdate();
                }
            };

            $scope.incrementalUpdates = IncrementalUpdate.getIncrementalUpdates();
            $scope.selectedContentHosts = IncrementalUpdate.getBulkContentHosts();
            $scope.contentHostIds = IncrementalUpdate.getContentHostIds();
            $scope.errataIds = IncrementalUpdate.getErrataIds();
        }
    ]);
/**
 * @ngdoc object
 * @name  Bastion.errata.controller:ErratumContentHostsController
 *
 * @requires $scope
 * @requires Nutupane
 * @requires Host
 * @requires IncrementalUpdate
 * @requires Environment
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality for the available host collection details action pane.
 */

angular.module('Bastion.errata').controller('ErratumContentHostsController',
    ['$scope', 'Nutupane', 'Host', 'IncrementalUpdate', 'Environment', 'CurrentOrganization',
    function ($scope, Nutupane, Host, IncrementalUpdate, Environment, CurrentOrganization) {
        var nutupane, params, nutupaneParams = {
            'overrideAutoLoad': true
        };

        $scope.restrictInstallable = false;

        params = {
            'erratum_id': $scope.$stateParams.errataId,
            'organization_id': CurrentOrganization
        };

        nutupane = new Nutupane(Host, params, 'postIndex', nutupaneParams);
        $scope.controllerName = 'hosts';
        nutupane.masterOnly = true;
        nutupane.enableSelectAllResults();

        $scope.nutupane = nutupane;
        $scope.table = nutupane.table;
        $scope.nutupane.searchTransform = function(term) {
            var addition, errataSearchStringAddition = $scope.errataSearchString($scope.restrictInstallable);
            if (errataSearchStringAddition) {
                addition = '( ' + errataSearchStringAddition + ' )';
            }
            if (angular.isDefined($scope.environmentFilter)) {
                addition = addition + ' and lifecycle_environment_id = ' + $scope.environmentFilter;
            }

            if (term === "" || angular.isUndefined(term)) {
                return addition;
            }
            return term + ' and ' + addition;
        };

        Environment.queryUnpaged(function (response) {
            $scope.environments = response.results;
        });

        $scope.toggleInstallable = function () {
            nutupane.refresh();
        };

        $scope.errataSearchString = function(installable) {
            var searchTerm = installable ? 'installable_errata' : 'applicable_errata',
                searchStatements, errataIds;

            if ($scope.errata) {
                errataIds = [$scope.errata['errata_id']];
            } else {
                errataIds = IncrementalUpdate.getErrataIds();
            }

            searchStatements = _.map(errataIds, function(errataId) {
                return searchTerm + ' = "' + errataId + '"';
            });
            return searchStatements.join(" or ");
        };
        nutupane.load();

        $scope.selectEnvironment = function (environmentId) {
            $scope.environmentFilter = environmentId;
            nutupane.refresh();
        };

        $scope.goToNextStep = function () {
            IncrementalUpdate.setBulkContentHosts(nutupane.getAllSelectedResults());
            if ($scope.errata) {
                IncrementalUpdate.setErrataIds([$scope.errata['errata_id']]);
            }

            if ($scope.errata && $scope.errata.id) {
                $scope.transitionTo('erratum.apply', {errataId: $scope.errata.id});
            } else {
                $scope.transitionTo('apply-errata.confirm');
            }
        };

        $scope.incrementalUpdates = IncrementalUpdate.getIncrementalUpdates();
    }
]);
/**
 * @ngdoc object
 * @name  Bastion.errata.controller:ErratumRepositoriesController
 *
 * @requires $scope
 * @requires $q
 * @requires Nutupane
 * @requires Repository
 * @requires Environment
 * @requires ContentView
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality for the errata details repositories page.
 */

angular.module('Bastion.errata').controller('ErratumRepositoriesController',
['$scope', '$q', 'Nutupane', 'Repository', 'Environment', 'ContentView', 'CurrentOrganization',
function ($scope, $q, Nutupane, Repository, Environment, ContentView, CurrentOrganization) {
    var repositoriesNutupane, environment, contentView, params = {
        'erratum_id': $scope.$stateParams.errataId,
        'organization_id': CurrentOrganization
    };

    repositoriesNutupane = new Nutupane(Repository, params);
    $scope.controllerName = 'katello_repositories';
    $scope.table = repositoriesNutupane.table;
    $scope.table.initialLoad = false;
    repositoriesNutupane.masterOnly = true;
    repositoriesNutupane.setSearchKey('repositoriesSearch');

    environment = Environment.queryUnpaged(function (response) {
        $scope.environments = response.results;
        $scope.environmentFilter = _.find($scope.environments, {library: true}).id;
    });

    contentView = ContentView.queryUnpaged(function (response) {
        $scope.contentViews = response.results;
        $scope.contentViewFilter = _.find($scope.contentViews, {'default': true});
    });

    $scope.table.working = true;
    $q.all([contentView.$promise, environment.$promise]).then(function () {
        $scope.filterErrata();
        $scope.table.working = false;
    });

    $scope.filterErrata = function () {
        params['environment_id'] = $scope.environmentFilter;
        params['content_view_version_id'] = $scope.contentViewFilter;

        if ($scope.contentViewFilter) {
            params['content_view_version_id'] = _.map($scope.contentViewFilter.versions, 'id');
        }

        repositoriesNutupane.setParams = (params);
        repositoriesNutupane.refresh();
    };
}]);
/**
 * @ngdoc object
 * @name  Bastion.errata.controller:ErratumController
 *
 * @requires $scope
 * @requires Errata
 * @requires ApiErrorHandler
 *
 * @description
 *   Provides the functionality for the errata pages.
 */

angular.module('Bastion.errata').controller('ErratumController', ['$scope', 'Erratum', 'ApiErrorHandler',
    function ($scope, Erratum, ApiErrorHandler) {
        $scope.panel = {
            error: false,
            loading: true
        };

        if ($scope.errata) {
            $scope.panel.loading = false;
        }

        $scope.errata = Erratum.get({id: $scope.$stateParams.errataId}, function () {
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });
    }
]);
/**
 * @ngdoc directive
 * @name bastion.errata:errataCounts
 *
 * @description
 *   Directive for displaying the counts of the various types of errata.
 *
 * @example
 * <div errata-counts="contentViewVersion.errata_counts"></div>
 */

angular.module('Bastion.errata').directive('errataCounts', function () {
    return {
        restrict: 'AE',
        replace: true,
        templateUrl: 'errata/views/errata-counts.html',
        scope: {
            errataCounts: '='
        }
    };
});
/**
 * @ngdoc filter
 * @name  Bastion.errata.filter:errataSeverity
 *
 * @requires translate
 *
 * @description
 *   A filter to turn am errata's severity into an easier to read, translated string.
 *
 * @example
 *   {{ 'moderate' | errataSeverity }} will produce the string "Moderate".
 */

angular.module('Bastion.errata').filter('errataSeverity', ['translate', function (translate) {
    return function (type) {
        var errataSeverity;

        switch (type) {
        case 'Moderate':
            errataSeverity = translate('Moderate');
            break;
        case 'Important':
            errataSeverity = translate('Important');
            break;
        case 'Critical':
            errataSeverity = translate('Critical');
            break;
        case '':
            errataSeverity = translate('N/A');
            break;
        default:
            errataSeverity = type;
        }

        return errataSeverity;
    };

}]);
/**
 * @ngdoc filter
 * @name  Bastion.errata.filter:errataType
 *
 * @requires translate
 *
 * @description
 *   A filter to turn an errata type into an easier to read, translated string.
 *
 * @example
 *   {{ 'bugfix' | errataType }} will produce the translated string "Bug Fix Advisory".
 */

angular.module('Bastion.errata').filter('errataType', ['translate', function (translate) {
    return function (type) {
        var errataType;

        switch (type) {
        case 'bugfix':
            errataType = translate('Bug Fix Advisory');
            break;
        case 'enhancement':
            errataType = translate('Product Enhancement Advisory');
            break;
        case 'security':
            errataType = translate('Security Advisory');
            break;
        default:
            errataType = type;
        }

        return errataType;
    };

}]);
/**
 * @ngdoc object
 * @name  Bastion.errata.controller:ErrataController
 *
 * @requires $scope
 * @requires $state
 * @requires $location
 * @requires translate
 * @requires Nutupane
 * @requires Erratum
 * @requires IncrementalUpdate
 * @requires Repository
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality specific to errata for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.errata').controller('ErrataController',
    ['$scope', '$state', '$location', 'translate', 'Nutupane', 'Erratum', 'IncrementalUpdate', 'Repository', 'CurrentOrganization',
    function ($scope, $state, $location, translate, Nutupane, Erratum, IncrementalUpdate, Repository, CurrentOrganization) {
        var nutupane, params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'sort_by': 'updated',
            'sort_order': 'DESC',
            'paged': true,
            'errata_restrict_applicable': true
        };

        nutupane = $scope.nutupane = new Nutupane(Erratum, params);
        $scope.controllerName = 'katello_errata';
        $scope.table = nutupane.table;
        $scope.removeRow = nutupane.removeRow;

        Erratum.queryPaged({'organization_id': CurrentOrganization}, function (result) {
            $scope.errataCount = result.total;
        });

        $scope.repository = {name: translate('All Repositories'), id: 'all'};

        Repository.queryUnpaged({'organization_id': CurrentOrganization, 'content_type': 'yum'}, function (response) {
            $scope.repositories = [$scope.repository];
            $scope.repositories = $scope.repositories.concat(response.results);

            if ($location.search().repositoryId) {
                $scope.repository = _.find($scope.repositories, function (repository) {
                    return repository.id === parseInt($location.search().repositoryId, 10);
                });
            }
        });

        $scope.showApplicable = true;
        $scope.showInstallable = false;

        $scope.toggleFilters = function () {
            if ($scope.showInstallable === true) {
                $scope.showApplicable = true;
            }

            nutupane.table.params['errata_restrict_applicable'] = $scope.showApplicable;
            nutupane.table.params['errata_restrict_installable'] = $scope.showInstallable;
            nutupane.refresh();
        };

        $scope.$watch('repository', function (repository) {
            var nutupaneParams = nutupane.getParams();

            if (repository.id === 'all') {
                nutupaneParams['repository_id'] = null;
                nutupane.setParams(nutupaneParams);
            } else {
                $location.search('repositoryId', repository.id);
                nutupaneParams['repository_id'] = repository.id;
                nutupane.setParams(nutupaneParams);
            }

            if (!nutupane.table.initialLoad) {
                nutupane.refresh();
            }
        });

        $scope.goToNextStep = function () {
            IncrementalUpdate.setBulkErrata(nutupane.getAllSelectedResults('errata_id'));
            $state.transitionTo('apply-errata.select-content-hosts');
        };

        $scope.incrementalUpdates = IncrementalUpdate.getIncrementalUpdates();
    }]
);
/**
 * @ngdoc object
 * @name Bastion.errata.config
 *
 * @requires $stateProvider
 *
 * @description
 *   State routes defined for the errata module.
 */

angular.module('Bastion.errata').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('errata', {
        url: '/errata',
        permission: ['view_products', 'view_content_views'],
        views: {
          '@': {
              controller: 'ErrataController',
              templateUrl: 'errata/views/errata.html'
          }
        },
        ncyBreadcrumb: {
            label: "{{ 'Errata' | translate }}"
        }
    });

    $stateProvider.state('errata.tasks', {
        url: '/tasks',
        abstract: true,
        permission: ['view_products', 'view_content_views'],
        views: {
            '@': {
                templateUrl: 'errata/views/errata-tasks.html'
            }
        }
    })
    .state('errata.tasks.list', {
        url: '',
        permission: ['view_products', 'view_content_views'],
        templateUrl: 'errata/views/errata-tasks-list.html',
        ncyBreadcrumb: {
            label: "{{ 'Tasks' | translate }}",
            parent: 'errata'

        }
    })
    .state('errata.tasks.task', {
        url: '/:taskId',
        permission: ['view_products', 'view_content_views'],
        controller: 'TaskDetailsController',
        templateUrl: 'errata/views/errata-task-details.html',
        ncyBreadcrumb: {
            label: '{{ task.id }}',
            parent: 'errata.tasks.list'
        }
    });

    $stateProvider.state('apply-errata', {
        url: '/errata/apply',
        abstract: true,
        templateUrl: 'errata/views/apply-errata.html'
    })
    .state('apply-errata.select-content-hosts', {
        url: '/select-content-hosts',
        permission: 'edit_hosts',
        controller: 'ErratumContentHostsController',
        templateUrl: 'errata/views/apply-errata-select-content-hosts.html',
        ncyBreadcrumb: {
            label: '{{ "Select Content Host(s)" | translate }}',
            parent: 'errata'
        }
    })
    .state('apply-errata.confirm', {
        url: '/confirm',
        permission: 'edit_hosts',
        controller: 'ApplyErrataController',
        templateUrl: 'errata/views/apply-errata-confirm.html',
        ncyBreadcrumb: {
            label: '{{ "Confirm" | translate }}',
            parent: 'apply-errata.select-content-hosts'
        }
    });

    $stateProvider.state('erratum', {
        abstract: true,
        url: '/errata/:errataId',
        permission: ['view_products', 'view_content_views'],
        controller: 'ErratumController',
        templateUrl: 'errata/details/views/erratum.html'
    })
    .state('erratum.info', {
        url: '',
        permission: ['view_products', 'view_content_views'],
        templateUrl: 'errata/details/views/erratum-info.html',
        ncyBreadcrumb: {
            label: "{{ errata.title }}",
            parent: 'errata'
        }
    })
    .state('erratum.apply', {
        url: '/apply',
        permission: 'edit_hosts',
        controller: 'ApplyErrataController',
        templateUrl: 'errata/views/apply-errata-confirm.html',
        ncyBreadcrumb: {
            label: "{{ 'Apply' | translate }}",
            parent: 'erratum.info'
        }
    })
    .state('erratum.content-hosts', {
        url: '/content-hosts',
        permission: ['view_products', 'view_content_views'],
        controller: 'ErratumContentHostsController',
        templateUrl: 'errata/details/views/erratum-content-hosts.html',
        ncyBreadcrumb: {
            label: "{{ 'Content Hosts' | translate }}",
            parent: 'erratum.info'
        }
    })
    .state('erratum.repositories', {
        url: '/repositories',
        permission: ['view_products', 'view_content_views'],
        controller: 'ErratumRepositoriesController',
        templateUrl: 'errata/details/views/erratum-repositories.html',
        ncyBreadcrumb: {
            label: "{{ 'Repositories' | translate }}",
            parent: 'erratum.info'
        }
    })
    .state('erratum.task', {
        url: '/tasks/:taskId',
        permission: ['view_products', 'view_content_views'],
        controller: 'TaskDetailsController',
        templateUrl: 'errata/views/errata-task-details.html',
        ncyBreadcrumb: {
            label: "{{ task.id }}",
            parent: 'erratum.info'
        }
    });
}]);
/**
 * @ngdoc service
 * @name  Bastion.errata.factory:Erratum
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for Errata
 */

angular.module('Bastion.errata').factory('Erratum',
    ['BastionResource', function (BastionResource) {

        return BastionResource('katello/api/v2/errata/:id/',
            {id: '@id', 'sort_by': 'issued', 'sort_order': 'DESC'},
            {
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}},
                applicableContentHosts: {method: 'GET', transformResponse: function (data) {
                    var erratum = angular.fromJson(data),
                        systems = erratum['hosts_applicable'];
                    return {results: systems, subtotal: systems.length, total: systems.length};
                }}
            }
        );

    }]
);
(function () {

    /**
     * @ngdoc service
     * @name Bastion.errata.service:IncrementalUpdate
     *
     * @description
     *   Provides a helper service for Incremental Updates.
     *
     */
    function IncrementalUpdate($q, Task, CurrentOrganization) {
        var getIdsFromBulk;

        getIdsFromBulk = function (bulkResource) {
            var ids = [];

            if (bulkResource && bulkResource.included && bulkResource.included.ids) {
                ids = bulkResource.included.ids;
            }

            return ids;
        };

        this.errataIds = [];
        this.contentHostIds = [];
        this.bulkErrata = {};
        this.bulkContentHosts = {};

        /**
         * Allows setting the Errata Ids.
         *
         * @param {Array} errataIds the errata Ids to store for update
         **/
        this.setErrataIds = function (errataIds) {
            this.errataIds = errataIds;
        };

        /**
         * Return the list of Errata Ids.
         *
         * @returns {Array} array of Errata Ids
         */
        this.getErrataIds = function () {
            return this.errataIds;
        };

        /**
         * Allows setting the Content Host Ids.
         *
         * @param {Array} contentHostIds the Content Host Ids to store for update
         **/
        this.setContentHostIds = function (contentHostIds) {
            this.contentHostIds = contentHostIds;
        };

        /**
         * Return the list of Content Host Ids.
         *
         * @returns {Array} array of Content Host Ids
         */
        this.getContentHostIds = function () {
            return this.contentHostIds;
        };

        /**
         * Allows setting the Bulk Content Hosts.
         *
         * @param {Object} bulkContentHosts the bulk selection of the content hosts.
         *
         * @returns {Boolean}
         */
        this.setBulkContentHosts = function (bulkContentHosts) {
            this.bulkContentHosts = bulkContentHosts;
            this.contentHostIds = getIdsFromBulk(bulkContentHosts);
        };

        /**
         * Return the Bulk Content Hosts.
         *
         * @returns {Object} the bulk selection of the content hosts
         */
        this.getBulkContentHosts = function () {
            return this.bulkContentHosts;
        };

        /**
         * Allows setting the Bulk Errata.
         *
         * @param {Object} bulkErratas the bulk selection of the content hosts.
         *
         * @returns {Boolean}
         */
        this.setBulkErrata = function (bulkErratas) {
            this.bulkErrata = bulkErratas;
            this.errataIds = getIdsFromBulk(bulkErratas);
        };

        /**
         * Return the Bulk Errata.
         *
         * @returns {Object} the bulk selection of the content hosts
         */
        this.getBulkErrata = function () {
            return this.bulkErrata;
        };

        /**
         * Return the incremental updates that are currently running.
         *
         * @returns $promise that resolves to a list of incremental updates.
         */
        this.getIncrementalUpdates = function () {
            var searchId, taskSearchParams, taskSearchComplete,
                deferred = $q.defer();

            taskSearchParams = {
                'type': 'all',
                "resource_type": "Organization",
                "resource_id": CurrentOrganization,
                "action_types": "Actions::Katello::ContentView::IncrementalUpdates",
                "active_only": true
            };

            taskSearchComplete = function (results) {
                Task.unregisterSearch(searchId);
                deferred.resolve(results);
            };

            searchId = Task.registerSearch(taskSearchParams, taskSearchComplete);
            return deferred.promise;
        };
    }

    angular.module('Bastion.errata').service('IncrementalUpdate', IncrementalUpdate);

    IncrementalUpdate.$inject = ['$q', 'Task', 'CurrentOrganization'];

})();
/**
 * @ngdoc module
 * @name  Bastion.environments
 *
 * @description
 *   Module for environments related functionality.
 */

angular.module('Bastion.environments', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.utils',
    'Bastion.components',
    'Bastion.errata',
    'Bastion.packages',
    'Bastion.ostree-branches',
    'Bastion.puppet-modules',
    'Bastion.repositories',
    'Bastion.content-views'
]);
(function () {
    'use strict';

    /**
     * @ngdoc service
     * @name  Bastion.environments.service:Content
     *
     * @description
     *   Provides common functions for managing content types and can build a Nutupane
     *   pre-configured for the content type based on params that are passed in and the
     *   current state of the application.
     */
    function ContentService($injector, Nutupane, $state, translate) {
        var contentTypes;

        function currentState() {
            return $state.current.name.split('.').pop();
        }

        function getContentType(state) {
            return _.find(contentTypes, function (type) {
                return state === type.state;
            });
        }

        this.contentTypes = contentTypes = [
            {
                state: 'content-views',
                resource: 'ContentView',
                display: translate('Content Views'),
                params: {nondefault: true}
            }, {
                state: 'repositories',
                resource: 'Repository',
                params: {'content_type': 'yum'},
                display: 'Yum Repositories'
            }, {
                state: 'errata',
                resource: 'Erratum',
                display: translate('Errata'),
                repositoryType: 'yum'
            }, {
                state: 'packages',
                resource: 'Package',
                display: translate('Packages'),
                repositoryType: 'yum'
            }, {
                state: 'puppet-modules',
                resource: 'PuppetModule',
                display: translate('Puppet Modules'),
                repositoryType: 'puppet'
            }, {
                state: 'docker',
                resource: 'DockerTag',
                display: translate('Docker Tags'),
                repositoryType: 'docker'
            }, {
                state: 'ostree',
                resource: 'OstreeBranch',
                display: translate('OSTree Branches'),
                repositoryType: 'ostree'
            }
        ];

        this.getRepositoryType = function () {
            return getContentType(currentState()).repositoryType;
        };

        this.buildNutupane = function (params) {
            var nutupane;

            params = params || {};
            params = angular.extend(params, getContentType(currentState()).params);
            nutupane = new Nutupane($injector.get(getContentType(currentState()).resource), params, 'queryPaged');

            return nutupane;
        };

    }

    angular
        .module('Bastion.environments')
        .service('ContentService', ContentService);

    ContentService.$inject = ['$injector', 'Nutupane', '$state', 'translate'];

})();
(function () {
    'use strict';

    /**
     * @ngdoc controller
     * @name  Bastion.environments.controller:EnvironmentContent
     *
     * @description
     *   Controls displaying content tables for a given environment along with appropriate filtering.
     */
    function EnvironmentContentController($scope, ContentService, ContentView, Repository, translate, $location) {
        var nutupane, allRepositories, nutupaneParams;

        function fetchContentViews(environmentId) {
            ContentView.queryUnpaged({'environment_id': environmentId}, function (data) {
                $scope.contentViews = [$scope.contentView].concat(data.results);
            });
        }

        function fetchRepositories(contentView) {
            var promise, params = {
                'environment_id': $scope.$stateParams.environmentId,
                'content_type': ContentService.getRepositoryType(),
                library: true
            };

            if (contentView && contentView.id !== 'all') {
                params['content_view_id'] = contentView.id;
            }

            promise = Repository.queryUnpaged(params).$promise;

            promise.then(function (data) {
                var repository;

                $scope.repositories = [allRepositories].concat(data.results);

                if ($location.search().repositoryId) {
                    repository = _.find($scope.repositories, function (repo) {
                        return repo.id.toString() === $location.search().repositoryId.toString();
                    });
                }

                if (repository === null) {
                    $location.search('repositoryId', null);
                }

                $scope.repository = repository || allRepositories;
            });

            return promise;
        }

        function getVersionId(contentView, environmentId) {
            var versionId, version;

            if (contentView.id !== 'all') {
                version = _.find(contentView.versions, function (vers) {
                    return vers['environment_ids'].indexOf(parseInt(environmentId, 10)) > -1;
                });
                versionId = version.id;
            }

            return versionId;
        }

        nutupaneParams = {'environment_id': $scope.$stateParams.environmentId, library: true};
        if ($location.search().repositoryId) {
            nutupaneParams['repository_id'] = $location.search().repositoryId;
        }

        nutupane = ContentService.buildNutupane(nutupaneParams);
        nutupane.masterOnly = true;

        $scope.nutupane = nutupane;
        $scope.table = nutupane.table;

        $scope.contentView = {id: 'all', name: translate('All Content Views')};

        allRepositories = {id: 'all', name: translate('All Repositories')};
        $scope.repository = allRepositories;

        fetchContentViews($scope.$stateParams.environmentId);

        if (ContentService.getRepositoryType()) {
            fetchRepositories();
        }

        $scope.contentViewSelected = function (contentView) {
            var params = nutupane.getParams();

            fetchRepositories(contentView).then(function (response) {
                var repo;

                params['content_view_version_id'] = getVersionId(contentView, $scope.$stateParams.environmentId);

                if (params['repository_id']) {
                    repo = _.find(response.results, function (repository) {
                        return repository.id.toString() === params['repository_id'].toString();
                    });
                }

                if (!repo) {
                    params['repository_id'] = null;
                }

                nutupane.setParams(params);
                nutupane.refresh();
            });
        };

        $scope.repositorySelected = function (repository) {
            var params = nutupane.getParams();

            if (repository.id === 'all') {
                $location.search('repositoryId', null);
            } else {
                $location.search('repositoryId', repository.id);
            }

            params['repository_id'] = $location.search().repositoryId;
            nutupane.setParams(params);
            nutupane.refresh();
        };
    }

    angular
        .module('Bastion.environments')
        .controller('EnvironmentContentController', EnvironmentContentController);

    EnvironmentContentController.$inject = ['$scope', 'ContentService', 'ContentView', 'Repository', 'translate', '$location'];

})();
(function () {
    'use strict';

    /**
     * @ngdoc controller
     * @name  Bastion.environments.controller:EnvironmentController
     *
     * @description
     *   Enter a description!
     */
    function EnvironmentController($scope, Environment, translate, ContentService, ApiErrorHandler, Notification) {

        $scope.contentTypes = ContentService.contentTypes;
        $scope.panel = {
            error: false,
            loading: true
        };

        $scope.environment = Environment.get({id: $scope.$stateParams.environmentId}, function () {
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        $scope.save = function (environment) {
            var promise;

            function success() {
                Notification.setSuccessMessage(translate('Environment saved'));
            }

            function error(response) {
                angular.forEach(response.data.errors, function (errorMessage) {
                    Notification.setErrorMessage((translate("An error occurred saving the Environment: ") + errorMessage));
                });
            }

            promise = environment.$update();
            promise.then(success, error);

            return promise;
        };

        $scope.remove = function (environment) {
            var promise;

            function success() {
                Notification.setSuccessMessage(translate('Remove Successful.'));
                $scope.transitionTo('environments');
            }

            function error(response) {
                Notification.setErrorMessage(translate("An error occurred removing the environment: ") +
                    response.data.displayMessage);
            }

            promise = environment.$delete();
            promise.then(success, error);

            return promise;
        };

    }

    angular
        .module('Bastion.environments')
        .controller('EnvironmentController', EnvironmentController);

    EnvironmentController.$inject = ['$scope', 'Environment', 'translate', 'ContentService', 'ApiErrorHandler'];

})();
/**
 * @ngdoc service
 * @name  Bastion.environments.factory:Environment
 *
 * @requires BastionResource
 * @requires CurrentOrganization
 *
 * @description
 *   Provides a BastionResource for interacting with environments.
 */

angular.module('Bastion.environments').factory('Environment',
    ['BastionResource', 'CurrentOrganization',
    function (BastionResource, CurrentOrganization) {

        return BastionResource('katello/api/v2/environments/:id/:action',
            {id: '@id', 'organization_id': CurrentOrganization},
            {
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}},
                update: {method: 'PUT'}
            }
        );

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.environments.controller:EnvironmentsController
 *
 * @requires $scope
 * @requires Organization
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality for the environments path page.
 */

angular.module('Bastion.environments').controller('EnvironmentsController',
    ['$scope', 'Organization', 'CurrentOrganization',
        function ($scope, Organization, CurrentOrganization) {

            Organization.paths({id: CurrentOrganization}, function (paths) {
                var actualPaths = [];

                $scope.library = paths[0].environments[0];

                angular.forEach(paths, function (path, index) {
                    paths[index].environments.splice(0, 1);

                    if (paths[index].environments.length !== 0) {
                        actualPaths.push(path);
                    }
                });

                $scope.paths = actualPaths;
            });

            $scope.lastEnvironment = function (path) {
                return path.environments[path.environments.length - 1];
            };

        }]
);
/**
 * @ngdoc object
 * @name Bastion.environments.config
 *
 * @requires $stateProvider
 *
 * @description
 *   Used for systems level configuration such as setting up the ui state machine.
 */

angular.module('Bastion.environments').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('environments', {
        url: '/lifecycle_environments',
        permission: 'view_lifecycle_environments',
        template: '<div ui-view></div>',
        views: {
            '@': {
                controller: 'EnvironmentsController',
                templateUrl: 'environments/views/environments.html'
            }
        },
        ncyBreadcrumb: {
            label: '{{ "Environments" | translate }}'
        }
    });

    $stateProvider.state('environments.new', {
        url: '/lifecycle_environments/:priorId/new',
        permission: 'create_lifecycle_environments',
        views: {
            '@': {
                controller: 'NewEnvironmentController',
                templateUrl: 'environments/views/new-environment.html'
            }
        },
        ncyBreadcrumb: {
            label: '{{ "New Environment" | translate }}',
            parent: 'environments'
        }
    });

    $stateProvider.state('environment', {
        abstract: true,
        url: '/lifecycle_environments/:environmentId',
        permission: 'view_lifecycle_environments',
        controller: 'EnvironmentController',
        templateUrl: 'environments/details/views/environment.html'
    })
    .state('environment.details', {
        url: '',
        permission: 'view_lifecycle_environments',
        templateUrl: 'environments/details/views/environment-details.html',
        ncyBreadcrumb: {
            label: '{{ environment.name }}',
            parent: 'environments'
        }
    })
    .state('environment.errata', {
        url: '/errata?repositoryId&contentViewId',
        reloadOnSearch: false,
        permission: 'view_lifecycle_environments',
        controller: 'EnvironmentContentController',
        templateUrl: 'environments/details/views/environment-errata.html',
        ncyBreadcrumb: {
            label: '{{ "Errata" | translate }}',
            parent: 'environment.details'
        }
    })
    .state('environment.repositories', {
        url: '/repositories?contentViewId',
        reloadOnSearch: false,
        permission: 'view_lifecycle_environments',
        controller: 'EnvironmentContentController',
        templateUrl: 'environments/details/views/environment-repositories.html',
        ncyBreadcrumb: {
            label: '{{ "Repositories" | translate }}',
            parent: 'environment.details'
        }
    })
    .state('environment.packages', {
        url: '/packages?repositoryId&contentViewId',
        reloadOnSearch: false,
        permission: 'view_lifecycle_environments',
        controller: 'EnvironmentContentController',
        templateUrl: 'environments/details/views/environment-packages.html',
        ncyBreadcrumb: {
            label: '{{ "Packages" | translate }}',
            parent: 'environment.details'
        }
    })
    .state('environment.puppet-modules', {
        url: '/puppet-modules?contentViewId',
        reloadOnSearch: false,
        permission: 'view_lifecycle_environments',
        controller: 'EnvironmentContentController',
        templateUrl: 'environments/details/views/environment-puppet-modules.html',
        ncyBreadcrumb: {
            label: '{{ "Puppet Modules" | translate }}',
            parent: 'environment.details'
        }
    })
    .state('environment.docker', {
        url: '/docker?repositoryId&contentViewId',
        reloadOnSearch: false,
        permission: 'view_lifecycle_environments',
        controller: 'EnvironmentContentController',
        templateUrl: 'environments/details/views/environment-docker.html',
        ncyBreadcrumb: {
            label: '{{ "Docker" | translate }}',
            parent: 'environment.details'
        }
    })
    .state('environment.ostree', {
        url: '/ostree?repositoryId&contentViewId',
        reloadOnSearch: false,
        permission: 'view_lifecycle_environments',
        controller: 'EnvironmentContentController',
        templateUrl: 'environments/details/views/environment-ostree.html',
        ncyBreadcrumb: {
            label: '{{ "OSTree Branches" | translate }}',
            parent: 'environment.details'
        }
    })
    .state('environment.content-views', {
        url: '/content-views',
        reloadOnSearch: false,
        permission: 'view_lifecycle_environments',
        controller: 'EnvironmentContentController',
        templateUrl: 'environments/details/views/environment-content-views.html',
        ncyBreadcrumb: {
            label: '{{ "Content Views" | translate }}',
            parent: 'environment.details'
        }
    });

}]);
(function () {
    'use strict';

    /**
     * @ngdoc controller
     * @name  Bastion.environments.controller:NewEnvironmentController
     *
     * @description
     *   Handles creating a new environment.
     */
    function NewEnvironmentController($scope, Environment, FormUtils, Notification) {

        function success() {
            $scope.transitionTo('environments');
        }

        function error(response) {
            $scope.working = false;
            angular.forEach(response.data.errors, function (errors, field) {
                if ($scope.environmentForm.hasOwnProperty(field)) {
                    $scope.environmentForm[field].$setValidity('server', false);
                    $scope.environmentForm[field].$error.messages = errors;
                } else {
                    Notification.setErrorMessage(errors);
                }
            });
        }

        $scope.loading = true;
        $scope.environment = new Environment();
        $scope.priorEnvironment = Environment.get({id: $scope.$stateParams.priorId});

        $scope.priorEnvironment.$promise.then(function () {
            $scope.loading = false;
        });

        $scope.save = function (environment) {
            environment['prior_id'] = $scope.$stateParams.priorId;
            environment.$save(success, error);
        };

        $scope.$watch('environment.name', function () {
            if ($scope.environmentForm.name) {
                $scope.environmentForm.name.$setValidity('server', true);
                FormUtils.labelize($scope.environment);
            }
        });

    }

    angular
        .module('Bastion.environments')
        .controller('NewEnvironmentController', NewEnvironmentController);

    NewEnvironmentController.$inject = ['$scope', 'Environment', 'FormUtils', 'Notification'];

})();
/**
 * @ngdoc module
 * @name  Bastion.capsules
 *
 * @description
 *   Module for Capsule related functionality.
 */

angular.module('Bastion.capsules', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.components'
]);
/**
 * @ngdoc service
 * @name  Bastion.capsules.factory:Capsule
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for capsules or list of capsules.
 */

angular.module('Bastion.capsules').factory('Capsule',
    ['BastionResource', function (BastionResource) {

        return BastionResource('katello/api/capsules/:id/:action', {id: '@id'}, {
        });

    }]
);
/**
 * @ngdoc module
 * @name  Bastion.capsule-content
 *
 * @description
 *   Module for Capsule content related functionality.
 */

angular.module('Bastion.capsule-content', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.components'
]);
/**
 * @ngdoc object
 * @name  Bastion.capsule-content.controller:CapsuleContentController
 *
 * @requires $scope
 * @requires $urlMatcherFactory
 * @requires $location
 * @requires translate
 * @requires CapsuleContent
 * @requires AggregateTask
 * @requires CurrentOrganization
 * @requires syncState
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the capsule-content page.
 */

angular.module('Bastion.capsule-content').controller('CapsuleContentController',
    ['$scope', '$urlMatcherFactory', '$location', 'translate', 'CapsuleContent', 'AggregateTask', 'CurrentOrganization', 'syncState', 'Notification',
    function ($scope, $urlMatcherFactory, $location, translate, CapsuleContent, AggregateTask, CurrentOrganization, syncState, Notification) {

        var refreshSyncStatus;
        var urlMatcher = $urlMatcherFactory.compile("/smart_proxies/:capsuleId");
        var capsuleId = urlMatcher.exec($location.path()).capsuleId;

        function processError(response) {
            if (response.data && response.data.displayMessage) {
                Notification.setErrorMessage(response.data.displayMessage);
            }
        }

        function pickLastTask(tasks) {
            var task = tasks.slice(-1)[0];
            task.progressbar = {
                value: task.progress * 100,
                type: $scope.progressbarType(task)
            };
            return task;
        }

        function isTaskInProgress(task) {
            return (task && (task.state === 'pending' || task.state === 'running'));
        }

        function stateFromTask(syncTask) {
            var state;

            if (isTaskInProgress(syncTask)) {
                state = syncState.SYNCING;
            }
            if (syncTask && syncTask.result !== 'success') {
                state = syncState.FAILURE;
            } else {
                state = syncState.DEFAULT;
            }
            return state;
        }

        function taskUpdated() {
            if (!angular.isUndefined($scope.syncTask.result) && $scope.syncTask.result !== 'pending') {
                $scope.syncTask.unregisterAll();
                refreshSyncStatus();
            }
        }

        function aggregateTasks(tasks) {
            var taskIds = _.map(tasks, function (task) {
                                return task.id;
                            });
            return AggregateTask.new(taskIds, taskUpdated);
        }

        refreshSyncStatus = function () {
            var params = {
                id: capsuleId
            };
            if ( CurrentOrganization !== '' ) {
                params['organization_id'] = CurrentOrganization;
            }

            CapsuleContent.syncStatus(params).$promise.then(function (syncStatus) {
                var errorCount, errorMessage;
                $scope.syncStatus = syncStatus;
                if (syncStatus['last_sync_time'] === null) {
                    $scope.syncStatus['last_sync_time'] = translate('Never');
                }

                if (syncStatus['active_sync_tasks'].length > 0) {
                    $scope.syncTask = aggregateTasks(syncStatus['active_sync_tasks']);

                } else if (syncStatus['last_failed_sync_tasks'].length > 0) {
                    errorCount = $scope.syncTask.humanized.errors.length;
                    $scope.syncTask = pickLastTask(syncStatus['last_failed_sync_tasks']);
                    if (errorCount > 0) {
                        errorMessage = $scope.syncTask.humanized.errors[0];
                        if (errorCount > 2) {
                            errorMessage += " " + translate("Plus %y more errors").replace("%y", errorCount - 1);
                        } else if (errorCount > 1) {
                            errorMessage += " " + translate("Plus 1 more error");
                        }
                        Notification.setErrorMessage(errorMessage);
                    }
                }
                $scope.syncState.set(stateFromTask($scope.syncTask));
            }, function (response) {
                $scope.syncStatus = {
                    'active_sync_tasks': [],
                    'last_failed_sync_tasks': []
                };
                processError(response);
            });
        };

        $scope.syncState = syncState;

        $scope.expandEnvironments = {};

        refreshSyncStatus();

        $scope.$on('$destroy', function () {
            if ($scope.syncTask) {
                $scope.syncTask.unregisterAll();
            }
        });

        $scope.isTaskInProgress = isTaskInProgress;

        $scope.syncCapsule = function (skipMetadataCheck) {
            if (!$scope.syncState.is(syncState.SYNCING)) {

                $scope.syncState.set(syncState.SYNC_TRIGGERED);

                CapsuleContent.sync({id: capsuleId, 'skip_metadata_check': skipMetadataCheck}).$promise.then(function (task) {
                    $scope.syncStatus['active_sync_tasks'].push(task);
                    $scope.syncTask = aggregateTasks($scope.syncStatus['active_sync_tasks']);
                    $scope.syncState.set(syncState.SYNCING);
                }, function (response) {
                    processError(response);
                    $scope.syncState.set(syncState.DEFAULT);
                });
            }
        };

        $scope.cancelSync = function () {
            if ($scope.syncState.is(syncState.SYNCING)) {

                $scope.syncState.set(syncState.CANCEL_TRIGGERED);
                CapsuleContent.cancelSync({id: capsuleId}).$promise.catch(processError);
            }
        };

        $scope.syncStatusText = function (currentSyncState, syncStatus) {
            var message, syncableEnvs, envNames;

            if (angular.isUndefined(syncStatus)) {
                return "";
            }

            if (currentSyncState.is(currentSyncState.SYNCING)) {
                message = translate("Smart proxy currently syncing to your locations...");
            } else if (currentSyncState.is(currentSyncState.SYNC_TRIGGERED)) {
                message = translate("Synchronization is about to start...");
            } else if (currentSyncState.is(currentSyncState.CANCEL_TRIGGERED)) {
                message = translate("Synchronization is being cancelled...");
            } else {
                syncableEnvs = _.filter(syncStatus['lifecycle_environments'], {syncable: true});

                if (syncableEnvs.length > 0) {
                    envNames = _.map(syncableEnvs, 'name').join(', ');
                    message = translate("%count environment(s) can be synchronized: %envs")
                                .replace('%count', syncableEnvs.length)
                                .replace('%envs', envNames);
                } else {
                    message = translate("Smart proxy is synchronized");
                }
            }
            return message;
        };

        $scope.toggleExpandEnvironment = function (environment) {
            $scope.expandEnvironments[environment.id] = !$scope.expandEnvironments[environment.id];
            return $scope.expandEnvironments[environment.id];
        };

        $scope.isEnvronmentExpanded = function (environment) {
            return $scope.expandEnvironments[environment.id];
        };

        $scope.progressbarType = function (syncTask) {
            var type;

            if (angular.isUndefined(syncTask) || syncTask.result === 'pending' || syncTask.result === 'success') {
                type = 'success';
            } else {
                type = 'danger';
            }
            return type;
        };

    }]
);
/**
 * @ngdoc service
 * @name  Bastion.capsule-content.factory:CapsuleContent
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for capsule content.
 */

angular.module('Bastion.capsule-content').factory('CapsuleContent',
    ['BastionResource', function (BastionResource) {

        return BastionResource('katello/api/capsules/:id/content/:action', {id: '@id'}, {
          syncStatus: {method: 'GET', isArray: false, params: {action: 'sync'}},
          sync: {method: 'post', isArray: false, params: {action: 'sync'}},
          cancelSync: {method: 'delete', isArray: false, params: {action: 'sync'}}
        });

    }]
);
/**
* @ngdoc object
* @name Bastion.capsule-content.config
*
* @requires $stateProvider
* @requires $urlRouterProvider
*
* @description
*   Used for systems level configuration such as setting up the ui state machine.
*/

angular.module('Bastion.capsule-content').config(['$stateProvider', '$urlRouterProvider', function ($stateProvider, $urlRouterProvider) {
    //Catch the url to prevent the router to perform redirect.
    $urlRouterProvider.when('/smart_proxies/:proxyId', ['$match', '$stateParams', function ($match, $stateParams) {
        $stateParams.pageName = 'smart_proxies/detail';
        return true;
    }]);

    // Add rule to redirect links on the smart proxy detail page.
    // Changing state doesn't work there since there's no <ui-view> element there
    $urlRouterProvider.rule(function ($injector, $location) {
        var $stateParams = $injector.get('$stateParams'),
            $window = $injector.get('$window');

        if ($stateParams.pageName === 'smart_proxies/detail') {
            $window.location.href = $location.path();
        }
    });
}]);

/**
 * @ngdoc run
 * @name Bastion.capsule-content.run
 *
 * @requires $rootScope
 * @requires $window
 * @requires $timeout
 *
 * @description
 *   Ensure foreman's setTab() function is called on capsule content pages.
 */
angular.module('Bastion.capsule-content').run(['$rootScope', '$window', '$timeout', function ($rootScope, $window, $timeout) {
    var smartProxiesRegex = new RegExp("/smart_proxies/.+#.+");
    $rootScope.$on('$locationChangeStart', function (event, newUrl) {
        if (newUrl.match(smartProxiesRegex)) {
            $timeout(function () {
                $window.setTab();
            });
        }
    });
}]);
/**
 * @ngdoc service
 * @name  Bastion.capsule-content.service:syncState
 *
 * @description
 *   Provides a syncState that keeps capsule sync UI state.
 */

angular.module('Bastion.capsule-content').service('syncState', function () {

    this.DEFAULT = 'DEFAULT';
    this.SYNCING = 'SYNCING';
    this.SYNC_TRIGGERED = 'SYNC_TRIGGERED';
    this.CANCEL_TRIGGERED = 'CANCEL_TRIGGERED';
    this.FAILURE = 'FAILURE';

    this.set = function (state) {
        this.state = state;
        return this.state;
    };

    this.is = function () {
        return _.includes(arguments, this.state || this.DEFAULT);
    };

});
/**
 * @ngdoc module
 * @name  Bastion.organizations
 *
 * @description
 *   Module for organization related functionality.
 */

angular.module('Bastion.organizations', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.components'
]);
(function () {

    /**
     * @ngdoc run
     * @name Bastion.run:CheckCurrentOrganization
     *
     * @description
     *   Checks whether a page requires a current organization to be set and if it does
     *   redirects the user to the Katello 403 page to instruct them to select an organization to proceed.
     */
    function CheckCurrentOrganization($rootScope, $window, CurrentOrganization, FencedPages) {

        $rootScope.$on('$stateChangeStart', function (event, toState, toParams) {
            if (CurrentOrganization === "" && FencedPages.isFenced(toState)) {
                event.preventDefault();
                $rootScope.transitionTo('organizations.select', {toState: $rootScope.$state.href(toState.name, toParams, {absolute: 'true'})});
            }
        });

    }

    angular
        .module('Bastion.organizations')
        .run(CheckCurrentOrganization);

    CheckCurrentOrganization.$inject = ['$rootScope', '$window', 'CurrentOrganization', 'FencedPages'];

})();
/**
 *@ngdoc service
 *@name Bastion.organizations.service:FencedPages
 *
 *@descriptions
 * Service that keeps track of pages that require an organization to be selected
 */


angular.module('Bastion.organizations').service('FencedPages', ['$state',
    function ($state) {
        var fencedPages = [
            'activation-keys',
            'content-hosts',
            'content-views',
            'docker-tags',
            'errata',
            'gpg-keys',
            'host-collections',
            'lifecycle-environments',
            'packages',
            'products',
            'puppet-modules',
            'subscriptions',
            'sync-plans'
        ];

        function getRootPath(path) {
            var rootPath = null;

            if (path && angular.isString(path)) {
                rootPath = path.replace('_', '-').split('/')[1];
            }
            return rootPath;
        }

        this.addPages = function (pages) {
            fencedPages = _.uniq(fencedPages.concat(pages));
        };

        this.list = function () {
            return fencedPages;
        };

        this.isFenced = function (toState) {
            var stateUrl = $state.href(toState);
            return fencedPages.indexOf(getRootPath(stateUrl)) !== -1;
        };
    }
]);
(function () {
    'use strict';

    /**
    * @ngdoc controller
    * @name Bastion.organizations.controller:OrganizationSelectorController
    *
    * @description
    *     Selecting an organization
    */
    function OrganizationSelectorController($scope, Organization, CurrentOrganization, $window) {
        var transitionState;

        $scope.selectedOrganization = {};

        // TODO: per_page hack necessary because of http://projects.theforeman.org/issues/21800
        Organization.queryUnpaged({'per_page': 99999999999}, function (response) {
            $scope.organizations = response.results;
        });

        $scope.selectOrganization = function (organization) {
            var label = organization.id + '-' + organization.name.replace("'", '').replace(".", '');

            Organization.select({label: label}).$promise.catch(function () {
                $window.location.href = transitionState;
            });
        };

        $scope.$on('$stateChangeSuccess', function (event, toState, toParams) {
            transitionState = toParams.toState;

            if (CurrentOrganization) {
                $window.location.href = transitionState;
            }
        });
    }

    angular
        .module('Bastion.organizations')
        .controller('OrganizationSelectorController', OrganizationSelectorController);

    OrganizationSelectorController.$inject = ['$scope', 'Organization', 'CurrentOrganization', '$window'];
})();
/**
 * @ngdoc service
 * @name  Bastion.organizations.factory:Organization
 *
 * @requires BastionResource
 * @requires CurrentOrganization
 *
 * @description
 *   Provides a BastionResource for organization(s).
 */

angular.module('Bastion.organizations').factory('Organization',
    ['BastionResource', 'CurrentOrganization', function (BastionResource, CurrentOrganization) {

        return BastionResource('katello/api/v2/organizations/:id/:action',
            {id: '@id'},
            {
                update: { method: 'PUT' },
                releaseVersions: {method: 'GET', params: {action: 'releases'}},

                select: {
                    method: 'GET',
                    url: 'organizations/:label/select'
                },
                repoDiscover: { method: 'POST', params: {action: 'repo_discover'}},
                cancelRepoDiscover: {method: 'POST', params: {action: 'cancel_repo_discover'}},
                paths: {
                    method: 'GET',
                    url: 'katello/api/v2/organizations/:id/environments/paths',
                    isArray: true,
                    transformResponse: function (data) {
                        return angular.fromJson(data).results;
                    }
                },
                readableEnvironments: {
                    method: 'GET',
                    url: 'katello/api/v2/organizations/:id/environments/paths',
                    isArray: true,
                    transformResponse: function (data) {
                        // transform [{environments : [{id, name, permissions: {readable : true}}]}]
                        // to [[{id, name, select: true}]]
                        return _.map(angular.fromJson(data).results, function (path) {
                            return _.map(path.environments, function (env) {
                                env.select = env.permissions.readable;
                                return env;
                            });
                        });
                    }
                },
                redhatProvider: {
                    method: 'GET',
                    url: 'katello/api/v2/organizations/:organization_id/redhat_provider',
                    params: {'organization_id': CurrentOrganization}
                }
            }
        );

    }]
);
(function () {

    /**
     * @ngdoc object
     * @name Bastion.organizations.config
     *
     * @requires $stateProvider
     *
     * @description
     *   State routes defined for the organizations module.
     */
    function OrganizationRoutes($stateProvider) {
        $stateProvider.state('organizations', {
            abstract: true,
            template: '<div ui-view></div>'
        })
        .state('organizations.select', {
            url: '/select_organization?toState',
            permission: 'view_organizations',
            controller: 'OrganizationSelectorController',
            templateUrl: 'organizations/views/organization-selector.html'
        });
    }

    angular.module('Bastion.organizations').config(OrganizationRoutes);

    OrganizationRoutes.$inject = ['$stateProvider'];

})();
/**
 * @ngdoc module
 * @name  Bastion.packages
 *
 * @description
 *   Module for package related functionality.
 */

angular.module('Bastion.packages', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.common'
]);
/**
 * @ngdoc object
 * @name  Bastion.packages.controller:PackageRepositoriesController
 *
 * @requires $scope
 * @requires $q
 * @requires Nutupane
 * @requires Repository
 * @requires Environment
 * @requires ContentView
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality for the package repositories page.
 */

angular.module('Bastion.packages').controller('PackageRepositoriesController',
    ['$scope', '$q', 'Nutupane', 'Repository', 'Environment', 'ContentView', 'CurrentOrganization',
    function ($scope, $q, Nutupane, Repository, Environment, ContentView, CurrentOrganization) {
        var repositoriesNutupane, environment, contentView;
        var params = {
            'rpm_id': $scope.$stateParams.packageId,
            'organization_id': CurrentOrganization
        };

        repositoriesNutupane = new Nutupane(Repository, params);
        $scope.controllerName = 'katello_repositories';
        $scope.table = repositoriesNutupane.table;
        $scope.table.initialLoad = false;
        repositoriesNutupane.masterOnly = true;
        repositoriesNutupane.setSearchKey('repositoriesSearch');

        environment = Environment.queryUnpaged(function (response) {
            $scope.environments = response.results;
            $scope.environmentFilter = _.find($scope.environments, {library: true}).id;
        });

        contentView = ContentView.queryUnpaged(function (response) {
            $scope.contentViews = response.results;
            $scope.contentViewFilter = _.find($scope.contentViews, {'default': true});
        });

        $scope.table.working = true;
        $q.all([contentView.$promise, environment.$promise]).then(function () {
            $scope.filterPackages();
            $scope.table.working = false;
        });

        $scope.filterPackages = function () {
            params['environment_id'] = $scope.environmentFilter;
            params['content_view_version_id'] = $scope.contentViewFilter;

            if ($scope.contentViewFilter) {
                params['content_view_version_id'] = _.map($scope.contentViewFilter.versions, 'id');
            }

            repositoriesNutupane.setParams = (params);
            repositoriesNutupane.refresh();
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.packages.controller:PackageController
 *
 * @requires $scope
 * @requires Package
 * @requires Host
 * @requires CurrentOrganization
 * @requires ApiErrorHandler
 *
 * @description
 *   Provides the functionality for the package page.
 */

angular.module('Bastion.packages').controller('PackageController',
    ['$scope', 'Package', 'Host', 'CurrentOrganization', 'ApiErrorHandler',
    function ($scope, Package, Host, CurrentOrganization, ApiErrorHandler) {
        $scope.panel = {
            error: false,
            loading: true
        };

        if ($scope.package) {
            $scope.panel.loading = false;
        }

        $scope.installedPackageCount = undefined;

        $scope.fetchHostCount = function() {
            Host.get({'per_page': 0, 'search': $scope.createSearchString('installed_package'), 'organization_id': CurrentOrganization}, function (data) {
                $scope.installedPackageCount = data.subtotal;
            });
        };

        $scope.createSearchString = function(field) {
            return field + '=' + $scope.package.name + '-' + $scope.package.version + '-' + $scope.package.release + '.' +
                            $scope.package.arch;
        };

        $scope.package = Package.get({id: $scope.$stateParams.packageId}, function () {
            $scope.panel.loading = false;
            $scope.fetchHostCount();
        }, function(response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });
    }
]);
/**
 * @ngdoc service
 * @name  Bastion.providers.factory:Package
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for product or list of providers.
 */

angular.module('Bastion.packages').factory('Package',
    ['BastionResource', 'CurrentOrganization', function (BastionResource) {

        return BastionResource('katello/api/v2/packages/:id',
            {'id': '@id'},
            {
                'autocomplete': {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}},
                'autocompleteName': {method: 'GET', isArray: false, params: {id: 'auto_complete_name'},
                    transformResponse: function (data) {
                        data = angular.fromJson(data);
                        return {results: data};
                    }
                },
                'autocompleteArch': {method: 'GET', isArray: false, params: {id: 'auto_complete_arch'},
                    transformResponse: function (data) {
                        data = angular.fromJson(data);
                        return {results: data};
                    }
                }
            });
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.errata.controller:ErrataController
 *
 * @requires $scope
 * @requires $location
 * @requires translate
 * @requires Nutupane
 * @requires Erratum
 * @requires Repository
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality specific to errata for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.packages').controller('PackagesController',
    ['$scope', '$location', 'translate', 'Nutupane', 'Package', 'Task', 'Repository', 'CurrentOrganization',
    function ($scope, $location, translate, Nutupane, Package, Task, Repository, CurrentOrganization) {
        var nutupane, params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'paged': true
        };

        nutupane = $scope.nutupane = new Nutupane(Package, params);
        $scope.controllerName = 'katello_erratum_packages';
        nutupane.masterOnly = true;
        $scope.table = nutupane.table;
        $scope.removeRow = nutupane.removeRow;

        $scope.repository = {name: translate('All Repositories'), id: 'all'};

        Repository.queryUnpaged({'organization_id': CurrentOrganization, 'content_type': 'yum'}, function (response) {
            $scope.repositories = [$scope.repository];
            $scope.repositories = $scope.repositories.concat(response.results);

            if ($location.search().repositoryId) {
                $scope.repository = _.find($scope.repositories, function (repository) {
                    return repository.id === parseInt($location.search().repositoryId, 10);
                });
            }
        });

        Package.queryPaged({'organization_id': CurrentOrganization}, function (result) {
            $scope.packageCount = result.total;
        });

        $scope.showApplicable = false;
        $scope.showUpgradable = false;

        $scope.toggleFilters = function () {
            if ($scope.showUpgradable === true) {
                $scope.showApplicable = true;
            }

            nutupane.table.params['packages_restrict_applicable'] = $scope.showApplicable;
            nutupane.table.params['packages_restrict_upgradable'] = $scope.showUpgradable;
            nutupane.refresh();
        };

        $scope.$watch('repository', function (repository) {
            var nutupaneParams = nutupane.getParams();

            if (repository.id === 'all') {
                nutupaneParams['repository_id'] = null;
                nutupane.setParams(nutupaneParams);
            } else {
                $location.search('repositoryId', repository.id);
                nutupaneParams['repository_id'] = repository.id;
                nutupane.setParams(nutupaneParams);
            }

            if (!nutupane.table.initialLoad) {
                nutupane.refresh();
            }
        });
    }]
);
/**
 * @ngdoc object
 * @name Bastion.packages.config
 *
 * @requires $stateProvider
 *
 * @description
 *   State routes defined for the packages module.
 */

angular.module('Bastion.packages').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('packages', {
        url: '/packages',
        permission: ['view_products', 'view_content_views'],
        views: {
            '@': {
                controller: 'PackagesController',
                templateUrl: 'packages/views/packages.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{ 'Packages' | translate }}"
        }
    })
    .state('package', {
        abstract: true,
        url: '/packages/:packageId',
        permission: ['view_products', 'view_content_views'],
        controller: 'PackageController',
        templateUrl: 'packages/details/views/package.html'
    })
    .state('package.info', {
        url: '',
        permission: ['view_products', 'view_content_views'],
        templateUrl: 'packages/details/views/package-info.html',
        ncyBreadcrumb: {
            label: "{{ package.nvra }}",
            parent: 'packages'
        }
    })
    .state('package.dependencies', {
        url: '/dependencies',
        permission: ['view_products', 'view_content_views'],
        templateUrl: 'packages/details/views/package-dependencies.html',
        ncyBreadcrumb: {
            label: "{{ 'Dependencies' | translate }}",
            parent: 'package.info'
        }
    })
    .state('package.files', {
        url: '/files',
        permission: ['view_products', 'view_content_views'],
        templateUrl: 'packages/details/views/package-files.html',
        ncyBreadcrumb: {
            label: "{{ 'Files' | translate }}",
            parent: 'package.info'
        }
    })
    .state('package.repositories', {
        url: '/repositories',
        permission: ['view_products', 'view_content_views'],
        controller: 'PackageRepositoriesController',
        templateUrl: 'packages/details/views/package-repositories.html',
        ncyBreadcrumb: {
            label: "{{ 'Repositories' | translate }}",
            parent: 'package.info'
        }
    });
}]);
(function () {
    'use strict';

    /**
     * @ngdoc module
     * @name  Bastion.package-groups
     *
     * @description
     *   Module for Package Group related functionality.
     */
    angular
        .module('Bastion.package-groups', [
            'ui.router',
            'ngResource',
            'Bastion',
            'Bastion.i18n',
            'Bastion.components'
        ]);

})();
(function () {
    'use strict';

    /**
     * @ngdoc factory
     * @name  Bastion.package-groups.factory:PackageGroup
     *
     * @description
     *   Provides a BastionResource for interacting with Package Groups
     */
    function PackageGroup(BastionResource) {
        return BastionResource('katello/api/v2/package_groups/:id',
            {'id': '@id'},
            {
              autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}}
            }
        );
    }

    angular
        .module('Bastion.package-groups')
        .factory('PackageGroup', PackageGroup);

    PackageGroup.$inject = ['BastionResource'];

})();
(function () {
    'use strict';

    /**
     * @ngdoc module
     * @name  Bastion.puppet-modules
     *
     * @description
     *   Module for Puppet Module related functionality.
     */
    angular
        .module('Bastion.puppet-modules', [
            'ngResource',
            'ui.router',
            'Bastion',
            'Bastion.common',
            'Bastion.i18n'
        ]);

})();
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.puppet-modules.controller:PuppetModuleContentViewsController
     *
     * @description
     *   Provides the functionality for the puppet modules details contentViews page.
     */
    function PuppetModuleContentViewsController($scope, Nutupane, ContentViewVersion, CurrentOrganization) {
        var contentViewsNutupane,
            params = {
                'puppet_module_id': $scope.$stateParams.puppetModuleId,
                'organization_id': CurrentOrganization
            };

        contentViewsNutupane = new Nutupane(ContentViewVersion, params);
        $scope.controllerName = 'katello_content_views';
        contentViewsNutupane.masterOnly = true;
        contentViewsNutupane.setSearchKey('contentViewsSearch');

        $scope.table = contentViewsNutupane.table;

        $scope.environmentNames = function (environments) {
            var names = _.map(environments, function (environment) {
                return environment.name;
            });

            return names.join(',');
        };
    }

    angular
        .module('Bastion.puppet-modules')
        .controller('PuppetModuleContentViewsController', PuppetModuleContentViewsController);

    PuppetModuleContentViewsController.$inject = [
        '$scope',
        'Nutupane',
        'ContentViewVersion',
        'CurrentOrganization'
    ];

})();
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.puppet-modules.controller:PuppetModuleRepositoriesController
     *
     * @description
     *   Provides the functionality for the puppet modules details repositories page.
     */
    function PuppetModuleRepositoriesController($scope, Nutupane, Repository, CurrentOrganization) {
        var repositoriesNutupane,
            params = {
                'puppet_module_id': $scope.$stateParams.puppetModuleId,
                'organization_id': CurrentOrganization
            };

        repositoriesNutupane = new Nutupane(Repository, params);
        $scope.controllerName = 'katello_repositories';
        repositoriesNutupane.masterOnly = true;
        repositoriesNutupane.setSearchKey('repositoriesSearch');

        $scope.table = repositoriesNutupane.table;
    }

    angular
        .module('Bastion.puppet-modules')
        .controller('PuppetModuleRepositoriesController', PuppetModuleRepositoriesController);

    PuppetModuleRepositoriesController.$inject = [
        '$scope',
        'Nutupane',
        'Repository',
        'CurrentOrganization'
    ];

})();
(function () {

    /**
     * @ngdoc object
     * @name  Bastion.puppet-modules.controller:PuppetModuleController
     *
     * @description
     *   Provides the functionality for the puppet modules details action pane.
     */
    function PuppetModuleController($scope, PuppetModule, ApiErrorHandler) {
        $scope.panel = {
            error: false,
            loading: true
        };

        if ($scope.puppetModule) {
            $scope.panel.loading = false;
        }

        $scope.puppetModule = PuppetModule.get({id: $scope.$stateParams.puppetModuleId}, function () {
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });
    }

    angular
        .module('Bastion.puppet-modules')
        .controller('PuppetModuleController', PuppetModuleController);

    PuppetModuleController.$inject = ['$scope', 'PuppetModule', 'ApiErrorHandler'];

})();
(function () {
    'use strict';

    /**
     * @ngdoc factory
     * @name  Bastion.puppet-modules.factory:PuppetModule
     *
     * @description
     *   Provides a BastionResource for interacting with Puppet Modules
     */
    function PuppetModule(BastionResource) {

        return BastionResource('katello/api/v2/puppet_modules/:id',
            {'id': '@id'},
            {
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}}
            }
        );

    }

    angular
        .module('Bastion.puppet-modules')
        .factory('PuppetModule', PuppetModule);

    PuppetModule.$inject = ['BastionResource'];

})();
(function () {
    'use strict';

    /**
     * @ngdoc controller
     * @name  Bastion.puppet-modules.controller:PuppetModulesController
     *
     * @description
     *   Handles fetching puppet modules content and populating Nutupane based on the current
     *   ui-router state.
     */
    function PuppetModulesController($scope, Nutupane, PuppetModule, CurrentOrganization) {
        var nutupane;

        var params = {
            'organization_id': CurrentOrganization,
            'sort_by': 'name',
            'sort_order': 'ASC'
        };

        nutupane = new Nutupane(PuppetModule, params);
        $scope.controllerName = 'katello_puppet_modules';
        nutupane.masterOnly = true;

        $scope.table = nutupane.table;
        $scope.controllerName = 'katello_puppet_modules';
    }

    angular
        .module('Bastion.puppet-modules')
        .controller('PuppetModulesController', PuppetModulesController);

    PuppetModulesController.$inject = ['$scope', 'Nutupane', 'PuppetModule', 'CurrentOrganization'];

})();
(function () {
    'use strict';

    /**
     * @ngdoc object
     * @name Bastion.puppet-modules.config
     *
     * @requires $stateProvider
     *
     * @description
     *   State routes defined for the docker tags module.
     */
    function PuppetModulesConfig($stateProvider) {
        $stateProvider.state('puppet-modules', {
            url: '/puppet_modules',
            permission: ['view_products', 'view_content_views'],
            views: {
                '@': {
                    controller: 'PuppetModulesController',
                    templateUrl: 'puppet-modules/views/puppet-modules.html'
                }
            },
            ncyBreadcrumb: {
                label: "{{ 'Puppet Modules' | translate }}"
            }
        })
        .state('puppet-module', {
            abstract: true,
            url: '/puppet_modules/:puppetModuleId',
            permission: ['view_products', 'view_content_views'],
            controller: 'PuppetModuleController',
            templateUrl: 'puppet-modules/details/views/puppet-module.html'
        })
        .state('puppet-module.info', {
            url: '',
            permission: ['view_products', 'view_content_views'],
            templateUrl: 'puppet-modules/details/views/puppet-module-info.html',
            ncyBreadcrumb: {
                label: "{{ puppetModule.name }}",
                parent: 'puppet-modules'
            }
        })
        .state('puppet-module.repositories', {
            url: '/repositories',
            permission: ['view_products'],
            controller: 'PuppetModuleRepositoriesController',
            templateUrl: 'puppet-modules/details/views/puppet-module-repositories.html',
            ncyBreadcrumb: {
                label: "{{ 'Repositories' | translate }}",
                parent: 'puppet-module.info'
            }
        })
        .state('puppet-module.content-views', {
            url: '/content_views',
            collapsed: true,
            permission: ['view_content_views'],
            controller: 'PuppetModuleContentViewsController',
            templateUrl: 'puppet-modules/details/views/puppet-module-content-views.html',
            ncyBreadcrumb: {
                label: "{{ 'Content Views' | translate }}",
                parent: 'puppet-module.info'
            }
        });
    }

    angular
        .module('Bastion.puppet-modules')
        .config(PuppetModulesConfig);

    PuppetModulesConfig.$inject = ['$stateProvider'];
})();
/**
 * @ngdoc module
 * @name  Bastion.repositories
 *
 * @description
 *   Module for repository related functionality.
 */

angular.module('Bastion.repositories', [
    'ngResource',
    'ui.router',
    'ngUpload',
    'Bastion',
    'Bastion.i18n',
    'Bastion.utils',
    'Bastion.common',
    'Bastion.components',
    'Bastion.components.formatters',
    'Bastion.packages',
    'Bastion.products',
    'Bastion.docker-manifests',
    'Bastion.docker-manifest-lists',
    'Bastion.files',
    'Bastion.debs'
]);

angular.module('Bastion.repositories').run(['$rootScope', '$state', '$stateParams',
    function ($rootScope, $state, $stateParams) {
        $rootScope.$state = $state;
        $rootScope.$stateParams = $stateParams;
    }
]);
/**
 * @ngdoc service
 * @name  Bastion.repository.service:checksum
 *
 * @requires translate
 *
 * @description
 *   Provides a checksum for repositories
 */

angular.module('Bastion.repositories').service('Checksum',
    ['translate', function (translate) {

        this.checksums = [{name: translate('Default'), id: null}, {id: 'sha256', name: 'sha256'}, {id: 'sha1', name: 'sha1'}];

        this.checksumType = function (checksum) {
            if (checksum === null) {
                checksum = translate('Default');
            }
            return checksum;
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.repositories.controller:RepositoryAdvancedSyncController
 *
 * @requires $scope
 * @requires $state
 * @requires Notification
 * @requires Repository
 *
 * @description
 *   Provides the functionality for advanced repository syncing
 */

angular.module('Bastion.repositories').controller('RepositoryAdvancedSyncController',
    ['$scope', '$state', 'Notification', 'Repository',
    function ($scope, $state, Notification, Repository) {
        var errorHandler = function errorHandler(response) {
            angular.forEach(response.data.errors, function (error) {
                Notification.setErrorMessage(error);
            });
            $scope.working = false;
        };

        $scope.advancedSync = function (syncType) {
            var params = {id: $scope.repository.id};
            if (syncType === "skipMetadataCheck") {
                params['skip_metadata_check'] = true;
            } else if (syncType === "validateContents") {
                params['validate_contents'] = true;
            }

            Repository.sync(params, function (task) {
                $state.go('product.repository.tasks.details', {taskId: task.id});
            }, errorHandler);
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.repositories.controller:RepositoryDetailsInfoController
 *
 * @requires $scope
 * @requires $q
 * @requires translate
 * @requires Notification
 * @requires GPGKey
 * @requires CurrentOrganization
 * @requires Checksum
 * @requires DownloadPolicy
 * @requires OstreeUpstreamSyncPolicy
 * @requires Architecture
 *
 * @description
 *   Provides the functionality for the repository details info page.
 */

angular.module('Bastion.repositories').controller('RepositoryDetailsInfoController',
    ['$scope', '$q', 'translate', 'Notification', 'GPGKey', 'CurrentOrganization', 'Checksum', 'DownloadPolicy', 'OstreeUpstreamSyncPolicy', 'Architecture',
    function ($scope, $q, translate, Notification, GPGKey, CurrentOrganization, Checksum, DownloadPolicy, OstreeUpstreamSyncPolicy, Architecture) {
        $scope.organization = CurrentOrganization;

        $scope.progress = {uploading: false};

        $scope.repository.$promise.then(function () {
            $scope.uploadURL = 'katello/api/v2/repositories/' + $scope.repository.id + '/upload_content';
        });

        $scope.gpgKeys = function () {
            var deferred = $q.defer();

            GPGKey.queryUnpaged(function (gpgKeys) {
                var results = gpgKeys.results;

                results.unshift({id: null});
                deferred.resolve(results);
            });

            return deferred.promise;
        };

        $scope.architectures = function () {
            var deferred = $q.defer();
            Architecture.queryUnpaged(function (architectures) {
                var results = architectures.results;
                results.map(function(i) {
                    i.id = i.name;
                });
                results.unshift({
                    id: 'noarch',
                    name: translate('Default'),
                    value: null
                });
                deferred.resolve(results);
            });
            return deferred.promise;
        };

        $scope.save = function (repository) {
            var deferred = $q.defer();

            repository.$update(function (response) {
                deferred.resolve(response);
                Notification.setSuccessMessage(translate('Repository Saved.'));
            }, function (response) {
                deferred.reject(response);
                _.each(response.data.errors, function (errorMessage) {
                    Notification.setErrorMessage(translate("An error occurred saving the Repository: ") + errorMessage);
                });
            });

            return deferred.promise;
        };

        $scope.uploadContent = function (content) {
            var returnData, error, uploaded;

            if (content) {
                try {
                    returnData = angular.fromJson(angular.element(content).html());
                } catch (err) {
                    returnData = content;
                }

                if (!returnData) {
                    returnData = content;
                }

                if (returnData !== null && returnData.status === 'success') {
                    uploaded = returnData.filenames.join(', ');
                    Notification.setSuccessMessage(translate('Successfully uploaded content: ') + uploaded);
                    $scope.repository.$get();
                } else {
                    error = returnData.displayMessage;
                    Notification.setErrorMessage(translate('Error during upload: ') + error);
                }

                $scope.progress.uploading = false;
            }
        };

        $scope.uploadError = function (error, content) {
            if (angular.isString(content) && content.indexOf("Request Entity Too Large")) {
                error = translate('File too large. Please use the CLI instead.');
            } else {
                error = content;
            }
            Notification.setErrorMessage(translate('Error during upload: ') + error);
            $scope.progress.uploading = false;
        };

        $scope.checksums = Checksum.checksums;
        $scope.downloadPolicies = DownloadPolicy.downloadPolicies;
        $scope.ostreeUpstreamSyncPolicies = OstreeUpstreamSyncPolicy.syncPolicies;

        $scope.checksumTypeDisplay = function (checksum) {
            return Checksum.checksumType(checksum);
        };

        $scope.downloadPolicyDisplay = function (downloadPolicy) {
            return DownloadPolicy.downloadPolicyName(downloadPolicy);
        };

        $scope.ostreeUpstreamSyncPolicyDisplay = function (repository) {
            var policy = repository["ostree_upstream_sync_policy"];
            if ( policy === "custom") {
                return OstreeUpstreamSyncPolicy.syncPolicyName(policy, repository["ostree_upstream_sync_depth"]);
            }
            return OstreeUpstreamSyncPolicy.syncPolicyName(policy);
        };

        $scope.clearUpstreamPassword = function () {
            $scope.repository['upstream_password'] = null;
            $scope.repository['upstream_password_exists'] = false;
            $scope.save($scope.repository);
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.repositories.controller:OstreeUpstreamSyncPolicyFilter
 *
 * @requires translate
 * @requires OstreeUpstreamSyncPolicy
 *
 * @description
 *   Provides the Ostree Upstream Sync policy filter functionality for the repository details info page.
 */


angular.module('Bastion.components.formatters').filter('ostreeUpstreamSyncPolicyFilter', ['translate', 'OstreeUpstreamSyncPolicy', function (translate, OstreeUpstreamSyncPolicy) {
    return function (displayValue, repository) {
        var policy = repository["ostree_upstream_sync_policy"];
        if ( policy === "custom") {
            return OstreeUpstreamSyncPolicy.syncPolicyName(policy, repository["ostree_upstream_sync_depth"]);
        }
        return OstreeUpstreamSyncPolicy.syncPolicyName(policy);
    };
}]);

angular.module('Bastion.components.formatters').filter('upstreamPasswordFilter', [function () {
    return function (displayValue, repository) {
        if (repository["upstream_password_exists"]) {
            return "******";
        }
        return null;
    };
}]);
/**
 * @ngdoc object
 * @name  Bastion.repositories.controller:RepositoryManageContentController
 *
 * @requires $scope
 * @requries $state
 * @requires translate
 * @requires Notification
 * @requires Nutupane
 * @requires Repository
 * @requires Package
 * @requires PackageGroup
 * @requires PuppetModule
 * @requires DockerManifest
 * @requires DockerManifestList
 * @requires OstreeBranch
 * @requires File
 * @requires Deb
 *
 * @description
 *   Provides the functionality for the repository details pane.
 */

angular.module('Bastion.repositories').controller('RepositoryManageContentController',
    ['$scope', '$state', 'translate', 'Notification', 'Nutupane', 'Repository', 'Package', 'PackageGroup', 'PuppetModule', 'DockerManifest', 'DockerManifestList', 'OstreeBranch', 'File', 'Deb',
    function ($scope, $state, translate, Notification, Nutupane, Repository, Package, PackageGroup, PuppetModule, DockerManifest, DockerManifestList, OstreeBranch, File, Deb) {
        var contentTypes;

        function success(response, selected) {
            var message;

            $scope.contentNutupane.refresh();
            $scope.table.working = true;

            if (selected.length === 1) {
                message = translate("Successfully removed 1 item.");
            } else {
                message = translate("Successfully removed %s items.").replace('%s', selected.length);
            }
            Notification.setSuccessMessage(message);
            $scope.generationTaskId = response.output['task_id'];
        }

        function error(data) {
            $scope.table.working = true;
            Notification.setErrorMessage(data.response.displayMessage);
        }

        $scope.repository = Repository.get({id: $scope.$stateParams.repositoryId}, function (repository) {
            $scope.product = repository.product;
        });

        $scope.currentState = $state.current.name.split('.').pop();

        contentTypes = {
            'packages': { type: Package },
            'package-groups': { type: PackageGroup },
            'puppet-modules': { type: PuppetModule },
            'docker-manifests': { type: DockerManifest },
            'docker-manifest-lists': { type: DockerManifestList },
            'ostree-branches': { type: OstreeBranch },
            'files': { type: File },
            'debs': { type: Deb }
        };

        $scope.contentNutupane = new Nutupane(contentTypes[$scope.currentState].type, {
            'repository_id': $scope.$stateParams.repositoryId
        });
        $scope.table = $scope.contentNutupane.table;
        $scope.contentNutupane.masterOnly = true;

        $scope.removeContent = function () {
            var selected = $scope.table.getSelected();
            $scope.table.working = true;
            Repository.removeContent({id: $scope.repository.id, ids: _.map(selected, 'id')},
                function (response) {
                    success(response, selected);
                }, error);
        };

        $scope.taskUrl = function () {
            return $scope.$state.href('product.tasks.details', {productId: $scope.product.id,
                taskId: $scope.generationTaskId});
        };

        $scope.clearTaskId = function () {
            $scope.generationTaskId = undefined;
        };

        $scope.tagsForManifest = function(manifest) {
            return _.filter(manifest.tags, function(tag) {
                return tag.repository_id + '' === $scope.$stateParams.repositoryId;
            });
        };

        $scope.updateSelectable = function(item) {
            if ($scope.currentState === "docker-manifests" && !_.isEmpty(item.manifest_lists)) {
                item.unselectable = true;
            }
            return item;
        };
    }]
);
(function () {
    function RepositoryDetailsController($scope, $state, translate, Repository, Product, ApiErrorHandler, Notification) {
        /**
         * @ngdoc object
         * @name  Bastion.repositories.controller:RepositoryDetailsController
         *
         * @requires $scope
         * @requires $state
         * @requires translate
         * @requires Repository
         * @requires ApiErrorHandler
         * @requires Notification
         *
         * @description
         *   Core functionality for Repository.
         */

        var errorHandler = function errorHandler(response) {
            angular.forEach(response.data.errors, function (error) {
                Notification.setErrorMessage(error);
            });
        };

        $scope.page = {
            error: false,
            loading: true
        };

        if ($scope.repository) {
            $scope.page.loading = false;
        }

        $scope.product = Product.get({id: $scope.$stateParams.productId}, function () {
            $scope.page.loading = false;
        }, function (response) {
            $scope.page.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        $scope.repository = Repository.get({
            'product_id': $scope.$stateParams.productId,
            'id': $scope.$stateParams.repositoryId
        }, function () {
            $scope.page.loading = false;
        }, function (response) {
            $scope.page.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        $scope.hideSyncButton = function(repository, advancedSync) {
            var result = $scope.syncInProgress(repository.last_sync) || !repository.url || $scope.denied('sync_products', $scope.product);
            if (advancedSync) {
                result = result || repository.content_type !== 'yum';
            }
            return result;
        };

        $scope.disableSyncLink = function (adavancedSync) {
            return $scope.hideSyncButton($scope.repository, adavancedSync);
        };

        $scope.syncRepository = function (repository) {
            Repository.sync({id: repository.id}, function (task) {
                $state.go('product.repository.tasks.details', {taskId: task.id});
            }, errorHandler);
        };

        $scope.republishRepository = function (repository) {
            Repository.republish({id: repository.id}, function (task) {
                $state.go('product.repository.tasks.details', {taskId: task.id});
            }, errorHandler);
        };

        $scope.syncInProgress = function (task) {
            var inProgress = false;
            if (task && (task.state === 'pending' || task.state === 'running')) {
                inProgress = true;
            }
            return inProgress;
        };

        $scope.removeRepository = function (repository) {
            var success, repositoryName = repository.name;

            success = function () {
                $scope.transitionTo('product.repositories', {productId: $scope.$stateParams.productId});
                Notification.setSuccessMessage(translate('Repository "%s" successfully deleted').replace('%s', repositoryName));
            };

            repository.$delete(success, errorHandler);
        };

        $scope.canRemove = function (repo, product) {
            return $scope.getRepoNonDeletableReason(repo, product) === null;
        };

        $scope.getRepoNonDeletableReason = function (repo, product) {
            var readOnlyReason = null;

            if (repo.$resolved && product.$resolved) {
                if (repo.promoted) {
                    readOnlyReason = 'published';
                } else if (repo['product_type'] === "redhat") {
                    readOnlyReason = 'redhat';
                } else if ($scope.denied('deletable', repo)) {
                    readOnlyReason = 'permissions';
                }
            }
            return readOnlyReason;
        };
    }

    angular.module('Bastion.repositories').controller('RepositoryDetailsController', RepositoryDetailsController);

    RepositoryDetailsController.$inject = ['$scope', '$state', 'translate', 'Repository', 'Product', 'ApiErrorHandler', 'Notification'];
})();
/**
 * @ngdoc service
 * @name  Bastion.repository.service:downloadPolicy
 *
 * @requires translate
 *
 * @description
 *   Provides a downloadPolicy for repositories
 */

angular.module('Bastion.repositories').service('DownloadPolicy',
    ['translate', function (translate) {

        this.downloadPolicies = {
            'on_demand': translate('On Demand'),
            'background': translate('Background'),
            'immediate': translate('Immediate')
        };

        this.downloadPolicyName = function (policy) {
            return this.downloadPolicies[policy];
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.repositories.controller:NewRepositoryController
 *
 * @requires $scope
 * @requires Repository
 * @requires Product
 * @requires GPGKey
 * @requires FormUtils
 * @requires translate
 * @requires Notification
 * @requires ApiErrorHandler
 * @requires BastionConfig
 * @requires Checksum
 * @requires DownloadPolicy
 * @requires OstreeUpstreamSyncPolicy
 * @requires Architecture
 *
 * @description
 *   Controls the creation of an empty Repository object for use by sub-controllers.
 */

angular.module('Bastion.repositories').controller('NewRepositoryController',
    ['$scope', 'Repository', 'Product', 'GPGKey', 'FormUtils', 'translate', 'Notification', 'ApiErrorHandler', 'BastionConfig', 'Checksum', 'DownloadPolicy', 'OstreeUpstreamSyncPolicy', 'Architecture',
    function ($scope, Repository, Product, GPGKey, FormUtils, translate, Notification, ApiErrorHandler, BastionConfig, Checksum, DownloadPolicy, OstreeUpstreamSyncPolicy, Architecture) {

        function success() {
            Notification.setSuccessMessage(translate('Repository %s successfully created.').replace('%s', $scope.repository.name));
            $scope.transitionTo('product.repositories', {productId: $scope.$stateParams.productId});
        }

        function error(response) {
            var foundError = false;
            $scope.working = false;

            angular.forEach($scope.repositoryForm, function (field) {
                if ($scope.repositoryForm.hasOwnProperty(field) && field.hasOwnProperty('$modelValue')) {
                    field.$setValidity('server', true);
                    $scope.repositoryForm[field].$error.messages = [];
                }
            });

            angular.forEach(response.data.errors, function (errors, field) {
                if ($scope.repositoryForm.hasOwnProperty(field)) {
                    foundError = true;
                    $scope.repositoryForm[field].$setValidity('server', false);
                    $scope.repositoryForm[field].$error.messages = errors;
                }
            });

            if (!foundError) {
                Notification.setErrorMessage(response.data.displayMessage);
            }
        }

        $scope.page = {
            error: false,
            loading: true
        };

        $scope.repository = new Repository({'product_id': $scope.$stateParams.productId, unprotected: true,
            'checksum_type': null, 'mirror_on_sync': true, 'verify_ssl_on_sync': true,
            'download_policy': BastionConfig.defaultDownloadPolicy, 'arch': null,
            'ostree_upstream_sync_policy': 'latest'});

        $scope.product = Product.get({id: $scope.$stateParams.productId}, function () {
            $scope.page.loading = false;
        }, function (response) {
            $scope.page.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        Repository.repositoryTypes({'creatable': true}, function (data) {
            $scope.repositoryTypes = data;
        });

        $scope.checksums = Checksum.checksums;
        $scope.downloadPolicies = DownloadPolicy.downloadPolicies;
        $scope.ostreeUpstreamSyncPolicies = OstreeUpstreamSyncPolicy.syncPolicies;

        $scope.$watch('repository.name', function () {
            if ($scope.repositoryForm && $scope.repositoryForm.name) {
                $scope.repositoryForm.name.$setValidity('server', true);
                FormUtils.labelize($scope.repository);
            }
        });

        GPGKey.queryUnpaged(function (gpgKeys) {
            $scope.gpgKeys = gpgKeys.results;
        });

        Architecture.queryUnpaged(function (architecture) {
            var results = architecture.results;
            var noarch = {
                id: 'noarch',
                name: translate('Default'),
                value: null
            };
            results.map(function(i) {
                i.id = i.name;
            });
            results.unshift(noarch);
            $scope.architecture = results;
            $scope.repository.arch = results[0].id;
        });

        $scope.save = function (repository) {
            if (repository.content_type === 'ostree') {
                repository.unprotected = false;
            }
            if (repository.content_type !== 'yum') {
                repository['download_policy'] = '';
            }
            if (repository.arch === 'Default') {
                repository.arch = null;
            }
            repository.$save(success, error);
        };

    }]
);
/**
 * @ngdoc service
 * @name  Bastion.repository.service:ostreeUpstreamSyncPolicy
 *
 * @requires translate
 *
 * @description
 *   Provides a ostree upstream syncPolicies for repositories
 */

angular.module('Bastion.repositories').service('OstreeUpstreamSyncPolicy',
    ['translate', function (translate) {

        this.syncPolicies = {
            'latest': translate('Latest Only'),
            'all': translate('All History'),
            'custom': translate('Custom Depth')
        };

        this.syncPolicyName = function (policy, depth) {
            if (policy === "custom") {
                return translate('Custom Depth (Currently %s)').replace('%s', depth.toString());
            }
            return this.syncPolicies[policy];
        };
    }]
);
(function () {
    function RepositoryRoutes($stateProvider) {
        /**
         * @ngdoc object
         * @name  Bastion.repositories:RepositoryRoutes
         *
         * @requires $stateProvider
         *
         * @description
         *   Routes for repository pages.
         */

        $stateProvider.state('product.repositories', {
            url: '/repositories',
            permission: 'view_products',
            controller: 'ProductRepositoriesController',
            templateUrl: 'products/details/repositories/views/product-repositories.html',
            ncyBreadcrumb: {
                label: "{{ 'Repositories' | translate }}",
                parent: 'product.info'
            }
        })
        .state('product.repositories.new', {
            url: '/new',
            permission: 'create_products',
            views: {
                '@product': {
                    controller: 'NewRepositoryController',
                    templateUrl: 'products/details/repositories/new/views/new-repository.html'
                }
            },
            ncyBreadcrumb: {
                label: "{{ 'New Repository' | translate }}",
                parent: 'product.repositories'
            }
        });

        $stateProvider.state("product.repository", {
            abstract: true,
            url: '/repositories/:repositoryId',
            permission: 'view_products',
            views: {
                '@': {
                    controller: 'RepositoryDetailsController',
                    templateUrl: 'products/details/repositories/details/views/repository-details.html'
                }
            }
        })
        .state('product.repository.info', {
            url: '',
            permission: 'view_products',
            controller: 'RepositoryDetailsInfoController',
            templateUrl: 'products/details/repositories/details/views/repository-info.html',
            ncyBreadcrumb: {
                label: "{{ repository.name }}",
                parent: 'product.repositories'
            }
        })
        .state('product.repository.advanced_sync', {
            url: '/advanced_sync',
            permission: 'view_products',
            controller: 'RepositoryAdvancedSyncController',
            templateUrl: 'products/details/repositories/details/views/repository-advanced-sync.html',
            ncyBreadcrumb: {
                label: "{{ 'Advanced Sync' | translate }}",
                parent: 'product.repository.info'
            }
        })
        .state('product.repository.manage-content', {
            abstract: true,
            url: '/content',
            controller: 'RepositoryManageContentController',
            template: '<div ui-view></div>'
        })
        .state('product.repository.manage-content.packages', {
            url: '/packages',
            permission: 'view_products',
            templateUrl: 'products/details/repositories/details/views/repository-manage-packages.html',
            ncyBreadcrumb: {
                label: "{{'Packages' | translate }}",
                parent: 'product.repository.info'
            }
        })
        .state('product.repository.manage-content.package-groups', {
            url: '/package_groups',
            permission: 'view_products',
            templateUrl: 'products/details/repositories/details/views/repository-manage-package-groups.html',
            ncyBreadcrumb: {
                label: "{{'Package Groups' | translate }}",
                parent: 'product.repository.info'
            }
        })
        .state('product.repository.manage-content.puppet-modules', {
            url: '/content/puppet_modules',
            permission: 'view_products',
            templateUrl: 'products/details/repositories/details/views/repository-manage-puppet-modules.html',
            ncyBreadcrumb: {
                label: "{{'Manage Puppet Modules' | translate }}",
                parent: 'product.repository.info'
            }
        })
        .state('product.repository.manage-content.docker-manifests', {
            url: '/content/docker_manifests',
            permission: 'view_products',
            templateUrl: 'products/details/repositories/details/views/repository-manage-docker-manifests.html',
            ncyBreadcrumb: {
                label: "{{'Docker Manifests' | translate }}",
                parent: 'product.repository.info'
            }
        })
        .state('product.repository.manage-content.docker-manifest-lists', {
            url: '/content/docker_manifest_lists',
            permission: 'view_products',
            templateUrl: 'products/details/repositories/details/views/repository-manage-docker-manifest-lists.html',
            ncyBreadcrumb: {
                label: "{{'Docker Manifest Lists' | translate }}",
                parent: 'product.repository.info'
            }
        })
        .state('product.repository.manage-content.files', {
            url: '/content/files',
            permission: 'view_products',
            templateUrl: 'products/details/repositories/details/views/repository-manage-files.html',
            ncyBreadcrumb: {
                label: "{{'Files' | translate }}",
                parent: 'product.repository.info'
            }
        })
        .state('product.repository.manage-content.ostree-branches', {
            url: '/content/ostree_branches',
            permission: 'view_products',
            templateUrl: 'products/details/repositories/details/views/repository-manage-ostree-branches.html',
            ncyBreadcrumb: {
                label: "{{'OSTree Branches' | translate }}",
                parent: 'product.repository.info'
            }
        })
        .state('product.repository.manage-content.debs', {
            url: '/content/debs',
            permission: 'view_products',
            templateUrl: 'products/details/repositories/details/views/repository-manage-debs.html',
            ncyBreadcrumb: {
                label: "{{'Debs' | translate }}",
                parent: 'product.repository.info'
            }
        });

        $stateProvider.state('product.repository.tasks', {
            abstract: true,
            template: '<div ui-view></div>'
        })
        .state('product.repository.tasks.index', {
            url: '/tasks',
            permission: 'view_repositories',
            templateUrl: 'products/details/repositories/details/views/repository-tasks.html',
            ncyBreadcrumb: {
                label: "{{'Tasks' | translate }}",
                parent: 'product.repository.info'
            }
        })
        .state('product.repository.tasks.details', {
            url: '/tasks/:taskId',
            permission: 'view_repositories',
            controller: 'TaskDetailsController',
            templateUrl: 'tasks/views/task-details.html',
            ncyBreadcrumb: {
                label: "{{ task.id }}",
                parent: 'product.repository.tasks.index'
            }
        });
    }

    angular.module('Bastion.repositories').config(RepositoryRoutes);
    RepositoryRoutes.$inject = ['$stateProvider'];
})();
/**
 * @ngdoc service
 * @name  Bastion.repositories.factory:Repository
 *
 * @requires BastionResource
 * @requires CurrentOrganization
 *
 * @description
 *   Provides a BastionResource for product or list of repositories.
 */

angular.module('Bastion.repositories').factory('Repository',
    ['BastionResource', 'CurrentOrganization',
    function (BastionResource, CurrentOrganization) {

        return BastionResource('katello/api/v2/repositories/:id/:action',
            {id: '@id', 'organization_id': CurrentOrganization},
            {
                update: { method: 'PUT' },
                sync: { method: 'POST', params: { action: 'sync' } },
                removePackages: { method: 'PUT', params: { action: 'remove_packages'}},
                removeContent: { method: 'PUT', params: { action: 'remove_content'}},
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}},
                repositoryTypes: {method: 'GET', isArray: true, params: {id: 'repository_types'}},
                republish: {method: 'PUT', params: { action: 'republish' }}
            }
        );

    }]
);

/**
 * @ngdoc service
 * @name  Bastion.repositories.factory:RepositoryBulkAction
 *
 * @requires BastionResource
 * @requires CurrentOrganization
 *
 * @description
 *   Provides a BastionResource for bulk actions on repositories.
 */
angular.module('Bastion.repositories').factory('RepositoryBulkAction',
    ['BastionResource', 'CurrentOrganization', function (BastionResource, CurrentOrganization) {

        return BastionResource('katello/api/v2/repositories/bulk/:action',
            {'organization_id': CurrentOrganization},
            {
                removeRepositories: {method: 'PUT', params: {action: 'destroy'}},
                syncRepositories: {method: 'POST', params: {action: 'sync'}}
            }
        );

    }]
);
/**
 * @ngdoc module
 * @name  Bastion.products
 *
 * @description
 *   Module for product related functionality.
 */

angular.module('Bastion.products', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.components',
    'Bastion.gpg-keys',
    'Bastion.architectures',
    'Bastion.i18n',
    'Bastion.sync-plans',
    'Bastion.tasks',
    'Bastion.utils'
]);
/**
 * @ngdoc service
 * @name  Bastion.products.factory:ProductBulkAction
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for bulk actions on products.
 */

angular.module('Bastion.products').factory('ProductBulkAction',
    ['BastionResource', function (BastionResource) {

        return BastionResource('katello/api/v2/products/bulk/:action', {}, {
            removeProducts: {method: 'PUT', params: {action: 'destroy'}},
            syncProducts: {method: 'PUT', params: {action: 'sync'}},
            updateProductSyncPlan: {method: 'PUT', params: {action: 'sync_plan'}}
        });
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.products.controller:ProductsBulkAdvancedSyncModalController
 *
 * @requires $scope
 * @requires $state
 * @requires $sce
 * @requires translate
 * @requires ProductBulkAction
 * @requires CurrentOrganization
 * @requires Notification
 * @requires $uibModalInstance
 * @requires bulkParams
 *
 * @description
 *   A controller for providing bulk action functionality to the products page.
 */

angular.module('Bastion.products').controller('ProductsBulkAdvancedSyncModalController',
    ['$scope', '$state', '$sce', 'translate', 'ProductBulkAction', 'CurrentOrganization', 'Notification', '$uibModalInstance', 'bulkParams',
        function ($scope, $state, $sce, translate, ProductBulkAction, CurrentOrganization, Notification, $uibModalInstance, bulkParams) {
            var success, error;

            success = function (task) {
                var url, message, taskLink;
                url = $state.href('task', {taskId: task.id});
                taskLink = $sce.trustAsHtml("<a href=" + url + ">here</a>");
                message = translate("Product syncs has been initiated in the background. " +
                    "Click %s to monitor the progress.");

                Notification.setRenderedSuccessMessage(message.replace('%s', taskLink));
            };

            error = function (response) {
                angular.forEach(response.data.errors, function(message) {
                    Notification.setErrorMessage(translate("An error occurred initiating the sync: " ) + message);
                });
            };

            $scope.ok = function () {
                var params = bulkParams;
                if ($scope.syncType === "skipMetadataCheck") {
                    params['skip_metadata_check'] = true;
                } else if ($scope.syncType === "validateContents") {
                    params['validate_contents'] = true;
                }
                ProductBulkAction.syncProducts(params, success, error);
                $uibModalInstance.close();
            };

            $scope.cancel = function () {
                $uibModalInstance.dismiss('cancel');
            };
        }]
);
/**
 * @ngdoc object
 * @name  Bastion.products.controller:ProductsBulkSyncPlanModalController
 *
 * @requires $scope
 * @requires $state
 * @requires translate
 * @requires Nutupane
 * @requires ProductBulkAction
 * @requires SyncPlan
 * @requires CurrentOrganization
 * @requires Notification
 * @requires $uibModalInstance
 * @requires bulkParams
 *
 * @description
 *   A controller for providing bulk action functionality to the products page.
 */

angular.module('Bastion.products').controller('ProductsBulkSyncPlanModalController',
    ['$scope', '$state', 'translate', 'Nutupane', 'ProductBulkAction', 'SyncPlan', 'CurrentOrganization', 'Notification', '$uibModalInstance', 'bulkParams',
        function ($scope, $state, translate, Nutupane, ProductBulkAction, SyncPlan, CurrentOrganization, Notification, $uibModalInstance, bulkParams) {
            var success, error, nutupane, params;

            params = {
                'organization_id': CurrentOrganization,
                'offset': 0,
                'sort_by': 'name',
                'sort_order': 'ASC',
                'paged': true
            };

            nutupane = new Nutupane(SyncPlan, params, 'queryPaged');
            $scope.controllerName = 'katello_sync_plans';
            nutupane.masterOnly = true;

            $scope.table = nutupane.table;

            success = function (response) {
                angular.forEach(response.displayMessages.success, function(message) {
                    Notification.setSuccessMessage(message);
                });

                angular.forEach(response.displayMessages.error, function(message) {
                    Notification.setErrorMessage(message);
                });

                nutupane.invalidate();
            };

            error = function (response) {
                angular.forEach(response.data.errors, function(message) {
                    Notification.setErrorMessage(translate("An error occurred updating the sync plan: " ) + message);
                });
            };

            $scope.selectSyncPlan = function (syncPlan) {
                $scope.selectedSyncPlan = syncPlan;
            };

            $scope.updateSyncPlan = function () {
                var updateParams = bulkParams;
                updateParams['plan_id'] = $scope.selectedSyncPlan.id;
                ProductBulkAction.updateProductSyncPlan(updateParams, success, error);
            };

            $scope.removeSyncPlan = function () {
                var removeParams = bulkParams;
                removeParams['plan_id'] = null;
                ProductBulkAction.updateProductSyncPlan(removeParams, success, error);
            };

            $scope.ok = function () {
                $uibModalInstance.close();
            };

            $scope.cancel = function () {
                $uibModalInstance.dismiss('cancel');
            };
        }]
);
/**
 * @ngdoc object
 * @name  Bastion.products.controller:ProductDetailsInfoController
 *
 * @requires $scope
 * @requires $q
 * @requires translate
 * @requires Product
 * @requires SyncPlan
 * @requires GPGKey
 * @requires MenuExpander
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the product details action pane.
 */

angular.module('Bastion.products').controller('ProductDetailsInfoController',
    ['$scope', '$state', '$q', 'translate', 'Product', 'SyncPlan', 'GPGKey', 'MenuExpander', 'Notification',
    function ($scope, $state, $q, translate, Product, SyncPlan, GPGKey, MenuExpander, Notification) {
        $scope.menuExpander = MenuExpander;
        $scope.page = $scope.page || {loading: false};

        $scope.product = $scope.product || Product.get({id: $scope.$stateParams.productId}, function () {
            $scope.page.loading = false;
        });

        $scope.gpgKeys = function () {
            var deferred = $q.defer();

            GPGKey.queryUnpaged(function (gpgKeys) {
                var results = gpgKeys.results;

                results.unshift({id: null});
                deferred.resolve(results);
            });

            return deferred.promise;
        };

        $scope.syncPlans = function () {
            return SyncPlan.queryUnpaged().$promise;
        };

        $scope.save = function (product) {
            var deferred = $q.defer();

            product.$update(function (response) {
                deferred.resolve(response);
                Notification.setSuccessMessage(translate('Product Saved'));
            }, function (response) {
                deferred.reject(response);
                angular.forEach(response.data.errors, function (errorMessage) {
                    Notification.setErrorMessage(translate("An error occurred saving the Product: ") + errorMessage);
                });
            });

            return deferred.promise;
        };

        $scope.$on('$stateChangeSuccess', function (event, toState, toParams, fromState) {
            if (fromState.name === 'product.info.new-sync-plan') {
                $scope.product.$update();
            }
        });
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.products.controller:ProductDetailsController
 *
 * @requires $scope
 * @requires $state
 * @requires $uibModal
 * @requires translate
 * @requires Product
 * @requires Notification
 * @requires ApiErrorHandler
 *
 * @description
 *   Provides the functionality for the product details action pane.
 */

angular.module('Bastion.products').controller('ProductDetailsController',
    ['$scope', '$state', '$uibModal', 'translate', 'Product', 'Notification', 'ApiErrorHandler', function ($scope, $state, $uibModal, translate, Product, Notification, ApiErrorHandler) {
        $scope.page = {
            error: false,
            loading: true
        };

        if ($scope.product) {
            $scope.page.loading = false;
        }

        $scope.product = Product.get({id: $scope.$stateParams.productId}, function () {
            $scope.page.loading = false;
        }, function (response) {
            $scope.page.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        $scope.removeProduct = function (product) {
            product.$delete(function (data) {
                $scope.$emit('productDelete', data.id);
                $scope.transitionTo('products');
            });
        };

        $scope.syncProduct = function () {
            Product.sync({id: $scope.product.id}, function (task) {
                $state.go('product.tasks.details', {taskId: task.id});
            },
            function (response) {
                angular.forEach(response.data.errors, function (error) {
                    Notification.setErrorMessage(error);
                });
            });
        };

        $scope.productDeletable = function(product) {
            return $scope.getReadOnlyReason(product) === null;
        };

        $scope.getReadOnlyReason = function (product) {
            var readOnlyReason = null;

            if (product.$resolved) {
                if (product.redhat) {
                    readOnlyReason = 'redhat';
                } else if ($scope.denied('destroy_products', product)) {
                    readOnlyReason = 'permissions';
                } else if (product['published_content_view_ids'].length > 0) {
                    readOnlyReason = 'published';
                }
            }

            return readOnlyReason;
        };

        $scope.updateProduct = function () {
            function success() {
                Notification.setSuccessMessage(translate('Sync Plan created and assigned to product.'));
            }

            function error(response) {
                ApiErrorHandler.handlePUTRequestErrors(response, $scope);
            }

            $scope.product.$update(success, error);
        };

        $scope.openSyncPlanModal = function () {
            $uibModal.open({
                templateUrl: 'products/new/views/new-sync-plan-modal.html',
                controller: 'NewSyncPlanModalController'
            }).result.then(function ($value) {
                $scope.product['sync_plan_id'] = $value.id;
                $scope.updateProduct();
            });
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.products.controller:ProductRepositoriesController
 *
 * @requires $scope
 * @requires $location
 * @requires Notification
 * @requires translate
 * @requires ApiErrorHandler
 * @requires Product
 * @requires Repository
 * @requires RepositoryBulkAction
 * @requires CurrentOrganization
 * @requires Nutupane
 *
 *
 * @description
 *   Provides the functionality for manipulating repositories attached to a product.
 */

angular.module('Bastion.products').controller('ProductRepositoriesController',
    ['$scope', '$state', '$location', 'Notification', 'translate', 'ApiErrorHandler', 'Product', 'Repository', 'RepositoryBulkAction', 'CurrentOrganization', 'Nutupane',
    function ($scope, $state, $location, Notification, translate, ApiErrorHandler, Product, Repository, RepositoryBulkAction, CurrentOrganization, Nutupane) {
        var repositoriesNutupane = new Nutupane(Repository, {
            'product_id': $scope.$stateParams.productId,
            'search': $location.search().search || "",
            'library': true,
            'organization_id': CurrentOrganization,
            'enabled': true,
            'paged': true
        });
        $scope.controllerName = 'katello_repositories';
        repositoriesNutupane.masterOnly = true;

        function getParams() {
            return {
                ids: repositoriesNutupane.getAllSelectedResults('id').included.ids
            };
        }

        $scope.page = $scope.page || {loading: false};

        $scope.product = Product.get({id: $scope.$stateParams.productId}, function () {
            $scope.page.loading = false;
        }, function (response) {
            $scope.page.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        $scope.table = repositoriesNutupane.table;

        $scope.syncSelectedRepositories = function () {
            var params = getParams();

            RepositoryBulkAction.syncRepositories(params, function (task) {
                $state.go('product.tasks.details', {taskId: task.id});
            },
            function (response) {
                angular.forEach(response.data.errors, function (error) {
                    Notification.setErrorMessage(error);
                });
            });
        };

        $scope.removeSelectedRepositories = function () {
            var success, error, params = getParams(), removalPromise;

            success = function (response) {
                var message = translate('Removal of selected repositories initiated successfully. ');

                message += translate('<a href="/foreman_tasks/tasks/%taskId">Click here to check the status of the task.</a>').
                    replace('%taskId', response.task.id);
                Notification.setSuccessMessage(message);
            };

            error = function (response) {
                angular.forEach(response.data.errors, function (errorMessage) {
                    Notification.setErrorMessage(errorMessage);
                });
            };

            $scope.removingRepositories = true;
            removalPromise = RepositoryBulkAction.removeRepositories(params, success, error).$promise;

            removalPromise.finally(function () {
                $scope.removingRepositories = false;
            });
        };

        $scope.removeRepository = function (repository) {
            repository.$delete(function () {
                $scope.transitionTo('product.repositories', {productId: $scope.$stateParams.productId});
            });
        };

        $scope.getRepositoriesNonDeletableReason = function (product) {
            var readOnlyReason = null;

            if (product.$resolved) {
                if ($scope.denied('destroy_products', product)) {
                    readOnlyReason = 'permissions';
                } else if (product.redhat) {
                    readOnlyReason = 'redhat';
                }
            }

            return readOnlyReason;
        };

        $scope.canRemoveRepositories = function (product) {
            return $scope.getRepositoriesNonDeletableReason(product) === null;
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.products.controller:DiscoveryCreateController
 *
 * @requires $scope
 * @requires $q
 * @requires Notification
 * @requires CurrentOrganization
 * @requires Product
 * @requires Repository
 * @requires GPGKey
 * @requires FormUtils
 * @requires DiscoveryRepositories
 * @requires translate
 *
 * @description
 *   Provides the functionality for the repo creation as part of
 *      repository discovery.
 */

angular.module('Bastion.products').controller('DiscoveryCreateController',
    ['$scope', '$q', 'Notification', 'CurrentOrganization', 'Product', 'Repository', 'GPGKey', 'FormUtils', 'DiscoveryRepositories', 'translate', 'ApiErrorHandler',
    function ($scope, $q, Notification, CurrentOrganization, Product, Repository, GPGKey, FormUtils, DiscoveryRepositories, translate, ApiErrorHandler) {

        $scope.table = {
            rows: DiscoveryRepositories.getRows(),
            resource: {
                total: DiscoveryRepositories.getRows().length,
                subtotal: DiscoveryRepositories.getRows().length
            }
        };

        function convertToResource(repo) {
            var repoParams = {
                'docker_upstream_name': repo.dockerUpstreamName,
                name: repo.name,
                label: repo.label,
                'content_type': repo.contentType,
                url: repo.url + repo.path,
                'product_id': $scope.createRepoChoices.existingProductId,
                unprotected: $scope.createRepoChoices.unprotected,
                'verify_ssl': $scope.createRepoChoices.verifySsl,
                'upstream_username': $scope.createRepoChoices.upstreamUsername,
                'upstream_password': $scope.createRepoChoices.upstreamPassword
            };

            if (repo.contentType === 'docker') {
                repoParams.url = $scope.createRepoChoices.repositoryUrl;
            }

            return new Repository(repoParams);
        }

        function getNextRepoToCreate() {
            var found;
            angular.forEach($scope.table.rows, function (repo) {
                if (angular.isUndefined(found) && repo.created !== true && repo.form.$invalid !== true) {
                    found = repo;
                }
            });
            return found;
        }

        function repoCreateError(response) {
            var currentlyCreating = $scope.currentlyCreating;
            $scope.currentlyCreating = undefined;
            currentlyCreating.messages = response.data.displayMessage;
            currentlyCreating.creating = false;
            currentlyCreating.form.$invalid = true;
        }

        function createNextRepo() {
            var toCreate, repoObject;
            toCreate = getNextRepoToCreate();

            if (toCreate) {
                toCreate.messages = translate("Creating repository...");
                $scope.currentlyCreating = toCreate;
                repoObject = convertToResource(toCreate);
                toCreate.creating = true;
                toCreate.form.$invalid = false;

                repoObject.$save(function (task) {
                    toCreate.creating = false;
                    toCreate.created = true;
                    toCreate.messages = translate("Success!");
                    toCreate.repositoryId = task.id;
                    toCreate.productId = task.product.id;
                    createNextRepo();
                }, function(response) {
                    repoCreateError(response);
                    createNextRepo();
                });
            } else {
                $scope.createRepoChoices.creating = false;
            }
        }

        function productCreateError(response) {
            $scope.createRepoChoices.creating = false;
            angular.forEach(response.data.errors, function (errors, field) {
                Notification.setErrorMessage(translate('An error occurred while creating the Product: %s').replace('%s', field + ' ' + errors));
                if (!angular.isUndefined($scope.productForm[field])) {
                    $scope.productForm[field].$error.messages = errors;
                }
            });
        }

        function productCreateSuccess(response) {
            $scope.createRepoChoices.existingProductId = response.id;
            $scope.createRepoChoices.newProduct = 'false';
            $scope.products.unshift(response);
            createNextRepo();
        }

        $scope.page = {loading: true};
        $scope.$watch('createRepoChoices.product.name', function () {
            FormUtils.labelize($scope.createRepoChoices.product);
        });

        $scope.createRepoChoices = {
            existingProductId: undefined,
            newProduct: 'false',
            product: new Product(),
            unprotected: true,
            creating: false,
            verifySsl: true,
            repositoryUrl: DiscoveryRepositories.getRepositoryUrl(),
            upstreamUsername: DiscoveryRepositories.getUpstreamUsername(),
            upstreamPassword: DiscoveryRepositories.getUpstreamPassword()
        };

        $scope.createRepoChoices.product['organization_id'] = CurrentOrganization;

        angular.forEach($scope.table.rows, function (repo) {
            //Add a fake form to keep track of validations
            repo.form = {
                    messages: '',
                    $invalid: false
                };
            FormUtils.labelize(repo);
        });

        Product.queryUnpaged({'organization_id': CurrentOrganization, custom: true}, function (values) {
            $scope.products = values.results;

            if ($scope.products.length > 0) {
                $scope.createRepoChoices.existingProductId = $scope.products[0].id;
            } else {
                $scope.createRepoChoices.newProduct = "true";
            }
            $scope.page.loading = false;
        });

        $scope.creating = function () {
            return $scope.createRepoChoices.creating;
        };

        $scope.requiredFieldsEnabled = function () {
            var fieldsEnabled = true;
            if ($scope.createRepoChoices.newProduct === "true") {
                fieldsEnabled = $scope.productForm.$valid;
            } else if (angular.isUndefined($scope.createRepoChoices.existingProductId)) {
                fieldsEnabled = false;
            }

            if (fieldsEnabled) {
                return $scope.productForm.$valid;
            }

            return fieldsEnabled;
        };

        $scope.gpgKeys = [];
        GPGKey.queryUnpaged(function (gpgKeys) {
            $scope.gpgKeys = gpgKeys.results;
        }, function (response) {
            $scope.gpgKeys = [];
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        $scope.$watch('table.rows', function (newList, oldList) {
            if (newList) {
                angular.forEach(newList, function (newItem, position) {
                    if (angular.isUndefined(oldList) || newItem.name !== oldList[position].name) {
                        FormUtils.labelize(newItem);
                    }
                });
            }
        }, function (newList, oldList) {
            var isEqual = true;
            angular.forEach(newList, function (newItem, position) {
                if (newList[position].name !== oldList[position].name) {
                    isEqual = false;
                }
            });
            return isEqual;
        });

        $scope.clearInvalidRepos = function () {
            angular.forEach($scope.table.rows, function (repo) {
                repo.form.$invalid = false;
            });
        };

        $scope.createRepos = function () {
            $scope.createRepoChoices.creating = true;
            $scope.clearInvalidRepos();

            if ($scope.createRepoChoices.newProduct === "true") {
                Product.save($scope.createRepoChoices.product, productCreateSuccess, productCreateError);
            } else {
                createNextRepo();
            }
        };

        $scope.createStatusIcon = function (repo) {
            var icon;
            if (repo.created === true) {
                icon = 'pficon pficon-ok';
            } else if (repo.creating === true) {
                icon = 'fa fa-spinner fa-spin';
            } else if (angular.isUndefined(repo.messages)) {
                icon = '';
            } else {
                icon = 'pficon pficon-error-circle-o';
            }

            return icon;
        };

        $scope.createStatusMessages = function (repo) {
            var message;
            if (angular.isUndefined(repo.messages)) {
                message = translate("Not started");
            } else {
                message = repo.messages;
            }
            return message;
        };
    }]
);
(function () {

    /**
     * @ngdoc service
     * @name Bastion.products.service:DiscoveryRepositories
     *
     * @description
     *   Provides a helper service for discovery repositories.
     *
     */
    function DiscoveryRepositories() {
        this.rows = [];
        this.repositoryUrl = '';
        this.upstreamUsername = '';
        this.upstreamPassword = '';

        /**
         * Allows setting the rows.
         *
         * @param {Array} selected rows
         **/
        this.setRows = function (rows) {
            this.rows = rows;
        };

        /**
         * Return the list of rows.
         *
         * @returns {Array} array of rows
         */
        this.getRows = function () {
            return this.rows;
        };

        /**
         * Allows setting the repository URL.
         *
         * @param {Array} selected repositoryUrl
         **/
        this.setRepositoryUrl = function (repositoryUrl) {
            this.repositoryUrl = repositoryUrl;
        };

        /**
         * Return the list of repository URL.
         *
         * @returns {Array} array of repositoryUrl
         */
        this.getRepositoryUrl = function () {
            return this.repositoryUrl;
        };

        /**
         * Allows setting the upstream username.
         *
         * @param {Array} selected upstreamUsername
         **/
        this.setUpstreamUsername = function (upstreamUsername) {
            this.upstreamUsername = upstreamUsername;
        };

        /**
         * Return the list of upstream username.
         *
         * @returns {Array} array of upstreamUsername
         */
        this.getUpstreamUsername = function () {
            return this.upstreamUsername;
        };

        /**
         * Allows setting the upstream password.
         *
         * @param {Array} selected upstreamPassword
         **/
        this.setUpstreamPassword = function (upstreamPassword) {
            this.upstreamPassword = upstreamPassword;
        };

        /**
         * Return the list of upstream password.
         *
         * @returns {Array} array of upstreamPassword
         */
        this.getUpstreamPassword = function () {
            return this.upstreamPassword;
        };
    }

    angular.module('Bastion.products').service('DiscoveryRepositories', DiscoveryRepositories);

    DiscoveryRepositories.$inject = [];

})();
/**
 * @ngdoc object
 * @name  Bastion.products.controller:DiscoveryController
 *
 * @requires $scope
 * @requires $q
 * @requires $timeout
 * @requires $http
 * @requires Notification
 * @requires Task
 * @requires Organization
 * @requires CurrentOrganization
 * @requires DiscoveryRepositories
 * @requires translate
 *
 * @description
 *   Provides the functionality for the repo discovery action pane.
 */

angular.module('Bastion.products').controller('DiscoveryController',
    ['$scope', '$q', '$timeout', '$http', 'Notification', 'Task', 'Organization', 'CurrentOrganization', 'DiscoveryRepositories', 'translate',
    function ($scope, $q, $timeout, $http, Notification, Task, Organization, CurrentOrganization, DiscoveryRepositories, translate) {
        var transformRows, setDiscoveryDetails;

        $scope.discovery = {
            url: '',
            contentType: 'yum'
        };

        $scope.page = {loading: false};

        $scope.contentTypes = [
            {id: "yum", name: "Yum Repositories"},
            {id: "docker", name: "Docker Images"}
        ];

        if (!$scope.table) {
            $scope.table = {
                rows: [],
                resource: {
                    total: 0,
                    subtotal: 0
                },
                numSelected: 0
            };
        }

        setDiscoveryDetails = function (task) {
            $scope.table.rows = transformRows(task.output);
            $scope.table.resource.total = $scope.table.rows.length;
            $scope.table.resource.subtotal = $scope.table.resource.total;
        };

        $scope.setupSelected = function () {
            var url;

            if (!_.startsWith($scope.discovery.url, 'http')) {
                url = 'http://' + $scope.discovery.url;
            } else {
                url = $scope.discovery.url;
            }
            $scope.page.loading = true;
            $scope.discovery.selected = $scope.table.getSelected();

            DiscoveryRepositories.setRows($scope.table.getSelected());
            DiscoveryRepositories.setRepositoryUrl(url);
            DiscoveryRepositories.setUpstreamUsername($scope.discovery.upstreamUsername);
            DiscoveryRepositories.setUpstreamPassword($scope.discovery.upstreamPassword);

            $scope.transitionTo('product-discovery.create').then(function () {
                $scope.page.loading = false;
            });
        };

        $scope.defaultName = function (basePath) {
            //Remove leading/trailing slash and replace rest with space
            return basePath.replace(/^\//, "").replace(/\/$/, "").replace(/\//g, ' ');
        };

        $scope.cancelDiscovery = function () {
            $scope.discovery.working = false;
            Task.unregisterSearch($scope.taskSearchId);
            Organization.cancelRepoDiscover({id: CurrentOrganization});
        };

        transformRows = function (urls) {
            var baseUrl, toRet;
            baseUrl = $scope.discovery.url;

            toRet = _.map(urls, function (url) {
                var params;

                params = {
                    url: $scope.discovery.url,
                    label: '',
                    contentType: $scope.discovery.contentType
                };
                if ($scope.discovery.contentType === 'yum') {
                    params.path = url.replace(baseUrl, "");
                    params.name = $scope.defaultName(params.path);
                } else {
                    params.dockerUpstreamName = url;
                    params.path = url;
                    params.name = url;
                }
                return params;
            });

            return _.sortBy(toRet, function (item) {
                return item.url;
            });
        };

        $scope.updateTask = function (task) {
            if ($scope.discovery.working) {
                setDiscoveryDetails(task);
                if (task.state !== "running" && task.state !== "planned") {
                    $scope.discovery.working = false;
                    Task.unregisterSearch($scope.taskSearchId);
                    if (task.result === "error") {
                        Notification.setErrorMessage(translate("Discovery failed. Error: %s").replace('%s', task.humanized.errors[0]));
                    }
                }
            }
        };

        $scope.discover = function () {
            var params;

            $scope.discovery.working = true;
            $scope.table.rows = [];
            $scope.table.selectAll(false);
            params = {
                id: CurrentOrganization, url: $scope.discovery.url,
                'content_type': $scope.discovery.contentType,
                'upstream_username': $scope.discovery.upstreamUsername,
                'upstream_password': $scope.discovery.upstreamPassword
            };

            Organization.repoDiscover(params, function (task) {
                // Hide pagination
                $scope.table.resource['per_page'] = false;
                $scope.taskSearchId = Task.registerSearch({ 'type': 'task', 'task_id': task.id }, $scope.updateTask);
            });
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.products.controller:NewProductController
 *
 * @requires $scope
 * @requires Product
 *
 * @description
 *   Controls the creation of an empty Product object for use by sub-controllers.
 */

angular.module('Bastion.products').controller('NewProductController',
    ['$scope', 'Product', 'CurrentOrganization',
    function ($scope, Product, CurrentOrganization) {
        $scope.product = new Product({'organization_id': CurrentOrganization});
        $scope.page = {loading: true};
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.proudcts.controller:NewSyncPlanModalController
 *
 * @requires $scope
 * @requires $uibModalInstance
 * @requires SyncPlan
 * @requires SyncPlanHelper
 *
 * @description
 *   A controller for creating a new sync plan in a modal.
 */

angular.module('Bastion.products').controller('NewSyncPlanModalController',
    ['$scope', '$uibModalInstance', 'SyncPlan', 'SyncPlanHelper',
        function ($scope, $uibModalInstance, SyncPlan, SyncPlanHelper) {
            function success(syncPlan) {
                $uibModalInstance.close(syncPlan);
            }

            function error(response) {
                var form = SyncPlanHelper.getForm();

                angular.forEach(response.data.errors, function (errors, field) {
                    form[field].$setValidity('server', false);
                    form[field].$error.messages = errors;
                });
            }

            $scope.ok = function (syncPlan) {
                SyncPlanHelper.createSyncPlan(syncPlan, success, error);
            };

            $scope.cancel = function () {
                $uibModalInstance.dismiss('cancel');
            };

            $scope.isFormDisabled = function () {
                var form = SyncPlanHelper.getForm();
                return form && !form.$valid;
            };

            $scope.intervals = SyncPlanHelper.getIntervals();
            $scope.syncPlan = new SyncPlan();
            $scope.syncPlan.interval = $scope.intervals[0].id;
        }]
);
/**
 * @ngdoc object
 * @name  Bastion.products.controller:ProductFormController
 *
 * @requires $scope
 * @requires $q
 * @requires $uibModal
 * @requires Product
 * @requires GPGKey
 * @requires SyncPlan
 * @requires FormUtils
 *
 *
 * @description
 *   Provides the functionality specific to Products for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.products').controller('ProductFormController',
    ['$scope', '$q', '$uibModal', 'Product', 'GPGKey', 'SyncPlan', 'FormUtils', 'Notification',
    function ($scope, $q, $uibModal, Product, GPGKey, SyncPlan, FormUtils, Notification) {

        function fetchGpgKeys() {
            return GPGKey.queryUnpaged(function (gpgKeys) {
                $scope.gpgKeys = gpgKeys.results;
            });
        }

        function fetchSyncPlans() {
            return SyncPlan.queryUnpaged(function (syncPlans) {
                $scope.syncPlans = syncPlans.results;
            });
        }

        function success() {
            $scope.transitionTo('product.repositories', {productId: $scope.product.id});
        }

        function error(response) {
            $scope.working = false;
            angular.forEach(response.data.errors, function (errors, field) {
                if ( $scope.productForm[field]) {
                    $scope.productForm[field].$setValidity('server', false);
                    $scope.productForm[field].$error.messages = errors;
                } else {
                    Notification.setErrorMessage("An error occurred while saving the Product: " + field + " " + errors);
                }
            });
        }

        $scope.product = $scope.product || new Product();

        $scope.$watch('product.name', function () {
            if ($scope.productForm.name) {
                $scope.productForm.name.$setValidity('server', true);
                FormUtils.labelize($scope.product);
            }
        });

        $scope.save = function (product) {
            product.$save(success, error);
        };

        $q.all([fetchSyncPlans().$promise, fetchGpgKeys().$promise]).finally(function () {
            $scope.page.loading = false;
        });

        $scope.openSyncPlanModal = function () {
            $uibModal.open({
                templateUrl: 'products/new/views/new-sync-plan-modal.html',
                controller: 'NewSyncPlanModalController'
            }).result.then(function ($value) {
                fetchSyncPlans().$promise.then(function () {
                    $scope.product['sync_plan_id'] = $value.id;
                });
            });
        };
    }]
);
/**
 * @ngdoc service
 * @name  Bastion.products.factory:Product
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for product or list of products.
 */

angular.module('Bastion.products').factory('Product',
    ['BastionResource', function (BastionResource) {

        return BastionResource('katello/api/v2/products/:id/:action', {id: '@id'}, {
            update: { method: 'PUT'},
            sync: { method: 'POST', params: { action: 'sync' }},
            updateSyncPlan: { method: 'POST', params: { action: 'sync_plan' }},
            autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}}
        });

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.products.controller:ProductsController
 *
 * @requires $scope
 * @requires $state
 * @requires $sce
 * @requires $location
 * @requires $uibModal
 * @requires translate
 * @requires Nutupane
 * @requires Product
 * @requires ProductBulkAction
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality specific to Products for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.products').controller('ProductsController',
    ['$scope', '$state', '$sce', '$location', '$uibModal', 'translate', 'Nutupane', 'Product', 'ProductBulkAction', 'CurrentOrganization', 'Notification',
    function ($scope, $state, $sce, $location, $uibModal, translate, Nutupane, Product, ProductBulkAction, CurrentOrganization, Notification) {
        var nutupane, taskUrl, taskLink, getBulkParams, bulkError, params;

        getBulkParams = function () {
            return {
                ids: _.map($scope.table.getSelected(), 'id'),
                'organization_id': CurrentOrganization
            };
        };

        bulkError = function (response) {
            angular.forEach(response.data.errors, function(message) {
                Notification.setErrorMessage(translate("An error occurred: ") + message);
            });

            nutupane.refresh();
        };

        params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'sort_by': 'name',
            'sort_order': 'ASC',
            'enabled': true,
            'paged': true
        };

        nutupane = new Nutupane(Product, params);
        $scope.controllerName = 'katello_products';
        nutupane.masterOnly = true;

        $scope.table = nutupane.table;

        $scope.$on('productDelete', function (event, taskId) {
            var message;
            taskUrl = $scope.taskUrl(taskId);
            taskLink = $sce.trustAsHtml("<a href=" + taskUrl + ">here</a>");
            message = translate("Product delete operation has been initiated in the background. Click %s to monitor the progress.");
            Notification.setRenderedSuccessMessage(message.replace("%", taskLink));
        });

        $scope.unsetProductDeletionTaskId = function () {
            $scope.productDeletionTaskId = undefined;
        };

        $scope.mostImportantSyncState = function (product) {
            var state = 'none';
            if (product['sync_summary'].pending > 0) {
                state = 'pending';
            } else if (product['sync_summary'].error > 0) {
                state = 'error';
            } else if (product['sync_summary'].warning > 0) {
                state = 'warning';
            } else if (product['sync_summary'].success > 0) {
                state = 'success';
            }
            return state;
        };

        $scope.syncProducts = function () {
            var success;

            success = function (task) {
                var url = $state.href('task', {taskId: task.id}), message;

                taskLink = $sce.trustAsHtml("<a href=" + url + ">here</a>");
                message = translate("Product sync has been initiated in the background. " +
                    "Click %s to monitor the progress.");

                Notification.setRenderedSuccessMessage(message.replace('%s', taskLink));
            };

            ProductBulkAction.syncProducts(getBulkParams(), success, bulkError);
        };

        $scope.goToDiscoveries = function () {
            nutupane.table.rows = [];
            nutupane.table.resource.results = [];
            nutupane.table.resource.total = 0;
            nutupane.table.resource.subtotal = 0;
            $state.go("product-discovery.scan");
        };

        $scope.openSyncPlanModal = function () {
            nutupane.invalidate();
            $uibModal.open({
                templateUrl: 'products/bulk/views/products-bulk-sync-plan-modal.html',
                controller: 'ProductsBulkSyncPlanModalController',
                size: 'lg',
                resolve: {
                    bulkParams: function () {
                        return getBulkParams();
                    }
                }
            }).closed.then(function () {
                nutupane.refresh();
            });
        };

        $scope.openAdvancedSyncModal = function () {
            nutupane.invalidate();
            $uibModal.open({
                templateUrl: 'products/bulk/views/products-bulk-advanced-sync-modal.html',
                controller: 'ProductsBulkAdvancedSyncModalController',
                size: 'lg',
                resolve: {
                    bulkParams: function () {
                        return getBulkParams();
                    }
                }
            });
        };

        $scope.removeProducts = function () {
            var success;

            success = function (response) {
                angular.forEach(response.displayMessages.success, function(message) {
                    Notification.setSuccessMessage(message);
                });

                angular.forEach(response.displayMessages.error, function(message) {
                    Notification.setErrorMessage(message);
                });

                nutupane.refresh();
            };

            ProductBulkAction.removeProducts(getBulkParams(), success, bulkError);
        };
    }]
);
/**
 * @ngdoc object
 * @name Bastion.products.config
 *
 * @requires $stateProvider
 *
 * @description
 *   Used for systems level configuration such as setting up the ui state machine.
 */

angular.module('Bastion.products').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('products', {
        url: '/products',
        permission: 'view_products',
        views: {
            '@': {
                controller: 'ProductsController',
                templateUrl: 'products/views/products.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{ 'Products' | translate }}"
        }
    });

    $stateProvider.state('products.new', {
        abstract: true,
        template: '<div ui-view></div>',
        views: {
            '@': {
                controller: 'NewProductController',
                templateUrl: 'products/new/views/product-new.html'
            }
        }
    })
    .state('products.new.form', {
        url: '/new',
        permission: 'create_products',
        controller: 'ProductFormController',
        templateUrl: 'products/new/views/product-new-form.html',
        ncyBreadcrumb: {
            label: "{{ 'New Product' | translate }}"
        }
    });

    $stateProvider.state("product-discovery", {
        abstract: true,
        url: '/products/discovery',
        controller: 'DiscoveryController',
        templateUrl: 'products/discovery/views/discovery-base.html'
    })
    .state("product-discovery.scan", {
        url: '/scan',
        permission: 'edit_products',
        templateUrl: 'products/discovery/views/discovery.html',
        ncyBreadcrumb: {
            label: "{{ 'Discover Repositories' | translate }}",
            parent: 'products'
        }
    })
    .state("product-discovery.create", {
        url: '/scan/create',
        permission: 'edit_products',
        templateUrl: 'products/discovery/views/discovery-create.html',
        controller: 'DiscoveryCreateController',
        ncyBreadcrumb: {
            label: "{{ 'Create Repositories' | translate }}",
            parent: 'product-discovery.scan'
        }
    });

    $stateProvider.state("product", {
        abstract: true,
        url: '/products/:productId',
        permission: 'view_products',
        controller: 'ProductDetailsController',
        templateUrl: 'products/details/views/product-details.html'
    })
    .state('product.info', {
        url: '',
        permission: 'view_products',
        controller: 'ProductDetailsInfoController',
        templateUrl: 'products/details/views/product-info.html',
        ncyBreadcrumb: {
            label: "{{ product.name }}",
            parent: 'products'
        }
    });

    $stateProvider.state('product.tasks', {
        abstract: true,
        template: '<div ui-view></div>'
    })
    .state('product.tasks.index', {
        url: '/tasks',
        permission: 'view_products',
        templateUrl: 'products/details/views/product-tasks.html',
        ncyBreadcrumb: {
            label: "{{'Tasks' | translate }}",
            parent: 'product.info'
        }
    })
    .state('product.tasks.details', {
        url: '/tasks/:taskId',
        permission: 'view_products',
        controller: 'TaskDetailsController',
        templateUrl: 'tasks/views/task-details.html',
        ncyBreadcrumb: {
            label: "{{ task.id }}",
            parent: 'product.tasks.index'
        }
    });
}]);
/**
 * @ngdoc module
 * @name  Bastion.repository-sets
 *
 * @description
 *   Module for common repository sets functionality.
 */

angular.module('Bastion.repository-sets', ['Bastion', 'Bastion.i18n', 'Bastion.components']);
(function () {
    /**
     * @ngdoc run
     * @name Bastion.run:ContentOverrideHelper
     *
     * @description
     *  Provides common functionality for repository sets content overrides.
     */
    function ContentOverrideHelper() {
        var self = this;

        this.getContentOverrides = function (productRepositorySets, overrideName, overrideValue, remove) {
            var contentOverrides = [];

            angular.forEach(productRepositorySets, function (productRepositorySet) {
                var repositorySet = {
                    'content_label': productRepositorySet.content.label,
                    name: overrideName,
                    value: overrideValue
                };

                if (remove) {
                    repositorySet.remove = true;
                }

                contentOverrides.push(repositorySet);
            });

            return {'content_overrides': contentOverrides};
        };

        this.getEnabledContentOverrides = function (productRepositorySets) {
            return self.getContentOverrides(productRepositorySets, 'enabled', true);
        };

        this.getDisabledContentOverrides = function (productRepositorySets) {
            return self.getContentOverrides(productRepositorySets, 'enabled', false);
        };

        this.getDefaultContentOverrides = function (productRepositorySets) {
            return self.getContentOverrides(productRepositorySets, 'enabled', false, true);

        };
    }

    angular.module('Bastion.repository-sets').service('ContentOverrideHelper', ContentOverrideHelper);
})();
/**
 * @ngdoc factory
 * @name  Bastion.host-collections.factory:RepositorySet
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for host collections.
 */

angular.module('Bastion.repository-sets').factory('RepositorySet',
    ['BastionResource', function (BastionResource) {
        return BastionResource('katello/api/v2/repository_sets/:id/:action', {id: '@id'}, {
        });

    }]
);
(function () {
    /**
     * @ngdoc filter
     * @name  Bastion.repository-sets.filter:repositorySetsEnabled.filter.js
     *
     * @requires translate
     *
     * @description
     *  Makes the value of the repository set filter human readable.
     */

    function repositorySetsEnabled(translate) {
        return function (repositorySet) {
            var enabledText;

            if (repositorySet.enabled_content_override === true) {
                enabledText = translate('Enabled (overridden)');
            } else if (repositorySet.enabled_content_override === false) {
                enabledText = translate('Disabled (overridden)');
            } else {
                if (repositorySet.enabled) {
                    enabledText = translate('Enabled');
                } else {
                    enabledText = translate('Disabled');
                }
            }

            return enabledText;
        };
    }

    angular.module('Bastion.repository-sets').filter('repositorySetsEnabled', repositorySetsEnabled);
    repositorySetsEnabled.$inject = ['translate'];
})();
/**
 * @ngdoc module
 * @name  Bastion.sync-plans
 *
 * @description
 *   Module for sync plan related functionality.
 */

angular.module('Bastion.sync-plans', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.common',
    'Bastion.components.formatters',
    'Bastion.products'
]);
/**
 * @ngdoc object
 * @name  Bastion.sync-plans.controller:SyncPlanAddProductsController
 *
 * @requires $scope
 * @requires $q
 * @requires $location
 * @requires translate
 * @requires SyncPlan
 * @requires Product
 * @requires CurrentOrganization
 * @requires Nutupane
 * @requires Notification
 *
 * @description
 *   Provides the functionality for adding products to a sync plan.
 */

angular.module('Bastion.sync-plans').controller('SyncPlanAddProductsController',
    ['$scope', '$q', '$location', 'translate', 'SyncPlan', 'Product', 'CurrentOrganization', 'Nutupane', 'Notification',
        function ($scope, $q, $location, translate, SyncPlan, Product, CurrentOrganization, Nutupane, Notification) {
            var productsNutupane, params;

            params = {
                'search': $location.search().search || "",
                'sort_by': 'name',
                'sort_order': 'ASC',
                'full_result': true,
                'organization_id': CurrentOrganization,
                'sync_plan_id': $scope.$stateParams.syncPlanId,
                'available_for': 'sync_plan'
            };

            productsNutupane = new Nutupane(Product, params);
            $scope.controllerName = 'katello_products';
            $scope.table = productsNutupane.table;

            $scope.addProducts = function () {
                var data,
                    success,
                    error,
                    deferred = $q.defer(),
                    productsToAdd = _.map($scope.table.getSelected(), 'id');

                data = {
                    "product_ids": productsToAdd
                };

                success = function (response) {
                    Notification.setSuccessMessage(translate('Added %x products to sync plan "%y".')
                        .replace('%x', $scope.table.numSelected).replace('%y', $scope.syncPlan.name));
                    $scope.table.working = false;
                    $scope.table.selectAll(false);
                    productsNutupane.refresh();
                    $scope.syncPlan.$get();
                    deferred.resolve(response);
                };

                error = function (response) {
                    deferred.reject(response.data.errors);
                    Notification.setErrorMessage(response.data.errors.base);
                    $scope.table.working = false;
                };

                $scope.table.working = true;
                SyncPlan.addProducts({id: $scope.syncPlan.id}, data, success, error);
                return deferred.promise;
            };
        }]
);
/**
 * @ngdoc object
 * @name  Bastion.syncPlans.controller:SyncPlanDetailsInfoController
 *
 * @requires $scope
 * @requires $q
 * @requires translate
 * @requires SyncPlan
 * @requires MenuExpander
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the sync plan details action pane.
 */

angular.module('Bastion.sync-plans').controller('SyncPlanDetailsInfoController',
    ['$scope', '$q', 'translate', 'SyncPlan', 'MenuExpander', 'Notification',
        function ($scope, $q, translate, SyncPlan, MenuExpander, Notification) {
            $scope.intervals = [
                {id: 'hourly', value: translate('hourly')},
                {id: 'daily', value: translate('daily')},
                {id: 'weekly', value: translate('weekly')}
            ];

            $scope.menuExpander = MenuExpander;
            $scope.panel = $scope.panel || {loading: false};

            function updateSyncPlan(syncPlan) {
                syncPlan.syncDate = new Date(syncPlan['sync_date']);
                syncPlan.syncTime = new Date(syncPlan['sync_date']);
                $scope.syncPlan = syncPlan;
            }

            SyncPlan.get({id: $scope.$stateParams.syncPlanId}, function (syncPlan) {
                $scope.panel.loading = false;
                updateSyncPlan(syncPlan);
            });

            $scope.save = function (syncPlan) {
                var deferred = $q.defer(),
                    syncDate = new Date(syncPlan.syncDate),
                    syncTime = new Date(syncPlan.syncTime || new Date());

                syncDate.setHours(syncTime.getHours());
                syncDate.setMinutes(syncTime.getMinutes());
                syncDate.setSeconds(0);
                syncPlan['sync_date'] = syncDate.toString();

                syncPlan.$update(function (response) {
                    updateSyncPlan(syncPlan);
                    deferred.resolve(response);
                    Notification.setSuccessMessage(translate('Sync Plan Saved'));
                }, function (response) {
                    deferred.reject(response);
                    angular.forEach(response.data.errors, function (errorMessage, key) {
                        if (angular.isString(key)) {
                            errorMessage = [key, errorMessage].join(' ');
                        }
                        Notification.setErrorMessage(translate("An error occurred saving the Sync Plan: ") + errorMessage);
                    });
                });

                return deferred.promise;
            };
        }]
);
/**
 * @ngdoc object
 * @name  Bastion.sync-plan.controller:SyncPlanDetailsController
 *
 * @requires $scope
 * @requires $state
 * @requires translate
 * @requires SyncPlan
 * @requires ApiErrorHandler
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the sync plan details action pane.
 */

angular.module('Bastion.sync-plans').controller('SyncPlanDetailsController',
    ['$scope', '$state', 'translate', 'SyncPlan', 'ApiErrorHandler', 'Notification',
        function ($scope, $state, translate, SyncPlan, ApiErrorHandler, Notification) {
            $scope.panel = {
                error: false,
                loading: true
            };

            if ($scope.syncPlan) {
                $scope.panel.loading = false;
            }

            $scope.syncPlan = SyncPlan.get({id: $scope.$stateParams.syncPlanId}, function () {
                $scope.panel.loading = false;
            }, function (response) {
                $scope.panel.loading = false;
                ApiErrorHandler.handleGETRequestErrors(response, $scope);
            });

            $scope.runSyncPlan = function () {
                SyncPlan.sync({id: $scope.$stateParams.syncPlanId}, function (task) {
                    $scope.task = task;
                }, function (response) {
                    Notification.setErrorMessage(response.data.errors[0]);
                });
            };

            $scope.removeSyncPlan = function (syncPlan) {
                syncPlan.$remove(function () {
                    Notification.setSuccessMessage(translate('Sync Plan %s has been deleted.').replace('%s', syncPlan.name));
                    $scope.transitionTo('sync-plans');
                });
            };
        }
    ]
);
/**
 * @ngdoc object
 * @name  Bastion.sync-plans.controller:SyncPlanProductsController
 *
 * @requires $scope
 * @requires $q
 * @requires $location
 * @requires translate
 * @requires SyncPlan
 * @requires Product
 * @requires CurrentOrganization
 * @requires Nutupane
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the sync plan list products details action pane.
 */

angular.module('Bastion.sync-plans').controller('SyncPlanProductsController',
    ['$scope', '$q', '$location', 'translate', 'SyncPlan', 'Product', 'CurrentOrganization', 'Nutupane', 'Notification',
        function ($scope, $q, $location, translate, SyncPlan, Product, CurrentOrganization, Nutupane, Notification) {
            var productsNutupane, params;

            $scope.table = {};

            params = {
                'search': $location.search().search || "",
                'sort_by': 'name',
                'sort_order': 'ASC',
                'full_result': true,
                'organization_id': CurrentOrganization,
                'sync_plan_id': $scope.$stateParams.syncPlanId
            };

            productsNutupane = new Nutupane(Product, params);
            $scope.controllerName = 'katello_products';
            $scope.table = productsNutupane.table;

            $scope.removeProducts = function () {
                var data,
                    success,
                    error,
                    deferred = $q.defer(),
                    productsToRemove = _.map($scope.table.getSelected(), 'id');

                data = {
                    "product_ids": productsToRemove
                };

                success = function (response) {
                    Notification.setSuccessMessage(translate('Removed %x products from sync plan "%y".')
                        .replace('%x', $scope.table.numSelected).replace('%y', $scope.syncPlan.name));
                    $scope.table.working = false;
                    $scope.table.selectAll(false);
                    productsNutupane.refresh();
                    $scope.syncPlan.$get();
                    deferred.resolve(response);
                };

                error = function (response) {
                    deferred.reject(response.data.errors);
                    angular.forEach(response.data.errors, function (responseError) {
                        Notification.setErrorMessage(responseError);
                    });
                    $scope.table.working = false;
                };

                $scope.table.working = true;
                SyncPlan.removeProducts({id: $scope.syncPlan.id}, data, success, error);
                return deferred.promise;
            };
        }]
);
/**
 * @ngdoc object
 * @name  Bastion.sync-plans.controller:NewSyncPlanController
 *
 * @requires $scope
 * @requires translate
 * @requires SyncPlan
 * @requires SyncPlanHelper
 * @requires Notification
 *
 * @description
 *   Controls the creation of an empty SyncPlan object for use by sub-controllers.
 */

angular.module('Bastion.sync-plans').controller('NewSyncPlanController',
    ['$scope', '$rootScope', 'translate', 'SyncPlan', 'SyncPlanHelper', 'Notification',
        function ($scope, $rootScope, translate, SyncPlan, SyncPlanHelper, Notification) {
            $scope.intervals = SyncPlanHelper.getIntervals();

            $scope.syncPlan = new SyncPlan();
            $scope.syncPlan.interval = $scope.intervals[0].id;
            $scope.syncPlan.startDate = new Date();

            function success(syncPlan) {
                $scope.working = false;
                $scope.$state.go('sync-plan.info', {syncPlanId: syncPlan.id});
                Notification.setSuccessMessage(translate('New sync plan successfully created.'));
            }

            function error(response) {
                var form = SyncPlanHelper.getForm();

                angular.forEach(response.data.errors, function (errors, field) {
                    form[field].$setValidity('server', false);
                    form[field].$error.messages = errors;
                });

                $scope.working = false;
            }

            $scope.createSyncPlan = function (syncPlan) {
                SyncPlanHelper.createSyncPlan(syncPlan, success, error);
            };

            $scope.setForm = function (form) {
                SyncPlanHelper.setForm(form);
            };
        }]
);
(function () {

    /**
     * @ngdoc service
     * @name Bastion.sync-plans.service:SyncPlanHelper
     *
     * @description
     *   Provides a helper service for Sync Plans
     *
     */
    function SyncPlanHelper(translate) {

        /**
         * Get the set form.  Allows the sharing of form data across controllers
         *
         * @param form the form to set
         */
        this.setForm = function (form) {
            this.form = form;
        };

        /**
         * Set the form.  Allows the sharing of form data across controllers
         *
         * @returns {form|*} the stored form
         */
        this.getForm = function () {
            return this.form;
        };

        /**
         * Returns the valid sync plan intervals.
         *
         * @returns {{id: string, value: *}[]}
         */
        this.getIntervals = function () {
            return [
                {id: 'hourly', value: translate('hourly')},
                {id: 'daily', value: translate('daily')},
                {id: 'weekly', value: translate('weekly')}
            ];

        };

        /**
         * Create a sync plan, including setting the dates correctly.
         *
         * @returns $resource sync plan
         */
        this.createSyncPlan = function (syncPlan, success, error) {
            var GMT_OFFSET_MILLISECONDS = syncPlan.startDate.getTimezoneOffset() * 60000,
                syncDate = new Date(syncPlan.startDate.getTime() + GMT_OFFSET_MILLISECONDS),
                syncTime = new Date(syncPlan.startTime || new Date());
            syncDate.setHours(syncTime.getHours());
            syncDate.setMinutes(syncTime.getMinutes());
            syncDate.setSeconds(0);

            syncPlan['sync_date'] = syncDate.toString();
            syncPlan.$save(success, error);
            return syncPlan;
        };
    }

    angular.module('Bastion.sync-plans').service('SyncPlanHelper', SyncPlanHelper);
    SyncPlanHelper.$inject = ['translate'];
})();
/**
 * @ngdoc factory
 * @name  Bastion.sync-plans.factory:SyncPlan
 *
 * @requires BastionResource
 * @requires CurrentOrganization
 *
 * @description
 *   Provides a BastionResource for product or list of repositories.
 */

angular.module('Bastion.sync-plans').factory('SyncPlan',
    ['BastionResource', 'CurrentOrganization',
    function (BastionResource, CurrentOrganization) {

        return BastionResource('katello/api/v2/organizations/:organizationId/sync_plans/:id/:action',
            {id: '@id', organizationId: CurrentOrganization}, {
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}},
                update: { method: 'PUT' },
                sync: { method: 'PUT', params: {action: 'sync'}},
                addProducts: {method: 'PUT', params: {action: 'add_products'}},
                removeProducts: {method: 'PUT', params: {action: 'remove_products'}}
            }
        );

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.syncPlans.controller:SyncPlansController
 *
 * @requires $scope
 * @requires $location
 * @requires translate
 * @requires Nutupane
 * @requires SyncPlan
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality specific to Sync Plans for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.sync-plans').controller('SyncPlansController',
    ['$scope', '$location', 'translate', 'Nutupane', 'SyncPlan', 'CurrentOrganization',
        function ($scope, $location, translate, Nutupane, SyncPlan, CurrentOrganization) {
            var params, nutupane;

            params = {
                'organization_id': CurrentOrganization,
                'search': $location.search().search || "",
                'sort_by': 'name',
                'sort_order': 'ASC'
            };

            nutupane = new Nutupane(SyncPlan, params);
            $scope.controllerName = 'katello_sync_plans';
            nutupane.masterOnly = true;

            $scope.syncPlanTable = nutupane.table;
            $scope.removeRow = nutupane.removeRow;
            $scope.nutupane = nutupane;

            nutupane.enableSelectAllResults();

            if ($location.search()['select_all']) {
                nutupane.table.selectAllResults(true);
            }

            $scope.table = $scope.syncPlanTable;
        }]
);
/**
 * @ngdoc object
 * @name Bastion.sync-plans.config
 *
 * @requires $stateProvider
 *
 * @description
 *   Used for sync plan level configuration such as setting up the ui state machine.
 */

angular.module('Bastion.sync-plans').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('sync-plans', {
        url: '/sync_plans',
        permission: 'view_sync_plans',
        template: '<div ui-view></div>',
        views: {
            '@': {
                controller: 'SyncPlansController',
                templateUrl: 'sync-plans/views/sync-plans.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{ 'Sync Plans' | translate }}"
        }
    });

    $stateProvider.state('sync-plans.new', {
        url: '/sync_plans/new',
        permission: 'create_sync_plans',
        views: {
            '@': {
                controller: 'NewSyncPlanController',
                templateUrl: 'sync-plans/new/views/new-sync-plan.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{ 'New Sync Plan' | translate }}"
        }
    });

    $stateProvider.state("sync-plan", {
        abstract: true,
        url: '/sync_plans/:syncPlanId',
        permission: 'view_sync_plans',
        controller: 'SyncPlanDetailsController',
        templateUrl: 'sync-plans/details/views/sync-plan-details.html'
    })
    .state('sync-plan.info', {
        url: '',
        permission: 'view_sync_plans',
        controller: 'SyncPlanDetailsInfoController',
        templateUrl: 'sync-plans/details/views/sync-plan-info.html',
        ncyBreadcrumb: {
            label: "{{ syncPlan.name }}",
            parent: 'sync-plans'
        }
    })

    .state('sync-plan.products', {
        abstract: true,
        url: '/products',
        permission: 'view_sync_plans',
        template: '<div ui-view></div>'
    })
    .state('sync-plan.products.list', {
        url: '',
        permission: 'view_sync_plans',
        controller: 'SyncPlanProductsController',
        templateUrl: 'sync-plans/details/views/sync-plan-products.html',
        ncyBreadcrumb: {
            label: "{{ 'List Products' | translate }}",
            parent: 'sync-plan.info'
        }
    })
    .state('sync-plan.products.add', {
        url: '/add',
        permission: 'edit_sync_plans',
        controller: 'SyncPlanAddProductsController',
        templateUrl: 'sync-plans/details/views/sync-plan-products.html',
        ncyBreadcrumb: {
            label: "{{ 'Add Products' | translate }}",
            parent: 'sync-plan.info'
        }
    });
}]);
/**
 * @ngdoc module
 * @name  Bastion.tasks
 *
 * @description
 *   Module for task related functionality.
 */

angular.module('Bastion.tasks', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.components',
    'Bastion.i18n'
]);

/**
 * @ngdoc object
 * @name Bastion.tasks.config
 *
 * @requires $stateProvider
 *
 * @description
 *   Set up the states for tasks.
 */
angular.module('Bastion.tasks').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('tasks', {
        abstract: true,
        templateUrl: 'tasks/views/tasks.html'
    })
    .state('tasks.index', {
        url: 'katello_tasks',
        permission: 'view_tasks',
        templateUrl: 'tasks/views/tasks-index.html'
    })
    .state('tasks.details', {
        url: 'katello_tasks/:taskId',
        permission: 'view_tasks',
        collapsed: true,
        controller: 'TaskDetailsController',
        templateUrl: 'tasks/views/task-details-standalone.html'
    })
    .state('task', {
        url: 'katello_tasks/single/:taskId',
        controller: 'TaskDetailsController',
        permission: 'view_tasks',
        templateUrl: 'tasks/views/task-details-standalone.html'
    });

}]);
/**
* @ngdoc service
* @name  Bastion.tasks.factory:AggregateTask
*
* @requires Task
 *
* @description
*   Provides a BastionResource for aggregating multiple tasks into a task like interface.
*
*/


angular.module('Bastion.tasks').factory('AggregateTask',
    ['Task', function (Task) {

        /**
          * @param {Array} taskIds ids of tasks to be polled for
          *
          * @param {Function} callback function to reflect the
          *        changes after the task was updated.
          *        The function is called with the updated aggregated task
          */
        var newAggregate = function (taskIds, callback) {
            var taskMap = {},
                taskSearches = {},
                taskRepresentation = {
                    state: undefined,
                    result: undefined,
                    progressbar: {}
                };

            var unregisterSearch = function (taskId) {
                if (taskSearches[taskId]) {
                    Task.unregisterSearch(taskSearches[taskId]);
                }
            },
            greatestType = function () {
                var found = 'success',
                    weights = {
                        error: 3,
                        danger: 2,
                        success: 1
                    };

                _.each(taskMap, function (task) {
                    if (weights[task.progressbar.type] > weights[found]) {

                        found = task.progressbar.type;
                    }
                });
                return found;
            },
            greatestState = function () {
                var found = 'stopped',
                    weights = {
                        running: 3,
                        pending: 2,
                        stopped: 1
                    };
                _.each(taskMap, function (task) {
                    if (weights[task.state] > weights[found]) {
                        found = task.state;
                    }
                });
                return found;
            },
            greatestResult = function () {
                var found = 'success',
                    weights = {
                        error: 4,
                        warning: 3,
                        pending: 2,
                        success: 1
                    };
                _.each(taskMap, function (task) {
                    if (weights[task.result] > weights[found]) {
                        found = task.result;
                    }
                });
                return found;
            },
            updateProgress = function () {
                var total = 0;
                _.each(taskMap, function (task) {
                    total = total + task.progressbar.value;
                });
                taskRepresentation.progressbar.value = total / _.size(taskMap);
                taskRepresentation.progressbar.type = greatestType();
                taskRepresentation.state = greatestState();
                taskRepresentation.result = greatestResult();
            },
            updateTask = function (task) {
                taskMap[task.id] = task;
                if (!task.pending) {
                    unregisterSearch(task.id);
                }
                updateProgress();
                if (callback) {
                    callback(task);
                }
            },
            unregisterAll = function () {
                _.each(taskSearches, function (searchId, taskId) {
                    unregisterSearch(taskId);
                });
            };

            taskRepresentation.unregisterAll = unregisterAll;
            taskRepresentation.unregisterSearch = unregisterSearch;


            _.each(taskIds, function (taskId) {
                if (angular.isUndefined(taskSearches[taskId])) {
                    taskSearches[taskId] = Task.registerSearch({ 'type': 'task', 'task_id': taskId }, updateTask);
                }
            });

            return taskRepresentation;
        };
        return {new: newAggregate};
    }]);
/**
 * @ngdoc object
 * @name  Bastion.tasks.controller:TaskDetailsController
 *
 * @requires $scope
 * @requires $rootScope
 * @requires Task
 *
 * @description
 *   Provides the functionality for the details of a task.
 */

angular.module('Bastion.tasks').controller('TaskDetailsController',
    ['$scope', 'Task',
    function ($scope, Task) {
        var taskId;

        taskId = $scope.$stateParams.taskId;

        $scope.unregisterSearch = function () {
            Task.unregisterSearch($scope.searchId);
            $scope.searchId = undefined;
        };

        $scope.updateTask = function (task) {
            $scope.task = task;
            if (!$scope.task.pending) {
                $scope.unregisterSearch();
            }
        };

        $scope.isArray = _.isArray;

        $scope.$on('$destroy', function () {
            $scope.unregisterSearch();
        });

        $scope.searchId = Task.registerSearch({ 'type': 'task', 'task_id': taskId }, $scope.updateTask);
    }
]);
/**
 * @ngdoc filter
 * @name  Bastion.tasks.filter:taskInputCompile
 *
 * @description
 *   Converts the task humanized structure into a flat string
 */

angular.module('Bastion.tasks')
    .filter('taskInputCompile', function () {
        return function (humanizedTaskInput) {
            var parts;

            if (!_.isArray(humanizedTaskInput) && !_.isObject(humanizedTaskInput)) {
                return humanizedTaskInput;
            }

            parts = _.map(humanizedTaskInput, function (part) {
                if (part.length === 2) {
                    return part[1].text;
                }

                return part;
            });

            return parts.join('; ');
        };
    });
/**
 * @ngdoc directive
 * @name  Bastion.tasks.directive:taskInputPart
 *
 * @description
 *   Converts part of task humanized structure into a link if possible
 */

angular.module('Bastion.tasks').directive('taskInputPart',
    function () {
        return {
            restrict: 'A',
            template: '<span ng-if="!link()">{{text()}}</span>' +
                      '<a ng-if="link()" href="{{link()}}">{{text()}}</a>',
            scope: {
                data: '='
            },
            link: function (scope) {
                scope.text = function () {
                    if (_.isString(scope.data)) {
                        return scope.data;
                    }

                    return scope.data[1].text;
                };

                scope.link = function () {
                    if (!_.isString(scope.data) && scope.data[1]) {
                        return scope.data[1].link;
                    }
                };
            }
        };
    }
);
/**
 * @ngdoc filter
 * @name  Bastion.tasks.filter:taskInputReduce
 *
 * @description
 *   Omits the parts of task humanized input that are not necessary to
 *   show (such as repository name in tasks list or a repository)
 */

angular.module('Bastion.tasks')
    .filter('taskInputReduce', function () {
        return function (humanizedTaskInput, skippedParts) {
            if (_.isString(humanizedTaskInput) || !skippedParts) {
                return humanizedTaskInput;
            }
            if (_.isString(skippedParts)) {
                skippedParts = skippedParts.split(',');
            }
            return _.reject(humanizedTaskInput, function (part) {
                if (part.length === 2) {
                    return _.includes(skippedParts, part[0]);
                }

                return false;
            });
        };
    });
/**
 * @ngdoc filter
 * @name  Bastion.tasks.filter:taskInputCompile
 *
 * @description
 *   Produces the shortest possible version of humanized input of a
 *   task, used for list of recent user tasks.
 */


angular.module('Bastion.tasks')
    .filter('taskInputShort', function () {
        return function (humanizedTaskInput) {
            if (_.isString(humanizedTaskInput)) {
                return humanizedTaskInput;
            }
            return _.first(humanizedTaskInput, 1);
        };
    });
 /**
 * @ngdoc service
 * @name  Bastion.tasks.factory:Task
 *
 * @requires BastionResource
 * @requires $timeout
 * @requires $log
 * @requires CurrentOrganization
 *
 * @description
 *   Provides a BastionResource for task(s).
 *
 *   Also provides a polling bulk search for tasks, define in
 *   foreman-tasks engine API. This allows polling for multiple task
 *   statuses with a single request (e.g. currently running tasks of a
 *   user, recent tasks of a user, tasks for given resource, detail of
 *   one task etc). See `resource.registerSearch` and
 *   `resource.unregisterSearch` for details.
 */


angular.module('Bastion.tasks').factory('Task',
    ['BastionResource', '$timeout', '$log', '$q', 'CurrentOrganization',
    function (BastionResource, $timeout, $log, $q, CurrentOrganization) {
        var bulkSearchRunning = false, searchIdGenerator = 0,
            searchParamsById = {}, callbackById = {}, pollCount = 0, maxPollInterval = 10000;

        var resource = BastionResource('katello/api/v2/tasks/:id/:action',
            {id: '@id', 'organization_id': CurrentOrganization}, {});

        var foremanTasksResource = BastionResource('foreman_tasks/api/tasks/:id/:action',
            {},
            {
                bulkSearch: {method: 'POST', isArray: true, params: { action: 'bulk_search'}}
            }
        );

        function bulkSearchParams() {
            var searches = [];
            _.each(searchParamsById, function (searchParams, id) {
                searchParams['search_id'] = id;
                searches.push(searchParams);
            });
            return { searches: searches };
        }

        function taskProgressbar(task) {
            var mapping = { 'error': 'danger', 'warning': 'warning', 'default': 'success' };
            var type = mapping[task.result] || mapping.default;
            return { value: task.progress * 100, type: type };
        }

        // add additional fields to the task model that help with
        // usage in the UI templates
        function normalizeTask(task) {
            task.progressbar = taskProgressbar(task);
            if (!task.humanized.errors) {
                task.humanized.errors = [];
            } else if (!_.isArray(task.humanized.errors)) {
                task.humanized.errors = [task.humanized.errors];
            }
        }

        function updateProgress(periodic) {
            if (_.keys(searchParamsById).length === 0) {
                bulkSearchRunning = false;
                return;
            }
            foremanTasksResource.bulkSearch(bulkSearchParams(), function (response) {
                try {
                    _.each(response, function (tasksSearch) {
                        var searchId = tasksSearch['search_params']['search_id'];
                        var callback = callbackById[searchId];
                        if (!callback) {
                            return;
                        }
                        try {
                            _.each(tasksSearch.results, function (task) {
                                normalizeTask(task);
                            });
                            if (tasksSearch['search_params'].type === 'task') {
                                callback(tasksSearch.results[0]);
                            } else {
                                callback(tasksSearch.results);
                            }
                        } catch (e) {
                            $log.error(e);
                        }
                    });
                } finally {
                    // schedule the next update
                    if (periodic) {
                        schedulePoll();
                    }
                }
            }, function () {
                if (periodic) {
                    schedulePoll();
                }
            });
        }

        /*eslint no-use-before-define:0*/
        function schedulePoll() {
            var pollTime;
            pollCount = pollCount + 1;
            pollTime = 1000 * pollCount;
            if (pollTime > maxPollInterval) {
                pollTime = maxPollInterval;
            }

            $timeout(function () {
                updateProgress(true);
            }, pollTime);
        }

        function ensureBulkSearchRunning() {
            if (!bulkSearchRunning) {
                bulkSearchRunning = true;
                updateProgress(true);
            }
        }

        function generateSearchId() {
            searchIdGenerator += 1;
            return searchIdGenerator;
        }

        function addSearch(searchParams, callback) {
            var searchId = generateSearchId();
            searchParamsById[searchId] = searchParams;
            callbackById[searchId] = callback;
            ensureBulkSearchRunning();
            return searchId;
        }

        function deleteSearch(searchId) {
            delete callbackById[searchId];
            delete searchParamsById[searchId];
        }

        resource.poll = function (task, returnFunction) {
            resource.get({id: task.id}, function (data) {
                if (data.pending) {
                    $timeout(function () {
                        resource.poll(data, returnFunction);
                    }, 8000);
                } else {
                    returnFunction(data);
                }
            }, function () {
                returnFunction({'human_readable_result': "Failed to fetch task", failed: true});
            });
        };

        /**
          * Registers a search for polling. The polling is
          * is performed using bulkSearch for all the registered searches at once
          * to avoid overloading the server with multiple requests
          * (since it is performed periodically)
          *
          * @param {Object} searchParams object specifying the params
          *        of the search.
          *
          * @param {Function} callback function to reflect the
          *        results. If searchParams.type === 'task', the
          *        function is called with a single task, otherwise
          *        it's passed with array of tasks sattisfying the
          *        conditions.
          *
          * @return {Number} the autogenerated id of the condition
          *        that can be used for unregistering the search later on
          */
        resource.registerSearch = function (searchParams, callback) {
            return addSearch(searchParams, callback);
        };


        /**
          * Unregisters the search from polling.
          *
          * @param {Number} id the value returned by the registerSearch
          */
        resource.unregisterSearch = function (id) {
            deleteSearch(id);
        };

        /**
          * monitors a single task, and runs corresponding callbacks
          * on given events.
          *
          * @param {Task} task - a task object to monitor. The task
          *    object can be either retrieved from polling or it might
          *    be a promised object from triggering a task.
          *
          * @returns {RunningTask} an object that represents the state
          *   of the running task with the following fields
          */
        resource.monitorTask = function (task) {
            var searchId, deferred = $q.defer();
            var runningTask = {
                // promise that the task finishes, rejected when some
                // error occurs
                promise: deferred.promise,
                // task object with details about the task
                task: task,
                // state of the running tasks, one of 'starting',
                // 'running', 'paused', 'stopped'
                state: null,
                // stops polling for the task updates
                stopMonitoring: function () {
                    if (searchId) {
                        resource.unregisterSearch(searchId);
                    }
                }
            };

            function updateTask(taskObject) {
                runningTask.task = taskObject;
                if (taskObject.state === 'paused' || taskObject.state === 'stopped') {
                    resource.unregisterSearch(searchId);
                    if (taskObject.result === 'success') {
                        runningTask.state = 'stopped';
                        deferred.resolve(taskObject);
                    }
                    if ((taskObject.result === 'error' || taskObject.result === 'warning')) {
                        runningTask.state = 'paused';
                        deferred.reject(taskObject.humanized.errors);
                    }
                } else {
                    runningTask.state = 'running';
                    deferred.notify(taskObject);
                }
            }

            function pollTask(taskObject) {
                searchId = resource.registerSearch({ 'type': 'task', 'task_id': taskObject.id }, updateTask);
            }

            if (task.$promise) {
                runningTask.state = 'starting';
                task.$promise.then(pollTask, function (error) {
                    var errors = [];
                    runningTask.state = 'stopped';
                    if (error.data.errors) {
                        errors = error.data.errors;
                    } else if (error.data.error) {
                        errors = error.data.error.message;
                    }
                    if (!_.isArray(errors)) {
                        errors = [errors];
                    }
                    deferred.reject(errors);
                });
            } else {
                updateTask(task);
                pollTask(task);
            }

            return runningTask;
        };

        return resource;
    }]
);
/**
 * @ngdoc factory
 * @name  Bastion.tasks.factory:TaskNutupane
 *
 * @requires Task
 * @requires Nutupane
 *
 * @description
 *   Defines TasksNutupane factory that uses Task.registerSearch as
 *   source of data instead of the standard `query` method. This
 *   effectively creates Nutupane table that is updated by polling.
 */

angular.module('Bastion.tasks').factory('TasksNutupane',
    ['Task', 'Nutupane', function (Task, Nutupane) {
        var TasksNutupane = function () {
            var self = this;

            Nutupane.call(self, Task, {});

            self.existingTasks = {};

            self.registerSearch = function (params) {
                if (!self.searchId) {
                    self.searchId = Task.registerSearch(params, self.updateTasks);
                }
            };

            self.unregisterSearch = function () {
                if (self.searchId) {
                    Task.unregisterSearch(self.searchId);
                    self.searchId = undefined;
                }
            };

            self.load = function () {
                self.table.working = true;
            };

            self.updateTasks = function (tasks) {
                self.refreshTasks(tasks);
                self.deleteOldRows(tasks);
                self.table.working = false;
                self.table.refreshing = false;
            };

            self.table.search = function () {
                self.table.resource.offset = 0;
                self.table.rows = [];
                if (!self.table.working) {
                    self.load();
                }
            };

            // Updates values for existing tasks and adds new rows
            self.refreshTasks = function (tasks) {
                // we reverse because we add new items to the top of
                // the table
                _.each(tasks.reverse(), function (task) {
                    var existingTask = self.existingTasks[task.id];
                    if (existingTask) {
                        _.each(task, function (value, key) {
                            existingTask[key] = value;
                        });
                    } else {
                        self.table.rows.unshift(task);
                        self.existingTasks[task.id] = task;
                    }
                });
            };

            // Removes rows that are no longer valid for the table
            self.deleteOldRows = function (tasks) {
                var newTaskIds = _.map(tasks, function (task) {
                        return task.id;
                    }),
                    oldTaskIds = _.keys(self.existingTasks),
                    taskIdsToDelete = _.difference(oldTaskIds, newTaskIds),
                    rowsToDelete = [];

                _.each(taskIdsToDelete, function (id) {
                    delete self.existingTasks[id];
                });

                _.each(self.table.rows, function (row, i) {
                    if (_.includes(taskIdsToDelete, row.id)) {
                        rowsToDelete.push(i);
                    }
                });

                _.each(rowsToDelete.reverse(), function (i) {
                    self.table.rows.splice(i, 1);
                });
            };
        };
        return TasksNutupane;
    }]);
/**
 * @ngdoc directive
 * @name  Bastion.tasks.directive:tasksTable
 *
 * @requires TasksNutupane
 *
 * @description
 *   Directive to show list of tasks for given resource/user using TasksNutupane.
 *   The basic conditions used for searching the tasks are provided in
 *   the attributes.
 *
 * @param {string} resourceType The type of resource that the tasks
 *   should be shown for, e.g. 'Katello::Repository'
 * @param {string} resourceId The id of resource that the tasks should
 *   be shown for (in case resourceType is specified)
 * @param {string} userId The id of user that the tasks should
 *   be shown for
 * @param {boolean} activeOnly Reduce the list of tasks only for active
 * @param {boolean} all Don't apply any implicit search params
 * @param {string} detailsState What state should be transitioned to
 *   for showing details of the task
 * @param {string} knownContext What parts of humanized task inputs
 *   can be skipped because are obvious from the context the table is in
 * @example
 *   <pre>
        <div tasks-table  details-state="product.tasks.details"
                          known-context="product,organization"
                          resource-type="Katello::Product"
                          resource-id="{{ product.id }}"/>
     </pre>
 */

angular.module('Bastion.tasks').directive('tasksTable',
    ['TasksNutupane',
    function (TasksNutupane) {
        return {
            restrict: 'A',
            templateUrl: function (element, attrs) {
                if (attrs.templateUrl) {
                    return attrs.templateUrl;
                }

                return 'tasks/views/tasks-table.html';
            },
            scope: {
                resourceId: '@',
                resourceType: '@',
                userId: '@',
                activeOnly: '@',
                all: '@',
                detailsState: '@',
                knownContext: '@'
            },
            controller: ['$scope', '$state', function ($scope, $state) {
                // we need to set the table before the template
                // is compiled. Therefore we're doing that in the
                // controller
                $scope.tasksNutupane = new TasksNutupane();
                $scope.table = $scope.tasksNutupane.table;

                // to be able to navigate to task details from the table
                $scope.tasksNutupane.table.gotoDetails = function (taskId) {
                    $state.go($scope.detailsState, { taskId: taskId });
                };
            }],
            link: function (scope, element) {
                scope.$watch('resourceId', function (resourceId) {
                    if (resourceId) {
                        scope.tasksNutupane.registerSearch({ 'type': 'resource',
                                                             'active_only': scope.activeOnly,
                                                             'resource_type': scope.resourceType,
                                                             'resource_id': resourceId });
                    }
                });
                scope.$watch('userId', function (userId) {
                    if (userId) {
                        scope.tasksNutupane.registerSearch({ 'type': 'user',
                                                             'active_only': scope.activeOnly,
                                                             'user_id': userId });
                    }
                });

                scope.$watch('all', function () {
                    scope.tasksNutupane.registerSearch({ 'type': 'all',
                                                         'active_only': scope.activeOnly });
                });

                element.bind('$destroy', function () {
                    scope.tasksNutupane.unregisterSearch();
                });
            }
        };
    }
]);
/**
 * @ngdoc module
 * @name  Bastion.gpg-keys
 *
 * @description
 *   Module for GPG key related functionality.
 */

angular.module('Bastion.gpg-keys', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.common',
    'Bastion.components',
    'ngUpload'
]);
/**
 * @ngdoc object
 * @name  Bastion.gpgKeys.controller:GPGKeyDetailsInfoController
 *
 * @requires $scope
 * @requires GPGKey
 * @requires translate
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the gpgKey details action pane.
 */

angular.module('Bastion.gpg-keys').controller('GPGKeyDetailsInfoController',
    ['$scope', 'GPGKey', 'translate', 'Notification', function ($scope, GPGKey, translate, Notification) {

        $scope.panel = $scope.panel || {loading: false};
        $scope.progress = {uploading: false};

        $scope.gpgKey = $scope.gpgKey || GPGKey.get({id: $scope.$stateParams.gpgKeyId}, function () {
            $scope.panel.loading = false;
        });

        $scope.gpgKey.$promise.then(function () {
            $scope.uploadURL = 'katello/api/v2/gpg_keys/' + $scope.gpgKey.id + '/content';
        });

        $scope.uploadContent = function (content) {
            if (content && (content !== "Please wait...")) {
                if (content.status === 'success') {
                    Notification.setSuccessMessage(translate('GPG Key successfully uploaded'));
                    $scope.uploadStatus = 'success';
                    $scope.gpgKey.$get();
                } else {
                    Notification.setErrorMessage(content.displayMessage);
                    $scope.uploadStatus = 'error';
                }

                $scope.progress.uploading = false;
            }
        };

        $scope.uploadError = function (error, content) {
            if (angular.isString(content) && content.indexOf("Request Entity Too Large")) {
                error = translate('File too large.');
            } else {
                error = content;
            }
            Notification.setErrorMessage(translate('Error during upload: ') + error);
            $scope.uploadStatus = 'error';
            $scope.progress.uploading = false;
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.gpg-keys.controller:GPGKeyDetailsController
 *
 * @requires $scope
 * @requires GPGKey
 * @requires $q
 * @requires translate
 * @requires ApiErrorHandler
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the gpgKey details action pane.
 */

angular.module('Bastion.gpg-keys').controller('GPGKeyDetailsController',
    ['$scope', 'GPGKey', '$q', 'translate', 'ApiErrorHandler', 'Notification', function ($scope, GPGKey, $q, translate, ApiErrorHandler, Notification) {
        $scope.panel = $scope.panel || {error: false, loading: false};

        $scope.gpgKey = GPGKey.get({id: $scope.$stateParams.gpgKeyId}, function () {
            $scope.panel.error = false;
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        $scope.save = function (gpgKey) {
            var deferred = $q.defer();

            gpgKey.$update(function (response) {
                deferred.resolve(response);
                Notification.setSuccessMessage(translate('GPG Key updated'));

            }, function (response) {
                deferred.reject(response);
                Notification.setErrorMessage(response.data.displayMessage);
            });

            return deferred.promise;
        };

        $scope.removeGPGKey = function (gpgKey) {
            gpgKey.$delete(function () {
                $scope.transitionTo('gpg-keys');
            });
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.gpg-keys.controller:GPGKeyProductsController
 *
 * @requires $scope
 * @requires Nutupane
 * @requires GPGKey
 * @requires ApiErrorHandler
 *
 * @description
 *   Page for GPG Key products
 */

(function () {
    function GPGKeyProductsController($scope, Nutupane, GPGKey, ApiErrorHandler) {
        var nutupane = new Nutupane(GPGKey, {
            id: $scope.$stateParams.gpgKeyId
        }, 'products');
        $scope.controllerName = 'katello_gpg_keys';
        nutupane.masterOnly = true;

        $scope.panel = $scope.panel || {error: false, loading: false};

        $scope.gpgKey = GPGKey.get({id: $scope.$stateParams.gpgKeyId}, function () {
            $scope.panel.error = false;
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        $scope.table = nutupane.table;
    }

    angular.module('Bastion.gpg-keys').controller('GPGKeyProductsController', GPGKeyProductsController);
    GPGKeyProductsController.$inject = ['$scope', 'Nutupane', 'GPGKey', 'ApiErrorHandler'];
})();
(function () {
    function GPGKeyRepositoriesController($scope, Nutupane, GPGKey, ApiErrorHandler) {
        /**
         * @ngdoc object
         * @name  Bastion.gpg-keys.controller:GPGKeyRepositoriesController
         *
         * @requires $scope
         * @requires Nutupane
         * @requires GPGKey
         * @requires ApiErrorHandler
         *
         * @description
         *   Page for GPG Key repositories
         */
        var nutupane = new Nutupane(GPGKey, {
            id: $scope.$stateParams.gpgKeyId
        }, 'repositories');
        $scope.controllerName = 'katello_gpg_keys';
        nutupane.masterOnly = true;

        $scope.panel = $scope.panel || {error: false, loading: false};

        $scope.gpgKey = GPGKey.get({id: $scope.$stateParams.gpgKeyId}, function () {
            $scope.panel.error = false;
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        $scope.table = nutupane.table;
    }

    angular.module('Bastion.gpg-keys').controller('GPGKeyRepositoriesController', GPGKeyRepositoriesController);
    GPGKeyRepositoriesController.$inject = ['$scope', 'Nutupane', 'GPGKey', 'ApiErrorHandler'];
})();
/**
 * @ngdoc service
 * @name  Katello.gpg-keys.factory:GPGKey
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for GPG keys.
 */

angular.module('Bastion.gpg-keys').factory('GPGKey',
    ['BastionResource', 'CurrentOrganization',
    function (BastionResource, CurrentOrganization) {

        return BastionResource('katello/api/v2/gpg_keys/:id/:action',
            {id: '@id', 'organization_id': CurrentOrganization},
            {
                autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}},
                update: {method: 'PUT'},
                products: {method: 'GET', transformResponse:
                    function (data) {
                        var gpgKey = angular.fromJson(data);
                        return {
                            total: gpgKey.products.length,
                            subtotal: gpgKey.products.length,
                            results: gpgKey.products
                        };
                    }
                },
                repositories: {method: 'GET', transformResponse:
                    function (data) {
                        var gpgKey = angular.fromJson(data);
                        return {
                            total: gpgKey.repositories.length,
                            subtotal: gpgKey.repositories.length,
                            results: gpgKey.repositories
                        };
                    }
                }
            }
        );

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.gpg-keys.controller:GPGKeysController
 *
 * @requires $scope
 * @requires $location
 * @requires Nutupane
 * @requires GPGKey
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality specific to GPGKeys for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.gpg-keys').controller('GPGKeysController',
    ['$scope', '$location', 'Nutupane', 'GPGKey', 'CurrentOrganization',
    function ($scope, $location, Nutupane, GPGKey, CurrentOrganization) {
        var params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'sort_by': 'name',
            'sort_order': 'ASC',
            'paged': true
        };

        var nutupane = new Nutupane(GPGKey, params);
        $scope.controllerName = 'katello_gpg_keys';
        $scope.table = nutupane.table;
        $scope.panel = {loading: false};
        $scope.removeRow = nutupane.removeRow;
    }]
);
/**
 * @ngdoc object
 * @name Bastion.gpg-keys.config
 *
 * @requires $stateProvider
 *
 * @description
 *   Used for systems level configuration such as setting up the ui state machine.
 */

angular.module('Bastion.gpg-keys').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('gpg-keys', {
        url: '/gpg_keys',
        permission: 'view_gpg_keys',
        template: '<div ui-view></div>',
        views: {
            '@': {
                controller: 'GPGKeysController',
                templateUrl: 'gpg-keys/views/gpg-keys.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{ 'GPG Keys' | translate}}"
        }
    })
    .state('gpg-keys.new', {
        url: '/new',
        permission: 'create_gpg_keys',
        views: {
            '@': {
                controller: 'NewGPGKeyController',
                templateUrl: 'gpg-keys/new/views/new-gpg-key.html'
            }
        },
        ncyBreadcrumb: {
            label: "{{'New GPG Key' | translate }}",
            parent: 'gpg-keys'
        }
    })
    .state("gpg-key", {
        abstract: true,
        url: '/gpg_keys/:gpgKeyId',
        permission: 'view_gpg_keys',
        controller: 'GPGKeyDetailsController',
        templateUrl: 'gpg-keys/details/views/gpg-key-details.html'
    })
    .state('gpg-key.info', {
        url: '',
        permission: 'view_gpg_keys',
        controller: 'GPGKeyDetailsInfoController',
        templateUrl: 'gpg-keys/details/views/gpg-key-info.html',
        ncyBreadcrumb: {
            label: "{{ gpgKey.name }}",
            parent: 'gpg-keys'
        }
    })
    .state('gpg-key.products', {
        url: '/products',
        permission: 'view_gpg_keys',
        controller: 'GPGKeyProductsController',
        templateUrl: 'gpg-keys/details/views/gpg-key-products.html',
        ncyBreadcrumb: {
            label: "{{ 'Products' | translate }}",
            parent: 'gpg-key.info'
        }
    })
    .state('gpg-key.repositories', {
        url: '/repositories',
        permission: 'view_gpg_keys',
        controller: 'GPGKeyRepositoriesController',
        templateUrl: 'gpg-keys/details/views/gpg-key-repositories.html',
        ncyBreadcrumb: {
            label: "{{ 'Repositories' | translate }}",
            parent: 'gpg-key.info'
        }
    });
}]);
/**
 * @ngdoc object
 * @name  Bastion.gpg-keys.controller:NewGPGKeyController
 *
 * @requires $scope
 * @requires $state
 * @requires translate
 * @requires GPGKey
 * @requires CurrentOrganization
 * @requires Notification
 *
 * @description
 *   Controls the creation of an empty GPGKey object for use by sub-controllers.
 */

angular.module('Bastion.gpg-keys').controller('NewGPGKeyController',
    ['$scope', '$state', 'translate', 'GPGKey', 'CurrentOrganization', 'Notification',
    function ($scope, $state, translate, GPGKey, CurrentOrganization, Notification) {
        $scope.panel = {loading: false};
        $scope.gpgKey = new GPGKey();
        $scope.CurrentOrganization = CurrentOrganization;
        $scope.progress = {uploading: false};
        $scope.uploadURL = 'katello/api/v2/gpg_keys?organization_id=' + CurrentOrganization;

        $scope.uploadContent = function (response) {
            if (response) {
                if (angular.isUndefined(response.errors)) {
                    $scope.uploadStatus = 'success';
                    $scope.transitionTo('gpg-key.info', {gpgKeyId: response.id});
                    Notification.setSuccessMessage(translate('GPG key %s has been created.').replace('%s', response.name));
                } else {
                    Notification.setErrorMessage(translate("An error occurred while creating the GPG key: ") + response.displayMessage);
                    $scope.uploadStatus = 'error';
                }

                $scope.progress.uploading = false;
            }
        };
    }]
);
/**
 * @ngdoc module
 * @name  Bastion.activation-keys
 *
 * @description
 *   Module for activation keys related functionality.
 */

angular.module('Bastion.activation-keys', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.utils',
    'Bastion.components',
    'Bastion.host-collections',
    'Bastion.repository-sets'
]);
/**
 * @ngdoc factory
 * @name  Bastion.activation-keys.factory:ActivationKey
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for activation keys.
 */

angular.module('Bastion.activation-keys').factory('ActivationKey',
    ['BastionResource', function (BastionResource) {

        return BastionResource('katello/api/v2/activation_keys/:id/:action/:action2', {id: '@id'}, {
            get: {method: 'GET', params: {fields: 'full'}},
            update: {method: 'PUT'},
            copy: {method: 'POST', params: {action: 'copy'}},
            releaseVersions: {method: 'GET', params: {action: 'releases'}},
            subscriptions: {method: 'GET', params: {action: 'subscriptions'}},
            contentHosts: {method: 'GET', params: {action: 'systems'}},
            removeSubscriptions: {method: 'PUT', isArray: false, params: {action: 'remove_subscriptions'}},
            addSubscriptions: {method: 'PUT', isArray: false, params: {action: 'add_subscriptions'}},
            hostCollections: {method: 'GET', params: {action: 'host_collections'}},
            availableHostCollections: {method: 'GET', params: {action: 'host_collections', action2: 'available'}},
            removeHostCollections: {method: 'PUT', isArray: false, params: {action: 'host_collections'}},
            addHostCollections: {method: 'POST', isArray: false, params: {action: 'host_collections'}},
            contentOverride: {method: 'PUT', isArray: false, params: {action: 'content_override'}},
            autocomplete: {method: 'GET', isArray: true, params: {id: 'auto_complete_search'}},
            repositorySets: {method: 'GET', params: {action: 'product_content'}}
        });
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.activation-keys.controller:ActivationKeysController
 *
 * @requires $scope
 * @requires $location
 * @requires translate
 * @requires Nutupane
 * @requires ActivationKey
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality specific to activation keys for use with the Nutupane UI pattern.
 *   Defines the columns to display and the transform function for how to generate each row
 *   within the table.
 */

angular.module('Bastion.activation-keys').controller('ActivationKeysController',
    ['$scope', '$location', 'translate', 'Nutupane', 'ActivationKey', 'CurrentOrganization',
    function ($scope, $location, translate, Nutupane, ActivationKey, CurrentOrganization) {

        var params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'sort_by': 'name',
            'sort_order': 'ASC',
            'paged': true
        };

        var nutupane = new Nutupane(ActivationKey, params);
        $scope.controllerName = 'katello_activation_keys';
        nutupane.masterOnly = true;
        $scope.table = nutupane.table;
    }]
);
angular.module('Bastion.activation-keys').config(['$stateProvider', function ($stateProvider) {
    $stateProvider.state('activation-keys', {
        url: '/activation_keys',
        permission: 'view_activation_keys',
        template: '<div ui-view></div>',
        views: {
            '@': {
                controller: 'ActivationKeysController',
                templateUrl: 'activation-keys/views/activation-keys.html'
            }
        },
        ncyBreadcrumb: {
            label: '{{ "Activation Keys" | translate }}'
        }
    })
    .state('activation-keys.new', {
        url: '/new',
        permission: 'create_activation_keys',
        views: {
            '@': {
                controller: 'NewActivationKeyController',
                templateUrl: 'activation-keys/new/views/activation-key-new.html'
            }
        },
        ncyBreadcrumb: {
            label: '{{ "New Activation Key" | translate }}',
            parent: 'activation-keys'
        }
    });

    $stateProvider.state("activation-key", {
        abstract: true,
        url: '/activation_keys/:activationKeyId',
        permission: 'view_activation_keys',
        controller: 'ActivationKeyDetailsController',
        templateUrl: 'activation-keys/details/views/activation-key-details.html'
    })
    .state('activation-key.info', {
        url: '',
        permission: 'view_activation_keys',
        controller: 'ActivationKeyDetailsInfoController',
        templateUrl: 'activation-keys/details/views/activation-key-info.html',
        ncyBreadcrumb: {
            label: '{{ activationKey.name }}',
            parent: 'activation-keys'
        }
    })
    .state('activation-key.products', {
        url: '/products',
        permission: 'view_activation_keys',
        controller: 'ActivationKeyRepositorySetsController',
        templateUrl: 'activation-keys/details/views/activation-key-repository-sets.html',
        ncyBreadcrumb: {
            label: '{{ "Repository Sets" | translate }}',
            parent: 'activation-key.info'
        }
    })
    .state('activation-key.subscriptions', {
        abstract: true,
        templateUrl: 'activation-keys/details/views/activation-key-subscriptions.html'
    })
    .state('activation-key.subscriptions.list', {
        url: '/subscriptions',
        permission: 'view_activation_keys',
        controller: 'ActivationKeySubscriptionsController',
        templateUrl: 'activation-keys/details/views/activation-key-subscriptions-list.html',
        ncyBreadcrumb: {
            label: '{{ "List Subscriptions" | translate }}',
            parent: 'activation-key.info'
        }
    })
    .state('activation-key.subscriptions.add', {
        url: '/add-subscriptions',
        permission: 'edit_activation_keys',
        controller: 'ActivationKeyAddSubscriptionsController',
        templateUrl: 'activation-keys/details/views/activation-key-add-subscriptions.html',
        ncyBreadcrumb: {
            label: '{{ "Add Subscriptions" | translate }}',
            parent: 'activation-key.info'
        }
    });

    $stateProvider.state('activation-key.host-collections', {
        abstract: true,
        templateUrl: 'activation-keys/details/views/activation-key-host-collections.html'
    })
    .state('activation-key.host-collections.list', {
        url: '/host-collections',
        permission: 'view_activation_keys',
        controller: 'ActivationKeyHostCollectionsController',
        templateUrl: 'activation-keys/details/views/activation-key-host-collections-table.html',
        ncyBreadcrumb: {
            label: '{{ "List Host Collections" | translate }}',
            parent: 'activation-key.info'
        }
    })
    .state('activation-key.host-collections.add', {
        url: '/host-collections/add',
        permission: 'edit_activation_keys',
        controller: 'ActivationKeyAddHostCollectionsController',
        templateUrl: 'activation-keys/details/views/activation-key-host-collections-table.html',
        ncyBreadcrumb: {
            label: '{{ "Add Host Collections" | translate }}',
            parent: 'activation-key.info'
        }
    })
    .state('activation-key.associations-content-hosts', {
        url: '/associations/content-hosts',
        permission: 'view_activation_keys',
        controller: 'ActivationKeyAssociationsController',
        templateUrl: 'activation-keys/details/views/activation-key-associations-content-hosts.html',
        ncyBreadcrumb: {
            label: '{{ "Content Hosts" | translate }}',
            parent: 'activation-key.info'
        }
    })
    .state('activation-key.copy', {
        url: '/copy',
        permission: 'create_activation_key',
        controller: 'ActivationKeyCopyController',
        templateUrl: 'activation-keys/details/views/activation-key-copy.html',
        ncyBreadcrumb: {
            label: "{{ 'Create Copy' | translate }}",
            parent: 'activation-key.info'
        }
    });
}]);
/**
 * @ngdoc filter
 * @name  Bastion.activation-keys.filter:activationKeyConsumedFilter
 */

angular.module('Bastion.activation-keys').filter('activationKeyConsumedFilter',
    ['$filter', 'translate',
    function ($filter, translate) {
        return function (activationKey) {
            var quantity = $filter('unlimitedFilter')(activationKey['max_hosts'], activationKey['unlimited_hosts']);
            return translate('%(consumed)s out of %(quantity)s')
                .replace('%(consumed)s', activationKey['usage_count'])
                .replace('%(quantity)s', quantity);
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.systems.controller:ActivationKeyAddHostCollectionsController
 *
 * @requires $scope
 * @requires $q
 * @requires $location
 * @requires translate
 * @requires ActivationKey
 * @requires Nutupane
 * @requires Notification
 *
 * @description
 *   Provides the functionality for adding host collections to an activation key.
 */

angular.module('Bastion.activation-keys').controller('ActivationKeyAddHostCollectionsController',
    ['$scope', '$q', '$location', 'translate', 'ActivationKey', 'Nutupane', 'Notification',
    function ($scope, $q, $location, translate, ActivationKey, Nutupane, Notification) {
        var hostCollectionsPane, params;

        params = {
            'search': $location.search().search || "",
            'sort_by': 'name',
            'sort_order': 'ASC',
            'full_result': true,
            'id': $scope.$stateParams.activationKeyId
        };

        hostCollectionsPane = new Nutupane(ActivationKey, params, 'availableHostCollections');
        $scope.controllerName = 'katello_host_collections';
        $scope.table = hostCollectionsPane.table;

        $scope.addHostCollections = function () {
            var data,
                success,
                error,
                deferred = $q.defer(),
                hostCollectionsToAdd = _.map($scope.table.getSelected(), 'id');

            data = {
                "activation_key": {
                    "host_collection_ids": hostCollectionsToAdd
                }
            };

            success = function (response) {
                var message = translate('Added %x host collections to activation key "%y".')
                    .replace('%x', $scope.table.numSelected)
                    .replace('%y', $scope.activationKey.name);

                Notification.setSuccessMessage(message);
                $scope.table.working = false;
                $scope.table.selectAll(false);
                hostCollectionsPane.refresh();
                $scope.activationKey.$get();
                deferred.resolve(response);
            };

            error = function (response) {
                deferred.reject(response.data.errors);
                Notification.setErrorMessage(response.data.errors.base);
                $scope.table.working = false;
            };

            $scope.table.working = true;
            ActivationKey.addHostCollections({id: $scope.activationKey.id}, data, success, error);
            return deferred.promise;
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.activation-keys.controller:ActivationKeyAddSubscriptionsController
 *
 * @requires $scope
 * @requires $state
 * @requires $location
 * @requires translate
 * @requires Nutupane
 * @requires CurrentOrganization
 * @requires Subscription
 * @requires ActivationKey
 * @requires SubscriptionsHelper
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the activation key add subscriptions pane.
 */

angular.module('Bastion.activation-keys').controller('ActivationKeyAddSubscriptionsController',
    ['$scope', '$state', '$location', 'translate', 'Nutupane', 'CurrentOrganization', 'Subscription', 'ActivationKey', 'SubscriptionsHelper', 'Notification',
    function ($scope, $state, $location, translate, Nutupane, CurrentOrganization, Subscription, ActivationKey, SubscriptionsHelper, Notification) {

        var params;

        params = {
            'activation_key_id': $scope.$stateParams.activationKeyId,
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'sort_order': 'ASC',
            'available_for': 'activation_key'
        };

        $scope.contentNutupane = new Nutupane(Subscription, params);
        $scope.controllerName = 'katello_subscriptions';
        $scope.table = $scope.contentNutupane.table;
        $scope.isAdding = false;
        $scope.contentNutupane.setSearchKey('subscriptionSearch');
        $scope.contentNutupane.masterOnly = true;

        $scope.groupedSubscriptions = {};
        $scope.$watch('table.rows', function (rows) {
            $scope.groupedSubscriptions = SubscriptionsHelper.groupByProductName(rows);
        });

        $scope.disableAddButton = function () {
            return $scope.table.numSelected === 0 || $scope.isAdding;
        };

        $scope.addSelected = function () {
            var selected;
            selected = SubscriptionsHelper.getSelectedSubscriptionAmounts($scope.table);

            $scope.isAdding = true;
            ActivationKey.addSubscriptions({id: $scope.activationKey.id, 'subscriptions': selected}, function () {
                Notification.setSuccessMessage(translate("Successfully added %s subscriptions.").replace('%s', selected.length));
                $scope.isAdding = false;
                $scope.contentNutupane.refresh();
            }, function (response) {
                Notification.setErrorMessage(response.data.displayMessage);
                $scope.isAdding = false;
                $scope.contentNutupane.refresh();
            });
        };

        $scope.amountSelectorValues = function (subscription) {
            var value, values;

            values = [];
            for (value = 1; value < subscription.quantity && values.length < 5; value += 1) {
                values.push(value);
            }
            values.push(subscription.quantity);
            return values;
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.activation-keys.controller:ActivationKeyAssociationsController
 *
 * @requires $scope
 * @requires $location
 * @requires translate
 * @requires Nutupane
 * @requires ActivationKey
 * @requires ContentHostsHelper
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality for activation key associations.
 */

angular.module('Bastion.activation-keys').controller('ActivationKeyAssociationsController',
    ['$scope', '$location', 'translate', 'Nutupane', 'ActivationKey', 'ContentHostsHelper', 'CurrentOrganization', 'Host',
    function ($scope, $location, translate, Nutupane, ActivationKey, ContentHostsHelper, CurrentOrganization, Host) {
        var contentHostsNutupane, nutupaneParams, params = {
            'organization_id': CurrentOrganization,
            'search': $location.search().search || "",
            'page': 1,
            'sort_by': 'name',
            'sort_order': 'ASC',
            'paged': true
        };

        nutupaneParams = {
            'disableAutoLoad': true
        };

        contentHostsNutupane = new Nutupane(Host, params, undefined, nutupaneParams);
        $scope.controllerName = 'hosts';
        contentHostsNutupane.searchTransform = function (term) {
            var searchQuery, addition = "activation_key_id=" + $scope.$stateParams.activationKeyId;
            if (term === "" || angular.isUndefined(term)) {
                searchQuery = addition;
            } else {
                searchQuery = term + " and " + addition;
            }
            return searchQuery;
        };

        contentHostsNutupane.masterOnly = true;
        contentHostsNutupane.setSearchKey('contentHostSearch');

        $scope.table = contentHostsNutupane.table;
        $scope.table.working = true;

        if ($scope.contentHosts) {
            $scope.table.working = false;
        }

        $scope.activationKey.$promise.then(function () {
            contentHostsNutupane.setParams(params);
            contentHostsNutupane.load();
        });

        $scope.getHostStatusIcon = ContentHostsHelper.getHostStatusIcon;

        $scope.memory = ContentHostsHelper.memory;
    }]
);
(function () {
    'use strict';

    /**
     * @ngdoc controller
     * @name  Bastion.activation-keys.controller:ActivationKeyCopyController
     *
     * @description
     *   For copying a host collection.
     */
    function ActivationKeyCopyController($scope, ActivationKey, Notification) {
        $scope.copy = function (newName) {
            ActivationKey.copy({id: $scope.activationKey.id, 'new_name': newName}, function (response) {
                $scope.transitionTo('activation-key.info', {activationKeyId: response.id});
            }, function (response) {
                Notification.setErrorMessage(response.data.displayMessage);
            });
        };
    }

    angular
        .module('Bastion.activation-keys')
        .controller('ActivationKeyCopyController', ActivationKeyCopyController);

    ActivationKeyCopyController.$inject = ['$scope', 'ActivationKey', 'Notification'];

})();
/**
 * @ngdoc object
 * @name  Bastion.activation-keys.controller:ActivationKeyDetailsInfoController
 *
 * @requires $scope
 * @requires $q
 * @requires translate
 * @requires ActivationKey
 * @requires ContentView
 * @requires Organization
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality for the activation key info details action pane.
 */

angular.module('Bastion.activation-keys').controller('ActivationKeyDetailsInfoController',
    ['$scope', '$q', 'translate', 'ActivationKey', 'ContentView', 'Organization', 'CurrentOrganization',
    function ($scope, $q, translate, ActivationKey, ContentView, Organization, CurrentOrganization) {

        $scope.editContentView = false;
        $scope.editEnvironment = false;
        $scope.disableEnvironmentSelection = false;
        $scope.selectionRequired = true;
        $scope.environments = [];

        $scope.environments = Organization.readableEnvironments({id: CurrentOrganization});

        $scope.$on('activationKey.loaded', function () {
            $scope.originalEnvironment = $scope.activationKey.environment;
        });

        $scope.$watch('activationKey.environment', function (environment) {
            if (environment && $scope.originalEnvironment) {
                if (environment.id !== $scope.originalEnvironment.id) {
                    $scope.editContentView = true;
                    $scope.disableEnvironmentSelection = true;
                }
            } else if (environment) {
                $scope.editEnvironment = true;
                $scope.editContentView = true;
                $scope.disableEnvironmentSelection = true;
            }
        });

        $scope.cancelContentViewUpdate = function () {
            if ($scope.editContentView) {
                $scope.editContentView = false;
                $scope.editEnvironment = false;
                $scope.selectionRequired = false;
                $scope.activationKey.environment = $scope.originalEnvironment;
                $scope.disableEnvironmentSelection = false;
            }
        };

        $scope.saveContentView = function (activationKey) {
            $scope.editContentView = false;
            $scope.editEnvironment = false;
            $scope.save(activationKey).then(function (actKey) {
                $scope.originalEnvironment = actKey.environment;
            });
            $scope.disableEnvironmentSelection = false;
        };

        $scope.releaseVersions = function () {
            var deferred = $q.defer();

            ActivationKey.releaseVersions({ id: $scope.activationKey.id }, function (response) {
                deferred.resolve(response.results);
            });

            return deferred.promise;
        };

        $scope.clearReleaseVersion = function () {
            $scope.activationKey['release_version'] = '';
            $scope.save($scope.activationKey);
        };

        $scope.clearServiceLevel = function () {
            $scope.activationKey['service_level'] = '';
            $scope.save($scope.activationKey);
        };

        $scope.contentViews = function () {
            var deferred = $q.defer();

            ContentView.queryUnpaged({ 'environment_id': $scope.activationKey.environment.id }, function (response) {
                deferred.resolve(response.results);
            });

            return deferred.promise;
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.activation-keys.controller:ActivationKeyDetailsController
 *
 * @requires $scope
 * @requires $state
 * @requires $q
 * @requires translate
 * @requires ActivationKey
 * @requires CurrentOrganization
 * @requires Organization
 * @requires Notification
 * @requires ApiErrorHandler
 *
 * @description
 *   Provides the functionality for the activation key details action pane.
 */

angular.module('Bastion.activation-keys').controller('ActivationKeyDetailsController',
    ['$scope', '$state', '$q', 'translate', 'ActivationKey', 'Organization', 'CurrentOrganization', 'Notification', 'ApiErrorHandler',
    function ($scope, $state, $q, translate, ActivationKey, Organization, CurrentOrganization, Notification, ApiErrorHandler) {
        $scope.panel = {
            error: false,
            loading: true
        };

        if ($scope.activationKey) {
            $scope.panel.loading = false;
        }

        $scope.activationKey = ActivationKey.get({id: $scope.$stateParams.activationKeyId}, function (activationKey) {
            $scope.$broadcast('activationKey.loaded', activationKey);
            $scope.panel.loading = false;
        }, function (response) {
            $scope.panel.loading = false;
            ApiErrorHandler.handleGETRequestErrors(response, $scope);
        });

        $scope.save = function (activationKey) {
            var deferred = $q.defer();

            activationKey.$update(function (response) {
                deferred.resolve(response);
                Notification.setSuccessMessage(translate('Activation Key updated'));
            }, function (response) {
                deferred.reject(response);
                Notification.setErrorMessage(translate("An error occurred saving the Activation Key: ") + response.data.displayMessage);
            });
            return deferred.promise;
        };

        $scope.setActivationKey = function (activationKey) {
            $scope.activationKey = activationKey;
        };

        $scope.removeActivationKey = function (activationKey) {
            activationKey.$delete(function () {
                $scope.transitionTo('activation-keys');
                Notification.setSuccessMessage(translate('Activation Key removed.'));
            }, function (response) {
                Notification.setErrorMessage(translate("An error occurred removing the Activation Key: ") + response.data.displayMessage);
            });
        };

        $scope.serviceLevels = function () {
            var deferred = $q.defer();

            Organization.get({id: CurrentOrganization}, function (organization) {
                deferred.resolve(organization['service_levels']);
            });

            return deferred.promise;
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.systems.controller:ActivationKeyHostCollectionsController
 *
 * @requires $scope
 * @requires $q
 * @requires $location
 * @requires translate
 * @requires ActivationKey
 * @requires Nutupane
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the list host collections details action pane.
 */

angular.module('Bastion.activation-keys').controller('ActivationKeyHostCollectionsController',
    ['$scope', '$q', '$location', 'translate', 'ActivationKey', 'Nutupane', 'Notification',
    function ($scope, $q, $location, translate, ActivationKey, Nutupane, Notification) {
        var hostCollectionsPane, params;

        params = {
            'id': $scope.$stateParams.activationKeyId,
            'search': $location.search().search || "",
            'sort_by': 'name',
            'sort_order': 'ASC',
            'full_result': true
        };

        hostCollectionsPane = new Nutupane(ActivationKey, params, 'hostCollections');
        $scope.controllerName = 'katello_host_collections';
        $scope.table = hostCollectionsPane.table;

        $scope.removeHostCollections = function () {
            var data,
                success,
                error,
                deferred = $q.defer(),
                hostCollectionsToRemove = _.map($scope.table.getSelected(), 'id');

            data = {
                "activation_key": {
                    "host_collection_ids": hostCollectionsToRemove
                }
            };

            success = function (response) {
                var message = translate('Removed %x host collections from activation key "%y".')
                    .replace('%x', $scope.table.numSelected)
                    .replace('%y', $scope.activationKey.name);

                Notification.setSuccessMessage(message);
                $scope.table.working = false;
                $scope.table.selectAll(false);
                hostCollectionsPane.refresh();
                $scope.activationKey.$get();
                deferred.resolve(response);
            };

            error = function (response) {
                deferred.reject(response.data.errors);
                angular.forEach(response.data.errors, function (responseError) {
                    Notification.setErrorMessage(responseError);
                });
                $scope.table.working = false;
            };

            $scope.table.working = true;
            ActivationKey.removeHostCollections({id: $scope.activationKey.id}, data, success, error);
            return deferred.promise;
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.activation-keys.controller:ActivationKeyRepositorySetsController
 *
 * @requires $scope
 * @requires translate
 * @requires Nutupane
 * @requires ActivationKey
 * @requires ContentOverrideHelper
 * @requires Notification
 * @requires CurrentOrganization
 *
 * @description
 *   Provides the functionality for the activation-key products action pane.
 */

angular.module('Bastion.activation-keys').controller('ActivationKeyRepositorySetsController',
    ['$scope', 'translate', 'Nutupane', 'ActivationKey', 'ContentOverrideHelper', 'Notification', 'CurrentOrganization',
    function ($scope, translate, Nutupane, ActivationKey, ContentOverrideHelper, Notification, CurrentOrganization) {
        var params, saveContentOverride, success, error;

        params = {
            id: $scope.$stateParams.activationKeyId,
            'organization_id': CurrentOrganization,
            enabled: true,
            'full_result': true,
            'include_available_content': true
        };

        $scope.controllerName = 'katello_products';
        $scope.nutupane = new Nutupane(ActivationKey, params, 'repositorySets');
        $scope.table = $scope.nutupane.table;

        $scope.contentAccessModes = {
            contentAccessModeAll: false,
            contentAccessModeEnv: false
        };
        $scope.toggleFilters = function () {
            $scope.nutupane.table.params['content_access_mode_env'] = $scope.contentAccessModes.contentAccessModeEnv;
            $scope.nutupane.table.params['content_access_mode_all'] = $scope.contentAccessModes.contentAccessModeAll;
            $scope.nutupane.refresh();
        };

        success = function () {
            $scope.table.working = false;
            Notification.setSuccessMessage(translate('Repository Sets settings saved successfully.'));
            $scope.nutupane.refresh();
        };

        error = function (response) {
            $scope.table.working = false;
            Notification.setErrorMessage(response.data.errors);
        };

        saveContentOverride = function (contentOverrides) {
            $scope.table.working = true;
            ActivationKey.contentOverride({id: $scope.$stateParams.activationKeyId}, contentOverrides, success, error);
        };

        $scope.overrideToEnabled = function () {
            var contentOverrides = ContentOverrideHelper.getEnabledContentOverrides($scope.table.getSelected());
            saveContentOverride(contentOverrides);
        };

        $scope.overrideToDisabled = function () {
            var contentOverrides = ContentOverrideHelper.getDisabledContentOverrides($scope.table.getSelected());
            saveContentOverride(contentOverrides);
        };

        $scope.resetToDefault = function () {
            var contentOverrides = ContentOverrideHelper.getDefaultContentOverrides($scope.table.getSelected());
            saveContentOverride(contentOverrides);
        };
    }]
);
/**
 * @ngdoc object
 * @name  Bastion.activation-keys.controller:ActivationKeySubscriptionsController
 *
 * @requires $scope
 * @requires $location
 * @requires translate
 * @requires Nutupane
 * @requires ActivationKey
 * @requires SubscriptionsHelper
 * @requires Notification
 *
 * @description
 *   Provides the functionality for the activation key subscriptions details action pane.
 */

angular.module('Bastion.activation-keys').controller('ActivationKeySubscriptionsController',
    ['$scope', '$location', 'translate', 'Nutupane', 'ActivationKey', 'Subscription', 'SubscriptionsHelper', 'Notification',
    function ($scope, $location, translate, Nutupane, ActivationKey, Subscription, SubscriptionsHelper, Notification) {
        var params;

        params = {
            'id': $scope.$stateParams.activationKeyId,
            'search': $location.search().search || "",
            'sort_order': 'ASC',
            'paged': true
        };

        $scope.contentNutupane = new Nutupane(ActivationKey, params, "subscriptions");
        $scope.controllerName = 'katello_subscriptions';
        $scope.table = $scope.contentNutupane.table;
        $scope.contentNutupane.masterOnly = true;
        $scope.contentNutupane.setSearchKey('subscriptionSearch');
        $scope.isRemoving = false;

        $scope.groupedSubscriptions = {};
        $scope.$watch('table.rows', function (rows) {
            $scope.groupedSubscriptions = SubscriptionsHelper.groupByProductName(rows);
        });

        $scope.disableRemoveButton = function () {
            return $scope.table.numSelected === 0 || $scope.isRemoving;
        };

        $scope.removeSelected = function () {
            var selected;
            selected = SubscriptionsHelper.getSelectedSubscriptionAmounts($scope.table);

            $scope.isRemoving = true;
            ActivationKey.removeSubscriptions({id: $scope.activationKey.id, 'subscriptions': selected}, function () {
                $scope.contentNutupane.table.selectAll(false);
                $scope.contentNutupane.refresh();
                Notification.setSuccessMessage(translate("Successfully removed %s subscriptions.").replace('%s', selected.length));
                $scope.isRemoving = false;
            }, function (response) {
                $scope.isRemoving = false;
                Notification.setErrorMessage(translate("An error occurred removing the subscriptions.") + response.data.displayMessage);
            });
        };

    }]
);
/**
 * @ngdoc object
 * @name  Bastion.activation-keys.controller:NewActivationKeyController
 *
 * @requires $scope
 * @requires $q
 * @requires FormUtils
 * @requires ActivationKey
 * @requires Organization
 * @requires CurrentOrganization
 * @requires ContentView
 *
 * @description
 *   Controls the creation of an empty ActivationKey object for use by sub-controllers.
 */

angular.module('Bastion.activation-keys').controller('NewActivationKeyController',
    ['$scope', '$q', 'FormUtils', 'ActivationKey', 'Organization', 'CurrentOrganization', 'ContentView',
    function ($scope, $q, FormUtils, ActivationKey, Organization, CurrentOrganization, ContentView) {

        function success() {
            $scope.transitionTo('activation-key.info', {activationKeyId: $scope.activationKey.id});
        }

        function error(response) {
            $scope.working = false;
            angular.forEach(response.data.errors, function (errors, field) {
                $scope.activationKeyForm[field].$setValidity('server', false);
                $scope.activationKeyForm[field].$error.messages = errors;
            });
        }

        $scope.activationKey = $scope.activationKey || new ActivationKey();
        $scope.activationKey['unlimited_hosts'] = true;

        $scope.panel = {loading: false};
        $scope.organization = CurrentOrganization;

        $scope.contentViews = [];
        $scope.editContentView = false;
        $scope.environments = [];

        $scope.environments = Organization.readableEnvironments({id: CurrentOrganization});

        $scope.$watch('activationKey.environment', function (environment) {
            if (environment) {
                $scope.editContentView = true;
                ContentView.queryUnpaged({ 'environment_id': environment.id }, function (response) {
                    $scope.contentViews = response.results;
                });
            }
        });

        $scope.save = function (activationKey) {
            activationKey['organization_id'] = CurrentOrganization;
            activationKey.$save(success, error);
        };

    }]
);
/**
 * @ngdoc module
 * @name  Bastion.settings
 *
 * @description
 *    Module for settings
 */

angular.module('Bastion.settings', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.common'
]);
/**
 * @ngdoc service
 * @name  Bastion.settings.factory:Setting
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for a Setting.
 */

angular.module('Bastion.settings').factory('Setting',
    ['BastionResource', function (BastionResource) {
        var resource = BastionResource('api/v2/settings/');
        return resource;
    }]
);
/**
 * @ngdoc module
 * @name  Bastion.architectures
 *
 * @description
 *    Module for architectures
 */

angular.module('Bastion.architectures', [
    'ngResource',
    'ui.router',
    'Bastion',
    'Bastion.i18n',
    'Bastion.common'
]);
/**
 * @ngdoc service
 * @name  Bastion.architectures.factory:Architecture
 *
 * @requires BastionResource
 *
 * @description
 *   Provides a BastionResource for architectures.
 */

angular.module('Bastion.architectures').factory('Architecture',
    ['BastionResource', function (BastionResource) {
        var resource = BastionResource('api/v2/architectures/');
        return resource;
    }]
);
angular.module('Bastion.i18n').run(['gettextCatalog', function (gettextCatalog) {
/* jshint -W100 */
    gettextCatalog.setStrings('de', {"%(consumed)s out of %(quantity)s":"%(consumed)s von %(quantity)s","(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)":"(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"Taskdetails\" | translate }}</a>)","(Not all Activation Keys editable )":"(Nicht alle Aktivierungsschlüssel sind änderbar)","(Not all Content Hosts editable )":"(Nicht alle Inhaltshosts sind änderbar)","* These marked Content View Versions are from Composite Content Views.  Their components needing updating are listed underneath.":"* Diese markierten Inhaltsansichtsversionen stammen aus Verbund-Inhaltsansichten. Deren Komponenten, die eine Aktualisierung erfordern, werden darunter aufgelistet.","+ New Sync Plan":"+ Neuer Synchronisationsplan","- Date and Type":"- Datum und Typ","-- select an interval --":"-- Intervall wählen --","1 Content Host":["1 Inhaltshost","{{ contentHost.virtual_guests.length }} Inhaltshosts"],"1 repository sync has errors.":["1 Repository-Synchronisation hat Fehler.","{{ product.sync_summary.error}} Repository-Synchronisationen haben Fehler."],"1 repository sync in progress.":["1 Repository-Synchronisation in Arbeit.","{{ product.sync_summary.pending}} Repository-Synchronisationen in Arbeit."],"1 successfully synced repository.":["1 erfolgreich synchronisiertes Repository.","{{ product.sync_summary.success}} erfolgreich synchronisierte Repositorys."],"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Version {{ version.version }}</a>":"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Version {{ version.version }}</a>","<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Create":"<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Erstellen","<span translate=\"\">\n        If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">here</a>\n      <span translate=\"\">to manage them individually.</span>":"<span translate=\"\">\n        Falls Sie Teile dieser Aktivierungsschlüssel lieber in andere Inhaltsansichten oder Lebenszyklusumgebungen verlegen möchten, klicken Sie\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">hier</a>\n      <span translate=\"\">, um sie einzeln zu verwalten.</span>","A composite view contains other content views.":"Eine Verbundinhaltsansicht enthält mehrere andere Inhaltsansichten.","A new version of {{ contentView.name }} will be created and promoted to the Library environment.\n    It can be promoted to other environments from the Versions tab of this Content View.":"Eine neue Version von {{ contentView.name }} wird erstellt und in die Library-Umgebung übertragen.\n    Auf dem Versionen-Reiter dieser Inhaltsansicht kann die Version in andere Umgebungen übertragen werden.","Account":"Account","Account Number":"Account-Nummer:","Action":"Aktion","Action Type":"Aktionstyp","Actions":"Aktionen","Actions for:":"Aktionen für:","Activation Key":["Aktivierungsschlüssel","Aktivierungsschlüssel"],"Activation Key Content":"Aktivierungsschlüsselinhalt","Activation Key Type:":"Aktivierungsschlüsseltyp:","Activation Key removed.":"Aktivierungsschlüssel entfernt","Activation Key updated":"Aktivierungsschlüssel aktualisiert","Activation Key:":"Aktivierungsschlüssel:","Activation Key: {{ activationKey.name }}":"Aktivierungsschlüssel: {{ activationKey.name }}","Activation Keys":"Aktivierungsschlüssel","Activation Keys for Subscription:":"Aktivierungsschlüssel für Subskription:","Activation Keys using Version {{ version.version }}":"Aktivierungsschlüssel mit Version {{ version.version }}","Add":"Hinzufügen","Add Content Hosts to:":"Inhaltshosts hinzufügen zu:","Add Content Views":"Inhaltshosts hinzufügen","Add Errata":"Errata hinzufügen","Add New Environment":"Hinzufügen einer neuen Umgebung","Add New Filter":"Neuen Filter hinzufügen","Add New Module":"Neues Modul hinzufügen","Add New Repository":"Neues Repository hinzufügen","Add Package Group":"Paketgruppe hinzufügen","Add Repositories":"Repositorys hinzufügen","Add Selected":"Auswahl hinzufügen","Add Subscriptions for Activation Key:":"Subskriptionen für Aktivierungsschlüssel hinzufügen:","Add To":"Hinzufügen zu","Added %x host collections to activation key \"%y\".":"%x Hostsammlungen wurden zu Aktivierungsschlüssel \"%y\" hinzugefügt.","Added %x host collections to content host \"%y\".":"%x Hostsammlungen wurden zu Inhaltshost \"%y\" hinzugefügt.","Added %x products to sync plan \"%y\".":"%x Produkte wurden zu Synchronisationsplan \"%y\" hinzugefügt.","Adding Lifecycle Environment to the end of \"{{ priorEnvironment.name }}\"":"Lebenszyklusumgebung wird nach \"{{ priorEnvironment.name }}\" hinzugefügt","Advanced Information":"Erweiterte Informationen","Advisory":"Advisory","Affected Activation Keys":"Betroffene Aktivierungsschlüssel","Affected Content Hosts":"Betroffene Inhaltshosts","Affected Packages":"Betroffene Pakete","Affected Systems":"Betroffene Systeme","Affected repositories":"Betroffene Repositorys","Affected repositories have been updated.":"Betroffene Repositorys wurden aktualisiert.","All Content Views":"Alle Inhaltsansichten","All Products":"Alle Produkte","All Repositories":"Alle Repositorys","All Versions":"Alle Versionen","Alter Sync Plans":"Synchronisationspläne ändern","An error occurred applying Subscriptions:":"Fehler beim Anwenden von Subskriptionen:","An error occurred removing the Activation Key:":"Fehler beim Entfernen des Aktivierungsschlüssels:","An error occurred removing the Host Collection:":"Fehler beim Entfernen der Hostsammlung:","An error occurred removing the Products:":"Fehler beim Entfernen der Produkte:","An error occurred removing the content hosts.":"Fehler beim Entfernen der Inhaltshosts:","An error occurred removing the environment:":"Fehler beim Entfernen der Umgebung:","An error occurred removing the subscriptions.":"Fehler beim Entfernen der Subskriptionen:","An error occurred saving the Activation Key:":"Fehler beim Speichern des Aktivierungsschlüssels:","An error occurred saving the Content Host:":"Fehler beim Speichern des Inhaltshosts:","An error occurred saving the Environment:":"Fehler beim Speichern der Umgebung:","An error occurred saving the Filter:":"Fehler beim Speichern des Filters:","An error occurred saving the Host Collection:":"Fehler beim Speichern der Hostsammlung:","An error occurred saving the Product:":"Fehler beim Speichern des Produkts:","An error occurred saving the Repository:":"Fehler beim Speichern des Repositorys:","An error occurred saving the Sync Plan:":"Fehler beim Speichern des Synchronisationsplans:","An error occurred saving the URL:":"Fehler beim Speichern der URL:","An error occurred trying to auto-attach subscriptions.  Please check your log for further information.":"Fehler beim automatischen Verknüpfen von Subskriptionen. Bitte werfen Sie einen Blick auf die Protokolldatei für weitere Informationen.","An error occurred updating the Content View:":"Fehler beim Aktualisieren der Inhaltsansicht:","Applicable":"Relevant","Applicable Content Hosts for Errata":"Relevante Inhaltshosts für Errata","Applicable Errata":"Relevante Errata","Applicable hosts for {{ erratum.errata_id }}":"Relevante Hosts für {{ erratum.errata_id }}","Apply":"Anwenden","Apply Errata":"Errata anwenden","Apply Errata to Content Host \"{{contentHost.name}}\"?":"Errata auf Inhaltshost \"{{contentHost.name}}\" anwenden?","Apply Errata to Content Hosts":"Errata auf Inhaltshosts anwenden","Apply Errata to Content Hosts immediately after publishing.":"Sofort nach Veröffentlichung Errata auf Inhaltshosts anwenden.","Apply Selected":"Auswahl anwenden","Apply to Content Hosts":"Auf Inhaltshosts anwenden","Apply to Hosts":"Auf Hosts anwenden","Apply {{ errata.errata_id }}":"{{ errata.errata_id }} anwenden","Apply {{ errata.errata_id }} to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"{{ errata.errata_id }} auf {{ numberOfContentHostsToUpdate }} Inhaltshost(s) anwenden?","Apply {{ errataIds.length }} Errata to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"{{ errataIds.length }} Errata auf {{ numberOfContentHostsToUpdate }} Inhaltshost(s) anwenden?","Applying":"Anwenden","Applying Bulk Errata Update":"Massen-Errata-Aktualisierung wird angewendet","Arch":"Architektur","Architecture(s)":"Architektur(en)","Archived Copy":"Archivierte Kopie","Are you sure you want to add the {{ table.numSelected }} content host(s) selected to the host collection(s) chosen?":"Sind Sie sicher, dass Sie die {{ table.numSelected }} gewählten Inhaltshosts zu den ausgewählten Hostsammlungen hinzufügen möchten?","Are you sure you want to apply Errata to content host \"{{ contentHost.name }}\"?":"Sind Sie sicher, dass Sie die Errata auf Inhaltshost \"{{ contentHost.name }}\" anwenden möchten?","Are you sure you want to apply the {{ detailsTable.numSelected }} selected errata to the content hosts chosen?":"Sind Sie sicher, dass Sie die {{ detailsTable.numSelected }} gewählten Errata auf die ausgewählten Inhaltshosts anwenden möchten?","Are you sure you want to assign the {{ table.numSelected }} content host(s) selected to {{ selected.contentView.name }} in {{ selected.environment.name }}?":"Sind Sie sicher, dass Sie die {{ table.numSelected }} gewählten Inhaltshosts zu {{ selected.contentView.name }} in {{ selected.environment.name }} zuweisen möchten?","Are you sure you want to auto-attach available subscriptions to all {{ table.resource.total }} content hosts?":"Sind Sie sicher, dass Sie verfügbare Subskriptionen automatisch mit allen {{ table.resource.total }} Inhaltshosts verknüpfen möchten?","Are you sure you want to delete Content View \"{{ contentView.name }}\"?":"Sind Sie sicher, dass Sie die Inhaltsansicht \"{{ contentView.name }}\" löschen möchten?","Are you sure you want to install {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"Sind Sie sicher, dass Sie {{ content.content }} auf den {{ getSelectedSystemIds().length }} ausgewählten Systemen installieren möchten?","Are you sure you want to remove Activation Key \"{{ activationKey.name }}\"?":"Sind Sie sicher, dass Sie den Aktivierungsschlüssel \"{{ activationKey.name }}\" entfernen möchten?","Are you sure you want to remove GPG key {{ gpgKey.name }}?":"Sind Sie sicher, dass Sie den GPG-Schlüssel  {{ gpgKey.name }} entfernen möchten?","Are you sure you want to remove Host Collection \"{{ hostCollection.name }}\"?":"Sind Sie sicher, dass Sie die Hostsammlung \"{{ hostCollection.name }}\" entfernen möchten?","Are you sure you want to remove Sync Plan \"{{ syncPlan.name }}\"?":"Sind Sie sicher, dass Sie den Synchronisationsplan \"{{ syncPlan.name }}\" entfernen möchten?","Are you sure you want to remove product \"{{ product.name }}\"?":"Sind Sie sicher, dass Sie das Produkt \"{{ product.name }}\" entfernen möchten?","Are you sure you want to remove repository \"{{ repository.name }}\"?":"Sind Sie sicher, dass Sie das Repository \"{{ repository.name }}\" entfernen möchten?","Are you sure you want to remove the {{ productTable.getSelected().length }} products(s) selected?":"Sind Sie sicher, dass Sie die {{ productTable.getSelected().length }} ausgewählten Produkte entfernen möchten?","Are you sure you want to remove the {{ table.numSelected }} content host(s) selected from the host collection(s) chosen?":"Sind Sie sicher, dass Sie die {{ table.numSelected }} gewählten Inhaltshosts aus der ausgewählten Hostsammlung entfernen möchten?","Are you sure you want to remove {{ content.content }} from the {{ getSelectedSystemIds().length }} system(s) selected?":"Sind Sie sicher, dass Sie {{ content.content }} von den {{ getSelectedSystemIds().length }} ausgewählten Systemen entfernen möchten?","Are you sure you want to unregister content host \"{{ contentHost.name }}\"?":"Sind Sie sicher, dass Sie den Inhaltshost \"{{ contentHost.name }}\" abmelden möchten?","Are you sure you want to unregister the {{ table.numSelected }} content host(s) selected?":"Sind Sie sicher, dass Sie die {{ table.numSelected }} ausgewählten Inhaltshosts abmelden möchten?","Are you sure you want to update {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"Sind Sie sicher, dass Sie {{ content.content }} auf den {{ getSelectedSystemIds().length }} ausgewählten Systemen aktualisieren möchten?","As part of this deletion, 1 Content View Version will be deleted.":["Im Rahmen dieses Löschvorgangs wird 1 Inhaltsansichtsversion gelöscht.","Im Rahmen dieses Löschvorgangs werden {{ versions.length }} Inhaltsansichtsversionen gelöscht."],"Assign":"Zuweisen","Assign Environment and Content View":"Umgebung und Inhaltsansicht zuweisen","Associations":"Zuweisungen","At least one Errata needs to be selected to Apply.":"Mindestens ein Erratum muss zum Anwenden ausgewählt werden.","At least one content host needs to be selected in order to perform a bulk action.":"Mindestens ein Inhaltshost muss ausgewählt werden, um eine Massenaktion durchzuführen.","At least one product needs to be selected in order to perform a bulk action.":"Mindestens ein Produkt muss ausgewählt werden, um eine Massenaktion durchzuführen","Attached":"Verknüpft","Attached to Content Hosts":"Verknüpft mit Inhaltshosts","Author":"Autor","Auto-Attach":"Automatische Verknüpfung","Auto-attach available subscriptions to all content hosts":"Automatische Verknüpfung verfügbarer Subskriptionen mit allen Inhaltshosts","Automatic":"Automatisch","Available Content Views for Composite Content View:":"Verfügbare Inhaltsansichten für Verbundinhaltsansicht:","Available Puppet Modules for Content View:":"Verfügbare Puppet-Module für Inhaltsansicht:","Back":"Zurück","Back To Errata List":"Zurück zur Errata-Liste","Back To Repository Selection":"Zurück zur Repository-Auswahl","Back to Module List":"Zurück zur Modulliste","Back to Repository Details":"Zurück zu Repository-Details","Back to Repository List":"Zurück zur Repository-Liste","Back to Version List":"Zurück zur Versionsliste","Basic Information":"Allgemeine Informationen","Below are the product's repository content sets currently available for this subscription. For Red Hat subscriptions, additional content can be made available through the":"Nachfolgend finden Sie die Repository-Inhaltsgruppen des Produkts, die derzeit für diese Subskription zur Verfügung stehen. Für Red Hat Subskriptionen können weitere Inhalte verfügbar gemacht werden auf der Seite","Below are the repository content sets currently available for this activation key through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"Nachfolgend finden Sie die Repository-Inhaltsgruppen, die derzeit für diesen Aktivierungsschlüssel über dessen Subskriptionen zur Verfügung stehen. Für Red Hat Subskriptionen können weitere Inhalte verfügbar gemacht werden auf der Seite","Below are the repository content sets currently available for this content host through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"Nachfolgend finden Sie die Repository-Inhaltsgruppen, die derzeit für diesen Inhaltshost über dessen Subskriptionen zur Verfügung stehen. Für Red Hat Subskriptionen können weitere Inhalte verfügbar gemacht werden auf der Seite","Bug Fix":"Bugfix","Bug Fix Advisory":"Bugfix-Advisory","Bugfix":"Bugfix","Bulk Actions":"Massenaktionen","Bulk Alter Products Sync Plan":"Produktsynchronisationsplan in Massenaktion ändern","Bulk Sync Products":"Produkte in Massenaktion synchronisieren","CVEs":"CVEs","Cancel":"Abbrechen","Capacity":"Kapazität","Change assigned Lifecycle Environment or Content View":"Ändern der zugewiesenen Lebenszyklusumgebung oder Inhaltsansicht","Changing default settings for content hosts that register with this activation key requires subscription-manager version 1.10 or newer to be installed on that host.":"Um die Standardeinstellungen für Inhaltshosts zu ändern, die sich mit diesem Aktivierungsschlüssel registrieren, muss subscription-manager Version 1.10 oder höher auf dem Host installiert sein.","Changing default settings requires subscription-manager version 1.10 or newer to be installed on this host.":"Um die Standardeinstellungen zu ändern, muss subscription-manager Version 1.10 oder höher auf diesem Host installiert sein.","Checkin":"Anmeldung","Checksum":"Prüfsumme","Choose a lifecycle environment from the available promotion paths.":"Wählen Sie eine Lebenszyklusumgebung aus den verfügbaren Übertragungspfaden.","Click here to check the status of the task.":"Klicken Sie hier, um den Status der Task zu prüfen.","Click here to refresh.":"Zum Aktualisieren hier klicken.","Click here to see the task for the successful removals.":"Klicken Sie hier, um die Task für die erfolgreichen Entfernungen zu sehen.","Click here to select Errata for an Incremental Update.":"Klicken Sie hier, um Errata für eine inkrementelle Aktualisierung auszuwählen.","Close":"Schließen","Collection Actions":"Sammlungsaktionen","Completely remove version?":"Version vollständig entfernen?","Components":"Komponenten","Components:":"Komponenten:","Composite Content View {{ contentView.name }}":"Verbundsinhaltsansicht {{ contentView.name }}","Composite View":"Verbundansicht","Composite View?":"Verbundansicht?","Composite?":"Verbund?","Confirm":"Bestätigen","Confirm Remove":"Entfernen bestätigen","Confirm Version Removal: Version {{ version.version }}":"Entfernen der Version bestätigen: Version {{ version.version }}","Confirmation":"Bestätigung","Consumed":"Verbraucht","Content":"Inhalt","Content Counts":"Anzahl der Inhalte","Content Download URL":"Inhaltsdownload-URL","Content Host":"Inhaltshost","Content Host %s has been deleted.":"Inhaltshost %s wurde gelöscht.","Content Host Bulk Actions":"Inhaltshost-Massenaktionen","Content Host Bulk Content":"Inhaltshost-Masseninhalt","Content Host Bulk Errata":"Inhaltshost-Massenerrata","Content Host Bulk Host Collections":"Inhaltshost-Massenhostsammlung","Content Host Bulk Packages":"Inhaltshost-Massenpakete","Content Host Bulk Subscriptions":"Inhaltshost-Massensubskriptionen","Content Host Content":"Inhalte des Inhaltshosts","Content Host Counts":"Inhaltshostanzahl","Content Host Errata Management":"Errataverwaltung für Inhaltshost","Content Host Limit":"Inhaltshostgrenze","Content Host Package Management":"Paketverwaltung für Inhaltshost","Content Host Properties":"Eigenschaften des Inhaltshosts","Content Host Registration":"Registrierung eines Inhaltshosts","Content Host Status":"Status des Inhaltshosts","Content Host Subscriptions":"Subskriptionen des Inhaltshosts","Content Host {{ contentHost.name }}":"Inhaltshost {{ contentHost.name }}","Content Host:":"Inhaltshost:","Content Hosts":"Inhaltshosts","Content Hosts for Activation Key:":"Inhaltshosts für Aktivierungsschlüssel:","Content Hosts for Subscription:":"Inhaltshosts für Subskription:","Content Hosts for:":"Inhaltshosts für:","Content Hosts using Version {{ version.version }}":"Inhaltshosts mit Version {{ version.version }}","Content Type":"Inhaltstyp","Content View":"Inhaltsansicht","Content View Name":"Name der Inhaltsansicht","Content View Version":"Version der Inhaltsansicht","Content View updated.":"Inhaltsansicht aktualisiert.","Content View:":"Inhaltsansicht:","Content Views":"Inhaltsansichten","Content Views <div>{{ library.counts.content_views || 0 }}</div>":"Inhaltsansichten <div>{{ library.counts.content_views || 0 }}</div>","Content Views for Composite Content View:":"Inhaltsansichten für Verbundinhaltsansicht:","Content successfully uploaded":"Inhalt erfolgreich hochgeladen","Contents":"Inhalte","Contract":"Vertrag","Contract Number":"Vertragsnummer","Copy":"Kopieren","Copy Activation Key":"Aktivierungsschlüssel kopieren","Copy Collection":"Sammlung kopieren","Copy View":"Ansicht kopieren","Cores per Socket":"Kerne pro Socket","Cores: %s":"Kerne: %s","Create":"Erstellen","Create New View":"Neue Ansicht erstellen","Create Repositories":"Repositorys erstellen","Create Repositories Within:":"Repositorys erstellen in:","Create Repository":"Repository erstellen","Create Selected":"Auswahl erstellen","Create Sync Plan":"Synchronisationsplan erstellen","Create new view":"Neue Ansicht erstellen","Create view from existing views":"Ansicht aus vorhandenen Ansichten erstellen","Critical":"Kritisch","Current Environment (%e/%cv)":"Derzeitige Umgebung (%e/%cv)","Current Subscriptions for Activation Key:":"Derzeitige Subskriptionen für Aktivierungsschlüssel:","Currently Selected Puppet Modules":"Derzeit ausgewählte Puppet-Module","Daily at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"Täglich um {{ product.sync_plan.sync_date | date:'mediumTime' }} (Serverzeit)","Date":"Datum","Default":"Standard","Delete":"Löschen","Delete Manifest":"Manifest löschen","Delete Version {{ version.version }}":"Version {{ version.version }} löschen","Delete {{ contentView.name }}":"{{ contentView.name }} löschen","Deleted from %s":"Gelöscht von %s","Deleting Manifest...":"Manifest wird gelöscht …","Deleting from %count environments.":"Von %count Umgebungen löschen.","Deleting from 1 environment.":"Von 1 Umgebung löschen.","Deletion from %s":"Löschen von %s","Description":"Beschreibung","Detail":"Detail","Details":"Details","Details for Activation Key:":"Details für Aktivierungsschlüssel:","Details for Content View:":"Details für Inhaltsansicht:","Details for Product:":"Details für Produkt:","Discover":"Suchen","Discovered URLs":"Gefundene URLs:","Docker Content":"Docker-Inhalt","Docker Images":"Docker-Images","Docker Images in {{ repository.name }}":"Docker-Images in {{ repository.name }}","Docker Repositories":"Docker-Repositorys","Docker Repositories <div>{{ library.counts.docker_repositories || 0 }}</div>":"Docker-Repositorys <div>{{ library.counts.docker_repositories || 0 }}</div>","Docker Tag:":"Docker-Tag:","Docker Tags":"Docker-Tags","Docker metadata generation has been initiated in the background.  Click\n        <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Die Generierung von Docker-Metadaten wurde im Hintergrund initiiert. Klicken Sie\n        <a ng-href=\"{{ taskUrl() }}\">hier</a>, um den Fortschritt nachzuverfolgen.","Edit":"Bearbeiten","Enabled?":"Aktiviert?","End Date":"Enddatum","Ends":"Endet","Enhancement":"Erweiterung","Enter Package Group Name(s)...":"Paketgruppennamen eingeben …","Enter Package Name(s)...":"Paketnamen eingeben …","Environment":"Umgebung","Environment saved":"Umgebung gespeichert","Environment: {{ environment.name }}":"Umgebung: {{ environment.name }}","Environments":"Umgebungen","Environments List":"Umgebungsliste","Environments:":"Umgebungen:","Equal To":"Ist gleich","Errata":"Errata","Errata <div>{{ library.counts.errata.total || 0 }}</div>":"Errata <div>{{ library.counts.errata.total || 0 }}</div>","Errata Details":"Errata-Details","Errata Filter:":"Errata-Filter:","Errata ID":"Errata-ID","Errata Installation":"Installieren von Errata","Errata List":"Errata-Liste","Errata Task List":"Errata-Taskliste","Errata Tasks":"Errata-Tasks","Errata Type":"Errata-Typ","Errata for:":"Errata für:","Errata successfully added.":"Errata erfolgreich hinzugefügt.","Errata successfully removed.":"Errata erfolgreich entfernt.","Errata:":"Errata:","Erratum - by Date and Type":"Erratum – nach Datum und Typ","Erratum - by ID":"Erratum – nach ID","Erratum Date Range":"Erratum-Datumsbereich","Error deleting manifest.":"Fehler beim Löschen von Manifest.","Error during upload:":"Fehler beim Hochladen:","Error importing manifest.":"Fehler beim Importieren von Manifest.","Error refreshing manifest.":"Fehler beim Aktualisieren von Manifest.","Event":"Ereignis","Events":"Ereignisse","Exclude":"Ausschließen","Existing Product":"Vorhandenes Produkt","Expires":"Endet","Failed":"Fehlgeschlagen","File too large. Please use the CLI instead.":"Datei zu groß. Bitte stattdessen die Befehlszeilenschnittstelle verwenden.","Filter":"Filter","Filter Saved":"Filter gespeichert","Filter information for:":"Informationen filtern für:","Filter...":"Filtern …","Filter:":"Filter:","Filters":"Filter","Filters List":"Filterliste","Filters for Content View:":"Filter für Inhaltsansicht:","Filters successfully removed.":"Filter erfolgreich entfernt.","Finished At":"Abgeschlossen am","For older operating systems such as Red Hat Enterprise Linux 5 or CentOS 5 it is recommended to use sha1.":"Für ältere Betriebssysteme wie Red Hat Enterprise Linux 5 oder CentOS 5 wird die Verwendung von sha1 empfohlen.","GPG Key":"GPG-Schlüssel","GPG Key URL":"GPG-Schlüssel-URL","GPG Key successfully uploaded":"GPG-Schlüssel erfolgreich hochgeladen","GPG Key {{ gpgKey.name }}":"GPG-Schlüssel {{ gpgKey.name }}","GPG Keys":"GPG-Schlüssel","General Information":"Allgemeine Informationen","Generated On":"Generiert am","Gpg Key updated":"GPG-Schlüssel aktualisiert","Gpg Keys":"GPG-Schlüssel","Greater Than":"Größer als","Group Install":"Gruppeninstallation","Group Remove":"Gruppenentfernung","Guests of":"Gäste von","Hide affected Activation Keys":"Betroffene Aktivierungsschlüssel ausblenden","Hide affected Content Hosts":"Betroffene Inhaltshosts ausblenden","History":"Chronik","History for Content View:":"Chronik für Inhaltsansicht:","Host Collection Management":"Verwaltung von Hostsammlungen","Host Collection Membership":"Mitgliedschaft in Hostsammlung","Host Collection Membership Management":"Verwaltung der Mitgliedschaft in Hostsammlung","Host Collection removed.":"Hostsammlung entfernt","Host Collection updated":"Hostsammlung aktualisiert","Host Collection:":"Hostsammlung:","Host Collection: {{ hostCollection.name }}":"Hostsammlung: {{ hostCollection.name }}","Host Collections":"Hostsammlungen","Host Collections for:":"Hostsammlungen für:","Host Count":"Hostanzahl","Host Group":"Hostgruppe","Hostname":"Hostname","Hourly at {{ product.sync_plan.sync_date | date:'m' }} minutes and {{ product.sync_plan.sync_date | date:'s' }} seconds":"Stündlich um {{ product.sync_plan.sync_date | date:'m' }} Minuten und {{ product.sync_plan.sync_date | date:'s' }} Sekunden","IPv4 Address":"IPv4-Adresse","IPv6 Address":"IPv6-Adresse","Id":"ID","If you would prefer to move portions of these Activation Keys to different Content Views or Lifecycle Environments click <a href=\"{{ activationKeyLink() }}\">here</a> to manage them individually.":"Falls Sie Teile dieser Aktivierungsschlüssel lieber in andere Inhaltsansichten oder Lebenszyklusumgebungen verlegen möchten, klicken Sie <a href=\"{{ activationKeyLink() }}\">hier</a>, um sie einzeln zu verwalten.","If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click":"Falls Sie Teile dieser Aktivierungsschlüssel lieber in andere Inhaltsansichten oder Lebenszyklusumgebungen verlegen möchten, klicken Sie","If you would prefer to move portions of these Content Hosts to different content views or environments click <a href=\"{{ contentHostsLink() }}\">here</a>  to manage these Content Hosts in bulk.":"Falls Sie Teile dieser Inhaltshosts lieber in andere Inhaltsansichten oder Umgebungen verlegen möchten, klicken Sie <a href=\"{{ contentHostsLink() }}\">hier</a>, um diese Inhaltshosts mittels Massenaktionen zu verwalten.","Image":"Image","Image ID":"Image-ID","Import History":"Importchronik","Important":"Wichtig","In order to add a content view to a composite view you must first publish an initial version of the content view.":"Um eine Inhaltsansicht zu einer Verbundansicht hinzuzufügen, müssen Sie zunächst eine anfängliche Version der Inhaltsansicht veröffentlichen.","Include":"Einschließen","Included in Activation Keys":"In Aktivierungsschlüsseln enthalten","Incremental Update":"Inkrementelle Aktualisierung","Install":"Installieren","Install Selected":"Auswahl installieren","Install katello-agent for remote actions and displaying errata information:":"Installieren Sie katello-agent zum Durchführen von entfernten Aktionen und zum Anzeigen von Errata-Informationen:","Install the pre-built bootstrap RPM:":"Installieren Sie das integrierte Bootstrap-RPM:","Installable":"Installierbar","Installable Errata":"Installierbare Errata","Installed":"Installiert","Installed Package":"Installiertes Paket","Installed Packages":"Installierte Pakete","Installed Products":"Installierte Produkte","Instance-based":"Instanzbasiert","Interfaces":"Schnittstellen","Interval":"Intervall","Issued":"Ausgegeben","Katello Agent":"Katello-Agents","Label":"Kennung","Last Checkin":"Letzte Anmeldung","Last Published":"Letzte Veröffentlichung","Last Puppet Report":"Letzter Puppet-Bericht","Last Sync":"Letzte Synchronisation","Last Updated":"Letzte Aktualisierung","Last Updated On":"Letzte Aktualisierung:","Last synced {{ product.last_sync_words }} ago.":"Letzte Synchronisation vor {{ product.last_sync_words }}.","Latest (Currently %s)":"Aktuellste (derzeit %s)","Less Than":"Kleiner als","Library":"Library","Library Synced Content":"Von Library synchronisierte Inhalte","Lifecycle Environment \"{{ priorEnvironment.name }}\" exists within the middle of a path. New Lifecycle Environments can only be added to the end of a path.":"Lebenszyklusumgebung \"{{ priorEnvironment.name }}\" befindet sich in der Mitte eines Übertragungspfads. Neue Lebenszyklusumgebungen können nur am Ende eines Pfads hinzugefügt werden.","Lifecycle Environment Paths":"Lebenszyklus-Umgebungspfade","Lifecycle Environments":"Lebenszyklusumgebungen","Limit":"Grenze","Limit:":"Grenze:","Limits":"Grenzen","List/Remove":"Auflisten/Entfernen","Loading Environment Paths...":"Umgebungspfade werden geladen …","Loading...":"Laden …","Manage Docker Images":"Docker-Images verwalten","Manage Docker Images for Repository:":"Docker-Images für Repository verwalten:","Manage Manifest":"Manifest verwalten","Manage Packages":"Pakete verwalten","Manage Packages for Repository:":"Pakete für Repository verwalten:","Manage Puppet Modules":"Puppet-Module verwalten","Manage Puppet Modules for Repository:":"Puppet-Module für Repository verwalten:","Manifest History":"Manifestchronik","Manifest successfully deleted.":"Manifest erfolgreich gelöscht.","Manifest successfully imported.":"Manifest erfolgreich importiert.","Manifest successfully refreshed.":"Manifest erfolgreich aktualisiert.","Message":"Nachricht","Model":"Modell","Moderate":"Moderat (Mittel)","Module %s removed from Content View.":"Module %s aus Inhaltsansicht entfernt.","Module List":"Modulliste","More Details":"Mehr Details","More...":"Mehr …","Multi-entitlement":"Multi-Berechtigung","N/A":"N/A","Name":"Name","Name of the upstream repository you want to sync. Example: 'busybox' or 'fedora/ssh'.":"Name des Upstream-Repositorys, das Sie synchronisieren möchten. Beispiel: \"busybox\" oder \"fedora/ssh\".","Networking":"Netzwerk","Never synced":"Nie synchronisiert","New Activation Key":"Neuer Aktivierungsschlüssel","New Content View":"Neue Inhaltsansicht","New Environment":"Neue Umgebung","New Environment Path":"Neuer Umgebungspfad","New Filter":"Neuer Filter","New Filter for Content View:":"Neuer Filter für Inhaltsansicht:","New GPG Key":"Neuer GPG-Schlüssel","New Gpg Key":"Neuer GPG-Schlüssel","New Host Collection":"Neue Hostsammlung","New Name:":"Neuer Name:","New Product":"Neues Produkt","New Repository":"Neues Repository","New Sync Plan":"Neuer Synchronisationsplan","New sync plan successfully created.":"Neuer Synchronisationsplan erfolgreich erstellt.","Next":"Weiter","Next Sync":"Nächste Synchronisation","No":"Nein","No \"Edit\" permissions on some of the Activation Keys in that Lifecycle Environment.":"Keine Berechtigung zum Bearbeiten auf einigen der Aktivierungsschlüssel in dieser Lebenszyklusumgebung.","No \"Edit\" permissions on some of the Content Hosts in that Lifecycle Environment.":"Keine Berechtigung zum Bearbeiten auf einigen der Inhaltshosts in dieser Lebenszyklusumgebung.","No \"Promote/Remove\" permissions on the Lifecycle Environment.":"Keine Berechtigung zum Übertragen/Entfernen in dieser Lebenszyklusumgebung.","No (Default)":"Nein (Standard)","No Content Hosts are affected.":"Keine Inhaltshosts sind betroffen.","No Content Hosts match this Erratum.":"Keine Inhaltshosts entsprechen diesem Erratum.","No Content Views available, please select another environment.":"Keine Inhaltshosts verfügbar, bitte wählen Sie eine andere Umgebung.","No Host Collections to show, you can add Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"Keine Hostsammlungen zum Anzeigen. Sie können Hostsammlungen hinzufügen, wenn Sie im Hauptmenü \"Hosts\" auf \"Hostsammlungen\" klicken.","No Host Collections to show, you can add Host Collections after selecting the 'Add' tab.":"Keine Hostsammlungen zum Anzeigen. Sie können Hostsammlungen hinzufügen, wenn Sie auf den Reiter \"Hinzufügen\" klicken.","No Products to show.":"Keine Produkte zum Anzeigen.","No Repositories contain this Erratum.":"Keine Repositorys enthalten dieses Erratum.","No activation keys are affected.":"Keine Aktivierungsschlüssel sind betroffen.","No alternate release version choices are available. The available releases are based upon what is available in \"{{ contentHost.content_view.name }}\", the selected <a href=\"/content_views\">content view</a> this content host is attached to for the given <a href=\"/lifecycle_environments\">lifecycle environment</a>, \"{{ contentHost.environment.name }}\".":"Es stehen keine alternativen Release-Versionen zur Verfügung. Die verfügbaren Releases basieren auf \"{{ contentHost.content_view.name }}\", der ausgewählten <a href=\"/content_views\">Inhaltsansicht</a>, die mit diesem Inhaltshost verknüpft ist für die angegebene <a href=\"/lifecycle_environments\">Lebenszyklusumgebung</a>, \"{{ contentHost.environment.name }}\".","No content views exist for {{selected.environment.name}}":"Keine Inhaltsansichten existieren für {{selected.environment.name}}","No host collections":"Keine Hostsammlungen","No provisioning details are available for this content host.":"Für diesen Inhaltshost stehen keine Bereitstellungsdetails zur Verfügung.","No repository content provided through subscriptions.":"Keine Repository-Inhalte von Subskriptionen bereitgestellt.","No repository content provided through this subscription.":"Keine Repository-Inhalte von dieser Subskription bereitgestellt.","No repository content.":"Kein Repository-Inhalt.","No subscription manifest imported":"Kein Subskriptionsmanifest importiert","None":"Keine","Not Synced":"Nicht synchronisiert","Not installed":"Nicht installiert","Not yet published":"Noch nicht veröffentlicht","Number of CPUs":"Anzahl der CPUs","OS":"Betriebssystem","Only show Errata that is Applicable to one or more Content Hosts":"Nur Errata zeigen, die einen oder mehrere Inhaltshosts betreffen","Only show Errata that is Installable on one or more Content Hosts":"Nur Errata zeigen, die auf einem oder mehreren Inhaltshosts installierbar sind","Operating System":"Betriebssystem","Options:":"Optionen:","Organization":"Organisation","Original Sync Date":"Erste Synchronisation","Override to No":"Mit \"Nein\" überschreiben","Override to Yes":"Mit \"Ja\" überschreiben","Package":"Paket","Package Actions":"Paketaktionen","Package Filter:":"Paketfilter:","Package Group":"Paketgruppe","Package Group Filter:":"Paketgruppenfilter:","Package Group successfully added.":"Paketgruppe erfolgreich hinzugefügt.","Package Group successfully removed.":"Paketgruppe erfolgreich entfernt.","Package Groups":"Paketgruppen","Package Install":"Paketinstallation","Package Installation, Removal, and Update":"Installieren, Entfernen und Aktualisieren von Paketen","Package Remove":"Paketentfernung","Package Update":"Paketaktualisierung","Package removal failed":"Paketentfernung fehlgeschlagen","Package successfully added.":"Paket erfolgreich hinzugefügt.","Package successfully removed.":"Paket erfolgreich entfernt.","Package successfully updated.":"Paket erfolgreich aktualisiert.","Package/Group Name":"Paket-/Gruppenname","Packages":"Pakete","Packages <div>{{ library.counts.packages || 0 }}</div>":"Pakete <div>{{ library.counts.packages || 0 }}</div>","Packages for:":"Pakete für:","Packages in {{ repository.name }}":"Pakete in {{ repository.name }}","Parameters":"Parameter","Paste GPG Key Contents":"GPG-Schlüsselinhalte einfügen","Paste contents...":"Inhalte hier einfügen …","Paused":"Pausiert","Perform":"Ausführen","Physical":"Physisch","Please select a Lifecycle Environment and Content View to move the affected Activation Keys to:":"Bitte wählen Sie eine Lebenszyklusumgebung und Inhaltsansicht, in die diese betroffenen Aktivierungsschlüssel verlegt werden sollen:","Please select a Lifecycle Environment and Content View to move these Content Hosts to:":"Bitte wählen Sie eine Lebenszyklusumgebung und Inhaltsansicht, in die diese Inhaltshosts verlegt werden sollen:","Please select a Lifecycle Environment.":"Bitte wählen Sie eine Lebenszyklusumgebung.","Please select an environment.":"Bitte wählen Sie eine Umgebung.","Please select one from the list below and you will be redirected.":"Bitte wählen Sie aus der Liste unten, Sie werden anschließend weitergeleitet.","Press 'Sync Now' below to begin syncing the following product(s).":"Klicken Sie auf \"Jetzt synchronisieren\", um die Synchronisation der folgenden Produkte zu beginnen.","Previous Environment (%e/%cv)":"Vorherige Umgebung (%e/%cv)","Processing":"Verarbeiten","Product":"Produkt","Product Bulk Actions":"Produkt-Massenaktionen","Product Content":"Produktinhalt","Product Count":"Produktanzahl","Product Enhancement Advisory":"Produkterweiterungs-Advisory","Product ID":"Produkt-ID","Product Management":"Produktverwaltung","Product Management for Sync Plan:":"Produktverwaltung für Synchronisationsplan:","Product Name":"Produktname","Product Saved":"Produkt gespeichert","Product Sync":"Produktsynchronisation","Product delete operation has been initiated in the background.  Click <a href=\"{{ taskUrl(productDeletionTaskId) }}\">Here</a> to monitor the progress.":"Produktlöschoperation wurde im Hintergrund initiiert. Klicken Sie <a href=\"{{ taskUrl(productDeletionTaskId) }}\">hier</a>, um den Fortschritt nachzuverfolgen.","Product information for:":"Produktinformationen für:","Product {{ product.name }}":"Produkt {{ product.name }}","Products":"Produkte","Products <div>{{ library.counts.products || 0 }}</div>":"Produkte <div>{{ library.counts.products || 0 }}</div>","Promote":"Übertragen","Promote Content View:":"Inhaltsansicht übertragen:","Promote Version":"Version übertragen","Promote Version {{ version.version }}":"Version {{ version.version }} übertragen","Promoted to %s":"Übertragen nach %s","Promoting to %count environments.":"Übertragen in %count Umgebungen.","Promoting to 1 environment.":"Übertragen in 1 Umgebung.","Promotion History":"Übertragungschronik","Provided Products":"Bereitgestellte Produkte","Provisioning Details":"Bereitstellungsdetails","Provisioning Host Details":"Bereitstellungshost-Details","Publish Content View:":"Inhaltsansicht veröffentlichen:","Publish New Version":"Neue Version veröffentlichen","Publish via HTTP":"Über HTTP veröffentlichen","Publish via HTTPS":"Über HTTPS veröffentlichen","Published":"Veröffentlicht","Published At":"Veröffentlicht unter","Published Repository Information":"Repository-Informationen veröffentlicht","Published new version":"Neue Version veröffentlicht","Publishing and promoting to %count environments.":"Veröffentlichen und übertragen in %count Umgebungen.","Publishing and promoting to 1 environment.":"Veröffentlichen und übertragen in 1 Umgebung.","Puppet Environment":"Puppet-Umgebung","Puppet Modules":"Puppet-Module","Puppet Modules <div>{{ library.counts.puppet_modules || 0 }}</div>":"Puppet-Module <div>{{ library.counts.puppet_modules || 0 }}</div>","Puppet Modules for Content View:":"Puppet-Module für Inhaltsansicht:","Puppet Modules in {{ repository.name }}":"Puppet-Module in {{ repository.name }}","Puppet module added to Content View":"Puppet-Modul zur Inhaltsansicht hinzugefügt","Puppet module metadata generation has been initiated in the background.\n        Click <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Die Generierung von Puppet-Modul-Metadaten wurde im Hintergrund initiiert.\n        Klicken Sie <a ng-href=\"{{ taskUrl() }}\">hier</a>, um den Fortschritt nachzuverfolgen.","Quantity":"Anzahl","RAM (GB)":"RAM (GB)","RAM: %s GB":"RAM: %s GB","Range":"Bereich","Reboot Suggested":"Neustart empfohlen","Reboot Suggested?":"Neustart empfohlen?","Red Hat CDN URL":"Red Hat CDN-URL","Red Hat Provider Details":"Red Hat Anbieterdetails","Red Hat Repositories page":"Red Hat Repositorys","Red Hat Repositories page.":"Red Hat Repositorys-Seite.","Refresh Manifest":"Manifest aktualisieren","Refresh Table":"Tabelle aktualisieren","Refreshing Manifest...":"Manifest wird aktualisiert …","Refreshing manifest not suppported by this manifest.":"Manifestaktualisierung wird von diesem Manifest nicht unterstützt.","Register Content Host":"Inhaltshosts registrieren","Register a Content Host":"Inhaltshost registrieren","Registered":"Registriert","Registered By":"Registriert durch","Registry URL":"Registry-URL","Release":"Release","Release Version":"Release-Version","Remove":"Entfernen","Remove Activation Key \"{{ activationKey.name }}\"?":"Aktivierungsschlüssel \"{{ activationKey.name }}\" entfernen?","Remove Content View":"Inhaltsansicht entfernen","Remove Content Views":"Inhaltsansichten entfernen","Remove Docker Images":"Docker-Images entfernen","Remove Errata":"Errata entfernen","Remove Filters":"Filter entfernen","Remove From":"Entfernen von","Remove GPG Key":"GPG-Schlüssel entfernen","Remove GPG Key {{ gpgKey.name }}":"GPG-Schlüssel {{ gpgKey.name }} entfernen","Remove Host Collection \"{{ hostCollection.name }}\"?":"Host-Sammlung \"{{ hostCollection.name }}\" entfernen?","Remove Module":"Modul entfernen","Remove Package Group":"Paketgruppe entfernen","Remove Packages":"Pakete entfernen","Remove Product":"Produkt entfernen","Remove Product \"{{ product.name }}\"?":"Produkt \"{{ product.name }}\" entfernen?","Remove Products":"Produkte entfernen","Remove Puppet Modules":"Puppet-Module entfernen","Remove Repositories":"Repositorys entfernen","Remove Repository":"Repository entfernen","Remove Repository \"{{ repository.name }}\"?":"Repository \"{{ repository.name }}\" entfernen?","Remove Selected":"Ausgewählte entfernen","Remove Successful.":"Erfolgreich entfernt.","Remove Sync Plan":"Synchronisationsplan entfernen","Remove Sync Plan \"{{ syncPlan.name }}\"?":"Synchronisationsplan \"{{ syncPlan.name }}\" entfernen?","Remove Version":"Version entfernen","Remove Version Confirmation":"Entfernen der Version bestätigen","Remove View":"Ansicht entfernen","Remove {{ productTable.getSelected().length }} Products?":"{{ productTable.getSelected().length }} Produkte entfernen?","Removed %x host collections from activation key \"%y\".":"%x Hostsammlungen wurden vom Aktivierungsschlüssel \"%y\" entfernt.","Removed %x host collections from content host \"%y\".":"%x Hostsammlungen wurden vom Inhaltshost \"%y\" entfernt.","Removed %x products from sync plan \"%y\".":"%x Produkte wurden vom Synchronisationsplan \"%y\" entfernt.","Removing Content Hosts":"Inhaltshosts werden entfernt","Removing Products":"Produkte werden entfernt","Removing Repositories":"Repositorys werden entfernt","Repo Discovery":"Repository-Suche","Repo Type":"Repository-Typ","Repositories":"Repositorys","Repositories containing Errata {{ errata.errata_id }}":"Repositorys mit Errata {{ errata.errata_id }}","Repositories for Content View:":"Repositorys für Inhaltsansicht:","Repositories for Errata:":"Repositorys für Errata:","Repositories for Filter:":"Repositorys für Filter:","Repositories for Product:":"Repositorys für Produkt:","Repository":"Repository","Repository \"%s\" successfully deleted":"Repository \"%s\" erfolgreich gelöscht","Repository %s successfully created.":"Repository %s erfolgreich erstellt.","Repository Count":"Repository-Anzahl","Repository Discovery":"Repository-Suche","Repository Name":"Repository-Name","Repository Saved.":"Repository gespeichert.","Repository Selection":"Repository-Auswahl","Repository URL updated":"Repository-URL aktualisiert","Repository:":"Repository:","Required Host":"Erforderlicher Host","Result":"Ergebnis","Results":"Ergebnisse","Rules Source":"Regelquelle","Rules Version":"Regelversion","Run Auto-Attach":"Automatische Verknüpfung ausführen","Run subscription-manager in a console on the client host.  You may use an Activation Key to register:":"Führen Sie subscription-manager in einer Konsole auf dem Client-Host aus. Sie können einen Aktivierungsschlüssel zur Registrierung verwenden:","Save Successful.":"Erfolgreich gespeichert.","Search...":"Suchen …","Security":"Sicherheit","Security Advisory":"Sicherheits-Advisory","Select A New Puppet Module To Add":"Neues Puppet-Modul zum Hinzufügen wählen","Select Content Hosts":"Inhaltshosts auswählen","Select Module":"Modul wählen","Select Version":"Version wählen","Select a Content Source:":"Inhaltsquelle wählen:","Select a Version":"Version wählen","Select an Available Version of {{ $stateParams.moduleName }}":"Verfügbare Version von {{ $stateParams.moduleName }} wählen","Select an Organization":"Organisation wählen","Select new version":"Neue Version wählen","Select the Lifecycle Environments you would like to remove Version {{ version.version }} from:":"Wählen Sie die Lebenszyklusumgebungen, von denen Sie Version {{ version.version }} entfernen möchten:","Selected URLs:":"Gewählte URLs:","Selecting this option will result in the Version being completely deleted and no longer being available for promotion.  The version must be removed from all Lifecycle Environments in order to select this option.":"Wenn Sie diese Option wählen, wird die Version vollständig gelöscht und ist nicht länger zur Übertragung verfügbar. Die Version muss aus allen Lebenszyklusumgebungen entfernt werden, um diese Option zu wählen.","Serve via HTTP.":"Über HTTP bereitstellen.","Service Level":"Servicelevel","Severity":"Schweregrad","Show affected Activation Keys":"Betroffene Aktivierungsschlüssel anzeigen","Show affected Content Hosts":"Betroffene Inhaltshosts anzeigen","Show from:":"Anzeigen von:","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"{{ detailsTable.rows.length }} von {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} insgesamt) Inhaltshosts werden angezeigt","Showing {{ addSubscriptionsTable.rows.length }} of {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} Total) Subscriptions":"{{ addSubscriptionsTable.rows.length }} von {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} insgesamt) Subskriptionen werden angezeigt","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"{{ detailsTable.rows.length }} von {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} insgesamt) Inhaltshosts werden angezeigt","Showing {{ currentPackagesTable.resource.total }} Packages":"{{ currentPackagesTable.resource.total }} Pakete werden angezeigt","Showing {{ subscriptionsTable.rows.length }} of {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} Total) Subscriptions":"{{ subscriptionsTable.rows.length }} von {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} insgesamt) Subskriptionen werden angezeigt","Sockets":"Sockets","Sockets: %s":"Sockets: %s","Solution":"Lösung","Some of the Errata shown below may not be installable as they are not in this Content Host's\n          Content View and Lifecycle Environment.  In order to apply such Errata an Incremental Update is required.":"Einige der unten angezeigten Errata sind ggf. nicht installierbar, da sie sich nicht in der Inhaltsansicht und\n          Lebenszyklusumgebung dieses Hosts befinden. Um diese Errata zu installieren, ist eine inkrementelle Aktualisierung erforderlich.","Stacking ID":"Bündelungs-ID","Start Date":"Startdatum","Start Time":"Startzeit","Started At":"Gestartet am","Starting":"Start","Starts":"Beginnt","State":"Status","Status":"Status","Subscription Details":"Subskriptionsdetails","Subscription Info":"Subskriptionsinformationen","Subscription Manifest":"Subskriptionsmanifest","Subscription Server Details":"Subskriptions-Serverdetails","Subscription Status":"Subskriptionsstatus","Subscriptions":"Subskriptionen","Subscriptions for Activation Key:":"Subskriptionen für Aktivierungsschlüssel:","Subscriptions for:":"Subskriptionen für:","Succesfully scheduled package installation":"Paketinstallation erfolgreich eingeplant","Succesfully scheduled package removal":"Paketentfernung erfolgreich eingeplant","Succesfully scheduled package update":"Paketaktualisierung erfolgreich eingeplant","Success":"Erfolgreich","Successfully Scheduled Auto-attach.":"Automatische Verknüpfung erfolgreich eingeplant.","Successfully added %s subscriptions.":"%s Subskriptionen erfolgreich hinzugefügt.","Successfully deleted %cv version %ver.":"%cv Version %ver erfolgreich gelöscht.","Successfully initiated deletion of %cv version %ver.":"Löschen von %cv Version %ver erfolgreich initiiert.","Successfully initiated promotion of %cv version %ver to %env.":"Übertragen von %cv Version %ver in %env erfolgreich initiiert.","Successfully initiated removal of %cv version %ver.":"Entfernen von %cv Version %ver erfolgreich initiiert.","Successfully promoted %cv version %ver to %env":"%cv Version %ver erfolgreich in %env übertragen","Successfully published %cv version %ver and promoted to Library":"%cv Version %ver erfolgreich veröffentlicht und in Library übertragen","Successfully removed %cv version %ver from environments: %env":"%cv Version %ver erfolgreich aus Umgebungen %env entfernt","Successfully removed %s items.":"%s Elemente erfolgreich entfernt.","Successfully removed %s subscriptions.":"%s Subskriptionen erfolgreich entfernt.","Successfully removed 1 item.":"1 Element erfolgreich entfernt.","Successfully updated subscriptions.":"Subskriptionen erfolgreich aktualisiert.","Summary":"Zusammenfassung","Support Level":"Supportlevel","Support Type":"Supporttyp","Sync Enabled":"Synchronisation aktiviert","Sync Information":"Synchronisationsinformationen","Sync Interval":"Synchronisationsintervall","Sync Now":"Jetzt synchronisieren","Sync Plan":"Synchronisationsplan","Sync Plan %s has been deleted.":"Synchronisationsplan %s wurde gelöscht.","Sync Plan Saved":"Synchronisationsplan gespeichert","Sync Plan {{ syncPlan.name }}":"Synchronisationsplan {{ syncPlan.name }}","Sync Plan:":"Synchronisationsplan:","Sync Plans":"Synchronisationspläne","Sync State":"Syncronisationsstatus","Sync Status":"Synchronisationsstatus","Synced manually, no interval set.":"Manuell synchronisiert, kein Intervall festgelegt.","Syncing...":"Synchronisieren …","Tags":"Tags","Target":"Ziel","Task Details":"Taskdetails","Tasks":"Tasks","Tasks for Content View:":"Tasks für Inhaltsansicht:","Temporary":"Temporär","The Activation Keys listed below are currently using this Content View Version.  Before deleting the Version you must move these Activation Keys to a Lifecycle Environment where this Version is not in use.":"Die nachfolgend aufgeführten Aktivierungsschlüssel verwenden derzeit diese Inhaltsansichtsversion. Bevor Sie die Version löschen, müssen Sie diese Aktivierungsschlüssel in eine Lebenszyklusumgebung verlegen, in der diese Version nicht in Gebrauch ist.","The Content Hosts listed below are currently using this Content View Version.  Before removing the version you must move these Content Hosts to an environment where this version is not in use.":"Die nachfolgend aufgeführten Inhaltshosts verwenden derzeit diese Inhaltsansichtsversion. Bevor Sie die Version entfernen, müssen Sie diese Inhaltshosts in eine Umgebung verlegen, in der diese Version nicht in Gebrauch ist.","The Content View or Lifecycle Environment needs to be updated in order to make errata available to these hosts.":"Die Inhaltsansicht oder Lebenszyklusumgebung muss aktualisiert werden, um Errata für diese Hosts verfügbar zu machen.","The archive of this Version will be deleted.  This Version will not be available once deletion is complete.":"Das Archiv dieser Version wird gelöscht. Diese Version wird nicht mehr verfügbar sein, sobald der Löschvorgang abgeschlossen ist.","The archive of this Version will not be deleted.  This Version will still be available to be promoted to a Lifecycle Environment.":"Das Archiv dieser Version wird nicht gelöscht. Diese Version wird weiterhin zur Übertragung in eine Lebenszyklusumgebung verfügbar sein.","The following actions can be performed on content hosts in this host collection:":"Die folgenden Aktionen können auf Inhaltshosts in dieser Hostsammlung ausgeführt werden:","The host requires being attached to a content view and the environment you have chosen has no content views promoted to it.\n            See the <a href=\"/content_views\">content views page</a> to manage and promote a content view.":"Der Host muss mit einer Inhaltsansicht verknüpft sein; in die von Ihnen gewählte Umgebung wurden jedoch keine Inhaltsansichten übertragen.\n            Auf der Seite <a href=\"/content_views\">Inhaltsansichten</a> können Sie eine Inhaltsansicht verwalten und übertragen.","The katello-agent package is required to manage errata on this Host.\n      Please install the katello-agent package to manage errata on\n      {{ contentHost.name }}.":"Das Paket \"katello-agent\" ist erforderlich, um Errata auf diesem Host\n      zu verwalten. Bitte installieren Sie das Paket \"katello-agent\", \n      um Errata auf {{ contentHost.name }} zu verwalten.","The katello-agent package is required to manage packages on this Host.\n      Please install the katello-agent package to manage packages on\n      {{ contentHost.name }}. To install the package, connect directly to the\n      host from a terminal and use yum.":"Das Paket \"katello-agent\" ist erforderlich, um Pakete auf diesem Host zu verwalten.\n      Bitte installieren Sie das Paket \"katello-agent\", um Pakete auf\n      {{ contentHost.name }} zu verwalten. Um das Paket zu installieren, verbinden\n      Sie sich von einem Terminal direkt mit dem Host und verwenden Sie yum.","The page you are attempting to access requires selecting a specific organization.":"Die Seite, auf die Sie zugreifen möchten, erfordert die Auswahl einer bestimmten Organisation.","The selected environment contains no Content Views, please select a different environment.":"Die gewählte Umgebung einhält keine Inhaltsansichten, bitte wählen Sie eine andere Umgebung.","There are currently no repositories associated with this Content View, you can add some by clicking on the \"Add\" tab above.":"Mit dieser Inhaltsansicht sind derzeit keine Repositorys verknüpft. Sie können welche hinzufügen, indem Sie auf den Reiter \"Hinzufügen\" klicken.","There are currently no repositories to add to this Content View,":"Es gibt keine Repositorys zum Hinzufügen zu dieser Inhaltsansicht,","There are no Content Views in this Environment.":"Es gibt keine Inhaltsansichten in dieser Umgebung.","There are no Content Views that match the criteria.":"Es gibt keine Inhaltsansichten, die mit diesen Kriterien übereinstimmen.","There are no Docker Tags that match the criteria.":"Es gibt keine Docker-Tags, die mit diesen Kriterien übereinstimmen.","There are no Errata associated with this Content Host to display.":"Es gibt keine anzuzeigenden Errata, die mit diesem Inhaltshost verknüpft sind.","There are no Errata in this organization.  Create one or more Products with Errata to view Errata on this page.":"Es gibt keine Errata in dieser Organisation. Erstellen Sie ein oder mehrere Produkte mit Errata, um auf dieser Seite Errata anzuzeigen.","There are no Errata that match the criteria.":"Es gibt keine Errata, die mit diesen Kriterien übereinstimmen.","There are no Errata to display.":"Es gibt keine anzuzeigenden Errata.","There are no Host Collections available. You can create new Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"Es sind keine Hostsammlungen verfügbar. Sie können neue Hostsammlungen erstellen, wenn Sie im Hauptmenü \"Hosts\" auf \"Hostsammlungen\" klicken.","There are no Packages that match the criteria.":"Es gibt keine Pakete, die mit diesen Kriterien übereinstimmen.","There are no Puppet Modules that match the criteria.":"Es gibt keine Puppet-Module, die mit diesen Kriterien übereinstimmen.","There are no Yum Repositories that match the criteria.":"Es gibt keine Yum-Repositorys, die mit diesen Kriterien übereinstimmen.","There is currently an Incremental Update task in progress.  This update must finish before applying existing updates.":"Derzeit läuft eine inkrementelle Aktualisierung. Diese Aktualisierung muss abgeschlossen sein, bevor vorhandene Aktualisierungen angewendet werden können.","There were errors while removing the following Repositories:":"Beim Entfernen der folgenden Repositorys traten Fehler auf:","This Content Host is not attached to any subscriptions that provide content":"Dieser Inhaltshost ist mit keinen Subskriptionen verknüpft, die Inhalte bereitstellen","This Content View does not have any versions, create your first Content View Version by using the \"Publish New Version\" button on the right.":"Diese Inhaltsansicht hat keine Versionen. Erstellen Sie Ihre erste Inhaltsansichtsversion über die Schaltfläche \"Neue Version veröffentlichen\" rechts.","This Version cannot be deleted from some of the Lifecycle Environments due to one of the reasons below.":"Diese Version kann aufgrund einer der unten genannten Gründe nicht von allen Lebenszyklusumgebungen gelöscht werden","This Version is not associated with any Lifecycle Environments.":"Diese Version ist mit keinen Lebenszyklusumgebungen verknüpft.","This activation key is not associated with any content hosts.":"Dieser Aktivierungsschlüssel ist derzeit mit keinen Inhaltshosts verknüpft.","This activation key may be used during system registration. For example:":"Dieser Aktivierungsschlüssel kann während der Systemregistrierung verwendet werden. Zum Beispiel:","This filter applies only to a subset of repositories in the content view.":"Dieser Filter gilt nur für eine Untergruppe von Repositorys in der Inhaltsansicht.","This filter applies to all repositories in the content view (current and future).":"Dieser Filter gilt für alle Repositorys in der Inhaltsansicht (aktuell und zukünftig).","This subscription is not associated with any activation keys.":"Diese Subskription ist mit keinen Aktivierungsschlüsseln verknüpft.","This subscription is not attached to any content hosts.":"Diese Subskription ist mit keinen Inhaltshosts verknüpft.","This version will be removed from:":"Diese Version wird entfernt von:","Time":"Zeit","Title":"Titel","To register a content host to this server, follow these steps.":"Um einen Inhaltshosts bei diesem Server zu registrieren, gehen Sie folgendermaßen vor.","Topic":"Thema","Type":"Typ","URL":"URL","URL of the registry you want to sync. Example: https://registry.hub.docker.com":"URL der Registry, die Sie synchronisieren möchten. Beispiel: https://registry.hub.docker.com","URL to discover":"Such-URL","UUID":"UUID","Unattach Sync Plan":"Synchronisationsplan lösen","Unattaching...":"Lösen …","Unlimited Content Hosts:":"Unbegrenzte Inhaltshosts:","Unregister Content Host":"Inhaltshost abmelden","Unregister Content Host \"{{contentHost.name}}\"?":"Inhaltshost \"{{contentHost.name}}\" abmelden?","Unregister Content Hosts":"Inhaltshosts abmelden","Unregister {{ table.numSelected  }} Content Hosts?":"{{ table.numSelected  }} Inhaltshosts abmelden?","Update":"Aktualisieren","Update All":"Alle aktualisieren","Update Repositories":"Repositorys aktualisieren","Update Sync Plan":"Synchronisationsplan aktualisieren","Updated":"Aktualisiert","Updated errata filter -":"Errata-Filter aktualisiert -","Updated override to \"%x\".":"Überschreibung aktualisiert auf \"%x\".","Updating...":"Aktualisieren …","Upload":"Hochladen","Upload Docker Image":"Docker-Image hochladen","Upload GPG Key":"GPG-Schlüssel hochladen","Upload New Manifest":"Neues Manifest hochladen","Upload Package":"Paket hochladen","Upload Puppet Module":"Puppet-Modul hochladen","Uploading...":"Hochladen …","Upstream Repository Name":"Upstream-Repository-Name","Upstream Subscription":"Upstream-Subskription","Upstream Subscription Management Application":"Upstream-Subskriptionsverwaltungsapplikation","Upstream Type":"Upstream-Typ","Upstream User":"Upstream-Benutzer","Use Latest (currently %s)":"Aktuellste verwenden (derzeit %s)","Use the cancel button on content view selection to revert your environment selection.":"Klicken Sie auf \"Abbrechen\" in der Inhaltsansichtsauswahl, um Ihre Umgebungsauswahl zurückzusetzen.","User":"Benutzer","Version":"Version","Version %s":"Version %s","Version 1":"Version 1","Version Deletion":"Versionslöschung","Version Details":"Versionsdetails","Version {{ version.version }}":"Version {{ version.version }}","Version:":"Version:","Versions":"Versionen","Versions for Content View:":"Versionen für Inhaltsansicht:","Versions for Puppet Module:":"Versionen für Puppet-Modul:","View Details":"Ansichtsdetails","Virtual":"Virtuell","Virtual Guest From":"Virtueller Gast von","Virtual Guests":"Virtuelle Gäste","Virtual Host":"Virtuelle Hosts","Weekly on {{ product.sync_plan.sync_date | date:'EEEE' }} at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"Wöchentlich am {{ product.sync_plan.sync_date | date:'EEEE' }} um {{ product.sync_plan.sync_date | date:'mediumTime' }} (Serverzeit)","When Auto Attach is disabled, registering systems will be attached to all associated subscriptions.":"Wenn die automatische Verknüpfung deaktiviert ist, werden neu registrierte Systeme mit allen zugewiesenen Subskriptionen verknüpft.","Working":"In Arbeit","Yes":"Ja","Yes (Default)":"Ja (Standard)","You cannot remove these repositories because they belong to a Red Hat product.":"Sie können diese Repositorys nicht entfernen, da Sie zu einem Red Hat Produkt gehören.","You cannot remove these repositories because you do not have permission.":"Sie können diese Repositorys nicht entfernen, da Sie keine Berechtigung dazu haben.","You cannot remove this product because it is a Red Hat product.":"Sie können dieses Produkt nicht entfernen, da es ein Red Hat Produkt ist.","You cannot remove this product because it was published to a content view.":"Sie können dieses Produkt nicht entfernen, da es in einer Inhaltsansicht veröffentlicht wurde.","You cannot remove this product because you do not have permission.":"Sie können dieses Produkt nicht entfernen, da Sie keine Berechtigung dazu haben.","You cannot remove this repository because it is a Red Hat repository.":"Sie können dieses Repository nicht entfernen, da es ein Red Hat Repository ist.","You cannot remove this repository because it was published to a content view.":"Sie können dieses Repository nicht entfernen, da es in einer Inhaltsansicht veröffentlicht wurde.","You cannot remove this repository because you do not have permission.":"Sie können dieses Repository nicht entfernen, da Sie keine Berechtigung dazu haben.","You currently do not have any Packages installed. You can use Package Actions above this message to install new Packages.":"Sie haben derzeit keine Pakete installiert. Sie können die Paketaktionen über dieser Meldung verwenden, um neue Pakete zu installieren.","You currently don't have any Activation Keys, you can add Activation Keys using the button on the right.":"Sie haben derzeit keine Aktivierungsschlüssel. Sie können über die Schaltfläche rechts neue Aktivierungsschlüssel hinzufügen.","You currently don't have any Content Hosts in this Host Group, you can add Content Hosts after selecting the 'Add' tab.":"Sie haben derzeit keine Inhaltshosts in dieser Hostgruppe. Sie können Inhaltshosts hinzufügen, wenn Sie auf den Reiter \"Hinzufügen\" klicken.","You currently don't have any Content Hosts, you can create new Content Hosts by selecting Contents Host from main menu and then clicking the button on the right.":"Sie haben derzeit keine Inhaltshosts. Sie können neue Inhaltshosts erstellen, indem Sie im Hauptmenü den Punkt \"Inhaltshosts\" wählen und anschließend auf die Schaltfläche rechts klicken.","You currently don't have any Content Hosts, you can register one by clicking the button on the right and following the instructions.":"Sie haben derzeit keine Inhaltshosts. Sie können neue Inhaltshosts registrieren, indem Sie auf die Schaltfläche rechts klicken und den Anweisungen folgen.","You currently don't have any Content Views.  A Content View can be added by using the button on the right.":"Sie haben derzeit keine Inhaltsansichten. Sie können neue Inhaltsansichten hinzufügen, indem Sie auf die Schaltfläche rechts klicken.","You currently don't have any Docker Tags.":"Sie haben derzeit keine Docker-Tags.","You currently don't have any Filters included in this Content View, you can add a new Filter by using the button on the right.":"In dieser Inhaltsansicht sind derzeit keine Filter enthalten. Sie können über die Schaltfläche rechts einen neuen Filter hinzufügen.","You currently don't have any Gpg keys, you can add Gpg keys using the button on the right.":"Sie haben derzeit keine GPG-Schlüssel. Sie können über die Schaltfläche rechts neue GPG-Schlüssel hinzufügen.","You currently don't have any Host Collections, you can add Host Collections using the button on the right.":"Sie haben derzeit keine Hostsammlungen. Sie können über die Schaltfläche rechts neue Hostsammlungen hinzufügen.","You currently don't have any Products associated with this Gpg Key.":"Sie haben diesem GPG-Schlüssel derzeit keine Produkte zugewiesen.","You currently don't have any Products to subscribe to, you can add Products after selecting 'Products' under 'Content' in the main menu.":"Sie haben derzeit keine Produkte zum Subskribieren. Sie können Produkte hinzufügen, indem Sie im Hauptmenü den Punkt \"Inhalt\" und anschließend \"Produkte\" wählen.","You currently don't have any Products<span bst-feature-flag=\"custom_products\">, you can add Products using the button on the right</span>.":"Sie haben derzeit keine Produkte<span bst-feature-flag=\"custom_products\">. Sie können über die Schaltfläche rechts Produkte hinzufügen</span>.","You currently don't have any Puppet Modules included in this Content View, you can add Puppet Modules using the button on the right.":"In dieser Inhaltsansicht sind derzeit keine Puppet-Module enthalten. Sie können über die Schaltfläche rechts Puppet-Module hinzufügen.","You currently don't have any Repositories associated with this Gpg Key.":"Sie haben diesem GPG-Schlüssel derzeit keine Repositorys zugewiesen.","You currently don't have any Repositories included in this Product, you can add Repositories using the button on the right.":"In diesem Produkt sind derzeit keine Repositorys enthalten. Sie können über die Schaltfläche rechts Repositorys hinzufügen.","You currently don't have any Subscriptions associated with this Content Host, you can add Subscriptions after selecting the 'Add' tab.":"Mit diesem Inhaltshost sind derzeit keine Subskriptionen verknüpft. Sie können Subskriptionen hinzufügen, wenn Sie auf den Reiter \"Hinzufügen\" klicken.","You currently don't have any Subscriptions, you can add Subscriptions by importing a manifest using the button on the right labeled 'Manage Manifest'.":"Sie haben derzeit keine Subskriptionen. Sie können Subskriptionen hinzufügen, indem Sie ein Manifest importieren über die Schaltfläche rechts \"Manifest verwalten\".","You currently don't have any Sync Plan, you can add Sync Plans using the button on the right.":"Sie haben derzeit keinen Synchronisationsplan. Sie können über die Schaltfläche rechts neue Synchronisationspläne hinzufügen.","You currently don't have any Sync Plans.  A Sync Plan can be created by using the button on the right.":"Sie haben derzeit keine Synchronisationspläne. Sie können über die Schaltfläche rechts einen neuen Synchronisationsplan erstellen.","You do not have any Installed Products":"Sie haben keine installierten Produkte","You must select a content view in order to save your environment.":"Sie müssen eine Inhaltsansicht auswählen, um Ihre Umgebung zu speichern.","You must select a new content view before your change of environment can be saved.":"Sie müssen eine neue Inhaltsansicht wählen, bevor die Änderung der Umgebung gespeichert werden kann.","You must select a new content view before your change of environment can be saved. Use the cancel button on content view selection to revert your environment selection.":"Sie müssen eine neue Inhaltsansicht wählen, bevor die Änderung der Umgebung gespeichert werden kann. Klicken Sie auf \"Abbrechen\" in der Inhaltsansichtsauswahl, um Ihre Umgebungsauswahl zurückzusetzen.","You must select at least one Content Host in order to apply Errata.":"Sie müssen mindestens einen Inhaltshost auswählen, um Errata anzuwenden.","You must select at least one repository.":"Sie müssen mindestens ein Repository auswählen.","Your selected content hosts have changed, it is highly recommended to refresh the calculated list of errata using the 'Refresh Errata' button below.":"Ihre ausgewählten Inhaltshosts haben sich verändert. Es wird dringend empfohlen, die Liste der Errata zu aktualisieren. Klicken Sie dazu unten auf die Schaltfläche \"Errata aktualisieren\".","Yum Content":"Yum-Inhalte","Yum Metadata Checksum":"Yum-Metadaten-Prüfsumme","Yum Repo Discovery":"Yum-Repository-Suche","Yum Repositories":"Yum-Repositorys","Yum Repositories <div>{{ library.counts.yum_repositories || 0 }}</div>":"Yum-Repositorys <div>{{ library.counts.yum_repositories || 0 }}</div>","Yum metadata generation has been initiated in the background.  Click <a href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Die Generierung von Yum-Metadaten wurde im Hintergrund initiiert. Klicken Sie <a href=\"{{ taskUrl() }}\">hier</a>, um den Fortschritt nachzuverfolgen.","add some repositories.":"fügen Sie Repositorys hinzu.","daily":"täglich","here":"hier","hourly":"stündlich","or authenticate with a username and password:":"oder authentifizieren Sie sich mit einem Benutzernamen und Passwort:","subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"":"subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"","task details":"Taskdetails","to manage them individually.":"um sie einzeln zu verwalten.","true":"true","weekly":"wöchentlich","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }} ausgewählt","{{ addSubscriptionsTable.numSelected }} Selected":"{{ addSubscriptionsTable.numSelected }} ausgewählt","{{ contentHost.registered_by }}":"{{ contentHost.registered_by }}","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }} ausgewählt","{{ contentView.name }} cannot be deleted as one or more Content View Versions are still promoted to a Lifecycle Environment.\n      Each Content View Version must be removed from their Lifecycle Environments before the Content View can be deleted.":"{{ contentView.name }} kann nicht gelöscht werden, da eine oder mehrere Inhaltsansichtsversionen noch in einer Lebenszyklusumgebung enthalten sind.\n      Jede Inhaltsansichtsversion muss aus ihren Lebenszyklusumgebungen entfernt werden, bevor die Inhaltsansicht gelöscht werden kann.","{{ contentView.puppet_modules.length || 0 }} Puppet Modules":"{{ contentView.puppet_modules.length || 0 }} Puppet-Module","{{ contentView.repositories.length || 0 }} Repositories":"{{ contentView.repositories.length || 0 }} Repositorys","{{ contentViewVersion.puppet_module_count || 0 }} Puppet Modules":"{{ contentViewVersion.puppet_module_count || 0 }} Puppet-Module","{{ contentViewVersion.repositories.length || 0 }} Repositories":"{{ contentViewVersion.repositories.length || 0 }} Repositorys","{{ discoveryTable.numSelected }} Selected":"{{ discoveryTable.numSelected }} ausgewählt","{{ environment.name }} Environment":"{{ environment.name }} Umgebung","{{ errata.hosts_applicable_count || 0 }} Applicable,":"{{ errata.hosts_applicable_count || 0 }} relevant","{{ errata.hosts_available_count || 0 }} Installable":"{{ errata.hosts_available_count || 0 }} installierbar","{{ hostCollectionsTable.numSelected }} Selected":"{{ hostCollectionsTable.numSelected }} ausgewählt","{{ package.removeTask.affected_units }} package removed":"{{ package.removeTask.affected_units }} Pakete entfernt","{{ package.removeTask.affected_units }} packages removed":"{{ package.removeTask.affected_units }} Pakete entfernt","{{ product.last_sync_words }} Ago ({{ product.sync_status.finish_time | date:'medium' }} Local Time)":"vor {{ product.last_sync_words }} ({{ product.sync_status.finish_time | date:'medium' }} Ortszeit)","{{ product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ product.sync_plan.next_sync | date:'medium' }} (Serverzeit)","{{ productsTable.numSelected }} Selected":"{{ productsTable.numSelected }} ausgewählt","{{ repository.content_counts.docker_image || 0 }} Docker Images":"{{ repository.content_counts.docker_image || 0 }} Docker-Images","{{ repository.content_counts.docker_image }} Docker Images":"{{ repository.content_counts.docker_image }} Docker-Images","{{ repository.content_counts.docker_tag || 0 }} Docker Tags":"{{ repository.content_counts.docker_tag || 0 }} Docker-Tags","{{ repository.content_counts.docker_tag }} Docker Tags":"{{ repository.content_counts.docker_tag }} Docker-Tags","{{ repository.content_counts.erratum || 0 }} Errata":"{{ repository.content_counts.erratum || 0 }} Errata","{{ repository.content_counts.erratum }} Errata":"{{ repository.content_counts.erratum }} Errata","{{ repository.content_counts.puppet_module || 0 }} Puppet Modules":"{{ repository.content_counts.puppet_module || 0 }} Puppet-Module","{{ repository.content_counts.rpm || 0 }} Packages":"{{ repository.content_counts.rpm || 0 }} Pakete","{{ repository.content_counts.rpm }} Packages":"{{ repository.content_counts.rpm }} Pakete","{{ repository.last_sync_words }} Ago ({{ repository.last_sync.ended_at | date:'medium' }} Local Time)":"vor {{ repository.last_sync_words }} ({{ repository.last_sync.ended_at | date:'medium' }} Ortszeit)","{{ repository.last_sync_words }} ago":"vor {{ repository.last_sync_words }}","{{ repository.product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ repository.product.sync_plan.next_sync | date:'medium' }} (Serverzeit)","{{ subscription.host.name }}":"{{ subscription.host.name }}","{{ subscriptionsTable.numSelected }} Selected":"{{ subscriptionsTable.numSelected }} ausgewählt","{{ totalActivationKeyCount() }} Activation Keys will be moved to {{ deleteOptions.activationKeys.contentView.name }} in {{ deleteOptions.activationKeys.environment.name }}":"{{ totalActivationKeyCount() }} Aktivierungsschlüssel werden nach {{ deleteOptions.activationKeys.contentView.name }} in {{ deleteOptions.activationKeys.environment.name }} verlegt","{{ totalHostCount() }} Content Hosts will be moved to {{ deleteOptions.contentHosts.contentView.name }} in {{ deleteOptions.contentHosts.environment.name }}":"{{ totalHostCount() }} Inhaltshosts werden nach {{ deleteOptions.contentHosts.contentView.name }} in {{ deleteOptions.contentHosts.environment.name }} verlegt","{{ type.display }}":"{{ type.display }}","{{ version.docker_image_count }} Docker Images":"{{ version.docker_image_count }} Docker-Images","{{ version.docker_tag_count }} Docker Tags":"{{ version.docker_tag_count }} Docker-Tags","{{ version.errata_counts.total }} Errata":"{{ version.errata_counts.total }} Errata","{{ version.package_count }} Packages":"{{ version.package_count }} Pakete","{{ version.puppet_module_count }} Puppet Modules":"{{ version.puppet_module_count }} Puppet-Module"});
    gettextCatalog.setStrings('es', {"%(consumed)s out of %(quantity)s":"%(consumed)s de %(quantity)s","(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)":"(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)","(Not all Activation Keys editable )":"(No todas las llaves de activación son modificables)","(Not all Content Hosts editable )":"(No todos los hosts de contenido son modificables)","* These marked Content View Versions are from Composite Content Views.  Their components needing updating are listed underneath.":"*Estas versiones de vista de contenido pertenecen a  Vistas de contenido compuestas. Los componentes que deben actualizarse se listan abajo. ","+ New Sync Plan":"+ Nuevo plan de sincronización","- Date and Type":"- Fecha y tipo","-- select an interval --":"-- seleccionar un intervalo --","1 Content Host":["1 Host de contenido","{{ contentHost.virtual_guests.length }} Hosts de contenido"],"1 repository sync has errors.":["1 sincronización de repositorio tiene errores.","{{ product.sync_summary.error}} sincronizaciones de repositorios tienen errores."],"1 repository sync in progress.":["1 sincronización de repositorio en progreso.","{{ product.sync_summary.pending}} sincronizaciones de repositorios en progreso."],"1 successfully synced repository.":["1 repositorio ha sido sincronizado correctamente. ","{{ product.sync_summary.success}} repositorios han sido sincronizados correctamente."],"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Version {{ version.version }}</a>":"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Versión {{ version.version }}</a>","<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Create":"<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Crear","<span translate=\"\">\n        If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">here</a>\n      <span translate=\"\">to manage them individually.</span>":"<span translate=\"\">\n        Si prefiere mover porciones de llaves de activación a diferentes vistas de contenido o entorno de ciclo de vida haga clic en \n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">aquí</a>\n      <span translate=\"\">para administrarlas individualmente.</span>","A composite view contains other content views.":"Una vista compuesta contiene otras vistas de contenido.","A new version of {{ contentView.name }} will be created and promoted to the Library environment.\n    It can be promoted to other environments from the Versions tab of this Content View.":"Una nueva versión de {{ contentView.name }} será creada y promovida al entorno de biblioteca.\n    Puede promoverse  a otros entornos desde la pestaña Versiones de esta vista de contenido.","Account":"Cuenta","Account Number":"Número de cuenta","Action":"Acción","Action Type":"Tipo de acción","Actions":"Acciones","Actions for:":"Acciones para:","Activation Key":["Llave de activación","Llaves de activación"],"Activation Key Content":"Contenido de la llave de activación","Activation Key Type:":"Tipo de llaves de activación: ","Activation Key removed.":"Se ha eliminado la llave de activación.","Activation Key updated":"Se ha actualizado la llave de activación.","Activation Key:":"Llave de activación:","Activation Key: {{ activationKey.name }}":"Llave de activación: {{ activationKey.name }}","Activation Keys":"Llaves de activación","Activation Keys for Subscription:":"Llaves de activación para suscripción:","Activation Keys using Version {{ version.version }}":"Llaves de activación  con la  versión {{ version.version }}","Add":"Añadir","Add Content Hosts to:":"Añadir hosts de contenido a:","Add Content Views":"Añadir vistas de contenido","Add Errata":"Añadir erratas","Add New Environment":"Añada un nuevo entorno","Add New Filter":"Añadir nuevo filtro","Add New Module":"Añadir nuevo módulo","Add New Repository":"Añadir nuevo repositorio","Add Package Group":"Añadir grupo de paquetes","Add Repositories":"Añadir repositorios","Add Selected":"Añadir seleccionado","Add Subscriptions for Activation Key:":"Añadir suscripciones para la llave de activación:","Add To":"Añadir a","Added %x host collections to activation key \"%y\".":"Se han añadido %x colecciones de hosts a la llave de activación \"%y\".","Added %x host collections to content host \"%y\".":"Se han añadido %x colecciones de hosts al host de contenido \"%y\".","Added %x products to sync plan \"%y\".":"Se han añadido %x productos al plan de sincronización \"%y\".","Adding Lifecycle Environment to the end of \"{{ priorEnvironment.name }}\"":"Adición de entorno de ciclo de vida al final de \"{{ priorEnvironment.name }}\"","Advanced Information":"Información avanzada","Advisory":"Recomendación","Affected Activation Keys":"Llaves de activación afectadas","Affected Content Hosts":"Hosts de contenido afectados","Affected Packages":"Paquetes afectados","Affected Systems":"Sistemas afectados","Affected repositories":"Repositorios afectados","Affected repositories have been updated.":"Se han actualizado los repositorios afectados.","All Content Views":"Todas las vistas de contenido","All Products":"Todos los productos","All Repositories":"Todos los repositorios","All Versions":"Todas las versiones","Alter Sync Plans":"Modificar planes de sincronización","An error occurred applying Subscriptions:":"Ocurrió un error al aplicar las suscripciones:","An error occurred removing the Activation Key:":"Ocurrió un error al eliminar la llave de activación:","An error occurred removing the Host Collection:":"Ocurrió un error al eliminar la colección de hosts:","An error occurred removing the Products:":"Ocurrió un error al eliminar los productos:","An error occurred removing the content hosts.":"Ocurrió un error al eliminar los hosts de contenido.","An error occurred removing the environment:":"Ocurrió un error al eliminar el entorno:","An error occurred removing the subscriptions.":"Ocurrió un error al eliminar las suscripciones.","An error occurred saving the Activation Key:":"Ocurrió un error al guardar la llave de activación:","An error occurred saving the Content Host:":"Ocurrió un error al guardar el host de contenido:","An error occurred saving the Environment:":"Ocurrió un error al guardar el entorno:","An error occurred saving the Filter:":"Ocurrió un error al guardar el filtro:","An error occurred saving the Host Collection:":"Ocurrió un error al guardar la colección de hosts:","An error occurred saving the Product:":"Ocurrió un error al guardar el producto:","An error occurred saving the Repository:":"Ocurrió un error al guardar el repositorio:","An error occurred saving the Sync Plan:":"Ocurrió un error al guardar el plan de sincronización:","An error occurred saving the URL:":"Ocurrió un error al guardar la URL:","An error occurred trying to auto-attach subscriptions.  Please check your log for further information.":"Ocurrió un error al intentar asociar automáticamente las suscripciones.  Revise el registro para obtener más información.","An error occurred updating the Content View:":"Ocurrió un error al actualizar la vista de contenido:","Applicable":"Aplicable","Applicable Content Hosts for Errata":"Hosts de contenido aplicables para las erratas","Applicable Errata":"Erratas aplicables","Applicable hosts for {{ erratum.errata_id }}":"Hosts aplicables para {{ erratum.errata_id }}","Apply":"Aplicar","Apply Errata":"Aplicar erratas","Apply Errata to Content Host \"{{contentHost.name}}\"?":"¿Aplicar erratas a host de contenido \"{{contentHost.name}}\"?","Apply Errata to Content Hosts":"Aplicar erratas a hosts de contenido","Apply Errata to Content Hosts immediately after publishing.":"Aplicar erratas a hosts de contenido instantes después de la publicación.","Apply Selected":"Aplicar seleccionado","Apply to Content Hosts":"Aplicar  a hosts de contenido","Apply to Hosts":"Aplicar  a hosts ","Apply {{ errata.errata_id }}":"Aplicar  {{ errata.errata_id }}","Apply {{ errata.errata_id }} to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"¿Aplicar {{ errata.errata_id }} a {{ numberOfContentHostsToUpdate }} host(s) de contenido?","Apply {{ errataIds.length }} Errata to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"¿Aplicar {{ errataIds.length }} erratas a { {{ numberOfContentHostsToUpdate }} host(s) de contenido?","Applying":"Aplicación","Applying Bulk Errata Update":"Aplicación de actualizaciones de erratas en conjunto","Arch":"Arquitectura","Architecture(s)":"Arquitectura(s)","Archived Copy":"Copia archivada","Are you sure you want to add the {{ table.numSelected }} content host(s) selected to the host collection(s) chosen?":"¿Está seguro de que desea añadir los {{ table.numSelected }} hosts de contenido seleccionados a las colecciones de hosts elegidas?","Are you sure you want to apply Errata to content host \"{{ contentHost.name }}\"?":"¿Está seguro de que desea aplicar erratas a hosts de contenido \"{{ contentHost.name }}\"?","Are you sure you want to apply the {{ detailsTable.numSelected }} selected errata to the content hosts chosen?":"¿Está seguro de que desea aplicar las {{ detailsTable.numSelected }} erratas seleccionadas a los hosts de contenido elegidos?","Are you sure you want to assign the {{ table.numSelected }} content host(s) selected to {{ selected.contentView.name }} in {{ selected.environment.name }}?":"¿Está seguro de que desea asignar los {{ table.numSelected }} hosts de contenido seleccionados a {{ selected.contentView.name }} en {{ selected.environment.name }}?","Are you sure you want to auto-attach available subscriptions to all {{ table.resource.total }} content hosts?":"¿Está seguro de que desea asociar automáticamente las suscripciones disponibles a todos los {{ table.resource.total }} hosts de contenido?","Are you sure you want to delete Content View \"{{ contentView.name }}\"?":"¿Está seguro de que desea borrar la vista de contenido \"{{ contentView.name }}\"?","Are you sure you want to install {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"¿Está seguro de que desea instalar {{ content.content }} en los {{ getSelectedSystemIds().length }} sistemas seleccionados?","Are you sure you want to remove Activation Key \"{{ activationKey.name }}\"?":"¿Está seguro de que desea eliminar la llave de activación  \"{{ activationKey.name }}\"?","Are you sure you want to remove GPG key {{ gpgKey.name }}?":"¿Está seguro de que desea eliminar la llave GPG  {{ gpgKey.name }}?","Are you sure you want to remove Host Collection \"{{ hostCollection.name }}\"?":"¿Está seguro de que desea retirar esta colección de hosts \"{{ hostCollection.name }}\"?","Are you sure you want to remove Sync Plan \"{{ syncPlan.name }}\"?":"¿Está seguro de que desea retirar el plan de sincronización \"{{ syncPlan.name }}\"?","Are you sure you want to remove product \"{{ product.name }}\"?":"¿Está seguro de desear retirar el producto \"{{ product.name }}\"?","Are you sure you want to remove repository \"{{ repository.name }}\"?":"¿Está seguro de desear retirar el repositorio  \"{{ repository.name }}\"?","Are you sure you want to remove the {{ productTable.getSelected().length }} products(s) selected?":"¿Está seguro de que desea eliminar  {{ productTable.getSelected().length }} producto(s) seleccionados?","Are you sure you want to remove the {{ table.numSelected }} content host(s) selected from the host collection(s) chosen?":"¿Está seguro de que desea eliminar los {{ table.numSelected }} hosts de contenido seleccionados de las colecciones de hosts elegidas?","Are you sure you want to remove {{ content.content }} from the {{ getSelectedSystemIds().length }} system(s) selected?":"¿Está seguro de que desea eliminar {{ content.content }} de los {{ getSelectedSystemIds().length }} sistemas seleccionados?","Are you sure you want to unregister content host \"{{ contentHost.name }}\"?":"¿Está seguro de que desea cancelar el registro de host de contenido \"{{ contentHost.name }}\"?","Are you sure you want to unregister the {{ table.numSelected }} content host(s) selected?":"¿Está seguro de que desea cancelar el registro  {{ table.numSelected }} host(s) de contenido seleccionados?","Are you sure you want to update {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"¿Está seguro de que desea actualizar {{ content.content }} en los {{ getSelectedSystemIds().length }} sistemas seleccionados?","As part of this deletion, 1 Content View Version will be deleted.":["Como parte del borrado, se borrará 1 versión de vista de contenido."," Como parte del borrado, se borrarán {{ versions.length }} versiones de vista de contenido."],"Assign":"Asignar","Assign Environment and Content View":"Asignar entorno y vista de contenido","Associations":"Asociaciones","At least one Errata needs to be selected to Apply.":"Se debe seleccionar por lo menos una errata para que aplique.","At least one content host needs to be selected in order to perform a bulk action.":"Se debe seleccionar al menos un host de contenido para realizar una acción masiva.","At least one product needs to be selected in order to perform a bulk action.":"Se debe seleccionar al menos un producto para realizar una acción masiva.","Attached":"Asociada","Attached to Content Hosts":"Asociada a hosts de contenido","Author":"Autor","Auto-Attach":"Asociar automáticamente","Auto-attach available subscriptions to all content hosts":"Asociar automáticamente suscripciones disponibles a todos los hosts de contenido","Automatic":"Automático","Available Content Views for Composite Content View:":"Vistas de contenido disponibles para vista de contenido compuesta:","Available Puppet Modules for Content View:":"Módulos Puppet disponibles para vista de contenido:","Back":"Atrás","Back To Errata List":"Volver a la lista de erratas","Back To Repository Selection":"Volver a la selección de repositorio","Back to Module List":"Volver a la lista de módulos","Back to Repository Details":"Volver a los detalles del repositorio","Back to Repository List":"Volver a la lista de repositorios","Back to Version List":"Volver a la lista de versiones","Basic Information":"Información básica","Below are the product's repository content sets currently available for this subscription. For Red Hat subscriptions, additional content can be made available through the":"A continuación se indican los sets de  contenido del repositorio del producto que está disponible para esta suscripción. En el caso de suscripciones Red Hat, se puede ofrecer contenido adicional a través de ","Below are the repository content sets currently available for this activation key through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"A continuación se indican los conjuntos de contenido de repositorio que están disponibles para esta llave de activación a través de sus suscripciones. En el caso de suscripciones Red Hat, podrá ofrecerse  contenido adicional a través de","Below are the repository content sets currently available for this content host through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"A continuación se indican los conjuntos de contenido de repositorio que están disponibles para este host de contenido a través de sus suscripciones. En el caso de suscripciones Red Hat, se puede ofrecer contenido adicional a través de","Bug Fix":"Corrección","Bug Fix Advisory":"Recomendación de corrección","Bugfix":"Corrección","Bulk Actions":"Acciones masivas","Bulk Alter Products Sync Plan":"Modificar de forma masiva el plan de sincronización de los productos","Bulk Sync Products":"Sincronizar de forma masiva los productos","CVEs":"CVE","Cancel":"Cancelar","Capacity":"Capacidad","Change assigned Lifecycle Environment or Content View":"Cambiar entorno del ciclo de vida o vista de contenido asignados","Changing default settings for content hosts that register with this activation key requires subscription-manager version 1.10 or newer to be installed on that host.":"Para modificar la configuración predeterminada de los hosts de contenido que se registran con esta llave de activación debe tener instalada en este host  la herramienta subscription-manager versión 1.10 o posterior.","Changing default settings requires subscription-manager version 1.10 or newer to be installed on this host.":"Para modificar la configuración predeterminada debe tener instalada en el host la herramienta subscription-manager versión 1.10  o posterior.","Checkin":"Registro","Checksum":"Suma de verificación","Choose a lifecycle environment from the available promotion paths.":"Elija un entorno del ciclo de vida de entre las rutas de promoción disponibles.","Click here to check the status of the task.":"Haga clic para verificar el estatus de la tarea.","Click here to refresh.":"Haga clic aquí para actualizar.","Click here to see the task for the successful removals.":"Haga clic para ver la tarea para los retiros correctos.","Click here to select Errata for an Incremental Update.":"Haga clic para seleccionar erratas para una actualización incremental. ","Close":"Cerrar","Collection Actions":"Acciones de recopilación","Completely remove version?":"¿Eliminar completamente la versión?","Components":"Componentes","Components:":"Componentes:","Composite Content View {{ contentView.name }}":"Vista de contenido compuesta {{ contentView.name }}","Composite View":"Vista de compuesto","Composite View?":"¿Vista compuesta?","Composite?":"¿Compuesto?","Confirm":"Confirmar","Confirm Remove":"Confirmar eliminación","Confirm Version Removal: Version {{ version.version }}":"Confirmar eliminación de versión: Versión {{ version.version }}","Confirmation":"Confirmación","Consumed":"Consumidas","Content":"Contenido","Content Counts":"Recuentos de contenido","Content Download URL":"URL de descarga de contenido","Content Host":"Host de contenido","Content Host %s has been deleted.":"Se ha borrado el host de contenido %s.","Content Host Bulk Actions":"Acciones masivas del host de contenido","Content Host Bulk Content":"Contenido masivo del host de contenido","Content Host Bulk Errata":"Erratas masivas del host de contenido","Content Host Bulk Host Collections":"colecciones de hosts masivas del host de contenido","Content Host Bulk Packages":"Paquetes masivos del host de contenido","Content Host Bulk Subscriptions":"Suscripciones masivas del host de contenido","Content Host Content":"Contenido del host de contenido","Content Host Counts":"Recuento de host de contenido","Content Host Errata Management":"Administración de erratas del host de contenido","Content Host Limit":"Límite del host de contenido","Content Host Package Management":"Administración de paquetes del host de contenido","Content Host Properties":"Propiedades del host de contenido","Content Host Registration":"Registro del host de contenido","Content Host Status":"Estado del host de contenido","Content Host Subscriptions":"Suscripciones del host de contenido","Content Host {{ contentHost.name }}":"Host de contenido {{ contentHost.name }}","Content Host:":"Host de contenido:","Content Hosts":"Hosts de contenido","Content Hosts for Activation Key:":"Hosts de contenido para la llave de activación:","Content Hosts for Subscription:":"Hosts de contenido para la suscripción:","Content Hosts for:":"Hosts de contenido para:","Content Hosts using Version {{ version.version }}":"Hosts de contenido que utilizan la versión {{ version.version }}","Content Type":"Tipo de contenido","Content View":"Vista de contenido","Content View Name":"Nombre de vista de contenido","Content View Version":"Versión de vista de contenido","Content View updated.":"Se ha actualizado la vista de contenido.","Content View:":"Vista de contenido:","Content Views":"Vistas de contenido","Content Views <div>{{ library.counts.content_views || 0 }}</div>":"Vista de contenido <div>{{ library.counts.content_views || 0 }}</div>","Content Views for Composite Content View:":"Vistas de contenido para la vista de contenido compuesta:","Content successfully uploaded":"El contenido se ha cargado correctamente","Contents":"Contenido","Contract":"Contrato","Contract Number":"Número de contrato","Copy":"Copiar","Copy Activation Key":"Copiar llave de activación","Copy Collection":"Copiar recopilación","Copy View":"Copiar vista","Cores per Socket":"Núcleos por socket","Cores: %s":"Núcleos: %s","Create":"Crear","Create New View":"Crear vista","Create Repositories":"Crear repositorios","Create Repositories Within:":"Crear repositorios dentro: ","Create Repository":"Crear repositorio","Create Selected":"Crear seleccionado","Create Sync Plan":"Crear plan de sincronización","Create new view":"Crear vista","Create view from existing views":"Crear vista de entre las vistas existentes","Critical":"Crítico","Current Environment (%e/%cv)":"Entorno actual  (%e/%cv)","Current Subscriptions for Activation Key:":"Suscripciones actuales para la llave de activación:","Currently Selected Puppet Modules":"Módulos Puppet seleccionados actualmente","Daily at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"Diariamente a las {{ product.sync_plan.sync_date | date:'mediumTime' }} (Hora del servidor)","Date":"Fecha","Default":"Predeterminado","Delete":"Borrar","Delete Manifest":"Borrar manifiesto","Delete Version {{ version.version }}":"Borrar la versión {{ version.version }}","Delete {{ contentView.name }}":"Borrar {{ contentView.name }}","Deleted from %s":"Borrado de %s","Deleting Manifest...":"Borrando el manifiesto...","Deleting from %count environments.":"Borrando de %count entornos.","Deleting from 1 environment.":"Borrado de 1 entorno.","Deletion from %s":"Borrado desde %s","Description":"Descripción","Detail":"Detalle","Details":"Información","Details for Activation Key:":"Detalles de la llave de activación:","Details for Content View:":"Detalles de la vista de contenido:","Details for Product:":"Detalles del producto:","Discover":"Descubrir","Discovered URLs":"URL detectadas","Docker Content":"Contenido de Docker","Docker Images":"Imágenes Docker","Docker Images in {{ repository.name }}":"Imágenes Docker en {{ repository.name }}","Docker Repositories":"Repositorios Docker","Docker Repositories <div>{{ library.counts.docker_repositories || 0 }}</div>":"Repositorios Docker <div>{{ library.counts.docker_repositories || 0 }}</div>","Docker Tag:":"Etiqueta Docker:","Docker Tags":"Etiquetas Docker","Docker metadata generation has been initiated in the background.  Click\n        <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"La  generación de metadatos Docker se ha iniciado en segundo plano. Haga clic en\n        <a ng-href=\"{{ taskUrl() }}\">Aquí</a> para supervisar el progreso.","Edit":"Editar","Enabled?":"¿Habilitado?","End Date":"Fecha de terminación","Ends":"Finaliza","Enhancement":"Mejora","Enter Package Group Name(s)...":"Introducir nombres de grupos de paquetes...","Enter Package Name(s)...":"Ingresar nombres de paquetes...","Environment":"Entorno","Environment saved":"Entorno guardado","Environment: {{ environment.name }}":"Entorno: {{ environment.name }}","Environments":"Entornos","Environments List":"Lista de entornos","Environments:":"Entornos:","Equal To":"Igual que","Errata":"Erratas","Errata <div>{{ library.counts.errata.total || 0 }}</div>":"Erratas <div>{{ library.counts.errata.total || 0 }}</div>","Errata Details":"Detalles de las erratas","Errata Filter:":"Filtro de erratas:","Errata ID":"ID de erratas","Errata Installation":"Instalación de erratas","Errata List":"Lista de erratas","Errata Task List":"Lista de tareas de erratas","Errata Tasks":"Tareas de erratas","Errata Type":"Tipo de erratas","Errata for:":"Erratas para:","Errata successfully added.":"Las erratas se han añadido correctamente.","Errata successfully removed.":"Las erratas se han eliminado correctamente.","Errata:":"Erratas:","Erratum - by Date and Type":"Errata - por fecha y tipo","Erratum - by ID":"Errata - por ID","Erratum Date Range":"Intervalos de fechas de la errata","Error deleting manifest.":"Ocurrió un error al borrar el manifiesto.","Error during upload:":"Error durante la carga:","Error importing manifest.":"Ocurrió un error al importar el manifiesto.","Error refreshing manifest.":"Ocurrió un error al actualizar el manifiesto.","Event":"Evento","Events":"Eventos","Exclude":"Excluir","Existing Product":"Productos existentes","Expires":"Expira","Failed":"Falló","File too large. Please use the CLI instead.":"Archivo demasiado grande.  Use CLI en su lugar.","Filter":"Filtro","Filter Saved":"Filtro guardado","Filter information for:":"Información de filtro para:","Filter...":"Filtrar...","Filter:":"Filtro:","Filters":"Filtros","Filters List":"Lista de filtros","Filters for Content View:":"Filtros de la vista de contenido:","Filters successfully removed.":"Los filtros se han eliminado correctamente.","Finished At":"Finalizado a las","For older operating systems such as Red Hat Enterprise Linux 5 or CentOS 5 it is recommended to use sha1.":"Para sistemas operativos tales como Red Hat Enterprise Linux 5 o CentOS 5 se recomienda usar sha1.","GPG Key":"llave GPG","GPG Key URL":"URL de la llave GPG","GPG Key successfully uploaded":"La llave GPG se ha cargado correctamente","GPG Key {{ gpgKey.name }}":"Llave GPG {{ gpgKey.name }}","GPG Keys":"llaves GPG","General Information":"Información general","Generated On":"Generado","Gpg Key updated":"Llave GPG actualizada","Gpg Keys":"Llaves GPG","Greater Than":"Mayor que","Group Install":"Instalación de grupo","Group Remove":"Eliminación de grupo","Guests of":"Huéspedes de","Hide affected Activation Keys":"Ocultar llaves de activación afectadas","Hide affected Content Hosts":"Ocultar hosts de contenido afectados","History":"Historial","History for Content View:":"Historial para la vista de contenido:","Host Collection Management":"Administración de colecciones de hosts","Host Collection Membership":"Afiliación de colecciones de hosts","Host Collection Membership Management":"Administración de la afiliación a colecciones de hosts","Host Collection removed.":"Se ha eliminado la colección de hosts.","Host Collection updated":"Colección de hosts actualizada","Host Collection:":"Colección de hosts:","Host Collection: {{ hostCollection.name }}":"Colección de hosts: {{ hostCollection.name }}","Host Collections":"colecciones de hosts","Host Collections for:":"Colecciones de hosts de:","Host Count":"Cuenta de host","Host Group":"Grupo de hosts","Hostname":"Nombre de host","Hourly at {{ product.sync_plan.sync_date | date:'m' }} minutes and {{ product.sync_plan.sync_date | date:'s' }} seconds":"Cada hora a los {{ product.sync_plan.sync_date | date:'m' }} minutos y {{ product.sync_plan.sync_date | date:'s' }} segundos","IPv4 Address":"Dirección IPv4","IPv6 Address":"Dirección IPv6","Id":"ID","If you would prefer to move portions of these Activation Keys to different Content Views or Lifecycle Environments click <a href=\"{{ activationKeyLink() }}\">here</a> to manage them individually.":"Si prefiere mover partes de estas teclas de activación a otras vistas de contenido o entornos del ciclo de vida, haga clic <a href=\"{{ activationKeyLink() }}\">aquí</a> para gestionarlas de forma individual.","If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click":"Si prefiere mover partes de estas teclas de activación a otras vistas de contenido o entornos del ciclo de vida, haga clic","If you would prefer to move portions of these Content Hosts to different content views or environments click <a href=\"{{ contentHostsLink() }}\">here</a>  to manage these Content Hosts in bulk.":"Si prefiere mover partes de estos hosts de contenido a otras vistas de contenido o entornos, haga clic <a href=\"{{ contentHostsLink() }}\">aquí</a>  para gestionar estos hosts de contenido de forma masiva.","Image":"Imagen","Image ID":"ID imagen","Import History":"Importar historia","Important":"Importante","In order to add a content view to a composite view you must first publish an initial version of the content view.":"Para añadir una vista de contenido a una vista compuesta, primero debe publicar una versión inicial de la vista de contenido.","Include":"Incluir","Included in Activation Keys":"Incluido en las teclas de activación","Incremental Update":"Actualización incremental","Install":"Instalación","Install Selected":"Instalar seleccionado","Install katello-agent for remote actions and displaying errata information:":"Instale katello-agent para acciones remotas y desplegar información de erratas:","Install the pre-built bootstrap RPM:":"Instale el RPM pre-built bootstrap","Installable":"Instalable","Installable Errata":"Erratas instalables","Installed":"Instalado","Installed Package":"Paquete instalado","Installed Packages":"Paquetes instalados","Installed Products":"Productos instalados","Instance-based":"Basado en instancia","Interfaces":"Interfaces","Interval":"Intervalo","Issued":"Emitido","Katello Agent":"Agente Katello","Label":"Etiqueta","Last Checkin":"Última conexión","Last Published":"Última publicación","Last Puppet Report":"Último informe de Puppet","Last Sync":"Última sincronización","Last Updated":"Última actualización","Last Updated On":"Última actualización","Last synced {{ product.last_sync_words }} ago.":"Última sincronización {{ product.last_sync_words }}. ","Latest (Currently %s)":"Última (actualmente, %s)","Less Than":"Menor que","Library":"Biblioteca","Library Synced Content":"Contenido de biblioteca sincronizado ","Lifecycle Environment \"{{ priorEnvironment.name }}\" exists within the middle of a path. New Lifecycle Environments can only be added to the end of a path.":"Entorno de ciclo de vida \"{{ priorEnvironment.name }}\" existe en la mitad de una ruta. Solo se pueden aplicar nuevos entornos de ciclo de vida al final de una ruta.","Lifecycle Environment Paths":"Rutas de entornos del ciclo de vida","Lifecycle Environments":"Entornos del ciclo de vida","Limit":"Límite","Limit:":"Límite:","Limits":"Límites","List/Remove":"Mostrar/Eliminar","Loading Environment Paths...":"Cargando rutas de entornos...","Loading...":"Cargando...","Manage Docker Images":"Administrar imágenes Docker","Manage Docker Images for Repository:":"Administrar imágenes Docker para repositorio:","Manage Manifest":"Administrar manifiesto","Manage Packages":"Administrar paquetes","Manage Packages for Repository:":"Administrar paquetes para el repositorio:","Manage Puppet Modules":"Administrar módulos Puppet","Manage Puppet Modules for Repository:":"Administrar módulos Puppet para repositorio","Manifest History":"Historial de manifiesto","Manifest successfully deleted.":"El manifiesto se ha borrado correctamente.","Manifest successfully imported.":"El manifiesto se ha importado correctamente.","Manifest successfully refreshed.":"El manifiesto se ha actualizado correctamente.","Message":"Mensaje","Model":"Modelo","Moderate":"Moderado","Module %s removed from Content View.":"Se ha eliminado el módulo %s de la vista de contenido.","Module List":"Lista de módulos","More Details":"Más información","More...":"Más...","Multi-entitlement":"Multi-derechos","N/A":"N/D","Name":"Nombre","Name of the upstream repository you want to sync. Example: 'busybox' or 'fedora/ssh'.":"Nombre del repositorio de la corriente de desarrollo principal que desea sincronizar. Por ejemplo:   'busybox' o 'fedora/ssh'.","Networking":"Red","Never synced":"Nunca sincronizado","New Activation Key":"Nueva llave de activación","New Content View":"Nueva vista de contenido","New Environment":"Nuevo Entorno","New Environment Path":"Nueva ruta de entorno","New Filter":"Nuevo filtro","New Filter for Content View:":"Nuevo filtro para la vista de contenido:","New GPG Key":"Nueva llave GPG","New Gpg Key":"Nueva llave GPG","New Host Collection":"Nueva recopilación de hosts","New Name:":"Nuevo nombre:","New Product":"Nuevo producto","New Repository":"Nuevo repositorio","New Sync Plan":"Nuevo plan de sincronización","New sync plan successfully created.":"El nuevo plan de sincronización se ha creado correctamente.","Next":"Siguiente","Next Sync":"Próxima sincronización","No":"No","No \"Edit\" permissions on some of the Activation Keys in that Lifecycle Environment.":"No hay permisos de edición en algunas de las llaves de activación de ese entorno del ciclo de vida.","No \"Edit\" permissions on some of the Content Hosts in that Lifecycle Environment.":"No hay permisos de edición en algunos de los hosts de contenido de ese entorno del ciclo de vida.","No \"Promote/Remove\" permissions on the Lifecycle Environment.":"No hay permisos de promoción/eliminación en el entorno del ciclo de vida.","No (Default)":"No (predeterminado)","No Content Hosts are affected.":"No hay hosts de contenido afectados.","No Content Hosts match this Erratum.":"No hay contenidos que coincidan con esta errata.","No Content Views available, please select another environment.":"No hay vistas de contenido disponibles; seleccione otro entorno.","No Host Collections to show, you can add Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"No hay colecciones a mostrar, puede agregar colecciones de hosts después de seleccionar 'Colecciones de hosts' bajo 'Hosts' en el menú principal. ","No Host Collections to show, you can add Host Collections after selecting the 'Add' tab.":"No hay colecciones a mostrar, puede agregar Colecciones de hosts después de seleccionar la pestaña 'Añadir'.","No Products to show.":"No hay productos para mostrar.","No Repositories contain this Erratum.":"No hay repositorios que contengan esta errata.","No activation keys are affected.":"No hay llaves de activación afectadas.","No alternate release version choices are available. The available releases are based upon what is available in \"{{ contentHost.content_view.name }}\", the selected <a href=\"/content_views\">content view</a> this content host is attached to for the given <a href=\"/lifecycle_environments\">lifecycle environment</a>, \"{{ contentHost.environment.name }}\".":"No hay opciones de versiones de lanzamiento alternativas disponibles. Los lanzamientos disponibles se basan en lo que está disponible en \"{{ contentHost.content_view.name }}\", la <a href=\"/content_views\">vista de contenido</a> seleccionada  a la cual se asocia este host de contenido  para un <a href=\"/lifecycle_environments\">entorno del ciclo de vida</a> determinado, \"{{ contentHost.environment.name }}\".","No content views exist for {{selected.environment.name}}":"No existen vistas de contenido para {{selected.environment.name}}","No host collections":"No hay colecciones de hosts","No provisioning details are available for this content host.":"No hay detalles de aprovisionamiento disponibles para este host de contenido.","No repository content provided through subscriptions.":"No se ha proporcionado contenido de repositorio a través de las suscripciones.","No repository content provided through this subscription.":"No se ha proporcionado contenido de repositorio a través de esta suscripción.","No repository content.":"No hay contenido de repositorio.","No subscription manifest imported":"No hay manifiesto de suscripción importado","None":"Ninguno","Not Synced":"No sincronizado","Not installed":"No instalado","Not yet published":"Aún no se ha publicado","Number of CPUs":"Número de CPU","OS":"SO","Only show Errata that is Applicable to one or more Content Hosts":"Solo muestra erratas que sean aplicables a uno o más hosts de contenido","Only show Errata that is Installable on one or more Content Hosts":"Solo muestra las erratas que son instalables en uno o más hosts de contenido","Operating System":"Sistema operativo","Options:":"Opciones:","Organization":"Organización","Original Sync Date":"Fecha de la sincronización original","Override to No":"Cambiar a No","Override to Yes":"Cambiar a Sí","Package":"Paquete","Package Actions":"Acciones de paquete","Package Filter:":"Filtro de paquete:","Package Group":"Grupo de paquetes","Package Group Filter:":"Filtro de grupo de paquetes:","Package Group successfully added.":"El grupo de paquetes se ha añadido correctamente.","Package Group successfully removed.":"El grupo de paquetes se ha eliminado correctamente.","Package Groups":"Grupos de paquetes","Package Install":"Instalación de paquetes","Package Installation, Removal, and Update":"Instalación, eliminación y actualización de paquetes","Package Remove":"Retiro de paquetes","Package Update":"Actualización de paquetes","Package removal failed":"No se ha podido eliminar el paquete","Package successfully added.":"El paquete se ha añadido correctamente.","Package successfully removed.":"El paquete se ha eliminado correctamente.","Package successfully updated.":"El paquete se ha actualizado correctamente.","Package/Group Name":"Nombre del grupo/paquete","Packages":"Paquetes","Packages <div>{{ library.counts.packages || 0 }}</div>":"Paquetes <div>{{ library.counts.packages || 0 }}</div>","Packages for:":"Paquetes para:","Packages in {{ repository.name }}":"Paquetes en {{ repository.name }}","Parameters":"Parámetros","Paste GPG Key Contents":"Pegar contenido de la llave GPG","Paste contents...":"Pegar contenido...","Paused":"En pausa","Perform":"Realizar","Physical":"Físicos","Please select a Lifecycle Environment and Content View to move the affected Activation Keys to:":"Seleccione un entorno del ciclo de vida y una vista de contenido a los que mover las llaves de activación afectadas:","Please select a Lifecycle Environment and Content View to move these Content Hosts to:":"Seleccione un entorno del ciclo de vida y una vista de contenido a los que mover estos hosts de contenido:","Please select a Lifecycle Environment.":"Seleccione un entorno del ciclo de vida.","Please select an environment.":"Seleccione un entorno.","Please select one from the list below and you will be redirected.":"Por favor, seleccione uno de la lista de abajo y será reenviado.","Press 'Sync Now' below to begin syncing the following product(s).":"A continuación, pulse 'Sincronizar ahora' para empezar a sincronizar los siguientes productos.","Previous Environment (%e/%cv)":"Entorno anterior (%e/%cv)","Processing":"Procesamiento","Product":"Producto","Product Bulk Actions":"Acciones masivas del producto","Product Content":"Contenido del producto","Product Count":"Recuento de productos","Product Enhancement Advisory":"Recomendación de mejoras de productos","Product ID":"ID de producto","Product Management":"Administración de productos","Product Management for Sync Plan:":"Administración de productos para el plan de sincronización:","Product Name":"Nombre de producto","Product Saved":"Producto guardado","Product Sync":"Sincronización de productos","Product delete operation has been initiated in the background.  Click <a href=\"{{ taskUrl(productDeletionTaskId) }}\">Here</a> to monitor the progress.":"La operación de borrado de productos se ha iniciado en segundo plano.  Haga clic <a href=\"{{ taskUrl(productDeletionTaskId) }}\">aquí</a> para supervisar el progreso.","Product information for:":"Información de producto para:","Product {{ product.name }}":"Producto {{ product.name }}","Products":"Productos","Products <div>{{ library.counts.products || 0 }}</div>":"Productos <div>{{ library.counts.products || 0 }}</div>","Promote":"Promover","Promote Content View:":"Promover vista de contenido:","Promote Version":"Promover versión","Promote Version {{ version.version }}":"Promover versión {{ version.version }}","Promoted to %s":"Promovida a %s","Promoting to %count environments.":"Promovida a %count entornos.","Promoting to 1 environment.":"Se está promoviendo a 1 entorno.","Promotion History":"Historial de promociones","Provided Products":"Productos suministrados","Provisioning Details":"Detalles de aprovisionamiento","Provisioning Host Details":"Detalles de host de aprovisionamiento","Publish Content View:":"Publicar vista de contenido:","Publish New Version":"Publicar nueva versión","Publish via HTTP":"Publicar mediante HTTP","Publish via HTTPS":"Publicar mediante HTTPS","Published":"Publicado","Published At":"Publicada a las","Published Repository Information":"Se ha publicado información de repositorio","Published new version":"Nueva versión publicada","Publishing and promoting to %count environments.":"Publicando y promoviendo a %count entornos.","Publishing and promoting to 1 environment.":"Publicando y promoviendo a 1 entorno.","Puppet Environment":"Entorno de Puppet","Puppet Modules":"Módulos Puppet","Puppet Modules <div>{{ library.counts.puppet_modules || 0 }}</div>":"Módulos Puppet <div>{{ library.counts.puppet_modules || 0 }}</div>","Puppet Modules for Content View:":"Módulos Puppet para la vista de contenido:","Puppet Modules in {{ repository.name }}":"Módulos Puppet en {{ repository.name }}","Puppet module added to Content View":"Se ha añadido el módulo Puppet a la vista de contenido","Puppet module metadata generation has been initiated in the background.\n        Click <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"La  generación de metadatos Puppet se ha iniciado en segundo plano. Haga clic \n        <a ng-href=\"{{ taskUrl() }}\">Aquí</a> para supervisar el progreso.","Quantity":"Cantidad","RAM (GB)":"RAM (GB)","RAM: %s GB":"RAM: %s GB","Range":"Rango","Reboot Suggested":"Se sugiere un reinicio","Reboot Suggested?":"¿Se sugiere reiniciar?","Red Hat CDN URL":"Red Hat CDN URL","Red Hat Provider Details":"Información sobre el proveedor Red Hat","Red Hat Repositories page":"Página de repositorios Red Hat","Red Hat Repositories page.":"Página de repositorios Red Hat","Refresh Manifest":"Recargar manifiesto","Refresh Table":"Actualizar tabla","Refreshing Manifest...":"Actualizando el manifiesto...","Refreshing manifest not suppported by this manifest.":"Este manifiesto no soporta la actualización de manifiestos.","Register Content Host":"Registrar host de contenido","Register a Content Host":"Registrar un host de contenido","Registered":"Registrado","Registered By":"Registrado por","Registry URL":"URL de registro","Release":"Lanzamiento","Release Version":"Versión de lanzamiento","Remove":"Borrar","Remove Activation Key \"{{ activationKey.name }}\"?":"¿Está seguro de que desea retirar la llave de activación  \"{{ activationKey.name }}\"?","Remove Content View":"Eliminar vista de contenido","Remove Content Views":"Eliminar vistas de contenido","Remove Docker Images":"Retirar imágenes Docker","Remove Errata":"Eliminar erratas","Remove Filters":"Eliminar filtros","Remove From":"Eliminar desde","Remove GPG Key":"Retirar llave GPG","Remove GPG Key {{ gpgKey.name }}":"Retirar la llave GPG {{ gpgKey.name }}","Remove Host Collection \"{{ hostCollection.name }}\"?":"¿Retirar colección de hosts \"{{ hostCollection.name }}\"?","Remove Module":"Eliminar módulo","Remove Package Group":"Eliminar grupo de paquetes","Remove Packages":"Eliminar paquetes","Remove Product":"Retirar producto","Remove Product \"{{ product.name }}\"?":"¿Retirar el producto \"{{ product.name }}\"?","Remove Products":"Eliminar productos","Remove Puppet Modules":"Retirar módulos Puppet.","Remove Repositories":"Eliminar repositorios","Remove Repository":"Retirar repositorio","Remove Repository \"{{ repository.name }}\"?":"¿Retirar el repositorio  \"{{ repository.name }}\"?","Remove Selected":"Retirar seleccionado","Remove Successful.":"La eliminación se ha realizado correctamente.","Remove Sync Plan":"Retirar plan de sincronización","Remove Sync Plan \"{{ syncPlan.name }}\"?":"¿Retirar el plan de sincronización \"{{ syncPlan.name }}\"?","Remove Version":"Eliminar versión","Remove Version Confirmation":"Eliminar confirmación de versión","Remove View":"Eliminar vista","Remove {{ productTable.getSelected().length }} Products?":"¿Retirar {{ productTable.getSelected().length }} productos?","Removed %x host collections from activation key \"%y\".":"Se han eliminado %x colecciones de hosts de la llave de activación \"%y\".","Removed %x host collections from content host \"%y\".":"Se han eliminado %x colecciones de hosts del host de contenido \"%y\".","Removed %x products from sync plan \"%y\".":"Se han eliminado %x productos del plan de sincronización \"%y\".","Removing Content Hosts":"Eliminando hosts de contenido","Removing Products":"Eliminando productos","Removing Repositories":"Eliminando repositorios","Repo Discovery":"Detección de repositorio","Repo Type":"Tipo de repositorio","Repositories":"Repositorios","Repositories containing Errata {{ errata.errata_id }}":"Repositorios que contengan las erratas {{ errata.errata_id }}","Repositories for Content View:":"Repositorios para la vista de contenido:","Repositories for Errata:":"Repositorios para erratas:","Repositories for Filter:":"Repositorios para el filtro:","Repositories for Product:":"Repositorios para el producto:","Repository":"Repositorio","Repository \"%s\" successfully deleted":"Repositorio \"%s\" ha sido borrado correctamente","Repository %s successfully created.":"El repositorio %s se ha creado correctamente.","Repository Count":"Recuento de repositorios","Repository Discovery":"Descubrimiento de repositorio","Repository Name":"Nombre de repositorio","Repository Saved.":"Se ha guardado el repositorio.","Repository Selection":"Selección de repositorio","Repository URL updated":"URL de repositorio actualizada","Repository:":"Repositorio:","Required Host":"Host requerido","Result":"Resultado:","Results":"Resultados","Rules Source":"Fuente de reglas","Rules Version":"Versión de reglas","Run Auto-Attach":"Ejecutar Auto-adjuntar","Run subscription-manager in a console on the client host.  You may use an Activation Key to register:":"Ejecute subscription-manager en una consola en el host de cliente. Puede usar una llave de activación para registrarse: ","Save Successful.":"Se ha guardado correctamente.","Search...":"Búsqueda...","Security":"Seguridad","Security Advisory":"Recomendación de seguridad ","Select A New Puppet Module To Add":"Seleccionar el nuevo módulo Puppet a añadir","Select Content Hosts":"Seleccionar hosts de contenido","Select Module":"Seleccionar módulo","Select Version":"Seleccionar versión","Select a Content Source:":"Seleccione una fuente de contenido: ","Select a Version":"Seleccionar una versión","Select an Available Version of {{ $stateParams.moduleName }}":"Seleccionar una versión disponible de {{ $stateParams.moduleName }}","Select an Organization":"Seleccionar una organización","Select new version":"Seleccionar nueva versión","Select the Lifecycle Environments you would like to remove Version {{ version.version }} from:":"Seleccione los entornos del ciclo de vida de los que desea eliminar la versión {{ version.version }}:","Selected URLs:":"URL seleccionadas:","Selecting this option will result in the Version being completely deleted and no longer being available for promotion.  The version must be removed from all Lifecycle Environments in order to select this option.":"Si se selecciona esta opción, la versión se borrará completamente y dejará de estar disponible para su promoción.  La versión se debe eliminar de todos los entornos del ciclo de vida para seleccionar esta opción.","Serve via HTTP.":"Servir a través de HTTP.","Service Level":"Nivel de servicio","Severity":"Gravedad","Show affected Activation Keys":"Mostrar llaves de activación afectadas","Show affected Content Hosts":"Mostrar host de contenido afectados","Show from:":"Mostrar desde:","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"Mostrando {{ detailsTable.rows.length }} de {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) hosts de contenido","Showing {{ addSubscriptionsTable.rows.length }} of {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} Total) Subscriptions":"Mostrando {{ addSubscriptionsTable.rows.length }} de {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} Total) suscripciones","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"Mostrando {{ detailsTable.rows.length }} de {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) hosts de contenido","Showing {{ currentPackagesTable.resource.total }} Packages":"Mostrando {{ currentPackagesTable.resource.total }} paquetes","Showing {{ subscriptionsTable.rows.length }} of {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} Total) Subscriptions":"Mostrando {{ subscriptionsTable.rows.length }} de {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} Total) suscripciones","Sockets":"Sockets","Sockets: %s":"Sockets: %s","Solution":"Solución","Some of the Errata shown below may not be installable as they are not in this Content Host's\n          Content View and Lifecycle Environment.  In order to apply such Errata an Incremental Update is required.":"Puede ser que algunas de las erratas de abajo no sean instalables en esta vista de contenido de host de contenido\n          entorno de ciclo de vida.  A fin de aplicar dichas erratas se requiere una  actualización incremental.","Stacking ID":"ID de apilamiento","Start Date":"Fecha de inicio","Start Time":"Hora de inicio","Started At":"Iniciado el","Starting":"Iniciando","Starts":"Inicia","State":"Estado","Status":"Estatus","Subscription Details":"Información de suscripción","Subscription Info":"Información de suscripción","Subscription Manifest":"Manifiesto de suscripción","Subscription Server Details":"Información de suscripción de servidor","Subscription Status":"Estatus de suscripción","Subscriptions":"Suscripciones","Subscriptions for Activation Key:":"Suscripciones para la llave de activación:","Subscriptions for:":"Suscripciones para:","Succesfully scheduled package installation":"La instalación del paquete se ha programado correctamente","Succesfully scheduled package removal":"La eliminación del paquete se ha programado correctamente","Succesfully scheduled package update":"La actualización del paquete se ha programado correctamente","Success":"Éxito","Successfully Scheduled Auto-attach.":"La acción de asociar automáticamente se ha programado correctamente.","Successfully added %s subscriptions.":"%s suscripciones se han añadido correctamente.","Successfully deleted %cv version %ver.":"%cv version %ver se ha borrado correctamente.","Successfully initiated deletion of %cv version %ver.":"El borrado de %cv version %ver se ha iniciado correctamente.","Successfully initiated promotion of %cv version %ver to %env.":"La promoción de %cv version %ver a %env se ha iniciado correctamente.","Successfully initiated removal of %cv version %ver.":"La eliminación de %cv version %ver se ha iniciado correctamente.","Successfully promoted %cv version %ver to %env":"%cv version %ver se ha promovido a %env correctamente","Successfully published %cv version %ver and promoted to Library":"%cv version %ver se ha publicado y se ha promovido a la biblioteca correctamente","Successfully removed %cv version %ver from environments: %env":"%cv version %ver se ha eliminado correctamente de los entornos: %env","Successfully removed %s items.":"Se retiraron  %s elementos.","Successfully removed %s subscriptions.":"%s suscripciones se han eliminado correctamente.","Successfully removed 1 item.":"Se retiró 1 ítem.","Successfully updated subscriptions.":"Las suscripciones se han actualizado correctamente.","Summary":"Resumen","Support Level":"Nivel de soporte","Support Type":"Tipo de soporte","Sync Enabled":"Se habilitó sincronización ","Sync Information":"Información de sincronización","Sync Interval":"Intervalo de sincronización","Sync Now":"Sincronizar ahora","Sync Plan":"Plan de sincronización","Sync Plan %s has been deleted.":"El plan de sincronización %s se ha borrado.","Sync Plan Saved":"Plan de sincronización guardado","Sync Plan {{ syncPlan.name }}":"Plan de sincronización {{ syncPlan.name }}","Sync Plan:":"Plan de sincronización:","Sync Plans":"Planes de sincronización","Sync State":"Estado de la sincronización","Sync Status":"Estatus de sincronización","Synced manually, no interval set.":"Se ha sincronizado de forma manual. No se ha establecido un intervalo.","Syncing...":"Sincronizando...","Tags":"Etiquetas","Target":"Destino","Task Details":"Detalles de tareas","Tasks":"Tareas","Tasks for Content View:":"Tareas para vista de contenido:","Temporary":"Temporal","The Activation Keys listed below are currently using this Content View Version.  Before deleting the Version you must move these Activation Keys to a Lifecycle Environment where this Version is not in use.":"Las teclas de activación que se muestran a continuación utilizan actualmente esta versión de vista de contenido.  Antes de borrar la versión, debe mover estas teclas de activación a un entorno del ciclo de vida donde no se esté utilizando esta versión.","The Content Hosts listed below are currently using this Content View Version.  Before removing the version you must move these Content Hosts to an environment where this version is not in use.":"Los hosts de contenido que se muestran a continuación utilizan actualmente esta versión de vista de contenido.  Antes de eliminar la versión, debe mover estos hosts de contenido a un entorno donde no se esté utilizando esta versión.","The Content View or Lifecycle Environment needs to be updated in order to make errata available to these hosts.":"La vista de contenido o entorno de ciclo de vida debe ser actualizada para que la errata esté disponible para estos hosts. ","The archive of this Version will be deleted.  This Version will not be available once deletion is complete.":"El archivo de esta versión se borrará.  Esta versión no estará disponible una vez que finalice el borrado.","The archive of this Version will not be deleted.  This Version will still be available to be promoted to a Lifecycle Environment.":"El archivo de esta versión no se borrará.  Esta versión aún estará disponible para ser promovida a un entorno del ciclo de vida.","The following actions can be performed on content hosts in this host collection:":"Las siguientes acciones se pueden realizar en hosts de contenido en esta colección de hosts:","The host requires being attached to a content view and the environment you have chosen has no content views promoted to it.\n            See the <a href=\"/content_views\">content views page</a> to manage and promote a content view.":"El host debe vincularse a la vista de contenido y el entorno que ha elegido no tiene vistas de contenido promovidas.\n   Consulte <a href=\"/content_views\">content views page</a> para manejar y promover una vista de contenido.","The katello-agent package is required to manage errata on this Host.\n      Please install the katello-agent package to manage errata on\n      {{ contentHost.name }}.":"Se requiere el paquete katello-agent  para administrar erratas en este host.\n    Instale el paquete katello-agent para administrar erratas en \n    {{ contentHost.name }}.","The katello-agent package is required to manage packages on this Host.\n      Please install the katello-agent package to manage packages on\n      {{ contentHost.name }}. To install the package, connect directly to the\n      host from a terminal and use yum.":"  Para administrar paquetes en este host, requiere el paquete katello-agent.  \n        Instale el paquete katello-agent para administrar paquetes en\n        {{ contentHost.name }}. Conéctese  directamente al host desde una terminal \n        y use Yum para instalar el paquete. ","The page you are attempting to access requires selecting a specific organization.":"Para acceder a la página debe seleccionar una organización específica.","The selected environment contains no Content Views, please select a different environment.":"El entorno seleccionado no contiene vistas de contenido. Seleccione otro entorno.","There are currently no repositories associated with this Content View, you can add some by clicking on the \"Add\" tab above.":"Actualmente no hay repositorios asociados a esta vista de contenido. Puede añadir alguno haciendo clic en la pestaña anterior \"Añadir\".","There are currently no repositories to add to this Content View,":"Actualmente no hay ningún repositorio que añadir a esta vista de contenido,","There are no Content Views in this Environment.":"No hay vistas de contenido en este entorno.","There are no Content Views that match the criteria.":"No hay vistas de contenido que coincidan con los criterios.","There are no Docker Tags that match the criteria.":"No hay etiquetas Docker que coincidan con los criterios.","There are no Errata associated with this Content Host to display.":"No hay erratas asociadas con este Host de contenido para mostrar.","There are no Errata in this organization.  Create one or more Products with Errata to view Errata on this page.":"No hay erratas en esta organización. Cree uno más productos con erratas para ver erratas en esta página.","There are no Errata that match the criteria.":"No hay erratas coincidentes con estos criterios.","There are no Errata to display.":"No hay erratas a mostrar.","There are no Host Collections available. You can create new Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"No hay colecciones de hosts disponibles. Cree una nueva colección de hosts después de seleccionar 'Colecciones de hosts' bajo 'Hosts' en el menú. ","There are no Packages that match the criteria.":"No hay paquetes que coincidan con los criterios.","There are no Puppet Modules that match the criteria.":"No hay módulos Puppet coincidentes con estos criterios.","There are no Yum Repositories that match the criteria.":"No hay repositorios Yum que coincidan con los criterios.","There is currently an Incremental Update task in progress.  This update must finish before applying existing updates.":"Actualmente hay una tarea de actualización incremental en progreso. Esta actualización debe terminar antes de aplicar las actualizaciones existentes.","There were errors while removing the following Repositories:":"Hubo errores al retirar los siguientes repositorios:","This Content Host is not attached to any subscriptions that provide content":"Este host de contenido no se ha asociado a ninguna suscripción que proporcione contenido","This Content View does not have any versions, create your first Content View Version by using the \"Publish New Version\" button on the right.":"Este contenido no tiene versiones, primero cree su versión de vista de contenido con el botón de la derecha \"Publicar nueva versión\".","This Version cannot be deleted from some of the Lifecycle Environments due to one of the reasons below.":"Esta versión no se puede borrar de algunos de los entornos del ciclo de vida debido a una de las siguientes razones.","This Version is not associated with any Lifecycle Environments.":"Esta versión no está asociada a ningún entorno del ciclo de vida.","This activation key is not associated with any content hosts.":"Esta llave de activación no está asociada a ningún host de contenido.","This activation key may be used during system registration. For example:":"Esta llave de activación se puede utilizar durante el registro de sistema. Por ejemplo:","This filter applies only to a subset of repositories in the content view.":"Este filtro solo se aplica a un subconjunto de repositorios en la vista de contenido.","This filter applies to all repositories in the content view (current and future).":"Este filtro se aplica a todos los repositorios en la vista de contenido (actuales y futuros).","This subscription is not associated with any activation keys.":"Esta suscripción no está asociada a ninguna tecla de activación.","This subscription is not attached to any content hosts.":"Esta suscripción no se ha asociado a ningún host de contenido.","This version will be removed from:":"Se eliminará esta versión de:","Time":"Hora","Title":"Título","To register a content host to this server, follow these steps.":"Para registrar un host de contenido para este servidor, siga los siguientes pasos:","Topic":"Tema:","Type":"Tipo","URL":"URL","URL of the registry you want to sync. Example: https://registry.hub.docker.com":"URL del registro que desea sincronizar. Por ejemplo: https://registry.hub.docker.com","URL to discover":"URL para detectar","UUID":"UUID","Unattach Sync Plan":"Dejar de asociar un plan de sincronización","Unattaching...":"Dejando de asociar...","Unlimited Content Hosts:":"Hosts de contenido ilimitados:","Unregister Content Host":"Anular el registro del host de contenido","Unregister Content Host \"{{contentHost.name}}\"?":"¿Cancelar el host de contenido  \"{{contentHost.name}}\"?","Unregister Content Hosts":"Anular el registro de hosts de contenido","Unregister {{ table.numSelected  }} Content Hosts?":"¿Cancelar {{ table.numSelected  }} hosts de contenido?","Update":"Actualizar","Update All":"Actualizar todos","Update Repositories":"Actualizar repositorios","Update Sync Plan":"Actualizar plan de sincronización","Updated":"Actualizado","Updated errata filter -":"Filtro de erratas actualizado -","Updated override to \"%x\".":"El cambio a \"%x\" se ha actualizado.","Updating...":"Actualizando...","Upload":"Cargar","Upload Docker Image":"Descargar imagen Docker","Upload GPG Key":"Cargar llave GPG","Upload New Manifest":"Cargar nuevo manifiesto","Upload Package":"Cargar paquete","Upload Puppet Module":"Cargar módulo puppet","Uploading...":"Cargando...","Upstream Repository Name":"Nombre de repositorio principal","Upstream Subscription":"Suscripción de desarrollo principal","Upstream Subscription Management Application":"Aplicación de administración de suscripción de desarrollo principal","Upstream Type":"Tipo de línea de desarrollo principal","Upstream User":"Usuario de línea de desarrollo principal","Use Latest (currently %s)":"Utilizar más reciente (actualmente %s)","Use the cancel button on content view selection to revert your environment selection.":"Utilice el botón de cancelación en la selección de vista de contenido para invertir la selección de entorno.","User":"Usuario","Version":"Versión","Version %s":"Versión %s","Version 1":"Versión 1","Version Deletion":"Borrado de versión","Version Details":"Detalles de la versión","Version {{ version.version }}":"Versión {{ version.version }}","Version:":"Versión:","Versions":"Versiones","Versions for Content View:":"Versiones de la vista de contenido:","Versions for Puppet Module:":"Versiones del módulo puppet:","View Details":"Ver detalles","Virtual":"Virtual","Virtual Guest From":"Huésped virtual de","Virtual Guests":"Huéspedes virtuales","Virtual Host":"Host virtual","Weekly on {{ product.sync_plan.sync_date | date:'EEEE' }} at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"Semanalmente el {{ product.sync_plan.sync_date | date:'EEEE' }} a las {{ product.sync_plan.sync_date | date:'mediumTime' }} (Hora del servidor)","When Auto Attach is disabled, registering systems will be attached to all associated subscriptions.":"Cuando auto-vincular está inhabilitado,los sistemas de registro se vincularán a todas las suscripciones asociadas. ","Working":"En funcionamiento","Yes":"Sí","Yes (Default)":"Sí (Predeterminado)","You cannot remove these repositories because they belong to a Red Hat product.":"No se pueden eliminar estos repositorios porque pertenecen a un producto de Red Hat.","You cannot remove these repositories because you do not have permission.":"No se pueden eliminar estos repositorios porque no tiene permiso.","You cannot remove this product because it is a Red Hat product.":"No se puede eliminar este producto porque es un producto de Red Hat.","You cannot remove this product because it was published to a content view.":"No se puede eliminar este producto porque se ha publicado en una vista de contenido.","You cannot remove this product because you do not have permission.":"No se puede eliminar este producto porque no tiene permiso.","You cannot remove this repository because it is a Red Hat repository.":"No se puede eliminar este repositorio porque es un repositorio Red Hat.","You cannot remove this repository because it was published to a content view.":"No se puede eliminar este repositorio porque se ha publicado en una vista de contenido.","You cannot remove this repository because you do not have permission.":"No se puede eliminar este repositorio porque no tiene permiso.","You currently do not have any Packages installed. You can use Package Actions above this message to install new Packages.":"Actualmente no tiene paquetes instalados.Use 'Acciones de paquetes' arriba de este mensaje para instalar nuevos paquetes.","You currently don't have any Activation Keys, you can add Activation Keys using the button on the right.":"Actualmente no tiene llaves de activación, agregue llaves de activación con el botón de la derecha.","You currently don't have any Content Hosts in this Host Group, you can add Content Hosts after selecting the 'Add' tab.":"Actualmente no tiene ningún host de contenido en este grupo de hosts, agregue hosts de contenido después de seleccionar la pestaña 'Añadir'","You currently don't have any Content Hosts, you can create new Content Hosts by selecting Contents Host from main menu and then clicking the button on the right.":"Actualmente no tiene ningún host de contenido, para crear un host de contenido, seleccione Host de contenido desde el menú principal y luego haga clic en el botón de la derecha.","You currently don't have any Content Hosts, you can register one by clicking the button on the right and following the instructions.":"Actualmente no tiene hosts de contenido. Para registrar  un host, haga clic en el  botón a la derecha y siga las instrucciones.  ","You currently don't have any Content Views.  A Content View can be added by using the button on the right.":"Actualmente no tiene vistas de contenido. Para agregar  una  vista de contenido use el botón de la derecha.","You currently don't have any Docker Tags.":"Actualmente no tiene etiquetas Docker.","You currently don't have any Filters included in this Content View, you can add a new Filter by using the button on the right.":"Actualmente no dispone de filtros en esta vista de contenido. Puede añadir un nuevo filtro mediante el botón que se encuentra a la derecha.","You currently don't have any Gpg keys, you can add Gpg keys using the button on the right.":"Actualmente no tiene llaves GPG, agregue llaves GPG con el botón de la derecha.","You currently don't have any Host Collections, you can add Host Collections using the button on the right.":"Actualmente no tiene colecciones de hosts, agregue colecciones de hosts con el botón de la derecha.","You currently don't have any Products associated with this Gpg Key.":"Actualmente no tiene llaves de productos asociados a esta llave GPG.","You currently don't have any Products to subscribe to, you can add Products after selecting 'Products' under 'Content' in the main menu.":"Actualmente no dispone de productos a suscribir,  agregue productos después de seleccionar 'Productos' bajo 'Contenido' en el menú principal.","You currently don't have any Products<span bst-feature-flag=\"custom_products\">, you can add Products using the button on the right</span>.":"Actualmente no tiene productos <span bst-feature-flag=\"custom_products\">, agregue productos con el botón a la derecha</span>.","You currently don't have any Puppet Modules included in this Content View, you can add Puppet Modules using the button on the right.":"Actualmente no tiene módulos Puppet incluidos en esta vista de contenido, agregue módulos Puppet con el botón de la derecha.","You currently don't have any Repositories associated with this Gpg Key.":"Actualmente no tiene repositorios asociados a esta llave GPG.","You currently don't have any Repositories included in this Product, you can add Repositories using the button on the right.":"Actualmente no tiene repositorios incluidos en este producto, agregue repositorios con el botón de la derecha.","You currently don't have any Subscriptions associated with this Content Host, you can add Subscriptions after selecting the 'Add' tab.":"Actualmente no tiene suscripciones asociadas al host de contenido, agregue suscripciones después de seleccionar la pestaña 'Añadir'","You currently don't have any Subscriptions, you can add Subscriptions by importing a manifest using the button on the right labeled 'Manage Manifest'.":"Actualmente no tiene suscripciones. Para agregar suscripciones, importe un manifiesto con el botón 'Administrar manifiesto'  a la derecha.","You currently don't have any Sync Plan, you can add Sync Plans using the button on the right.":"Actualmente no tiene ningún plan de sincronización. Para agregar un plan de sincronización use el botón a la derecha.","You currently don't have any Sync Plans.  A Sync Plan can be created by using the button on the right.":"Actualmente no tiene ningún plan de sincronización. Para agregar un plan de sincronización use el botón a la derecha.","You do not have any Installed Products":"No tiene productos instalados","You must select a content view in order to save your environment.":"Seleccione una vista de contenido para guardar un entorno ","You must select a new content view before your change of environment can be saved.":"Debe seleccionar una nueva vista de contenido antes de  guardar el cambio de entorno.","You must select a new content view before your change of environment can be saved. Use the cancel button on content view selection to revert your environment selection.":"Debe seleccionar una nueva vista de contenido antes de que el cambio de entorno pueda ser guardado. Utilice el botón Cancelar en la selección de vista de contenido para revertir la selección de entorno.","You must select at least one Content Host in order to apply Errata.":"Debe seleccionar un host de contenido para aplicar erratas. ","You must select at least one repository.":"Debe seleccionar al menos un repositorio.","Your selected content hosts have changed, it is highly recommended to refresh the calculated list of errata using the 'Refresh Errata' button below.":"Los hosts de contenido seleccionados han cambiado. Se recomienda enfáticamente que actualice la lista calculada de erratas con el botón de abajo: 'Actualizar erratas'.","Yum Content":"Contenido Yum","Yum Metadata Checksum":"Suma de verificación de metadatos Yum","Yum Repo Discovery":"Descubrimiento de repositorio Yum","Yum Repositories":"Repositorios Yum","Yum Repositories <div>{{ library.counts.yum_repositories || 0 }}</div>":"Repositorios Yum <div>{{ library.counts.yum_repositories || 0 }}</div>","Yum metadata generation has been initiated in the background.  Click <a href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"La generación de metadatos Yum se ha iniciado en segundo plano.  Haga clic <a href=\"{{ taskUrl() }}\">aquí</a> para supervisar el progreso.","add some repositories.":"añadir algunos repositorios.","daily":"A diario","here":"aquí","hourly":"Por hora","or authenticate with a username and password:":"o autentíquese con un nombre de usuario y contraseña: ","subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"":"subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"","task details":"Información de tareas","to manage them individually.":"para administrarlas de forma individual.","true":"true","weekly":"Semanal","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }} seleccionados","{{ addSubscriptionsTable.numSelected }} Selected":"{{ addSubscriptionsTable.numSelected }} seleccionadas","{{ contentHost.registered_by }}":"{{ contentHost.registered_by }}","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }} seleccionados","{{ contentView.name }} cannot be deleted as one or more Content View Versions are still promoted to a Lifecycle Environment.\n      Each Content View Version must be removed from their Lifecycle Environments before the Content View can be deleted.":"{{ contentView.name }} no se puede borrar, ya que una o más versiones de visión de contenido, aún se están promoviendo a un entorno de ciclo de vida.\n   Cada versión de vista de contenido debe ser retirada de los entornos de ciclo de vida antes de borrar la vista de contenido.","{{ contentView.puppet_modules.length || 0 }} Puppet Modules":"{{ contentView.puppet_modules.length || 0 }} módulos Puppet","{{ contentView.repositories.length || 0 }} Repositories":"{{ contentView.repositories.length || 0 }} repositorios","{{ contentViewVersion.puppet_module_count || 0 }} Puppet Modules":"{{ contentViewVersion.puppet_module_count || 0 }} Módulos Puppet","{{ contentViewVersion.repositories.length || 0 }} Repositories":"{{ contentViewVersion.repositories.length || 0 }} Repositorios","{{ discoveryTable.numSelected }} Selected":"{{ discoveryTable.numSelected }} seleccionados","{{ environment.name }} Environment":"{{ environment.name }} Entorno","{{ errata.hosts_applicable_count || 0 }} Applicable,":"{{ errata.hosts_applicable_count || 0 }} Aplicable,","{{ errata.hosts_available_count || 0 }} Installable":"{{ errata.hosts_available_count || 0 }} Instalable","{{ hostCollectionsTable.numSelected }} Selected":"{{ hostCollectionsTable.numSelected }} seleccionadas","{{ package.removeTask.affected_units }} package removed":"Se ha eliminado {{ package.removeTask.affected_units }} paquete","{{ package.removeTask.affected_units }} packages removed":"Se han eliminado {{ package.removeTask.affected_units }} paquetes","{{ product.last_sync_words }} Ago ({{ product.sync_status.finish_time | date:'medium' }} Local Time)":"Hace {{ product.last_sync_words }} ({{ product.sync_status.finish_time | date:'medium' }} Hora local)","{{ product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ product.sync_plan.next_sync | date:'medium' }} (Hora del servidor)","{{ productsTable.numSelected }} Selected":"{{ productsTable.numSelected }} seleccionados","{{ repository.content_counts.docker_image || 0 }} Docker Images":"{{ repository.content_counts.docker_image || 0 }} Imágenes Docker","{{ repository.content_counts.docker_image }} Docker Images":"{{ repository.content_counts.docker_image }} Imágenes Docker","{{ repository.content_counts.docker_tag || 0 }} Docker Tags":"{{ repository.content_counts.docker_tag || 0 }} Etiquetas Docker","{{ repository.content_counts.docker_tag }} Docker Tags":"{{ repository.content_counts.docker_tag }} Etiquetas Docker","{{ repository.content_counts.erratum || 0 }} Errata":"{{ repository.content_counts.erratum || 0 }} erratas","{{ repository.content_counts.erratum }} Errata":"{{ repository.content_counts.erratum }} Erratas","{{ repository.content_counts.puppet_module || 0 }} Puppet Modules":"{{ repository.content_counts.puppet_module || 0 }} módulos Puppet","{{ repository.content_counts.rpm || 0 }} Packages":"{{ repository.content_counts.rpm || 0 }} paquetes","{{ repository.content_counts.rpm }} Packages":"{{ repository.content_counts.rpm }} Paquetes","{{ repository.last_sync_words }} Ago ({{ repository.last_sync.ended_at | date:'medium' }} Local Time)":"{{ repository.last_sync_words }} Hace ({{ repository.last_sync.ended_at | date:'medium' }} Local Time)","{{ repository.last_sync_words }} ago":"Hace {{ repository.last_sync_words }}","{{ repository.product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ repository.product.sync_plan.next_sync | date:'medium' }} (Hora del servidor)","{{ subscription.host.name }}":"{{ subscription.host.name }}","{{ subscriptionsTable.numSelected }} Selected":"{{ subscriptionsTable.numSelected }} seleccionadas","{{ totalActivationKeyCount() }} Activation Keys will be moved to {{ deleteOptions.activationKeys.contentView.name }} in {{ deleteOptions.activationKeys.environment.name }}":"{{ totalActivationKeyCount() }} Llaves de activación se moverán a {{ deleteOptions.activationKeys.contentView.name }} en {{ deleteOptions.activationKeys.environment.name }}","{{ totalHostCount() }} Content Hosts will be moved to {{ deleteOptions.contentHosts.contentView.name }} in {{ deleteOptions.contentHosts.environment.name }}":"{{ totalHostCount() }} hosts de contenido se moverán a {{ deleteOptions.contentHosts.contentView.name }} en {{ deleteOptions.contentHosts.environment.name }}","{{ type.display }}":"{{ type.display }}","{{ version.docker_image_count }} Docker Images":"{{ version.docker_image_count }} Imágenes Docker","{{ version.docker_tag_count }} Docker Tags":"{{ version.docker_tag_count }} Etiquetas  Docker","{{ version.errata_counts.total }} Errata":"{{ version.errata_counts.total }} Erratas","{{ version.package_count }} Packages":"{{ version.package_count }} Paquetes","{{ version.puppet_module_count }} Puppet Modules":"{{ version.puppet_module_count }} Módulos Puppet"});
    gettextCatalog.setStrings('fr', {"%(consumed)s out of %(quantity)s":"%(consumed)s sur %(quantity)s","(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)":"(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)","(Not all Activation Keys editable )":"(toutes les clés d'activation ne sont pas modifiables)","(Not all Content Hosts editable )":"(tous les hôtes de contenu ne sont pas modifiables)","* These marked Content View Versions are from Composite Content Views.  Their components needing updating are listed underneath.":"* Ces versions d'affichages de contenu  marked proviennent d'affichages de contenu composites. Les composants nécessitant une mise à jour sont énumérés ci-dessous.","+ New Sync Plan":"+ Nouveau plan de synchronisation","- Date and Type":"- Date et type","-- select an interval --":"-- sélectionnez un intervalle --","1 Content Host":["1 hôte de contenu","{{ contentHost.virtual_guests.length }} Hôtes de contenu"],"1 repository sync has errors.":["1 synchronisation de référentiel contient des erreurs.","{{ product.sync_summary.error}} synchronisations de référentiel contiennent des erreurs."],"1 repository sync in progress.":["1 synchronisation de référentiel en cours.","{{ product.sync_summary.pending}} synchronisations de référentiel en cours."],"1 successfully synced repository.":["1 synchronisation de référentiel réussie.","{{ product.sync_summary.success}} synchronisations de référentiels réussies."],"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Version {{ version.version }}</a>":"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Version {{ version.version }}</a>","<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Create":"<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n        Créez","<span translate=\"\">\n        If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">here</a>\n      <span translate=\"\">to manage them individually.</span>":"<span translate=\"\">\n       Si vous préférez déplacer des portions de ces clés d'activations vers différents affichages du contenu ou environnements de cycle de vie, veuillez cliquer\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">ici</a>\n      <span translate=\"\">pour les gérer indivuellement.</span>","A composite view contains other content views.":"Une vue composite contient d'autres affichages du contenu.","A new version of {{ contentView.name }} will be created and promoted to the Library environment.\n    It can be promoted to other environments from the Versions tab of this Content View.":"Une nouvelle version de {{ contentView.name }} sera créée et transférée vers l'environnement de la Bibliothèque.\n  Elle peut être transférée vers d'autres environnements à partir de l'onglet Versions de cet Affichage de Contenu.","Account":"Compte","Account Number":"Numéro de compte","Action":"Action","Action Type":"Type d'action","Actions":"Actions","Actions for:":"Actions pour :","Activation Key":["Clé d'activation","Clés d'activation"],"Activation Key Content":"Contenu de la clé d'activation","Activation Key Type:":"Type de clé d'activation :","Activation Key removed.":"Clé d'activation supprimée.","Activation Key updated":"Clé d'activation mise à jour","Activation Key:":"Clé d'activation :","Activation Key: {{ activationKey.name }}":"Clé d'activation : {{ activationKey.name }}","Activation Keys":"Clés d'activation","Activation Keys for Subscription:":"Clés d'activation pour l'abonnement :","Activation Keys using Version {{ version.version }}":"Clés d'activation en utilisant la Version {{ version.version }}","Add":"Ajouter","Add Content Hosts to:":"Ajouter des hôtes de contenu à :","Add Content Views":"Ajouter les affichages du contenu","Add Errata":"Ajouter des errata","Add New Environment":"Ajouter un nouvel environnement","Add New Filter":"Ajouter un nouveau filtre","Add New Module":"Ajouter un nouveau module","Add New Repository":"Ajouter un nouveau référentiel","Add Package Group":"Ajouter un groupe de package","Add Repositories":"Ajouter des référentiels","Add Selected":"Ajouter les éléments sélectionnés","Add Subscriptions for Activation Key:":"Ajouter des abonnements pour la clé d'activation :","Add To":"Ajouter à","Added %x host collections to activation key \"%y\".":"%x collections d'hôtes ajoutées à la clé d'activation \"%y\".","Added %x host collections to content host \"%y\".":"%x collections d'hôtes ajoutées à l'hôte de contenu \"%y\".","Added %x products to sync plan \"%y\".":"%x produits ajoutés au plan de synchronisation \"%y\".","Adding Lifecycle Environment to the end of \"{{ priorEnvironment.name }}\"":"Ajouter un environnement de cycle de vie à la fin de \"{{ priorEnvironment.name }}\"","Advanced Information":"Informations détaillées","Advisory":"Alerte","Affected Activation Keys":"Clés d'activation affectées","Affected Content Hosts":"Hôtes de contenu affectés","Affected Packages":"Packages affectés","Affected Systems":"Systèmes affectés","Affected repositories":"Référentiels affectés","Affected repositories have been updated.":"Les référentiels affectés ont été mis à jour","All Content Views":"Tous les affichages de contenu","All Products":"Tous les produits","All Repositories":"Tous les référentiels","All Versions":"Toutes les versions","Alter Sync Plans":"Modifier les plans de synchronisation","An error occurred applying Subscriptions:":"Une erreur s'est produite lors de l'application des abonnements :","An error occurred removing the Activation Key:":"Une erreur s’est produite lors de la suppression de la clé d'activation :","An error occurred removing the Host Collection:":"Une erreur s’est produite lors de la suppression de la collection d'hôtes :","An error occurred removing the Products:":"Une erreur s’est produite lors de la suppression des produits :","An error occurred removing the content hosts.":"Une erreur s’est produite lors de la suppression des hôtes de contenu.","An error occurred removing the environment:":"Une erreur s’est produite lors de la suppression de l'environnement :","An error occurred removing the subscriptions.":"Une erreur s’est produite lors de la suppression des produits.","An error occurred saving the Activation Key:":"Une erreur s’est produite lors de l'enregistrement de la clé d'activation :","An error occurred saving the Content Host:":"Une erreur s’est produite lors de l'enregistrement de l'hôte de contenu :","An error occurred saving the Environment:":"Une erreur s'est produite lors de l'enregistrement de l'Environnement :","An error occurred saving the Filter:":"Une erreur s’est produite lors de l'enregistrement du filtre :","An error occurred saving the Host Collection:":"Une erreur s’est produite lors de l'enregistrement de la collection d'hôtes :","An error occurred saving the Product:":"Une erreur s’est produite lors de l'enregistrement du produit :","An error occurred saving the Repository:":"Une erreur s’est produite lors de l'enregistrement du référentiel :","An error occurred saving the Sync Plan:":"Une erreur s’est produite lors de l'enregistrement du plan de synchronisation :","An error occurred saving the URL:":"Une erreur s’est produite lors de l'enregistrement de l'URL :","An error occurred trying to auto-attach subscriptions.  Please check your log for further information.":"Une erreur s'est produite lors de la tentative d'attachement automatiquement des abonnements. Veuillez vérifier votre fichier journal pour obtenir plus d'informations.","An error occurred updating the Content View:":"Une erreur s’est produite lors de la mise à jour de l'affichage du contenu :","Applicable":"Applicable","Applicable Content Hosts for Errata":"Hôtes de contenu applicables pour les errata","Applicable Errata":"Errata applicables","Applicable hosts for {{ erratum.errata_id }}":"Hôtes applicables pour {{ erratum.errata_id }}","Apply":"Appliquer","Apply Errata":"Appliquer l'errata","Apply Errata to Content Host \"{{contentHost.name}}\"?":"Appliquer l'errata à l'hôte de contenu \"{{contentHost.name}}\" ?","Apply Errata to Content Hosts":"Appliquer l'errata aux hôtes de contenu","Apply Errata to Content Hosts immediately after publishing.":"Appliquer les errata aux hôtes de contenu tout de suite après la publication.","Apply Selected":"Appliquer Sélectionné","Apply to Content Hosts":"Appliquer aux hôtes de contenu","Apply to Hosts":"Appliquer aux hôtes","Apply {{ errata.errata_id }}":"Appliquer {{ errata.errata_id }}","Apply {{ errata.errata_id }} to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"Appliquer {{ errata.errata_id }} à {{ numberOfContentHostsToUpdate }} hôte(s) de contenu ?","Apply {{ errataIds.length }} Errata to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"Appliquer {{ errataIds.length }} Errata à {{ numberOfContentHostsToUpdate }} hôte(s) de contenu ?","Applying":"Application","Applying Bulk Errata Update":"Application de la mise à jour des errata par lot","Arch":"Arch","Architecture(s)":"Architecture(s)","Archived Copy":"Copie archivée","Are you sure you want to add the {{ table.numSelected }} content host(s) selected to the host collection(s) chosen?":"Êtes-vous sûr de vouloir ajouter le(s) {{ table.numSelected }} hôte(s) de contenu sélectionnés à la/aux collection(s) d'hôtes choisie(s) ?","Are you sure you want to apply Errata to content host \"{{ contentHost.name }}\"?":"Êtes-vous sûr de vouloir appliquer les Errata à l'hôte de contenu \"{{ contentHost.name }}\" ?","Are you sure you want to apply the {{ detailsTable.numSelected }} selected errata to the content hosts chosen?":"Êtes-vous sûr de vouloir appliquer les {{ detailsTable.numSelected }} errata sélectionnés aux hôtes de contenu choisis ?","Are you sure you want to assign the {{ table.numSelected }} content host(s) selected to {{ selected.contentView.name }} in {{ selected.environment.name }}?":"Êtes-vous sûr de vouloir attribuer le(s) {{ table.numSelected }} hôte(s) de contenu sélectionnés à {{ selected.contentView.name }} dans {{ selected.environment.name }} ?","Are you sure you want to auto-attach available subscriptions to all {{ table.resource.total }} content hosts?":"Êtes-vous sûr de vouloir attacher automatiquement les abonnements disponibles aux {{ table.resource.total }} hôtes de contenu ?","Are you sure you want to delete Content View \"{{ contentView.name }}\"?":"Êtes-vous sûr de vouloir supprimer l'affichage du contenu « {{ contentView.name }} » ?","Are you sure you want to install {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"Êtes-vous sûr de vouloir installer {{ content.content }} sur le(s) {{ getSelectedSystemIds().length }} système(s) sélectionné(s) ?","Are you sure you want to remove Activation Key \"{{ activationKey.name }}\"?":"Êtes-vous sûr de vouloir supprimer la clé d'activation \"{{ activationKey.name }}\" ?","Are you sure you want to remove GPG key {{ gpgKey.name }}?":"Êtes-vous sûr de vouloir supprimer la clé GPG {{ gpgKey.name }} ?","Are you sure you want to remove Host Collection \"{{ hostCollection.name }}\"?":"Êtes-vous sûr de vouloir supprimer la collection d'hôtes \"{{ hostCollection.name }}\" ?","Are you sure you want to remove Sync Plan \"{{ syncPlan.name }}\"?":"Êtes-vous sûr de vouloir supprimer le plan de synchronisation \"{{ syncPlan.name }}\" ?","Are you sure you want to remove product \"{{ product.name }}\"?":"Êtes-vous sûr de vouloir supprimer le produit \"{{ product.name }}\" ?","Are you sure you want to remove repository \"{{ repository.name }}\"?":"Êtes-vous sûr de vouloir supprimer le référentiel \"{{ repository.name }}\" ?","Are you sure you want to remove the {{ productTable.getSelected().length }} products(s) selected?":"Êtes-vous sûr de vouloir supprimer le(s) {{ productTable.getSelected().length }} produit(s) sélectionnés ?","Are you sure you want to remove the {{ table.numSelected }} content host(s) selected from the host collection(s) chosen?":"Êtes-vous sûr de vouloir supprimer le(s) {{ table.numSelected }} hôte(s) de contenu sélectionnés de la/des collection(s) d'hôtes choisie(s) ?","Are you sure you want to remove {{ content.content }} from the {{ getSelectedSystemIds().length }} system(s) selected?":"Êtes-vous sûr de vouloir supprimer {{ content.content }} du/des {{ getSelectedSystemIds().length }} système(s) sélectionné(s) ?","Are you sure you want to unregister content host \"{{ contentHost.name }}\"?":"Êtes-vous sûr de vouloir annuler l'enregistrement de l'hôte de contenu \"{{ contentHost.name }}\" ?","Are you sure you want to unregister the {{ table.numSelected }} content host(s) selected?":"Êtes-vous sûr de vouloir annuler l'enregistrement de  {{ table.numSelected }} hôte(s) de contenu sélectionnés ?","Are you sure you want to update {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"Êtes-vous sûr de vouloir mettre à jour {{ content.content }} sur le(s) {{ getSelectedSystemIds().length }} système(s) sélectionné(s) ?","As part of this deletion, 1 Content View Version will be deleted.":["Pour cette suppression, 1 version d'affichage du contenu sera supprimée.","Pour cette suppression, {{ versions.length }} versions d'affichage du contenu seront supprimées."],"Assign":"Attribuer","Assign Environment and Content View":"Attribuer l'environnement et l'affichage du contenu","Associations":"Associations","At least one Errata needs to be selected to Apply.":"Au moins un errata doit être sélectionné pour appliquer.","At least one content host needs to be selected in order to perform a bulk action.":"Au moins un hôte de contenu doit être sélectionné pour effectuer une action par lots.","At least one product needs to be selected in order to perform a bulk action.":"Au moins un produit doit être sélectionné pour effectuer une action par lots.","Attached":"Attaché","Attached to Content Hosts":"Attaché aux hôtes de contenu","Author":"Auteur","Auto-Attach":"Attacher automatiquement","Auto-attach available subscriptions to all content hosts":"Attacher automatiquement les abonnements disponibles à tous les hôtes de contenu","Automatic":"Automatique","Available Content Views for Composite Content View:":"Affichages du contenu disponibles pour l'affichage du contenu composite :","Available Puppet Modules for Content View:":"Modules puppet disponibles pour l'affichage du contenu :","Back":"Retour","Back To Errata List":"Retour à la liste d'errata","Back To Repository Selection":"Retour à la sélection de référentiels","Back to Module List":"Retour à la liste des modules","Back to Repository Details":"Retour aux détails de référentiels","Back to Repository List":"Retour à la liste des référentiels","Back to Version List":"Retour à la liste des versions","Basic Information":"Informations de base","Below are the product's repository content sets currently available for this subscription. For Red Hat subscriptions, additional content can be made available through the":"Les ensembles de contenus de référentiels de produits actuellement disponibles  pour cet abonnement sont situés ci-dessous. Pour les abonnements Red Hat, du contenu supplémentaire est disponible via","Below are the repository content sets currently available for this activation key through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"Les ensembles de contenus de référentiels actuellement disponibles  pour cette clé d'activation à travers ses abonnements sont situés ci-dessous. Pour les abonnements Red Hat, du contenu supplémentaire est disponible via","Below are the repository content sets currently available for this content host through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"Les ensembles de contenus de référentiels actuellement disponibles  pour cet hôte de contenu à travers ses abonnements sont situés ci-dessous. Pour les abonnements Red Hat, du contenu supplémentaire est disponible via","Bug Fix":"Correctif de bogue","Bug Fix Advisory":"Alerte de correctif de bogue","Bugfix":"Correction de bogues","Bulk Actions":"Actions par lots","Bulk Alter Products Sync Plan":"Modifier le plan de synchronisation de produits par lots","Bulk Sync Products":"Synchroniser les produits par lots","CVEs":"CVE","Cancel":"Annuler","Capacity":"Capacité","Change assigned Lifecycle Environment or Content View":"Modifier l'environnement de cycle de vie attribué ou l'affichage du contenu","Changing default settings for content hosts that register with this activation key requires subscription-manager version 1.10 or newer to be installed on that host.":"Une version 1.10 ou ultérieure de subscription-manager doit être installée sur cet hôte pour modifier les paramètres par défaut pour les hôtes de contenu qui s'enregistrent avec cette clé d'activation.","Changing default settings requires subscription-manager version 1.10 or newer to be installed on this host.":"Une version 1.10 ou ultérieure de subscription-manager doit être installée sur cet hôte pour modifier les paramètres par défaut.","Checkin":"Archiver","Checksum":"Somme de contrôle","Choose a lifecycle environment from the available promotion paths.":"Choisir un environnement de cycle de vie à partir des chemins d'accès de promotion disponibles.","Click here to check the status of the task.":"Cliquez ici pour vérifier le statut de la tâche.","Click here to refresh.":"Cliquer ici pour actualiser.","Click here to see the task for the successful removals.":"Cliquez ici pour vérifier la tâche des suppressions réussies.","Click here to select Errata for an Incremental Update.":"Cliquez ici pour sélectionner les erratas pour une mise à jour croissante.","Close":"Fermer","Collection Actions":"Actions de collection","Completely remove version?":"Supprimer la version entièrement ?","Components":"Composants","Components:":"Composants :","Composite Content View {{ contentView.name }}":"Affichage du contenu composite {{ contentView.name }}","Composite View":"Affichage  composite","Composite View?":"Affichage composite ?","Composite?":"Composite ?","Confirm":"Confirmer","Confirm Remove":"Confirmer la suppression","Confirm Version Removal: Version {{ version.version }}":"Confirmer la suppression de la version : Version {{ version.version }}","Confirmation":"Confirmation","Consumed":"Consommé","Content":"Contenu","Content Counts":"Nombre de contenus","Content Download URL":"URL du téléchargement de contenu","Content Host":"Hôte du contenu","Content Host %s has been deleted.":"L'hôte du contenu %s a été supprimé.","Content Host Bulk Actions":"Actions par lots d'hôte de contenu","Content Host Bulk Content":"Contenu par lot d'hôte de contenu","Content Host Bulk Errata":"Errata par lots d'hôte de contenu","Content Host Bulk Host Collections":"Collections d'hôtes par lots d'hôtes de contenu","Content Host Bulk Packages":"Packages par lots d'hôte de contenu","Content Host Bulk Subscriptions":"Abonnements par lots d'hôte de contenu","Content Host Content":"Contenu d'hôte de contenu","Content Host Counts":"Nombre d'hôtes de contenu","Content Host Errata Management":"Gestion d'errata d'hôtes de contenu","Content Host Limit":"Limite d'hôte de contenu","Content Host Package Management":"Gestion de packages d'hôte de contenu","Content Host Properties":"Propriétés d'hôte de contenu","Content Host Registration":"Enregistrement d'hôte de contenu","Content Host Status":"Statut d'hôte de contenu","Content Host Subscriptions":"Abonnements d'hôte de contenu","Content Host {{ contentHost.name }}":"Hôte de contenu {{ contentHost.name }}","Content Host:":"Hôte de contenu :","Content Hosts":"Hôtes de contenu","Content Hosts for Activation Key:":"Hôtes de contenu pour clé d'activation :","Content Hosts for Subscription:":"Hôtes de contenu pour abonnement :","Content Hosts for:":"Hôtes de contenu pour :","Content Hosts using Version {{ version.version }}":"Hôtes de contenu utilisant la Version {{ version.version }}","Content Type":"Type de contenu","Content View":"Affichage du contenu","Content View Name":"Nom de l'affichage de contenu","Content View Version":"Version de l'affichage de contenu","Content View updated.":"Affichage du contenu mis à jour.","Content View:":"Affichage du contenu :","Content Views":"Affichages du contenu","Content Views <div>{{ library.counts.content_views || 0 }}</div>":"Affichages de contenu <div>{{ library.counts.content_views || 0 }}</div>","Content Views for Composite Content View:":"Affichages du contenu pour l'affichage du contenu composite :","Content successfully uploaded":"Contenu mis à jour","Contents":"Contenu","Contract":"Contrat","Contract Number":"Numéro de contrat","Copy":"Copier","Copy Activation Key":"Copier la clé d'activation","Copy Collection":"Copier la collection","Copy View":"Copier l'affichage","Cores per Socket":"Noyaux par Socket","Cores: %s":"Cœurs : %s","Create":"Créer","Create New View":"Créer un nouvel affichage","Create Repositories":"Créer des référentiels","Create Repositories Within:":"Créer des référentiels dans :","Create Repository":"Créer le référentiel","Create Selected":"Créer sélectionné","Create Sync Plan":"Créer un plan de synchronisation","Create new view":"Créer un nouvel affichage","Create view from existing views":"Créer un affichage à partir d'affichages existants","Critical":"Critique","Current Environment (%e/%cv)":"Environnement actuel (%e/%cv)","Current Subscriptions for Activation Key:":"Abonnements actuels pour la clé d'activation :","Currently Selected Puppet Modules":"Modules Puppet sélectionnés actuellement","Daily at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"Chaque jour à {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)","Date":"Date","Default":"Défaut","Delete":"Supprimer","Delete Manifest":"Supprimer le manifeste","Delete Version {{ version.version }}":"Supprimer la version {{ version.version }}","Delete {{ contentView.name }}":"Supprimer {{ contentView.name }}","Deleted from %s":"Supprimé de %s","Deleting Manifest...":"Suppression du fichier manifeste...","Deleting from %count environments.":"Suppression à partir de %count  environnements ","Deleting from 1 environment.":"Suppression à partir de 1 environnement.","Deletion from %s":"Suppression à partir de %s","Description":"Description","Detail":"Détail","Details":"Détails","Details for Activation Key:":"Détails de la clé d'activation :","Details for Content View:":"Détails de l'affichage du contenu :","Details for Product:":"Détails du produit :","Discover":"Découvrir","Discovered URLs":"URL découvertes","Docker Content":"Contenu Docker","Docker Images":"Images Docker","Docker Images in {{ repository.name }}":"Images Docker dans {{ repository.name }}","Docker Repositories":"Référentiels Docker","Docker Repositories <div>{{ library.counts.docker_repositories || 0 }}</div>":"Référentiels Docker <div>{{ library.counts.docker_repositories || 0 }}</div>","Docker Tag:":"Balise Docker :","Docker Tags":"Balises Docker","Docker metadata generation has been initiated in the background.  Click\n        <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"La génération de métadonnées Docker a été initiée en arrière-plan. Cliquez <a ng-href=\"{{ taskUrl() }}\">ici</a> pour vérifier la progression.","Edit":"Modifier","Enabled?":"Activé ?","End Date":"Date de fin","Ends":"Se termine","Enhancement":"Amélioration","Enter Package Group Name(s)...":"Veuillez saisir le(s) nom(s) de groupe(s) de packages...","Enter Package Name(s)...":"Veuillez saisir le(s) nom(s) de package(s)...","Environment":"Environnement","Environment saved":"Environnement enregistré","Environment: {{ environment.name }}":"Environment : {{ environment.name }}","Environments":"Environnements","Environments List":"Liste des environnements","Environments:":"Environnements :","Equal To":"Égal à","Errata":"Errata","Errata <div>{{ library.counts.errata.total || 0 }}</div>":"Errata <div>{{ library.counts.errata.total || 0 }}</div>","Errata Details":"Détails d'errata","Errata Filter:":"Filtre d'errata :","Errata ID":"ID de l'errata","Errata Installation":"Installation d'errata","Errata List":"Liste des errata","Errata Task List":"Liste des tâches d'errata","Errata Tasks":"Tâches d'errata","Errata Type":"Type d'errata","Errata for:":"Errata pour :","Errata successfully added.":"Errata ajoutés.","Errata successfully removed.":"Errata supprimés.","Errata:":"Errata :","Erratum - by Date and Type":"Erratum : par date et par type","Erratum - by ID":"Erratum : par id","Erratum Date Range":"Plage de dates d'erratum","Error deleting manifest.":"Erreur lors de la suppression du fichier manifeste.","Error during upload:":"Erreur lors du téléchargement :","Error importing manifest.":"Erreur lors de l'importation du fichier manifeste.","Error refreshing manifest.":"Erreur lors de l'actualisation du fichier manifeste.","Event":"Événement","Events":"Événements","Exclude":"Exclure","Existing Product":"Produit existant","Expires":"Expire","Failed":"Échec","File too large. Please use the CLI instead.":"Fichier trop volumineux. Veuillez utiliser le CLI.","Filter":"Filtre","Filter Saved":"Filtre enregistré","Filter information for:":"Information de filtre pour :","Filter...":"Filtre...","Filter:":"Filtre :","Filters":"Filtres","Filters List":"Liste de filtres","Filters for Content View:":"Filtres pour affichage du contenu :","Filters successfully removed.":"Filtres supprimés.","Finished At":"Fini à","For older operating systems such as Red Hat Enterprise Linux 5 or CentOS 5 it is recommended to use sha1.":"Pour les systèmes d'exploitation plus anciens, tels que Red Hat Enterprise Linux 5 ou CentOS 5, il est recommandé d'utiliser sha1.","GPG Key":"Clé GPG","GPG Key URL":"URL de clé GPG","GPG Key successfully uploaded":"Clé GPG téléchargée","GPG Key {{ gpgKey.name }}":"Clé GPG {{ gpgKey.name }}","GPG Keys":"Clés GPG","General Information":"Informations générales","Generated On":"Généré le","Gpg Key updated":"Clé GPG mise à jour","Gpg Keys":"Clés GPG","Greater Than":"Supérieur à","Group Install":"Installation de groupe","Group Remove":"Suppression de groupe","Guests of":"Invités de","Hide affected Activation Keys":"Masquer les clés d'activation affectées","Hide affected Content Hosts":"Masquer les hôtes de contenu affectés","History":"Historique","History for Content View:":"Historique de l'affichage du contenu :","Host Collection Management":"Gestion de collection d'hôtes","Host Collection Membership":"Abonnement de collection d'hôtes","Host Collection Membership Management":"Gestion d'abonnement de collection d'hôtes","Host Collection removed.":"Collections d'hôtes supprimée","Host Collection updated":"Collection d'hôtes mise à jour","Host Collection:":"Collection d'hôtes :","Host Collection: {{ hostCollection.name }}":"Collection d'hôtes : {{ hostCollection.name }}","Host Collections":"Collections d'hôtes","Host Collections for:":"Collections d'hôtes pour :","Host Count":"Nombre d'hôtes","Host Group":"Groupe d'hôtes","Hostname":"Nom d'hôte","Hourly at {{ product.sync_plan.sync_date | date:'m' }} minutes and {{ product.sync_plan.sync_date | date:'s' }} seconds":"Chaque heure à {{ product.sync_plan.sync_date | date:'m' }} minutes et {{ product.sync_plan.sync_date | date:'s' }} secondes","IPv4 Address":"Adresse IPv4","IPv6 Address":"Adresse IPv6","Id":"Id","If you would prefer to move portions of these Activation Keys to different Content Views or Lifecycle Environments click <a href=\"{{ activationKeyLink() }}\">here</a> to manage them individually.":"Si vous préférez déplacer des portions de ces clés d'activations vers différents affichages du contenu ou environnements de cycle de vie, cliquez  <a href=\"{{ activationKeyLink() }}\">ici</a> pour les gérer individuellement.","If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click":"Si vous préférez déplacer des portions de ces clés d'activations vers différents affichages du contenu ou environnements de cycle de vie, cliquez","If you would prefer to move portions of these Content Hosts to different content views or environments click <a href=\"{{ contentHostsLink() }}\">here</a>  to manage these Content Hosts in bulk.":"Si vous préférez déplacer des portions de ces hôtes de contenu vers différents environnements ou affichages du contenu, cliquez <a href=\"{{ contentHostsLink() }}\">ici</a> pour gérer ces hôtes de contenu par lots.","Image":"Image","Image ID":"ID de l'image","Import History":"Importer l'historique","Important":"Important","In order to add a content view to a composite view you must first publish an initial version of the content view.":"Pour ajouter un affichage du contenu à un affichage composite, vous devez d'abord publier une version initiale de l'affichage du contenu.","Include":"Inclure","Included in Activation Keys":"Inclus dans les clés d'activation","Incremental Update":"Mise à jour croissante","Install":"Installer","Install Selected":"Installer sélectionné","Install katello-agent for remote actions and displaying errata information:":"Installez katello-agent pour les actions distantes et l'affichage d''informations d'erratas :","Install the pre-built bootstrap RPM:":"Installez le RPM bootstrap existant :","Installable":"Installable","Installable Errata":"Errata installable","Installed":"Installé","Installed Package":"Package installé","Installed Packages":"Packages installés","Installed Products":"Produits installés","Instance-based":"Basé sur instance","Interfaces":"Interfaces","Interval":"Intervalle","Issued":"Émis","Katello Agent":"Agent Katello","Label":"Étiquette","Last Checkin":"Dernière connexion","Last Published":"Dernière publication","Last Puppet Report":"Dernier rapport puppet","Last Sync":"Dernière synchronisation","Last Updated":"Dernière mise à jour","Last Updated On":"Dernière mise à jour effectuée le","Last synced {{ product.last_sync_words }} ago.":"Dernière synchronisation il y a {{ product.last_sync_words }}.","Latest (Currently %s)":"Les (Currently %s) les plus récents","Less Than":"Inférieur à","Library":"Bibliothèque","Library Synced Content":"Contenu synchronisé de bibliothèque","Lifecycle Environment \"{{ priorEnvironment.name }}\" exists within the middle of a path. New Lifecycle Environments can only be added to the end of a path.":"L'environnement de cycle de vie \"{{ priorEnvironment.name }}\" existe au milieu d'un chemin d'accès. Les nouveaux environnements de cycle de vie peuvent uniquement être ajoutés à la fin d'un chemin d'accès.","Lifecycle Environment Paths":"Chemins d'environnement de cycle de vie","Lifecycle Environments":"Environnements de cycle de vie","Limit":"Limite","Limit:":"Limite :","Limits":"Limites","List/Remove":"Répertorier/Supprimer","Loading Environment Paths...":"Chargement de chemins d'environnement...","Loading...":"Chargement...","Manage Docker Images":"Gestion d'Images Docker","Manage Docker Images for Repository:":"Gérez les images Docker pour les référentiels :","Manage Manifest":"Gérer le fichier manifeste","Manage Packages":"Gérer les packages","Manage Packages for Repository:":"Gérer les packages pour les référentiels :","Manage Puppet Modules":"Gérez les modules Puppet","Manage Puppet Modules for Repository:":"Gérez les modules Puppet pour les référentiels :","Manifest History":"Historique du manifeste","Manifest successfully deleted.":"Fichier manifeste supprimé.","Manifest successfully imported.":"Fichier manifeste importé.","Manifest successfully refreshed.":"Fichier manifeste actualisé.","Message":"Message","Model":"Modèle","Moderate":"Modéré","Module %s removed from Content View.":"Module %s supprimé de l'affichage du contenu.","Module List":"Liste de module","More Details":"Plus de détails","More...":"Plus...","Multi-entitlement":"Droits d'accès multiples","N/A":"N/A","Name":"Nom","Name of the upstream repository you want to sync. Example: 'busybox' or 'fedora/ssh'.":"Nom du référentiel en amont que vous souhaitez synchroniser. Exemple : 'busybox' ou 'fedora/ssh'.","Networking":"Réseau","Never synced":"Jamais synchronisé","New Activation Key":"Nouvelle clé d'activation","New Content View":"Nouvel affichage du contenu","New Environment":"Nouvel environnement","New Environment Path":"Nouveau chemin d'environnement ","New Filter":"Nouveau filtre","New Filter for Content View:":"Nouveau filtre pour l'affichage du contenu :","New GPG Key":"Nouvelle clé GPG","New Gpg Key":"Nouvelle clé GPG","New Host Collection":"Nouvelle collection d'hôte","New Name:":"Nouveau nom :","New Product":"Nouveau produit","New Repository":"Nouveau référentiel","New Sync Plan":"Nouveau plan de synchronisation","New sync plan successfully created.":"Nouveau plan de synchronisation créé.","Next":"Suivant","Next Sync":"Prochaine synchronisation","No":"Non","No \"Edit\" permissions on some of the Activation Keys in that Lifecycle Environment.":"Aucune permission de « Modification » sur certaines clés d'activation dans cet environnement de cycle de vie.","No \"Edit\" permissions on some of the Content Hosts in that Lifecycle Environment.":"Aucune permission de « Modification » sur certains hôtes de contenu dans cet environnement de cycle de vie.","No \"Promote/Remove\" permissions on the Lifecycle Environment.":"Aucune permission de « Promouvoir/Supprimer » sur l'environnement de cycle de vie.","No (Default)":"Non (par défaut)","No Content Hosts are affected.":"Aucun hôte de contenu affecté","No Content Hosts match this Erratum.":"Aucun hôte de contenu ne correspond à cet erratum.","No Content Views available, please select another environment.":"Aucun affichage du contenu disponible, veuillez sélectionner un autre environnement","No Host Collections to show, you can add Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"Aucune collection d'hôte à afficher. Vous pouvez ajouter des collections d'hôtes après avoir sélectionné « Collections d'hôtes » sous « Hôtes » dans le menu principal.","No Host Collections to show, you can add Host Collections after selecting the 'Add' tab.":"Aucune collection d'hôte à afficher. Vous pouvez ajouter des collections d'hôtes après avoir sélectionné l'onglet « Ajouter ».","No Products to show.":"Aucun produit à afficher","No Repositories contain this Erratum.":"Aucun référentiel ne contient cet erratum.","No activation keys are affected.":"Aucune clé d'activation n'est affectée.","No alternate release version choices are available. The available releases are based upon what is available in \"{{ contentHost.content_view.name }}\", the selected <a href=\"/content_views\">content view</a> this content host is attached to for the given <a href=\"/lifecycle_environments\">lifecycle environment</a>, \"{{ contentHost.environment.name }}\".":"Aucune version alternative n'est disponible. Les versions alternatives se basent sur ce qui est disponible dans  \"{{ contentHost.content_view.name }}\", l'<a href=\"/content_views\">affichage du contenu</a> sélectionné auquel l'hôte de contenu est attaché pour l'<a href=\"/lifecycle_environments\">environnement de cycle de vie</a> donné, \"{{ contentHost.environment.name }}\".","No content views exist for {{selected.environment.name}}":"Il n'existe aucun affichage du contenu pour {{selected.environment.name}}.","No host collections":"Aucune collection d'hôtes","No provisioning details are available for this content host.":"Aucun détail de provisioning n'est disponible pour cet hôte de contenu.","No repository content provided through subscriptions.":"Aucun contenu de référentiel n'est fourni avec les abonnements.","No repository content provided through this subscription.":"Aucun contenu de référentiel fourni avec cet abonnement.","No repository content.":"Aucun contenu de référentiel.","No subscription manifest imported":"Aucun manifeste des abonnements n'a été importé","None":"Aucun","Not Synced":"Pas synchronisé","Not installed":"Non installé.","Not yet published":"Pas encore publié","Number of CPUs":"Nombre de CPU","OS":"SE","Only show Errata that is Applicable to one or more Content Hosts":"N'afficher que les errata applicables à un ou plusieurs hôtes de contenu.","Only show Errata that is Installable on one or more Content Hosts":"N'afficher que les errata pouvant être installés sur un ou plusieurs hôtes de contenu","Operating System":"Système d'Exploitation","Options:":"Options :","Organization":"Organisation","Original Sync Date":"Date de synchronisation d'origine","Override to No":"Remplacer par non","Override to Yes":"Remplacer par oui","Package":"Package","Package Actions":"Actions de packages","Package Filter:":"Filtre de package :","Package Group":"Groupe de packages","Package Group Filter:":"Filtre de groupe de package :","Package Group successfully added.":"Groupe de package ajouté.","Package Group successfully removed.":"Groupe de package supprimé.","Package Groups":"Groupes de packages","Package Install":"Installation de packages","Package Installation, Removal, and Update":"Installation, suppression et mise à jour de package","Package Remove":"Suppression de packages","Package Update":"Mise à jour de packages","Package removal failed":"Suppression de package échouée","Package successfully added.":"Package ajouté.","Package successfully removed.":"Package supprimé.","Package successfully updated.":"Package mis à jour.","Package/Group Name":"Nom de package/groupe","Packages":"Packages","Packages <div>{{ library.counts.packages || 0 }}</div>":"Packages <div>{{ library.counts.packages || 0 }}</div>","Packages for:":"Packages pour :","Packages in {{ repository.name }}":"Packages dans {{ repository.name }}","Parameters":"Paramètres","Paste GPG Key Contents":"Coller le contenu de la clé GPG","Paste contents...":"Coller le contenu...","Paused":"En pause","Perform":"Effectuer","Physical":"Physique","Please select a Lifecycle Environment and Content View to move the affected Activation Keys to:":"Veuillez sélectionner un environnement de cycle de vie et un affichage du contenu pour déplacer les clés d'activation affectées vers :","Please select a Lifecycle Environment and Content View to move these Content Hosts to:":"Veuillez sélectionner un environnement de cycle de vie et un affichage du contenu pour déplacer ces hôtes de contenu vers :","Please select a Lifecycle Environment.":"Veuillez sélectionner un environnement de cycle de vie.","Please select an environment.":"Veuillez sélectionner un environnement","Please select one from the list below and you will be redirected.":"Veuillez en sélectionner un à partir de la liste ci-dessous pour être redirigé.","Press 'Sync Now' below to begin syncing the following product(s).":"Veuillez appuyer sur le bouton « Synchroniser » ci-dessous pour commencer à synchroniser le(s) produit(s) suivant(s).","Previous Environment (%e/%cv)":"Environnement précédent (%e/%cv)","Processing":"Traitement","Product":"Produit","Product Bulk Actions":"Actions de produits par lots","Product Content":"Contenu de produit","Product Count":"Nombre de produits","Product Enhancement Advisory":"Alerte d'amélioration de produit","Product ID":"ID de produit","Product Management":"Gestion de produits","Product Management for Sync Plan:":"Gestion de produits pour le plan de synchronisation :","Product Name":"Nom de produit","Product Saved":"Produit enregistré","Product Sync":"Synchronisation de produit","Product delete operation has been initiated in the background.  Click <a href=\"{{ taskUrl(productDeletionTaskId) }}\">Here</a> to monitor the progress.":"L'opération de suppression de produit a démarré en arrière-plan. Cliquez <a href=\"{{ taskUrl(productDeletionTaskId) }}\">ici</a> pour vérifier sa progression.","Product information for:":"Information de produit pour :","Product {{ product.name }}":"Produit {{ product.name }}","Products":"Produits","Products <div>{{ library.counts.products || 0 }}</div>":"Produits <div>{{ library.counts.products || 0 }}</div>","Promote":"Promouvoir","Promote Content View:":"Promouvoir l'affichage du contenu","Promote Version":"Promouvoir la version","Promote Version {{ version.version }}":"Promouvoir la version {{ version.version }}","Promoted to %s":"Promu à %s","Promoting to %count environments.":"Promouvoir à %count environnements.","Promoting to 1 environment.":"Promouvoir à 1 environnement.","Promotion History":"Historique de promotion","Provided Products":"Produits fournis","Provisioning Details":"Détails de provisioning","Provisioning Host Details":"Détails d'hôte de provisioning","Publish Content View:":"Publier l'affichage du contenu :","Publish New Version":"Publier la nouvelle version","Publish via HTTP":"Publier via HTTP","Publish via HTTPS":"Publier via HTTPS","Published":"Publié le","Published At":"Publié à","Published Repository Information":"Informations de référentiel publiées","Published new version":"Nouvelle version publiée","Publishing and promoting to %count environments.":"Publier et promouvoir à %count environnements.","Publishing and promoting to 1 environment.":"Publier et promouvoir à 1 environnement.","Puppet Environment":"Environnement Puppet","Puppet Modules":"Modules Puppet","Puppet Modules <div>{{ library.counts.puppet_modules || 0 }}</div>":"Modules puppet <div>{{ library.counts.puppet_modules || 0 }}</div>","Puppet Modules for Content View:":"Modules puppet pour l'affichage du contenu :","Puppet Modules in {{ repository.name }}":"Modules puppet dans {{ repository.name }}","Puppet module added to Content View":"Module puppet ajouté à l'affichage du contenu","Puppet module metadata generation has been initiated in the background.\n        Click <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"La génération de métadonnées de module Puppet a été initiée en arrière-plan.\n        Cliquez <a ng-href=\"{{ taskUrl() }}\">ici</a> pour vérifier la progression.","Quantity":"Quantité","RAM (GB)":"RAM (Go)","RAM: %s GB":"RAM : %s Go","Range":"Gamme","Reboot Suggested":"Redémarrage suggéré","Reboot Suggested?":"Redémarrage suggéré ?","Red Hat CDN URL":"URL du CDN (« Content Distribution Network », réseau de distribution de contenu)  de Red Hat","Red Hat Provider Details":"Détails du fournisseur Red Hat","Red Hat Repositories page":"Page des référentiels Red Hat","Red Hat Repositories page.":"Page des référentiels Red Hat.","Refresh Manifest":"Actualiser le manifeste","Refresh Table":"Actualiser la table","Refreshing Manifest...":"Actualisation du manifeste...","Refreshing manifest not suppported by this manifest.":"Actualisation du fichier manifeste non pris en charge par ce fichier manifeste.","Register Content Host":"Enregistrer l'hôte de contenu","Register a Content Host":"Enregistrer un hôte de contenu","Registered":"Enregistré","Registered By":"Enregistré par","Registry URL":"URL de registre","Release":"Version","Release Version":"Version de publication","Remove":"Supprimer","Remove Activation Key \"{{ activationKey.name }}\"?":"Supprimer la  clé d'activation \"{{ activationKey.name }}\" ?","Remove Content View":"Supprimer l'affichage du contenu","Remove Content Views":"Supprimer les affichages du contenu","Remove Docker Images":"Supprimer les images Docker","Remove Errata":"Supprimer les errata","Remove Filters":"Supprimer les filtres","Remove From":"Supprimer de","Remove GPG Key":"Supprimer la clé GPG","Remove GPG Key {{ gpgKey.name }}":"Supprimer la clé GPG {{ gpgKey.name }}","Remove Host Collection \"{{ hostCollection.name }}\"?":"Supprimer la collection d'hôtes \"{{ hostCollection.name }}\" ?","Remove Module":"Supprimer le module","Remove Package Group":"Supprimer le groupe de package","Remove Packages":"Supprimer les packages","Remove Product":"Supprimer le produit","Remove Product \"{{ product.name }}\"?":"Supprimer le produit \"{{ product.name }}\" ?","Remove Products":"Supprimer les produits","Remove Puppet Modules":"Supprimer les modules Puppet","Remove Repositories":"Supprimer les référentiels","Remove Repository":"Supprimer le référentiel","Remove Repository \"{{ repository.name }}\"?":"Supprimer le référentiel \"{{ repository.name }}\" ?","Remove Selected":"Supprimer la sélection","Remove Successful.":"Suppression réussie.","Remove Sync Plan":"Supprimer le plan de synchronisation","Remove Sync Plan \"{{ syncPlan.name }}\"?":"Supprimer le plan de synchronisation \"{{ syncPlan.name }}\" ?","Remove Version":"Supprimer la version","Remove Version Confirmation":"Supprimer la confirmation de la version","Remove View":"Supprimer l'affichage","Remove {{ productTable.getSelected().length }} Products?":"Supprimer {{ productTable.getSelected().length }} produits ?","Removed %x host collections from activation key \"%y\".":"%x collections d'hôtes supprimées de la clé d'activation \"%y\".","Removed %x host collections from content host \"%y\".":"%x collections d'hôtes supprimées de l'hôte de contenu \"%y\".","Removed %x products from sync plan \"%y\".":"%x produits supprimés du plan de synchronisation \"%y\".","Removing Content Hosts":"Supprimer les hôtes de contenu","Removing Products":"Supprimer les produits","Removing Repositories":"Supprimer les référentiels","Repo Discovery":"Découverte de référentiels","Repo Type":"Type de référentiels","Repositories":"Référentiels","Repositories containing Errata {{ errata.errata_id }}":"Référentiels contenant des Errata {{ errata.errata_id }}","Repositories for Content View:":"Référentiels pour l'affichage du contenu :","Repositories for Errata:":"Référentiels pour errata :","Repositories for Filter:":"Référentiels pour filtre :","Repositories for Product:":"Référentiels pour produit :","Repository":"Référentiel","Repository \"%s\" successfully deleted":"Le référentiel \"%s\" a été supprimé avec succès.","Repository %s successfully created.":"Référentiel %s créé.","Repository Count":"Nombre de référentiels","Repository Discovery":"Découverte de référentiels","Repository Name":"Nom du référentiel","Repository Saved.":"Référentiel enregistré.","Repository Selection":"Sélection de référentiel","Repository URL updated":"URL de référentiel mis à jour","Repository:":"Référentiel :","Required Host":"Hôte requis","Result":"Résultat","Results":"Résultats","Rules Source":"Source des règles","Rules Version":"Version des règles","Run Auto-Attach":"Exécuter Auto-attach","Run subscription-manager in a console on the client host.  You may use an Activation Key to register:":"Exécuter subscription-manager dans une console sur l'hôte client. Vous pouvez utiliser une clé d'activation pour enregistrer :","Save Successful.":"Enregistrement réussi.","Search...":"Rechercher...","Security":"Sécurité","Security Advisory":"Alerte de sécurité","Select A New Puppet Module To Add":"Sélectionner un nouveau module puppet à ajouter","Select Content Hosts":"Sélectionnez les hôtes de contenu","Select Module":"Sélectionner module","Select Version":"Sélectionner version","Select a Content Source:":"Sélectionnez une source de contenu :","Select a Version":"Sélectionner une version","Select an Available Version of {{ $stateParams.moduleName }}":"Sélectionner une version disponible de {{ $stateParams.moduleName }}","Select an Organization":"Sélectionner une organisation","Select new version":"Sélectionner une nouvelle version","Select the Lifecycle Environments you would like to remove Version {{ version.version }} from:":"Sélectionner les environnements de cycle de vie desquels vous souhaiteriez supprimer Version {{ version.version }} :","Selected URLs:":"URL sélectionnés :","Selecting this option will result in the Version being completely deleted and no longer being available for promotion.  The version must be removed from all Lifecycle Environments in order to select this option.":"La sélection de cette option entraînera la suppression totale de la Version et ne la rendra plus disponible à la promotion. La version doit être supprimée de tous les environnements de cycle de vie pour sélectionner cette option.","Serve via HTTP.":"Servir par HTTP.","Service Level":"Niveau de service","Severity":"Gravité","Show affected Activation Keys":"Afficher les clés d'activation affectées","Show affected Content Hosts":"Afficher les hôtes de contenu affectés","Show from:":"Afficher à partir de :","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"Afficher {{ detailsTable.rows.length }} sur {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} au total) hôtes de contenu","Showing {{ addSubscriptionsTable.rows.length }} of {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} Total) Subscriptions":"Afficher {{ addSubscriptionsTable.rows.length }} sur {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} au total) abonnements","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"Afficher {{ detailsTable.rows.length }} sur {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} au total) hôtes de contenu","Showing {{ currentPackagesTable.resource.total }} Packages":"Afficher {{ currentPackagesTable.resource.total }} packages","Showing {{ subscriptionsTable.rows.length }} of {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} Total) Subscriptions":"Afficher {{ subscriptionsTable.rows.length }} sur {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} au total) abonnements","Sockets":"Sockets","Sockets: %s":"Sockets : %s","Solution":"Solution","Some of the Errata shown below may not be installable as they are not in this Content Host's\n          Content View and Lifecycle Environment.  In order to apply such Errata an Incremental Update is required.":"Certains errata affichés ci-dessous ne peuvent pas être installés puisqu'ils n'appartiennent pas à l'affichage de contenu\n  et environnement de cycle de vie  de l'hôte de contenu. Pour appliquer cet errata, une mise à jour croissante est requise.","Stacking ID":"ID de la pile","Start Date":"Date de début","Start Time":"Date de lancement","Started At":"Date de début","Starting":"Démarrage","Starts":"Commence","State":"État ","Status":"État","Subscription Details":"Détails de l'abonnement","Subscription Info":"Informations d'abonnement","Subscription Manifest":"Manifeste des abonnements","Subscription Server Details":"Détails du serveur des abonnements","Subscription Status":"Statut de l'abonnement","Subscriptions":"Abonnements","Subscriptions for Activation Key:":"Abonnements pour la clé d'activation :","Subscriptions for:":"Abonnements pour :","Succesfully scheduled package installation":"Installation du package programmée.","Succesfully scheduled package removal":"Suppression du package programmée.","Succesfully scheduled package update":"Mise à jour du package programmée.","Success":"Réussite","Successfully Scheduled Auto-attach.":"Attachement automatique programmé.","Successfully added %s subscriptions.":"%s abonnements ajoutés.","Successfully deleted %cv version %ver.":"%cv version %ver supprimée.","Successfully initiated deletion of %cv version %ver.":"Suppression de %cv version %ver initiée.","Successfully initiated promotion of %cv version %ver to %env.":"Promotion de %cv version %ver à %env initiée.","Successfully initiated removal of %cv version %ver.":"Suppression de %cv version %ver initiée.","Successfully promoted %cv version %ver to %env":"%cv version %ver promue à %env.","Successfully published %cv version %ver and promoted to Library":"%cv version %ver publiée et promue à la Bibliothèque.","Successfully removed %cv version %ver from environments: %env":"%cv version %ver supprimée des environnements : %env","Successfully removed %s items.":"%s éléments supprimés.","Successfully removed %s subscriptions.":"%s abonnements supprimés.","Successfully removed 1 item.":"1 élément supprimé.","Successfully updated subscriptions.":"Abonnements mis à jour.","Summary":"Sommaire","Support Level":"Niveau de service","Support Type":"Type de support","Sync Enabled":"Synchronisation activée.","Sync Information":"Informations de synchronisation.","Sync Interval":"Intervalle de synchronisation","Sync Now":"Synchroniser","Sync Plan":"Plan de synchronisation","Sync Plan %s has been deleted.":"Le plan de synchronisation %s a été supprimé.","Sync Plan Saved":"Plan de synchronisation enregistré","Sync Plan {{ syncPlan.name }}":"Plan de synchronisation {{ syncPlan.name }}","Sync Plan:":"Plan de synchronisation :","Sync Plans":"Plans de synchronisation","Sync State":"État de synchronisation","Sync Status":"Statut de la synchronisation","Synced manually, no interval set.":"Synchronisé manuellement, aucun intervalle défini.","Syncing...":"Synchronisation...","Tags":"Balises","Target":"Cible","Task Details":"Détails de la tâche","Tasks":"Tâches","Tasks for Content View:":"Tâches pour l'affichage du contenu :","Temporary":"Temporaire","The Activation Keys listed below are currently using this Content View Version.  Before deleting the Version you must move these Activation Keys to a Lifecycle Environment where this Version is not in use.":"Les clés d'activation répertoriées ci-dessous utilisent actuellement cette version d'affichage du contenu. Avant de supprimer la version, veuillez déplacer ces clés d'activation vers un environnement de cycle de vie où cette version n'est pas utilisée.","The Content Hosts listed below are currently using this Content View Version.  Before removing the version you must move these Content Hosts to an environment where this version is not in use.":"Les hôtes de contenu répertoriés ci-dessous utilisent actuellement cette version d'affichage du contenu. Avant de supprimer cette version, veuillez déplacer ces hôtes de contenu vers un environnement où cette version n'est pas utilisée.","The Content View or Lifecycle Environment needs to be updated in order to make errata available to these hosts.":"L'affichage de contenu ou environnement de cycle de vie doit être mis à jour pour rendre les errata disponibles pour ces hôtes.","The archive of this Version will be deleted.  This Version will not be available once deletion is complete.":"L'archive de cette version sera supprimée. Cette version ne sera plus disponible une fois la suppression terminée.","The archive of this Version will not be deleted.  This Version will still be available to be promoted to a Lifecycle Environment.":"L'archive de cette version ne sera pas supprimée. Cette version pourra toujours être promue à l'environnement du cycle de vie.","The following actions can be performed on content hosts in this host collection:":"Les actions suivantes peuvent être effectuées sur les hôtes de contenu dans cette collection d'hôtes :","The host requires being attached to a content view and the environment you have chosen has no content views promoted to it.\n            See the <a href=\"/content_views\">content views page</a> to manage and promote a content view.":"L'hôte doit être relié à un affichage de contenu, or l'environnement que vous avez choisi ne possède pas d'affichage de contenu.\n     Veuillez consulter la <a href=\"/content_views\">page d'affichage de contenu</a> pour gérer et attribuer un affichage de contenu.","The katello-agent package is required to manage errata on this Host.\n      Please install the katello-agent package to manage errata on\n      {{ contentHost.name }}.":"Le package katello-agent est requis pour gérer les errata sur cet hôte.\n      Veuillez installer le package  katello-agent pour gérer les errata sur\n      {{ contentHost.name }}.","The katello-agent package is required to manage packages on this Host.\n      Please install the katello-agent package to manage packages on\n      {{ contentHost.name }}. To install the package, connect directly to the\n      host from a terminal and use yum.":"Le package katello-agent est nécessaire pour gérer les packages sur cet Hôte.\n   Veuillez installer le package  katello-agent pour gérer les packages sur\n      {{ contentHost.name }}. Pour installer le package, connectez-vous directement à l'hôte\n      à partir d'un terminal et utilisez yum.","The page you are attempting to access requires selecting a specific organization.":"Veuillez sélectionner une organisation spécifique afin d'accéder à cette page.","The selected environment contains no Content Views, please select a different environment.":"L'environnement sélectionné ne contient pas d'affichage du contenu. Veuillez sélectionner un autre environnement.","There are currently no repositories associated with this Content View, you can add some by clicking on the \"Add\" tab above.":"Il n'y a actuellement aucun référentiel associé à cet affichage du contenu. Vous pouvez en ajouter quelques uns en cliquant sur l'onglet « Ajouter » ci-dessus.","There are currently no repositories to add to this Content View,":"Il n'y a actuellement aucun référentiel à ajouter à cet affichage du contenu,","There are no Content Views in this Environment.":"Cet environnement ne contient pas d'affichage de contenu.","There are no Content Views that match the criteria.":"Aucun affichage de contenu ne correspond à ce critère.","There are no Docker Tags that match the criteria.":"Aucun Tag Docker ne correspond à ce critère.","There are no Errata associated with this Content Host to display.":"Il n'y a aucun errata associé à cet hôte de contenu à afficher.","There are no Errata in this organization.  Create one or more Products with Errata to view Errata on this page.":"Il n'existe aucun errata dans cette organisation. Créer un ou plusieurs Produit(s) avec Errata pour afficher les Errata sur cette page.","There are no Errata that match the criteria.":"Il n'existe aucun Errata correspondant à ce critère.","There are no Errata to display.":"Aucun errata à afficher.","There are no Host Collections available. You can create new Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"Aucune collection d'hôte disponible. Vous pouvez créer de nouvelles collections d'hôtes après avoir sélectionné « Collections d'hôtes » sous « Hôtes » dans le menu principal.","There are no Packages that match the criteria.":"Il n'existe aucun package correspondant à ce critère.","There are no Puppet Modules that match the criteria.":"Il n'existe aucune Module Puppet correspondant à ce critère.","There are no Yum Repositories that match the criteria.":"Il n'existe aucun Référentiel Yum correspondant à ce critère.","There is currently an Incremental Update task in progress.  This update must finish before applying existing updates.":"Une tâche de mise à jour croissante est actuellement en cours. Cette mise à jour doit se terminer avant d'appliquer les mises à jour existantes.","There were errors while removing the following Repositories:":"Des erreurs sont apparues lors de la suppression des référentiels suivants :","This Content Host is not attached to any subscriptions that provide content":"Cet affichage du contenu n'est attaché à aucun abonnement fournissant du contenu","This Content View does not have any versions, create your first Content View Version by using the \"Publish New Version\" button on the right.":"Cet affichage de contenu ne possède aucune version, créez votre première version d'affichage de contenu à l'aide du bouton Publier une nouvelle version situé à droite.","This Version cannot be deleted from some of the Lifecycle Environments due to one of the reasons below.":"Cette version ne peut pas être supprimée de l'un des environnements de cycle de vie à cause de l'une des raisons ci-dessous.","This Version is not associated with any Lifecycle Environments.":"Cette version n'est associée à aucun environnement de cycle de vie.","This activation key is not associated with any content hosts.":"Cette clé d'activation n'est associée à aucun hôte de contenu.","This activation key may be used during system registration. For example:":"Cette clé d'activation peut être utilisée au cours de l'enregistrement du système. Exemple :","This filter applies only to a subset of repositories in the content view.":"Le filtre s'applique uniquement à un sous-ensemble de référentiels dans l'affichage du contenu.","This filter applies to all repositories in the content view (current and future).":"Ce filtre s'applique à tous les référentiels dans l'affichage du contenu (actuel et futur).","This subscription is not associated with any activation keys.":"Cet abonnement n'est associé à aucune clé d'activation.","This subscription is not attached to any content hosts.":"Cet abonnement n'est attaché à aucun hôte de contenu.","This version will be removed from:":"Cette version sera supprimée de :","Time":"Heure","Title":"Titre","To register a content host to this server, follow these steps.":"Pour enregistrer un hôte de contenu sur ce serveur, suivez ces étapes.","Topic":"Sujet","Type":"Type","URL":"URL","URL of the registry you want to sync. Example: https://registry.hub.docker.com":"URL du registre que vous souhaitez synchroniser. Exemple : https://registry.hub.docker.com","URL to discover":"URL à découvrir","UUID":"UUID","Unattach Sync Plan":"Détacher le plan de synchronisation","Unattaching...":"Détachement...","Unlimited Content Hosts:":"Hôtes de contenu illimité :","Unregister Content Host":"Annuler l'enregistrement de l'hôte de contenu","Unregister Content Host \"{{contentHost.name}}\"?":"Annuler l'enregistrement de l'hôte de contenu \"{{contentHost.name}}\" ?","Unregister Content Hosts":"Annuler l'enregistrement des hôtes de contenu","Unregister {{ table.numSelected  }} Content Hosts?":"Annuler l'enregistrement des hôtes de contenu {{ table.numSelected  }} ?","Update":"Mettre à jour","Update All":"Tout mettre à jour","Update Repositories":"Mettre à jour les référentiels","Update Sync Plan":"Mettre à jour le plan de synchronisation","Updated":"Mis à jour","Updated errata filter -":"Mettre à jour le filtre d'errata -","Updated override to \"%x\".":"Remplacement de \"%x\" mis à jour.","Updating...":"Mise à jour...","Upload":"Télécharger","Upload Docker Image":"Téléchargez l'image Docker","Upload GPG Key":"Télécharger la clé GPG","Upload New Manifest":"Télécharger un nouveau manifeste","Upload Package":"Télécharger le package","Upload Puppet Module":"Télécharger le module puppet","Uploading...":"Téléchargement...","Upstream Repository Name":"Nom du référentiel en amont","Upstream Subscription":"Abonnement en amont","Upstream Subscription Management Application":"Application de Gestion d'abonnement en amont","Upstream Type":"Type en amont","Upstream User":"Utilisateur en amont","Use Latest (currently %s)":"Utiliser les (Currently %s) les plus récents","Use the cancel button on content view selection to revert your environment selection.":"Utiliser le bouton Annuler sur la sélection d'affichage du contenu pour rétablir la sélection de votre environnement.","User":"Utilisateur","Version":"Version","Version %s":"Version %s","Version 1":"Version 1","Version Deletion":"Suppression de version","Version Details":"Détails de la version","Version {{ version.version }}":"Version {{ version.version }}","Version:":"Version :","Versions":"Versions","Versions for Content View:":"Versions pour l'affichage du contenu :","Versions for Puppet Module:":"Versions pour le module puppet :","View Details":"Afficher les détails","Virtual":"Virtuel","Virtual Guest From":"Invité virtuel à partir de","Virtual Guests":"Invités virtuels","Virtual Host":"Hôte virtuel","Weekly on {{ product.sync_plan.sync_date | date:'EEEE' }} at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"Hebdomadaire le {{ product.sync_plan.sync_date | date:'EEEE' }} à {{ product.sync_plan.sync_date | date:'mediumTime' }} (heure du serveur)","When Auto Attach is disabled, registering systems will be attached to all associated subscriptions.":"Lorsque Auto Attach est désactivé, l'enregistrement de systèmes sera attaché à tous les abonnements associés.","Working":"Chargement","Yes":"Oui","Yes (Default)":"Oui (par défaut)","You cannot remove these repositories because they belong to a Red Hat product.":"Vous ne pouvez pas supprimer ces référentiels car ils appartiennent à un produit Red Hat.","You cannot remove these repositories because you do not have permission.":"Vous ne pouvez pas supprimer ces référentiels car nous n'avez pas la permission.","You cannot remove this product because it is a Red Hat product.":"Vous ne pouvez pas supprimer ce produit car c'est un produit Red Hat.","You cannot remove this product because it was published to a content view.":"Vous ne pouvez pas supprimer ce produit car il a été publié sur un affichage du contenu.","You cannot remove this product because you do not have permission.":"Vous ne pouvez pas supprimer ces produits car nous n'avez pas de permission.","You cannot remove this repository because it is a Red Hat repository.":"Vous ne pouvez pas supprimer ce référentiel car c'est un référentiel Red Hat.","You cannot remove this repository because it was published to a content view.":"Vous ne pouvez pas supprimer ce référentiel car il a été publié sur un affichage du contenu.","You cannot remove this repository because you do not have permission.":"Vous ne pouvez pas supprimer ce référentiel car nous n'avez pas de permission.","You currently do not have any Packages installed. You can use Package Actions above this message to install new Packages.":"Vous n'avez pour l'instant aucun package d'installé. Vous pouvez utiliser Actions de Package au-dessus de ce message pour installer de nouveaux packages.","You currently don't have any Activation Keys, you can add Activation Keys using the button on the right.":"Vous n'avez pour l'instant aucune Clé d'Activation. Vous pouvez ajouter des Clés d'Activation en utilisant le bouton situé sur la droite.","You currently don't have any Content Hosts in this Host Group, you can add Content Hosts after selecting the 'Add' tab.":"Vous n'avez pour l'instant aucun Hôte de Contenu dans ce groupe d'hôtes. Vous pouvez ajouter des Hôtes de Contenu après avoir sélectionné l'onglet « Ajouter ».","You currently don't have any Content Hosts, you can create new Content Hosts by selecting Contents Host from main menu and then clicking the button on the right.":"Vous n'avez pour l'instant aucun hôte de contenu. Vous pouvez créer des nouveaux hôtes de contenu en sélectionnant Hôte de Contenu à partir du menu principal et en cliquant ensuite sur le bouton situé à droite.","You currently don't have any Content Hosts, you can register one by clicking the button on the right and following the instructions.":"Vous ne possédez aucun hôte de contenu, vous pouvez en enregistrer un en cliquant sur le bouton situé à droite et en suivant les instructions.","You currently don't have any Content Views.  A Content View can be added by using the button on the right.":"Vous ne possédez aucun affichage de contenu. Un affichage de contenu peut être ajouté à l'aide du bouton situé à droite.","You currently don't have any Docker Tags.":"Vous ne possédez aucun Tag Docker.","You currently don't have any Filters included in this Content View, you can add a new Filter by using the button on the right.":"Vous ne possédez actuellement aucun filtre inclus dans cet affichage du contenu. Vous pouvez ajouter un nouveau filtre à l'aide du bouton situé sur la droite.","You currently don't have any Gpg keys, you can add Gpg keys using the button on the right.":"Vous n'avez pour l'instant aucune Clé Gpg. Vous pouvez ajouter des Clés Gpg  à l'aide du bouton situé sur la droite.","You currently don't have any Host Collections, you can add Host Collections using the button on the right.":"Vous ne possédez actuellement aucune Collection d'hôtes. Vous pouvez ajouter des Collections d'hôtes à l'aide du bouton situé à droite.","You currently don't have any Products associated with this Gpg Key.":"Vous ne possédez actuellement aucun Produit associé à cette clé Gpg.","You currently don't have any Products to subscribe to, you can add Products after selecting 'Products' under 'Content' in the main menu.":"Vous ne possédez actuellement aucun Produit auquel vous abonner, vous pouvez ajouter des Produits après avoir sélectionné « Produits » sous « Contenu » dans le menu principal.","You currently don't have any Products<span bst-feature-flag=\"custom_products\">, you can add Products using the button on the right</span>.":"Vous ne possédez actuellement aucun produit<span bst-feature-flag=\"custom_products\">. Vous pouvez ajouter des Produits en utilisant le bouton sur la droite</span>.","You currently don't have any Puppet Modules included in this Content View, you can add Puppet Modules using the button on the right.":"Vous ne possédez actuellement aucun Module Puppet inclus dans cet affichage du contenu. Vous pouvez ajouter des Modules Puppet à l'aide du bouton situé sur la droite.","You currently don't have any Repositories associated with this Gpg Key.":"Vous ne possédez actuellement aucun référentiel associé à cette clé Gpg.","You currently don't have any Repositories included in this Product, you can add Repositories using the button on the right.":"Vous ne possédez actuellement aucun filtre inclus dans cet affichage de contenu. Vous pouvez ajouter des référentiels à l'aide du bouton situé sur la droite.","You currently don't have any Subscriptions associated with this Content Host, you can add Subscriptions after selecting the 'Add' tab.":"Vous ne possédez actuellement aucun Abonnement associé à cet Hôte de Contenu. Vous pouvez ajouter des Abonnements après avoir sélectionné l'onglet « Ajouter ».","You currently don't have any Subscriptions, you can add Subscriptions by importing a manifest using the button on the right labeled 'Manage Manifest'.":"Vous ne possédez actuellement aucun Abonnement. Vous pouvez ajouter des Abonnements en important un manifeste à l'aide du bouton « Gérer un manifeste » situé à droite.","You currently don't have any Sync Plan, you can add Sync Plans using the button on the right.":"Vous ne possédez actuellement aucun Plan de Synchronisation. Vous pouvez ajouter des Plans de Synchronisation à l'aide du bouton situé à droite.","You currently don't have any Sync Plans.  A Sync Plan can be created by using the button on the right.":"Vous ne possédez aucun plan de synchronisation. Un plan de synchronisation peut être créé à l'aide du bouton situé à droite.","You do not have any Installed Products":"Vous n'avez aucun produit d'installé","You must select a content view in order to save your environment.":"Vous devez sélectionner un affichage de contenu pour enregistrer votre environnement.","You must select a new content view before your change of environment can be saved.":"Vous devez sélectionner un nouvel affichage du contenu pour que votre modification d'environnement puisse être enregistrée.","You must select a new content view before your change of environment can be saved. Use the cancel button on content view selection to revert your environment selection.":"Vous devez sélectionner un nouvel affichage du contenu pour que votre modification d'environnement puisse être enregistrée. Utiliser le bouton Annuler sur la sélection d'affichage du contenu pour rétablir la sélection de votre environnement.","You must select at least one Content Host in order to apply Errata.":"Vous devez sélectionner au moins un Hôte de Contenu afin d'appliquer les Errata.","You must select at least one repository.":"Vous devez sélectionner au moins un référentiel.","Your selected content hosts have changed, it is highly recommended to refresh the calculated list of errata using the 'Refresh Errata' button below.":"Vos hôtes de contenu sélectionnés ont changé, nous vous conseillons vivement d'actualiser la liste d'errata calculés à l'aide du bouton « Actualiser les errata » ci-dessous.","Yum Content":"Contenu Yum","Yum Metadata Checksum":"Somme de contrôle de métadonnées Yum","Yum Repo Discovery":"Découverte du référentiel yum","Yum Repositories":"Référentiels Yum","Yum Repositories <div>{{ library.counts.yum_repositories || 0 }}</div>":"Référentiels Yum <div>{{ library.counts.yum_repositories || 0 }}</div>","Yum metadata generation has been initiated in the background.  Click <a href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"La génération de métadonnées Yum a été initiée en arrière-plan. Cliquez  <a href=\"{{ taskUrl() }}\">ici</a> pour vérifier la progression.","add some repositories.":"ajouter des référentiels.","daily":"quotidien","here":"ici","hourly":"horaire","or authenticate with a username and password:":"ou effectuer une authentification avec un nom d'utilisateur et un mot de passe :","subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"":"subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"","task details":"détails des tâches","to manage them individually.":"pour les gérer individuellement.","true":"vrai","weekly":"hebdomadaire","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }} sélectionné","{{ addSubscriptionsTable.numSelected }} Selected":"{{ addSubscriptionsTable.numSelected }} sélectionné","{{ contentHost.registered_by }}":"{{ contentHost.registered_by }}","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }} sélectionné","{{ contentView.name }} cannot be deleted as one or more Content View Versions are still promoted to a Lifecycle Environment.\n      Each Content View Version must be removed from their Lifecycle Environments before the Content View can be deleted.":"{{ contentView.name }} ne peut pas être supprimé car une ou plusieurs versions d'affichage de contenu sont encore rattachées à un environnement de cycle de vie.\n     Chaque version d'affichage de contenu doit être supprimée de son environnement de cycle de vie avant que l'affichage de contenu ne puisse être supprimé.","{{ contentView.puppet_modules.length || 0 }} Puppet Modules":"{{ contentView.puppet_modules.length || 0 }} modules puppet","{{ contentView.repositories.length || 0 }} Repositories":"{{ contentView.repositories.length || 0 }} référentiels","{{ contentViewVersion.puppet_module_count || 0 }} Puppet Modules":"{{ contentViewVersion.puppet_module_count || 0 }} Modules Puppet","{{ contentViewVersion.repositories.length || 0 }} Repositories":"{{ contentViewVersion.repositories.length || 0 }} référentiels","{{ discoveryTable.numSelected }} Selected":"{{ discoveryTable.numSelected }} sélectionné","{{ environment.name }} Environment":"{{ environment.name }} Environment","{{ errata.hosts_applicable_count || 0 }} Applicable,":"{{ errata.hosts_applicable_count || 0 }} Applicable,","{{ errata.hosts_available_count || 0 }} Installable":"{{ errata.hosts_available_count || 0 }} pouvant être installés","{{ hostCollectionsTable.numSelected }} Selected":"{{ hostCollectionsTable.numSelected }} sélectionné","{{ package.removeTask.affected_units }} package removed":"{{ package.removeTask.affected_units }} packages supprimés","{{ package.removeTask.affected_units }} packages removed":"{{ package.removeTask.affected_units }} packages supprimés","{{ product.last_sync_words }} Ago ({{ product.sync_status.finish_time | date:'medium' }} Local Time)":"Il y a {{ product.last_sync_words }} ({{ product.sync_status.finish_time | date:'medium' }} heure locale)","{{ product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ product.sync_plan.next_sync | date:'medium' }} (Heure du serveur)","{{ productsTable.numSelected }} Selected":"{{ productsTable.numSelected }} sélectionné","{{ repository.content_counts.docker_image || 0 }} Docker Images":"{{ repository.content_counts.docker_image || 0 }} images Docker","{{ repository.content_counts.docker_image }} Docker Images":"{{ repository.content_counts.docker_image }} images Docker","{{ repository.content_counts.docker_tag || 0 }} Docker Tags":"{{ repository.content_counts.docker_tag || 0 }} balises Docker","{{ repository.content_counts.docker_tag }} Docker Tags":"{{ repository.content_counts.docker_tag }} balises Docker","{{ repository.content_counts.erratum || 0 }} Errata":"{{ repository.content_counts.erratum || 0 }} Errata","{{ repository.content_counts.erratum }} Errata":"{{ repository.content_counts.erratum }} Errata","{{ repository.content_counts.puppet_module || 0 }} Puppet Modules":"{{ repository.content_counts.puppet_module || 0 }} modules puppet","{{ repository.content_counts.rpm || 0 }} Packages":"{{ repository.content_counts.rpm || 0 }} Packages","{{ repository.content_counts.rpm }} Packages":"{{ repository.content_counts.rpm }} Packages","{{ repository.last_sync_words }} Ago ({{ repository.last_sync.ended_at | date:'medium' }} Local Time)":"il y a {{ repository.last_sync_words }} ({{ repository.last_sync.ended_at | date:'medium' }} Heure locale)","{{ repository.last_sync_words }} ago":"Il y a {{ repository.last_sync_words }}","{{ repository.product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ repository.product.sync_plan.next_sync | date:'medium' }} (heure du serveur)","{{ subscription.host.name }}":"{{ subscription.host.name }}","{{ subscriptionsTable.numSelected }} Selected":"{{ subscriptionsTable.numSelected }} sélectionné","{{ totalActivationKeyCount() }} Activation Keys will be moved to {{ deleteOptions.activationKeys.contentView.name }} in {{ deleteOptions.activationKeys.environment.name }}":"{{ totalActivationKeyCount() }} clés d'activation seront déplacées vers {{ deleteOptions.activationKeys.contentView.name }} dans {{ deleteOptions.activationKeys.environment.name }}","{{ totalHostCount() }} Content Hosts will be moved to {{ deleteOptions.contentHosts.contentView.name }} in {{ deleteOptions.contentHosts.environment.name }}":"{{ totalHostCount() }} Les hôtes de contenu seront déplacés vers {{ deleteOptions.contentHosts.contentView.name }} dans {{ deleteOptions.contentHosts.environment.name }}","{{ type.display }}":"{{ type.display }}","{{ version.docker_image_count }} Docker Images":"{{ version.docker_image_count }} Images Docker","{{ version.docker_tag_count }} Docker Tags":"{{ version.docker_tag_count }} Balises Docker","{{ version.errata_counts.total }} Errata":"{{ version.errata_counts.total }} Errata","{{ version.package_count }} Packages":"{{ version.package_count }} Packages","{{ version.puppet_module_count }} Puppet Modules":"{{ version.puppet_module_count }} Modules Puppet"});
    gettextCatalog.setStrings('it', {"%(consumed)s out of %(quantity)s":"%(consumed)s di %(quantity)s","(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)":"(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)","(Not all Activation Keys editable )":"(Non tutte le chiavi di attivazione sono modificabili )","(Not all Content Hosts editable )":"(Non tutti gli host di contenuto sono modificabili )","* These marked Content View Versions are from Composite Content Views.  Their components needing updating are listed underneath.":"* Queste versioni contrassegnate di Visualizzazione del contenuto fanno parte di Visualizzazioni composite. I componenti relativi da aggiornare sono di seguito elencati. ","+ New Sync Plan":"+ Nuova programmazione di sinc","- Date and Type":"- Data e Tipo","-- select an interval --":"-- selezionare un intervallo --","1 Content Host":["1 host di contenuto","{{ contentHost.virtual_guests.length }} host di contenuto"],"1 repository sync has errors.":["sono presenti alcuni errori in 1 sinc del repository.","sono presenti alcuni errori in {{ product.sync_summary.error}} sinc del repository."],"1 repository sync in progress.":["1 sincronizzazione del repository in corso.","{{ product.sync_summary.pending}} sinc dei repository in corso."],"1 successfully synced repository.":["1 repository sincronizzato con successo.","{{ product.sync_summary.success}} repository sincronizzati con successo."],"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Version {{ version.version }}</a>":"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Versione {{ version.version }}</a>","<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Create":"<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Crea","<span translate=\"\">\n        If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">here</a>\n      <span translate=\"\">to manage them individually.</span>":"<span translate=\"\">\n          Se preferisci spostare parte di queste Chiavi di attivazione su visualizzazioni diverse del contenuto, o Ambienti Ciclo di vita, fare clic </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">qui</a>\n      <span translate=\"\">per una loro gestione individuale.\n</span>","A composite view contains other content views.":"Una visualizzazione composita presenta altre visualizzazioni del contenuto.","A new version of {{ contentView.name }} will be created and promoted to the Library environment.\n    It can be promoted to other environments from the Versions tab of this Content View.":"Una nuova versione di {{ contentView.name }} verrà creata e avanzata all'ambiente Libreria.\n Questa versione può essere avanzata su altri ambienti dalla scheda Versioni di questa Visualizzazione dei contenuti.","Account":"Account","Account Number":"Numero account","Action":"Azione","Action Type":"Tipo di Azione","Actions":"Azioni","Actions for:":"Azioni per:","Activation Key":["Chiave di attivazione","Chiavi di attivazione"],"Activation Key Content":"Contenuto chiavi di attivazione","Activation Key Type:":"Tipo di chiave di attivazione:","Activation Key removed.":"Chiave di attivazione rimossa.","Activation Key updated":"Chiave di attivazione aggiornata","Activation Key:":"Chiave di attivazione:","Activation Key: {{ activationKey.name }}":"Chiave di attivazione: {{ activationKey.name }}","Activation Keys":"Chiavi di attivazione","Activation Keys for Subscription:":"Chiavi di attivazione per la sottoscrizione:","Activation Keys using Version {{ version.version }}":"Chiavi di attivazione usando la versione {{ version.version }}","Add":"Aggiungi","Add Content Hosts to:":"Aggiungere host di contenuto a:","Add Content Views":"Aggiungere le visualizzazioni del contenuto","Add Errata":"Aggiungi errata","Add New Environment":"Aggiungere un nuovo ambiente","Add New Filter":"Aggiungi nuovo filtro","Add New Module":"Aggiungi nuovo modulo","Add New Repository":"Aggiungi nuovo repositorio","Add Package Group":"Aggiungi gruppo di pacchetti","Add Repositories":"Aggiungi repositori","Add Selected":"Aggiungi selezione","Add Subscriptions for Activation Key:":"Aggiungi sottoscrizioni per la chiave di attivazione:","Add To":"Aggiungi a","Added %x host collections to activation key \"%y\".":"Aggiunti gli host %x alla chiave di attivazione \"%y\".","Added %x host collections to content host \"%y\".":"Aggiunti %x insiemi di host all'host di contenuto \"%y\".","Added %x products to sync plan \"%y\".":"Aggiunti %x prodotti per la programmazione della sincronizzazione \"%y\".","Adding Lifecycle Environment to the end of \"{{ priorEnvironment.name }}\"":"Aggiunta ambiente ciclo di vita alla fine di \"{{ priorEnvironment.name }}\"","Advanced Information":"Informazioni avanzate","Advisory":"Advisory","Affected Activation Keys":"Chiavi di attivazione interessate","Affected Content Hosts":"Host di contenuto interessati","Affected Packages":"Pacchetti interessati","Affected Systems":"Sistemi interessati","Affected repositories":"Repositori interessati","Affected repositories have been updated.":"I repositori interessati sono stati aggiornati.","All Content Views":"Visualizzazione di tutti i contenuti","All Products":"Tutti i prodotti","All Repositories":"Tutti i repositori","All Versions":"Tutte le versioni","Alter Sync Plans":"Modifica delle programmazioni di sincronizzazione","An error occurred applying Subscriptions:":"Si è verificato un errore durante l'applicazione delle Sottoscrizioni:","An error occurred removing the Activation Key:":"Si è verificato un errore durante la rimozione della chiave di attivazione.","An error occurred removing the Host Collection:":"Si è verificato un errore durante la rimozione dell'insieme di host.","An error occurred removing the Products:":"Si è verificato un errore durante la rimozione dei Prodotti:","An error occurred removing the content hosts.":"Si è verificato un errore durante la rimozione degli host di contenuto.","An error occurred removing the environment:":"Si è verificato un errore durante la rimozione dell'ambiente:","An error occurred removing the subscriptions.":"Si è verificato un errore durante la rimozione delle sottoscrizioni.","An error occurred saving the Activation Key:":"Si è verificato un errore durante il salvataggio della chiave di attivazione:","An error occurred saving the Content Host:":"Si è verificato un errore durante il salvataggio dell'host di contenuto:","An error occurred saving the Environment:":"Si è verificato un errore durante il salvataggio dell'Ambiente:","An error occurred saving the Filter:":"Si è verificato un errore durante il salvataggio del Filtro:","An error occurred saving the Host Collection:":"Si è verificato un errore durante il salvataggio dell'insieme di host.","An error occurred saving the Product:":"Si è verificato un errore durante il salvataggio del Prodotto:","An error occurred saving the Repository:":"Si è verificato un errore durante il salvataggio del Repositorio:","An error occurred saving the Sync Plan:":"Si è verificato un errore durante il salvataggio della Programmazione di sincronizzazione:","An error occurred saving the URL:":"Si è verificato un errore durante il salvataggio dell'URL:","An error occurred trying to auto-attach subscriptions.  Please check your log for further information.":"Si è verificato un errore durante l'assegnazione automatica delle sottoscrizioni. Controllare il log per maggiori informazioni.","An error occurred updating the Content View:":"Si è verificato un errore durante l'aggiornamento della Visualizzazione del contenuto:","Applicable":"Applicabile","Applicable Content Hosts for Errata":"Host di contenuto applicabili per l'Errata","Applicable Errata":"Errata applicabile","Applicable hosts for {{ erratum.errata_id }}":"Host applicabili per {{ erratum.errata_id }}","Apply":"Applica","Apply Errata":"Applica Errata","Apply Errata to Content Host \"{{contentHost.name}}\"?":"Applicare errata all'host di contenuto \"{{contentHost.name}}\"?","Apply Errata to Content Hosts":"Applica errata agli host di contenuto","Apply Errata to Content Hosts immediately after publishing.":"Applica errata agli host di contenuto immediatamente dopo la pubblicazione.","Apply Selected":"Applica quelli selezionati","Apply to Content Hosts":"Applica agli host di contenuto","Apply to Hosts":"Applica agli host","Apply {{ errata.errata_id }}":"Applica {{ errata.errata_id }}","Apply {{ errata.errata_id }} to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"Applica {{ errata.errata_id }} ai {{ numberOfContentHostsToUpdate }} Host di contenuto?","Apply {{ errataIds.length }} Errata to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"Applica {{ errataIds.length }} Errata agli host di contenuto {{ numberOfContentHostsToUpdate }} ?","Applying":"Applicazione in corso","Applying Bulk Errata Update":"Applicazione in gruppo aggiornamento errata","Arch":"Arch","Architecture(s)":"Architetture","Archived Copy":"Copia archiviata","Are you sure you want to add the {{ table.numSelected }} content host(s) selected to the host collection(s) chosen?":"Sei sicuro di voler aggiungere l'host di contenuto {{ table.numSelected }} selezionato all'insieme di host desiderati?","Are you sure you want to apply Errata to content host \"{{ contentHost.name }}\"?":"Sei sicuro di voler applicare l'Errata all'host di contenuto \"{{ contentHost.name }}\"?","Are you sure you want to apply the {{ detailsTable.numSelected }} selected errata to the content hosts chosen?":"Sei sicuro di voler applicare l'errata  {{ detailsTable.numSelected }} selezionato agli host di contenuto desiderati?","Are you sure you want to assign the {{ table.numSelected }} content host(s) selected to {{ selected.contentView.name }} in {{ selected.environment.name }}?":"Sei sicuro di voler assegnare gli host di contenuto {{ table.numSelected }} selezionati su {{ selected.contentView.name }} in {{ selected.environment.name }}?","Are you sure you want to auto-attach available subscriptions to all {{ table.resource.total }} content hosts?":"Sei sicuro di voler assegnare automaticamente le sottoscrizioni disponibili a tutti gli host di contenuto {{ table.resource.total }}?","Are you sure you want to delete Content View \"{{ contentView.name }}\"?":"Sei sicuro di voler rimuovere la Visualizzazione del contenuto \"{{ contentView.name }}\"?","Are you sure you want to install {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"Sei sicuro di voler installare {{ content.content }} sui sistemi {{ getSelectedSystemIds().length }} selezionati?","Are you sure you want to remove Activation Key \"{{ activationKey.name }}\"?":"Sei sicuro di voler rimuovere la chiave di attivazione \"{{ activationKey.name }}\"?","Are you sure you want to remove GPG key {{ gpgKey.name }}?":"Sei sicuro di voler rimuovere la chiave GPG {{ gpgKey.name }}?","Are you sure you want to remove Host Collection \"{{ hostCollection.name }}\"?":"Sei sicuro di voler rimuovere l'insieme \"{{ hostCollection.name }}\"?","Are you sure you want to remove Sync Plan \"{{ syncPlan.name }}\"?":"Sei sicuro di voler rimuovere la programmazione di sincronizzazione \"{{ syncPlan.name }}\"?","Are you sure you want to remove product \"{{ product.name }}\"?":"Sei sicuro di voler rimuovere il prodotto \"{{ product.name }}\"?","Are you sure you want to remove repository \"{{ repository.name }}\"?":"Sei sicuro di voler rimuovere il repository \"{{ repository.name }}\"?","Are you sure you want to remove the {{ productTable.getSelected().length }} products(s) selected?":"Sei sicuro di voler rimuovere i {{ productTable.getSelected().length }} prodotti selezionati?","Are you sure you want to remove the {{ table.numSelected }} content host(s) selected from the host collection(s) chosen?":"Sei sicuro di voler rimuovere gli host di contenuto {{ table.numSelected }} dall'insieme di host desiderato?","Are you sure you want to remove {{ content.content }} from the {{ getSelectedSystemIds().length }} system(s) selected?":"Sei sicuro di voler rimuovere {{ content.content }} dai sistemi {{ getSelectedSystemIds().length }} selezionati?","Are you sure you want to unregister content host \"{{ contentHost.name }}\"?":"Sei sicuro di voler rimuovere la registrazione dell'host di contenuto \"{{ contentHost.name }}\"?","Are you sure you want to unregister the {{ table.numSelected }} content host(s) selected?":"Sei sicuro di voler rimuovere la registrazione di {{ table.numSelected }} host di contenuto selezionati?","Are you sure you want to update {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"Sei sicuro di voler aggiornare {{ content.content }} sui sistemi {{ getSelectedSystemIds().length }} selezionati?","As part of this deletion, 1 Content View Version will be deleted.":["Come parte di questo processo, 1 Versione della visualizzazione del contenuto verrà cancellata.","Come parte di questo processo, {{ versions.length }} Versioni della visualizzazione del contenuto verranno cancellate."],"Assign":"Assegna","Assign Environment and Content View":"Assegna l'Ambiente e la Visualizzazione del contenuto","Associations":"Associazioni","At least one Errata needs to be selected to Apply.":"Selezionare almeno un errata da applicare.","At least one content host needs to be selected in order to perform a bulk action.":"Selezionare almeno un host di contenuto per poter eseguire operazioni in gruppo.","At least one product needs to be selected in order to perform a bulk action.":"Selezionare almeno un prodotto per poter eseguire operazioni in gruppo.","Attached":"Connesso","Attached to Content Hosts":"Assegnato agli host di contenuto","Author":"Autore","Auto-Attach":"Auto-Attach","Auto-attach available subscriptions to all content hosts":"Assegnazione automatica (auto-attach) delle sottoscrizioni disponibili a tutti gli host del contentuto","Automatic":"Automatico","Available Content Views for Composite Content View:":"Visualizzazioni del contenuto disponibili per una visualizzazione multipla:","Available Puppet Modules for Content View:":"Moduli Puppet disponibili per una Visualizzazione del contenuto:","Back":"Indietro","Back To Errata List":"Torna all'Elenco di errata","Back To Repository Selection":"Torna alla Selezione dei repositori","Back to Module List":"Torna all'Elenco dei moduli","Back to Repository Details":"Torna alle Informazioni repositorio","Back to Repository List":"Torna all'Elenco dei Repositori","Back to Version List":"Torna all'Elenco delle versioni","Basic Information":"Informazioni di base","Below are the product's repository content sets currently available for this subscription. For Red Hat subscriptions, additional content can be made available through the":"Di seguito vengono riportati gli insiemi del contenuto dei repositori del prodotto, attualmente disponibili per questa sottoscrizione. Per le sottoscrizioni di Red Hat, è possibile ottenere altri contenuti attraverso","Below are the repository content sets currently available for this activation key through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"Di seguito vengono riportati gli insiemi del contenuto dei repositori disponibili per questa chiave di attivazione per mezzo delle proprie sottoscrizioni. Per le sottoscrizioni di Red Hat, è possibile ottenere altri contenuti attraverso","Below are the repository content sets currently available for this content host through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"Di seguito vengono riportati gli insiemi del contenuto dei repositori attualmente disponibili per questo host per mezzo delle proprie sottoscrizioni. Per le sottoscrizioni di Red Hat, è possibile ottenere altri contenuti attraverso","Bug Fix":"Bug Fix","Bug Fix Advisory":"Bug Fix Advisory","Bugfix":"Bugfix","Bulk Actions":"Operazioni in gruppo","Bulk Alter Products Sync Plan":"Modifica in gruppo la Programmazione di sincronizzazione dei prodotti","Bulk Sync Products":"Sincronizza in gruppo i Prodotti","CVEs":"CVE","Cancel":"Cancella","Capacity":"Capacità","Change assigned Lifecycle Environment or Content View":"Modifica la Visualizzazione del contenuto o l'Ambiente Ciclo di vita assegnati","Changing default settings for content hosts that register with this activation key requires subscription-manager version 1.10 or newer to be installed on that host.":"Per la modifica delle impostazioni predefinite degli host che eseguono la registrazione con questa chiave di attivazione, è necessario avere la versione 1.10 o più recente di subscription-manager sull'host in questione.","Changing default settings requires subscription-manager version 1.10 or newer to be installed on this host.":"Per la modifica delle impostazioni predefinite è necessario aver installato la versione 1.10, o versione più recente, di subscription-manager su questo host.","Checkin":"Checkin","Checksum":"Checksum","Choose a lifecycle environment from the available promotion paths.":"Selezionere un ambiente Ciclo di vita dai percorsi di avanzamento disponibili.","Click here to check the status of the task.":"Fai clic qui per controllare lo stato dell'attività.","Click here to refresh.":"Fai clic qui per l'aggiornamento.","Click here to see the task for the successful removals.":"Clicca qui per vedere l'attività relativa alle rimozioni con successo.","Click here to select Errata for an Incremental Update.":"Fai clic qui per selezionare l'Errata per un Aggiornamento incrementale","Close":"Chiudi","Collection Actions":"Insieme azioni","Completely remove version?":"Rimuovere completamente la versione?","Components":"Componenti","Components:":"Componenti:","Composite Content View {{ contentView.name }}":"Visualizzazione contenuto composita {{ contentView.name }}","Composite View":"Visualizzazione composita","Composite View?":"Visualizzazione composita?","Composite?":"Composita?","Confirm":"Conferma","Confirm Remove":"Conferma rimozione","Confirm Version Removal: Version {{ version.version }}":"Conferma rimozione versione: Versione {{ version.version }}","Confirmation":"Conferma","Consumed":"Consumati","Content":"Contenuto","Content Counts":"Conteggi contenuto","Content Download URL":"URL per scaricare il contenuto","Content Host":"Host contenuto","Content Host %s has been deleted.":"Host di contenuto %s rimosso.","Content Host Bulk Actions":"Azioni in gruppo per l'host di contenuto","Content Host Bulk Content":"Contenuto in gruppo per l'host di contenuto","Content Host Bulk Errata":"Errata in gruppo per l'host di contenuto","Content Host Bulk Host Collections":"Insiemi host in gruppo per l'host di contenuto","Content Host Bulk Packages":"Pacchetti in gruppo per l'host di contenuto","Content Host Bulk Subscriptions":"Sottoscrizioni in gruppo per l'host di contenuto","Content Host Content":"Contenuto per l'host","Content Host Counts":"Conteggi host di contenuto","Content Host Errata Management":"Gestione errata per l'host di contenuto","Content Host Limit":"Limite host di contenuto","Content Host Package Management":"Gestione pacchetti per l'host di contenuto","Content Host Properties":"Proprietà per l'host di contenuto","Content Host Registration":"Registrazione per l'host di contenuto","Content Host Status":"Stato dell'host di contenuto","Content Host Subscriptions":"Sottoscrizioni host di contenuto","Content Host {{ contentHost.name }}":"Host di contenuto {{ contentHost.name }}","Content Host:":"Host di contenuto:","Content Hosts":"Host di contenuto","Content Hosts for Activation Key:":"Host di contenuto per la chiave di attivazione:","Content Hosts for Subscription:":"Host di contenuto per la sottoscrizione:","Content Hosts for:":"Host di contenuto per:","Content Hosts using Version {{ version.version }}":"Host di contenuto usando la Versione {{ version.version }}","Content Type":"Tipo di Contenuto","Content View":"Visualizzazione contenuto","Content View Name":"Nome visualizzazione del contenuto","Content View Version":"Versione visualizzazione del contenuto","Content View updated.":"Visualizzazione contenuto aggiornata.","Content View:":"Visualizzazione contenuto:","Content Views":"Visualizzazioni contenuto","Content Views <div>{{ library.counts.content_views || 0 }}</div>":"Visualizzazioni contenuto <div>{{ library.counts.content_views || 0 }}</div>","Content Views for Composite Content View:":"Visualizzazioni del contenuto per una visualizzazione composita:","Content successfully uploaded":"Contenuto caricato con successo","Contents":"Contenuti","Contract":"Contratto","Contract Number":"Numero contratto","Copy":"Copia","Copy Activation Key":"Copia chiave di attivazione","Copy Collection":"Copia insieme","Copy View":"Copia visualizzazione","Cores per Socket":"Core per socket","Cores: %s":"Core: %s","Create":"Crea","Create New View":"Crea nuova visualizzazione","Create Repositories":"Crea repositorio","Create Repositories Within:":"Crea repository all'interno:","Create Repository":"Crea repositorio","Create Selected":"Crea i selezionati","Create Sync Plan":"Crea programmazione per la sincronizzazione","Create new view":"Crea nuova visualizzazione","Create view from existing views":"Crea visualizzazione dalle visualizzazionii esistenti","Critical":"Critico","Current Environment (%e/%cv)":"Ambiente corrente (%e/%cv)","Current Subscriptions for Activation Key:":"Sottoscrizioni correnti per la chiave di attivazione:","Currently Selected Puppet Modules":"Moduli Puppet correnti selezionati","Daily at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"Giornalmente il {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)","Date":"Data","Default":"Predefinita","Delete":"Cancella","Delete Manifest":"Cancella il manifesto","Delete Version {{ version.version }}":"Cancella versione {{ version.version }}","Delete {{ contentView.name }}":"Cancella {{ contentView.name }}","Deleted from %s":"Cancellato da %s","Deleting Manifest...":"Rimozione manifesto...","Deleting from %count environments.":"Rimozione da %count ambienti.","Deleting from 1 environment.":"Rimozione da 1 ambiente.","Deletion from %s":"Rimozione da %s","Description":"Descrizione","Detail":"Dettagli","Details":"Dettagli","Details for Activation Key:":"Dettagli per la chiave di attivazione:","Details for Content View:":"Dettagli per la visualizzazione del contenuto:","Details for Product:":"Dettagli per il prodotto:","Discover":"Trova","Discovered URLs":"URL trovati","Docker Content":"Contenuto Docker","Docker Images":"Immagini Docker","Docker Images in {{ repository.name }}":"Immagini Docker in {{ repository.name }}","Docker Repositories":"Repository Docker","Docker Repositories <div>{{ library.counts.docker_repositories || 0 }}</div>":"Repository Docker <div>{{ library.counts.docker_repositories || 0 }}</div>","Docker Tag:":"Tag Docker:","Docker Tags":"Tag Docker:","Docker metadata generation has been initiated in the background.  Click\n        <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"L'operazione di creazione dei metadati Docker è stata iniziata nel background. Fare clic       \n<a ng-href=\"{{ taskUrl() }}\">Qui</a> per controllare \nlo stato.","Edit":"Modifica","Enabled?":"Abilitato?","End Date":"Data di fine","Ends":"Termina","Enhancement":"Miglioramenti","Enter Package Group Name(s)...":"Inserire i nomi dei gruppi di pacchetti...","Enter Package Name(s)...":"Inserire i nomi dei pacchetti...","Environment":"Ambiente","Environment saved":"Ambiente salvato","Environment: {{ environment.name }}":"Ambiente: {{ environment.name }}","Environments":"Ambienti","Environments List":"Elenco ambienti","Environments:":"Ambienti:","Equal To":"Uguale a ","Errata":"Errata","Errata <div>{{ library.counts.errata.total || 0 }}</div>":"Errata <div>{{ library.counts.errata.total || 0 }}</div>","Errata Details":"Informazioni errata","Errata Filter:":"Filtro errata:","Errata ID":"ID errata","Errata Installation":"Installazione errata","Errata List":"Elenco errata","Errata Task List":"Elenco attività errata","Errata Tasks":"Attività errata","Errata Type":"Tipo di errata","Errata for:":"Errata per:","Errata successfully added.":"Errata aggiunto con successo.","Errata successfully removed.":"Errata rimosso con successo.","Errata:":"Errata:","Erratum - by Date and Type":"Erratum - per Data e Tipo","Erratum - by ID":"Erratum - per ID","Erratum Date Range":"Intervallo date per l'Erratum","Error deleting manifest.":"Errore durante la rimozione del manifesto.","Error during upload:":"Errore durante il caricamento:","Error importing manifest.":"Errore durante l'importazione del manifesto.","Error refreshing manifest.":"Errore durante l'aggiornamento del manifesto.","Event":"Evento","Events":"Eventi","Exclude":"Escludi","Existing Product":"Prodotto esistente","Expires":"Scade","Failed":"Fallito","File too large. Please use the CLI instead.":"Il file è troppo grande. Usare CLI.","Filter":"Filtro","Filter Saved":"Filtro salvato","Filter information for:":"Informazioni filtro per:","Filter...":"Filtro...","Filter:":"Filtro:","Filters":"Filtri","Filters List":"Elenco filtri","Filters for Content View:":"Filtri per la Visualizzazione del contenuto:","Filters successfully removed.":"Filtri rimossi con successo.","Finished At":"Terminato ","For older operating systems such as Red Hat Enterprise Linux 5 or CentOS 5 it is recommended to use sha1.":"Per sistemi operativi come Red Hat Enterprise Linux 5 o CentOS 5 è consigliato usare sha1.","GPG Key":"Chiave GPG","GPG Key URL":"URL chiave GPG","GPG Key successfully uploaded":"Chiave GPG caricata con successo","GPG Key {{ gpgKey.name }}":"Chiave GPG {{ gpgKey.name }}","GPG Keys":"Chiavi GPG","General Information":"Informazioni generali","Generated On":"Generato su","Gpg Key updated":"Chiave GPG caricata","Gpg Keys":"Chiavi GPG","Greater Than":"Maggiore di","Group Install":"Installazione gruppo","Group Remove":"Rimozione gruppo","Guests of":"Guest di","Hide affected Activation Keys":"Nascondi le chiavi di attivazione interessate","Hide affected Content Hosts":"Nascondi gli host di contenuto interessati","History":"Cronologia","History for Content View:":"Cronologia per la Visualizzazione del contenuto","Host Collection Management":"Gestione insieme di host","Host Collection Membership":"Appartenenza insieme di host","Host Collection Membership Management":"Gestione appartenenza insieme di host","Host Collection removed.":"Insieme di host rimosso.","Host Collection updated":"Insieme di host aggiornato","Host Collection:":"Insieme di host:","Host Collection: {{ hostCollection.name }}":"Insieme di host: {{ hostCollection.name }}","Host Collections":"Insiemi di host","Host Collections for:":"Insiemi di host per:","Host Count":"Conteggio host","Host Group":"Gruppo di host","Hostname":"Hostname","Hourly at {{ product.sync_plan.sync_date | date:'m' }} minutes and {{ product.sync_plan.sync_date | date:'s' }} seconds":"Ogni ora {{ product.sync_plan.sync_date | date:'m' }} minuti e {{ product.sync_plan.sync_date | date:'s' }} secondi","IPv4 Address":"Indirizzo IPv4","IPv6 Address":"Indirizzo IPv6","Id":"Id","If you would prefer to move portions of these Activation Keys to different Content Views or Lifecycle Environments click <a href=\"{{ activationKeyLink() }}\">here</a> to manage them individually.":"Se preferisci spostare parte di queste Chiavi di attivazione su visualizzazioni diverse del contenuto, o Ambienti Ciclo di vita, fare clic <a href=\"{{ activationKeyLink() }}\">qui</a> per una loro gestione individuale.","If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click":"Se preferisci spostare parte di queste Chiavi di attivazione su visualizzazioni diverse del contenuto o Ambienti Ciclo di vita.","If you would prefer to move portions of these Content Hosts to different content views or environments click <a href=\"{{ contentHostsLink() }}\">here</a>  to manage these Content Hosts in bulk.":"Se preferisci spostare parte di questi host di contenuto su visualizzazioni diverse o ambienti fare clic <a href=\"{{ contentHostsLink() }}\">qui</a> per una loro gestione di gruppo.","Image":"Immagine","Image ID":"ID immagine","Import History":"Importa cronologia","Important":"Importante","In order to add a content view to a composite view you must first publish an initial version of the content view.":"Per aggiungere una visualizzazione del contenuto ad una visualizzazione composita, è necessario prima pubblicare una versione iniziale della visualizzazione stessa.","Include":"Includi","Included in Activation Keys":"Incluso nelle chiavi di attivazione","Incremental Update":"Aggiornamento incrementale","Install":"Installa","Install Selected":"Installa i selezionati","Install katello-agent for remote actions and displaying errata information:":"Installare katello-agent per le azioni remote e la visualizzazione delle informazioni sull'errata:","Install the pre-built bootstrap RPM:":"Installare l'RPM bootstrap esistente:","Installable":"Installabile","Installable Errata":"Errata installabile","Installed":"Installato","Installed Package":"Pacchetto installato","Installed Packages":"Pacchetti installati","Installed Products":"Prodotti installati","Instance-based":"Basato sull'istanza","Interfaces":"Interfacce","Interval":"Intervallo","Issued":"Emesso","Katello Agent":"Katello Agent","Label":"Etichetta","Last Checkin":"Ultimo checkin","Last Published":"Ultimo pubblicato","Last Puppet Report":"Ultimo riporto del Puppet","Last Sync":"Ultima sincronizzazione","Last Updated":"Ultimo aggiornato","Last Updated On":"Ultimo aggiornato il","Last synced {{ product.last_sync_words }} ago.":"Ultimo aggiornato {{ product.last_sync_words }} fa.","Latest (Currently %s)":"Ultimissimo (Corrente %s)","Less Than":"Inferiore a","Library":"Libreria ","Library Synced Content":"Contenuto libreria sincronizzato","Lifecycle Environment \"{{ priorEnvironment.name }}\" exists within the middle of a path. New Lifecycle Environments can only be added to the end of a path.":"Ambiente ciclo di vita \"{{ priorEnvironment.name }}\" esistente in posizione centrale nel percorso. È possibile aggiungere nuovi ambienti solo alla fine di un percorso.","Lifecycle Environment Paths":"Percorsi ambiente Ciclo di vita","Lifecycle Environments":"Ambienti Ciclo di vita","Limit":"Limite","Limit:":"Limite:","Limits":"Limiti","List/Remove":"Elenca/Rimuovi","Loading Environment Paths...":"Caricamento percorsi ambiente in corso...","Loading...":"Caricamento in corso...","Manage Docker Images":"Gestisci immagini Docker","Manage Docker Images for Repository:":"Gestisci immagini Docker per il repositoy:","Manage Manifest":"Gestisci il manifesto","Manage Packages":"Gestisci i pacchetti","Manage Packages for Repository:":"Gestisci i pacchetti per il repositorio:","Manage Puppet Modules":"Gestisci moduli Puppet","Manage Puppet Modules for Repository:":"Gestisci moduli Puppet per il repository:","Manifest History":"Cronologia del manifesto","Manifest successfully deleted.":"Manifesto rimosso con successo.","Manifest successfully imported.":"Manifesto importato con successo.","Manifest successfully refreshed.":"Manifesto aggiornato con successo.","Message":"Messaggio","Model":"Modello","Moderate":"Moderato","Module %s removed from Content View.":"Modulo %s rimosso dalla Visualizzazione del contenuto.","Module List":"Elenco modulo","More Details":"Più informazioni","More...":"Più...","Multi-entitlement":"Entitlement-multipli","N/A":"N/A","Name":"Nome","Name of the upstream repository you want to sync. Example: 'busybox' or 'fedora/ssh'.":"Nome del repository upstream da sincronizzare. Esempio: 'busybox' or 'fedora/ssh'.","Networking":"Networking","Never synced":"Mai sincronizzato","New Activation Key":"Nuova chiave di attivazione","New Content View":"Nuova Visualizzazione del contenuto","New Environment":"Nuovo ambiente","New Environment Path":"Nuovo Percorso dell'ambiente","New Filter":"Nuovo filtro","New Filter for Content View:":"Nuovo filtro per la Visualizzazione dell'ambiente:","New GPG Key":"Nuova chiave GPG","New Gpg Key":"Nuova chiave GPG","New Host Collection":"Nuovo insieme di host","New Name:":"Nuovo nome:","New Product":"Nuovo prodotto","New Repository":"Nuovo repositorio","New Sync Plan":"Nuova programmazione di sinc","New sync plan successfully created.":"Nuovo programma di sincronizzazione creato con successo.","Next":"Successivo","Next Sync":"Sincronizzazione successiva","No":"No","No \"Edit\" permissions on some of the Activation Keys in that Lifecycle Environment.":"Nessun permesso di \"Modifica\" su alcune Chiavi di attivazione nell'ambiente Ciclo di vita.","No \"Edit\" permissions on some of the Content Hosts in that Lifecycle Environment.":"Nessun permesso di \"Modifica\" su alcuni degli host di contenuto nell'ambiente Ciclo di vita.","No \"Promote/Remove\" permissions on the Lifecycle Environment.":"Nessun permesso \"Avanza/Rimuovi\" sull'ambiente Ciclo di vita.","No (Default)":"No (Predefinito)","No Content Hosts are affected.":"Nessun host di contenuto è interessato.","No Content Hosts match this Erratum.":"Nessun host di contenuto corrisponde a questo erratum.","No Content Views available, please select another environment.":"Nessuna visualizzazione del contenuto disponibile, selezionare un altro ambiente.","No Host Collections to show, you can add Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"Nessun insieme di host da visualizzare, è possibile aggiungere un insieme dopo aver selezionato 'Insiemi di host' in 'Host' nel menu principale.","No Host Collections to show, you can add Host Collections after selecting the 'Add' tab.":"Nessun insieme di host da visualizzare, è possibile aggiungere un insieme dopo aver selezionato la scheda 'Aggiungi'.","No Products to show.":"Nessun prodotto da visualizzare.","No Repositories contain this Erratum.":"Nessun repository presenta questo erratum.","No activation keys are affected.":"Nessuna chiave di attivazione è interessata.","No alternate release version choices are available. The available releases are based upon what is available in \"{{ contentHost.content_view.name }}\", the selected <a href=\"/content_views\">content view</a> this content host is attached to for the given <a href=\"/lifecycle_environments\">lifecycle environment</a>, \"{{ contentHost.environment.name }}\".":"Non è disponibile alcuna selezione alternativa per la versione della release. Le release disponibili si basano su quelle presenti in \"{{ contentHost.content_view.name }}\", <a href=\"/content_views\">visualizzazione del contenuto</a> selezionata alla quale è assegnato questo host di contenuto per <a href=\"/lifecycle_environments\">l'ambiente ciclo di vita</a>, \"{{ contentHost.environment.name }}\".","No content views exist for {{selected.environment.name}}":"Non esiste alcuna visualizzazione del contenuto per {{selected.environment.name}}","No host collections":"Nessun insieme di host","No provisioning details are available for this content host.":"Nessuna informazione sul provisioning è disponibile per questo host di contenuto.","No repository content provided through subscriptions.":"Non è stato fornito alcun contenuto del repositorio attraverso le sottoscrizioni.","No repository content provided through this subscription.":"Non è stato fornito alcun contenuto del repositorio attraverso questa sottoscrizione.","No repository content.":"Nessun contenuto del repositorio.","No subscription manifest imported":"Nessun manifesto di sottoscrizione importato","None":"Nessuno","Not Synced":"Non sincronizzato","Not installed":"Non installati","Not yet published":"Non ancora pubblicato","Number of CPUs":"Numero di CPU","OS":"OS","Only show Errata that is Applicable to one or more Content Hosts":"Mostra solo gli errata applicabili ad uno o più host di contenuto","Only show Errata that is Installable on one or more Content Hosts":"Mostra solo gli errata installabili ad uno o più host di contenuto","Operating System":"Sistema Operativo","Options:":"Opzioni:","Organization":"Organizzazione","Original Sync Date":"Data originale per la sincronizzazione","Override to No":"Override su No","Override to Yes":"Override su Si","Package":"Pacchetto","Package Actions":"Azioni pacchetto","Package Filter:":"Filtro pacchetto:","Package Group":"Gruppo di pacchetti","Package Group Filter:":"Filtro gruppo pacchetti:","Package Group successfully added.":"Gruppo pacchetti aggiunti con successo.","Package Group successfully removed.":"Gruppo pacchetti rimossi con successo.","Package Groups":"Gruppi di pacchetti","Package Install":"Installazione pacchetto","Package Installation, Removal, and Update":"Installazione, rimozione e aggiornamento del pacchetto","Package Remove":"Rimozione pacchetto","Package Update":"Aggiornmento pacchetto","Package removal failed":"Rimozione pacchetto fallita","Package successfully added.":"Pacchetto aggiunto con successo.","Package successfully removed.":"Pacchetto rimosso con successo.","Package successfully updated.":"Pacchetto aggiornato con successo.","Package/Group Name":"Nome Gruppo/Pacchetto","Packages":"Pacchetti","Packages <div>{{ library.counts.packages || 0 }}</div>":"Pacchetti <div>{{ library.counts.packages || 0 }}</div>","Packages for:":"Pacchetti per:","Packages in {{ repository.name }}":"Pacchetti in {{ repository.name }}","Parameters":"Parametri","Paste GPG Key Contents":"Incolla i contenuti della chiave GPG","Paste contents...":"Incolla i contenuti...","Paused":"In pausa","Perform":"Esegui","Physical":"Fisico","Please select a Lifecycle Environment and Content View to move the affected Activation Keys to:":"Per poter spostare le Chiavi di attivazione interessate selezionare un Ambiente Ciclo di vita e una Visualizzazione del contenuto:","Please select a Lifecycle Environment and Content View to move these Content Hosts to:":"Per poter spostare gli host di contenuto selezionare un ambiente Ciclo di vita e una Visualizzazione del contenuto:","Please select a Lifecycle Environment.":"Selezionare un ambiente Ciclo di vita.","Please select an environment.":"Selezionare un ambiente.","Please select one from the list below and you will be redirected.":"Selezionarne uno dall'elenco di seguito riportato e verrai ridirezionato.","Press 'Sync Now' below to begin syncing the following product(s).":"Premere \"Sincronizza ora' qui di seguito per iniziare la sincronizzazione dei seguenti prodotti.","Previous Environment (%e/%cv)":"Ambiente precedente (%e/%cv)","Processing":"Processazione in corso","Product":"Prodotto","Product Bulk Actions":"Azioni in gruppo per il prodotto","Product Content":"Contenuto del prodotto","Product Count":"Conteggio del prodotto","Product Enhancement Advisory":"Advisory sul miglioramento del prodotto","Product ID":"ID del prodotto","Product Management":"Gestione del prodotto","Product Management for Sync Plan:":"Gestione del prodotto per la programmazione della sincronizzazione:","Product Name":"Nome prodotto","Product Saved":"Prodotto salvato","Product Sync":"Sincronizzazione del prodotto","Product delete operation has been initiated in the background.  Click <a href=\"{{ taskUrl(productDeletionTaskId) }}\">Here</a> to monitor the progress.":"L'operazione di rimozione del prodotto è stata iniziata nel background. Fare clic <a href=\"{{ taskUrl(productDeletionTaskId) }}\">Qui</a> per controllare lo stato.","Product information for:":"Informazioni sul prodotto per:","Product {{ product.name }}":"Prodotto {{ product.name }}","Products":"Prodotti","Products <div>{{ library.counts.products || 0 }}</div>":"Prodotti <div>{{ library.counts.products || 0 }}</div>","Promote":"Avanzamento","Promote Content View:":"Visualizzazione avanzamento del contenuto:","Promote Version":"Avanzamento versione","Promote Version {{ version.version }}":"Avanzamento versione {{ version.version }}","Promoted to %s":"Avanzato a %s","Promoting to %count environments.":"Avanzamento su %count ambienti","Promoting to 1 environment.":"Avanzamento a 1 ambiente","Promotion History":"Cronologia avanzamento","Provided Products":"Prodotti forniti","Provisioning Details":"Informazioni sul provisioning","Provisioning Host Details":"Informazioni host di provisioning","Publish Content View:":"Visualizzazione pubblicazione del contenuto:","Publish New Version":"Pubblica nuova versione","Publish via HTTP":"Pubblica tramite HTTP","Publish via HTTPS":"Pubblica tramite HTTPS","Published":"Pubblicato","Published At":"Pubblicato su","Published Repository Information":"Informazioni repository pubblicate","Published new version":"Nuova versione pubblicata","Publishing and promoting to %count environments.":"Pubblicazione e avanzamento su %count ambienti.","Publishing and promoting to 1 environment.":"Pubblicazione e avanzamento su 1 ambiente.","Puppet Environment":"Ambiente Puppet","Puppet Modules":"Moduli Puppet","Puppet Modules <div>{{ library.counts.puppet_modules || 0 }}</div>":"Moduli Puppet <div>{{ library.counts.puppet_modules || 0 }}</div>","Puppet Modules for Content View:":"Moduli Puppet per la visualizzazione del contenuto:","Puppet Modules in {{ repository.name }}":"Moduli Puppet in {{ repository.name }}","Puppet module added to Content View":"Moduli Puppet aggiunti alla visualizzazione del contenuto","Puppet module metadata generation has been initiated in the background.\n        Click <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"L'operazione di creazione dei metadati del modulo Puppet è stata iniziata nel background. \nFare clic <a ng-href=\"{{ taskUrl() }}\">Qui</a> per controllare \nlo stato.","Quantity":"Quantità","RAM (GB)":"RAM (GB)","RAM: %s GB":"RAM: %s GB","Range":"Gamma","Reboot Suggested":"Riavvio consigliato","Reboot Suggested?":"Riavvio consigliato?","Red Hat CDN URL":"Red Hat CDN URL","Red Hat Provider Details":"Dettagli sul provider di Red Hat","Red Hat Repositories page":"Pagina repositori di Red Hat","Red Hat Repositories page.":"Pagina repository di Red Hat.","Refresh Manifest":"Ricarica il manifesto","Refresh Table":"Aggiorna tabella","Refreshing Manifest...":"Aggiornamento manifesto in corso...","Refreshing manifest not suppported by this manifest.":"L'aggiornamento del manifesto non è supportato da questo manifesto.","Register Content Host":"Registra host di contenuto","Register a Content Host":"Registra un host di contenuto","Registered":"Registrati","Registered By":"Registrato da","Registry URL":"URL registro","Release":"Release","Release Version":"Versione release","Remove":"Rimuovi","Remove Activation Key \"{{ activationKey.name }}\"?":"Rimuovere la chiave di attivazione \"{{ activationKey.name }}\"?","Remove Content View":"Rimuovi visualizzazione del contenuto","Remove Content Views":"Rimuovi visualizzazioni del contenuto","Remove Docker Images":"Rimuovi immagini del Docker","Remove Errata":"Rimuovi errata","Remove Filters":"Rimuovi filtri","Remove From":"Rimuovi da","Remove GPG Key":"Rimuovi chiave GPG","Remove GPG Key {{ gpgKey.name }}":"Rimuovi chiave GPG {{ gpgKey.name }}","Remove Host Collection \"{{ hostCollection.name }}\"?":"Rimuovere l'insieme di host \"{{ hostCollection.name }}\"?","Remove Module":"Rimuovi modulo","Remove Package Group":"Rimuovi gruppo di pacchetti","Remove Packages":"Rimuovi pacchetti","Remove Product":"Rimuovi prodotto","Remove Product \"{{ product.name }}\"?":"Rimuovere il prodotto \"{{ product.name }}\"?","Remove Products":"Rimuovi prodotti","Remove Puppet Modules":"Rimuovi moduli puppet","Remove Repositories":"Rimuovi repository","Remove Repository":"Rimuovi repository","Remove Repository \"{{ repository.name }}\"?":"Rrimuovere il repository \"{{ repository.name }}\"?","Remove Selected":"Rimuovi selezionati","Remove Successful.":"Rimosso con successo.","Remove Sync Plan":"Rimuovi programmazione di sinc","Remove Sync Plan \"{{ syncPlan.name }}\"?":"Rimuovere la programmazione di sincronizzazione \"{{ syncPlan.name }}\"?","Remove Version":"Rimozione versione","Remove Version Confirmation":"Conferma rimozione della versione","Remove View":"Rimuovi visualizzazione","Remove {{ productTable.getSelected().length }} Products?":"Rimuovere i {{ productTable.getSelected().length }} prodotti?","Removed %x host collections from activation key \"%y\".":"Rimossi %x insiemi di host dalla chiave di attivazione \"%y\".","Removed %x host collections from content host \"%y\".":"Rimossi %x insiemi di host dall'host di contenuto \"%y\".","Removed %x products from sync plan \"%y\".":"Rimossi %x prodotti dalla programmazione per la sincronizzazione \"%y\".","Removing Content Hosts":"Rimozione host di contenuto in corso","Removing Products":"Rimozione prodotti in corso","Removing Repositories":"Rimozione repository in corso","Repo Discovery":"Scoperta repository","Repo Type":"Tipo di repository","Repositories":"Repository","Repositories containing Errata {{ errata.errata_id }}":"Repository contenente gli Errata {{ errata.errata_id }}","Repositories for Content View:":"Repository per la visualizzazione del contenuto:","Repositories for Errata:":"Repositori per l'errata:","Repositories for Filter:":"Repositori per il filtro:","Repositories for Product:":"Repositori per il prodotto:","Repository":"Repository","Repository \"%s\" successfully deleted":"Repository \"%s\" cancellato con successo","Repository %s successfully created.":"Repository %s creato con successo.","Repository Count":"Conteggio repository","Repository Discovery":"Scoperta repository","Repository Name":"Il nome del repository","Repository Saved.":"Repositori salvati.","Repository Selection":"Selezione repositorio","Repository URL updated":"URL repositorio aggiornato","Repository:":"Repositorio:","Required Host":"Host obbligatorio","Result":"Risultato","Results":"Risultati","Rules Source":"Origine regole","Rules Version":"Versione regole","Run Auto-Attach":"Esegui Auto-Attach","Run subscription-manager in a console on the client host.  You may use an Activation Key to register:":"Esegui subscription-manager in una console sull'host del client. È possibile usare una Chiave di attivazione per la registrazione:","Save Successful.":"Salvato con successo.","Search...":"Cerca...","Security":"Sicurezza","Security Advisory":"Security Advisory","Select A New Puppet Module To Add":"Seleziona un nuovo modulo Puppet da aggiungere","Select Content Hosts":"Seleziona gli host di contenuto","Select Module":"Seleziona modulo","Select Version":"Seleziona versione","Select a Content Source:":"Seleziona un sorgente del contenuto:","Select a Version":"Seleziona una versione","Select an Available Version of {{ $stateParams.moduleName }}":"Seleziona una versione disponibile di {{ $stateParams.moduleName }}","Select an Organization":"Seleziona una organizzazione","Select new version":"Seleziona una nuova versione","Select the Lifecycle Environments you would like to remove Version {{ version.version }} from:":"Seleziona gli ambienti Ciclo di vita dai quali desideri rimuovere la versione {{ version.version }}:","Selected URLs:":"URL seleionati:","Selecting this option will result in the Version being completely deleted and no longer being available for promotion.  The version must be removed from all Lifecycle Environments in order to select this option.":"La selezione di questa opzione comporterà la cancellazione della Versione, la quale a sua volta non sarà più disponibile per l'avanzamento. Per poter selezionare questa opzione rimuovere la versione dagli ambienti Ciclo di vita.","Serve via HTTP.":"Servire via HTTP.","Service Level":"Livello di servizio","Severity":"Severitá","Show affected Activation Keys":"Mostra le chiavi di attivazione interessate","Show affected Content Hosts":"Mostra gli host di contenuto interessati","Show from:":"Mostra da:","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"Visualizzazione {{ detailsTable.rows.length }} degli host di contenuto {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Totale)","Showing {{ addSubscriptionsTable.rows.length }} of {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} Total) Subscriptions":"Visualizzazione {{ addSubscriptionsTable.rows.length }} delle sottoscrizioni {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} Totale)","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"Visualizzazione {{ detailsTable.rows.length }} degli host di contenuto {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Totale)","Showing {{ currentPackagesTable.resource.total }} Packages":"Visualizzazione {{ currentPackagesTable.resource.total }} pacchetti","Showing {{ subscriptionsTable.rows.length }} of {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} Total) Subscriptions":"Visualizzazione {{ subscriptionsTable.rows.length }} delle sottoscrizioni {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} Totale)","Sockets":"Socket","Sockets: %s":"Socket: %s","Solution":"Soluzione","Some of the Errata shown below may not be installable as they are not in this Content Host's\n          Content View and Lifecycle Environment.  In order to apply such Errata an Incremental Update is required.":"Alcuni degli errata di seguito indicati non possono essere installati poichè non sono presenti nell'Ambiente del ciclo di vita e\nnella Visualizzazione del contenuto dell'host. Per applicare questo tipo di Errata è necessario eseguire un Aggiornamento incrementale.","Stacking ID":"ID dello stack","Start Date":"Data d'inizio","Start Time":"Ora d'inizio","Started At":"Iniziato","Starting":"Avviata","Starts":"Inizia","State":"Stato","Status":"Stato","Subscription Details":"Informazioni sulla sottoscrizione","Subscription Info":"Informazioni sottoscrizione","Subscription Manifest":"Manifesto della sottoscrizione","Subscription Server Details":"Informazioni server delle sottoscrizioni","Subscription Status":"Stato sottoscrizione","Subscriptions":"Sottoscrizioni","Subscriptions for Activation Key:":"Sottoscrizioni per la chiave di attivazione:","Subscriptions for:":"Sottoscrizioni per:","Succesfully scheduled package installation":"Installazione pacchetto programmata con successo","Succesfully scheduled package removal":"Rimozione pacchetto programmata con successo","Succesfully scheduled package update":"Aggiornamento pacchetto programmato con successo","Success":"Con successo","Successfully Scheduled Auto-attach.":"Programmato con successo Auto-attach.","Successfully added %s subscriptions.":"Aggiunte con successo %s sottoscrizioni.","Successfully deleted %cv version %ver.":"Rimossa con successo %cv versione %ver.","Successfully initiated deletion of %cv version %ver.":"Inizializzata con successo la rimozione di %cv versione %ver.","Successfully initiated promotion of %cv version %ver to %env.":"Inizializzato con successo l'avanzamento di %cv versione %ver a %env.","Successfully initiated removal of %cv version %ver.":"Inizializzata con successo la rimozione di %cv versione %ver.","Successfully promoted %cv version %ver to %env":"Inizializzato con successo l'avanzamento di %cv versione %ver a %env.","Successfully published %cv version %ver and promoted to Library":"Pubblicazione con successo di %cv versione %ver e avanzamento su Libreria ","Successfully removed %cv version %ver from environments: %env":"Rimozione con successo di %cv versione %ver dagli ambienti: %env","Successfully removed %s items.":"Rimossi %s elementi con successo.","Successfully removed %s subscriptions.":"Rimosse con successo %s sottoscrizioni.","Successfully removed 1 item.":"Rimosse con successo 1 elemento.","Successfully updated subscriptions.":"Sottoscrizioni aggiornate con successo.","Summary":"Sommario","Support Level":"Livello supporto","Support Type":"Tipo di supporto","Sync Enabled":"Sinc abilitata","Sync Information":"Informazioni di sinc","Sync Interval":"Intervallo di sincronizzazione","Sync Now":"Sincronizza ora","Sync Plan":"Programmazione di sincronizzazione","Sync Plan %s has been deleted.":"La programmazione di sincronizzazione %s è stata rimossa.","Sync Plan Saved":"Programmazione sincronizzazione salvata","Sync Plan {{ syncPlan.name }}":"Programmazione sincronizzazione {{ syncPlan.name }}","Sync Plan:":"Programmazione sincronizzazione:","Sync Plans":"Programmazione di sinc","Sync State":"Stato sincronizzazione","Sync Status":"Stato di sinc","Synced manually, no interval set.":"Sincronizzato manualmente, nessun intervallo impostato.","Syncing...":"Sincronizzazione in corso...","Tags":"Tag","Target":"Destinazione","Task Details":"Informazione attività","Tasks":"Attività","Tasks for Content View:":"Attività per la visualizzazione dei contenuti:","Temporary":"Temporaneo","The Activation Keys listed below are currently using this Content View Version.  Before deleting the Version you must move these Activation Keys to a Lifecycle Environment where this Version is not in use.":"Le chiavi di attivazione qui di seguito elencate attualmente utilizzano questa versione per la visualizzazione dei contenuti. Prima di rimuovere la versione sarà necessario spostare le chiavi di attivazione in un ambiente Ciclo di vita, dove questa versione non è in uso.","The Content Hosts listed below are currently using this Content View Version.  Before removing the version you must move these Content Hosts to an environment where this version is not in use.":"Gli host di contenuto qui di seguito elencati attualmente utilizzano questa versione per la visualizzazione dei contenuti. Prima di rimuovere la versione sarà necessario spostare gli host di contenuto in un ambiente dove questa versione non è in uso.","The Content View or Lifecycle Environment needs to be updated in order to make errata available to these hosts.":"La Visualizzazione del contenuto o l'Ambiente del ciclo di vita devono essere aggiornati per rendere disponibili gli errata per gli host rilevanti.","The archive of this Version will be deleted.  This Version will not be available once deletion is complete.":"L'archivia di questa versione verrà rimosso. Dopo la sua rimozione la versione non risulterà più disponibile.","The archive of this Version will not be deleted.  This Version will still be available to be promoted to a Lifecycle Environment.":"L'archivio di questa versione non sarà rimosso. Questa versione potrà essere avanzata in un ambiente Ciclo di vita.","The following actions can be performed on content hosts in this host collection:":"È possibile eseguire le seguenti azioni su host di contenuto presenti in questo insieme:","The host requires being attached to a content view and the environment you have chosen has no content views promoted to it.\n            See the <a href=\"/content_views\">content views page</a> to manage and promote a content view.":"L'host necessita di una visualizzazione del contenuto, e non è stata assegnata alcuna visualizzazione all'ambiente selezionato. \n    Consultare la <a href=\"/content_views\">pagina delle visualizzazioni del contenuto</a> per gestire e promuovere una visualizzazione.","The katello-agent package is required to manage errata on this Host.\n      Please install the katello-agent package to manage errata on\n      {{ contentHost.name }}.":"È necessario aver installato il pacchetto katello-agent per gestire gli errata su questo host. Installare il pacchetto katello-agent per gestire gli errata\nsu \n{{ contentHost.name }}.","The katello-agent package is required to manage packages on this Host.\n      Please install the katello-agent package to manage packages on\n      {{ contentHost.name }}. To install the package, connect directly to the\n      host from a terminal and use yum.":"È necessario aver installato il pacchetto katello-agent per gestire i pacchetti su questo host. Installare il pacchetto katello-agent per gestire i pacchetti\nsu \n{{ contentHost.name }}. Per installare il pacchetto collegarsi\ndirettamente all'host dal terminale e usare yum.","The page you are attempting to access requires selecting a specific organization.":"Per accedere alla pagina desiderata selezionare una organizzazione specifica.","The selected environment contains no Content Views, please select a different environment.":"L'ambiente selezionato non presenta alcuna visualizzazione del contenuto, selezionare un altro ambiente.","There are currently no repositories associated with this Content View, you can add some by clicking on the \"Add\" tab above.":"Attualmente non è stato associato alcun repositorio a questa visualizzazione del contenuto, per questa operazione selezionare la scheda \"Aggiungi\".","There are currently no repositories to add to this Content View,":"Attualmente non è disponibile alcun repositorio da aggiungere a questa visualizzazione.","There are no Content Views in this Environment.":"Non è presente alcuna Visualizzazione del contenuto in questo Ambiente.","There are no Content Views that match the criteria.":"Non è disponibile alcuna Visualizzazione del contenuto corrispondente a questo criterio.","There are no Docker Tags that match the criteria.":"Non è disponibile alcun Tag Docker corrispondente al criterio.","There are no Errata associated with this Content Host to display.":"Non è presente alcun errata associato con questo Host di contenuto da visualizzare.","There are no Errata in this organization.  Create one or more Products with Errata to view Errata on this page.":"Non è presente alcun Errata in questa organizzazione. Creare uno o più prodotti con Errata per poterli visualizzare su questa pagina.","There are no Errata that match the criteria.":"Non è disponibile alcun errata corrispondente al criterio.","There are no Errata to display.":"Non è presente alcun errata da visualizzare.","There are no Host Collections available. You can create new Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"Non è disponibile alcun insieme di host. È  possibile creare un insieme dopo aver selezionato 'Insiemi di host' in 'Host' nel menu principale.","There are no Packages that match the criteria.":"Non è disponibile alcun pacchetto corrispondente al criterio.","There are no Puppet Modules that match the criteria.":"Non è disponibile alcun Modulo puppet corrispondente al criterio.","There are no Yum Repositories that match the criteria.":"Non è disponibile alcun Repository Yum corrispondente al criterio.","There is currently an Incremental Update task in progress.  This update must finish before applying existing updates.":"È attualmente in corso un Aggiornamento incrementale. Il suddetto aggiornamento deve terminare prima di applicare gli aggiornamenti esistenti.","There were errors while removing the following Repositories:":"Si sono verificati alcuni errori durante la rimozione dei reguenti repository:","This Content Host is not attached to any subscriptions that provide content":"Questo host non è stato assegnato ad alcuna sottoscrizione che fornisce un contenuto","This Content View does not have any versions, create your first Content View Version by using the \"Publish New Version\" button on the right.":"Questa Visualizzazione del contenuto non presenta alcuna versione, creare la prima Versione usando il pulsante \"Pubblica nuova versione\" sulla destra.","This Version cannot be deleted from some of the Lifecycle Environments due to one of the reasons below.":"Questa versione non può essere rimossa da alcuni ambienti Ciclo di vita a causa di uno dei seguenti motivi.","This Version is not associated with any Lifecycle Environments.":"Questa versione non è stata associata con alcun ambiente Ciclo di vita.","This activation key is not associated with any content hosts.":"Questa chiave di attivazione non è stata associata con alcun host di contenuto.","This activation key may be used during system registration. For example:":"Questa chiave di attivazione può essere usata durante la registrazione del sistema. Per esempio:","This filter applies only to a subset of repositories in the content view.":"Questo filtro viene applicato solo a una sottorete di repositori presenti nella visualizzazione del contenuto.","This filter applies to all repositories in the content view (current and future).":"Questo filtro viene applicato su tutti i repositori presenti nella visualizzazione del contenuto (attuali e futuri).","This subscription is not associated with any activation keys.":"Questa sottoscrizioni non è associata con alcuna chiave di attivazione.","This subscription is not attached to any content hosts.":"Questa sottoscrizioni non è stata assegnata ad alcun host di contenuto.","This version will be removed from:":"Questa versione verrà rimossa da:","Time":"Ora","Title":"Titolo","To register a content host to this server, follow these steps.":"Per registrare un host di contenuto a questo server, eseguire le fasi di seguito riportate.","Topic":"Argomento","Type":"Tipo","URL":"URL","URL of the registry you want to sync. Example: https://registry.hub.docker.com":"URL del registro da sincronizzare. Esempio: https://registry.hub.docker.com","URL to discover":"URL per la scoperta","UUID":"UUID","Unattach Sync Plan":"Rimuovi assegnazione programmazione per la sincronizzazione","Unattaching...":"Rimozione in corso...","Unlimited Content Hosts:":"Host di contenuto illimitati:","Unregister Content Host":"Rimozione registrazione host di contenuto","Unregister Content Host \"{{contentHost.name}}\"?":"Rimozione registrazione host di contenuto \"{{contentHost.name}}\"?","Unregister Content Hosts":"Rimozione registrazione host di contenuto","Unregister {{ table.numSelected  }} Content Hosts?":"Rimuovere la registrazione di {{ table.numSelected  }} host di contenuto?","Update":"Aggiorna","Update All":"Aggiorna tutti","Update Repositories":"Aggiorna repositori","Update Sync Plan":"Aggiorna programmazione sincronizzazione","Updated":"Aggiornato","Updated errata filter -":"Filtro errata aggiornato - ","Updated override to \"%x\".":"Override aggiornato a \"%x\".","Updating...":"Aggiornamento in corso...","Upload":"Carica","Upload Docker Image":"Carica immagine Docker","Upload GPG Key":"Carica chiave GPG","Upload New Manifest":"Carica un nuovo manifesto","Upload Package":"Carica pacchetto","Upload Puppet Module":"Carica modulo Puppet","Uploading...":"Caricamento in corso...","Upstream Repository Name":"Nome repository upstream","Upstream Subscription":"Sottoscrizione originale","Upstream Subscription Management Application":"Applicazione di gestione della sottoscrizione originale","Upstream Type":"Tipo upstream","Upstream User":"Utente upstream","Use Latest (currently %s)":"Usa ultimissima (attualmente %s)","Use the cancel button on content view selection to revert your environment selection.":"Usa il pulsante cancella sulla selezione della visualizzazione del contenuto, per repristinare la selezione dell'ambiente.","User":"Utente","Version":"Versione","Version %s":"Versione %s","Version 1":"Versione 1","Version Deletion":"Rimozione versione","Version Details":"Informazioni versione","Version {{ version.version }}":"Versione {{ version.version }}","Version:":"Versione:","Versions":"Versioni","Versions for Content View:":"Versioni per la visualizzazione del contenuto:","Versions for Puppet Module:":"Versioni per il modulo Puppet:","View Details":"Dettagli visualizzazione","Virtual":"Virtuale","Virtual Guest From":"Guest virtuale da","Virtual Guests":"Guest virtuali","Virtual Host":"Host virtuale","Weekly on {{ product.sync_plan.sync_date | date:'EEEE' }} at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"Settimanale il {{ product.sync_plan.sync_date | date:'EEEE' }} alle {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)","When Auto Attach is disabled, registering systems will be attached to all associated subscriptions.":"Quando Auto Attach è disabilitato, la registrazione dei sistemi verrà implementata a tutte le sottoscrizioni associate","Working":"In funzione","Yes":"Si","Yes (Default)":"Si (Predefinito)","You cannot remove these repositories because they belong to a Red Hat product.":"Impossibile rimuovere questi repositori poichè appartengono ad un prodotto di Red Hat. ","You cannot remove these repositories because you do not have permission.":"Impossibile rimuovere questi repositori poichè non sei in possesso dei permessi necessari. ","You cannot remove this product because it is a Red Hat product.":"Impossibile rimuovere questo prodotto poichè appartiene ad un prodotto di Red Hat. ","You cannot remove this product because it was published to a content view.":"Impossibile rimuovere questo prodotto poichè è stato pubblicato su una visualizzazione del contenuto.","You cannot remove this product because you do not have permission.":"Impossibile rimuovere questo prodotto poichè non sei in possesso dei permessi necessari. ","You cannot remove this repository because it is a Red Hat repository.":"Impossibile rimuovere questo repositorio poichè appartiene ad un repositorio di Red Hat. ","You cannot remove this repository because it was published to a content view.":"Impossibile rimuovere questo repositorio poichè è stato pubblicato su una visualizzazione del contenuto.","You cannot remove this repository because you do not have permission.":"Impossibile rimuovere questo repositorio poichè non sei in possesso dei permessi necessari. ","You currently do not have any Packages installed. You can use Package Actions above this message to install new Packages.":"Attualmente non è stato installato alcun pacchetto. Usare le 'Azioni per i pacchetti' sopra questo messaggio, per installare nuovi pacchetti.","You currently don't have any Activation Keys, you can add Activation Keys using the button on the right.":"Attualmente non è presente alcuna Chiave di attivazione, è possibile aggiungerne una usando il pulsante sulla destra.","You currently don't have any Content Hosts in this Host Group, you can add Content Hosts after selecting the 'Add' tab.":"Attualmente non è presente alcun host di contenuto all'interno del gruppo, è possibile aggiungerne uno dopo aver selezionato la scheda 'Aggiungi'.","You currently don't have any Content Hosts, you can create new Content Hosts by selecting Contents Host from main menu and then clicking the button on the right.":"Attualmente non è presente alcun host di contenuto, è possibile crearne uno nuovo selezionando Host di contenuto dal menu principale, e successivamente il pulsante sulla destra.","You currently don't have any Content Hosts, you can register one by clicking the button on the right and following the instructions.":"Attualmente non sei inpossesso di alcun Host di contenuto, per registrarne uno seleziona il pulsante sulla destra e segui le istruzioni.","You currently don't have any Content Views.  A Content View can be added by using the button on the right.":"Attualemente non è presente alcuna Visualizzazione del contenuto. È possibile aggiungerne una usando il pulsante sulla destra.","You currently don't have any Docker Tags.":"Attualmente non è presente alcun Tag Docker.","You currently don't have any Filters included in this Content View, you can add a new Filter by using the button on the right.":"Attualmente non hai incluso alcun filtro nella visualizzazione del contenuto, aggiungere un nuovo filtro usando il pulsante sulla destra.","You currently don't have any Gpg keys, you can add Gpg keys using the button on the right.":"Attualmente non è presente alcuna Chiave GPG, è possibile aggiungerne una usando il pulsante sulla destra.","You currently don't have any Host Collections, you can add Host Collections using the button on the right.":"Attualmente non è presente alcun Insieme di host, è possibile aggiungerne uno usando il pulsante sulla destra.","You currently don't have any Products associated with this Gpg Key.":"Attualmente nessun prodotto è stato associato con questa Chiave di attivazione.","You currently don't have any Products to subscribe to, you can add Products after selecting 'Products' under 'Content' in the main menu.":"Attualmente non è presente alcun prodotto per la sottoscrizione, è possibile aggiungere i Prodotti dopo aver selezionato 'Prodotti' in \"Contenuto' nel menu principale.","You currently don't have any Products<span bst-feature-flag=\"custom_products\">, you can add Products using the button on the right</span>.":"Attualmente non hai alcun prodotto <span bst-feature-flag=\"custom_products\">, è possibile aggiungere i Prodotti usando il pulsante sulla destra</span>.","You currently don't have any Puppet Modules included in this Content View, you can add Puppet Modules using the button on the right.":"Attualmente non hai incluso alcun Modulo Puppet nella Visualizzazione del contenuto, è possibile aggiungerne uno usando il pulsante sulla destra.","You currently don't have any Repositories associated with this Gpg Key.":"Attualmente nessun Repository è stato associato con questa Chiave GPG.","You currently don't have any Repositories included in this Product, you can add Repositories using the button on the right.":"Attualmente non hai incluso alcun Repository in questo Prodotto, è possibile aggiungerne uno usando il pulsante sulla destra.","You currently don't have any Subscriptions associated with this Content Host, you can add Subscriptions after selecting the 'Add' tab.":"Attualmente non è presente alcuna Sottoscrizione associata a questo Host di contenuti. È possibile aggiungerne una dopo aver selezionato la scheda 'Aggiungi'.","You currently don't have any Subscriptions, you can add Subscriptions by importing a manifest using the button on the right labeled 'Manage Manifest'.":"Attualmente sei sprovvisto di Sottoscrizioni. È possibile aggiungere una Sottoscrizione importando un manifesto usando il pulsante sulla destra, 'Gestisci manifesto'.","You currently don't have any Sync Plan, you can add Sync Plans using the button on the right.":"Attualmente non sei in possesso di una Programmazione di sincronizzazione. È possibile aggiungere una Programmazione usando il pulsante sulla destra.","You currently don't have any Sync Plans.  A Sync Plan can be created by using the button on the right.":"Attualmente non sei in possesso di una Programmazione di sincronizzazione. Per crearne una usare il pulsante sulla destra.","You do not have any Installed Products":"Non hai installato alcun prodotto","You must select a content view in order to save your environment.":"Per salvare il tuo ambiente è necessario selezionare una visualizzazione del contenuto.","You must select a new content view before your change of environment can be saved.":"È necessario selezionare una nuova visualizzazione prima di poter salvare il nuovo ambiente.","You must select a new content view before your change of environment can be saved. Use the cancel button on content view selection to revert your environment selection.":"È necessario selezionare una nuova visualizzazione prima di poter salvare il nuovo ambiente. Usa il pulsante cancella sulla selezione della visualizzazione del contenuto, per repristinare la selezione dell'ambiente.","You must select at least one Content Host in order to apply Errata.":"Per applicare un Errata è necessario selezionare un Host di contenuto.","You must select at least one repository.":"Selezionare almeno un repositorio.","Your selected content hosts have changed, it is highly recommended to refresh the calculated list of errata using the 'Refresh Errata' button below.":"Gli host di contenuto selezionati sono stati modificati, è fortemente consigliato aggiornare l'elenco di errata usando il pulsante \"Aggiorna errata\".","Yum Content":"Contenuto di yum","Yum Metadata Checksum":"Checksum metadati di yum","Yum Repo Discovery":"Scoperta repositorio Yum","Yum Repositories":"Repository yum","Yum Repositories <div>{{ library.counts.yum_repositories || 0 }}</div>":"Repositori yum <div>{{ library.counts.yum_repositories || 0 }}</div>","Yum metadata generation has been initiated in the background.  Click <a href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"L'operazione di creazione dei metadati Yum stata iniziata nel background. Fare clic <a href=\"{{ taskUrl() }}\">Qui</a> per controllare lo stato.","add some repositories.":"aggiungi alcuni repositori.","daily":"giornaliero","here":"qui","hourly":"ad ogni ora","or authenticate with a username and password:":"o eseguire una autenticazione con un nome utente e password:","subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"":"subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"","task details":"Dettagli dell'attività","to manage them individually.":"per una loro gestione individuale.","true":"vero","weekly":"settimanalmente","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }} Selezionato","{{ addSubscriptionsTable.numSelected }} Selected":"{{ addSubscriptionsTable.numSelected }} Seleczionato","{{ contentHost.registered_by }}":"{{ contentHost.registered_by }}","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }} Selezionato","{{ contentView.name }} cannot be deleted as one or more Content View Versions are still promoted to a Lifecycle Environment.\n      Each Content View Version must be removed from their Lifecycle Environments before the Content View can be deleted.":"{{ contentView.name }} non può essere rimosso poichè una o più versioni della Visualizzazione del contenuto sono state avanzate nell'ambiente Ciclo di vita.\n      Ogni versione deve essere rimossa dal rispettivo ambiente Ciclo di vita, prima di poter cancellare una Visualizzazione del contenuto.","{{ contentView.puppet_modules.length || 0 }} Puppet Modules":"{{ contentView.puppet_modules.length || 0 }} Moduli Puppet","{{ contentView.repositories.length || 0 }} Repositories":"{{ contentView.repositories.length || 0 }} Repositori","{{ contentViewVersion.puppet_module_count || 0 }} Puppet Modules":"{{ contentViewVersion.puppet_module_count || 0 }} Moduli Puppet","{{ contentViewVersion.repositories.length || 0 }} Repositories":"{{ contentViewVersion.repositories.length || 0 }} Repositori","{{ discoveryTable.numSelected }} Selected":"{{ discoveryTable.numSelected }} Selezionato","{{ environment.name }} Environment":"Ambiente {{ environment.name }}","{{ errata.hosts_applicable_count || 0 }} Applicable,":"{{ errata.hosts_applicable_count || 0 }} applicabile,","{{ errata.hosts_available_count || 0 }} Installable":"{{ errata.hosts_available_count || 0 }} Installabile","{{ hostCollectionsTable.numSelected }} Selected":"{{ hostCollectionsTable.numSelected }} Selezionato","{{ package.removeTask.affected_units }} package removed":"{{ package.removeTask.affected_units }} pacchetto rimosso","{{ package.removeTask.affected_units }} packages removed":"{{ package.removeTask.affected_units }} pacchetti  rimossi","{{ product.last_sync_words }} Ago ({{ product.sync_status.finish_time | date:'medium' }} Local Time)":"{{ product.last_sync_words }} fa ({{ product.sync_status.finish_time | date:'medium' }} Ora locale)","{{ product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ product.sync_plan.next_sync | date:'medium' }} (Server Time)","{{ productsTable.numSelected }} Selected":"{{ productsTable.numSelected }} Selezionato","{{ repository.content_counts.docker_image || 0 }} Docker Images":"Immagini Docker {{ repository.content_counts.docker_image || 0 }}","{{ repository.content_counts.docker_image }} Docker Images":"Immagini Docker {{ repository.content_counts.docker_image }}","{{ repository.content_counts.docker_tag || 0 }} Docker Tags":"{{ repository.content_counts.docker_tag || 0 }} Docker Tag","{{ repository.content_counts.docker_tag }} Docker Tags":"{{ repository.content_counts.docker_tag }} Docker Tag","{{ repository.content_counts.erratum || 0 }} Errata":"{{ repository.content_counts.erratum || 0 }} Errata","{{ repository.content_counts.erratum }} Errata":"{{ repository.content_counts.erratum }} Errata","{{ repository.content_counts.puppet_module || 0 }} Puppet Modules":"{{ repository.content_counts.puppet_module || 0 }} Moduli Puppet","{{ repository.content_counts.rpm || 0 }} Packages":"{{ repository.content_counts.rpm || 0 }} Pacchetti","{{ repository.content_counts.rpm }} Packages":"Pacchetti {{ repository.content_counts.rpm }}","{{ repository.last_sync_words }} Ago ({{ repository.last_sync.ended_at | date:'medium' }} Local Time)":"{{ repository.last_sync_words }} fa ({{ repository.last_sync.ended_at | date:'medium' }} Ora locale)","{{ repository.last_sync_words }} ago":"{{ repository.last_sync_words }} fa","{{ repository.product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ repository.product.sync_plan.next_sync | date:'medium' }} (Server Time)","{{ subscription.host.name }}":"{{ subscription.host.name }}","{{ subscriptionsTable.numSelected }} Selected":"{{ subscriptionsTable.numSelected }} Selezionato","{{ totalActivationKeyCount() }} Activation Keys will be moved to {{ deleteOptions.activationKeys.contentView.name }} in {{ deleteOptions.activationKeys.environment.name }}":"{{ totalActivationKeyCount() }} Chiavi di attivazione saranno spostate su {{ deleteOptions.activationKeys.contentView.name }} in {{ deleteOptions.activationKeys.environment.name }}","{{ totalHostCount() }} Content Hosts will be moved to {{ deleteOptions.contentHosts.contentView.name }} in {{ deleteOptions.contentHosts.environment.name }}":"{{ totalHostCount() }} Host di contenuto saranno spostati su {{ deleteOptions.contentHosts.contentView.name }} in {{ deleteOptions.contentHosts.environment.name }}","{{ type.display }}":"{{ type.display }}","{{ version.docker_image_count }} Docker Images":"Immagini Docker {{ version.docker_image_count }}","{{ version.docker_tag_count }} Docker Tags":"{{ version.docker_tag_count }} Docker Tag","{{ version.errata_counts.total }} Errata":"{{ version.errata_counts.total }} Errata","{{ version.package_count }} Packages":"Pacchetti {{ version.package_count }}","{{ version.puppet_module_count }} Puppet Modules":"Moduli Puppet {{ version.puppet_module_count }}"});
    gettextCatalog.setStrings('ja', {"%(consumed)s out of %(quantity)s":"%(quantity)s 件数中 %(consumed)s 件 ","(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)":"(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"タスクの詳細\" | translate }}</a>)","(Not all Activation Keys editable )":"(すべてのアクティベーションキーが編集可能である訳ではない )","(Not all Content Hosts editable )":"(すべてのコンテンツホストが編集可能である訳ではない )","* These marked Content View Versions are from Composite Content Views.  Their components needing updating are listed underneath.":"* これらのマークが付けられたコンテンツビューのバージョンは複合コンテンツビューのバージョンです。更新する必要のあるそれらのコンポーネントは以下に一覧表示されています。","+ New Sync Plan":"+ 新規同期プラン","- Date and Type":"- 日付およびタイプ","-- select an interval --":"-- 間隔の選択 --","1 Content Host":"{{ contentHost.virtual_guests.length }} コンテンツホスト","1 repository sync has errors.":"{{ product.sync_summary.error}} リポジトリーの同期にエラーがあります。","1 repository sync in progress.":"{{ product.sync_summary.pending}} リポジトリーの同期が進行中です。","1 successfully synced repository.":"{{ product.sync_summary.success}} リポジトリーが正常に同期されました。","<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Version {{ version.version }}</a>":"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">バージョン {{ version.version }}</a>","<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Create":"<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n           作成","<span translate=\"\">\n        If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">here</a>\n      <span translate=\"\">to manage them individually.</span>":"<span translate=\"\">\n   これらのアクティベーションキーの一部を別のコンテンツビューまたはライフサイクル環境に移動させる場合は、\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">こちら</a>\n      <span translate=\"\">をクリックして、それらを別々に管理します。</span>","A composite view contains other content views.":"複合ビューには他のコンテンツビューが組み込まれています。","A new version of {{ contentView.name }} will be created and promoted to the Library environment.\n    It can be promoted to other environments from the Versions tab of this Content View.":"新規バージョンの {{ contentView.name }} が作成され、ライブラリー環境にプロモートされます。\n  このコンテンツビューのバージョンタブから、他の環境にプロモートすることができます。","Account":"アカウント","Account Number":"アカウント番号","Action":"アクション","Action Type":"アクションタイプ","Actions":"アクション","Actions for:":"アクションの対象:","Activation Key":"アクティベーションキー","Activation Key Content":"アクティベーションキーのコンテンツ","Activation Key Type:":"アクティベーションキーのタイプ:","Activation Key removed.":"アクティベーションキーが削除されました。","Activation Key updated":"アクティベーションキーが更新されました","Activation Key:":"アクティベーションキー:","Activation Key: {{ activationKey.name }}":"アクティベーションキー: {{ activationKey.name }}","Activation Keys":"アクティベーションキー","Activation Keys for Subscription:":"サブスクリプションのアクティベーションキー:","Activation Keys using Version {{ version.version }}":"バージョン {{ version.version }} を使用したアクティベーションキー","Add":"追加","Add Content Hosts to:":"コンテンツホストの追加先:","Add Content Views":"コンテンツビューの追加","Add Errata":"エラータの追加","Add New Environment":"新規環境の追加","Add New Filter":"新規フィルターの追加","Add New Module":"新規モジュールの追加","Add New Repository":"新規リポジトリーの追加","Add Package Group":"パッケージグループの追加","Add Repositories":"リポジトリーの追加","Add Selected":"選択を追加","Add Subscriptions for Activation Key:":"アクティベーションキーのサブスクリプションを追加:","Add To":"追加先","Added %x host collections to activation key \"%y\".":"%x ホストコレクションをアクティベーションキー \"%y\" に追加。","Added %x host collections to content host \"%y\".":"%x ホストコレクションをコンテンツホスト \"%y\" に追加。","Added %x products to sync plan \"%y\".":"%x 製品を同期プラン \"%y\" に追加。","Adding Lifecycle Environment to the end of \"{{ priorEnvironment.name }}\"":"ライフサイクル環境を \"{{ priorEnvironment.name }}\" の末尾に追加します","Advanced Information":"詳細情報","Advisory":"アドバイザリー","Affected Activation Keys":"影響するアクティベーションキー","Affected Content Hosts":"影響するコンテンツホスト","Affected Packages":"影響するパッケージ","Affected Systems":"影響するシステム","Affected repositories":"影響するリポジトリー","Affected repositories have been updated.":"影響するリポジトリーが更新されました","All Content Views":"すべてのコンテンツビュー","All Products":"すべての製品","All Repositories":"すべてのリポジトリー","All Versions":"すべてのバージョン","Alter Sync Plans":"同期プランの変更","An error occurred applying Subscriptions:":"サブスクリプションの適用時にエラーが発生しました:","An error occurred removing the Activation Key:":"アクティベーションキーの削除時にエラーが発生しました:","An error occurred removing the Host Collection:":"ホストコレクションの削除時にエラーが発生しました:","An error occurred removing the Products:":"製品の削除時にエラーが発生しました:","An error occurred removing the content hosts.":"コンテンツホストの削除時にエラーが発生しました。","An error occurred removing the environment:":"環境の削除時にエラーが発生しました:","An error occurred removing the subscriptions.":"サブスクリプションの削除時にエラーが発生しました。","An error occurred saving the Activation Key:":"アクティベーションキーの保存時にエラーが発生しました:","An error occurred saving the Content Host:":"コンテンツホストの保存時にエラーが発生しました:","An error occurred saving the Environment:":"環境の保存時にエラーが発生しました:","An error occurred saving the Filter:":"フィルターの保存時にエラーが発生しました:","An error occurred saving the Host Collection:":"ホストコレクションの保存時にエラーが発生しました:","An error occurred saving the Product:":"製品の保存時にエラーが発生しました:","An error occurred saving the Repository:":"リポジトリーの保存時にエラーが発生しました:","An error occurred saving the Sync Plan:":"同期プランの保存時にエラーが発生しました:","An error occurred saving the URL:":"URL の保存時にエラーが発生しました:","An error occurred trying to auto-attach subscriptions.  Please check your log for further information.":"サブスクリプションの自動割り当てを試行する際にエラーが発生しました。ログで詳細情報を確認してください。","An error occurred updating the Content View:":"コンテンツビューの更新時にエラーが発生しました。","Applicable":"適用可能","Applicable Content Hosts for Errata":"エラータの適用可能なコンテンツホスト","Applicable Errata":"適用可能なエラータ","Applicable hosts for {{ erratum.errata_id }}":"{{ erratum.errata_id }} の適用可能なホスト","Apply":"適用","Apply Errata":"エラータを適用","Apply Errata to Content Host \"{{contentHost.name}}\"?":"エラータをコンテンツホスト \"{{contentHost.name}}\" に適用しますか?","Apply Errata to Content Hosts":"エラータをコンテンツホストに適用","Apply Errata to Content Hosts immediately after publishing.":"公開直後にエラータをコンテンツホストに適用","Apply Selected":"選択を適用","Apply to Content Hosts":"コンテンツホストに適用","Apply to Hosts":"ホストに適用","Apply {{ errata.errata_id }}":"{{ errata.errata_id }} を適用","Apply {{ errata.errata_id }} to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"{{ errata.errata_id }} を {{ numberOfContentHostsToUpdate }} コンテンツホストに適用しますか?","Apply {{ errataIds.length }} Errata to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"{{ errataIds.length }} エラータを {{ numberOfContentHostsToUpdate }} コンテンツホストに適用しますか?","Applying":"適用中","Applying Bulk Errata Update":"一括エラータ更新を適用しています","Arch":"アーキテクチャー","Architecture(s)":"アーキテクチャー","Archived Copy":"アーカイブされたコピー","Are you sure you want to add the {{ table.numSelected }} content host(s) selected to the host collection(s) chosen?":"選択済みのホストコレクションに対して選択された {{ table.numSelected }} コンテンツホストを追加してもよろしいですか?","Are you sure you want to apply Errata to content host \"{{ contentHost.name }}\"?":"エラータをコンテンツホスト \"{{ contentHost.name }}\" に適用してもよろしいですか?","Are you sure you want to apply the {{ detailsTable.numSelected }} selected errata to the content hosts chosen?":"選択済みのコンテンツホストに対して選択された  {{ detailsTable.numSelected }} エラータを適用してもよろしいですか?","Are you sure you want to assign the {{ table.numSelected }} content host(s) selected to {{ selected.contentView.name }} in {{ selected.environment.name }}?":"{{ selected.environment.name }} 内の  {{ selected.contentView.name }} に対して選択された {{ table.numSelected }} コンテンツホストを割り当ててもよろしいですか?","Are you sure you want to auto-attach available subscriptions to all {{ table.resource.total }} content hosts?":"すべての {{ table.resource.total }} コンテンツホストに対する利用可能なサブスクリプションの自動割り当てを実行してもよろしいですか?","Are you sure you want to delete Content View \"{{ contentView.name }}\"?":"コンテンツビュー \"{{ contentView.name }}\" を削除してもよろしいですか?","Are you sure you want to install {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"選択済みの  {{ getSelectedSystemIds().length }} システムに {{ content.content }} をインストールしてもよろしいですか?","Are you sure you want to remove Activation Key \"{{ activationKey.name }}\"?":"アクティベーションキー \"{{ activationKey.name }}\" を削除してもよろしいですか?","Are you sure you want to remove GPG key {{ gpgKey.name }}?":"GPG キー {{ gpgKey.name }} を削除してもよろしいですか?","Are you sure you want to remove Host Collection \"{{ hostCollection.name }}\"?":"ホストコレクション \"{{ hostCollection.name }}\" を削除してもよろしいですか?","Are you sure you want to remove Sync Plan \"{{ syncPlan.name }}\"?":"同期プラン \"{{ syncPlan.name }}\" を削除してもよろしいですか?","Are you sure you want to remove product \"{{ product.name }}\"?":"製品 \"{{ product.name }}\" を削除してもよろしいですか?","Are you sure you want to remove repository \"{{ repository.name }}\"?":"リポジトリー \"{{ repository.name }}\" を削除してもよろしいですか?","Are you sure you want to remove the {{ productTable.getSelected().length }} products(s) selected?":"選択された {{ productTable.getSelected().length }} 製品を削除してもよろしいですか?","Are you sure you want to remove the {{ table.numSelected }} content host(s) selected from the host collection(s) chosen?":"選択済みのホストコレクションから選択された {{ table.numSelected }} コンテンツホストを削除してもよろしいですか?","Are you sure you want to remove {{ content.content }} from the {{ getSelectedSystemIds().length }} system(s) selected?":"選択済みの {{ getSelectedSystemIds().length }} システムから {{ content.content }} を削除してもよろしいですか?","Are you sure you want to unregister content host \"{{ contentHost.name }}\"?":"コンテンツホスト \"{{ contentHost.name }}\" の登録を解除してもよろしいですか?","Are you sure you want to unregister the {{ table.numSelected }} content host(s) selected?":"選択された {{ table.numSelected }} コンテンツホストの登録を解除してもよろしいですか?","Are you sure you want to update {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"選択済みの {{ getSelectedSystemIds().length }} システムで {{ content.content }} を更新してもよろしいですか?","As part of this deletion, 1 Content View Version will be deleted.":" この削除の一部として、 {{ versions.length }} コンテンツビューバージョンは削除されます。","Assign":"割り当て","Assign Environment and Content View":"環境およびコンテンツビューの割り当て","Associations":"関連付け","At least one Errata needs to be selected to Apply.":"適用する 1 つ以上のエラータを選択する必要があります。","At least one content host needs to be selected in order to perform a bulk action.":"一括処理を実行するために、1 つ以上のコンテンツホストを選択する必要があります。","At least one product needs to be selected in order to perform a bulk action.":"一括処理を実行するために、1 つ以上の製品を選択する必要があります。","Attached":"割り当て済み","Attached to Content Hosts":"コンテンツホストに割り当て済み","Author":"作成者","Auto-Attach":"自動割り当て","Auto-attach available subscriptions to all content hosts":"利用可能なサブスクリプションのすべてのコンテンツホストへの自動割り当て","Automatic":"自動","Available Content Views for Composite Content View:":"複合コンテンツビューに利用できるコンテンツビュー:","Available Puppet Modules for Content View:":"コンテンツビューに利用できる Puppet モジュール:","Back":"戻る","Back To Errata List":"エラータ一覧に戻る","Back To Repository Selection":"リポジトリー選択に戻る","Back to Module List":"モジュール一覧に戻る","Back to Repository Details":"リポジトリーの詳細に戻る","Back to Repository List":"リポジトリー一覧に戻る","Back to Version List":"バージョン一覧に戻る","Basic Information":"基本情報","Below are the product's repository content sets currently available for this subscription. For Red Hat subscriptions, additional content can be made available through the":"以下は、このサブスクリプションで現在利用できる製品のリポジトリーコンテンツセットです。Red Hat サブスクリプションの場合、以下から追加のコンテンツを利用できます:","Below are the repository content sets currently available for this activation key through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"以下は、サブスクリプションにより、このアクティベーションキーに対して現在利用できるリポジトリーコンテンツセットです。Red Hat サブスクリプションの場合、以下から追加のコンテンツを利用できます:","Below are the repository content sets currently available for this content host through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"以下は、サブスクリプションにより、このコンテンツホストに対して現在利用できるリポジトリーコンテンツセットです。Red Hat サブスクリプションの場合、以下から追加のコンテンツを利用できます:","Bug Fix":"バグ修正","Bug Fix Advisory":"バグ修正アドバイザリー","Bugfix":"バグ修正","Bulk Actions":"一括処理","Bulk Alter Products Sync Plan":"製品同期プランの一括変更","Bulk Sync Products":"製品の一括同期","CVEs":"CVE","Cancel":"取り消し","Capacity":"容量","Change assigned Lifecycle Environment or Content View":"割り当て済みのライフサイクル環境またはコンテンツビューの変更","Changing default settings for content hosts that register with this activation key requires subscription-manager version 1.10 or newer to be installed on that host.":"このアクティベーションキーで登録されるコンテンツホストのデフォルト設定を変更するには、subscription-manager バージョン 1.10 以降をホストにインストールする必要があります。","Changing default settings requires subscription-manager version 1.10 or newer to be installed on this host.":"デフォルト設定を変更するには、subscription-manager バージョン 1.10 以降をこのホスト上にインストールする必要があります。","Checkin":"チェックイン","Checksum":"チェックサム","Choose a lifecycle environment from the available promotion paths.":"利用可能なプロモーションパスからライフサイクル環境を選択します。","Click here to check the status of the task.":"こちらをクリックしてタスクの状態を確認します。","Click here to refresh.":"こちらをクリックして更新します。","Click here to see the task for the successful removals.":"こちらをクリックして正常に実行された削除のタスクを確認します。","Click here to select Errata for an Incremental Update.":"こちらをクリックして増分更新のエラータを選択します。","Close":"閉じる","Collection Actions":"コレクションの各種アクション","Completely remove version?":"バージョンを完全に削除しますか?","Components":"コンポーネント","Components:":"コンポーネント:","Composite Content View {{ contentView.name }}":"複合コンテンツビュー {{ contentView.name }}","Composite View":"複合ビュー","Composite View?":"複合ビュー?","Composite?":"複合ですか?","Confirm":"確定します","Confirm Remove":"削除の確認","Confirm Version Removal: Version {{ version.version }}":"バージョンの削除を確認: バージョン {{ version.version }}","Confirmation":"確認","Consumed":"使用済み","Content":"コンテンツ","Content Counts":"コンテンツ数","Content Download URL":"コンテンツダウンロード URL","Content Host":"コンテンツホスト","Content Host %s has been deleted.":"コンテンツホスト %s が削除されました。","Content Host Bulk Actions":"コンテンツホストの一括処理","Content Host Bulk Content":"コンテンツホストの一括コンテンツ","Content Host Bulk Errata":"コンテンツホストの一括エラータ","Content Host Bulk Host Collections":"コンテンツホストの一括ホストコレクション","Content Host Bulk Packages":"コンテンツホストの一括パッケージ","Content Host Bulk Subscriptions":"コンテンツホストの一括サブスクリプション","Content Host Content":"コンテンツホストのコンテンツ","Content Host Counts":"コンテンツホスト数","Content Host Errata Management":"コンテンツホストのエラータ管理","Content Host Limit":"コンテンツホストの制限","Content Host Package Management":"コンテンツホストのパッケージ管理","Content Host Properties":"コンテンツホストのプロパティー","Content Host Registration":"コンテンツホストの登録","Content Host Status":"コンテンツホストの状態","Content Host Subscriptions":"コンテンツホストのサブスクリプション","Content Host {{ contentHost.name }}":"コンテンツホスト {{ contentHost.name }}","Content Host:":"コンテンツホスト:","Content Hosts":"コンテンツホスト","Content Hosts for Activation Key:":"アクティベーションキーのコンテンツホスト:","Content Hosts for Subscription:":"サブスクリプションのコンテンツホスト:","Content Hosts for:":"次のコンテンツホスト:","Content Hosts using Version {{ version.version }}":"バージョン {{ version.version }} を使用するコンテンツホスト","Content Type":"コンテンツタイプ","Content View":"コンテンツビュー","Content View Name":"コンテンツビュー名","Content View Version":"コンテンツビューのバージョン","Content View updated.":"コンテンツビューが更新されました。","Content View:":"コンテンツビュー:","Content Views":"コンテンツビュー","Content Views <div>{{ library.counts.content_views || 0 }}</div>":"コンテンツビュー: <div>{{ library.counts.content_views || 0 }}</div>","Content Views for Composite Content View:":"複合コンテンツビューのコンテンツビュー:","Content successfully uploaded":"コンテンツが正常にアップロードされました","Contents":"コンテンツ","Contract":"コントラクト","Contract Number":"コントラクト番号","Copy":"コピー","Copy Activation Key":"アクティベーションキーのコピー","Copy Collection":"コレクションのコピー","Copy View":"ビューのコピー","Cores per Socket":"ソケットあたりのコア数","Cores: %s":"コア: %s","Create":"作成","Create New View":"新規ビューの作成","Create Repositories":"リポジトリーの作成","Create Repositories Within:":"リポジトリーの作成場所:","Create Repository":"リポジトリーの作成","Create Selected":"選択を作成","Create Sync Plan":"同期プランの作成","Create new view":"新規ビューの作成","Create view from existing views":"既存ビューからビューを作成","Critical":"危険","Current Environment (%e/%cv)":"現在の環境 (%e/%cv)","Current Subscriptions for Activation Key:":"アクティベーションキーの現在のサブスクリプション:","Currently Selected Puppet Modules":"現在選択されている Puppet モジュール","Daily at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"毎日: {{ product.sync_plan.sync_date | date:'mediumTime' }} (サーバー時間)","Date":"日付","Default":"デフォルト","Delete":"削除","Delete Manifest":"マニフェストの削除","Delete Version {{ version.version }}":"バージョン {{ version.version }} の削除","Delete {{ contentView.name }}":"{{ contentView.name }} の削除","Deleted from %s":"%s から削除","Deleting Manifest...":"マニフェストの削除中...","Deleting from %count environments.":"%count 環境から削除中。","Deleting from 1 environment.":"1 環境から削除中。","Deletion from %s":"%s からの削除","Description":"説明","Detail":"詳細","Details":"詳細","Details for Activation Key:":"アクティベーションキーの詳細:","Details for Content View:":"コンテンツビューの詳細:","Details for Product:":"製品の詳細:","Discover":"検出","Discovered URLs":"検出された URL","Docker Content":"Docker コンテンツ","Docker Images":"Docker イメージ","Docker Images in {{ repository.name }}":"{{ repository.name }} の Docker イメージ","Docker Repositories":"Docker リポジトリー","Docker Repositories <div>{{ library.counts.docker_repositories || 0 }}</div>":"Docker リポジトリー: <div>{{ library.counts.docker_repositories || 0 }}</div>","Docker Tag:":"Docker タグ:","Docker Tags":"Docker タグ","Docker metadata generation has been initiated in the background.  Click\n        <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Docker メタデータの生成がバックグラウンドで開始されました。\n        <a ng-href=\"{{ taskUrl() }}\">こちら</a> をクリックして進捗をモニターしてください。","Edit":"編集","Enabled?":"有効にされていますか?","End Date":"終了日","Ends":"終了","Enhancement":"機能強化","Enter Package Group Name(s)...":"パッケージグループ名の入力...","Enter Package Name(s)...":"パッケージ名の入力...","Environment":"環境","Environment saved":"環境が保存されました","Environment: {{ environment.name }}":"環境: {{ environment.name }}","Environments":"環境","Environments List":"環境一覧","Environments:":"環境:","Equal To":"等しい","Errata":"エラータ","Errata <div>{{ library.counts.errata.total || 0 }}</div>":"エラータ: <div>{{ library.counts.errata.total || 0 }}</div>","Errata Details":"エラータの詳細","Errata Filter:":"エラータのフィルター:","Errata ID":"エラータ ID","Errata Installation":"エラータのインストール","Errata List":"エラータ一覧","Errata Task List":"エラータのタスク一覧","Errata Tasks":"エラータのタスク","Errata Type":"エラータのタイプ","Errata for:":"エラータの対象:","Errata successfully added.":"エラータが正常に追加されました。","Errata successfully removed.":"エラータが正常に削除されました。","Errata:":"エラータ:","Erratum - by Date and Type":"エラータ - 日付およびタイプ別","Erratum - by ID":"エラータ - ID 別","Erratum Date Range":"エラータの日付範囲","Error deleting manifest.":"マニフェストの削除時のエラー。","Error during upload:":"アップロード時のエラー:","Error importing manifest.":"マニフェストのインポート時のエラー。","Error refreshing manifest.":"マニフェストの更新時のエラー。","Event":"イベント","Events":"イベント","Exclude":"除外","Existing Product":"既存の製品","Expires":"期限切れ","Failed":"失敗","File too large. Please use the CLI instead.":"ファイルが大きすぎます。CLI を代わりに使用してください。","Filter":"フィルター","Filter Saved":"フィルターが保存されました","Filter information for:":"次の情報のフィルター:","Filter...":"フィルター...","Filter:":"フィルター:","Filters":"フィルター","Filters List":"フィルター一覧","Filters for Content View:":"コンテンツビューのフィルター:","Filters successfully removed.":"フィルターが正常に削除されました。","Finished At":"終了: ","For older operating systems such as Red Hat Enterprise Linux 5 or CentOS 5 it is recommended to use sha1.":"Red Hat Enterprise Linux 5 または CentOS 5 などの古いオペレーティングシステムの場合は、sha1 を使用することを推奨します。","GPG Key":"GPG キー","GPG Key URL":"GPG キー URL","GPG Key successfully uploaded":"GPG キーが正常にアップロードされました","GPG Key {{ gpgKey.name }}":"GPG キー {{ gpgKey.name }}","GPG Keys":"GPG キー","General Information":"一般情報","Generated On":"生成日","Gpg Key updated":"GPG キーが更新されました","Gpg Keys":"GPG キー","Greater Than":"次より大きい","Group Install":"グループのインストール","Group Remove":"グループの削除","Guests of":"次のゲスト:","Hide affected Activation Keys":"影響するアクティベーションキーを非表示","Hide affected Content Hosts":"影響するコンテンツホストを非表示","History":"履歴","History for Content View:":"コンテンツビューの履歴:","Host Collection Management":"ホストコレクションの管理","Host Collection Membership":"ホストコレクションのメンバーシップ","Host Collection Membership Management":"ホストコレクションのメンバーシップ管理","Host Collection removed.":"ホストコレクションが削除されました。","Host Collection updated":"ホストコレクションが更新されました","Host Collection:":"ホストコレクション:","Host Collection: {{ hostCollection.name }}":"ホストコレクション: {{ hostCollection.name }}","Host Collections":"ホストコレクション","Host Collections for:":"次のホストコレクション:","Host Count":"ホスト数","Host Group":"ホストグループ","Hostname":"ホスト名","Hourly at {{ product.sync_plan.sync_date | date:'m' }} minutes and {{ product.sync_plan.sync_date | date:'s' }} seconds":"毎時: {{ product.sync_plan.sync_date | date:'m' }} 分 {{ product.sync_plan.sync_date | date:'s' }} 秒","IPv4 Address":"IPv4 アドレス","IPv6 Address":"IPv6 アドレス","Id":"ID","If you would prefer to move portions of these Activation Keys to different Content Views or Lifecycle Environments click <a href=\"{{ activationKeyLink() }}\">here</a> to manage them individually.":"これらのアクティベーションキーの一部を別のコンテンツビューまたはライフサイクル環境に移動させる場合は、<a href=\"{{ activationKeyLink() }}\">こちら</a> をクリックして、それらを別々に管理します。","If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click":"これらのアクティベーションキーの一部を別のコンテンツビューまたはライフサイクル環境に移動させる場合は、次をクリックします。","If you would prefer to move portions of these Content Hosts to different content views or environments click <a href=\"{{ contentHostsLink() }}\">here</a>  to manage these Content Hosts in bulk.":"これらのコンテンツホストの一部を別のコンテンツビューまたは環境に移動させる場合は、<a href=\"{{ contentHostsLink() }}\">こちら</a>  をクリックして、これらのコンテンツホストを一括管理します。","Image":"イメージ","Image ID":"イメージ ID","Import History":"インポートの履歴","Important":"重要","In order to add a content view to a composite view you must first publish an initial version of the content view.":"コンテンツビューを複合ビューに追加するには、まずコンテンツビューの初期バージョンを公開する必要があります。","Include":"組み込み","Included in Activation Keys":"アクティベーションキーに組み込み済み","Incremental Update":"増分更新","Install":"インストール","Install Selected":"選択をインストール","Install katello-agent for remote actions and displaying errata information:":"リモートアクション用の katello-agent のインストールとエラータ情報の表示:","Install the pre-built bootstrap RPM:":"事前作成のブートストラップ RPM のインストール:","Installable":"インストール可能","Installable Errata":"インストールできるエラータ","Installed":"インストール済み","Installed Package":"インストール済みパッケージ","Installed Packages":"インストール済みパッケージ","Installed Products":"インストール済み製品","Instance-based":"インスタンスベース","Interfaces":"インターフェース","Interval":"間隔","Issued":"発行","Katello Agent":"Katello エージェント","Label":"ラベル","Last Checkin":"最終チェックイン","Last Published":"最終公開","Last Puppet Report":"最終 Puppet レポート","Last Sync":"最終同期","Last Updated":"最終更新","Last Updated On":"最終更新日時","Last synced {{ product.last_sync_words }} ago.":"最終同期: {{ product.last_sync_words }} 前。","Latest (Currently %s)":"最新 (現在 %s)","Less Than":"次より小さい","Library":"ライブラリー","Library Synced Content":"ライブラリーと同期したコンテンツ","Lifecycle Environment \"{{ priorEnvironment.name }}\" exists within the middle of a path. New Lifecycle Environments can only be added to the end of a path.":"ライフサイクル環境 \"{{ priorEnvironment.name }}\" は、パスの途中にあります。新規ライフサイクル環境は、パスの終わりにのみ追加することができます。","Lifecycle Environment Paths":"ライフサイクル環境パス","Lifecycle Environments":"ライフサイクル環境","Limit":"制限","Limit:":"制限:","Limits":"制限","List/Remove":"一覧/削除","Loading Environment Paths...":"環境パスをロードしています...","Loading...":"ロード中...","Manage Docker Images":"Docker イメージの管理","Manage Docker Images for Repository:":"リポジトリーの Docker イメージを管理:","Manage Manifest":"マニフェストの管理","Manage Packages":"パッケージの管理","Manage Packages for Repository:":"リポジトリーのパッケージを管理:","Manage Puppet Modules":"Puppet モジュールの管理","Manage Puppet Modules for Repository:":"リポジトリーの Puppet モジュールを管理:","Manifest History":"マニフェストの履歴","Manifest successfully deleted.":"マニフェストが正常に削除されました。","Manifest successfully imported.":"マニフェストが正常にインポートされました。","Manifest successfully refreshed.":"マニフェストが正常に更新されました。","Message":"メッセージ","Model":"モデル","Moderate":"モデレーション","Module %s removed from Content View.":"モジュール %s がコンテンツビューから削除されました。","Module List":"モジュール一覧","More Details":"詳細","More...":"詳細...","Multi-entitlement":"複数エンタイトルメント","N/A":"該当なし","Name":"名前","Name of the upstream repository you want to sync. Example: 'busybox' or 'fedora/ssh'.":"同期するアップストリームリポジトリーの名前です。例: 'busybox' または 'fedora/ssh'。","Networking":"ネットワーク","Never synced":"同期しない","New Activation Key":"新規アクティベーションキー","New Content View":"新規コンテンツビュー","New Environment":"新規環境","New Environment Path":"新規環境パス","New Filter":"新規フィルター","New Filter for Content View:":"コンテンツビューの新規フィルター:","New GPG Key":"新規 GPG キー","New Gpg Key":"新規 GPG キー","New Host Collection":"新規ホストコレクション","New Name:":"新規の名前:","New Product":"新製品","New Repository":"新規リポジトリー","New Sync Plan":"新規同期プラン","New sync plan successfully created.":"新規同期プランが正常に作成されました。","Next":"次へ","Next Sync":"次の同期","No":"No","No \"Edit\" permissions on some of the Activation Keys in that Lifecycle Environment.":"そのライフサイクル環境にある一部のアクティベーションキーには「編集」パーミッションがありません。","No \"Edit\" permissions on some of the Content Hosts in that Lifecycle Environment.":"そのライフサイクル環境にある一部のコンテンツホストには「編集」パーミッションがありません。","No \"Promote/Remove\" permissions on the Lifecycle Environment.":"ライフサイクル環境には、プロモート/削除パーミッションがありません。","No (Default)":"No (デフォルト)","No Content Hosts are affected.":"影響するコンテンツホストはありません。","No Content Hosts match this Erratum.":"このエラータに一致するコンテンツホストはありません。","No Content Views available, please select another environment.":"利用可能なコンテンツビューはありません。別の環境を選択してください。","No Host Collections to show, you can add Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"表示するホストコレクションがありません。メインメニューの「ホスト」の下にある「ホストコレクション」を選択してホストコレクションを追加できます。","No Host Collections to show, you can add Host Collections after selecting the 'Add' tab.":"表示するホストコレクションがありません。「追加」タブを選択してホストコレクションを追加できます。","No Products to show.":"表示する製品がありません。","No Repositories contain this Erratum.":"このエラータに含まれるリポジトリーはありません。","No activation keys are affected.":"影響するアクティベーションキーはありません。","No alternate release version choices are available. The available releases are based upon what is available in \"{{ contentHost.content_view.name }}\", the selected <a href=\"/content_views\">content view</a> this content host is attached to for the given <a href=\"/lifecycle_environments\">lifecycle environment</a>, \"{{ contentHost.environment.name }}\".":"選択可能な別のリリースバージョンはありません。利用可能なリリースは、\"{{ contentHost.content_view.name }}\" で利用可能なものに基づきます。\"{{ contentHost.content_view.name }}\" は、このコンテンツホストが所定の  <a href=\"/lifecycle_environments\">ライフサイクル環境</a>、\"{{ contentHost.environment.name }}\" に対して割り当てられる選択済みの <a href=\"/content_views\">コンテンツビュー</a> です。","No content views exist for {{selected.environment.name}}":"{{selected.environment.name}} についてのコンテンツビューは存在しません","No host collections":"ホストコレクションがありません","No provisioning details are available for this content host.":"このコンテンツホストについて利用できるプロビジョニングの詳細はありません。","No repository content provided through subscriptions.":"サブスクリプションで提供されるリポジトリーコンテンツはありません。","No repository content provided through this subscription.":"このサブスクリプションで提供されるリポジトリーコンテンツはありません。","No repository content.":"リポジトリーコンテンツはありません。","No subscription manifest imported":"サブスクリプションマニフェストはインポートされていません","None":"なし","Not Synced":"同期されていません","Not installed":"インストールされていません","Not yet published":"公開されていません","Number of CPUs":"CPU 数","OS":"OS","Only show Errata that is Applicable to one or more Content Hosts":"1 つ以上のコンテンツホストに適用可能なエラータのみを表示","Only show Errata that is Installable on one or more Content Hosts":"1 つ以上のコンテンツホストにインストール可能なエラータのみを表示","Operating System":"オペレーティングシステム","Options:":"オプション:","Organization":"組織","Original Sync Date":"最初の同期日","Override to No":"「No」に上書き","Override to Yes":"「Yes」に上書き","Package":"パッケージ","Package Actions":"パッケージのアクション","Package Filter:":"パッケージのフィルター:","Package Group":"パッケージグループ","Package Group Filter:":"パッケージグループフィルター:","Package Group successfully added.":"パッケージグループが正常に追加されました。","Package Group successfully removed.":"パッケージグループが正常に削除されました。","Package Groups":"パッケージグループ","Package Install":"パッケージのインストール","Package Installation, Removal, and Update":"パッケージのインストール、削除、および更新","Package Remove":"パッケージの削除","Package Update":"パッケージの更新","Package removal failed":"パッケージの削除が失敗しました","Package successfully added.":"パッケージが正常に追加されました。","Package successfully removed.":"パッケージが正常に削除されました。","Package successfully updated.":"パッケージが正常に更新されました。","Package/Group Name":"パッケージ/グループ名","Packages":"パッケージ","Packages <div>{{ library.counts.packages || 0 }}</div>":"パッケージ: <div>{{ library.counts.packages || 0 }}</div>","Packages for:":"次のパッケージ:","Packages in {{ repository.name }}":"{{ repository.name }} のパッケージ","Parameters":"パラメーター","Paste GPG Key Contents":"GPG キーコンテンツの貼り付け","Paste contents...":"コンテンツの貼り付け...","Paused":"一時停止中","Perform":"実行","Physical":"物理","Please select a Lifecycle Environment and Content View to move the affected Activation Keys to:":"影響するアクティベーションキーの移動先となるライフサイクル環境およびコンテンツビューを選択してください。","Please select a Lifecycle Environment and Content View to move these Content Hosts to:":"これらのコンテンツホストの移動先となるライフサイクル環境およびコンテンツビューを選択してください。","Please select a Lifecycle Environment.":"ライフサイクル環境を選択してください。","Please select an environment.":"環境を選択してください。","Please select one from the list below and you will be redirected.":"以下の一覧から 1 つ選択してください。自動的に移動します。","Press 'Sync Now' below to begin syncing the following product(s).":"以下の「同期開始」を押して、次の製品の同期を開始します。","Previous Environment (%e/%cv)":"直前の環境 (%e/%cv)","Processing":"処理中","Product":"製品","Product Bulk Actions":"製品の一括処理","Product Content":"製品コンテンツ","Product Count":"製品数","Product Enhancement Advisory":"製品の機能強化アドバイザリー","Product ID":"製品 ID","Product Management":"製品管理","Product Management for Sync Plan:":"同期プランのための製品管理:","Product Name":"製品名","Product Saved":"製品を保存","Product Sync":"製品の同期","Product delete operation has been initiated in the background.  Click <a href=\"{{ taskUrl(productDeletionTaskId) }}\">Here</a> to monitor the progress.":"製品の削除操作がバックグラウンドで開始されました。<a href=\"{{ taskUrl(productDeletionTaskId) }}\">こちら</a> をクリックして、進捗をモニターしてください。","Product information for:":"製品情報:","Product {{ product.name }}":"製品 {{ product.name }}","Products":"製品","Products <div>{{ library.counts.products || 0 }}</div>":"製品: <div>{{ library.counts.products || 0 }}</div>","Promote":"プロモート","Promote Content View:":"コンテンツビューのプロモート:","Promote Version":"バージョンのプロモート","Promote Version {{ version.version }}":"バージョン {{ version.version }} のプロモート","Promoted to %s":"%s にプロモート","Promoting to %count environments.":"%count 環境にプロモートしています。","Promoting to 1 environment.":"1 環境にプロモートしています。","Promotion History":"プロモーション履歴","Provided Products":"提供製品","Provisioning Details":"プロビジョニングの詳細","Provisioning Host Details":"ホストのプロビジョニングの詳細","Publish Content View:":"コンテンツビューの公開:","Publish New Version":"新規バージョンの公開","Publish via HTTP":"HTTP での公開","Publish via HTTPS":"HTTPS での公開","Published":"公開","Published At":"公開:","Published Repository Information":"リポジトリー情報の公開","Published new version":"新規バージョンの公開","Publishing and promoting to %count environments.":"%count 環境に公開およびプロモート中です。","Publishing and promoting to 1 environment.":"1 環境に公開およびプロモート中です。","Puppet Environment":"Puppet 環境","Puppet Modules":"Puppet モジュール","Puppet Modules <div>{{ library.counts.puppet_modules || 0 }}</div>":"Puppet モジュール: <div>{{ library.counts.puppet_modules || 0 }}</div>","Puppet Modules for Content View:":"コンテンツビューの Puppet モジュール:","Puppet Modules in {{ repository.name }}":"{{ repository.name }} の Puppet モジュール","Puppet module added to Content View":"コンテンツビューに追加された Puppet モジュール:","Puppet module metadata generation has been initiated in the background.\n        Click <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Puppet モジュールのメタデータ生成がバックグラウンドで開始されました。\n   <a ng-href=\"{{ taskUrl() }}\">こちら</a> をクリックして進捗をモニターしてください。","Quantity":"数量","RAM (GB)":"RAM (GB)","RAM: %s GB":"RAM: %s GB","Range":"範囲","Reboot Suggested":"再起動が推奨されています","Reboot Suggested?":"再起動が推奨されていますか?","Red Hat CDN URL":"Red Hat CDN URL","Red Hat Provider Details":"Red Hat プロバイダーの詳細","Red Hat Repositories page":"Red Hat リポジトリーのページ","Red Hat Repositories page.":"Red Hat リポジトリーのページ。","Refresh Manifest":"マニフェストの更新","Refresh Table":"テーブルの更新","Refreshing Manifest...":"マニフェストの更新中...","Refreshing manifest not suppported by this manifest.":"このマニフェストでサポートされていないマニフェストを更新中。","Register Content Host":"コンテンツホストの登録","Register a Content Host":"コンテンツホストの登録","Registered":"登録済み","Registered By":"登録:","Registry URL":"レジストリー URL","Release":"リリース","Release Version":"リリースバージョン","Remove":"削除","Remove Activation Key \"{{ activationKey.name }}\"?":"アクティベーションキー \"{{ activationKey.name }}\" を削除しますか?","Remove Content View":"コンテンツビューの削除","Remove Content Views":"コンテンツビューの削除","Remove Docker Images":"Docker イメージの削除","Remove Errata":"エラータの削除","Remove Filters":"フィルターの削除","Remove From":"削除元","Remove GPG Key":"GPG キーを削除","Remove GPG Key {{ gpgKey.name }}":"GPG キー {{ gpgKey.name }} の削除","Remove Host Collection \"{{ hostCollection.name }}\"?":"ホストコレクション \"{{ hostCollection.name }}\" を削除しますか?","Remove Module":"モジュールの削除","Remove Package Group":"パッケージグループの削除","Remove Packages":"パッケージの削除","Remove Product":"製品の削除","Remove Product \"{{ product.name }}\"?":"製品 \"{{ product.name }}\" を削除しますか?","Remove Products":"製品の削除","Remove Puppet Modules":"Puppet モジュールの削除","Remove Repositories":"リポジトリーの削除","Remove Repository":"リポジトリーの削除","Remove Repository \"{{ repository.name }}\"?":"リポジトリー \"{{ repository.name }}\" を削除しますか?","Remove Selected":"選択を削除","Remove Successful.":"削除が成功しました。","Remove Sync Plan":"同期プランの削除","Remove Sync Plan \"{{ syncPlan.name }}\"?":"同期プラン \"{{ syncPlan.name }}\" を削除しますか?","Remove Version":"バージョンの削除","Remove Version Confirmation":"バージョンの削除を確認","Remove View":"ビューの削除","Remove {{ productTable.getSelected().length }} Products?":"{{ productTable.getSelected().length }} 製品を削除しますか?","Removed %x host collections from activation key \"%y\".":" %x ホストコレクションがアクティベーションキー \"%y\" から削除されました。","Removed %x host collections from content host \"%y\".":"%x ホストコレクションがコンテンツホスト \"%y\" から削除されました。","Removed %x products from sync plan \"%y\".":"%x 製品が同期プラン \"%y\" から削除されました。","Removing Content Hosts":"コンテンツホストの削除中","Removing Products":"製品の削除中","Removing Repositories":"リポジトリーの削除中","Repo Discovery":"リポジトリーの検出","Repo Type":"リポジトリーのタイプ","Repositories":"リポジトリー","Repositories containing Errata {{ errata.errata_id }}":"エラータ {{ errata.errata_id }} を含むリポジトリー","Repositories for Content View:":"コンテンツビューのリポジトリー:","Repositories for Errata:":"エラータのリポジトリー:","Repositories for Filter:":"フィルターのリポジトリー:","Repositories for Product:":"製品のリポジトリー:","Repository":"リポジトリー","Repository \"%s\" successfully deleted":"リポジトリー \"%s\" が正常に削除されました","Repository %s successfully created.":"リポジトリー %s が正常に作成されました。","Repository Count":"リポジトリー数","Repository Discovery":"リポジトリーの検出","Repository Name":"リポジトリー名","Repository Saved.":"リポジトリーが保存されました。","Repository Selection":"リポジトリーの選択","Repository URL updated":"リポジトリー URL が更新されました","Repository:":"リポジトリー:","Required Host":"必要なホスト","Result":"結果","Results":"結果","Rules Source":"ルールソース","Rules Version":"ルールバージョン","Run Auto-Attach":"自動割り当ての実行","Run subscription-manager in a console on the client host.  You may use an Activation Key to register:":"クライアントホストのコンソールで subscription-manager を実行します。アクティベーションキーを使って登録できます。","Save Successful.":"保存が成功しました。","Search...":"検索...","Security":"セキュリティー","Security Advisory":"セキュリティーアドバイザリー","Select A New Puppet Module To Add":"追加する新規 Puppet モジュールの選択","Select Content Hosts":"コンテンツホストの選択","Select Module":"モジュールの選択","Select Version":"バージョンの選択","Select a Content Source:":"コンテンツソースの選択:","Select a Version":"バージョンの選択","Select an Available Version of {{ $stateParams.moduleName }}":"{{ $stateParams.moduleName }} の利用可能なバージョンを選択","Select an Organization":"組織の選択","Select new version":"新規バージョンの選択","Select the Lifecycle Environments you would like to remove Version {{ version.version }} from:":"バージョン {{ version.version }} を削除するライフサイクル環境を選択:","Selected URLs:":"選択された URL:","Selecting this option will result in the Version being completely deleted and no longer being available for promotion.  The version must be removed from all Lifecycle Environments in order to select this option.":"このオプションを選択すると、該当バージョンが完全に削除され、プロモーション対象として使用不可になります。このオプションを選択するには、バージョンをすべてのライフサイクル環境から取り除く必要があります。","Serve via HTTP.":"HTTP 経由で提供。","Service Level":"サービスレベル","Severity":"重大度","Show affected Activation Keys":"影響するアクティベーションキーの表示","Show affected Content Hosts":"影響するコンテンツホストの表示","Show from:":"次から表示:","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"{{ detailsTable.resource.subtotal }} (合計 {{ detailsTable.resource.total }}) コンテンツホストの {{ detailsTable.rows.length }} を表示中","Showing {{ addSubscriptionsTable.rows.length }} of {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} Total) Subscriptions":" {{ addSubscriptionsTable.resource.subtotal }} (合計 {{ addSubscriptionsTable.resource.total }}) サブスクリプションの {{ addSubscriptionsTable.rows.length }} を表示中","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"{{ detailsTable.resource.subtotal }} (合計 {{ detailsTable.resource.total }}) コンテンツホストの {{ detailsTable.rows.length }} を表示中","Showing {{ currentPackagesTable.resource.total }} Packages":"{{ currentPackagesTable.resource.total }} パッケージを表示中","Showing {{ subscriptionsTable.rows.length }} of {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} Total) Subscriptions":"{{ subscriptionsTable.resource.subtotal }} (合計 {{ subscriptionsTable.resource.total }}) サブスクリプションの {{ subscriptionsTable.rows.length }} を表示中","Sockets":"ソケット","Sockets: %s":"ソケット: %s","Solution":"ソリューション","Some of the Errata shown below may not be installable as they are not in this Content Host's\n          Content View and Lifecycle Environment.  In order to apply such Errata an Incremental Update is required.":"以下に表示されているエラータの一部は、このコンテンツホストの\nコンテンツビューとライフサイクル環境にないためにインストールできない場合があります。このエラータを適用するには、増分更新が必要です。","Stacking ID":"スタック ID","Start Date":"開始日","Start Time":"開始時刻","Started At":"開始時刻","Starting":"開始","Starts":"開始","State":"状態","Status":"状態","Subscription Details":"サブスクリプションの詳細","Subscription Info":"サブスクリプション情報","Subscription Manifest":"サブスクリプションマニフェスト","Subscription Server Details":"サブスクリプションサーバーの詳細","Subscription Status":"サブスクリプションの状態","Subscriptions":"サブスクリプション","Subscriptions for Activation Key:":"アクティベーションキーのサブスクリプション:","Subscriptions for:":"次のサブスクリプション:","Succesfully scheduled package installation":"パッケージのインストールが正常にスケジュールされました","Succesfully scheduled package removal":"パッケージの削除が正常にスケジュールされました","Succesfully scheduled package update":"パッケージの更新が正常にスケジュールされました","Success":"成功","Successfully Scheduled Auto-attach.":"自動割り当てが正常にスケジュールされました。","Successfully added %s subscriptions.":"%s サブスクリプションが正常に追加されました","Successfully deleted %cv version %ver.":"%cv バージョン %ver が正常に削除されました。","Successfully initiated deletion of %cv version %ver.":"%cv バージョン %ver の削除が正常に開始されました。","Successfully initiated promotion of %cv version %ver to %env.":"%cv バージョン %ver の %env へのプロモーションが正常に開始されました。","Successfully initiated removal of %cv version %ver.":"%cv バージョン %ver の削除が正常に開始されました。","Successfully promoted %cv version %ver to %env":"%cv バージョン %ver が %env に正常にプロモートされました ","Successfully published %cv version %ver and promoted to Library":"%cv バージョン %ver が正常に公開され、ライブラリーにプロモートされました","Successfully removed %cv version %ver from environments: %env":"%cv バージョン %ver が以下の環境から正常に削除されました: %env","Successfully removed %s items.":"%s 項目が正常に削除されました。","Successfully removed %s subscriptions.":"%s サブスクリプションが正常に削除されました。","Successfully removed 1 item.":"1 項目が正常に削除されました。","Successfully updated subscriptions.":"サブスクリプションが正常に更新されました。","Summary":"要約","Support Level":"サポートレベル","Support Type":"サポートタイプ","Sync Enabled":"同期が有効にされました","Sync Information":"同期情報","Sync Interval":"同期間隔","Sync Now":"同期開始","Sync Plan":"同期プラン","Sync Plan %s has been deleted.":"同期プラン %s が削除されました","Sync Plan Saved":"同期プランが保存されました","Sync Plan {{ syncPlan.name }}":"同期プラン {{ syncPlan.name }}","Sync Plan:":"同期プラン:","Sync Plans":"同期プラン","Sync State":"同期の状態","Sync Status":"同期の状態","Synced manually, no interval set.":"手動で同期されました。間隔が設定されていません。","Syncing...":"同期中...","Tags":"タグ","Target":"ターゲット","Task Details":"タスクの詳細","Tasks":"タスク","Tasks for Content View:":"コンテンツビューのタスク:","Temporary":"一時","The Activation Keys listed below are currently using this Content View Version.  Before deleting the Version you must move these Activation Keys to a Lifecycle Environment where this Version is not in use.":"以下に一覧表示されているアクティベーションキーは、現在このコンテンツビューバージョンを使用しています。バージョンを削除する前に、これらのアクティベーションキーを、このバージョンが使用されていないライフサイクル環境に移動する必要があります。","The Content Hosts listed below are currently using this Content View Version.  Before removing the version you must move these Content Hosts to an environment where this version is not in use.":"以下に一覧表示されているコンテンツホストは、現在このコンテンツビューバージョンを使用しています。バージョンを削除する前に、これらのコンテンツホストを、このバージョンが使用されていない環境に移動する必要があります。","The Content View or Lifecycle Environment needs to be updated in order to make errata available to these hosts.":"これらのホストに対してエラータを利用可能にするには、コンテンツビューまたはライフサイクル環境を更新する必要があります。","The archive of this Version will be deleted.  This Version will not be available once deletion is complete.":"このバージョンのアーカイブは削除されます。このバージョンは削除が完了すると利用できなくなります。","The archive of this Version will not be deleted.  This Version will still be available to be promoted to a Lifecycle Environment.":"このバージョンのアーカイブは削除されません。このバージョンは依然として、ライフサイクル環境にプロモートするために使用できます。","The following actions can be performed on content hosts in this host collection:":"以下のアクションは、このホストコレクション内のコンテンツホストで実行できます:","The host requires being attached to a content view and the environment you have chosen has no content views promoted to it.\n            See the <a href=\"/content_views\">content views page</a> to manage and promote a content view.":"ホストをコンテンツビューに割り当てる必要があります。選択した環境には、コンテンツビューがプロモートされていません。\n       コンテンツビューを管理し、プロモートするには、<a href=\"/content_views\">コンテンツビューのページ</a> を参照してください。","The katello-agent package is required to manage errata on this Host.\n      Please install the katello-agent package to manage errata on\n      {{ contentHost.name }}.":"このホストでエラータを管理するには katello-agent パッケージが必要です。\n    {{ contentHost.name }} でエラータを管理するには、katello-agent パッケージを\nインストールしてください。","The katello-agent package is required to manage packages on this Host.\n      Please install the katello-agent package to manage packages on\n      {{ contentHost.name }}. To install the package, connect directly to the\n      host from a terminal and use yum.":"このホストでパッケージを管理するには、katello-agent パッケージが必要です。\n       {{ contentHost.name }} でパッケージを管理するには katello-agent パッケー\nジをインストールしてください。パッケージをインストールするには、端末からホストに\n直接接続して yum を使用します。","The page you are attempting to access requires selecting a specific organization.":"アクセスしようとしているページでは、特定の組織を選択する必要があります。","The selected environment contains no Content Views, please select a different environment.":"選択された環境にはコンテンツビューがありません。別の環境を選択してください。","There are currently no repositories associated with this Content View, you can add some by clicking on the \"Add\" tab above.":"現在、このコンテンツビューに関連付けられたリポジトリーはありません。上部の「追加」タブをクリックすると、いくつか追加することができます。","There are currently no repositories to add to this Content View,":"現在、このコンテンツビューに追加するリポジトリーはありません。","There are no Content Views in this Environment.":"この環境にはコンテンツビューがありません。","There are no Content Views that match the criteria.":"条件に一致するコンテンツビューがありません。","There are no Docker Tags that match the criteria.":"条件に一致する Docker タグがありません。","There are no Errata associated with this Content Host to display.":"このコンテンツホストに関連した、表示できるエラータがありません。","There are no Errata in this organization.  Create one or more Products with Errata to view Errata on this page.":"この組織にはエラータがありません。このページにエラータを表示するには、エラータを含む 1 つ以上の製品を作成します。","There are no Errata that match the criteria.":"条件に一致するエラータがありません。","There are no Errata to display.":"表示するエラータがありません。","There are no Host Collections available. You can create new Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"利用可能なホストコレクションがありません。メインメニューの「ホスト」の下にある「ホストコレクション」を選択して新規ホストコレクションを作成できます。","There are no Packages that match the criteria.":"条件に一致するパッケージがありません。","There are no Puppet Modules that match the criteria.":"条件に一致する Puppet モジュールがありません。","There are no Yum Repositories that match the criteria.":"条件に一致する Yum リポジトリーがありません。","There is currently an Incremental Update task in progress.  This update must finish before applying existing updates.":"現在、増分更新が進行中です。既存の更新を適用する前に、この更新を終了する必要があります。","There were errors while removing the following Repositories:":"以下のリポジトリーを削除している間にエラーが発生しました:","This Content Host is not attached to any subscriptions that provide content":"このコンテンツホストは、コンテンツを提供するいずれのサブスクリプションにも割り当てられていません。","This Content View does not have any versions, create your first Content View Version by using the \"Publish New Version\" button on the right.":"このコンテンツビューにはバージョンがありません。右側の「新規バージョンの公開」ボタンを使用して最初のコンテンツビューバージョンを作成します。","This Version cannot be deleted from some of the Lifecycle Environments due to one of the reasons below.":"このバージョンは、以下のいずれかの理由により、一部のライフサイクル環境から削除できません。","This Version is not associated with any Lifecycle Environments.":"このバージョンは、いずれのライフサイクル環境にも関連付けられていません。","This activation key is not associated with any content hosts.":"このアクティベーションキーは、いずれのコンテンツホストにも関連付けられていません。","This activation key may be used during system registration. For example:":"このアクティベーションキーは、システムの登録時に使用できます。例:","This filter applies only to a subset of repositories in the content view.":"このフィルターは、コンテンツビューのリポジトリーのサブセットにのみ適用されます。","This filter applies to all repositories in the content view (current and future).":"このフィルターは、コンテンツビューのすべてのリポジトリーに適用されます (現在および将来)。","This subscription is not associated with any activation keys.":"このサブスクリプションは、いずれのアクティベーションキーにも関連付けられていません。","This subscription is not attached to any content hosts.":"このサブスクリプションは、いずれのコンテンツホストにも割り当てられていません。","This version will be removed from:":"このバージョンは次から削除されます:","Time":"時間","Title":"タイトル","To register a content host to this server, follow these steps.":"コンテンツホストをこのサーバーに登録するには、以下のステップを実行します。","Topic":"トピック","Type":"タイプ","URL":"URL","URL of the registry you want to sync. Example: https://registry.hub.docker.com":"同期するレジストリーの URL です。例: https://registry.hub.docker.com","URL to discover":"検出する URL","UUID":"UUID","Unattach Sync Plan":"同期プランの割り当て解除","Unattaching...":"割り当て解除中...","Unlimited Content Hosts:":"無制限のコンテンツホスト:","Unregister Content Host":"コンテンツホストの登録解除","Unregister Content Host \"{{contentHost.name}}\"?":"コンテンツホスト \"{{contentHost.name}}\" の登録を解除しますか?","Unregister Content Hosts":"コンテンツホストの登録解除","Unregister {{ table.numSelected  }} Content Hosts?":"{{ table.numSelected  }} コンテンツホストの登録を解除しますか?","Update":"更新","Update All":"すべてを更新","Update Repositories":"リポジトリーの更新","Update Sync Plan":"同期プランの更新","Updated":"更新済み","Updated errata filter -":"エラータフィルターの更新 -","Updated override to \"%x\".":"\"%x\" への上書きを更新しました。","Updating...":"更新中...","Upload":"アップロード","Upload Docker Image":"Docker イメージのアップロード","Upload GPG Key":"GPG キーのアップロード","Upload New Manifest":"新規マニフェストをアップロード","Upload Package":"パッケージのアップロード","Upload Puppet Module":"Puppet モジュールのアップロード","Uploading...":"アップロード中...","Upstream Repository Name":"アップストリームリポジトリー名","Upstream Subscription":"アップストリームのサブスクリプション","Upstream Subscription Management Application":"アップストリームのサブスクリプション管理アプリケーション","Upstream Type":"アップストリームのタイプ","Upstream User":"アップストリームユーザー","Use Latest (currently %s)":"最新を使用 (現在 %s)","Use the cancel button on content view selection to revert your environment selection.":"コンテンツビュー選択の「取り消し」ボタンを使って、環境の選択を元に戻します。","User":"ユーザー","Version":"バージョン","Version %s":"バージョン %s","Version 1":"バージョン 1","Version Deletion":"バージョンの削除","Version Details":"バージョンの詳細","Version {{ version.version }}":"バージョン {{ version.version }}","Version:":"バージョン: ","Versions":"バージョン","Versions for Content View:":"コンテンツビューのバージョン:","Versions for Puppet Module:":"Puppet モジュールのバージョン:","View Details":"ビューの詳細","Virtual":"仮想","Virtual Guest From":"仮想ゲスト","Virtual Guests":"仮想ゲスト","Virtual Host":"仮想ホスト","Weekly on {{ product.sync_plan.sync_date | date:'EEEE' }} at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"毎週: {{ product.sync_plan.sync_date | date:'EEEE' }}、{{ product.sync_plan.sync_date | date:'mediumTime' }} (サーバー時間)","When Auto Attach is disabled, registering systems will be attached to all associated subscriptions.":"自動割り当てが無効にされている場合、システムの登録により、すべての関連付けられたサブスクリプションに割り当てられます。","Working":"稼働中","Yes":"Yes","Yes (Default)":"Yes (デフォルト)","You cannot remove these repositories because they belong to a Red Hat product.":"これらのリポジトリーは Red Hat 製品に属するので、削除することはできません。","You cannot remove these repositories because you do not have permission.":"パーミッションがないので、これらのリポジトリーを削除することはできません。","You cannot remove this product because it is a Red Hat product.":"この製品は Red Hat 製品のため、削除することはできません。","You cannot remove this product because it was published to a content view.":"この製品はコンテンツビューに公開されたため、削除することはできません。","You cannot remove this product because you do not have permission.":"パーミッションがないので、この製品を削除することはできません。","You cannot remove this repository because it is a Red Hat repository.":"このリポジトリーは Red Hat リポジトリーのため、削除することはできません。","You cannot remove this repository because it was published to a content view.":"このリポジトリーはコンテンツビューに公開されたため、削除することはできません。","You cannot remove this repository because you do not have permission.":"パーミッションがないので、このリポジトリーを削除することはできません。","You currently do not have any Packages installed. You can use Package Actions above this message to install new Packages.":"現在、インストールされているパッケージがありません。このメッセージの上にあるパッケージのアクションを使用すると、新規パッケージをインストールすることができます。","You currently don't have any Activation Keys, you can add Activation Keys using the button on the right.":"現在、アクティベーションキーがありません。右側のボタンを使ってアクティベーションキーを追加することができます。","You currently don't have any Content Hosts in this Host Group, you can add Content Hosts after selecting the 'Add' tab.":"現在、このホストグループにはコンテンツホストがありません。「追加」タブをクリックすると、コンテンツホストを追加することができます。","You currently don't have any Content Hosts, you can create new Content Hosts by selecting Contents Host from main menu and then clicking the button on the right.":"現在、コンテンツホストがありません。メインメニューからコンテンツホストを選択してから右側のボタンをクリックすると、新規コンテンツホストを作成することができます。","You currently don't have any Content Hosts, you can register one by clicking the button on the right and following the instructions.":"現在、コンテンツホストがありません。右側のボタンをクリックし、指示に従ってコンテンツホストを登録することができます。","You currently don't have any Content Views.  A Content View can be added by using the button on the right.":"現在、コンテンツビューがありません。右側のボタンを使ってコンテンツビューを追加することができます。","You currently don't have any Docker Tags.":"現在、Docker タグがありません。","You currently don't have any Filters included in this Content View, you can add a new Filter by using the button on the right.":"現在、このコンテンツビューにはフィルターが含まれていません。右側のボタンを使って新規フィルターを追加することができます。","You currently don't have any Gpg keys, you can add Gpg keys using the button on the right.":"現在、GPG キーがありません。右側のボタンを使って GPG キーを追加することができます。","You currently don't have any Host Collections, you can add Host Collections using the button on the right.":"現在、ホストコレクションがありません。右側のボタンを使ってホストコレクションを追加することができます。","You currently don't have any Products associated with this Gpg Key.":"現在、GPG キーに関連付けられた製品がありません。","You currently don't have any Products to subscribe to, you can add Products after selecting 'Products' under 'Content' in the main menu.":"現在、サブスクライブさせる製品がありません。メインメニューの「コンテンツ」の下にある「製品」を選択して製品を追加できます。","You currently don't have any Products<span bst-feature-flag=\"custom_products\">, you can add Products using the button on the right</span>.":"現在、製品<span bst-feature-flag=\"custom_products\">がありません。右側のボタンを使って製品を追加することができます</span>。","You currently don't have any Puppet Modules included in this Content View, you can add Puppet Modules using the button on the right.":"現在、このコンテンツビューには Puppet モジュールが含まれていません。右側のボタンを使って Puppet モジュールを追加することができます。","You currently don't have any Repositories associated with this Gpg Key.":"現在、GPG キーに関連付けられたリポジトリーはありません。","You currently don't have any Repositories included in this Product, you can add Repositories using the button on the right.":"現在、この製品にはリポジトリーが含まれていません。右側のボタンを使ってリポジトリーを追加することができます。","You currently don't have any Subscriptions associated with this Content Host, you can add Subscriptions after selecting the 'Add' tab.":"現在、このコンテンツホストに関連付けられたサブスクリプションはありません。「追加」タブをクリックすると、サブスクリプションを追加することができます。","You currently don't have any Subscriptions, you can add Subscriptions by importing a manifest using the button on the right labeled 'Manage Manifest'.":"現在、サブスクリプションがありません。右側にある「マニフェストの管理」というラベルの付いた「ボタン」を使ってマニフェストをインポートして、サブスクリプションを追加できます。","You currently don't have any Sync Plan, you can add Sync Plans using the button on the right.":"現在、同期プランがありません。右側のボタンを使って同期プランを追加することができます。","You currently don't have any Sync Plans.  A Sync Plan can be created by using the button on the right.":"現在、同期プランがありません。右側のボタンを使って同期プランを作成することができます。","You do not have any Installed Products":"インストール済みの製品がありません","You must select a content view in order to save your environment.":"環境を保存するにはコンテンツビューを選択する必要があります。","You must select a new content view before your change of environment can be saved.":"環境の変更を保存するには、まず新規コンテンツビューを選択する必要があります。","You must select a new content view before your change of environment can be saved. Use the cancel button on content view selection to revert your environment selection.":"環境の変更を保存するには、まず新規コンテンツビューを選択する必要があります。コンテンツビュー選択の「取り消し」ボタンを使って、環境の選択を元に戻します。","You must select at least one Content Host in order to apply Errata.":"エラータを適用するために 1 つ以上のコンテンツホストを選択する必要があります。","You must select at least one repository.":"1 つ以上のリポジトリーを選択する必要があります。","Your selected content hosts have changed, it is highly recommended to refresh the calculated list of errata using the 'Refresh Errata' button below.":"選択済みのコンテンツホストが変更されました。以下の「エラータの更新」ボタンを使ってエラータの計算された一覧を更新することを強く推奨します。","Yum Content":"Yum コンテンツ","Yum Metadata Checksum":"Yum メタデータのチェックサム","Yum Repo Discovery":"Yum リポジトリーの検出","Yum Repositories":"Yum リポジトリー","Yum Repositories <div>{{ library.counts.yum_repositories || 0 }}</div>":"Yum リポジトリー: <div>{{ library.counts.yum_repositories || 0 }}</div>","Yum metadata generation has been initiated in the background.  Click <a href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Yum メタデータの生成がバックグラウンドで開始されました。<a href=\"{{ taskUrl() }}\">ここ</a>  をクリックして進捗をモニターしてください。","add some repositories.":"一部のリポジトリーの追加。","daily":"毎日","here":"こちら","hourly":"毎時","or authenticate with a username and password:":"またはユーザー名とパスワードを使った認証:","subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"":"subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"","task details":"タスクの詳細","to manage them individually.":"それぞれを別々に管理する","true":"true","weekly":"毎週","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }} 件選択済み","{{ addSubscriptionsTable.numSelected }} Selected":"{{ addSubscriptionsTable.numSelected }}  件選択済み","{{ contentHost.registered_by }}":"{{ contentHost.registered_by }}","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }}  件選択済み","{{ contentView.name }} cannot be deleted as one or more Content View Versions are still promoted to a Lifecycle Environment.\n      Each Content View Version must be removed from their Lifecycle Environments before the Content View can be deleted.":"1 つ以上のコンテンツビューバージョンがライフサイクル環境にプロモートされているため、{{ contentView.name }} を削除することができません。\n  コンテンツビューを削除するには、 各コンテンツビューバージョンをライフサイクル環境から取り除いておく必要があります。","{{ contentView.puppet_modules.length || 0 }} Puppet Modules":"{{ contentView.puppet_modules.length || 0 }} Puppet モジュール","{{ contentView.repositories.length || 0 }} Repositories":"{{ contentView.repositories.length || 0 }} リポジトリー","{{ contentViewVersion.puppet_module_count || 0 }} Puppet Modules":"{{ contentViewVersion.puppet_module_count || 0 }} Puppet モジュール","{{ contentViewVersion.repositories.length || 0 }} Repositories":"{{ contentViewVersion.repositories.length || 0 }} リポジトリー","{{ discoveryTable.numSelected }} Selected":"{{ discoveryTable.numSelected }} 件選択済み","{{ environment.name }} Environment":"{{ environment.name }} 環境","{{ errata.hosts_applicable_count || 0 }} Applicable,":"{{ errata.hosts_applicable_count || 0 }} 件適用可能","{{ errata.hosts_available_count || 0 }} Installable":"{{ errata.hosts_available_count || 0 }} 件インストール可能","{{ hostCollectionsTable.numSelected }} Selected":"{{ hostCollectionsTable.numSelected }} 件選択済み","{{ package.removeTask.affected_units }} package removed":"{{ package.removeTask.affected_units }} パッケージが削除されました","{{ package.removeTask.affected_units }} packages removed":"{{ package.removeTask.affected_units }} パッケージが削除されました","{{ product.last_sync_words }} Ago ({{ product.sync_status.finish_time | date:'medium' }} Local Time)":"{{ product.last_sync_words }} 前 ({{ product.sync_status.finish_time | date:'medium' }} ローカル時間)","{{ product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ product.sync_plan.next_sync | date:'medium' }} (サーバー時間)","{{ productsTable.numSelected }} Selected":"{{ productsTable.numSelected }} 件選択済み","{{ repository.content_counts.docker_image || 0 }} Docker Images":"{{ repository.content_counts.docker_image || 0 }} Docker イメージ","{{ repository.content_counts.docker_image }} Docker Images":"{{ repository.content_counts.docker_image }} Docker イメージ","{{ repository.content_counts.docker_tag || 0 }} Docker Tags":"{{ repository.content_counts.docker_tag || 0 }} Docker タグ","{{ repository.content_counts.docker_tag }} Docker Tags":"{{ repository.content_counts.docker_tag }} Docker タグ","{{ repository.content_counts.erratum || 0 }} Errata":"{{ repository.content_counts.erratum || 0 }} エラータ","{{ repository.content_counts.erratum }} Errata":"{{ repository.content_counts.erratum }} エラータ","{{ repository.content_counts.puppet_module || 0 }} Puppet Modules":"{{ repository.content_counts.puppet_module || 0 }} Puppet モジュール","{{ repository.content_counts.rpm || 0 }} Packages":"{{ repository.content_counts.rpm || 0 }} パッケージ","{{ repository.content_counts.rpm }} Packages":"{{ repository.content_counts.rpm }} パッケージ","{{ repository.last_sync_words }} Ago ({{ repository.last_sync.ended_at | date:'medium' }} Local Time)":"{{ repository.last_sync_words }} 前 ({{ repository.last_sync.ended_at | date:'medium' }} ローカル時間)","{{ repository.last_sync_words }} ago":"{{ repository.last_sync_words }} 前","{{ repository.product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ repository.product.sync_plan.next_sync | date:'medium' }} (サーバー時間)","{{ subscription.host.name }}":"{{ subscription.host.name }}","{{ subscriptionsTable.numSelected }} Selected":"{{ subscriptionsTable.numSelected }} 件選択済み","{{ totalActivationKeyCount() }} Activation Keys will be moved to {{ deleteOptions.activationKeys.contentView.name }} in {{ deleteOptions.activationKeys.environment.name }}":"{{ totalActivationKeyCount() }} アクティベーションキーが  {{ deleteOptions.activationKeys.environment.name }} の {{ deleteOptions.activationKeys.contentView.name }} に移行します","{{ totalHostCount() }} Content Hosts will be moved to {{ deleteOptions.contentHosts.contentView.name }} in {{ deleteOptions.contentHosts.environment.name }}":"{{ totalHostCount() }} コンテンツホストが  {{ deleteOptions.contentHosts.environment.name }} の {{ deleteOptions.contentHosts.contentView.name }} に移行します","{{ type.display }}":"{{ type.display }}","{{ version.docker_image_count }} Docker Images":"{{ version.docker_image_count }} Docker イメージ","{{ version.docker_tag_count }} Docker Tags":"{{ version.docker_tag_count }} Docker タグ","{{ version.errata_counts.total }} Errata":"{{ version.errata_counts.total }} エラータ","{{ version.package_count }} Packages":"{{ version.package_count }} パッケージ","{{ version.puppet_module_count }} Puppet Modules":"{{ version.puppet_module_count }} Puppet モジュール"});
    gettextCatalog.setStrings('ko', {"%(consumed)s out of %(quantity)s":"%(quantity)s 중 %(consumed)s","(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)":"(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"작업 정보\" | translate }}</a>)","(Not all Activation Keys editable )":"(모든 활성키를 편집할 수 없습니다)","(Not all Content Hosts editable )":"(모든 컨텐츠 호스트를 편집할 수 없습니다)","* These marked Content View Versions are from Composite Content Views.  Their components needing updating are listed underneath.":"*로 표시된 컨텐츠 뷰 버전은 복합 컨텐츠 뷰 버전입니다.  업데이트해야할 구성 요소는 아래에 나열되어 있습니다.","+ New Sync Plan":"+ 새 동기화 계획 ","- Date and Type":"- 날짜 및 유형 ","-- select an interval --":"-- 간격 선택 --","1 Content Host":"{{ contentHost.virtual_guests.length }} 컨텐츠 호스트 ","1 repository sync has errors.":"{{ product.sync_summary.error}}개의 리포지터리 동기화에 오류가 있습니다.","1 repository sync in progress.":"{{ product.sync_summary.pending}}개의 리포지터리 동기화가 진행 중입니다.","1 successfully synced repository.":"{{ product.sync_summary.success}}개의 리포지터리가 성공적으로 동기화되었습니다.","<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Version {{ version.version }}</a>":"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">버전 {{ version.version }}</a>","<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Create":"<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n        생성","<span translate=\"\">\n        If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">here</a>\n      <span translate=\"\">to manage them individually.</span>":"<span translate=\"\">\n   이러한 활성키의 일부를 다른 컨텐츠 뷰 또는 라이프 사이클 환경으로 이동하고자 하는 경우\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">여기</a>를\n      <span translate=\"\">클릭하여 이를 개별적으로 관리합니다.</span>","A composite view contains other content views.":"복합 뷰에 다른 컨텐츠 뷰가 포함되어 있습니다. ","A new version of {{ contentView.name }} will be created and promoted to the Library environment.\n    It can be promoted to other environments from the Versions tab of this Content View.":"새 {{ contentView.name }} 버전이 생성되어 라이브러리 환경으로 승격됩니다.\n 이 컨텐츠 뷰 버전 탭에서 다른 환경으로 승격할 수 있습니다.","Account":"계정 ","Account Number":"계정 번호 ","Action":"동작 ","Action Type":"동작 유형 ","Actions":"동작 ","Actions for:":"동작: ","Activation Key":"활성키 ","Activation Key Content":"활성키 컨텐츠 ","Activation Key Type:":"활성키 유형:","Activation Key removed.":"활성키가 삭제되었습니다. ","Activation Key updated":"활성키가 업데이트되었습니다 ","Activation Key:":"활성키:","Activation Key: {{ activationKey.name }}":"활성키: {{ activationKey.name }}","Activation Keys":"활성키 ","Activation Keys for Subscription:":"서브스크립션의 활성키: ","Activation Keys using Version {{ version.version }}":"{{ version.version }} 버전을 사용하는 활성키 ","Add":"추가 ","Add Content Hosts to:":"컨텐츠 호스트 추가: ","Add Content Views":"컨텐츠 뷰 추가 ","Add Errata":"에라타 추가 ","Add New Environment":"새 환경 추가 ","Add New Filter":"새 필터 추가 ","Add New Module":"새 모듈 추가 ","Add New Repository":"새 리포지터리 추가 ","Add Package Group":"패키지 그룹 추가 ","Add Repositories":"리포지터리 추가 ","Add Selected":"선택 사항 추가 ","Add Subscriptions for Activation Key:":"활성키의 서브스크립션 추가:","Add To":"추가 ","Added %x host collections to activation key \"%y\".":"활성키 \"%y\"에 %x 호스트 컬렉션을 추가했습니다. ","Added %x host collections to content host \"%y\".":"컨텐츠 호스트 \"%y\"에 %x 호스트 컬렉션을 추가했습니다. ","Added %x products to sync plan \"%y\".":"동기화 계획 \"%y\"에 %x 제품을 추가했습니다. ","Adding Lifecycle Environment to the end of \"{{ priorEnvironment.name }}\"":"\"{{ priorEnvironment.name }}\" 마지막에 라이프사이클 환경을 추가하고 있습니다 ","Advanced Information":"고급 정보 ","Advisory":"권고","Affected Activation Keys":"영향을 받은 활성키 ","Affected Content Hosts":"영향을 받은 컨텐츠 호스트 ","Affected Packages":"영향을 받은 패키지 ","Affected Systems":"영향을 받은 시스템 ","Affected repositories":"영향을 받은 리포지터리 ","Affected repositories have been updated.":"영향을 받은 리포지터리가 업데이트되었습니다. ","All Content Views":"모든 컨텐츠 보기 ","All Products":"전체 제품","All Repositories":"전체 리포지터리 ","All Versions":"모든 버전 ","Alter Sync Plans":"동기화 계획 변경 ","An error occurred applying Subscriptions:":"서브스크립션을 적용하는 도중 오류가 발생했습니다: ","An error occurred removing the Activation Key:":"활성키를 제거하는 도중 오류가 발생했습니다: ","An error occurred removing the Host Collection:":"호스트 모음을 삭제하는 도중 오류가 발생했습니다: ","An error occurred removing the Products:":"제품을 삭제하는 도중 오류가 발생했습니다: ","An error occurred removing the content hosts.":"컨텐츠 호스트를 삭제하는 도중 오류가 발생했습니다. ","An error occurred removing the environment:":"환경을 삭제하는 도중 오류가 발생했습니다: ","An error occurred removing the subscriptions.":"서브스크립션을 삭제하는 도중 오류가 발생했습니다. ","An error occurred saving the Activation Key:":"활성키를 저장하는 도중 오류가 발생했습니다: ","An error occurred saving the Content Host:":"컨텐츠 호스트를 저장하는 도중 오류가 발생했습니다: ","An error occurred saving the Environment:":"환경을 저장하는 도중 오류가 발생했습니다: ","An error occurred saving the Filter:":"필터를 저장하는 도중 오류가 발생했습니다: ","An error occurred saving the Host Collection:":"호스트 모음을 저장하는 도중 오류가 발생했습니다: ","An error occurred saving the Product:":"제품을 저장하는 도중 오류가 발생했습니다: ","An error occurred saving the Repository:":"리포지터리를 저장하는 도중 오류가 발생했습니다: ","An error occurred saving the Sync Plan:":"동기화 계획을 저장하는 도중 오류가 발생했습니다: ","An error occurred saving the URL:":"URL을 저장하는 도중 오류가 발생했습니다: ","An error occurred trying to auto-attach subscriptions.  Please check your log for further information.":"서브스크립션 자동 첨부를 시도하는 도중 오류가 발생했습니다. 보다 자세한 내용은 로그에서 확인하십시오. ","An error occurred updating the Content View:":"컨텐츠 뷰를 업데이트하는 도중 오류가 발생했습니다: ","Applicable":"적용 가능 ","Applicable Content Hosts for Errata":"적용 가능한 에라타의 컨텐츠 호스트 ","Applicable Errata":"적용 가능한 에라타 ","Applicable hosts for {{ erratum.errata_id }}":"{{ erratum.errata_id }}에 대해 적용 가능한 호스트 ","Apply":"적용","Apply Errata":"에라타 적용","Apply Errata to Content Host \"{{contentHost.name}}\"?":"컨텐츠 호스트 \"{{contentHost.name}}\"에 에라타를 적용하시겠습니까?","Apply Errata to Content Hosts":"컨텐츠 호스트에 에라타 적용 ","Apply Errata to Content Hosts immediately after publishing.":"공개 후 에라타를 컨텐츠 호스트에 적용합니다. ","Apply Selected":"선택 사항 적용 ","Apply to Content Hosts":"컨텐츠 호스트에 적용 ","Apply to Hosts":"호스트에 적용 ","Apply {{ errata.errata_id }}":"{{ errata.errata_id }} 적용 ","Apply {{ errata.errata_id }} to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"{{ errata.errata_id }}을/를 {{ numberOfContentHostsToUpdate }} 컨텐츠 호스트에 적용하시겠습니까?","Apply {{ errataIds.length }} Errata to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"{{ errataIds.length }} 에라타를 {{ numberOfContentHostsToUpdate }} 컨텐츠 호스트에 적용하시겠습니까?","Applying":"적용 중 ","Applying Bulk Errata Update":"대량 에라타 업데이트 적용 중 ","Arch":"아키텍처 ","Architecture(s)":"아키텍처 ","Archived Copy":"아카이브 복사 ","Are you sure you want to add the {{ table.numSelected }} content host(s) selected to the host collection(s) chosen?":"선택한 호스트 컬렉션에 선택된 {{ table.numSelected }} 컨텐츠 호스트를 추가하시겠습니까? ","Are you sure you want to apply Errata to content host \"{{ contentHost.name }}\"?":"에라타를 컨텐츠 호스트 \"{{ contentHost.name }}\"에 적용하시겠습니까?","Are you sure you want to apply the {{ detailsTable.numSelected }} selected errata to the content hosts chosen?":"선택한 컨텐츠 호스트에 선택된 {{ detailsTable.numSelected }} 에라타를 적용하시겠습니까? ","Are you sure you want to assign the {{ table.numSelected }} content host(s) selected to {{ selected.contentView.name }} in {{ selected.environment.name }}?":"{{ selected.environment.name }}에서 {{ selected.contentView.name }}에 대해 선택된 {{ table.numSelected }} 컨텐츠 호스트를 할당하시겠습니까? ","Are you sure you want to auto-attach available subscriptions to all {{ table.resource.total }} content hosts?":"모든 {{ table.resource.total }}  컨텐츠 호스트에 사용 가능한 서브스크립션을 자동 첨부하시겠습니까? ","Are you sure you want to delete Content View \"{{ contentView.name }}\"?":"정말로 컨텐츠 뷰 \"{{ contentView.name }}\"를 삭제하시겠습니까? ","Are you sure you want to install {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"선택한 {{ getSelectedSystemIds().length }} 시스템에 {{ content.content }}을(를) 설치하시겠습니까? ","Are you sure you want to remove Activation Key \"{{ activationKey.name }}\"?":"활성키 \"{{ activationKey.name }}\"를 삭제하시겠습니까?","Are you sure you want to remove GPG key {{ gpgKey.name }}?":"GPG 키 {{ gpgKey.name }}를 삭제하시겠습니까?","Are you sure you want to remove Host Collection \"{{ hostCollection.name }}\"?":"호스트 컬렉션 \"{{ hostCollection.name }}\"을 삭제하시겠습니까?","Are you sure you want to remove Sync Plan \"{{ syncPlan.name }}\"?":"동기화 계획 \"{{ syncPlan.name }}\"을 삭제하시겠습니까?","Are you sure you want to remove product \"{{ product.name }}\"?":"제품 \"{{ product.name }}\"을 삭제하시겠습니까?","Are you sure you want to remove repository \"{{ repository.name }}\"?":"리포지터리 \"{{ repository.name }}\"를 삭제하시겠습니까?","Are you sure you want to remove the {{ productTable.getSelected().length }} products(s) selected?":"선택한 {{ productTable.getSelected().length }} 제품을 삭제하시겠습니까?","Are you sure you want to remove the {{ table.numSelected }} content host(s) selected from the host collection(s) chosen?":"선택한 호스트 컬렉션에서 선택된 {{ table.numSelected }} 컨텐츠 호스트를 삭제하시겠습니까?","Are you sure you want to remove {{ content.content }} from the {{ getSelectedSystemIds().length }} system(s) selected?":"선택한 {{ getSelectedSystemIds().length }} 시스템에서 {{ content.content }}을(를) 삭제하시겠습니까? ","Are you sure you want to unregister content host \"{{ contentHost.name }}\"?":"컨텐츠 호스트 \"{{ contentHost.name }}\"을 등록 해제하시겠습니까?","Are you sure you want to unregister the {{ table.numSelected }} content host(s) selected?":"선택한 컨텐츠 호스트 {{ table.numSelected }}을 등록 해제하시겠습니까?","Are you sure you want to update {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"선택한 {{ getSelectedSystemIds().length }} 시스템에서 {{ content.content }}을(를) 업데이트하시겠습니까? ","As part of this deletion, 1 Content View Version will be deleted.":"삭제의 일부로 {{ versions.length }} 컨텐츠 뷰 버전은 삭제됩니다.","Assign":"할당","Assign Environment and Content View":"환경 및 컨텐츠 뷰 할당 ","Associations":"연결 ","At least one Errata needs to be selected to Apply.":"적용하려면 최소 하나의 에라타를 선택해야 합니다. ","At least one content host needs to be selected in order to perform a bulk action.":"대량 작업을 실행하려면 최소 하나의 컨텐츠 호스트를 선택해야 합니다. ","At least one product needs to be selected in order to perform a bulk action.":"대량 작업을 실행하려면 최소 하나의 제품을 선택해야 합니다. ","Attached":"추가됨","Attached to Content Hosts":"컨텐츠 호스트에 첨부 ","Author":"작성자","Auto-Attach":"자동 첨부 ","Auto-attach available subscriptions to all content hosts":"모든 컨텐츠 호스트에 사용 가능한 서브스크립션을 자동 첨부 ","Automatic":"자동","Available Content Views for Composite Content View:":"복합적인 컨텐츠 뷰에 대해 사용 가능한 컨텐츠 뷰: ","Available Puppet Modules for Content View:":"컨텐츠 뷰에 대해 사용 가능한 Puppet 모듈: ","Back":"뒤로 ","Back To Errata List":"에라타 목록으로 돌아가기 ","Back To Repository Selection":"리포지터리 선택으로 돌아가기 ","Back to Module List":"모듈 목록으로 돌아가기 ","Back to Repository Details":"리포지터리 상세 정보로 돌아가기 ","Back to Repository List":"리포지터리 목록으로 돌아가기 ","Back to Version List":"버전 목록으로 돌아가기 ","Basic Information":"기본 정보 ","Below are the product's repository content sets currently available for this subscription. For Red Hat subscriptions, additional content can be made available through the":"다음은 이 서브스크립션에서 현재 사용할 수 있는 제품 리포지터리 컨텐츠 모음입니다. Red Hat 서브스크립션의 경우 추가 컨텐츠를 사용할 수 있습니다 ","Below are the repository content sets currently available for this activation key through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"다음은 이 서브스크립션을 통해 활성키에 대해 현재 사용할 수 있는 리포지터리 컨텐츠 모음입니다. Red Hat 서브스크립션의 경우 추가 컨텐츠를 사용할 수 있습니다.","Below are the repository content sets currently available for this content host through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"다음은 이 서브스크립션을 통해 컨텐츠 호스트에 대해 현재 사용할 수 있는 리포지터리 컨텐츠 모음입니다. Red Hat 서브스크립션의 경우 추가 컨텐츠를 사용할 수 있습니다.","Bug Fix":"버그 수정","Bug Fix Advisory":"버그 수정 권고","Bugfix":"버그 수정 ","Bulk Actions":"대량 작업 ","Bulk Alter Products Sync Plan":"제품 동기화 계획 대량 변경 ","Bulk Sync Products":"대량 동기화 제품 ","CVEs":"CVE","Cancel":"취소 ","Capacity":"용량 ","Change assigned Lifecycle Environment or Content View":"할당된 라이프 사이클 환경 또는 컨텐츠 뷰 변경 ","Changing default settings for content hosts that register with this activation key requires subscription-manager version 1.10 or newer to be installed on that host.":"이 활성키에 등록된 컨텐츠 호스트의 기본값 설증을 변경하려면 subscription-manager 버전 1.10 이상을 호스트에 설치해야 합니다.  ","Changing default settings requires subscription-manager version 1.10 or newer to be installed on this host.":"기본값 설정을 변경하려면 호스트에 subscription-manager 버전 1.10 이상을 설치해야 합니다. ","Checkin":"체크인 ","Checksum":"체크섬 ","Choose a lifecycle environment from the available promotion paths.":"사용 가능한 승격 경로에서 라이프 사이클 환경을 선택합니다.","Click here to check the status of the task.":"여기를 클릭하여 작업 상태를 확인하십시오.","Click here to refresh.":"여기를 클릭하여 새로 고침합니다. ","Click here to see the task for the successful removals.":"여기를 클릭하여 성공적으로 제거하기 위한 작업을 확인합니다.","Click here to select Errata for an Incremental Update.":"여기를 클릭하여 증분 업데이트를 위한 에라타를 선택합니다.","Close":"종료 ","Collection Actions":"컬렉션 동작 ","Completely remove version?":"버전을 완전하게 제거하시겠습니까?","Components":"구성 요소","Components:":"구성 요소:","Composite Content View {{ contentView.name }}":"복합적인 컨텐츠 뷰 {{ contentView.name }}","Composite View":"복합적인 뷰","Composite View?":"복합 뷰입니까?","Composite?":"복합된 것입니까?","Confirm":"확인","Confirm Remove":"삭제 확인 ","Confirm Version Removal: Version {{ version.version }}":"버전 삭제 확인: 버전 {{ version.version }}","Confirmation":"확인 ","Consumed":"사용됨 ","Content":"컨텐츠 ","Content Counts":"컨텐츠 수 ","Content Download URL":"컨텐츠 다운로드 URL","Content Host":"컨텐츠 호스트 ","Content Host %s has been deleted.":"컨텐츠 호스트 %s이(가) 삭제되었습니다. ","Content Host Bulk Actions":"컨텐츠 호스트 대량 작업 ","Content Host Bulk Content":"컨텐츠 호스트 대량 컨텐츠 ","Content Host Bulk Errata":"컨텐츠 호스트 대량 에라타 ","Content Host Bulk Host Collections":"컨텐츠 호스트 대량 호스트 컬렉션 ","Content Host Bulk Packages":"컨텐츠 호스트 대량 패키지 ","Content Host Bulk Subscriptions":"컨텐츠 호스트 대량 서브스크립션 ","Content Host Content":"컨텐츠 호스트 컨텐츠 ","Content Host Counts":"컨텐츠 호스트 수 ","Content Host Errata Management":"컨텐츠 호스트 에라타 관리 ","Content Host Limit":"컨텐츠 호스트 제한 ","Content Host Package Management":"컨텐츠 호스트 패키지 관리 ","Content Host Properties":"컨텐츠 호스트 우선 순위 ","Content Host Registration":"컨텐츠 호스트 등록 ","Content Host Status":"컨텐츠 호스트 상태 ","Content Host Subscriptions":"컨텐츠 호스트 서브스크립션 ","Content Host {{ contentHost.name }}":"컨텐츠 호스트 {{ contentHost.name }}","Content Host:":"컨텐츠 호스트: ","Content Hosts":"컨텐츠 호스트 ","Content Hosts for Activation Key:":"활성키의 컨텐츠 호스트: ","Content Hosts for Subscription:":"서브스크립션의 컨텐츠 호스트: ","Content Hosts for:":"컨텐츠 호스트: ","Content Hosts using Version {{ version.version }}":"{{ version.version }} 버전을 사용하는 컨텐츠 호스트 ","Content Type":"컨텐츠 유형 ","Content View":"컨텐츠 뷰 ","Content View Name":"컨텐츠 뷰 이름 ","Content View Version":"컨텐츠 뷰 버전 ","Content View updated.":"컨텐츠 뷰가 업데이트되었습니다.","Content View:":"컨텐츠 뷰:","Content Views":"컨텐츠 뷰 ","Content Views <div>{{ library.counts.content_views || 0 }}</div>":"컨텐츠 뷰 <div>{{ library.counts.content_views || 0 }}</div>","Content Views for Composite Content View:":"복합적인 컨텐츠 뷰에 대한 컨텐츠 뷰: ","Content successfully uploaded":"컨텐츠가 성공적으로 업로드되었습니다 ","Contents":"컨텐츠 ","Contract":"계약 ","Contract Number":"계약 번호 ","Copy":"복사","Copy Activation Key":"활성키 복사 ","Copy Collection":"컬렉션 복사 ","Copy View":"뷰 복사 ","Cores per Socket":"소켓 당 코어 ","Cores: %s":"코어: %s","Create":"생성 ","Create New View":"새 뷰 생성 ","Create Repositories":"리포지터리 생성 ","Create Repositories Within:":"리포지터리 생성: ","Create Repository":"리포지터리 생성 ","Create Selected":"선택 사항 생성 ","Create Sync Plan":"동기화 계획 생성 ","Create new view":"새 뷰 생성 ","Create view from existing views":"기존 뷰에서 뷰 생성 ","Critical":"중요함","Current Environment (%e/%cv)":"현재 환경 (%e/%cv)","Current Subscriptions for Activation Key:":"활성키의 현재 서브스크립션:","Currently Selected Puppet Modules":"현재 선택된 Puppet 모듈 ","Daily at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"매일 {{ product.sync_plan.sync_date | date:'mediumTime' }} (서버 시간)","Date":"날짜 ","Default":"기본값 ","Delete":"삭제 ","Delete Manifest":"매니페스트 삭제 ","Delete Version {{ version.version }}":"버전 {{ version.version }} 삭제 ","Delete {{ contentView.name }}":"{{ contentView.name }} 삭제 ","Deleted from %s":"%s에서 삭제됨 ","Deleting Manifest...":"매니페스트 삭제 중...","Deleting from %count environments.":"%count 환경에서 삭제 중 ","Deleting from 1 environment.":"1 환경에서 삭제 중 ","Deletion from %s":"%s에서 삭제","Description":"설명 ","Detail":"상세 정보 ","Details":"상세 정보 ","Details for Activation Key:":"활성키 상세 정보:","Details for Content View:":"컨텐츠 뷰 상세 정보: ","Details for Product:":"제품 상세 정보: ","Discover":"검색","Discovered URLs":"검색된 URL","Docker Content":"Docker 컨텐츠 ","Docker Images":"Docker 이미지 ","Docker Images in {{ repository.name }}":"{{ repository.name }}에 있는 Docker 이미지 ","Docker Repositories":"Docker 리포지터리 ","Docker Repositories <div>{{ library.counts.docker_repositories || 0 }}</div>":"Docker 리포지터리 <div>{{ library.counts.docker_repositories || 0 }}</div>","Docker Tag:":"Docker 태그:","Docker Tags":"Docker 태그","Docker metadata generation has been initiated in the background.  Click\n        <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Docker 메타데이터 생성이 백그라운드에서 시작되었습니다.\n        <a ng-href=\"{{ taskUrl() }}\">여기</a>를 클릭하여 진행 상태를 모니터합니다.","Edit":"수정","Enabled?":"활성화되어 있습니까?","End Date":"종료 날짜  ","Ends":"종료","Enhancement":"기능 강화","Enter Package Group Name(s)...":"패키지 그룹 이름 입력...","Enter Package Name(s)...":"패키지 이름 입력...","Environment":"환경 ","Environment saved":"환경이 저장되었습니다 ","Environment: {{ environment.name }}":"환경: {{ environment.name }}","Environments":"환경 ","Environments List":"환경 목록 ","Environments:":"환경: ","Equal To":"동일 ","Errata":"에라타 ","Errata <div>{{ library.counts.errata.total || 0 }}</div>":"에라타 <div>{{ library.counts.errata.total || 0 }}</div>","Errata Details":"에라타 상세 정보 ","Errata Filter:":"에라타 필터: ","Errata ID":"에라타 ID","Errata Installation":"에라타 설치 ","Errata List":"에라타 목록 ","Errata Task List":"에라타 작업 목록 ","Errata Tasks":"에라타 작업 ","Errata Type":"에라타 유형 ","Errata for:":"에라타 대상: ","Errata successfully added.":"에라타가 성공적으로 추가되었습니다. ","Errata successfully removed.":"에라타가 성공적으로 삭제되었습니다. ","Errata:":"에라타: ","Erratum - by Date and Type":"에라타 - 날짜 및 유형에 따라 ","Erratum - by ID":"에라타 - ID에 따라  ","Erratum Date Range":"에라타 날짜 범위 ","Error deleting manifest.":"매니페스트 삭제 도중 오류가 발생했습니다. ","Error during upload:":"업로드 도중 오류가 발생했습니다: ","Error importing manifest.":"이미지를 가져오는 도중 오류가 발생했습니다","Error refreshing manifest.":"매니페스트를 새로 고침하는 도중 오류가 발생했습니다 ","Event":"이벤트 ","Events":"이벤트","Exclude":"제외 ","Existing Product":"기존 제품 ","Expires":"만료 ","Failed":"실패 ","File too large. Please use the CLI instead.":"파일이 너무 큽니다. 대신 CLI를 사용하십시오. ","Filter":"필터 ","Filter Saved":"필터가 저장되었습니다 ","Filter information for:":"필터 정보: ","Filter...":"필터...","Filter:":"필터: ","Filters":"필터 ","Filters List":"필터 목록 ","Filters for Content View:":"컨텐츠 뷰 필터: ","Filters successfully removed.":"필터가 성공적으로 삭제되었습니다. ","Finished At":"종료 일시: ","For older operating systems such as Red Hat Enterprise Linux 5 or CentOS 5 it is recommended to use sha1.":"Red Hat Enterprise Linux 5 또는 CentOS 5와 같은 이전 운영 체제의 경우 sha1를 사용할 것을 권장합니다. ","GPG Key":"GPG 키 ","GPG Key URL":"GPG 키 URL","GPG Key successfully uploaded":"GPG 키가 성공적으로 업로드되었습니다 ","GPG Key {{ gpgKey.name }}":"GPG 키 {{ gpgKey.name }}","GPG Keys":"GPG 키 ","General Information":"일반 정보 ","Generated On":"생성 ","Gpg Key updated":"GPG 키가 업데이트되었습니다 ","Gpg Keys":"GPG 키 ","Greater Than":"이상 ","Group Install":"그룹 설치 ","Group Remove":"그룹 삭제 ","Guests of":"게스트 ","Hide affected Activation Keys":"영향을 받은 활성키 숨기기 ","Hide affected Content Hosts":"영향을 받은 컨텐츠 호스트 숨기기 ","History":"기록 ","History for Content View:":"컨텐츠 뷰 기록: ","Host Collection Management":"호스트 컬렉션 관리 ","Host Collection Membership":"호스트 컬렉션 멤버쉽 ","Host Collection Membership Management":"호스트 컬렉션 멤버쉽 관리 ","Host Collection removed.":"호스트 컬렉션이 삭제되었습니다. ","Host Collection updated":"호스트 컬렉션이 업데이트되었습니다 ","Host Collection:":"호스트 컬렉션: ","Host Collection: {{ hostCollection.name }}":"호스트 컬렉션: {{ hostCollection.name }}","Host Collections":"호스트 컬렉션 ","Host Collections for:":"호스트 컬렉션: ","Host Count":"호스트 수 ","Host Group":"호스트 그룹 ","Hostname":"호스트 이름","Hourly at {{ product.sync_plan.sync_date | date:'m' }} minutes and {{ product.sync_plan.sync_date | date:'s' }} seconds":"매시 {{ product.sync_plan.sync_date | date:'m' }} 분 {{ product.sync_plan.sync_date | date:'s' }} 초","IPv4 Address":"IPv4 주소 ","IPv6 Address":"IPv6 주소 ","Id":"Id","If you would prefer to move portions of these Activation Keys to different Content Views or Lifecycle Environments click <a href=\"{{ activationKeyLink() }}\">here</a> to manage them individually.":"이러한 활성키의 일부를 다른 컨텐츠 뷰 또는 라이프 사이클 환경으로 이동하고자 하는 경우 <a href=\"{{ activationKeyLink() }}\">여기</a>를 클릭하여 이를 개별적으로 관리합니다. ","If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click":"이러한 활성키의 일부를 다른 컨텐츠 뷰 또는 라이프 사이클 환경으로 이동하고자 하는 경우 클릭합니다 ","If you would prefer to move portions of these Content Hosts to different content views or environments click <a href=\"{{ contentHostsLink() }}\">here</a>  to manage these Content Hosts in bulk.":"이 컨텐츠 호스트의 일부를 다른 컨텐츠 뷰 또는 환경에 이동하고자 할 경우 <a href=\"{{ contentHostsLink() }}\">여기</a>를 클릭하여 컨텐츠 호스트를 대량으로 관리합니다. ","Image":"이미지 ","Image ID":"이미지 ID","Import History":"가져오기 기록 ","Important":"중요","In order to add a content view to a composite view you must first publish an initial version of the content view.":"컨텐츠 뷰를 복합 뷰에 추가하려면 먼저 컨텐츠 뷰의 초기 버전을 공개해야 합니다. ","Include":"포함 ","Included in Activation Keys":"활성키에 포함됨 ","Incremental Update":"증분 업데이트 ","Install":"설치 ","Install Selected":"선택 사항 설치 ","Install katello-agent for remote actions and displaying errata information:":"원격 작업을 위해 katello-agent 설치 및 에라타 정보 표시:","Install the pre-built bootstrap RPM:":"사전 생성된 부트스트랩 RPM을 설치합니다:","Installable":"설치 가능 ","Installable Errata":"설치 가능한 에라타 ","Installed":"설치됨 ","Installed Package":"설치된 패키지 ","Installed Packages":"설치된 패키지 ","Installed Products":"설치된 제품 ","Instance-based":"인스턴스 기반 ","Interfaces":"인터페이스 ","Interval":"간격 ","Issued":"발행됨 ","Katello Agent":"Katello 에이전트 ","Label":"레이블 ","Last Checkin":"마지막 체크인 ","Last Published":"마지막 공개 일시 ","Last Puppet Report":"마지막 Puppet 보고 ","Last Sync":"마지막 동기화","Last Updated":"마지막 업데이트 일시","Last Updated On":"마지막 업데이트: ","Last synced {{ product.last_sync_words }} ago.":"마지막 동기화: {{ product.last_sync_words }} 전.","Latest (Currently %s)":"최신 (현재 %s)","Less Than":"이하 ","Library":"라이브러리","Library Synced Content":"라이브러리와 동기화된 컨텐츠 ","Lifecycle Environment \"{{ priorEnvironment.name }}\" exists within the middle of a path. New Lifecycle Environments can only be added to the end of a path.":"라이프 사이클 환경 \"{{ priorEnvironment.name }}이 경로 중간에 있습니다. 새 라이프 사이클 환경은 경로 마지막에만 추가할 수 있습니다. ","Lifecycle Environment Paths":"라이프사이클 환경 경로 ","Lifecycle Environments":"라이프사이클 환경 ","Limit":"제한 ","Limit:":"제한: ","Limits":"제한 ","List/Remove":"목록/삭제","Loading Environment Paths...":"환경 경로 로딩 중...","Loading...":"로딩...","Manage Docker Images":"Docker 이미지 관리 ","Manage Docker Images for Repository:":"리포지터리의 Docker 이미지 관리: ","Manage Manifest":"매니페스트 관리 ","Manage Packages":"패키지 관리 ","Manage Packages for Repository:":"리포지터리의 패키지 관리: ","Manage Puppet Modules":"Puppet 모듈 관리 ","Manage Puppet Modules for Repository:":"리포지터리의 Puppet 모듈 관리: ","Manifest History":"매니페스트 기록 ","Manifest successfully deleted.":"매니페스트가 성공적으로 삭제되었습니다. ","Manifest successfully imported.":"매니페스트을 성공적으로 불러왔습니다. ","Manifest successfully refreshed.":"매니페스트가 성공적으로 새로고침되었습니다.","Message":"메세지 ","Model":"모델 ","Moderate":"중간 수준","Module %s removed from Content View.":"모듈 %s이 컨텐츠 뷰에서 삭제되었습니다.","Module List":"모듈 목록 ","More Details":"상세 정보 ","More...":"상세 정보...","Multi-entitlement":"멀티 인타이틀먼트 ","N/A":"해당 없음 ","Name":"이름 ","Name of the upstream repository you want to sync. Example: 'busybox' or 'fedora/ssh'.":"동기화하고자 하는 업스트림 리포지터리 이름입니다. 예: 'busybox' 또는 'fedora/ssh'.","Networking":"네트워크 구성 ","Never synced":"동기화하지 않음 ","New Activation Key":"새 활성화 키 ","New Content View":"새 컨텐츠 뷰 ","New Environment":"새 환경 ","New Environment Path":"새 환경 경로 ","New Filter":"새 필터 ","New Filter for Content View:":"컨텐츠 뷰의 새 필터: ","New GPG Key":"새 GPG 키 ","New Gpg Key":"새 GPG 키 ","New Host Collection":"새 호스트 컬렉션 ","New Name:":"새로운 이름:","New Product":"새 제품 ","New Repository":"새 리포지터리 ","New Sync Plan":"새 동기화 계획 ","New sync plan successfully created.":"새 동기화 계획이 성공적으로 생성되었습니다. ","Next":"다음","Next Sync":"다음 동기화 ","No":"아니오  ","No \"Edit\" permissions on some of the Activation Keys in that Lifecycle Environment.":"라이프 사이클 환경에 있는 활성키 일부에는 \"편집\" 권한이 없습니다. ","No \"Edit\" permissions on some of the Content Hosts in that Lifecycle Environment.":"라이프 사이클 환경에 있는 컨텐츠 호스트 일부에는 \"편집\" 권한이 없습니다. ","No \"Promote/Remove\" permissions on the Lifecycle Environment.":"라이프 사이클 환경에는 \"승격/삭제\" 권한이 없습니다.","No (Default)":"아니요 (기본값)","No Content Hosts are affected.":"영향을 받는 컨텐츠 호스트가 없습니다. ","No Content Hosts match this Erratum.":"이 에라타에 일치하는 컨텐츠 호스트가 없습니다. ","No Content Views available, please select another environment.":"사용 가능한 컨텐츠 뷰가 없습니다. 다른 환경을 선택하십시오. ","No Host Collections to show, you can add Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"표시할 호스트 컬렉션이 없습니다. 메인 메뉴의 '호스트'에서 '호스트 컬렉션'을 선택한 후 호스트 컬렉션을 추가할 수 있습니다. ","No Host Collections to show, you can add Host Collections after selecting the 'Add' tab.":"표시할 호스트 컬렉션이 없습니다. '추가' 탭을 선택하여 호스트 컬렉션을 추가할 수 있습니다.","No Products to show.":"표시할 제품이 없습니다.","No Repositories contain this Erratum.":"이 에라타에 포함된 리포지터리가 없습니다. ","No activation keys are affected.":"영향을 받는 활성키가 없습니다. ","No alternate release version choices are available. The available releases are based upon what is available in \"{{ contentHost.content_view.name }}\", the selected <a href=\"/content_views\">content view</a> this content host is attached to for the given <a href=\"/lifecycle_environments\">lifecycle environment</a>, \"{{ contentHost.environment.name }}\".":"선택할 수 있는 다른 릴리즈 버전이 없습니다. 사용 가능한 릴리즈는 \"{{ contentHost.content_view.name }}\"에서 사용 가능한 릴리즈에 근거합니다.  이러한 컨텐츠 호스트는 <a href=\"/lifecycle_environments\">라이프 사이클 환경</a>, \"{{ contentHost.environment.name }}\"에 대해 할당된 선택한 <a href=\"/content_views\">컨텐츠 뷰</a>입니다. ","No content views exist for {{selected.environment.name}}":"{{selected.environment.name}}의 컨텐츠 뷰가 존재하지 않습니다 ","No host collections":"호스트 컬렉션이 없습니다 ","No provisioning details are available for this content host.":"컨텐츠 호스트에 사용 가능한 프로비저닝 상세 정보가 없습니다. ","No repository content provided through subscriptions.":"서브스크립션을 통해 제공되는 리포지터리 컨텐츠가 없습니다. ","No repository content provided through this subscription.":"이 서브스크립션을 통해 제공되는 리포지터리 컨텐츠가 없습니다. ","No repository content.":"리포지터리 컨텐츠가 없습니다. ","No subscription manifest imported":"서브스크립션 매니페스트를 가져오지 않았습니다 ","None":"없음 ","Not Synced":"동기화되지 않았습니다 ","Not installed":"설치되지 않음 ","Not yet published":"아직 공개되지 않았습니다 ","Number of CPUs":"CPU 수 ","OS":"OS","Only show Errata that is Applicable to one or more Content Hosts":"하나 이상의 컨텐츠 호스트에 적용 가능한 에라타만 표시 ","Only show Errata that is Installable on one or more Content Hosts":"하나 이상의 컨텐츠 호스트에 설치 가능한 에라타만 표시 ","Operating System":"운영체제","Options:":"옵션:","Organization":"조직 ","Original Sync Date":"원래 동기화 날짜 ","Override to No":"\"아니요\"로 덮어쓰기 ","Override to Yes":"\"예\"로 덮어쓰기 ","Package":"패키지 ","Package Actions":"패키지 동작 ","Package Filter:":"패키지 필터: ","Package Group":"패키지 그룹 ","Package Group Filter:":"패키지 그룹 필터: ","Package Group successfully added.":"패키지 그룹이 성공적으로 추가되었습니다. ","Package Group successfully removed.":"패키지 그룹이 성공적으로 삭제되었습니다. ","Package Groups":"패키지 그룹 ","Package Install":"패키지 설치 ","Package Installation, Removal, and Update":" 패키지 설치, 제거, 업데이트 ","Package Remove":"패키지 제거 ","Package Update":"패키지 업데이트 ","Package removal failed":"패키지 삭제에 실패했습니다 ","Package successfully added.":"패키지가 성공적으로 추가되었습니다. ","Package successfully removed.":"패키지가 성공적으로 삭제되었습니다. ","Package successfully updated.":"패키지가 성공적으로 업데이트되었습니다.","Package/Group Name":"패키지/그룹 이름 ","Packages":"패키지 ","Packages <div>{{ library.counts.packages || 0 }}</div>":"패키지 <div>{{ library.counts.packages || 0 }}</div>","Packages for:":"패키지 대상:  ","Packages in {{ repository.name }}":"{{ repository.name }}에 있는 패키지 ","Parameters":"매개 변수 ","Paste GPG Key Contents":"GPG 키 컨텐츠 붙여넣기 ","Paste contents...":"컨텐츠 붙여넣기...","Paused":"일시정지 ","Perform":"실행 ","Physical":"물리적 ","Please select a Lifecycle Environment and Content View to move the affected Activation Keys to:":"영향을 받은 활성키를 옮기려면 라이프사이클 환경 및 컨텐츠 뷰를 선택하십시오: ","Please select a Lifecycle Environment and Content View to move these Content Hosts to:":"컨텐츠 호스트를 옮기려면 라이프 사이클 환경 및 컨텐츠 뷰를 선택하십시오:","Please select a Lifecycle Environment.":"라이프 사이클 환경을 선택하십시오. ","Please select an environment.":"환경을 선택하십시오. ","Please select one from the list below and you will be redirected.":"다음 목록 중 하나를 선택하면 선택한 곳으로 이동합니다.","Press 'Sync Now' below to begin syncing the following product(s).":"다음의 '지금 동기화'를 눌러 다음 제품을 동기화 시작합니다.","Previous Environment (%e/%cv)":"이전 환경 (%e/%cv)","Processing":"처리 중 ","Product":"제품 ","Product Bulk Actions":"제품 대량 작업 ","Product Content":"제품 컨텐츠 ","Product Count":"제품 수 ","Product Enhancement Advisory":"제품 기능 개선 권고","Product ID":"제품 ID","Product Management":"제품 관리 ","Product Management for Sync Plan:":"동기화 계획을 위한 제품 관리: ","Product Name":"제품 이름 ","Product Saved":"저장된 제품 ","Product Sync":"제품 동기화 ","Product delete operation has been initiated in the background.  Click <a href=\"{{ taskUrl(productDeletionTaskId) }}\">Here</a> to monitor the progress.":"제품 삭제 작업이 백그라운드에서 시작되었습니다. <a href=\"{{ taskUrl(productDeletionTaskId) }}\">여기</a>를 클릭하여 진행 상태를 모니터하십시오. ","Product information for:":"제품 정보: ","Product {{ product.name }}":"제품 {{ product.name }}","Products":"제품 ","Products <div>{{ library.counts.products || 0 }}</div>":"제품 <div>{{ library.counts.products || 0 }}</div>","Promote":"승격 ","Promote Content View:":"컨텐츠 뷰 승격: ","Promote Version":"버전 승격 ","Promote Version {{ version.version }}":"버전 {{ version.version }} 승격 ","Promoted to %s":"%s로 승격 ","Promoting to %count environments.":"%count 환경으로 승격되고 있습니다. ","Promoting to 1 environment.":"1 환경으로 승격 중 ","Promotion History":"승격 기록 ","Provided Products":"제공되는 제품 ","Provisioning Details":"프로비저닝 상세 정보 ","Provisioning Host Details":"프로비저닝 호스트 상세 정보 ","Publish Content View:":"컨텐츠 뷰 공개: ","Publish New Version":"새 버전 공개 ","Publish via HTTP":"HTTP를 통해 공개 ","Publish via HTTPS":"HTTPS를 통해 공개 ","Published":"공개 일시 ","Published At":"공개 일시 ","Published Repository Information":"리포지터리 정보 공개 ","Published new version":"새 버전 공개 ","Publishing and promoting to %count environments.":"%count 환경으로 공개 및 승격 중 ","Publishing and promoting to 1 environment.":"1 환경으로 공개 및 승격 중 ","Puppet Environment":"Puppet 환경 ","Puppet Modules":"Puppet 모듈 ","Puppet Modules <div>{{ library.counts.puppet_modules || 0 }}</div>":"Puppet 모듈 <div>{{ library.counts.puppet_modules || 0 }}</div>","Puppet Modules for Content View:":"컨텐츠 뷰의 Puppet 모듈: ","Puppet Modules in {{ repository.name }}":"{{ repository.name }}에 있는 Puppet 모듈","Puppet module added to Content View":"Puppet 모듈이 컨텐츠 뷰에 추가되었습니다 ","Puppet module metadata generation has been initiated in the background.\n        Click <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Puppet 모듈 메타데이터 생성이 백그라운드에서 시작되었습니다\n <a ng-href=\"{{ taskUrl() }}\">여기</a>를 클릭하여 진행 상태를 모니터합니다.","Quantity":"수량 ","RAM (GB)":"RAM (GB)","RAM: %s GB":"RAM: %s GB","Range":"범위 ","Reboot Suggested":"재부팅이 권장됩니다 ","Reboot Suggested?":"재부팅하는 것이 좋습니까?","Red Hat CDN URL":"Red Hat CDN URL","Red Hat Provider Details":"Red Hat 공급자 정보 ","Red Hat Repositories page":"Red Hat 리포지터리 페이지 ","Red Hat Repositories page.":"Red Hat 리포지터리 페이지.","Refresh Manifest":"매니페스트 새로 고침 ","Refresh Table":"표 새로 고침 ","Refreshing Manifest...":"매니페스트 새로 고침하는 중 ","Refreshing manifest not suppported by this manifest.":"이 매니페스트에 의해 지원되지 않는 매니페스트를 새로 고침하고 있습니다. ","Register Content Host":"컨텐츠 호스트 등록 ","Register a Content Host":"컨텐츠 호스트 등록 ","Registered":"등록됨 ","Registered By":"등록: ","Registry URL":"레지스트리 URL","Release":"릴리즈 ","Release Version":"릴리즈 버전 ","Remove":"삭제 ","Remove Activation Key \"{{ activationKey.name }}\"?":"활성키 \"{{ activationKey.name }}\"을/를 삭제하시겠습니까?","Remove Content View":"컨텐츠 뷰 삭제 ","Remove Content Views":"컨텐츠 뷰 삭제 ","Remove Docker Images":"Docker 이미지 삭제 ","Remove Errata":"에라타 삭제 ","Remove Filters":"필터 삭제 ","Remove From":"삭제 대상 ","Remove GPG Key":"GPG 키 삭제 ","Remove GPG Key {{ gpgKey.name }}":"GPG 키 {{ gpgKey.name }}을/를 삭제하시겠습니까?","Remove Host Collection \"{{ hostCollection.name }}\"?":"호스트 컬렉션 \"{{ hostCollection.name }}\"을/를 삭제하시겠습니까?","Remove Module":"모듈 삭제 ","Remove Package Group":"패키지 그룹 삭제 ","Remove Packages":"패키지 삭제 ","Remove Product":"제품 삭제 ","Remove Product \"{{ product.name }}\"?":"제품 \"{{ product.name }}\"을/를 삭제하시겠습니까?","Remove Products":"제품 삭제 ","Remove Puppet Modules":"Puppet 모듈 삭제 ","Remove Repositories":"리포지터리 삭제 ","Remove Repository":"리포지터리 삭제 ","Remove Repository \"{{ repository.name }}\"?":"리포지터리 \"{{ repository.name }}\"을/를 삭제하시겠습니까?","Remove Selected":"선택한 사항 삭제 ","Remove Successful.":"성공적으로 삭제되었습니다. ","Remove Sync Plan":"동기화 계획 삭제 ","Remove Sync Plan \"{{ syncPlan.name }}\"?":"동기화 계획 \"{{ syncPlan.name }}\"을/를 삭제하시겠습니까?","Remove Version":"버전 삭제 ","Remove Version Confirmation":"버전 삭제 확인 ","Remove View":"뷰 삭제 ","Remove {{ productTable.getSelected().length }} Products?":"{{ productTable.getSelected().length }} 제품을 삭제하시겠습니까?","Removed %x host collections from activation key \"%y\".":"활성키 \"%y\"에서 %x 호스트 컬렉션이 삭제되었습니다. ","Removed %x host collections from content host \"%y\".":"컨텐츠 호스트 \"%y\"에서 %x 호스트 컬렉션이 삭제되었습니다. ","Removed %x products from sync plan \"%y\".":"동기화 계획 \"%y\"에서 %x 제품이 삭제되었습니다. ","Removing Content Hosts":"컨텐츠 호스트 삭제 중 ","Removing Products":"제품 삭제 중 ","Removing Repositories":"리포지터리 삭제 중 ","Repo Discovery":"리포지터리 검색 ","Repo Type":"리포지터리 유형 ","Repositories":"리포지터리 ","Repositories containing Errata {{ errata.errata_id }}":"에라타 {{ errata.errata_id }}가 들어 있는 리포지터리 ","Repositories for Content View:":"컨텐츠 뷰의 리포지터리: ","Repositories for Errata:":"에라타 용 리포지터리: ","Repositories for Filter:":"필터 용 리포지터리: ","Repositories for Product:":"제품 용 리포지터리: ","Repository":"리포지터리 ","Repository \"%s\" successfully deleted":"리포지터리 \"%s\"이/가 성공적으로 삭제되었습니다","Repository %s successfully created.":"리포지터리 %s가 성공적으로 생성되었습니다.","Repository Count":"리포지터리 수 ","Repository Discovery":"리포지터리 검색 ","Repository Name":"리포지터리 이름 ","Repository Saved.":"리포지터리가 저장되었습니다. ","Repository Selection":"리포지터리 선택 ","Repository URL updated":"리포지터리 URL이 업데이트되었습니다 ","Repository:":"리포지터리: ","Required Host":"필요한 호스트 ","Result":"결과 ","Results":"결과 ","Rules Source":"규칙 소스 ","Rules Version":"규칙 버전 ","Run Auto-Attach":"자동 첨부 실행 ","Run subscription-manager in a console on the client host.  You may use an Activation Key to register:":"클라이언트 호스트의 콘솔에서 subscription-manager를 실행합니다. 활성키를 사용하여 등록할 수 있습니다:","Save Successful.":"성공적으로 저장되었습니다. ","Search...":"검색...","Security":"보안 ","Security Advisory":"보안 권고 ","Select A New Puppet Module To Add":"추가할 새 Puppet 모듈 선택 ","Select Content Hosts":"컨텐츠 호스트 선택 ","Select Module":"모듈 선택 ","Select Version":"버전 선택 ","Select a Content Source:":"컨텐츠 소스 선택:","Select a Version":"버전 선택 ","Select an Available Version of {{ $stateParams.moduleName }}":"{{ $stateParams.moduleName }}의 사용 가능한 버전 선택 ","Select an Organization":"조직 선택 ","Select new version":"새 버전 선택 ","Select the Lifecycle Environments you would like to remove Version {{ version.version }} from:":"버전 {{ version.version }}을 제거하고자 하는 라이프 사이클 환경 선택: ","Selected URLs:":"선택된 URL:","Selecting this option will result in the Version being completely deleted and no longer being available for promotion.  The version must be removed from all Lifecycle Environments in order to select this option.":"이 옵션을 선택하면 버전이 완전하게 삭제되고 더 이상 승격 대상으로 사용할 수 없게 됩니다. 이 옵션을 선택하려면 버전을 모든 라이프 사이클 환경에서 삭제해야 합니다. ","Serve via HTTP.":"HTTP를 통해 제공합니다. ","Service Level":"서비스 레벨 ","Severity":"심각성 ","Show affected Activation Keys":"영향을 받은 활성키 표시 ","Show affected Content Hosts":"영향을 받은 컨텐츠 호스트 표시 ","Show from:":"보기: ","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"{{ detailsTable.resource.subtotal }} (총 {{ detailsTable.resource.total }} ) 컨텐츠 호스트 중 {{ detailsTable.rows.length }}를 표시 중 ","Showing {{ addSubscriptionsTable.rows.length }} of {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} Total) Subscriptions":"{{ addSubscriptionsTable.resource.subtotal }} (총 {{ addSubscriptionsTable.resource.total }}) 서브스크립션 중 {{ addSubscriptionsTable.rows.length }}를 표시 중 ","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"{{ detailsTable.resource.subtotal }} (총 {{ detailsTable.resource.total }}) 컨텐츠 호스트 중 {{ detailsTable.rows.length }}를 표시 중 ","Showing {{ currentPackagesTable.resource.total }} Packages":"{{ currentPackagesTable.resource.total }} 패키지 표시 ","Showing {{ subscriptionsTable.rows.length }} of {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} Total) Subscriptions":"{{ subscriptionsTable.resource.subtotal }} (총 {{ subscriptionsTable.resource.total }}) 서브스크립션 중 {{ subscriptionsTable.rows.length }}를 표시 중  ","Sockets":"소켓 ","Sockets: %s":"소켓: %s","Solution":"솔루션 ","Some of the Errata shown below may not be installable as they are not in this Content Host's\n          Content View and Lifecycle Environment.  In order to apply such Errata an Incremental Update is required.":"다음에 표시된 에라타 일부는 컨텐츠 호스트의 컨텐츠 뷰 및 라이프 사이클 환경에 없기 때문에 \n      설치할 수 없는 경우가 있습니다. 이러한 에라타를 적용하려면 증분 업데이트가 필요합니다.","Stacking ID":"스태킹 ID","Start Date":"시작 날짜 ","Start Time":"시작 시간 ","Started At":"시작 일시 ","Starting":"시작하는 중","Starts":"시작 ","State":"상태 ","Status":"상태 ","Subscription Details":"서브스크립션 정보  ","Subscription Info":"서브스크립션 정보 ","Subscription Manifest":"서브스크립션 매니페스트","Subscription Server Details":"서브스크립션 서버 상세 정보 ","Subscription Status":"서브스크립션 상태 ","Subscriptions":"서브스크립션 ","Subscriptions for Activation Key:":"활성키의 서브스크립션: ","Subscriptions for:":"서브스크립션 대상: ","Succesfully scheduled package installation":"패키지 설치를 성공적으로 스케줄했습니다 ","Succesfully scheduled package removal":"패키지 삭제를 성공적으로 스케줄했습니다 ","Succesfully scheduled package update":"패키지 업데이트를 성공적으로 스케줄했습니다 ","Success":"성공 ","Successfully Scheduled Auto-attach.":"자동 첨부를 성공적으로 스케줄했습니다.","Successfully added %s subscriptions.":"%s 서브스크립션을 성공적으로 추가했습니다 ","Successfully deleted %cv version %ver.":"%cv 버전 %ver을 성공적으로 삭제했습니다. ","Successfully initiated deletion of %cv version %ver.":"%cv 버전 %ver 삭제가 성공적으로 시작되었습니다. ","Successfully initiated promotion of %cv version %ver to %env.":"%cv 버전 %ver to %env 승격이 성공적으로 시작되었습니다. ","Successfully initiated removal of %cv version %ver.":"%cv 버전 %ver 삭제가 성공적으로 시작되었습니다. ","Successfully promoted %cv version %ver to %env":"%cv 버전 %ver이 %env로 성공적으로 승격되었습니다 ","Successfully published %cv version %ver and promoted to Library":"%cv 버전 %ver이 성공적으로 공개되어 라이브러리로 승격되었습니다 ","Successfully removed %cv version %ver from environments: %env":"%cv 버전 %ver이 다음 환경에서 성공적으로 삭제되었습니다: %env","Successfully removed %s items.":"%s 항목을 성공적으로 삭제했습니다. ","Successfully removed %s subscriptions.":"%s 서브스크립션을 성공적으로 삭제했습니다. ","Successfully removed 1 item.":"1 개의 항목이 성공적으로 삭제되었습니다. ","Successfully updated subscriptions.":"서브스크립션을 성공적으로 업데이트했습니다. ","Summary":"요약 ","Support Level":"지원 수준 ","Support Type":"지원 유형 ","Sync Enabled":"동기화가 활성화됨 ","Sync Information":"동기화 정보 ","Sync Interval":"동기화 간격 ","Sync Now":"지금 동기화 ","Sync Plan":"동기화 계획 ","Sync Plan %s has been deleted.":"동기화 계획 %s을(를) 삭제했습니다. ","Sync Plan Saved":"동기화 계획이 저장되었습니다 ","Sync Plan {{ syncPlan.name }}":"동기화 계획 {{ syncPlan.name }}","Sync Plan:":"동기화 계획: ","Sync Plans":"동기화 계획 ","Sync State":"동기화 상태 ","Sync Status":"동기화 상태 ","Synced manually, no interval set.":"수동으로 동기화되었습니다. 간격이 설정되어 있지 않습니다. ","Syncing...":"동기화 중...","Tags":"태그","Target":"대상 ","Task Details":"작업 상세 정보 ","Tasks":"작업","Tasks for Content View:":"컨텐츠 뷰 작업: ","Temporary":"임시","The Activation Keys listed below are currently using this Content View Version.  Before deleting the Version you must move these Activation Keys to a Lifecycle Environment where this Version is not in use.":"다음에 나열된 활성키는 현재 이 컨텐츠 뷰 버전을 사용하고 있습니다. 버전을 삭제하기 전 이러한 활성키를 버전이 사용되고 있지 않은 라이프 사이클 환경으로 이동해야 합니다. ","The Content Hosts listed below are currently using this Content View Version.  Before removing the version you must move these Content Hosts to an environment where this version is not in use.":"다음에 나열된 컨텐츠 호스트는 현재 이 컨텐츠 뷰 버전을 사용하고 있습니다. 버전을 삭제하기 전 이러한 컨텐츠 호스트를 버전이 사용되고 있지 않은 환경으로 이동해야 합니다. ","The Content View or Lifecycle Environment needs to be updated in order to make errata available to these hosts.":"이 호스트에서 에라타를 사용 가능하게 하려면 컨텐츠 뷰 또는 라이프 사이클 환경을 업데이트해야 합니다.","The archive of this Version will be deleted.  This Version will not be available once deletion is complete.":"이 버전의 아카이브는 삭제됩니다. 삭제 완료되면 이 버전을 사용할 수 없습니다. ","The archive of this Version will not be deleted.  This Version will still be available to be promoted to a Lifecycle Environment.":"이 버전의 아카이브는 삭제되지 않습니다. 이 버전은 라이프 사이클 환경으로 승격하는데 사용할 수 있습니다. ","The following actions can be performed on content hosts in this host collection:":"다음과 같은 작업이 호스트 컬렉션에 있는 컨텐츠 호스트에서 실행될 수 있습니다: ","The host requires being attached to a content view and the environment you have chosen has no content views promoted to it.\n            See the <a href=\"/content_views\">content views page</a> to manage and promote a content view.":"호스트는 컨텐츠 뷰에 할당되어 있어야 합니다. 선택한 환경에는 승격된 컨텐츠 뷰가 없습니다. \n   컨텐츠 뷰를 관리 및 승격하려면 <a href=\"/content_views\">컨텐츠 뷰 페이지</a>를 참조하십시오.","The katello-agent package is required to manage errata on this Host.\n      Please install the katello-agent package to manage errata on\n      {{ contentHost.name }}.":"이 호스트에서 에라타를 관리하려면 katello-agent 패키지가 필요합니다.\n    {{ contentHost.name }}에 있는 에라타를 관리하려면  katello-agent \n패키지를 설치하십시오.","The katello-agent package is required to manage packages on this Host.\n      Please install the katello-agent package to manage packages on\n      {{ contentHost.name }}. To install the package, connect directly to the\n      host from a terminal and use yum.":"이 호스트에서 패키지를 관리하려면 katello-agent 패키지가 필요합니다.\n      {{ contentHost.name }}에 있는 패키지를 관리하려면 katello-agent \n    패키지를 설치하십시오. 패키지를 설치하려면 터미널에서 호스트로 직접\n   연결하여 yum을 사용합니다.","The page you are attempting to access requires selecting a specific organization.":"액세스하려는 페이지에서 특정 조직을 선택해야 합니다.","The selected environment contains no Content Views, please select a different environment.":"선택한 환경에 컨텐츠 뷰가 없습니다. 다른 환경을 선택하십시오. ","There are currently no repositories associated with this Content View, you can add some by clicking on the \"Add\" tab above.":"현재 이 컨텐츠 뷰와 연결된 리포지터리가 없습니다. 위의 \"추가\" 탭을 클릭하여 이를 추가할 수 있습니다. ","There are currently no repositories to add to this Content View,":"현재 컨텐츠 뷰에 추가할 리포지터리가 없습니다.","There are no Content Views in this Environment.":"이 환경에는 컨텐츠 뷰가 없습니다. ","There are no Content Views that match the criteria.":"조건에 일치하는 컨텐츠 뷰가 없습니다.","There are no Docker Tags that match the criteria.":"조건에 일치하는 Docker 태그가 없습니다.","There are no Errata associated with this Content Host to display.":"표시할 컨텐츠 호스트에 연결된 에라타가 없습니다.","There are no Errata in this organization.  Create one or more Products with Errata to view Errata on this page.":"이 조직에는 에라타가 없습니다. 이 페이지에 에라타를 표시하려면 에라타와 함께 하나 이상의 제품을 생성합니다.","There are no Errata that match the criteria.":"조건에 일치하는 에라타가 없습니다.","There are no Errata to display.":"표시할 에라타가 없습니다. ","There are no Host Collections available. You can create new Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"사용 가능한 호스트 컬렉션이 없습니다. 메인 메뉴의 '호스트'에서 '호스트 컬렉션'을 선택한 후 새 호스트 컬렉션을 생성합니다. ","There are no Packages that match the criteria.":"조건에 일치하는 패키지가 없습니다.","There are no Puppet Modules that match the criteria.":"조건에 일치하는 Puppet 모듈이 없습니다.","There are no Yum Repositories that match the criteria.":"조건에 일치하는 Yum 리포지터리가 없습니다.","There is currently an Incremental Update task in progress.  This update must finish before applying existing updates.":"현재 증분 업데이트 작업이 진행 중입니다. 기존 업데이트를 적용하기 전 이 업데이트를 완료해야 합니다.","There were errors while removing the following Repositories:":"다음의 리포지터리를 제거하는 도중 오류가 발생했습니다:","This Content Host is not attached to any subscriptions that provide content":"이 컨텐츠 호스트는 컨텐츠를 제공하는 서브스크립션에 첨부되어 있지 않습니다. ","This Content View does not have any versions, create your first Content View Version by using the \"Publish New Version\" button on the right.":"이 컨텐츠 뷰에는 버전이 없습니다. 오른쪽의 \"새 버전 공개\" 버튼을 사용하여 첫 번째 컨텐츠 뷰 버전을 생성합니다.","This Version cannot be deleted from some of the Lifecycle Environments due to one of the reasons below.":"이 버전은 다음과 같은 이유로 라이프사이클 환경의 일부에서 삭제할 수 없습니다. ","This Version is not associated with any Lifecycle Environments.":"이 버전은  라이프사이클 환경과 연결되어 있지 않습니다. ","This activation key is not associated with any content hosts.":"이 활성키는 다른 컨텐츠 호스트와 연결되어 있지 않습니다. ","This activation key may be used during system registration. For example:":"시스템 등록 도중 이 활성키를 사용할 수 있습니다. 예: ","This filter applies only to a subset of repositories in the content view.":"이 필터는 컨텐츠 뷰에 있는 리포지터리의 서브셋에만 적용합니다. ","This filter applies to all repositories in the content view (current and future).":"이 필터는 컨텐츠 뷰 (지금 및 나중)에 있는 모든 리포지터리에 적용됩니다. ","This subscription is not associated with any activation keys.":"이 서브스크립션은 활성키와 연결되어 있지 않습니다. ","This subscription is not attached to any content hosts.":"이 서브스크립션은 컨텐츠 호스트에 첨부되어 있지 않습니다. ","This version will be removed from:":"이 버전은 다음에서 삭제됩니다: ","Time":"시간 ","Title":"제목","To register a content host to this server, follow these steps.":"서버에 컨텐츠 호스트를 등록하려면 다음 단계를 실행합니다:","Topic":"주제","Type":"유형","URL":"URL","URL of the registry you want to sync. Example: https://registry.hub.docker.com":"동기화하고자 하는 레지스트리의 URL입니다. 예: https://registry.hub.docker.com","URL to discover":"검색할 URL ","UUID":"UUID","Unattach Sync Plan":"동기화 계획 첨부 해제 ","Unattaching...":"첨부 해제 중...","Unlimited Content Hosts:":"무제한 컨텐츠 호스트: ","Unregister Content Host":" 컨텐츠 호스트 등록 해제 ","Unregister Content Host \"{{contentHost.name}}\"?":"컨텐츠 호스트 \"{{contentHost.name}}\"을 등록 해제하시겠습니까?","Unregister Content Hosts":"컨텐츠 호스트 등록 해제 ","Unregister {{ table.numSelected  }} Content Hosts?":"{{ table.numSelected  }} 컨텐츠 호스트를 등록 해제하시겠습니까?","Update":"업데이트 ","Update All":"모두 업데이트 ","Update Repositories":"리포지터리 업데이트 ","Update Sync Plan":"동기화 계획 업데이트 ","Updated":"업데이트 일시 ","Updated errata filter -":"업데이트된 에라타 필터 - ","Updated override to \"%x\".":"\"%x\"로 덮어쓰기를 업데이트했습니다. ","Updating...":"업데이트 중...","Upload":"업로드 ","Upload Docker Image":"Docker 이미지 업로드 ","Upload GPG Key":"GPG 키 업로드 ","Upload New Manifest":"새 매니페스트를 업로드 ","Upload Package":"패키지 업로드 ","Upload Puppet Module":"Puppet 모듈 업로드 ","Uploading...":"업로드 중...","Upstream Repository Name":"업스트림 리포지터리 이름 ","Upstream Subscription":"업스트림 서브스크립션 ","Upstream Subscription Management Application":"업스트림 서브스크립션 관리 애플리케이션 ","Upstream Type":"업스트림 유형 ","Upstream User":"업스트림 사용자 ","Use Latest (currently %s)":"최신 (현재 %s) 사용 ","Use the cancel button on content view selection to revert your environment selection.":"컨텐츠 뷰 선택에서 취소 버튼을 사용하여 환경 선택을 취소합니다. ","User":"사용자 ","Version":"버전 ","Version %s":"버전 %s","Version 1":"버전 1 ","Version Deletion":"버전 삭제","Version Details":"버전 상세 정보 ","Version {{ version.version }}":"버전 {{ version.version }}","Version:":"버전: ","Versions":"버전 ","Versions for Content View:":"컨텐츠 뷰 버전: ","Versions for Puppet Module:":"Puppet 모듈 버전 ","View Details":"뷰 상세 정보 ","Virtual":"가상 ","Virtual Guest From":"가상 게스트 ","Virtual Guests":"가상 게스트 ","Virtual Host":"가상 호스트 ","Weekly on {{ product.sync_plan.sync_date | date:'EEEE' }} at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"매주 {{ product.sync_plan.sync_date | date:'EEEE' }} {{ product.sync_plan.sync_date | date:'mediumTime' }} (서버 시간)","When Auto Attach is disabled, registering systems will be attached to all associated subscriptions.":"자동 첨부를 비활성화할 때 등록된 시스템은 모든 연관된 서브스크립션에 첨부됩니다.","Working":"작업 중 ","Yes":"예  ","Yes (Default)":"예 (기본값) ","You cannot remove these repositories because they belong to a Red Hat product.":"이 리포지터리는 Red Hat 제품에 속해 있기 때문에 이를 삭제할 수 없습니다. ","You cannot remove these repositories because you do not have permission.":"권한이 없기 때문에 이 리포지터리를 삭제할 수 없습니다. ","You cannot remove this product because it is a Red Hat product.":"Red Hat 제품이기 때문에 이를 삭제할 수 없습니다. ","You cannot remove this product because it was published to a content view.":"컨텐츠 뷰에 공개되어 있기 때문에 이 제품을 삭제할 수 없습니다. ","You cannot remove this product because you do not have permission.":"권한이 없기 때문에 이 제품을 삭제할 수 없습니다. ","You cannot remove this repository because it is a Red Hat repository.":"Red Hat 리포지터리이기 때문에 이를 삭제할 수 없습니다. ","You cannot remove this repository because it was published to a content view.":"컨텐츠 보기에 공개되어 있기 때문에 이 리포지터리를 삭제할 수 없습니다. ","You cannot remove this repository because you do not have permission.":"권한이 없기 때문에 이 리포지터리를 삭제할 수 없습니다. ","You currently do not have any Packages installed. You can use Package Actions above this message to install new Packages.":"현재 설치되어 있는 패키지가 없습니다. 이 메세지에 있는 패키지 동작을 사용하여 새 패키지를 설치할 수 있습니다. ","You currently don't have any Activation Keys, you can add Activation Keys using the button on the right.":"현재 활성키가 없습니다. 오른쪽에 있는 버튼을 사용하여 활성키를 추가할 수 있습니다. ","You currently don't have any Content Hosts in this Host Group, you can add Content Hosts after selecting the 'Add' tab.":"현재 호스트 그룹에 컨텐츠 호스트가 없습니다. '추가' 탭을 선택하여 컨텐츠 호스트를 추가할 수 있습니다. ","You currently don't have any Content Hosts, you can create new Content Hosts by selecting Contents Host from main menu and then clicking the button on the right.":"현재 컨텐츠 호스트가 없습니다. 메인 메뉴에서 컨텐츠 호스트를 선택한 후 오른쪽에 있는 버튼을 클릭하면 새 컨텐츠 호스트를 생성할 수 있습니다. ","You currently don't have any Content Hosts, you can register one by clicking the button on the right and following the instructions.":"현재 컨텐츠 호스트가 없습니다. 오른쪽 버튼을 클릭하고 지시에 따라 컨텐츠 호스트를 등록할 수 있습니다.","You currently don't have any Content Views.  A Content View can be added by using the button on the right.":"현재 컨텐츠 뷰가 없습니다. 오른쪽 버튼을 사용하여 컨텐츠 뷰를 추가할 수 있습니다.","You currently don't have any Docker Tags.":"현재 Docker 태그가 없습니다.","You currently don't have any Filters included in this Content View, you can add a new Filter by using the button on the right.":"현재 이 컨텐츠 뷰에 필터가 포함되어 있지 않습니다. 오른쪽에 있는 버튼을 사용하여 새 필터를 추가할 수 있습니다. ","You currently don't have any Gpg keys, you can add Gpg keys using the button on the right.":"현재 GPG 키가 없습니다. 오른쪽에 있는 버튼을 사용하여 GPG 키를 추가할 수 있습니다. ","You currently don't have any Host Collections, you can add Host Collections using the button on the right.":"현재 호스트 컬렉션이 없습니다. 오른쪽에 있는 버튼을 사용하여 호스트 컬렉션을 추가할 수 있습니다. ","You currently don't have any Products associated with this Gpg Key.":"현재 GPG 키와 관련된 제품이 없습니다. ","You currently don't have any Products to subscribe to, you can add Products after selecting 'Products' under 'Content' in the main menu.":"현재 등록할 제품이 없습니다. 메인 메뉴의 '컨텐츠'에서 '제품'을 선택하여 제품을 추가할 수 있습니다. ","You currently don't have any Products<span bst-feature-flag=\"custom_products\">, you can add Products using the button on the right</span>.":"현재 제품 <span bst-feature-flag=\"custom_products\">이 없습니다. 오른쪽에 있는 버튼을 사용하여 제품을 추가할 수 있습니다</span>.","You currently don't have any Puppet Modules included in this Content View, you can add Puppet Modules using the button on the right.":"현재 컨텐츠 뷰에 Puppet 모듈이 포함되어 있지 않습니다. 오른쪽에 있는 버튼을 사용하여 Puppet 모듈을 추가할 수 있습니다. ","You currently don't have any Repositories associated with this Gpg Key.":"현재 GPG 키와 관련된 리포지터리가 없습니다. ","You currently don't have any Repositories included in this Product, you can add Repositories using the button on the right.":"현재 이 제품에 리포지터리가 포함되어 있지 않습니다. 오른쪽에 있는 버튼을 사용하여 리포지터리를 추가할 수 있습니다. ","You currently don't have any Subscriptions associated with this Content Host, you can add Subscriptions after selecting the 'Add' tab.":"현재 이 컨텐츠 호스트와 관련된 서브스크립션이 없습니다. '추가' 탭을 클릭하여 서브스크립션을 추가할 수 있습니다. ","You currently don't have any Subscriptions, you can add Subscriptions by importing a manifest using the button on the right labeled 'Manage Manifest'.":"현재 서브스크립션이 없습니다. 오른쪽에 있는 '매니페스트 관리' 버튼을 사용하여 매니페스트를 가져오기한 후 서브스크립션을 추가할 수 있습니다. ","You currently don't have any Sync Plan, you can add Sync Plans using the button on the right.":"동기화 일정이 없습니다. 오른쪽 버튼을 사용하여 동기화 일정을 추가할 수 있습니다. ","You currently don't have any Sync Plans.  A Sync Plan can be created by using the button on the right.":"현재 동기화 계획이 없습니다. 오른쪽 버튼을 사용하여 동기화 계획을 생성할 수 있습니다.","You do not have any Installed Products":"설치된 제품이 없습니다 ","You must select a content view in order to save your environment.":"환경을 저장하려면 컨텐츠 뷰를 선택해야 합니다.","You must select a new content view before your change of environment can be saved.":"변경한 환경을 저장하기 전 새 컨텐츠 뷰를 선택해야 합니다. ","You must select a new content view before your change of environment can be saved. Use the cancel button on content view selection to revert your environment selection.":"환경 변경 사항을 저장하기 전 새 컨텐츠 뷰를 선택해야 합니다. 컨텐츠 뷰 선택 취소 버튼을 사용하여 환경 선택을 취소합니다. ","You must select at least one Content Host in order to apply Errata.":"에라타를 적용하려면 최소 하나의 컨텐츠 호스트를 선택하셔야 합니다. ","You must select at least one repository.":"최소 하나의 리포지터리를 선택해야 합니다. ","Your selected content hosts have changed, it is highly recommended to refresh the calculated list of errata using the 'Refresh Errata' button below.":"선택한 컨텐츠 호스트가 변경되었습니다. 아래의 \"에라타 새로 고침\" 버튼을 사용하여 에라카의 계산된 목록을 새로 고침할 것을 권장합니다. ","Yum Content":"Yum 컨텐츠 ","Yum Metadata Checksum":"Yum 메타데이터 체크섬 ","Yum Repo Discovery":"Yum 리포지터리 검색 ","Yum Repositories":"Yum 리포지터리 ","Yum Repositories <div>{{ library.counts.yum_repositories || 0 }}</div>":"Yum 리포지터리 <div>{{ library.counts.yum_repositories || 0 }}</div>","Yum metadata generation has been initiated in the background.  Click <a href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Yum 메타데이터 생성이 백그라운드에서 시작되었습니다. <a href=\"{{ taskUrl() }}\">여기</a>를 클릭하여 진행 상태를 모니터하십시오. ","add some repositories.":"일부 리포지터리를 추가합니다. ","daily":"매일 ","here":"여기 ","hourly":"한 시간 마다 ","or authenticate with a username and password:":"또는 사용자 이름 및 암호로 인증:","subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"":"subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"","task details":"작업 상세 정보 ","to manage them individually.":"개별적으로 관리 ","true":"true","weekly":"매주 ","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }} 선택됨 ","{{ addSubscriptionsTable.numSelected }} Selected":"{{ addSubscriptionsTable.numSelected }} 선택됨 ","{{ contentHost.registered_by }}":"{{ contentHost.registered_by }}","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }} 선택됨 ","{{ contentView.name }} cannot be deleted as one or more Content View Versions are still promoted to a Lifecycle Environment.\n      Each Content View Version must be removed from their Lifecycle Environments before the Content View can be deleted.":"하나 이상의 컨텐츠 뷰 버전이 라이프 사이클 환경으로 승격되고 있기 때문에 {{ contentView.name }}을 삭제할 수 없습니다.\n  컨텐츠 뷰를 삭제하기 전 각 컨텐츠 뷰 버전을 라이프 사이클 환경에서 삭제해야 합니다.","{{ contentView.puppet_modules.length || 0 }} Puppet Modules":"{{ contentView.puppet_modules.length || 0 }} Puppet 모듈 ","{{ contentView.repositories.length || 0 }} Repositories":"{{ contentView.repositories.length || 0 }} 리포지터리 ","{{ contentViewVersion.puppet_module_count || 0 }} Puppet Modules":"{{ contentViewVersion.puppet_module_count || 0 }} Puppet 모듈","{{ contentViewVersion.repositories.length || 0 }} Repositories":"{{ contentViewVersion.repositories.length || 0 }} 리포지터리 ","{{ discoveryTable.numSelected }} Selected":"{{ discoveryTable.numSelected }} 선택됨 ","{{ environment.name }} Environment":"{{ environment.name }} 환경 ","{{ errata.hosts_applicable_count || 0 }} Applicable,":"{{ errata.hosts_applicable_count || 0 }} 적용 가능,","{{ errata.hosts_available_count || 0 }} Installable":"{{ errata.hosts_available_count || 0 }} 설치 가능 ","{{ hostCollectionsTable.numSelected }} Selected":"{{ hostCollectionsTable.numSelected }} 선택됨 ","{{ package.removeTask.affected_units }} package removed":"{{ package.removeTask.affected_units }} 패키지 삭제됨 ","{{ package.removeTask.affected_units }} packages removed":"{{ package.removeTask.affected_units }} 패키지 삭제됨 ","{{ product.last_sync_words }} Ago ({{ product.sync_status.finish_time | date:'medium' }} Local Time)":"{{ product.last_sync_words }} 전 ({{ product.sync_status.finish_time | date:'medium' }} 현지 시간)","{{ product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ product.sync_plan.next_sync | date:'medium' }} (서버 시간)","{{ productsTable.numSelected }} Selected":"{{ productsTable.numSelected }} 선택됨 ","{{ repository.content_counts.docker_image || 0 }} Docker Images":"{{ repository.content_counts.docker_image || 0 }} Docker 이미지 ","{{ repository.content_counts.docker_image }} Docker Images":"{{ repository.content_counts.docker_image }} Docker 이미지 ","{{ repository.content_counts.docker_tag || 0 }} Docker Tags":"{{ repository.content_counts.docker_tag || 0 }} Docker 태그 ","{{ repository.content_counts.docker_tag }} Docker Tags":"{{ repository.content_counts.docker_tag }} Docker 태그 ","{{ repository.content_counts.erratum || 0 }} Errata":"{{ repository.content_counts.erratum || 0 }} 에라타 ","{{ repository.content_counts.erratum }} Errata":"{{ repository.content_counts.erratum }} 에라타 ","{{ repository.content_counts.puppet_module || 0 }} Puppet Modules":"{{ repository.content_counts.puppet_module || 0 }} Puppet 모듈 ","{{ repository.content_counts.rpm || 0 }} Packages":"{{ repository.content_counts.rpm || 0 }} 패키지 ","{{ repository.content_counts.rpm }} Packages":"{{ repository.content_counts.rpm }} 패키지 ","{{ repository.last_sync_words }} Ago ({{ repository.last_sync.ended_at | date:'medium' }} Local Time)":"{{ repository.last_sync_words }} 전 ({{ repository.last_sync.ended_at | date:'medium' }} 현지 시간)","{{ repository.last_sync_words }} ago":"{{ repository.last_sync_words }} 전 ","{{ repository.product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ repository.product.sync_plan.next_sync | date:'medium' }} (서버 시간)","{{ subscription.host.name }}":"{{ subscription.host.name }}","{{ subscriptionsTable.numSelected }} Selected":"{{ subscriptionsTable.numSelected }} 선택됨 ","{{ totalActivationKeyCount() }} Activation Keys will be moved to {{ deleteOptions.activationKeys.contentView.name }} in {{ deleteOptions.activationKeys.environment.name }}":"{{ totalActivationKeyCount() }} 활성키가 {{ deleteOptions.activationKeys.environment.name }}에 있는 {{ deleteOptions.activationKeys.contentView.name }}으로 이동합니다. ","{{ totalHostCount() }} Content Hosts will be moved to {{ deleteOptions.contentHosts.contentView.name }} in {{ deleteOptions.contentHosts.environment.name }}":"{{ totalHostCount() }} 컨텐츠 호스트가 {{ deleteOptions.contentHosts.environment.name }}에 있는 {{ deleteOptions.contentHosts.contentView.name }}로 이동합니다. ","{{ type.display }}":"{{ type.display }}","{{ version.docker_image_count }} Docker Images":"{{ version.docker_image_count }} Docker 이미지 ","{{ version.docker_tag_count }} Docker Tags":"{{ version.docker_tag_count }} Docker 태그 ","{{ version.errata_counts.total }} Errata":"{{ version.errata_counts.total }} 에라타 ","{{ version.package_count }} Packages":"{{ version.package_count }} 패키지 ","{{ version.puppet_module_count }} Puppet Modules":"{{ version.puppet_module_count }} Puppet 모듈 "});
    gettextCatalog.setStrings('pt-BR', {"%(consumed)s out of %(quantity)s":"%(consumed)s entre %(quantity)s","(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)":"(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)","(Not all Activation Keys editable )":"(Nem todas as Chaves de Ativação são editáveis)","(Not all Content Hosts editable )":"(Nem todas as Máquinas de Conteúdo são editáveis)","* These marked Content View Versions are from Composite Content Views.  Their components needing updating are listed underneath.":"*Estas Versões de Visualização de Conteúdo marcadas, são de Visualizações de Conteúdo de Composição. Seus componentes que precisam ser atualizados são listados abaixo.","+ New Sync Plan":"+ Novo Plano de Sinc","- Date and Type":"- Data e Tipo","-- select an interval --":"-- selecione um intervalo --","1 Content Host":["1 Host de Conteúdo","{{ contentHost.virtual_guests.length }} Hosts de Conteúdo "],"1 repository sync has errors.":["1 sincronização de repositório apresenta erros.","{{ product.sync_summary.error}} sincronizações de repositório apresentam erros."],"1 repository sync in progress.":["1 sincronização de repositório está em desenvolvimento.","{{ product.sync_summary.error}} sincronizações de repositório está em desenvolvimento."],"1 successfully synced repository.":["1 repositório sincronizado com sucesso.","{{ product.sync_summary.success}} repositórios sincronizados com sucesso."],"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Version {{ version.version }}</a>":"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Versão {{ version.version }}</a>","<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Create":"<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Criar","<span translate=\"\">\n        If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">here</a>\n      <span translate=\"\">to manage them individually.</span>":"<span translate=\"\">\n       Caso você prefira mover as partes destas Chaves de Ativação para uma visualização de conteúdo diferente ou clique em Ambientes de Ciclo de Vida\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">aqui</a>\n      <span translate=\"\">para gerenciá-los individualmente.</span>","A composite view contains other content views.":"Uma visualização da composição contém outras visualizações de conteúdo.","A new version of {{ contentView.name }} will be created and promoted to the Library environment.\n    It can be promoted to other environments from the Versions tab of this Content View.":"Uma nova versão do {{ contentView.name }}  será criada e promovida para o ambiente da Biblioteca.    \nEle pode ser promovido para outros ambientes da aba Versões desta Visualização de Conteúdo.","Account":"Conta","Account Number":"Número da Conta","Action":"Ação","Action Type":"Tipo de Ação","Actions":"Ações","Actions for:":"Ação para:","Activation Key":["Chave de Ativação","Chaves de Ativação"],"Activation Key Content":"Conteúdo de Chave de Ativação","Activation Key Type:":"Tipo de Chave de Ativação:","Activation Key removed.":"Chave de Ativação removida.","Activation Key updated":"Chave de Ativação atualizada","Activation Key:":"Chave de Ativação:","Activation Key: {{ activationKey.name }}":"Chave de Ativação:  {{ activationKey.name }}","Activation Keys":"Chaves de Ativação","Activation Keys for Subscription:":"Chaves de Ativação para Subscrição","Activation Keys using Version {{ version.version }}":"Chaves de Ativação usando a Versão  {{ version.version }}","Add":"Adicionar","Add Content Hosts to:":"Adicionar Hosts de Conteúdo ao:","Add Content Views":"Adicionar Visualizações de Conteúdo","Add Errata":"Adicionar Errata","Add New Environment":"Adicione Novo Ambiente","Add New Filter":"Adicionar Novo Filtro","Add New Module":"Adicionar Novo","Add New Repository":"Adicionar Novo Repositório","Add Package Group":"Adicionar Grupo de Pacote","Add Repositories":"Adicionar Repositórios","Add Selected":"Adicionar Selecionados","Add Subscriptions for Activation Key:":"Adicionar Subscrições para Chave de Ativação:","Add To":"Adicionar ao","Added %x host collections to activation key \"%y\".":"Foram adicionadas coleções de host %x para ativar a chave \"%y\".","Added %x host collections to content host \"%y\".":"Foram adicionadas coleções de host %x para host de conteúdo \"%y\".","Added %x products to sync plan \"%y\".":"Foram adicionados produtos %x para plano de sincronização \"%y\".","Adding Lifecycle Environment to the end of \"{{ priorEnvironment.name }}\"":"Adicionando Ambiente de Ciclo de Vida ao final de \"{{ priorEnvironment.name }}\"","Advanced Information":"Informações avançadas","Advisory":"Relatório","Affected Activation Keys":"Chaves de Ativação Afetadas","Affected Content Hosts":"Hosts de Conteúdo Afetados","Affected Packages":"Pacotes Afetados","Affected Systems":"Sistemas Afetados","Affected repositories":"Repositórios afetados","Affected repositories have been updated.":"Repositórios afetados foram atualizados.","All Content Views":"Todas as Visualizações de Conteúdo","All Products":"Todos os Produtos","All Repositories":"Todos os Repositórios","All Versions":"Todas as Versões","Alter Sync Plans":"Alterar Planos de Sinc.","An error occurred applying Subscriptions:":"Ocorreu um erro ao aplicar as Subscrições:","An error occurred removing the Activation Key:":"Ocorreu um erro ao remover a Chave de Ativação:","An error occurred removing the Host Collection:":"Ocorreu um erro ao remover a Coleção de Host:","An error occurred removing the Products:":"Ocorreu um erro ao remover os Produtos:","An error occurred removing the content hosts.":"Ocorreu um erro durante a remoção de hosts de conteúdo.","An error occurred removing the environment:":"Ocorreu um erro ao remover o ambiente.","An error occurred removing the subscriptions.":"Ocorreu um erro ao remover as subscrições.","An error occurred saving the Activation Key:":"Ocorreu um erro ao salvar a Chave de Ativação:","An error occurred saving the Content Host:":"Ocorreu um erro ao salvar o Host de Conteúdo:","An error occurred saving the Environment:":"Ocorreu um erro ao salvar o Ambiente:","An error occurred saving the Filter:":"Ocorreu um erro ao salvar o Filtro:","An error occurred saving the Host Collection:":"Ocorreu um erro ao salvar a Coleção de Host:","An error occurred saving the Product:":"Ocorreu um erro ao salvar o Produto:","An error occurred saving the Repository:":"Ocorreu um erro ao salvar o Repositório:","An error occurred saving the Sync Plan:":"Ocorreu um erro ao salvar o Plano de Sinc:","An error occurred saving the URL:":"Ocorreu um erro ao salvar a URL:","An error occurred trying to auto-attach subscriptions.  Please check your log for further information.":"Ocorreu um erro ao tentar auto-conectar subscrições. Por favor, verifique seu log para mais informações.","An error occurred updating the Content View:":"Ocorreu um erro ao atualizar a Visualização de Conteúdo:","Applicable":"Aplicável","Applicable Content Hosts for Errata":"Hosts de Conteúdo Aplicáveis para Errata","Applicable Errata":"Errata Aplicável","Applicable hosts for {{ erratum.errata_id }}":"Hosts aplicáveis para  {{ erratum.errata_id }}","Apply":"Aplicar","Apply Errata":"Aplicar Errata","Apply Errata to Content Host \"{{contentHost.name}}\"?":"Aplicar Errata ao Host de Conteúdo  \"{{contentHost.name}}\"?","Apply Errata to Content Hosts":"Aplicar Errata aos Hosts de Conteúdo","Apply Errata to Content Hosts immediately after publishing.":"Aplicar Errata aos Hosts de Conteúdo imediatamente após publicar.","Apply Selected":"Aplicar Selecionado","Apply to Content Hosts":"Aplicar aos Hosts de Conteúdo","Apply to Hosts":"Aplicar aos Hosts","Apply {{ errata.errata_id }}":"Aplicar {{ errata.errata_id }}","Apply {{ errata.errata_id }} to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"Aplicar {{ errata.errata_id }} ao {{ numberOfContentHostsToUpdate }} Host(s) de Conteúdo?","Apply {{ errataIds.length }} Errata to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"Aplicar {{ errataIds.length }} Errata to {{ numberOfContentHostsToUpdate }} Hosts de Conteúdo(s)?","Applying":"Aplicando","Applying Bulk Errata Update":"Aplicando Atualização da Errata em Massa","Arch":"Arq","Architecture(s)":"Arquitetura(s)","Archived Copy":"Cópia Arquivada","Are you sure you want to add the {{ table.numSelected }} content host(s) selected to the host collection(s) chosen?":"Você tem certeza de que deseja adicionar o(s) host(s) de conteúdo(s) {{ table.numSelected }} selecionado(s) à(s) coleção(ões) escolhida(s)?","Are you sure you want to apply Errata to content host \"{{ contentHost.name }}\"?":"Você tem certeza de que deseja aplicar Errata ao host de conteúdo \"{{ contentHost.name }}\"?","Are you sure you want to apply the {{ detailsTable.numSelected }} selected errata to the content hosts chosen?":"Você tem certeza de que deseja aplicar a errata selecionada {{ detailsTable.numSelected }} aos host de conteúdo escolhido(s)?","Are you sure you want to assign the {{ table.numSelected }} content host(s) selected to {{ selected.contentView.name }} in {{ selected.environment.name }}?":"Você tem certeza de que deseja atribuir o(s) host(s) de conteúdo(s) {{ table.numSelected }} selecionado(s) ao(s) {{ selected.contentView.name }} em {{ selected.environment.name }}?","Are you sure you want to auto-attach available subscriptions to all {{ table.resource.total }} content hosts?":"Você tem certeza de que deseja auto-anexar as subscrições disponíveis a todos os hosts de conteúdo {{ table.resource.total }} ?","Are you sure you want to delete Content View \"{{ contentView.name }}\"?":"Você tem certeza de que deseja remover a Visualização de Conteúdo \"{{ contentView.name }}\"?","Are you sure you want to install {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"Você tem certeza de que deseja instalar o {{ content.content }} no(s) sistema(s) {{ getSelectedSystemIds().length }} selecionado(s)?","Are you sure you want to remove Activation Key \"{{ activationKey.name }}\"?":"Você tem certeza de que deseja remover a Chave de Ativação \"{{ activationKey.name }}\"?","Are you sure you want to remove GPG key {{ gpgKey.name }}?":"Você tem certeza de que deseja remover a chave GPG {{ gpgKey.name }}?","Are you sure you want to remove Host Collection \"{{ hostCollection.name }}\"?":"Você tem certeza de que deseja remover o Host Collection \"{{ hostCollection.name }}\"?","Are you sure you want to remove Sync Plan \"{{ syncPlan.name }}\"?":"Você tem certeza de que deseja remover o Sync Plan \"{{ syncPlan.name }}\"?","Are you sure you want to remove product \"{{ product.name }}\"?":"Você tem certeza de que deseja remover o produto \"{{ product.name }}\"?","Are you sure you want to remove repository \"{{ repository.name }}\"?":"Você tem certeza de que deseja remover repositório \"{{ repository.name }}\"?","Are you sure you want to remove the {{ productTable.getSelected().length }} products(s) selected?":"Você tem certeza de que deseja remover o(s) produto(s) {{ productTable.getSelected().length }} selecionado(s)?","Are you sure you want to remove the {{ table.numSelected }} content host(s) selected from the host collection(s) chosen?":"Você tem certeza de que deseja remover o(s) host(s) de conteúdo(s) {{ table.numSelected }} content host(s) selecionado(s) a partir da(s) coleção (ões) escolhida(s)?","Are you sure you want to remove {{ content.content }} from the {{ getSelectedSystemIds().length }} system(s) selected?":"Você tem certeza de que deseja remover o {{ content.content }} do(s) sistema(s) {{ getSelectedSystemIds().length }} selecionado(s)?","Are you sure you want to unregister content host \"{{ contentHost.name }}\"?":"Você tem certeza de que deseja cancelar o registro do host de conteúdo \"{{ contentHost.name }}\"?","Are you sure you want to unregister the {{ table.numSelected }} content host(s) selected?":"Você tem certeza de que deseja cancelar o registro do host de conteúdo {{ table.numSelected }} selecionado?","Are you sure you want to update {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"Você tem certeza de que deseja atualizar o {{ content.content }} no(s) sistema(s) {{ getSelectedSystemIds().length }} selecionado(s)?","As part of this deletion, 1 Content View Version will be deleted.":["Como parte desta remoção, a Versão de Visualização de Conteúdo 1 será removida.","Como parte desta remoção, as Versões de Visualização de Conteúdo {{ versions.length }} serão removidas."],"Assign":"Atribuir","Assign Environment and Content View":"Atribuir Ambiente e Visualização de Conteúdo","Associations":"Associações","At least one Errata needs to be selected to Apply.":"Ao menos uma Errata precisa ser selecionada para Aplicar.","At least one content host needs to be selected in order to perform a bulk action.":"Ao menos um host de conteúdo precisa ser selecionado para realizar uma ação em massa.","At least one product needs to be selected in order to perform a bulk action.":"Ao menos um produto precisa ser selecionado para realizar uma ação em massa.","Attached":"Conectado","Attached to Content Hosts":"Anexado aos Hosts de Conteúdo","Author":"Autor","Auto-Attach":"Auto-conectar","Auto-attach available subscriptions to all content hosts":"Subscrições Auto-Attach diponíveis à todos os hosts de conteúdo.","Automatic":"Automático","Available Content Views for Composite Content View:":"Visualizações de Conteúdo Disponíveis para a Visualização de Conteúdo de Composição:","Available Puppet Modules for Content View:":"Módulos Puppet Disponíveis para a Visualização de Conteúdo:","Back":"Retornar","Back To Errata List":"Retornar à Lista de Errata","Back To Repository Selection":"Retornar à Seleção de Repositório","Back to Module List":"Retornar à Lista de Módulo","Back to Repository Details":"Retornar aos Detalhes de Repositório","Back to Repository List":"Retornar à Lista de Repositório","Back to Version List":"Retornar à Lista de Versão","Basic Information":"Informações Básicas","Below are the product's repository content sets currently available for this subscription. For Red Hat subscriptions, additional content can be made available through the":"Abaixo seguem os conjuntos de conteúdo de repositórios de produtos disponíveis atualmente para esta subscrição. Para as subscrições da Red Hat, o conteúdo adicional pode ser disponibilizado através do","Below are the repository content sets currently available for this activation key through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"Abaixo seguem os conjuntos de conteúdo de repositórios disponíveis atualmente para esta chave de ativação através de sua subscrição. Para as subscrições da Red Hat, o conteúdo adicional pode ser disponibilizado através do","Below are the repository content sets currently available for this content host through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"Abaixo seguem os conjuntos de conteúdo de repositórios disponíveis atualmente para este host de conteúdo através de sua subscrição. Para as subscrições da Red Hat, o conteúdo adicional pode ser disponibilizado através do","Bug Fix":"Reparo de Erro","Bug Fix Advisory":"Relatório de Conserto de Bug","Bugfix":"Bugfix","Bulk Actions":"Ações em Massa","Bulk Alter Products Sync Plan":"Plano de Sincronização dos Produtos de Alteração em Massa ","Bulk Sync Products":"Produtos de Sinc em Massa","CVEs":"CVEs","Cancel":"Cancelar","Capacity":"Capacidade","Change assigned Lifecycle Environment or Content View":"Mudar Ambiente de Ciclo de Vida atribuído ou Visualização de Conteúdo","Changing default settings for content hosts that register with this activation key requires subscription-manager version 1.10 or newer to be installed on that host.":"Mudar configurações padrão para hosts de conteúdo que registram com esta chave de ativação, requer a versão 1.10 do subscription-manager ou mais recente instalada neste host.","Changing default settings requires subscription-manager version 1.10 or newer to be installed on this host.":"Mudar as configurações padrão requer uma versão 1.10 do subscription-manager ou mais recente instalado neste host.","Checkin":"Checkin","Checksum":"Checksum","Choose a lifecycle environment from the available promotion paths.":"Escolha um ambiente de ciclo de vida a partir dos caminhos de promoção disponíveis.","Click here to check the status of the task.":"Clique aqui para verificar o estado da tarefa.","Click here to refresh.":"Clique aqui para atualizar.","Click here to see the task for the successful removals.":"Clique aqui para ver a tarefa para remoções bem sucedidas.","Click here to select Errata for an Incremental Update.":"Clique aqui para selecionar Errata para uma Atualização Adicional.","Close":"Fechar","Collection Actions":"Ações de Coleção","Completely remove version?":"Remover completamente a versão?","Components":"Componentes","Components:":"Componentes:","Composite Content View {{ contentView.name }}":"Visualização de Conteúdo de Composição {{ contentView.name }}?","Composite View":"Visualização de Composição","Composite View?":"Visualização de Composição?","Composite?":"Composição?","Confirm":"Confirmação","Confirm Remove":"Confirmar Remover","Confirm Version Removal: Version {{ version.version }}":"Confirmar Remoção da Versão: Versão  {{ version.version }}","Confirmation":"Confirmação","Consumed":"Consumido","Content":"Conteúdo","Content Counts":"Contas de Conteúdo","Content Download URL":"URL de Download do Conteúdo","Content Host":"Host de Conteúdo","Content Host %s has been deleted.":"Host de Conteúdo %s foi removido.","Content Host Bulk Actions":"Ações em Massa de Host de Conteúdo","Content Host Bulk Content":"Conteúdo em Massa de Host de Conteúdo","Content Host Bulk Errata":"Errata em Massa de Host de Conteúdo","Content Host Bulk Host Collections":"Conteúdo em Massa de Coleções de Host","Content Host Bulk Packages":"Pacotes em Massa de Host de Conteúdo","Content Host Bulk Subscriptions":"Subscrições em Massa de Host de Conteúdo","Content Host Content":"Conteúdo de Host de Conteúdo","Content Host Counts":"Contas de Host de Conteúdo","Content Host Errata Management":"Gerenciamento de Errata do Host de Conteúdo","Content Host Limit":"Limite do Host de Conteúdo","Content Host Package Management":"Gerenciamento de Pacotes do Host de Conteúdo","Content Host Properties":"Propriedades do Host de Conteúdo","Content Host Registration":"Registro de Host de Conteúdo","Content Host Status":"Status do Host de Conteúdo","Content Host Subscriptions":"Subscrições de Host de Conteúdo","Content Host {{ contentHost.name }}":"Host de Conteúdo {{ contentHost.name }}","Content Host:":"Host de Conteúdo:","Content Hosts":"Hosts de Conteúdo","Content Hosts for Activation Key:":"Hosts de Conteúdo para Chave de Ativação:","Content Hosts for Subscription:":"Hosts de Conteúdo para Subscrição:","Content Hosts for:":"Hosts de Conteúdo para:","Content Hosts using Version {{ version.version }}":"Hosts de Conteúdo usando a Versão  {{ version.version }}","Content Type":"Tipo de Conteúdo","Content View":"Visualização de Conteúdo","Content View Name":"Nome de visualização de conteúdo","Content View Version":"Versão de Visualização de Conteúdo","Content View updated.":"Visualização de Conteúdo atualizada","Content View:":"Visualização de Conteúdo:","Content Views":"Visualização de Conteúdo","Content Views <div>{{ library.counts.content_views || 0 }}</div>":"Visualizações de Conteúdo <div>{{ library.counts.content_views || 0 }}</div>","Content Views for Composite Content View:":"Visualizações de Conteúdo para a Visualização de Conteúdo de Composição:","Content successfully uploaded":"O  Conteúdo foi carregado com êxito","Contents":"Conteúdos","Contract":"Contrato","Contract Number":"Número de Contrato","Copy":"Copiar","Copy Activation Key":"Copiar Chaves de Ativação","Copy Collection":"Copiar Coleção","Copy View":"Copiar Visualização","Cores per Socket":"Núcleos para Soquete","Cores: %s":"Núcleos: %s","Create":"Criar","Create New View":"Criar Nova Visualização","Create Repositories":"Criar repositórios","Create Repositories Within:":"Criar Repositórios Dentro:","Create Repository":"Criar Repositório","Create Selected":" Criar selecionado","Create Sync Plan":"Criar Plano de Sinc","Create new view":"Criar Nova Visualização","Create view from existing views":"Criar Visualizações de visualizações Existentes","Critical":"Crítico","Current Environment (%e/%cv)":"Ambiente Atual (%e/%cv)","Current Subscriptions for Activation Key:":"Subscrições atuais para Chave de Ativação:","Currently Selected Puppet Modules":"Módulos Puppet Selecionado Atualmente","Daily at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"Diariamente em  {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)","Date":"Data","Default":"Padrão","Delete":"Remover","Delete Manifest":"Remover Manifesto","Delete Version {{ version.version }}":"Remover Versão {{ version.version }}","Delete {{ contentView.name }}":"Remover  {{ contentView.name }}","Deleted from %s":"Remover de %s","Deleting Manifest...":"Removendo Manifesto...","Deleting from %count environments.":"Removendo de ambientes %count","Deleting from 1 environment.":"Removendo de 1 ambiente","Deletion from %s":"Remoção de %s","Description":"Descrição","Detail":"Detalhe","Details":"Detalhes","Details for Activation Key:":"Detalhes para a Chave de Ativação:","Details for Content View:":"Detalhes para Visualização de Contéudo:","Details for Product:":"Detalhes para Produto:","Discover":"Descoberta","Discovered URLs":"URLs descobertas","Docker Content":"Conteúdo do Docker","Docker Images":"Imagens Docker","Docker Images in {{ repository.name }}":"Imagens de Docker em {{ repository.name }}","Docker Repositories":"Repositórios do Docker","Docker Repositories <div>{{ library.counts.docker_repositories || 0 }}</div>":"Repositórios do Docker <div>{{ library.counts.docker_repositories || 0 }}</div>","Docker Tag:":"Docker Tag:","Docker Tags":"Docker Tags","Docker metadata generation has been initiated in the background.  Click\n        <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Geração de metadados de Docker foi iniciado no pano de fundo. Clique \nem  <a ng-href=\"{{ taskUrl() }}\">Here</a> para monitorar o progresso.","Edit":"Editar","Enabled?":"Habilitado?","End Date":"Data Final","Ends":"Termina","Enhancement":"Melhoria","Enter Package Group Name(s)...":"insira um Nome de Grupo de Pacote...","Enter Package Name(s)...":"Insira Nome(s) de Pacote","Environment":"Ambiente","Environment saved":"Ambiente Salvo","Environment: {{ environment.name }}":"Ambiente: {{ environment.name }}","Environments":"Ambientes","Environments List":"Lista de Ambientes","Environments:":"Ambientes:","Equal To":"Igual a","Errata":"Errata","Errata <div>{{ library.counts.errata.total || 0 }}</div>":"Errata <div>{{ library.counts.errata.total || 0 }}</div>","Errata Details":"Detalhes da Errata","Errata Filter:":"Filtro da Errata:","Errata ID":"ID de Errata ","Errata Installation":"Instalação da Errata","Errata List":"Lista de Errata","Errata Task List":"Lista de tarefas de Errata","Errata Tasks":"Tarefas de Errata","Errata Type":"Tipo de Errata","Errata for:":"Errata para:","Errata successfully added.":"Errata adicionada com sucesso.","Errata successfully removed.":"Errata foi removida com sucesso.","Errata:":"Errata:","Erratum - by Date and Type":"Erratum - por Data e Tipo","Erratum - by ID":"Erratum - por ID","Erratum Date Range":"Classe da Data da Errata","Error deleting manifest.":"Erro ao remover manifesto.","Error during upload:":"Erro durante a carga:","Error importing manifest.":"Erro ao importar manifesto.","Error refreshing manifest.":"Erro ao atualizar manifesto.","Event":"Evento","Events":"Eventos","Exclude":"Excluir","Existing Product":"Produto Existente","Expires":"Expira","Failed":"Falhou","File too large. Please use the CLI instead.":"Arquivo muito grande. Por favor use o CLI.","Filter":"Filtro","Filter Saved":"Filtro salvo","Filter information for:":"Informações de filtro para:","Filter...":"Filtro...","Filter:":"Filtro:","Filters":"Filtros","Filters List":"Lista de filtros","Filters for Content View:":"Filtros para Visualização de Conteúdo","Filters successfully removed.":"Filtros foram removidos com sucesso.","Finished At":"Termino em","For older operating systems such as Red Hat Enterprise Linux 5 or CentOS 5 it is recommended to use sha1.":"Para sistemas operacionais como o Red Hat Enterprise Linux 5 ou CentOS 5 recomendamos que utilize o sha1.","GPG Key":"Chave GPG","GPG Key URL":"URL de Chave de GPG","GPG Key successfully uploaded":"Chave GPG foi carregada com sucesso","GPG Key {{ gpgKey.name }}":"Chave GPG {{ gpgKey.name }}","GPG Keys":"Chaves GPG","General Information":"Informações Gerais","Generated On":"Gerar em","Gpg Key updated":"Chave GPG atualizada","Gpg Keys":"Chaves GPG","Greater Than":"Maior que","Group Install":"Instalação de Grupo","Group Remove":"Remover Grupo","Guests of":"Sistemas Virtuais de","Hide affected Activation Keys":"Esconder as Chaves de Ativação Afetadas","Hide affected Content Hosts":"Esconder Hosts de Conteúdo Afetados","History":"Histórico","History for Content View:":"Histórico para Visualização de Conteúdo:","Host Collection Management":"Gerenciamento da Coleção de Host","Host Collection Membership":"Inscrição de Coleção do Host","Host Collection Membership Management":"Gerenciamento da Inscrição da Coleção do Host","Host Collection removed.":"Coleção de host removida.","Host Collection updated":"Coleção do host atualizada","Host Collection:":"Coleção do host:","Host Collection: {{ hostCollection.name }}":"Coleção do Host: {{ hostCollection.name }}","Host Collections":"Coleções de Máquina","Host Collections for:":"Coleções de Máquina para:","Host Count":"Contagem de Host","Host Group":"Grupo de Host","Hostname":"Hostname","Hourly at {{ product.sync_plan.sync_date | date:'m' }} minutes and {{ product.sync_plan.sync_date | date:'s' }} seconds":"A cada hora em {{ product.sync_plan.sync_date | date:'m' }} minutos e {{ product.sync_plan.sync_date | date:'s' }} segundos","IPv4 Address":"Endereço IPv4","IPv6 Address":"Endereço IPv6","Id":"Id","If you would prefer to move portions of these Activation Keys to different Content Views or Lifecycle Environments click <a href=\"{{ activationKeyLink() }}\">here</a> to manage them individually.":"se você preferir transferir partes destas Chaves de Ativação para visualizações de conteúdo diferentes ou Ambientes de Ciclo de Vida clique em   <a href=\"{{ activationKeyLink() }}\">here</a> para gerenciá-los individualmente.","If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click":"se você preferir transferir partes destas Chaves de Ativação para visualizações de conteúdo diferentes ou Ambientes de Ciclo de Vida clique em ","If you would prefer to move portions of these Content Hosts to different content views or environments click <a href=\"{{ contentHostsLink() }}\">here</a>  to manage these Content Hosts in bulk.":"se você preferir transferir partes destas Hosts de Conteúdo para visualizações de conteúdo diferentes ou ambientes clique em  <a href=\"{{ contentHostsLink() }}\">here</a>  para gerenciar estes Hosts de conteúdo em massa.","Image":"Imagem","Image ID":"ID de Imagem","Import History":"Importar Histórico","Important":"Importante","In order to add a content view to a composite view you must first publish an initial version of the content view.":"Para adicionar uma visualização de conteúdo a uma visualização de composição você deve primeiro publicar uma versão inicial de visualização de conteúdo.","Include":"Incluir ","Included in Activation Keys":"Incluído em Chaves de Ativação","Incremental Update":"Atualização Incremental","Install":"Instalar","Install Selected":"Instalar Selecionado","Install katello-agent for remote actions and displaying errata information:":"Instalar katello-agent para ações remotas e exibição de informações de errata:","Install the pre-built bootstrap RPM:":"Instalar o RPM do bootstrap pré-construído:","Installable":"Instaláveis","Installable Errata":"Errata Instalável","Installed":"Instalado","Installed Package":"Pacote instalado","Installed Packages":"Pacotes Instalados","Installed Products":"Produtos Instalados","Instance-based":"Baseado em instância","Interfaces":"Interfaces","Interval":"Intervalo","Issued":"Emitido","Katello Agent":"Agente do Katello ","Label":"Rótulo","Last Checkin":"Último Checkin","Last Published":"Último Publicado","Last Puppet Report":"Último Relatório de Puppet","Last Sync":"Última Sinc","Last Updated":"Atualizado Em","Last Updated On":"Atualizado em","Last synced {{ product.last_sync_words }} ago.":"Última sincronização  {{ product.last_sync_words }} atrás.","Latest (Currently %s)":"Mais recente (Atualmente %s)","Less Than":"Menor do que","Library":"Biblioteca","Library Synced Content":"Conteúdo de Biblioteca Sincronizado ","Lifecycle Environment \"{{ priorEnvironment.name }}\" exists within the middle of a path. New Lifecycle Environments can only be added to the end of a path.":"Ambiente de Ciclo de Vida \"{{ priorEnvironment.name }}\" existe dentro do meio de um caminho. Ambientes Novos de Ciclo de Vida só podem ser adicionados ao final de um caminho.","Lifecycle Environment Paths":"Caminho de Ambiente de Ciclo de Vida","Lifecycle Environments":"Ambientes de Ciclo de Vida","Limit":"Limite","Limit:":"Limite:","Limits":"Limites","List/Remove":"Listar/Remover","Loading Environment Paths...":"Carregando Caminhos de Ambiente...","Loading...":"Carregando...","Manage Docker Images":"Gerenciar Imagens do Docker","Manage Docker Images for Repository:":"Gerenciar Imagens de Docker para o Repositório:","Manage Manifest":"Gerenciar Manifesto","Manage Packages":"Gerenciar Pacotes","Manage Packages for Repository:":"Gerenciar Pacotes para repositório","Manage Puppet Modules":"Gerenciar Módulos do Puppet","Manage Puppet Modules for Repository:":"Gerenciar Módulos do Puppet para Repositório:","Manifest History":"Histórico do Manifesto","Manifest successfully deleted.":"Manifesto removido com sucesso.","Manifest successfully imported.":"Manifesto importado com sucesso","Manifest successfully refreshed.":"Manifesto atualizado com sucesso.","Message":"Mensagem","Model":"Modelo","Moderate":"Moderada","Module %s removed from Content View.":"Foi removido o módulo %s da visualização de conteúdo","Module List":"Lista de Módulo","More Details":"Mais detalhes","More...":"Mais...","Multi-entitlement":"Direitos à Serviços Múltiplos","N/A":"Não está disponível","Name":"Nome","Name of the upstream repository you want to sync. Example: 'busybox' or 'fedora/ssh'.":"Nome do repositório de upstream que deseja sincronizar. Exemplo: \"busybox\" ou \"fedora / ssh '. ","Networking":"Networking","Never synced":"Nunca sincronizado","New Activation Key":"Nova Chave de Ativação","New Content View":"Nova Visualização de Conteúdo","New Environment":"Novo Ambiente","New Environment Path":"Novo Caminho de Ambiente","New Filter":"Novo Filtro","New Filter for Content View:":"Novo Filtro para Visualização de Conteúdo","New GPG Key":"Nova Chave GPG","New Gpg Key":"Nova Chave GPG","New Host Collection":"Nova Coleção de Host","New Name:":"Novo Nome:","New Product":"NovoProduto","New Repository":"Novo Repositório","New Sync Plan":"Novo Plano de Sincronização","New sync plan successfully created.":"Novo Plano de sinc criado com sucesso.","Next":"Próximo","Next Sync":"Próxima Sinc","No":"Não","No \"Edit\" permissions on some of the Activation Keys in that Lifecycle Environment.":"Não há permissões 'Editar' em algumas Chaves de Ativação naquele Ambiente de Ciclo de Vida.","No \"Edit\" permissions on some of the Content Hosts in that Lifecycle Environment.":"Não há permissões 'Editar' em algumas Hosts de Conteúdo naquele Ambiente de Ciclo de Vida.","No \"Promote/Remove\" permissions on the Lifecycle Environment.":"Não há permissões 'Promoção/Remover' no Ambiente de Ciclo de Vida.","No (Default)":"Nenhum (Padrão)","No Content Hosts are affected.":"Nenhum Host de Conteúdo foi afetado.","No Content Hosts match this Erratum.":"Nenhum Host de Conteúdo coincide com esta Errata.","No Content Views available, please select another environment.":"Nenhuma Visualização de Conteúdo disponível, por favor selecione outro ambiente","No Host Collections to show, you can add Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"Nenhum  Coleções de Host a exibir, você pode adicionar Coleções de Host  após a seleção de 'Host Collections' em \"Hosts\"no menu principal.","No Host Collections to show, you can add Host Collections after selecting the 'Add' tab.":"Nenhuma Coleção de Host para mostrar, você pode adicionar Coleções de Hosts depois de selecionar a opção \"Adicionar\".","No Products to show.":"Nenhum Produto para exibir.","No Repositories contain this Erratum.":"Nenhum Repositório contém esta Errata.","No activation keys are affected.":"Nenhuma chave de ativação foi afetada.","No alternate release version choices are available. The available releases are based upon what is available in \"{{ contentHost.content_view.name }}\", the selected <a href=\"/content_views\">content view</a> this content host is attached to for the given <a href=\"/lifecycle_environments\">lifecycle environment</a>, \"{{ contentHost.environment.name }}\".":"Não há escolhas de versão de lançamento alternada disponíveis. Os lançamentos disponíveis são baseados no que estiver disponível na  {{ contentHost.content_view.name }}\", a visualização de conteúdo <a href=\"/content_views\">selecionada</a> que este host de conteúdo está anexado para o ciclo de vida<a href=\"/lifecycle_environments\">fornecido</a>, \"{{ contentHost.environment.name }}\".","No content views exist for {{selected.environment.name}}":"Nenhuma visualização de conteúdo existe para  {{selected.environment.name}}","No host collections":"Nenhuma coleção de host","No provisioning details are available for this content host.":"Nenhum detalhe de provisionamento está disponível para este host de conteúdo.","No repository content provided through subscriptions.":"Nenhum conteúdo de repositório fornecido através das subscrições.","No repository content provided through this subscription.":"Nenhum conteúdo de repositório fornecido através desta subscrição.","No repository content.":"Nenhum conteúdo de repositório.","No subscription manifest imported":"Nenhum manifesto de subscrição importado","None":"Nenhum","Not Synced":"Não foi Sincronizado","Not installed":"Não instalado","Not yet published":"Não foi ainda publicado","Number of CPUs":"Número de CPUs","OS":"SO","Only show Errata that is Applicable to one or more Content Hosts":"Apenas mostrar Errata que seja aplicável a um ou mais hosts de conteúdo","Only show Errata that is Installable on one or more Content Hosts":"Apenas mostrar Errata que seja instalável em um ou mais hosts de conteúdo","Operating System":"Sistema Operacional","Options:":"Opções:","Organization":"Organização","Original Sync Date":"Data de Sinc original","Override to No":"Sobrepor para Não","Override to Yes":"Sobrepor para Sim","Package":"Pacote","Package Actions":"Ações de Pacote","Package Filter:":"Filtro de Pacote:","Package Group":"Grupo do Pacote","Package Group Filter:":"Filtro de Grupo de Pacote:","Package Group successfully added.":"Grupo de Pacote foi adicionado com sucesso.","Package Group successfully removed.":"Grupo de Pacote foi removido com sucesso.","Package Groups":"Grupos de Pacote","Package Install":"Instalar Pacote","Package Installation, Removal, and Update":"Instalação do Pacote, Remoção e Atualização.","Package Remove":"Remover Pacote","Package Update":"Atualizar Pacote","Package removal failed":"Remoção do pacote falhou","Package successfully added.":"Pacote foi adicionado com sucesso","Package successfully removed.":"Pacote foi removido com sucesso.","Package successfully updated.":"Pacote atualizado com sucesso.","Package/Group Name":"Pacote/Nome do Grupo","Packages":"Pacotes","Packages <div>{{ library.counts.packages || 0 }}</div>":"Pacotes <div>{{ library.counts.packages || 0 }}</div>","Packages for:":"Pacotes para:","Packages in {{ repository.name }}":"Pacotes em {{ repository.name }}","Parameters":"Parâmetros","Paste GPG Key Contents":"Colar o Conteúdo da Chave GPG","Paste contents...":"Colar conteúdo...","Paused":"Pausado","Perform":"Desempenhar","Physical":"Físico","Please select a Lifecycle Environment and Content View to move the affected Activation Keys to:":"Por favor selecionar um Ambiente de Ciclo de Vida e Visualização de Conteúdo para mover as Chaves de Ativação afetadas para:","Please select a Lifecycle Environment and Content View to move these Content Hosts to:":"Por favor selecionar um Ambiente de Ciclo de Vida e Visualização de Conteúdo para mover estes Host de Conteúdo para:","Please select a Lifecycle Environment.":"Por favor selecionar um Ambiente de Ciclo de Vida.","Please select an environment.":"Por favor selecione um Ambiente.","Please select one from the list below and you will be redirected.":"Por favor, selecione um a partir da lista abaixo e você será redirecionado.","Press 'Sync Now' below to begin syncing the following product(s).":"Pressione 'Sincronizar Agora' abaixo para começar a sincronizar o(s) seguinte(s) produto(s).","Previous Environment (%e/%cv)":"Ambiente Anterior (%e/%cv)","Processing":"Processando","Product":"Produto","Product Bulk Actions":"Ações em Massa de Produto","Product Content":"Conteúdo de Produto","Product Count":"Conta de Produto","Product Enhancement Advisory":"Relatório de Aprimoramento do Produto","Product ID":"ID do Produto","Product Management":"Gerenciamento de Produto","Product Management for Sync Plan:":"Gerenciamento de Produto para Plano de Sinc:","Product Name":"Nome do Produto","Product Saved":"Produto Salvo","Product Sync":"Sinc Produto","Product delete operation has been initiated in the background.  Click <a href=\"{{ taskUrl(productDeletionTaskId) }}\">Here</a> to monitor the progress.":"Operação de remoção do produto foi iniciada no pano de fundo. Clique em  <a href=\"{{ taskUrl(productDeletionTaskId) }}\">Aqui</a> para monitorar o progresso.","Product information for:":"Informações de Produto para:","Product {{ product.name }}":"Produto  {{ product.name }}","Products":"Produtos","Products <div>{{ library.counts.products || 0 }}</div>":"Produtos <div>{{ library.counts.products || 0 }}</div>","Promote":"Promover","Promote Content View:":"Promover Visualização de Conteúdo:","Promote Version":"Promover Versão","Promote Version {{ version.version }}":"Promover Versão  {{ version.version }}","Promoted to %s":"Promovido para %s","Promoting to %count environments.":"Promovendo para ambientes de %count","Promoting to 1 environment.":"Promovendo para ambiente 1","Promotion History":"Histórico de Promoção","Provided Products":"Produtos Fornecidos","Provisioning Details":"Provisionando Detalhes","Provisioning Host Details":"Provisionando Detalhes do Host","Publish Content View:":"Publicar uma visualização de conteúdo","Publish New Version":"Publicado Nova Versão","Publish via HTTP":"Publicar Via HTTP","Publish via HTTPS":"Publicar Via HTTPS","Published":"Publicado","Published At":"Publicado em","Published Repository Information":"Informação de Repositório Publicada","Published new version":"Nova versão Publicada","Publishing and promoting to %count environments.":"Publicando e promovendo para ambientes de %count","Publishing and promoting to 1 environment.":"Publicando e promovendo para ambiente 1","Puppet Environment":"Ambiente Puppet","Puppet Modules":"Módulos puppet","Puppet Modules <div>{{ library.counts.puppet_modules || 0 }}</div>":"Puppet Modules <div>{{ library.counts.puppet_modules || 0 }}</div>","Puppet Modules for Content View:":"Módulos Puppet para a Visualização de Conteúdo:","Puppet Modules in {{ repository.name }}":"Puppet Modules em{{ repository.name }}","Puppet module added to Content View":"Foi adicionado o módulo puppet à visualização de conteúdo","Puppet module metadata generation has been initiated in the background.\n        Click <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Geração de metadados de Módulo de Puppet foi iniciada no pano de fundo. \nClique em  <a ng-href=\"{{ taskUrl() }}\">Here</a> para monitorar o progresso.","Quantity":"Quantidade","RAM (GB)":"RAM (GB)","RAM: %s GB":"RAM: %s GB","Range":"Classe","Reboot Suggested":"Reinicialização sugerida","Reboot Suggested?":"Reinicialização sugerida?","Red Hat CDN URL":"Red Hat CDN URL","Red Hat Provider Details":"Detalhes do Provedor da Red Hat","Red Hat Repositories page":"Página de Repositórios da Red Hat","Red Hat Repositories page.":"Página de Repositórios da Red Hat.","Refresh Manifest":"Atualizar Manifesto","Refresh Table":"Atualizar Tabela","Refreshing Manifest...":"Atualizando Manifesto...","Refreshing manifest not suppported by this manifest.":"Atualizando o manifesto não suportado por este manifesto","Register Content Host":"Registrar Host de Conteúdo","Register a Content Host":"Registrar um Host de Conteúdo","Registered":"Registrado","Registered By":"Registrado por","Registry URL":"URL de Registro","Release":"Lançamento","Release Version":"Versão de Lançamento","Remove":"Remover","Remove Activation Key \"{{ activationKey.name }}\"?":"Remover a Chave de Ativação \"{{ activationKey.name }}\"?","Remove Content View":"Remover visualização de conteúdo","Remove Content Views":"Remover Visualizações de Conteúdo","Remove Docker Images":"Remover Imagens do Docker","Remove Errata":"Remover errata","Remove Filters":"Remover Filtros","Remove From":"Remover a partir de","Remove GPG Key":"Remover Chave GPG","Remove GPG Key {{ gpgKey.name }}":"Remover Chave GPG {{ gpgKey.name }}","Remove Host Collection \"{{ hostCollection.name }}\"?":"Remover o Host Collection \"{{ hostCollection.name }}\"?","Remove Module":"Remover Módulo","Remove Package Group":"Remover Grupo de Pacote","Remove Packages":"Remover pacotes","Remove Product":"Remover Produto","Remove Product \"{{ product.name }}\"?":"Remover o Produto \"{{ product.name }}\"?","Remove Products":"Remover Produtos","Remove Puppet Modules":"Remover Módulos do Puppet","Remove Repositories":"Remover Repositórios","Remove Repository":"Remover Repositório","Remove Repository \"{{ repository.name }}\"?":"Remover Repositório \"{{ repository.name }}\"?","Remove Selected":"Remover Selecionado","Remove Successful.":"Remoção foi bem sucedido","Remove Sync Plan":"Remover Plano de Sincronização","Remove Sync Plan \"{{ syncPlan.name }}\"?":"Remover o Sync Plan \"{{ syncPlan.name }}\"?","Remove Version":"Versão a Remover","Remove Version Confirmation":"Remover Confirmação de Versão","Remove View":"Remover Visualização","Remove {{ productTable.getSelected().length }} Products?":"Remover Produtos {{ productTable.getSelected().length }}?","Removed %x host collections from activation key \"%y\".":"Foram removidas coleções de host %x da chave de  ativação \"%y\".","Removed %x host collections from content host \"%y\".":"Foram removidas coleções de host %x do host de conteúdo \"%y\".","Removed %x products from sync plan \"%y\".":"Foram removidos produtos %x do plano de sincronização \"%y\".","Removing Content Hosts":"Removendo Hosts de Conteúdo ","Removing Products":"Removendo Produtos","Removing Repositories":"Removendo Repositórios","Repo Discovery":"Descoberta do Repositório","Repo Type":"Tipo de repo","Repositories":"Repositórios","Repositories containing Errata {{ errata.errata_id }}":"Repositórios contendo Errata {{ errata.errata_id }}","Repositories for Content View:":"Repositório para Visualização de Conteúdo:","Repositories for Errata:":"Repositórios para Errata:","Repositories for Filter:":"Repositórios para Filtrar:","Repositories for Product:":"Repositórios para Produto:","Repository":"Repositório","Repository \"%s\" successfully deleted":"Repositório \"%s\" excluído com êxito","Repository %s successfully created.":"Repositório %s foi criado com sucesso.","Repository Count":"Contagem do Repositório","Repository Discovery":"Descoberta do Repositório","Repository Name":"Nome do Repositório","Repository Saved.":"Repositório Salvo.","Repository Selection":"Seleção de Repositório","Repository URL updated":"URL do repositório","Repository:":"Repositório:","Required Host":"Host Requerido","Result":"Resultado","Results":"Resultados","Rules Source":"Fonte de regras","Rules Version":"Versão de regras","Run Auto-Attach":"Executar Auto-Attach","Run subscription-manager in a console on the client host.  You may use an Activation Key to register:":"Execute o subscription-manager em um console em um host de cliente. Você pode usar uma Chave de Ativação para registrar:","Save Successful.":"Foi salvo com sucesso","Search...":"Buscar...","Security":"Segurança","Security Advisory":"Aviso de Segurança","Select A New Puppet Module To Add":"Selecionar um Novo Módulo Puppet para Adicionar","Select Content Hosts":"Selecionar os Hosts de Conteúdo","Select Module":"Selecionar Módulo","Select Version":"Selecionar Versão","Select a Content Source:":"Selecione uma Fonte de Conteúdo:","Select a Version":"Selecionar uma Versão","Select an Available Version of {{ $stateParams.moduleName }}":"Selecionar uma Versão Disponível do  {{ $stateParams.moduleName }}","Select an Organization":"Selecionar uma organização","Select new version":"Selecionar nova versão","Select the Lifecycle Environments you would like to remove Version {{ version.version }} from:":"Selecionar os Ambientes de Ciclo de Vida de onde você deseja remover a Versão {{ version.version }}:","Selected URLs:":"URLs Selecionadas:","Selecting this option will result in the Version being completely deleted and no longer being available for promotion.  The version must be removed from all Lifecycle Environments in order to select this option.":"A seleção dessa opção resultará na versão que está sendo completamente excluída e não estará disponível para a promoção. A versão deve ser removida de todos os ambientes de ciclo de vida, para que possa selecionar essa opção.","Serve via HTTP.":"Servir via HTTP.","Service Level":"Nível de Serviço","Severity":"Gravidade","Show affected Activation Keys":"Exibir as Chaves de Ativação Afetadas","Show affected Content Hosts":"Exibir Hosts de Conteúdo Afetados","Show from:":"Exibir a partir de:","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"Exibindo {{ detailsTable.rows.length }} de {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Hosts de Conteúdo","Showing {{ addSubscriptionsTable.rows.length }} of {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} Total) Subscriptions":"Exibindo {{ addSubscriptionsTable.rows.length }} de {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} Total) Subscrições","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"Exibindo {{ detailsTable.rows.length }} de {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Hosts de Conteúdo","Showing {{ currentPackagesTable.resource.total }} Packages":"Exibindo Pacotes {{ currentPackagesTable.resource.total }} ","Showing {{ subscriptionsTable.rows.length }} of {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} Total) Subscriptions":"Exibir {{ subscriptionsTable.rows.length }} of {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} Total) Subscrições","Sockets":"Soquetes","Sockets: %s":"Soquetes: %s","Solution":"Solução","Some of the Errata shown below may not be installable as they are not in this Content Host's\n          Content View and Lifecycle Environment.  In order to apply such Errata an Incremental Update is required.":"Algumas Erratas exibidas abaixo podem não ser instaláveis pois não se encontram na Visualização de Conteúdo e Ambiente de Ciclo de Vida\n    do Host de Conteúdo. Para aplicar tal Errata, é necessário uma Atualização Adicional.","Stacking ID":"Empilhando o ID","Start Date":"Data de Início","Start Time":"Hora inicial","Started At":"Iniciou em","Starting":"Iniciando","Starts":"Começa","State":"Estado","Status":"Status","Subscription Details":"Detalhes de Subscrição","Subscription Info":"Info da Subscrição","Subscription Manifest":"Manifesto da Subscrição","Subscription Server Details":"Detalhes de Servidor de Subscrição","Subscription Status":"Status de Subscrição","Subscriptions":"Subscrições","Subscriptions for Activation Key:":"Subscrições para Chave de Ativação:","Subscriptions for:":"Subscrições para:","Succesfully scheduled package installation":"Instalação de pacotes agendada com sucesso","Succesfully scheduled package removal":"Remoção de pacote agendada com sucesso","Succesfully scheduled package update":"Atualização de pacote agendada com sucesso","Success":"Sucesso","Successfully Scheduled Auto-attach.":"Auto-Anexar Agendado com Sucesso","Successfully added %s subscriptions.":"Subscrições %s adicionadas com sucessso","Successfully deleted %cv version %ver.":"Versão %ver removida %cv com sucesso","Successfully initiated deletion of %cv version %ver.":"Remoção da %cv versão %ver iniciada com sucesso.","Successfully initiated promotion of %cv version %ver to %env.":"Promoção da %cv versão %ver iniciada em %env.","Successfully initiated removal of %cv version %ver.":"Remoção da %cv versão %ver iniciada com sucesso.","Successfully promoted %cv version %ver to %env":"Foi promovida a %cv versão %ver em %env com sucesso.","Successfully published %cv version %ver and promoted to Library":"Publicação de %cv versão %ver e promovida para a Biblioteca com sucesso","Successfully removed %cv version %ver from environments: %env":"Foi removida a %cv versão %ver de ambientes com sucesso: %env","Successfully removed %s items.":"Itens foram removidos %s com sucesso.","Successfully removed %s subscriptions.":"Foram removidas subscrições %s com sucesso.","Successfully removed 1 item.":"1 Item foi removido com sucesso.","Successfully updated subscriptions.":"Subscrições atualizadas com sucesso.","Summary":"Sumário","Support Level":"Nível de Suporte","Support Type":"Tipo de Suporte","Sync Enabled":"Sincronizar  Habilitada","Sync Information":"Sincronizar Informação","Sync Interval":"Intervalo de Sinc","Sync Now":"Sinc Agora","Sync Plan":"Plano de Sinc","Sync Plan %s has been deleted.":"Plano de Sinc %s foi removido.","Sync Plan Saved":"Plano de Sinc foi Salvo","Sync Plan {{ syncPlan.name }}":"Plano de Sinc {{ syncPlan.name }}","Sync Plan:":"Plano de Sinc:","Sync Plans":"Planos de Sincronização","Sync State":"Estado de Sinc","Sync Status":"Status de Sincronização","Synced manually, no interval set.":"Sincronizado manualmente, nenhum intervalo foi definido.","Syncing...":"Sincronizando...","Tags":"Marcações","Target":"Alvo","Task Details":"Detalhes de Tarefa","Tasks":"Tarefas","Tasks for Content View:":"Tarefas para Visualização de Conteúdo:","Temporary":"Temporário","The Activation Keys listed below are currently using this Content View Version.  Before deleting the Version you must move these Activation Keys to a Lifecycle Environment where this Version is not in use.":"As Chaves de Ativação listadas abaixo estão usando esta Versão de Visualização de Conteúdo. Antes de apagar a versão,  mova essas chaves de ativação para um ambiente de Ciclo de Vida, onde esta versão não será utilizada.","The Content Hosts listed below are currently using this Content View Version.  Before removing the version you must move these Content Hosts to an environment where this version is not in use.":"Os hosts de conteúdo listados abaixo estão utilizando esta Versão de Visualização de Conteúdo. Antes de remover a versão, mova esses hosts de conteúdo para um ambiente onde esta versão não esteja sendo utilizada.","The Content View or Lifecycle Environment needs to be updated in order to make errata available to these hosts.":"O Visualizador de Conteúdo ou Ambiente de Ciclo de Vida precisa ser atualizado para disponibilizar a errata à estes hosts.","The archive of this Version will be deleted.  This Version will not be available once deletion is complete.":"O arquivo desta versão será excluído. Esta versão não estará disponível depois que a exclusão for concluída.","The archive of this Version will not be deleted.  This Version will still be available to be promoted to a Lifecycle Environment.":"O arquivo desta versão não será excluído. Esta versão ainda estará disponível para ser promovida para um Ambiente de Ciclo de vida.","The following actions can be performed on content hosts in this host collection:":"As seguintes ações podem ser realizadas nos hosts de conteúdo nesta coleção de host:","The host requires being attached to a content view and the environment you have chosen has no content views promoted to it.\n            See the <a href=\"/content_views\">content views page</a> to manage and promote a content view.":"O host exige ser anexado a uma exibição de conteúdo e que o ambiente que você escolheu não tenha visualizações de conteúdo promovidos a ele.\n                Veja a  <a href=\"/content_views\">content views page</a> para gerenciar e promover a exibição de conteúdo.","The katello-agent package is required to manage errata on this Host.\n      Please install the katello-agent package to manage errata on\n      {{ contentHost.name }}.":"O pacote katello-agent precisa gerenciar errata neste Host.\n   Por favor instale o pacote katello-agent para gerenciar a errata no\n     {{ contentHost.name }}.","The katello-agent package is required to manage packages on this Host.\n      Please install the katello-agent package to manage packages on\n      {{ contentHost.name }}. To install the package, connect directly to the\n      host from a terminal and use yum.":" O pacote Katello-agent é necessário para gerenciar pacotes neste host.\n          Por favor, instale o pacote Katello-agent para gerenciar pacotes em\n          {{ contentHost.name }}. Para instalar o pacote, conecte-se diretamente ao\n          host de um terminal e use o yum.","The page you are attempting to access requires selecting a specific organization.":"A página que você está tentando acessar requer a seleção de uma organização específica.","The selected environment contains no Content Views, please select a different environment.":"O ambiente selecionado não contém visualizações de conteúdo, por favor selecione um outro ambiente.","There are currently no repositories associated with this Content View, you can add some by clicking on the \"Add\" tab above.":"Não há repositórios no momento associados à Visualização de Conteúdo, você pode adicionar alguns clicando em \"Adicionar\" acima.","There are currently no repositories to add to this Content View,":"Não existem repositórios no momento para adicionar a esta Visualização de Conteúdo.","There are no Content Views in this Environment.":"Não existem Visualizações de Conteúdo neste Ambiente","There are no Content Views that match the criteria.":"Não existem Visualizações de Conteúdo que coincide com o critério.","There are no Docker Tags that match the criteria.":"Não existem Marcações de Docker que coincidem com este critério.","There are no Errata associated with this Content Host to display.":"Não existem Erratas associadas ao Host de Conteúdo para exibir.","There are no Errata in this organization.  Create one or more Products with Errata to view Errata on this page.":"Não existem Erratas nesta organização. Crie um ou mais Produtos com a Errata para visualizar a Errata nesta página.","There are no Errata that match the criteria.":"Não existem Erratas que coincidem com este critério.","There are no Errata to display.":"Não existem Erratas para exibir","There are no Host Collections available. You can create new Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"Nao existe nenhuma  Coleção de Host disponível. Você pode adicionar Coleções de Host  após a seleção de 'Host Collections' em \"Hosts\"no menu principal.","There are no Packages that match the criteria.":"Não existem Pacotes que coincidam com este critério.","There are no Puppet Modules that match the criteria.":"Não existem Módulos de Puppet que coincidam com este critério.","There are no Yum Repositories that match the criteria.":"Não existem Repositórios Yum que coincidam com este critério.","There is currently an Incremental Update task in progress.  This update must finish before applying existing updates.":"Existe atualmente uma tarefa de Atualização Adicional em desenvolvimento. Esta atualização deve ser concluída antes de aplicar as atualizações existentes.","There were errors while removing the following Repositories:":"Foram encontrados alguns erros ao remover os seguintes Repositórios:","This Content Host is not attached to any subscriptions that provide content":"Este Host de Conteúdo não está anexado à qualquer subscrição que forneça conteúdo","This Content View does not have any versions, create your first Content View Version by using the \"Publish New Version\" button on the right.":"Esta Visualização de Conteúdo não possui nenhuma versão, crie sua primeira Versão de Visualização de Conteúdo utilizando o botão \"Publicar Nova Versão\" à direita.","This Version cannot be deleted from some of the Lifecycle Environments due to one of the reasons below.":"Esta Versão não pode ser removida de alguns Ambienes de Ciclo de Vida devido à uma destas razões abaixo:","This Version is not associated with any Lifecycle Environments.":"Esta Versão não está associada à qualquer Ambiente de Ciclo de Vida.","This activation key is not associated with any content hosts.":"Esta chave de ativação não está associada à qualquer host de conteúdo.","This activation key may be used during system registration. For example:":"Esta chave de ativação pode ser utilizada durante o registro de sistema. Por exemplo:","This filter applies only to a subset of repositories in the content view.":"Este filtro se aplica somente ao subconjunto de repositórios na visualização de conteúdo.","This filter applies to all repositories in the content view (current and future).":"Este filtro se aplica à todos os repositórios na visualização de conteúdo (atual e futuro)","This subscription is not associated with any activation keys.":"Esta subscrição não está associada à qualquer chave de ativação.","This subscription is not attached to any content hosts.":"Esta subscrição não está anexada à qualquer hosts de conteúdo.","This version will be removed from:":"Esta versão será removida de:","Time":"Tempo","Title":"Título","To register a content host to this server, follow these steps.":"Para registrar um host de conteúdo à este servidor, siga estes passos.","Topic":"Tópico","Type":"Tipo","URL":"URL","URL of the registry you want to sync. Example: https://registry.hub.docker.com":"URL do registro que você deseja sincronizar. Exemplo: https://registry.hub.docker.com","URL to discover":"URL para descrobrir","UUID":"UUID","Unattach Sync Plan":"Desanexar o Plano de Sinc","Unattaching...":"Desanexando...","Unlimited Content Hosts:":"Hosts de Conteúdo Ilimitados:","Unregister Content Host":"Cancelar registro de Host de Conteúdo","Unregister Content Host \"{{contentHost.name}}\"?":"Cancele o Registro do Host de Conteúdo \"{{contentHost.name}}\"?","Unregister Content Hosts":"Cancelar registro de Hosts de Conteúdo","Unregister {{ table.numSelected  }} Content Hosts?":"Cancelar o registro dos Hosts de Conteúdo  {{ table.numSelected  }} ?","Update":"Atualizar","Update All":"Atualizar Todos","Update Repositories":"Atualizar Repositórios","Update Sync Plan":"Atualizar Plano de Sinc","Updated":"Atualizado","Updated errata filter -":"Filtro de errata atualizado - ","Updated override to \"%x\".":"Sobreposição atualizada para \"%x\".","Updating...":"Atualizando...","Upload":"Carregar","Upload Docker Image":"Carregar a Imagem de Docker","Upload GPG Key":"Carregar a Chave GPG","Upload New Manifest":"Carregar Manifesto Novo","Upload Package":"Carregar o Pacote","Upload Puppet Module":"Carregar o Módulo Puppet","Uploading...":"Carregando...","Upstream Repository Name":"Nome do Repositório de Upstream","Upstream Subscription":"Subscrição de Upstream","Upstream Subscription Management Application":"Aplicativo de Gerenciamento de Subscrição Upstream","Upstream Type":"Tipo de Upstream","Upstream User":"Usuário de Upstream","Use Latest (currently %s)":"Utilizar o Mais Recente (atualmente %s)","Use the cancel button on content view selection to revert your environment selection.":"Usar o botão de cancelar na seleção de visualização de conteúdo para reverter sua seleção de ambiente.","User":"Usuário","Version":"Versão","Version %s":"Versão %s","Version 1":"Versão 1","Version Deletion":"Remoção da Versão","Version Details":"Detalhes da Versão","Version {{ version.version }}":"Versão {{ version.version }}","Version:":"Versão:","Versions":"Versões","Versions for Content View:":"Versões para Visualização de Conteúdo:","Versions for Puppet Module:":"Versões para Módulo Puppet:","View Details":"Visualizar Detalhes","Virtual":"Virtual","Virtual Guest From":"Convidado Virtual de","Virtual Guests":"Convidados Virtuais","Virtual Host":"Host Virtual","Weekly on {{ product.sync_plan.sync_date | date:'EEEE' }} at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"Semanalmente em  {{ product.sync_plan.sync_date | date:'EEEE' }} em {{ product.sync_plan.sync_date | date:'mediumTime' }} (Tempo do Servidor)","When Auto Attach is disabled, registering systems will be attached to all associated subscriptions.":"Quando o Auto Attach é desabilitado, o registro de sistemas será anexado à todas as subscrições associadas.","Working":"Funcionando","Yes":"Sim","Yes (Default)":"Sim (Padrão)","You cannot remove these repositories because they belong to a Red Hat product.":"Você não pode remover estes repositórios porque eles pertencem ao produto da Red Hat.","You cannot remove these repositories because you do not have permission.":"Você não pode remover estes repositórios porque você não tem permissão.","You cannot remove this product because it is a Red Hat product.":"Você não pode remover estes produto porque é um produto da Red Hat.","You cannot remove this product because it was published to a content view.":"Você não pode remover este produto porque ele foi publicado à visualização de conteúdo.","You cannot remove this product because you do not have permission.":"Você não pode remover este produto porque você não possui permissão.","You cannot remove this repository because it is a Red Hat repository.":"Você não pode remover estes repositório porque é um repositório da Red Hat.","You cannot remove this repository because it was published to a content view.":"Você não pode remover este repositório porque ele foi publicado à visualização de conteúdo.","You cannot remove this repository because you do not have permission.":"Você não pode remover este repositório porque você não possui permissão.","You currently do not have any Packages installed. You can use Package Actions above this message to install new Packages.":"No momento você não tem nenhum pacote instalado. Você pode usar ações do pacote acima desta mensagem para instalar novos pacotes.","You currently don't have any Activation Keys, you can add Activation Keys using the button on the right.":" No momento você não tem nenhuma Chave de Ativação, você pode adicionar Chaves de Ativação utilizando o botão à direita.","You currently don't have any Content Hosts in this Host Group, you can add Content Hosts after selecting the 'Add' tab.":" No momento você não tem nenhum Host de conteúdo neste grupo de hosts, você pode adicionar Hosts de Conteúdo depois de selecionar a opção \"Adicionar\".","You currently don't have any Content Hosts, you can create new Content Hosts by selecting Contents Host from main menu and then clicking the button on the right.":"No momento você não tem nenhum host de conteúdo, você pode criar novos hosts de conteúdo, selecionando Host de Conteúdo do menu principal e, em seguida, clique no botão à direita.","You currently don't have any Content Hosts, you can register one by clicking the button on the right and following the instructions.":"No momento você não tem nenhum Host de Conteúdo, você pode registrar um utilizando o botão à direita e seguindo as instruções.","You currently don't have any Content Views.  A Content View can be added by using the button on the right.":"No momento você não tem nenhuma Visualização de Conteúdo. Pode-se adicionar uma Visualização de Conteúdo utilizando o botão à direita.","You currently don't have any Docker Tags.":"Você não possui nenhuma Marcação de Docker atualmente.","You currently don't have any Filters included in this Content View, you can add a new Filter by using the button on the right.":"Você não possui atualmente nenhum Filtro incluso nesta Visualização de Conteúdo, você pode adicionar um novo Filtro utilizando o botão à direita.","You currently don't have any Gpg keys, you can add Gpg keys using the button on the right.":"No momento você não tem nenhuma chave GPG, você pode adicionar as chaves GPG usando o botão à direita.","You currently don't have any Host Collections, you can add Host Collections using the button on the right.":"No momento você não tem nenhuma Coleção de Host, você pode adicionar Coleção de Host utilizando o botão à direita.","You currently don't have any Products associated with this Gpg Key.":"No momento você não tem nenhum produto associado a essa chave GPG.","You currently don't have any Products to subscribe to, you can add Products after selecting 'Products' under 'Content' in the main menu.":"No momento você não tem nenhum produto para assinar, você pode adicionar os produtos após a seleção de 'produtos' sob 'Conteúdo' no menu principal.","You currently don't have any Products<span bst-feature-flag=\"custom_products\">, you can add Products using the button on the right</span>.":"No momento você não possui nenhum produto <span bst-feature-flag=\"custom_products\">, você pode adicionar os produtos usando o botão à direita</span>.","You currently don't have any Puppet Modules included in this Content View, you can add Puppet Modules using the button on the right.":"No momento você não possui nenhum Módulos Puppet incluído nesta visualização de conteúdo, você pode adicionar Puppet Módulos utilizando o botão à direita.","You currently don't have any Repositories associated with this Gpg Key.":"No momento você não tem nenhum Repositório associado a esta chave GPG.","You currently don't have any Repositories included in this Product, you can add Repositories using the button on the right.":"No momento você não tem nenhum Repositório incluído neste produto, você pode adicionar Repositórios usando o botão à direita.","You currently don't have any Subscriptions associated with this Content Host, you can add Subscriptions after selecting the 'Add' tab.":"No momento você não tem nenhuma Assinatura associada a este Host de Conteúdo, você pode adicionar assinaturas depois de selecionar a opção \"Adicionar\".","You currently don't have any Subscriptions, you can add Subscriptions by importing a manifest using the button on the right labeled 'Manage Manifest'.":"No momento você não tem nenhuma Assinatura, você pode adicionar assinaturas importando um manifesto usando o botão à direita chamado \"Gerenciar Manifesto\".","You currently don't have any Sync Plan, you can add Sync Plans using the button on the right.":" No momento você não tem nenhum plano de sincronização, você pode adicionar Planos de sincronização usando o botão do lado direito.","You currently don't have any Sync Plans.  A Sync Plan can be created by using the button on the right.":" No momento você não tem nenhum plano de sincronização. Pode-se criar um Plano de Sincronização usando o botão do lado direito.","You do not have any Installed Products":"Você não possui qualquer Produto Instalado","You must select a content view in order to save your environment.":"Você precisa selecionar uma visualização de conteúdo para salvar seu ambiente.","You must select a new content view before your change of environment can be saved.":"Você precisa selecionar uma nova visualização de conteúdo antes que a mudança do ambiente  seja salva.","You must select a new content view before your change of environment can be saved. Use the cancel button on content view selection to revert your environment selection.":"Você precisa selecionar uma nova visualização de conteúdo antes que sua mudança de ambiente seja salva. Use o botão de cancelar na seleção de visualização de conteúdo para reverter sua seleção de ambiente.","You must select at least one Content Host in order to apply Errata.":"Você precisa selecionar ao menos um Host de Conteúdo para aplicar a Errata.","You must select at least one repository.":"Você precisa selecionar ao menos um repositório.","Your selected content hosts have changed, it is highly recommended to refresh the calculated list of errata using the 'Refresh Errata' button below.":"Seus hosts de conteúdo selecionados mudaram, recomendamos atualizar a lista calculada de erratas utilizando o botão abaixo \"Atualizar Errata\".","Yum Content":"Conteúdo do Yum","Yum Metadata Checksum":"Yum Metadata Checksum","Yum Repo Discovery":"Descoberta do Yum Repo","Yum Repositories":"Repositórios do Yum","Yum Repositories <div>{{ library.counts.yum_repositories || 0 }}</div>":"Yum Repositories <div>{{ library.counts.yum_repositories || 0 }}</div>","Yum metadata generation has been initiated in the background.  Click <a href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Geração de metadados Yum foi iniciada no pano de fundo. Clique em  <a href=\"{{ taskUrl() }}\">Here</a> para monitorar o progresso.","add some repositories.":"adicionar alguns repositórios","daily":"por dia","here":"aqui","hourly":"por hora","or authenticate with a username and password:":"ou autenticar com um username e senha:","subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"":"subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"","task details":"detalhes da tarefa","to manage them individually.":"para gerencia-los individualmente","true":"verdadeiro","weekly":"semanal","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }} Selecionado","{{ addSubscriptionsTable.numSelected }} Selected":"{{ addSubscriptionsTable.numSelected }} Selecionado","{{ contentHost.registered_by }}":"{{ contentHost.registered_by }}","{{ detailsTable.numSelected }} Selected":"{{ detailsTable.numSelected }} Selecionado","{{ contentView.name }} cannot be deleted as one or more Content View Versions are still promoted to a Lifecycle Environment.\n      Each Content View Version must be removed from their Lifecycle Environments before the Content View can be deleted.":"{{ contentView.name }} não pode ser apagado, pois um ou mais  Versões de Visualização de Conteúdo ainda estão promovidas a um Ambiente de Ciclo de Vida.\n       Cada Versão de Visualização de Conteúdo deve ser removida de seus Ambientes de Ciclo de Vida antes que a visualização de conteúdo seja apagada.","{{ contentView.puppet_modules.length || 0 }} Puppet Modules":"{{ contentView.puppet_modules.length || 0 }} Módulos do Puppet","{{ contentView.repositories.length || 0 }} Repositories":"{{ contentView.repositories.length || 0 }} Repositórios","{{ contentViewVersion.puppet_module_count || 0 }} Puppet Modules":"{{ contentViewVersion.puppet_module_count || 0 }} Puppet Modules","{{ contentViewVersion.repositories.length || 0 }} Repositories":"{{ contentViewVersion.repositories.length || 0 }} Repositórios","{{ discoveryTable.numSelected }} Selected":"{{ discoveryTable.numSelected }} Selecionado","{{ environment.name }} Environment":"{{ environment.name }} Ambiente","{{ errata.hosts_applicable_count || 0 }} Applicable,":"{{ errata.hosts_applicable_count || 0 }} Aplicável,","{{ errata.hosts_available_count || 0 }} Installable":"{{ errata.hosts_available_count || 0 }} Instalável","{{ hostCollectionsTable.numSelected }} Selected":"{{ hostCollectionsTable.numSelected }} Selecionado","{{ package.removeTask.affected_units }} package removed":"{{ package.removeTask.affected_units }} pacote removido","{{ package.removeTask.affected_units }} packages removed":"{{ package.removeTask.affected_units }} pacote removido","{{ product.last_sync_words }} Ago ({{ product.sync_status.finish_time | date:'medium' }} Local Time)":"{{ product.last_sync_words }} Atrás ({{ product.sync_status.finish_time | date:'medium' }} Horário Local)","{{ product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ product.sync_plan.next_sync | date:'medium' }} (Horário do Servidor)","{{ productsTable.numSelected }} Selected":"{{ productsTable.numSelected }} Selecionado","{{ repository.content_counts.docker_image || 0 }} Docker Images":"{{ repository.content_counts.docker_image || 0 }} Imagens de Docker","{{ repository.content_counts.docker_image }} Docker Images":"{{ repository.content_counts.docker_image }} Imagens de Docker","{{ repository.content_counts.docker_tag || 0 }} Docker Tags":"{{ repository.content_counts.docker_tag || 0 }} Docker Tags","{{ repository.content_counts.docker_tag }} Docker Tags":"{{ repository.content_counts.docker_tag }} Docker Tags","{{ repository.content_counts.erratum || 0 }} Errata":"{{ repository.content_counts.erratum || 0 }} Errata","{{ repository.content_counts.erratum }} Errata":"{{ repository.content_counts.erratum }} Errata","{{ repository.content_counts.puppet_module || 0 }} Puppet Modules":"{{ repository.content_counts.puppet_module || 0 }} Módulos do Puppet","{{ repository.content_counts.rpm || 0 }} Packages":"{{ repository.content_counts.rpm || 0 }} Pacotes","{{ repository.content_counts.rpm }} Packages":"{{ repository.content_counts.rpm }} Pacotes","{{ repository.last_sync_words }} Ago ({{ repository.last_sync.ended_at | date:'medium' }} Local Time)":"{{ repository.last_sync_words }} Atrás ({{ repository.last_sync.ended_at | date:'medium' }} Horário Local)","{{ repository.last_sync_words }} ago":"{{ repository.last_sync_words }} atrás","{{ repository.product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ repository.product.sync_plan.next_sync | date:'medium' }} (Horário do Servidor)","{{ subscription.host.name }}":"{{ subscription.host.name }}","{{ subscriptionsTable.numSelected }} Selected":"{{ subscriptionsTable.numSelected }} Selecionado","{{ totalActivationKeyCount() }} Activation Keys will be moved to {{ deleteOptions.activationKeys.contentView.name }} in {{ deleteOptions.activationKeys.environment.name }}":"{{ totalActivationKeyCount() }} Chave de Ativação será movida para {{ deleteOptions.activationKeys.contentView.name }} em {{ deleteOptions.activationKeys.environment.name }}","{{ totalHostCount() }} Content Hosts will be moved to {{ deleteOptions.contentHosts.contentView.name }} in {{ deleteOptions.contentHosts.environment.name }}":"{{ totalHostCount() }} Hosts de Conteúdo será movido para {{ deleteOptions.contentHosts.contentView.name }} em {{ deleteOptions.contentHosts.environment.name }}","{{ type.display }}":"{{ type.display }}","{{ version.docker_image_count }} Docker Images":"{{ version.docker_image_count }} Imagens de Docker","{{ version.docker_tag_count }} Docker Tags":"{{ version.docker_tag_count }} Alvos de Docker","{{ version.errata_counts.total }} Errata":"{{ version.errata_counts.total }} Errata","{{ version.package_count }} Packages":"{{ version.package_count }} Pacotes","{{ version.puppet_module_count }} Puppet Modules":"{{ version.puppet_module_count }} Puppet Modules"});
    gettextCatalog.setStrings('ru', {"%(consumed)s out of %(quantity)s":"%(consumed)s из %(quantity)s","(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)":"(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"сведения о задаче\" | translate }}</a>)","(Not all Activation Keys editable )":"(не все ключи активации могут быть изменены)","(Not all Content Hosts editable )":"(не все узлы могут быть изменены)","* These marked Content View Versions are from Composite Content Views.  Their components needing updating are listed underneath.":"* Отмеченные версии представлений получены из сложных представлений. Ниже перечислены их компоненты, нуждающиеся в обновлении.","+ New Sync Plan":"+ Добавить план","- Date and Type":"- Дата и тип","-- select an interval --":"-- выберите интервал --","1 Content Host":["1 узел","{{ contentHost.virtual_guests.length }} узла","{{ contentHost.virtual_guests.length }} узлов"],"1 repository sync has errors.":["Обнаружены ошибки при синхронизации репозитория.","Обнаружены ошибки при синхронизации {{ product.sync_summary.error}} репозиториев.","Обнаружены ошибки при синхронизации {{ product.sync_summary.error}} репозиториев."],"1 repository sync in progress.":["Синхронизация репозитория...","Синхронизация {{ product.sync_summary.pending}} репозиториев...","Синхронизация {{ product.sync_summary.pending}} репозиториев..."],"1 successfully synced repository.":["Синхронизация репозитория завершена.","Синхронизация {{ product.sync_summary.success}} репозиториев завершена.","Синхронизация {{ product.sync_summary.success}} репозиториев завершена."],"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Version {{ version.version }}</a>":"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Версия {{ version.version }}</a>","<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Create":"<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Создать","<span translate=\"\">\n        If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">here</a>\n      <span translate=\"\">to manage them individually.</span>":"<span translate=\"\">\n        Чтобы перенести отдельные ключи активации в разные представления или окружения, нажмите\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">здесь</a>","A composite view contains other content views.":"Сложное представление состоит из других представлений.","A new version of {{ contentView.name }} will be created and promoted to the Library environment.\n    It can be promoted to other environments from the Versions tab of this Content View.":"Новая версия {{ contentView.name }} будет добавлена в окружение Library.\nЧтобы перенести ее в другое окружение, перейдите на вкладку «Версии».","Account":"Учетная запись","Account Number":"Номер учетной записи","Action":"Действие","Action Type":"Тип действия","Actions":"Действия","Actions for:":"Действия для:","Activation Key":["Ключ активации","Ключи активации","Ключи активации"],"Activation Key Content":"Содержимое","Activation Key Type:":"Тип ключа:","Activation Key removed.":"Ключ активации удален.","Activation Key updated":"Ключ активации обновлен","Activation Key:":"Ключ активации:","Activation Key: {{ activationKey.name }}":"Ключи активации: {{ activationKey.name }}","Activation Keys":"Ключи активации","Activation Keys for Subscription:":"Ключи активации для подписки:","Activation Keys using Version {{ version.version }}":"Ключи активации с версией  {{ version.version }}","Add":"Добавить","Add Content Hosts to:":"Добавить узлы в:","Add Content Views":"Добавить","Add Errata":"Добавить","Add New Environment":"Добавить окружение","Add New Filter":"Добавить фильтр","Add New Module":"Добавить","Add New Repository":"Добавить репозиторий","Add Package Group":"Добавить","Add Repositories":"Добавить","Add Selected":"Добавить выбранные","Add Subscriptions for Activation Key:":"Добавить подписки для ключа активации:","Add To":"Добавить в","Added %x host collections to activation key \"%y\".":"Коллекции добавлены в ключ «%y» (всего: %x)","Added %x host collections to content host \"%y\".":"Коллекции для узла «%y» добавлены (всего: %x) ","Added %x products to sync plan \"%y\".":"Продукты добавлены в «%y» (всего: %x)","Adding Lifecycle Environment to the end of \"{{ priorEnvironment.name }}\"":"Окружение будет добавлено после \"{{ priorEnvironment.name }}\"","Advanced Information":"Дополнительная информация","Advisory":"Рекомендация","Affected Activation Keys":"Ключи активации","Affected Content Hosts":"Затронутые узлы","Affected Packages":"Пакеты","Affected Systems":"Системы","Affected repositories":"Репозитории","Affected repositories have been updated.":"Репозитории были обновлены.","All Content Views":"Все представления","All Products":"Все продукты","All Repositories":"Все репозитории","All Versions":"все версии","Alter Sync Plans":"Изменить план","An error occurred applying Subscriptions:":"Ошибка при назначении подписок:","An error occurred removing the Activation Key:":"Ошибка при удалении ключа активации:","An error occurred removing the Host Collection:":"Ошибка при удалении коллекции узлов:","An error occurred removing the Products:":"Ошибка при удалении продукта:","An error occurred removing the content hosts.":"Ошибка при удалении узлов:","An error occurred removing the environment:":"Ошибка при удалении окружения:","An error occurred removing the subscriptions.":"Ошибка при удалении подписок.","An error occurred saving the Activation Key:":"Ошибка сохранения ключа активации:","An error occurred saving the Content Host:":"Ошибка сохранения узла:","An error occurred saving the Environment:":"Ошибка сохранения окружения:","An error occurred saving the Filter:":"Ошибка сохранения фильтра:","An error occurred saving the Host Collection:":"Ошибка сохранения коллекции узлов:","An error occurred saving the Product:":"Ошибка сохранения продукта:","An error occurred saving the Repository:":"Ошибка сохранения репозитория:","An error occurred saving the Sync Plan:":"Ошибка сохранения плана синхронизации:","An error occurred saving the URL:":"Ошибка сохранения адреса:","An error occurred trying to auto-attach subscriptions.  Please check your log for further information.":"Произошла ошибка при попытке автоматического выбора подписок. Подробную информацию можно найти в журнале.","An error occurred updating the Content View:":"Произошла ошибка при обновлении представления:","Applicable":"Доступно","Applicable Content Hosts for Errata":"Доступные узлы для исправлений","Applicable Errata":"Доступные исправления","Applicable hosts for {{ erratum.errata_id }}":"Узлы для {{ erratum.errata_id }}","Apply":"Применить","Apply Errata":"Применить","Apply Errata to Content Host \"{{contentHost.name}}\"?":"Применить исправления на \"{{contentHost.name}}\"?","Apply Errata to Content Hosts":"Применить исправления к узлам содержимого","Apply Errata to Content Hosts immediately after publishing.":"Применить исправления сразу после публикации.","Apply Selected":"Применить выбранные","Apply to Content Hosts":"Применить к узлам содержимого","Apply to Hosts":"Применить к узлам","Apply {{ errata.errata_id }}":"Установить {{ errata.errata_id }}","Apply {{ errata.errata_id }} to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"Установить {{ errata.errata_id }} на выбранных узлах (всего {{ numberOfContentHostsToUpdate }})?","Apply {{ errataIds.length }} Errata to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"Установить эти исправления (всего {{ errataIds.length }}) на {{ numberOfContentHostsToUpdate }} узле(ах)?","Applying":"Установка...","Applying Bulk Errata Update":"Установка исправлений...","Arch":"Архитектура","Architecture(s)":"Архитектура","Archived Copy":"Архивная копия","Are you sure you want to add the {{ table.numSelected }} content host(s) selected to the host collection(s) chosen?":"Вы действительно хотите добавить выбранное число узлов ({{ table.numSelected }}) в коллекцию?","Are you sure you want to apply Errata to content host \"{{ contentHost.name }}\"?":"Вы действительно хотите установить исправления на «{{ contentView.name }}»?","Are you sure you want to apply the {{ detailsTable.numSelected }} selected errata to the content hosts chosen?":"Вы действительно хотите применить выбранное число исправлений ({{ detailsTable.numSelected }}) к узлам?","Are you sure you want to assign the {{ table.numSelected }} content host(s) selected to {{ selected.contentView.name }} in {{ selected.environment.name }}?":"Связать выбранное число узлов ({{ table.numSelected }}) с представлением {{ selected.contentView.name }} в окружении {{ selected.environment.name }}?","Are you sure you want to auto-attach available subscriptions to all {{ table.resource.total }} content hosts?":"Подобрать подписки автоматически для {{ table.resource.total }} узлов?","Are you sure you want to delete Content View \"{{ contentView.name }}\"?":"Вы действительно хотите удалить представление «{{ contentView.name }}»?","Are you sure you want to install {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"Установить {{ content.content }} в выбранном числе систем ({{ getSelectedSystemIds().length }})?","Are you sure you want to remove Activation Key \"{{ activationKey.name }}\"?":"Вы действительно хотите удалить ключ «{{ activationKey.name }}»?","Are you sure you want to remove GPG key {{ gpgKey.name }}?":"Вы действительно хотите удалить ключ GPG «{{ gpgKey.name }}»?","Are you sure you want to remove Host Collection \"{{ hostCollection.name }}\"?":"Удалить коллекцию «{{ hostCollection.name }}»?","Are you sure you want to remove Sync Plan \"{{ syncPlan.name }}\"?":"Удалить план синхронизации «{{ syncPlan.name }}»?","Are you sure you want to remove product \"{{ product.name }}\"?":"Удалить продукт «{{ product.name }}}»?","Are you sure you want to remove repository \"{{ repository.name }}\"?":"Удалить репозиторий «{{ repository.name }}»?","Are you sure you want to remove the {{ productTable.getSelected().length }} products(s) selected?":"Удалить выбранные продукты (всего {{ productTable.getSelected().length }})?","Are you sure you want to remove the {{ table.numSelected }} content host(s) selected from the host collection(s) chosen?":"Удалить выбранные узлы (всего {{ table.numSelected }}) из коллекций?","Are you sure you want to remove {{ content.content }} from the {{ getSelectedSystemIds().length }} system(s) selected?":"Вы действительно хотите удалить {{ content.content }} из {{ getSelectedSystemIds().length }} систем(ы)?","Are you sure you want to unregister content host \"{{ contentHost.name }}\"?":"Отменить регистрацию узла «{{ contentHost.name }}»?","Are you sure you want to unregister the {{ table.numSelected }} content host(s) selected?":"Отменить регистрацию выбранных узлов (всего {{ table.numSelected }})?","Are you sure you want to update {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"Вы действительно хотите обновить {{ content.content }} в {{ getSelectedSystemIds().length }} систем(е)(ах)?","As part of this deletion, 1 Content View Version will be deleted.":["В ходе операции будет удалена одна версия представления.","В ходе операции будут удалены {{ versions.length }} версии представления.","В ходе операции будет удалено {{ versions.length }} версий представления."],"Assign":"Выбрать","Assign Environment and Content View":"Выбор окружения и представления","Associations":"Связи","At least one Errata needs to be selected to Apply.":"Необходимо выбрать хотя бы одно исправление.","At least one content host needs to be selected in order to perform a bulk action.":"Для выполнения групповой операции необходимо выбрать узел.","At least one product needs to be selected in order to perform a bulk action.":"Для выполнения групповой операции необходимо выбрать продукт.","Attached":"Назначен","Attached to Content Hosts":"Узлы","Author":"Автор","Auto-Attach":"Автоподбор","Auto-attach available subscriptions to all content hosts":"Автоматический выбор подписок для всех узлов","Automatic":"Автоматически","Available Content Views for Composite Content View:":"Доступные представления для сложного представления:","Available Puppet Modules for Content View:":"Доступные модули Puppet для сложного представления:","Back":"Назад","Back To Errata List":"Назад к списку исправлений","Back To Repository Selection":"Назад к выбору репозитория","Back to Module List":"Назад к списку модулей","Back to Repository Details":"Назад к свойствам репозитория","Back to Repository List":"Назад к списку","Back to Version List":"Назад к списку версий","Basic Information":"Общая информация","Below are the product's repository content sets currently available for this subscription. For Red Hat subscriptions, additional content can be made available through the":"Ниже приведен список наборов репозиториев для этой подписки. Чтобы включить дополнительное содержимое, см.","Below are the repository content sets currently available for this activation key through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"Ниже приведен список наборов репозиториев для этой подписки. Чтобы включить дополнительное содержимое, см.","Below are the repository content sets currently available for this content host through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"Ниже приведен список наборов репозиториев для этой подписки. Чтобы включить дополнительное содержимое, см.","Bug Fix":"Исправление ошибки","Bug Fix Advisory":"Исправление ошибки","Bugfix":"Исправление ошибки","Bulk Actions":"Групповые операции","Bulk Alter Products Sync Plan":"Групповое изменение планов синхронизации","Bulk Sync Products":"Групповая синхронизация","CVEs":"CVE","Cancel":"Отмена","Capacity":"Размер","Change assigned Lifecycle Environment or Content View":"Изменение окружений жизненного цикла и представлений","Changing default settings for content hosts that register with this activation key requires subscription-manager version 1.10 or newer to be installed on that host.":"Чтобы изменить исходные настройки для узлов, зарегистрированных с помощью этого ключа активации, необходимо, чтобы на этих узлах был установлен subscription-manager версии 1.10 или новее.","Changing default settings requires subscription-manager version 1.10 or newer to be installed on this host.":"Для изменения исходных параметров необходимо, чтобы на узле была установлена как минимум версия subscription-manager 1.10.","Checkin":"Вход","Checksum":"Контрольная сумма","Choose a lifecycle environment from the available promotion paths.":"Выберите окружение.","Click here to check the status of the task.":"Нажмите здесь, чтобы проверить статус задачи.","Click here to refresh.":"Нажмите, чтобы обновить.","Click here to see the task for the successful removals.":"Нажмите здесь для просмотра задачи для успешного удаления.","Click here to select Errata for an Incremental Update.":"Нажмите здесь, чтобы выбрать исправление для инкрементного обновления.","Close":"Закрыть","Collection Actions":"Действия","Completely remove version?":"Удалить версию окончательно?","Components":"Компоненты","Components:":"Компоненты:","Composite Content View {{ contentView.name }}":"Сложное представление {{ contentView.name }}","Composite View":"Сложное представление","Composite View?":"Сложное","Composite?":"Сложное?","Confirm":"Подтвердить","Confirm Remove":"Подтвердить удаление","Confirm Version Removal: Version {{ version.version }}":"Подтвердите удаление версии {{ version.version }}","Confirmation":"Подтверждение","Consumed":"Занято","Content":"Содержимое","Content Counts":"Содержимое","Content Download URL":"URL загрузки содержимого","Content Host":"Узел","Content Host %s has been deleted.":"%s удален.","Content Host Bulk Actions":"Групповые операции над узлами","Content Host Bulk Content":"Групповые операции над содержимым узлов","Content Host Bulk Errata":"Групповые исправления для узлов","Content Host Bulk Host Collections":"Групповые операции над коллекциями узлов","Content Host Bulk Packages":"Групповые пакеты для узлов","Content Host Bulk Subscriptions":"Групповые подписки для узлов","Content Host Content":"Содержимое","Content Host Counts":"Узлы","Content Host Errata Management":"Управление исправлениями","Content Host Limit":"Число узлов","Content Host Package Management":"Управление пакетами","Content Host Properties":"Свойства узла","Content Host Registration":"Регистрация узла","Content Host Status":"Статус узла","Content Host Subscriptions":"Подписки узла ","Content Host {{ contentHost.name }}":"Узел {{ contentHost.name }}","Content Host:":"Узел:","Content Hosts":"Узлы","Content Hosts for Activation Key:":"Узлы для ключа активации:","Content Hosts for Subscription:":"Узлы для подписки:","Content Hosts for:":"Узлы для:","Content Hosts using Version {{ version.version }}":"Узлы с версией {{ version.version }}","Content Type":"Тип содержимого","Content View":"Представление","Content View Name":"Имя представления","Content View Version":"Версия представления","Content View updated.":"Представление обновлено.","Content View:":"Представление:","Content Views":"Представления","Content Views <div>{{ library.counts.content_views || 0 }}</div>":"Представления <div>{{ library.counts.content_views || 0 }}</div>","Content Views for Composite Content View:":"Состав сложного представления:","Content successfully uploaded":"Содержимое добавлено","Contents":"Содержимое","Contract":"Контракт","Contract Number":"Номер контракта","Copy":"Копировать","Copy Activation Key":"Копировать ключ","Copy Collection":"Копировать","Copy View":"Копировать","Cores per Socket":"Ядер на сокет","Cores: %s":"Ядра: %s","Create":"Создать","Create New View":"Создать","Create Repositories":"Создать репозитории","Create Repositories Within:":"Создать репозитории в:","Create Repository":"Создать репозиторий","Create Selected":"Создать выбранные","Create Sync Plan":"Создать план","Create new view":"Создать представление","Create view from existing views":"Создать представление на основе существующего","Critical":"Критично","Current Environment (%e/%cv)":"Текущее окружение (%e/%cv)","Current Subscriptions for Activation Key:":"Текущие подписки ключа активации:","Currently Selected Puppet Modules":"Выбранные модули Puppet","Daily at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"Ежедневно в {{ product.sync_plan.sync_date | date:'mediumTime' }} (время сервера)","Date":"Дата","Default":"По умолчанию","Delete":"Удалить","Delete Manifest":"Удалить манифест","Delete Version {{ version.version }}":"Удалить версию {{ version.version }}","Delete {{ contentView.name }}":"Удалить {{ contentView.name }}","Deleted from %s":"Удалено из %s","Deleting Manifest...":"Удаление манифеста...","Deleting from %count environments.":"Удаление из %count окружений(я)","Deleting from 1 environment.":"Удаление из одного окружения...","Deletion from %s":"Удаление из %s","Description":"Описание","Detail":"Свойства","Details":"Свойства","Details for Activation Key:":"Свойства ключа активации:","Details for Content View:":"Свойства представления:","Details for Product:":"Свойства продукта:","Discover":"Поиск","Discovered URLs":"Найдено","Docker Content":"Контент Docker","Docker Images":"Образы Docker","Docker Images in {{ repository.name }}":"Образы Docker в {{ repository.name }}","Docker Repositories":"Репозитории Docker","Docker Repositories <div>{{ library.counts.docker_repositories || 0 }}</div>":"Репозитории Docker <div>{{ library.counts.docker_repositories || 0 }}</div>","Docker Tag:":"Тег Docker:","Docker Tags":"Теги Docker","Docker metadata generation has been initiated in the background.  Click\n        <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Метаданные Docker будут созданы в фоновом режиме.\n        Нажмите  <a ng-href=\"{{ taskUrl() }}\">здесь</a> для наблюдения за ходом выполнения.","Edit":"Изменить","Enabled?":"Включено?","End Date":"Срок действия","Ends":"Конец","Enhancement":"Дополнение","Enter Package Group Name(s)...":"Введите имена групп пакетов...","Enter Package Name(s)...":"Введите имена пакетов...","Environment":"Окружение","Environment saved":"Окружение сохранено","Environment: {{ environment.name }}":"Окружение: {{ environment.name }}","Environments":"Окружения","Environments List":"Список окружений","Environments:":"Окружения:","Equal To":"равно","Errata":"Исправления","Errata <div>{{ library.counts.errata.total || 0 }}</div>":"Исправления  <div>{{ library.counts.errata.total || 0 }}</div>","Errata Details":"Свойства исправления","Errata Filter:":"Фильтр:","Errata ID":"Идентификатор","Errata Installation":"Установка исправлений","Errata List":"Список исправлений","Errata Task List":"Список задач исправлений","Errata Tasks":"Задачи исправлений","Errata Type":"Тип","Errata for:":"Исправления для:","Errata successfully added.":"Исправления добавлены.","Errata successfully removed.":"Исправления удалены.","Errata:":"Исправления: ","Erratum - by Date and Type":"Исправление (дата,тип)","Erratum - by ID":"Исправление (ID)","Erratum Date Range":"Диапазон","Error deleting manifest.":"Ошибка при удалении манифеста.","Error during upload:":"Не удалось  отправить:","Error importing manifest.":"Ошибка при импорте манифеста.","Error refreshing manifest.":"Ошибка при обновлении манифеста.","Event":"Событие","Events":"События","Exclude":"Исключение","Existing Product":"Существующий продукт","Expires":"Истекает","Failed":"Ошибка","File too large. Please use the CLI instead.":"Слишком большой файл. Используйте инструменты командной строки.","Filter":"Фильтр","Filter Saved":"Фильтр сохранен","Filter information for:":"Фильтр:","Filter...":"Фильтр...","Filter:":"Фильтр:","Filters":"Фильтры","Filters List":"Список фильтров","Filters for Content View:":"Фильтры представлений:","Filters successfully removed.":"Фильтры удалены.","Finished At":"Завершено:","For older operating systems such as Red Hat Enterprise Linux 5 or CentOS 5 it is recommended to use sha1.":"Для старых операционных систем (таких как Red Hat Enterprise Linux 5 и CentOS 5) рекомендуется использовать SHA1. ","GPG Key":"Ключ GPG","GPG Key URL":"URL ключа GPG","GPG Key successfully uploaded":"Ключ GPG отправлен","GPG Key {{ gpgKey.name }}":"Ключ {{ gpgKey.name }}","GPG Keys":"Ключи GPG","General Information":"Общие сведения","Generated On":"Создано","Gpg Key updated":"Ключ GPG обновлен","Gpg Keys":"Ключи GPG","Greater Than":"больше чем","Group Install":"Установка группы","Group Remove":"Удаление группы","Guests of":"Гости","Hide affected Activation Keys":"Скрыть затронутые ключи активации","Hide affected Content Hosts":"Скрыть затронутые узлы","History":"Журнал","History for Content View:":"Журнал представления:","Host Collection Management":"Управление коллекциями","Host Collection Membership":"Изменение состава коллекции","Host Collection Membership Management":"Состав коллекции узлов","Host Collection removed.":"Коллекция удалена.","Host Collection updated":"Коллекция обновлена.","Host Collection:":"Коллекция:","Host Collection: {{ hostCollection.name }}":"Коллекция:  {{ hostCollection.name }}","Host Collections":"Коллекции","Host Collections for:":"Коллекции узлов для:","Host Count":"Число узлов","Host Group":"Группа узлов","Hostname":"Имя узла","Hourly at {{ product.sync_plan.sync_date | date:'m' }} minutes and {{ product.sync_plan.sync_date | date:'s' }} seconds":"Каждый час в {{ product.sync_plan.sync_date | date:'m' }} мин. {{ product.sync_plan.sync_date | date:'s' }} сек.","IPv4 Address":"Адрес IPv4","IPv6 Address":"Адрес IPv6","Id":"ID","If you would prefer to move portions of these Activation Keys to different Content Views or Lifecycle Environments click <a href=\"{{ activationKeyLink() }}\">here</a> to manage them individually.":"Чтобы сразу переместить несколько ключей активации в другое окружение и представление, нажмите <a href=\"{{ activationKeyLink() }}\">здесь</a>","If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click":"Чтобы сразу переместить несколько ключей активации в другое окружение и представление, нажмите","If you would prefer to move portions of these Content Hosts to different content views or environments click <a href=\"{{ contentHostsLink() }}\">here</a>  to manage these Content Hosts in bulk.":"Чтобы сразу переместить несколько узлов в другое окружение и представление, нажмите  <a href=\"{{ contentHostsLink() }}\">здесь</a>","Image":"Образ","Image ID":"ID образа","Import History":"Журнал импорта","Important":"Важно","In order to add a content view to a composite view you must first publish an initial version of the content view.":"Чтобы добавить представления, сначала надо опубликовать исходную версию представления.","Include":"Включение","Included in Activation Keys":"Ключи активации","Incremental Update":"Инкрементное обновление","Install":"Установить","Install Selected":"Установить выбранные","Install katello-agent for remote actions and displaying errata information:":"Установите katello-agent для выполнения удаленных операций и получения информации об исправлениях:","Install the pre-built bootstrap RPM:":"Установите пакет:","Installable":"Можно установить","Installable Errata":"Разрешенные к установке исправления","Installed":"Установлено","Installed Package":"Установлено","Installed Packages":"Установленные пакеты","Installed Products":"Установленные","Instance-based":"Экземпляр","Interfaces":"Интерфейсы","Interval":"Интервал","Issued":"Выпущен","Katello Agent":"Агент Katello","Label":"Метка","Last Checkin":"Последний вход","Last Published":"Опубликовано","Last Puppet Report":"Последний отчет Puppet","Last Sync":"Синхронизация","Last Updated":"Обновлено","Last Updated On":"Обновлено","Last synced {{ product.last_sync_words }} ago.":"{{ product.last_sync_words }} назад","Latest (Currently %s)":"Последняя версия (%s)","Less Than":"меньше чем","Library":"Library","Library Synced Content":"Синхронизировано с Library","Lifecycle Environment \"{{ priorEnvironment.name }}\" exists within the middle of a path. New Lifecycle Environments can only be added to the end of a path.":"«{{ priorEnvironment.name }}» находится в центре цепочки. Новое окружение может быть добавлено только в конец схемы.","Lifecycle Environment Paths":"Схемы окружений жизненного цикла","Lifecycle Environments":"Окружения жизненного цикла","Limit":"Число","Limit:":"Число:","Limits":"Пределы","List/Remove":"Список","Loading Environment Paths...":"Загрузка структуры окружения...","Loading...":"Загрузка...","Manage Docker Images":"Управление образами Docker","Manage Docker Images for Repository:":"Управление образами Docker в репозитории:","Manage Manifest":"Манифест","Manage Packages":"Управление пакетами","Manage Packages for Repository:":"Управление пакетами в репозитории:","Manage Puppet Modules":"Управление модулями Puppet","Manage Puppet Modules for Repository:":"Управление модулями Puppet в репозитории:","Manifest History":"Журнал","Manifest successfully deleted.":"Манифест удален.","Manifest successfully imported.":"Манифест импортирован.","Manifest successfully refreshed.":"Манифест обновлен.","Message":"Сообщение","Model":"Модель","Moderate":"Модерация","Module %s removed from Content View.":"%s удален из представления.","Module List":"Список модулей","More Details":"Подробнее","More...":"Подробнее...","Multi-entitlement":"Многократные полномочия","N/A":"нет","Name":"Имя","Name of the upstream repository you want to sync. Example: 'busybox' or 'fedora/ssh'.":"Имя репозитория для синхронизации. Пример: «busybox», «fedora/sh»","Networking":"Сеть","Never synced":"Никогда","New Activation Key":"Новый ключ активации","New Content View":"Новое представление","New Environment":"Новое окружение","New Environment Path":"Создать схему","New Filter":"Новый фильтр","New Filter for Content View:":"Новый фильтр представления:","New GPG Key":"Новый ключ GPG","New Gpg Key":"Добавить","New Host Collection":"Новая коллекция","New Name:":"Новое имя:","New Product":"Новый продукт","New Repository":"Новый репозиторий","New Sync Plan":"Новый план","New sync plan successfully created.":"План синхронизации создан.","Next":"Далее","Next Sync":"Следующая синхронизация","No":"Нет","No \"Edit\" permissions on some of the Activation Keys in that Lifecycle Environment.":"Для некоторых ключей в этом окружении нет полномочий изменения.","No \"Edit\" permissions on some of the Content Hosts in that Lifecycle Environment.":"Для некоторых узлов в этом окружении нет полномочий изменения.","No \"Promote/Remove\" permissions on the Lifecycle Environment.":"Нет полномочий для продвижения и удаления объектов из этого окружения.","No (Default)":"Нет (по умолчанию)","No Content Hosts are affected.":"Узлы не были изменены.","No Content Hosts match this Erratum.":"Нет узлов для этого исправления.","No Content Views available, please select another environment.":"Нет представлений. Выберите другое окружение.","No Host Collections to show, you can add Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"Нет коллекций. Чтобы добавить коллекцию узлов, в главном меню выберите «Узлы» - «Коллекции».","No Host Collections to show, you can add Host Collections after selecting the 'Add' tab.":"Нет коллекций. Чтобы добавить коллекцию узлов, перейдите на вкладку «Добавить».","No Products to show.":"Нет продуктов.","No Repositories contain this Erratum.":"Нет репозиториев с этим исправлением.","No activation keys are affected.":"Ключи не были перенесены.","No alternate release version choices are available. The available releases are based upon what is available in \"{{ contentHost.content_view.name }}\", the selected <a href=\"/content_views\">content view</a> this content host is attached to for the given <a href=\"/lifecycle_environments\">lifecycle environment</a>, \"{{ contentHost.environment.name }}\".":"Нет других версий. Список версий определяется <a href=\"/content_views\">представлением</a> «{{ contentHost.content_view.name }}», которое назначено узлу в <a href=\"/lifecycle_environments\">окружении</a> «{{ contentHost.environment.name }}».","No content views exist for {{selected.environment.name}}":"Нет представлений для {{selected.environment.name}}","No host collections":"Нет коллекций узлов","No provisioning details are available for this content host.":"Нет данных для этого узла.","No repository content provided through subscriptions.":"Подписки не предоставляют доступ к репозиториям.","No repository content provided through this subscription.":"Подписка не предоставляет доступ к репозиториям.","No repository content.":"Нет содержимого.","No subscription manifest imported":"Нет манифеста подписок","None":"Нет","Not Synced":"Не синхронизирован","Not installed":"Не установлено","Not yet published":"Нет","Number of CPUs":"Число процессоров","OS":"ОС","Only show Errata that is Applicable to one or more Content Hosts":"Показать доступные исправления","Only show Errata that is Installable on one or more Content Hosts":"Показать исправления, которые можно будет установить","Operating System":"Операционная система","Options:":"Аргументы:","Organization":"Организация","Original Sync Date":"Первая синхронизация","Override to No":"Нет","Override to Yes":"Да","Package":"Пакет","Package Actions":"Операции над пакетами","Package Filter:":"Фильтр пакетов:","Package Group":"Группа пакетов","Package Group Filter:":"Фильтр групп:","Package Group successfully added.":"Группа добавлена.","Package Group successfully removed.":"Группа удалена.","Package Groups":"Группы пакетов","Package Install":"Установка пакетов","Package Installation, Removal, and Update":"Установка, удаление и обновление пакетов","Package Remove":"Удаление пакетов","Package Update":"Обновление пакетов","Package removal failed":"Не удалось удалить пакет.","Package successfully added.":"Пакет добавлен.","Package successfully removed.":"Пакет удален.","Package successfully updated.":"Пакет обновлен.","Package/Group Name":"Пакет/группа","Packages":"Пакеты","Packages <div>{{ library.counts.packages || 0 }}</div>":"Пакеты <div>{{ library.counts.packages || 0 }}</div>","Packages for:":"Пакеты для:","Packages in {{ repository.name }}":"Пакеты в {{ repository.name }}","Parameters":"Параметры","Paste GPG Key Contents":"Вставить содержимое ключа GPG","Paste contents...":"Вставьте текст...","Paused":"Приостановлена","Perform":"Выполнить","Physical":" Физическая","Please select a Lifecycle Environment and Content View to move the affected Activation Keys to:":"Выберите окружение и представление, куда следует перенести ключи активации:","Please select a Lifecycle Environment and Content View to move these Content Hosts to:":"Выберите окружение и представление, куда следует перенести узлы:","Please select a Lifecycle Environment.":"Выберите окружение жизненного цикла","Please select an environment.":"Выберите окружение","Please select one from the list below and you will be redirected.":"Чтобы перейти на страницу, выберите организацию из списка.","Press 'Sync Now' below to begin syncing the following product(s).":"Нажмите «Синхронизировать», чтобы начать синхронизацию.","Previous Environment (%e/%cv)":"Предыдущее окружение (%e/%cv)","Processing":"Обработка...","Product":"Компонент","Product Bulk Actions":"Групповые операции над продуктами","Product Content":"Состав","Product Count":"Число продуктов","Product Enhancement Advisory":"Дополнение","Product ID":"Идентификатор","Product Management":"Управление продуктами","Product Management for Sync Plan:":"Продукты в плане синхронизации:","Product Name":"Имя продукта","Product Saved":"Продукт сохранен","Product Sync":"Синхронизация","Product delete operation has been initiated in the background.  Click <a href=\"{{ taskUrl(productDeletionTaskId) }}\">Here</a> to monitor the progress.":"Удаление выполняется в фоновом режиме.  <a href=\"{{ taskUrl(productDeletionTaskId) }}\">Нажмите</a>  для просмотра подробной информации.","Product information for:":"Информация о","Product {{ product.name }}":"{{ product.name }}","Products":"Продукты","Products <div>{{ library.counts.products || 0 }}</div>":"Продукты <div>{{ library.counts.products || 0 }}</div>","Promote":"Перенести","Promote Content View:":"Продвинуть представление:","Promote Version":"Перенести","Promote Version {{ version.version }}":"Перенос версии {{ version.version }}","Promoted to %s":"Перенесено в %s","Promoting to %count environments.":"Перенос в несколько окружений (всего %count).","Promoting to 1 environment.":"Перенос в одно окружение.","Promotion History":"Журнал переносов","Provided Products":"Продукты","Provisioning Details":"Подготовка","Provisioning Host Details":"Свойства подготовки узла","Publish Content View:":"Опубликовать представление:","Publish New Version":"Опубликовать версию","Publish via HTTP":"Доступ HTTP","Publish via HTTPS":"Доступ HTTPS","Published":"Опубликовано","Published At":"Опубликован","Published Repository Information":"Информация о репозитории опубликована","Published new version":"Версия опубликована","Publishing and promoting to %count environments.":"Публикация в %count окружениях.","Publishing and promoting to 1 environment.":"Публикация и добавление в окружение.","Puppet Environment":"Окружение Puppet","Puppet Modules":"Модули Puppet","Puppet Modules <div>{{ library.counts.puppet_modules || 0 }}</div>":"Модули Puppet <div>{{ library.counts.puppet_modules || 0 }}</div>","Puppet Modules for Content View:":"Модули Puppet в представлении:","Puppet Modules in {{ repository.name }}":"Модули Puppet в {{ repository.name }}","Puppet module added to Content View":"Модуль Puppet добавлен в представление","Puppet module metadata generation has been initiated in the background.\n        Click <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Метаданные модуля Puppet будут созданы в фоновом режиме.\n        Нажмите <a ng-href=\"{{ taskUrl() }}\">здесь</a> для наблюдения за ходом выполнения.","Quantity":"Количество","RAM (GB)":"ОЗУ (ГБ)","RAM: %s GB":"ОЗУ: %s ГБ","Range":"диапазон","Reboot Suggested":"Рекомендуется перезагрузить","Reboot Suggested?":"Рекомендуется перезагрузить","Red Hat CDN URL":"Адрес сети доставки содержимого Red Hat","Red Hat Provider Details":"Провайдер Red Hat","Red Hat Repositories page":"Страница репозиториев Red Hat","Red Hat Repositories page.":"Страница репозиториев Red Hat","Refresh Manifest":"Обновить манифест","Refresh Table":"Обновить таблицу","Refreshing Manifest...":"Обновление манифеста","Refreshing manifest not suppported by this manifest.":"Манифест не поддерживает возможность обновления","Register Content Host":"Зарегистрировать узел","Register a Content Host":"Регистрация узла","Registered":"Зарегистрирован","Registered By":"Регистрация:","Registry URL":"Адрес реестра","Release":"Выпуск","Release Version":"Версия","Remove":"Удалить","Remove Activation Key \"{{ activationKey.name }}\"?":"Удалить ключ «{{ activationKey.name }}»?","Remove Content View":"Удалить представление","Remove Content Views":"Удалить представления","Remove Docker Images":"Удалить образы Docker","Remove Errata":"Удалить","Remove Filters":"Удалить","Remove From":"Удалить из","Remove GPG Key":"Удалить","Remove GPG Key {{ gpgKey.name }}":"Удалить ключ GPG {{ gpgKey.name }}","Remove Host Collection \"{{ hostCollection.name }}\"?":"Удалить коллекцию «{{ hostCollection.name }}»?","Remove Module":"Удалить модуль","Remove Package Group":"Удалить","Remove Packages":"Удалить","Remove Product":"Удалить","Remove Product \"{{ product.name }}\"?":"Удалить продукт «{{ product.name }}}»?","Remove Products":"Удалить продукты","Remove Puppet Modules":"Удалить модули Puppet","Remove Repositories":"Удалить репозитории","Remove Repository":"Удалить","Remove Repository \"{{ repository.name }}\"?":"Удалить репозиторий «{{ repository.name }}»?","Remove Selected":"Удалить выбранные","Remove Successful.":"Удаление успешно.","Remove Sync Plan":"Удалить план","Remove Sync Plan \"{{ syncPlan.name }}\"?":"Удалить план синхронизации «{{ syncPlan.name }}»?","Remove Version":"Удалить версию","Remove Version Confirmation":"Подтверждение удаления версии","Remove View":"Удалить","Remove {{ productTable.getSelected().length }} Products?":"Удалить выбранные продукты (всего {{ productTable.getSelected().length }})?","Removed %x host collections from activation key \"%y\".":"Коллекции удалены из ключа активации «%y» (всего %x)","Removed %x host collections from content host \"%y\".":"Коллекции узла «%y» удалены (всего %x)","Removed %x products from sync plan \"%y\".":"Продукты удалены из плана синхронизации «%y» (всего %x)","Removing Content Hosts":"Удаление узлов...","Removing Products":"Удаление продуктов...","Removing Repositories":"Удаление репозиториев...","Repo Discovery":"Поиск репозиториев","Repo Type":"Тип репозитория","Repositories":"Репозитории","Repositories containing Errata {{ errata.errata_id }}":"Репозитории с {{ errata.errata_id }}","Repositories for Content View:":"Репозитории для представления:","Repositories for Errata:":"Репозитории для исправления:","Repositories for Filter:":"Репозитории для фильтра:","Repositories for Product:":"Репозитории для продукта:","Repository":"Репозиторий","Repository \"%s\" successfully deleted":"Репозиторий «%s» удален.","Repository %s successfully created.":"Репозиторий %s создан.","Repository Count":"Число репозиториев","Repository Discovery":"Поиск репозиториев","Repository Name":"Имя репозитория","Repository Saved.":"Репозиторий сохранен.","Repository Selection":"Выбор репозитория","Repository URL updated":"Адрес репозитория обновлен","Repository:":"Репозиторий:","Required Host":"Обязательный узел","Result":"Результат","Results":"Результаты","Rules Source":"Источник правил","Rules Version":"Версия правил","Run Auto-Attach":"Назначить автоматически","Run subscription-manager in a console on the client host.  You may use an Activation Key to register:":"Зарегистрируйте систему клиента с использованием ключа активации:","Save Successful.":" Сохранено.","Search...":"Поиск...","Security":"Безопасность","Security Advisory":"Рекомендация безопас","Select A New Puppet Module To Add":"Выберите модуль Puppet","Select Content Hosts":"Выберите узлы","Select Module":"Выберите модуль","Select Version":"Выберите версию","Select a Content Source:":"Выберите источник содержимого:","Select a Version":"Выберите версию","Select an Available Version of {{ $stateParams.moduleName }}":"Выберите версию {{ $stateParams.moduleName }}","Select an Organization":"Выберите организацию","Select new version":"Выбрать другую версию","Select the Lifecycle Environments you would like to remove Version {{ version.version }} from:":"Выберите окружения, из которых будет удалена версия {{ version.version }}:","Selected URLs:":"URL:","Selecting this option will result in the Version being completely deleted and no longer being available for promotion.  The version must be removed from all Lifecycle Environments in order to select this option.":"Эта версия будет безвозвратно удалена. Прежде чем продолжить, удалите ее из всех окружений.","Serve via HTTP.":"Доступ HTTP","Service Level":"Уровень обслуживания","Severity":"Приоритет","Show affected Activation Keys":"Показать затронутые ключи активации","Show affected Content Hosts":"Показать затронутые узлы","Show from:":"Показать из:","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"Показано: {{ detailsTable.rows.length }} из {{ detailsTable.resource.subtotal }} (всего: {{ detailsTable.resource.total }})","Showing {{ addSubscriptionsTable.rows.length }} of {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} Total) Subscriptions":"Показано:  {{ addSubscriptionsTable.rows.length }} из {{ addSubscriptionsTable.resource.subtotal }} (всего: {{ addSubscriptionsTable.resource.total }})","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"Показано: {{ detailsTable.rows.length }} из {{ detailsTable.resource.subtotal }} (всего: {{ detailsTable.resource.total }})","Showing {{ currentPackagesTable.resource.total }} Packages":"Показано: {{ currentPackagesTable.resource.total }} ","Showing {{ subscriptionsTable.rows.length }} of {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} Total) Subscriptions":"Показано:  {{ subscriptionsTable.rows.length }} из {{ subscriptionsTable.resource.subtotal }} (всего {{ subscriptionsTable.resource.total }})","Sockets":"Сокеты","Sockets: %s":"Сокеты: %s","Solution":"Решение","Some of the Errata shown below may not be installable as they are not in this Content Host's\n          Content View and Lifecycle Environment.  In order to apply such Errata an Incremental Update is required.":"Некоторых исправлений нет в окружениях и представлениях этого узла.\nЧтобы их установить, потребуется выполнить инкрементное обновление.","Stacking ID":"Объединенный ID","Start Date":"Дата начала","Start Time":"Время","Started At":"Время","Starting":"Начало","Starts":"Начало","State":"Статус","Status":"Статус","Subscription Details":"Подписка","Subscription Info":"Сведения о подписке","Subscription Manifest":"Манифест подписок","Subscription Server Details":"Сервер подписок","Subscription Status":"Состояние","Subscriptions":"Подписки","Subscriptions for Activation Key:":"Подписки для ключа активации:","Subscriptions for:":"Подписки для:","Succesfully scheduled package installation":"Установка пакетов назначена","Succesfully scheduled package removal":"Удаление пакетов назначено","Succesfully scheduled package update":"Обновление пакетов назначено","Success":"Успешно","Successfully Scheduled Auto-attach.":"Автоматический подбор назначен","Successfully added %s subscriptions.":"Добавлено подписок: %s","Successfully deleted %cv version %ver.":"Версия %cv %ver удалена.","Successfully initiated deletion of %cv version %ver.":"Началось удаление версии %cv %ver.","Successfully initiated promotion of %cv version %ver to %env.":"Начат перенос версии %cv %ver в %env.","Successfully initiated removal of %cv version %ver.":"Началось удаление версии %cv %ver.","Successfully promoted %cv version %ver to %env":"Версия %cv %ver перенесена в %env.","Successfully published %cv version %ver and promoted to Library":"Версия %cv %ver опубликована в Library.","Successfully removed %cv version %ver from environments: %env":"Версия %cv %ver удалена из %env","Successfully removed %s items.":"Удалено: %s","Successfully removed %s subscriptions.":"Удалено подписок: %s","Successfully removed 1 item.":"Удален один элемент.","Successfully updated subscriptions.":"Подписки обновлены.","Summary":"Сводка","Support Level":"Уровень поддержки","Support Type":"Тип поддержки","Sync Enabled":"Синхронизация","Sync Information":"Информация о синхронизации","Sync Interval":"Интервал","Sync Now":"Синхронизировать","Sync Plan":"План синхронизации","Sync Plan %s has been deleted.":"План %s удален.","Sync Plan Saved":"План синхронизации сохранен.","Sync Plan {{ syncPlan.name }}":"План {{ syncPlan.name }}","Sync Plan:":"План синхронизации:","Sync Plans":"Планы синхронизации","Sync State":"Статус синхронизации","Sync Status":"Синхронизация","Synced manually, no interval set.":"Вручную. Интервал не задан.","Syncing...":"Синхронизация...","Tags":"Теги","Target":"Получатель","Task Details":"Свойства задачи","Tasks":"Задачи","Tasks for Content View:":"Задачи для представления:","Temporary":"Временные","The Activation Keys listed below are currently using this Content View Version.  Before deleting the Version you must move these Activation Keys to a Lifecycle Environment where this Version is not in use.":"Ниже перечислены ключи активации, использующие эту версию представления. Прежде чем удалить версию, перенесите ключи в окружение с другой версией.","The Content Hosts listed below are currently using this Content View Version.  Before removing the version you must move these Content Hosts to an environment where this version is not in use.":"Ниже перечислены узлы, использующие эту версию представления. Прежде чем удалить версию, следует их перенести в окружение с другой версией.","The Content View or Lifecycle Environment needs to be updated in order to make errata available to these hosts.":"Чтобы эти узлы могли получать исправления, необходимо обновить представление или окружение.","The archive of this Version will be deleted.  This Version will not be available once deletion is complete.":"После удаления архива версия будет недоступна.","The archive of this Version will not be deleted.  This Version will still be available to be promoted to a Lifecycle Environment.":"Архив не будет удален. Версия все так же будет доступна для продвижения в другие окружения.","The following actions can be performed on content hosts in this host collection:":"Над узлами в этой коллекции могут выполняться операции:","The host requires being attached to a content view and the environment you have chosen has no content views promoted to it.\n            See the <a href=\"/content_views\">content views page</a> to manage and promote a content view.":"В выбранном окружении нет представлений, с которыми можно было бы связать узел. \nЧтобы добавить представление, перейдите на <a href=\"/content_views\">страницу представлений</a>.","The katello-agent package is required to manage errata on this Host.\n      Please install the katello-agent package to manage errata on\n      {{ contentHost.name }}.":"Для управления исправлениями на {{ contentHost.name }} \nнеобходимо установить katello-agent.","The katello-agent package is required to manage packages on this Host.\n      Please install the katello-agent package to manage packages on\n      {{ contentHost.name }}. To install the package, connect directly to the\n      host from a terminal and use yum.":"Для управления пакетами на {{ contentHost.name }} необходимо \nустановить katello-agent. Подключитесь к узлу из терминала \nи установите пакет при помощи yum.","The page you are attempting to access requires selecting a specific organization.":"Для доступа к этой странице надо выбрать организацию.","The selected environment contains no Content Views, please select a different environment.":"В выбранном окружении нет представлений. Выберите другое окружение.","There are currently no repositories associated with this Content View, you can add some by clicking on the \"Add\" tab above.":"Чтобы добавить репозиторий, перейдите на вкладку «Добавить».","There are currently no repositories to add to this Content View,":"Нет репозиториев для этого представления.","There are no Content Views in this Environment.":"В окружении нет представлений.","There are no Content Views that match the criteria.":"Нет представлений, отвечающих условиям поиска.","There are no Docker Tags that match the criteria.":"Нет образов Docker, отвечающих условиям поиска.","There are no Errata associated with this Content Host to display.":"Нет исправлений для этого узла.","There are no Errata in this organization.  Create one or more Products with Errata to view Errata on this page.":"Нет исправлений. Вы можете создать продукт и добавить исправления.","There are no Errata that match the criteria.":"Нет исправлений, отвечающих условиям поиска.","There are no Errata to display.":"Нет исправлений.","There are no Host Collections available. You can create new Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"Нет коллекций узлов. Чтобы создать коллекцию, в главном меню выберите «Узлы» - «Коллекции».","There are no Packages that match the criteria.":"Нет пакетов, отвечающих условиям поиска.","There are no Puppet Modules that match the criteria.":"Нет модулей Puppet, отвечающих условиям поиска.","There are no Yum Repositories that match the criteria.":"Нет репозиториев Yum, отвечающих условиям поиска.","There is currently an Incremental Update task in progress.  This update must finish before applying existing updates.":"В настоящий момент выполняется инкрементное обновление. Дождитесь его завершения.","There were errors while removing the following Repositories:":"Произошли ошибки при удалении репозиториев:","This Content Host is not attached to any subscriptions that provide content":"Этот узел не связан с подписками, предоставляющими содержимое","This Content View does not have any versions, create your first Content View Version by using the \"Publish New Version\" button on the right.":"Это представление еще не публиковалось. Нажмите «Опубликовать версию».","This Version cannot be deleted from some of the Lifecycle Environments due to one of the reasons below.":"Версия не может быть удалена из некоторых окружений по перечисленным ниже причинам.","This Version is not associated with any Lifecycle Environments.":"Эта версия не сопоставлена окружениям.","This activation key is not associated with any content hosts.":"Этот ключ активации не связан с узлами.","This activation key may be used during system registration. For example:":"Чтобы зарегистрировать систему с использованием этого ключа:","This filter applies only to a subset of repositories in the content view.":"Выборочно","This filter applies to all repositories in the content view (current and future).":"Все репозитории в представлении","This subscription is not associated with any activation keys.":"Подписка не связана с ключами активации.","This subscription is not attached to any content hosts.":"Подписка не связана с узлами.","This version will be removed from:":"Версия будет удалена из:","Time":"Время","Title":"Название","To register a content host to this server, follow these steps.":"Чтобы зарегистрировать узел содержимого на этом сервере, выполните следующие действия.","Topic":"Тема","Type":"Тип","URL":"Адрес","URL of the registry you want to sync. Example: https://registry.hub.docker.com":"Адрес реестра для синхронизации (например: https://registry.hub.docker.com)","URL to discover":"URL","UUID":"UUID","Unattach Sync Plan":"Отменить план","Unattaching...":"Отмена плана синхронизации...","Unlimited Content Hosts:":"Не ограничено","Unregister Content Host":"Отменить регистрацию узла","Unregister Content Host \"{{contentHost.name}}\"?":"Отменить регистрацию узла «{{contentHost.name}}»?","Unregister Content Hosts":"Отменить регистрацию узлов","Unregister {{ table.numSelected  }} Content Hosts?":"Отменить регистрацию {{ table.numSelected  }} узлов?","Update":"Обновить","Update All":"Обновить все","Update Repositories":"Обновить","Update Sync Plan":"Обновить план","Updated":"Обновлено","Updated errata filter -":"Фильтр обновлен:","Updated override to \"%x\".":"Успешно изменено на «%x»","Updating...":"Обновление...","Upload":"Отправить","Upload Docker Image":"Добавить образ Docker","Upload GPG Key":"Добавить ключ GPG","Upload New Manifest":"Добавить манифест","Upload Package":"Добавить пакет","Upload Puppet Module":"Добавить модуль Puppet","Uploading...":"Отправляется...","Upstream Repository Name":"Имя основного репозитория","Upstream Subscription":"Официальная подписка","Upstream Subscription Management Application":"Официальная программа управления подписками","Upstream Type":"Официальный тип","Upstream User":"Официальный пользователь","Use Latest (currently %s)":"Выбрать последнюю (%s)","Use the cancel button on content view selection to revert your environment selection.":"Чтобы вернуться к выбору окружения, нажмите кнопку отмены.","User":"Пользователь","Version":"Версия","Version %s":"Версия %s","Version 1":"Версия 1","Version Deletion":"Удаление версии","Version Details":"Свойства версии","Version {{ version.version }}":"Версия {{ version.version }}","Version:":"Версия:","Versions":"Версии","Versions for Content View:":"Версии представления:","Versions for Puppet Module:":"Версии модуля Puppet:","View Details":"Подробности","Virtual":"Виртуальная","Virtual Guest From":"Виртуальный гость из","Virtual Guests":"Виртуальные гости","Virtual Host":"Виртуальный хост","Weekly on {{ product.sync_plan.sync_date | date:'EEEE' }} at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"Еженедельно в {{ product.sync_plan.sync_date | date:'EEEE' }} в {{ product.sync_plan.sync_date | date:'mediumTime' }} (время сервера)","When Auto Attach is disabled, registering systems will be attached to all associated subscriptions.":"Если автоматический выбор отключен, при регистрации системы будут выбраны только настроенные подписки","Working":"Выполнение","Yes":"Да","Yes (Default)":"Да (по умолчанию)","You cannot remove these repositories because they belong to a Red Hat product.":"Репозитории принадлежат Red Hat и не могут быть удалены.","You cannot remove these repositories because you do not have permission.":"Недостаточно полномочий для удаления этих репозиториев","You cannot remove this product because it is a Red Hat product.":"Продукт Red Hat не может быть удален.","You cannot remove this product because it was published to a content view.":"Продукт опубликован в представлении и не может быть удален.","You cannot remove this product because you do not have permission.":"Недостаточно полномочий для удаления этого продукта.","You cannot remove this repository because it is a Red Hat repository.":"Репозиторий Red Hat не может быть удален.","You cannot remove this repository because it was published to a content view.":"Репозиторий опубликован в представлении и не может быть удален.","You cannot remove this repository because you do not have permission.":"Недостаточно полномочий для удаления этого репозитория.","You currently do not have any Packages installed. You can use Package Actions above this message to install new Packages.":"Нет пакетов. Чтобы установить пакеты, выберите «Управление пакетами».","You currently don't have any Activation Keys, you can add Activation Keys using the button on the right.":"Нет ключей активации. Чтобы добавить ключ, нажмите кнопку справа.","You currently don't have any Content Hosts in this Host Group, you can add Content Hosts after selecting the 'Add' tab.":"Пустая группа. Чтобы добавить узлы, перейдите на вкладку «Добавить».","You currently don't have any Content Hosts, you can create new Content Hosts by selecting Contents Host from main menu and then clicking the button on the right.":"Нет узлов содержимого. Чтобы добавить узел, перейдите на страницу «Узлы» - «Узлы содержимого».","You currently don't have any Content Hosts, you can register one by clicking the button on the right and following the instructions.":"Нет узлов. Чтобы зарегистрировать узел, нажмите кнопку справа.","You currently don't have any Content Views.  A Content View can be added by using the button on the right.":"Нет представлений. Чтобы создать представление, нажмите кнопку справа.","You currently don't have any Docker Tags.":"Нет образов Docker.","You currently don't have any Filters included in this Content View, you can add a new Filter by using the button on the right.":"Нет фильтров. Чтобы добавить фильтр, нажмите кнопку справа.","You currently don't have any Gpg keys, you can add Gpg keys using the button on the right.":"Нет ключей GPG. Чтобы добавить ключ, нажмите кнопку справа.","You currently don't have any Host Collections, you can add Host Collections using the button on the right.":"Нет коллекций узлов. Чтобы добавить коллекцию, нажмите кнопку справа.","You currently don't have any Products associated with this Gpg Key.":"Этот ключ GPG не связан с продуктами.","You currently don't have any Products to subscribe to, you can add Products after selecting 'Products' under 'Content' in the main menu.":"Нет продуктов, на которые можно подписаться. Чтобы их добавить, в главном меню выберите «Содержимое» - «Продукты».","You currently don't have any Products<span bst-feature-flag=\"custom_products\">, you can add Products using the button on the right</span>.":"Нет продуктов<span bst-feature-flag=\"custom_products\">. Чтобы их добавить, нажмите кнопку справа</span>.","You currently don't have any Puppet Modules included in this Content View, you can add Puppet Modules using the button on the right.":"В этом представлении нет модулей Puppet. Чтобы добавить модули, нажмите кнопку справа.","You currently don't have any Repositories associated with this Gpg Key.":"Этот ключ GPG не связан с репозиториями.","You currently don't have any Repositories included in this Product, you can add Repositories using the button on the right.":"Нет репозиториев. Чтобы их добавить, нажмите кнопку справа.","You currently don't have any Subscriptions associated with this Content Host, you can add Subscriptions after selecting the 'Add' tab.":"Нет подписок для этого узла. Перейдите на вкладку «Добавить».","You currently don't have any Subscriptions, you can add Subscriptions by importing a manifest using the button on the right labeled 'Manage Manifest'.":"Нет подписок. Чтобы их добавить, нажмите кнопку «Манифест» и настройте манифест.","You currently don't have any Sync Plan, you can add Sync Plans using the button on the right.":"Нет плана синхронизации. Чтобы его настроить, нажмите кнопку справа.","You currently don't have any Sync Plans.  A Sync Plan can be created by using the button on the right.":"Нет плана синхронизации. Чтобы его настроить, нажмите кнопку справа.","You do not have any Installed Products":"Нет установленных продуктов","You must select a content view in order to save your environment.":"Чтобы сохранить окружение, выберите представление.","You must select a new content view before your change of environment can be saved.":"Прежде чем изменить окружение, выберите новое представление.","You must select a new content view before your change of environment can be saved. Use the cancel button on content view selection to revert your environment selection.":"Прежде чем изменить окружение, выберите новое представление. Чтобы отменить выбор окружения, нажмите кнопку отмены.","You must select at least one Content Host in order to apply Errata.":"Выберите узел, чтобы применить исправления.","You must select at least one repository.":"Необходимо выбрать репозиторий.","Your selected content hosts have changed, it is highly recommended to refresh the calculated list of errata using the 'Refresh Errata' button below.":"Выбранные узлы были изменены. Обновите список исправлений, нажав кнопку «Обновить исправления».","Yum Content":"Содержимое Yum","Yum Metadata Checksum":"Контрольная сумма метаданных Yum","Yum Repo Discovery":"Поиск репозиториев Yum","Yum Repositories":"Репозитории Yum","Yum Repositories <div>{{ library.counts.yum_repositories || 0 }}</div>":"Репозитории Yum <div>{{ library.counts.yum_repositories || 0 }}</div>","Yum metadata generation has been initiated in the background.  Click <a href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Метаданные yum будут созданы в фоновом режиме. <a href=\"{{ taskUrl() }}\">Нажмите</a> для просмотра подробной информации.","add some repositories.":"Добавьте репозитории","daily":"каждый день","here":"здесь","hourly":"каждый час","or authenticate with a username and password:":"или укажите свое имя пользователя и пароль:","subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"":"subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"","task details":"описание задачи","to manage them individually.":".","true":"true","weekly":"каждую неделю","{{ detailsTable.numSelected }} Selected":"Выбрано: {{ detailsTable.numSelected }}","{{ addSubscriptionsTable.numSelected }} Selected":"Выбрано: {{ addSubscriptionsTable.numSelected }}","{{ contentHost.registered_by }}":"{{ contentHost.registered_by }}","{{ detailsTable.numSelected }} Selected":"Выбрано: {{ detailsTable.numSelected }}","{{ contentView.name }} cannot be deleted as one or more Content View Versions are still promoted to a Lifecycle Environment.\n      Each Content View Version must be removed from their Lifecycle Environments before the Content View can be deleted.":"{{ contentView.name }} не может быть удалено, так как его версии опубликованы в окружениях жизненного цикла.\nПрежде чем удалить представление, необходимо удалить все его версии из окружений. ","{{ contentView.puppet_modules.length || 0 }} Puppet Modules":"Модулей Puppet: {{ contentView.puppet_modules.length || 0 }}","{{ contentView.repositories.length || 0 }} Repositories":"Репозиториев: {{ contentView.repositories.length || 0 }}","{{ contentViewVersion.puppet_module_count || 0 }} Puppet Modules":"Модулей Puppet: {{ contentViewVersion.puppet_module_count || 0 }}","{{ contentViewVersion.repositories.length || 0 }} Repositories":"Репозиториев: {{ contentViewVersion.repositories.length || 0 }}","{{ discoveryTable.numSelected }} Selected":"Выбрано: {{ discoveryTable.numSelected }}","{{ environment.name }} Environment":"Окружение {{ environment.name }}","{{ errata.hosts_applicable_count || 0 }} Applicable,":"доступно {{ errata.hosts_applicable_count || 0 }},","{{ errata.hosts_available_count || 0 }} Installable":"можно установить — {{ errata.hosts_available_count || 0 }}","{{ hostCollectionsTable.numSelected }} Selected":"Выбрано: {{ hostCollectionsTable.numSelected }}","{{ package.removeTask.affected_units }} package removed":"Удален {{ package.removeTask.affected_units }} пакет","{{ package.removeTask.affected_units }} packages removed":"Удалено пакетов: {{ package.removeTask.affected_units }}","{{ product.last_sync_words }} Ago ({{ product.sync_status.finish_time | date:'medium' }} Local Time)":"{{ product.last_sync_words }} назад ({{ product.sync_status.finish_time | date:'medium' }} по местному времени)","{{ product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ product.sync_plan.next_sync | date:'medium' }} (время сервера)","{{ productsTable.numSelected }} Selected":"Выбрано: {{ productsTable.numSelected }}","{{ repository.content_counts.docker_image || 0 }} Docker Images":"{{ repository.content_counts.docker_image || 0 }} образов Docker","{{ repository.content_counts.docker_image }} Docker Images":"образы Docker: {{ repository.content_counts.docker_image }}","{{ repository.content_counts.docker_tag || 0 }} Docker Tags":"{{ repository.content_counts.docker_tag || 0 }} тегов Docker","{{ repository.content_counts.docker_tag }} Docker Tags":"теги Docker {{ repository.content_counts.docker_tag }}","{{ repository.content_counts.erratum || 0 }} Errata":"{{ repository.content_counts.erratum || 0 }} исправлений","{{ repository.content_counts.erratum }} Errata":"исправления: {{ repository.content_counts.erratum }}","{{ repository.content_counts.puppet_module || 0 }} Puppet Modules":"{{ repository.content_counts.puppet_module || 0 }} модулей Puppet","{{ repository.content_counts.rpm || 0 }} Packages":"{{ repository.content_counts.rpm || 0 }} пакетов","{{ repository.content_counts.rpm }} Packages":"пакеты: {{ repository.content_counts.rpm }}","{{ repository.last_sync_words }} Ago ({{ repository.last_sync.ended_at | date:'medium' }} Local Time)":"{{ repository.last_sync_words }} назад ({{ repository.last_sync.ended_at | date:'medium' }} по местному времени)","{{ repository.last_sync_words }} ago":"{{ repository.last_sync_words }} назад","{{ repository.product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ repository.product.sync_plan.next_sync | date:'medium' }} (время сервера)","{{ subscription.host.name }}":"{{ subscription.host.name }}","{{ subscriptionsTable.numSelected }} Selected":"Выбрано: {{ subscriptionsTable.numSelected }}","{{ totalActivationKeyCount() }} Activation Keys will be moved to {{ deleteOptions.activationKeys.contentView.name }} in {{ deleteOptions.activationKeys.environment.name }}":"Ключи (всего {{ totalActivationKeyCount() }}) будут перемещены в {{ deleteOptions.activationKeys.contentView.name }} в {{ deleteOptions.activationKeys.environment.name }}","{{ totalHostCount() }} Content Hosts will be moved to {{ deleteOptions.contentHosts.contentView.name }} in {{ deleteOptions.contentHosts.environment.name }}":"Узлы (всего {{ totalHostCount() }}) будут перемещены в {{ deleteOptions.contentHosts.contentView.name }} в {{ deleteOptions.contentHosts.environment.name }}","{{ type.display }}":"{{ type.display }}","{{ version.docker_image_count }} Docker Images":"образы Docker: {{ version.docker_image_count }}","{{ version.docker_tag_count }} Docker Tags":"теги Docker: {{ version.docker_tag_count }}","{{ version.errata_counts.total }} Errata":"исправления: {{ version.errata_counts.total }}","{{ version.package_count }} Packages":"пакеты: {{ version.package_count }}","{{ version.puppet_module_count }} Puppet Modules":"модули Puppet: {{ version.puppet_module_count }}"});
    gettextCatalog.setStrings('zh-CN', {"%(consumed)s out of %(quantity)s":"%(quantity)s 中的 %(consumed)s","(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)":"(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)","(Not all Activation Keys editable )":"（不是所有激活码都可以编辑）","(Not all Content Hosts editable )":"（不是所有内容主机都可以编辑）","* These marked Content View Versions are from Composite Content Views.  Their components needing updating are listed underneath.":"*这些标记的内容视图版本来自综合内容视图。下面列出了这些版本需要更新的组件。","+ New Sync Plan":"+ 新建同步计划","- Date and Type":"- 日期和类型","-- select an interval --":"-- 选择间隔 --","1 Content Host":"1 内容主机","1 repository sync has errors.":"1 储存库同步出错。","1 repository sync in progress.":"1 储存库同步正在进行。","1 successfully synced repository.":"1 成功同步的储存库。","<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Version {{ version.version }}</a>":"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Version {{ version.version }}</a>","<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Create":"<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Create","<span translate=\"\">\n        If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">here</a>\n      <span translate=\"\">to manage them individually.</span>":"<span translate=\"\">\n        If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">here</a>\n      <span translate=\"\">to manage them individually.</span>","A composite view contains other content views.":"包含其他内容视图的综合视图","A new version of {{ contentView.name }} will be created and promoted to the Library environment.\n    It can be promoted to other environments from the Versions tab of this Content View.":"将创建 {{ contentView.name }} 的新版本，并推广至 Library 环境。\n    可在这个内容视图的版本标签中将其推广至其他环境。","Account":"帐户","Account Number":"帐户号","Action":"操作","Action Type":"操作类型","Actions":"操作","Actions for:":"操作：","Activation Key":"激活码","Activation Key Content":"激活码内容","Activation Key Type:":"激活码类型：","Activation Key removed.":"已删除激活码。","Activation Key updated":"已更新激活码","Activation Key:":"激活码：","Activation Key: {{ activationKey.name }}":"激活码：{{ activationKey.name }}","Activation Keys":"激活码","Activation Keys for Subscription:":"用于订阅的激活码：","Activation Keys using Version {{ version.version }}":"使用版本 {{ version.version }} 的激活码","Add":"添加","Add Content Hosts to:":"将内容主机添加到：","Add Content Views":"添加内容视图","Add Errata":"添加勘误","Add New Environment":"添加新环境","Add New Filter":"添加新过滤器","Add New Module":"添加新模块","Add New Repository":"添加新库","Add Package Group":"添加软件包组","Add Repositories":"添加库","Add Selected":"添加选定内容","Add Subscriptions for Activation Key:":"为激活码添加订阅： ","Add To":"添加至","Added %x host collections to activation key \"%y\".":"在激活码 \"%y\" 中添加 %x 主机集合。","Added %x host collections to content host \"%y\".":"在内容主机 \"%y\" 中添加 %x 主机集合。","Added %x products to sync plan \"%y\".":"在同步计划 \"%y\" 中添加 %x 产品。","Adding Lifecycle Environment to the end of \"{{ priorEnvironment.name }}\"":"在 \"{{ priorEnvironment.name }}\" 结尾添加生命周期环境","Advanced Information":"高级信息","Advisory":"公告","Affected Activation Keys":"受影响的激活码","Affected Content Hosts":"受影响的内容主机","Affected Packages":"受影响的软件包","Affected Systems":"受影响的系统","Affected repositories":"受影响的库","Affected repositories have been updated.":"已更新受影响的库","All Content Views":"所有内容视图","All Products":"全部产品","All Repositories":"所有库","All Versions":"所有版本","Alter Sync Plans":"修改同步计划","An error occurred applying Subscriptions:":"应用订阅时出错：","An error occurred removing the Activation Key:":"删除激活码时出错：","An error occurred removing the Host Collection:":"删除主机集合时出错：","An error occurred removing the Products:":"删除产品时出错：","An error occurred removing the content hosts.":"删除内容主机时出错：","An error occurred removing the environment:":"删除环境时出错：","An error occurred removing the subscriptions.":"删除订阅时出错：","An error occurred saving the Activation Key:":"保存激活码时出错：","An error occurred saving the Content Host:":"保存内容主机时出错：","An error occurred saving the Environment:":"保存环境时出错：","An error occurred saving the Filter:":"保存过滤器时出错：","An error occurred saving the Host Collection:":"保存主机集合时出错：","An error occurred saving the Product:":"保存产品时出错：","An error occurred saving the Repository:":"保存库时出错：","An error occurred saving the Sync Plan:":"保存产同步计划时出错：","An error occurred saving the URL:":"保存 URL 时出错：","An error occurred trying to auto-attach subscriptions.  Please check your log for further information.":"尝试自动附加订阅时出错。请查看日志了解具体情况。","An error occurred updating the Content View:":"更新内容视图时出错：","Applicable":"可用","Applicable Content Hosts for Errata":"勘误的可用内容主机","Applicable Errata":"可用勘误","Applicable hosts for {{ erratum.errata_id }}":"{{ erratum.errata_id }} 的可用主机","Apply":"应用","Apply Errata":"应用勘误","Apply Errata to Content Host \"{{contentHost.name}}\"?":"是否在内容主机 \"{{contentHost.name}}\" 中应用勘误？","Apply Errata to Content Hosts":"在内容主机中应用勘误","Apply Errata to Content Hosts immediately after publishing.":"发布后立即在内容主机中应用勘误","Apply Selected":"应用所选内容","Apply to Content Hosts":"在内容主机中应用","Apply to Hosts":"在主机中应用","Apply {{ errata.errata_id }}":"应用 {{ errata.errata_id }}","Apply {{ errata.errata_id }} to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"是否在 {{ numberOfContentHostsToUpdate }} 内容主机中应用 {{ errata.errata_id }}？","Apply {{ errataIds.length }} Errata to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"是否在 {{ numberOfContentHostsToUpdate }} 内容主机中应用 {{ errataIds.length }}？","Applying":"正在应用","Applying Bulk Errata Update":"正在用于批勘误更新","Arch":"归档","Architecture(s)":"架构","Archived Copy":"归档的副本","Are you sure you want to add the {{ table.numSelected }} content host(s) selected to the host collection(s) chosen?":"您确定要在所选主机集合中添加选择的 {{ table.numSelected }} 内容主机吗？","Are you sure you want to apply Errata to content host \"{{ contentHost.name }}\"?":"您确定要在内容主机 \"{{ contentHost.name }}\" 中应用勘误吗？","Are you sure you want to apply the {{ detailsTable.numSelected }} selected errata to the content hosts chosen?":"您确定要在所选内容主机中应用选择的 {{ detailsTable.numSelected }} 勘误吗？","Are you sure you want to assign the {{ table.numSelected }} content host(s) selected to {{ selected.contentView.name }} in {{ selected.environment.name }}?":"您确定要为在 {{ selected.environment.name }} 中选择的 {{ selected.contentView.name }} 分配选择的 {{ table.numSelected }} 内容主机吗？","Are you sure you want to auto-attach available subscriptions to all {{ table.resource.total }} content hosts?":"您确定要在所有 {{ table.resource.total }} 内容主机中自动附加可用订阅吗？","Are you sure you want to delete Content View \"{{ contentView.name }}\"?":"您确定要删除内容视图 \"{{ contentView.name }}\" 吗？","Are you sure you want to install {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"您确定要在所选 {{ getSelectedSystemIds().length }} 系统中安装 {{ content.content }} 吗？","Are you sure you want to remove Activation Key \"{{ activationKey.name }}\"?":"您确定要删除激活码 \"{{ activationKey.name }}\" 吗？","Are you sure you want to remove GPG key {{ gpgKey.name }}?":"您确定要删除 GPG 密钥 {{ gpgKey.name }} 吗？","Are you sure you want to remove Host Collection \"{{ hostCollection.name }}\"?":"您确定要删除主机集合 \"{{ hostCollection.name }}\" 吗？","Are you sure you want to remove Sync Plan \"{{ syncPlan.name }}\"?":"您确定要删除同步计划 {{ syncPlan.name }} 吗？","Are you sure you want to remove product \"{{ product.name }}\"?":"您确定要删除产品 {{ product.name }} 吗？","Are you sure you want to remove repository \"{{ repository.name }}\"?":"您确定您想要删除储存库 \"{{ repository.name }}\" 吗？","Are you sure you want to remove the {{ productTable.getSelected().length }} products(s) selected?":"您确定要删除所选 {{ productTable.getSelected().length }} 产品吗？","Are you sure you want to remove the {{ table.numSelected }} content host(s) selected from the host collection(s) chosen?":"您确定要从所选主机集合中删除选择的 {{ table.numSelected }} 内容主机吗？","Are you sure you want to remove {{ content.content }} from the {{ getSelectedSystemIds().length }} system(s) selected?":"您确定要从所选 {{ getSelectedSystemIds().length }} 系统中删除 {{ content.content }} 吗？","Are you sure you want to unregister content host \"{{ contentHost.name }}\"?":"您确定要取消注册内容主机 \"{{ contentHost.name }}\" 吗？","Are you sure you want to unregister the {{ table.numSelected }} content host(s) selected?":"您确定要取消注册所选内容主机 {{ table.numSelected }} 吗？","Are you sure you want to update {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"您确定要在所选 {{ getSelectedSystemIds().length }} 系统中更新 {{ content.content }} 吗？","As part of this deletion, 1 Content View Version will be deleted.":"作为删除操作的一部分，已删除 1 内容视图版本。","Assign":"分配","Assign Environment and Content View":"分配环境和内容视图","Associations":"关联","At least one Errata needs to be selected to Apply.":"至少要选择一个要应用的勘误。","At least one content host needs to be selected in order to perform a bulk action.":"至少要选择一个内容主机方可执行批量操作。","At least one product needs to be selected in order to perform a bulk action.":"至少要选择一个产品方可执行批量操作。","Attached":"已添加","Attached to Content Hosts":"附加到内容主机","Author":"作者","Auto-Attach":"自动附加","Auto-attach available subscriptions to all content hosts":"将可用订阅自动附加到所有内容主机","Automatic":"自动","Available Content Views for Composite Content View:":"可用于综合内容视图的内容视图：","Available Puppet Modules for Content View:":"内容视图的可用 Puppet 模块：","Back":"返回","Back To Errata List":"返回勘误列表","Back To Repository Selection":"返回库选择","Back to Module List":"返回模块列表","Back to Repository Details":"返回库详情","Back to Repository List":"返回库列表","Back to Version List":"返回版本列表","Basic Information":"基本信息","Below are the product's repository content sets currently available for this subscription. For Red Hat subscriptions, additional content can be made available through the":"以下是目前可用于这个订阅的产品库内容集合。在 Red Hat 订阅中使用的附加内容可通过","Below are the repository content sets currently available for this activation key through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"以下是可通过其订阅在激活码中使用的库内容集合。在 Red Hat 订阅中使用的附加内容可通过","Below are the repository content sets currently available for this content host through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"以下是可通过其订阅在内容主机中使用的库内容集合。在 Red Hat 订阅中使用的附加内容可通过","Bug Fix":"Bug 修复","Bug Fix Advisory":"Bug 修复通知","Bugfix":"Bugfix","Bulk Actions":"批量操作","Bulk Alter Products Sync Plan":"批量更改产品同步计划","Bulk Sync Products":"批量同步产品","CVEs":"CVE","Cancel":"取消","Capacity":"容量","Change assigned Lifecycle Environment or Content View":"更改分配的生命周期环境或者内容视图","Changing default settings for content hosts that register with this activation key requires subscription-manager version 1.10 or newer to be installed on that host.":"要更改使用这个激活码所注册内容主机的默认设置，需要在该主机中安装 subscription-manager  版本 1.0 或者更新的版本。","Changing default settings requires subscription-manager version 1.10 or newer to be installed on this host.":"更改默认设置需要在该主机中安装 subscription-manager  版本 1.0 或者更新的版本。","Checkin":"登录","Checksum":"Checksum","Choose a lifecycle environment from the available promotion paths.":"中可用推广路径中选择生命周期环境。","Click here to check the status of the task.":"点这里查看任务状态。","Click here to refresh.":"点这里刷新。","Click here to see the task for the successful removals.":"点这里查看成功删除的任务。","Click here to select Errata for an Incremental Update.":"点这里为增量更新选择勘误。","Close":"关闭","Collection Actions":"集合操作","Completely remove version?":"完全删除版本？","Components":"组件","Components:":"组件：","Composite Content View {{ contentView.name }}":"综合内容视图 {{ contentView.name }}","Composite View":"综合视图","Composite View?":"综合视图？","Composite?":"合成？","Confirm":"确认","Confirm Remove":"确认删除","Confirm Version Removal: Version {{ version.version }}":"确认版本删除：版本 {{ version.version }}","Confirmation":"确认","Consumed":"已消耗","Content":"内容","Content Counts":"内容计数","Content Download URL":"内容需啊中 URL","Content Host":"内容主机","Content Host %s has been deleted.":"已删除内容主机 %s。","Content Host Bulk Actions":"内容主机批量操作。","Content Host Bulk Content":"内容主机批量内容","Content Host Bulk Errata":"主机内容批量勘误","Content Host Bulk Host Collections":"主机内容批量主机集合","Content Host Bulk Packages":"内容主机批量软件包","Content Host Bulk Subscriptions":"内容主机批量订阅","Content Host Content":"内容主机内容","Content Host Counts":"内容主机计数","Content Host Errata Management":"内容主机勘误管理","Content Host Limit":"内容主机限制","Content Host Package Management":"内容主机软件包管理","Content Host Properties":"内容主机属性","Content Host Registration":"内容主机注册","Content Host Status":"内容主机状态","Content Host Subscriptions":"内容主机订阅","Content Host {{ contentHost.name }}":"内容主机 {{ contentHost.name }}","Content Host:":"内容主机：","Content Hosts":"内容主机","Content Hosts for Activation Key:":"用于激活码的内容主机：","Content Hosts for Subscription:":"用于订阅的内容主机：","Content Hosts for:":"内容主机：","Content Hosts using Version {{ version.version }}":"使用版本 {{ version.version }} 的内容主机","Content Type":"内容类型","Content View":"内容视图","Content View Name":"内容查看名称","Content View Version":"内容视图版本","Content View updated.":"已更新内容视图","Content View:":"内容视图：","Content Views":"内容视图","Content Views <div>{{ library.counts.content_views || 0 }}</div>":"内容视图 <div>{{ library.counts.content_views || 0 }}</div>","Content Views for Composite Content View:":"用于综合内容视图的内容视图：","Content successfully uploaded":"成功上传内容","Contents":"内容","Contract":"合同","Contract Number":"合同号","Copy":"复制","Copy Activation Key":"复制激活码","Copy Collection":"复制集合","Copy View":"复制视图","Cores per Socket":"每个插槽的 core","Cores: %s":"核: %s","Create":"创建","Create New View":"创建新视图","Create Repositories":"创建库","Create Repositories Within:":"创建库于：","Create Repository":"创建库","Create Selected":"创建所选","Create Sync Plan":"创建同步计划","Create new view":"创建新视图","Create view from existing views":"从现有视图创建视图","Critical":"关键","Current Environment (%e/%cv)":"当前环境 (%e/%cv)","Current Subscriptions for Activation Key:":"用于激活码的当前订阅：","Currently Selected Puppet Modules":"当前所选 Puppet 模块","Daily at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"每日于 {{ product.sync_plan.sync_date | date:'mediumTime' }} (服务器时间)","Date":"日期","Default":"默认","Delete":"删除","Delete Manifest":"删除清单","Delete Version {{ version.version }}":"删除版本 {{ version.version }}","Delete {{ contentView.name }}":"删除 {{ contentView.name }}","Deleted from %s":"从 %s 中删除","Deleting Manifest...":"正在删除清单......","Deleting from %count environments.":"正在从 %count 环境中删除。","Deleting from 1 environment.":"正在从 1 环境中删除。","Deletion from %s":"从 %s 中删除","Description":"描述","Detail":"明细","Details":"详情","Details for Activation Key:":"激活码详情：","Details for Content View:":"内容视图详情：","Details for Product:":"产品详情：","Discover":"发现","Discovered URLs":"已发现 URL","Docker Content":"Docker 内容","Docker Images":"Docker 映像","Docker Images in {{ repository.name }}":"{{ repository.name }} 中的 Docker 映像","Docker Repositories":"Docker 库","Docker Repositories <div>{{ library.counts.docker_repositories || 0 }}</div>":"Docker 库 <div>{{ library.counts.docker_repositories || 0 }}</div>","Docker Tag:":"Docker 标签：","Docker Tags":"Docker 标签","Docker metadata generation has been initiated in the background.  Click\n        <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"已知后台启动 Docker 元数据生成。请点\n        <a ng-href=\"{{ taskUrl() }}\">这里</a> 监控此进程。","Edit":"编辑","Enabled?":"已启用？","End Date":"终止日期","Ends":"结束","Enhancement":"加强型","Enter Package Group Name(s)...":"输入软件包组名称......","Enter Package Name(s)...":"输入软件包名称......","Environment":"环境","Environment saved":"已保存环境","Environment: {{ environment.name }}":"环境：{{ environment.name }}","Environments":"环境","Environments List":"环境列表","Environments:":"环境：","Equal To":"等于","Errata":"勘误","Errata <div>{{ library.counts.errata.total || 0 }}</div>":"勘误 <div>{{ library.counts.errata.total || 0 }}</div>","Errata Details":"勘误详情","Errata Filter:":"勘误过滤器：","Errata ID":"勘误 ID","Errata Installation":"勘误安装","Errata List":"勘误列表","Errata Task List":"勘误任务列表","Errata Tasks":"勘误任务","Errata Type":"勘误类型","Errata for:":"勘误使用：","Errata successfully added.":"成功添加勘误。","Errata successfully removed.":"成功删除勘误。","Errata:":"勘误：","Erratum - by Date and Type":"勘误 - 根据日期和类型排列","Erratum - by ID":"勘误 - 根据 ID 排列","Erratum Date Range":"勘误日期范围","Error deleting manifest.":"删除清单时出错。","Error during upload:":"上传时出错：","Error importing manifest.":"导入清单时出错。","Error refreshing manifest.":"刷新清单时出错。","Event":"事件","Events":"事件","Exclude":"排除","Existing Product":"现有产品","Expires":"过期","Failed":"失败的","File too large. Please use the CLI instead.":"文件太大。请使用 CLI。","Filter":"过滤器","Filter Saved":"已保存过滤器","Filter information for:":"过滤器信息：","Filter...":"过滤器......","Filter:":"过滤器:","Filters":"过滤器","Filters List":"过滤器列表","Filters for Content View:":"内容视图的过滤器：","Filters successfully removed.":"成功删除过滤器。","Finished At":"完成于","For older operating systems such as Red Hat Enterprise Linux 5 or CentOS 5 it is recommended to use sha1.":"在老的操作系统，比如 Red Hat Enterprise Linux 5 或 CentOS 5 中，建议使用 sha1。","GPG Key":"GPG 密钥","GPG Key URL":"GPG 密钥 URL","GPG Key successfully uploaded":"成功上传 GPG 密钥","GPG Key {{ gpgKey.name }}":"GPG 密钥 {{ gpgKey.name }}","GPG Keys":"GPG 密钥","General Information":"常规信息","Generated On":"已生成于","Gpg Key updated":"已更新 Gpg 密钥","Gpg Keys":"Gpg 密钥","Greater Than":"大于","Group Install":"组安装","Group Remove":"组删除","Guests of":"虚拟机","Hide affected Activation Keys":"隐藏受影响的激活码","Hide affected Content Hosts":"隐藏受影响的内容主机","History":"历史记录","History for Content View:":"内容视图的历史记录：","Host Collection Management":"主机集合管理 ","Host Collection Membership":"主机集合成员","Host Collection Membership Management":"主机集合成员管理","Host Collection removed.":"已删除主机集合。","Host Collection updated":"已更新主机集合。","Host Collection:":"主机集合：","Host Collection: {{ hostCollection.name }}":"主机集合：{{ hostCollection.name }}","Host Collections":"主机集合","Host Collections for:":"主机集合：","Host Count":"主机计数","Host Group":"主机组","Hostname":"主机名","Hourly at {{ product.sync_plan.sync_date | date:'m' }} minutes and {{ product.sync_plan.sync_date | date:'s' }} seconds":"每小时于{{ product.sync_plan.sync_date | date:'m' }} 分 {{ product.sync_plan.sync_date | date:'s' }} 秒","IPv4 Address":"IPv4 地址","IPv6 Address":"IPv6 地址 ","Id":"Id","If you would prefer to move portions of these Activation Keys to different Content Views or Lifecycle Environments click <a href=\"{{ activationKeyLink() }}\">here</a> to manage them individually.":"如果您要将这些激活码的部分内容移动到其他内容视图或者生命周期环境中，请点击 <a href=\"{{ activationKeyLink() }}\">这里</a> 对其进行单独管理。","If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click":"如果您要将这些激活码的部分内容移动到其他内容视图或者生命周期环境中，请点击","If you would prefer to move portions of these Content Hosts to different content views or environments click <a href=\"{{ contentHostsLink() }}\">here</a>  to manage these Content Hosts in bulk.":"如果您要将这些内容主机的部分内容移动到其他内容视图或者生命周期环境中，请点击 <a href=\"{{ contentHostsLink() }}\">这里</a> 对这些内容主机进行批量管理。","Image":"介质","Image ID":"介质编号","Import History":"导入历史记录","Important":"重要","In order to add a content view to a composite view you must first publish an initial version of the content view.":"要在综合视图中添加内容视图，必须首先发布该内容视图的初始版本。","Include":"包括","Included in Activation Keys":"包含在激活码中","Incremental Update":"增量更新","Install":"安装","Install Selected":"安装所选","Install katello-agent for remote actions and displaying errata information:":"为远程操作安装 katello-agent 并显示勘误信息：","Install the pre-built bootstrap RPM:":"安装预建 bootstrap RPM：","Installable":"可安装","Installable Errata":"可安装勘误","Installed":"已安装","Installed Package":"已安装软件包","Installed Packages":"已安装软件包","Installed Products":"安装的产品","Instance-based":"基于实例的","Interfaces":"接口","Interval":"间隔","Issued":"发出的","Katello Agent":"Katello Agent","Label":"标签","Last Checkin":"上一次登录","Last Published":"上一次发布","Last Puppet Report":"上一个 Puppet 报告","Last Sync":"上一次同步","Last Updated":"上一次更新","Last Updated On":"上一次更新于","Last synced {{ product.last_sync_words }} ago.":"上次同步于 {{ product.last_sync_words }} 前。","Latest (Currently %s)":"最新（当前 %s）","Less Than":"小于","Library":"库程序","Library Synced Content":"与库程序同步的内容","Lifecycle Environment \"{{ priorEnvironment.name }}\" exists within the middle of a path. New Lifecycle Environments can only be added to the end of a path.":"生命周期 \"{{ priorEnvironment.name }}\" 存在于路径中间。只能在路径结尾处添加新的生命周期。","Lifecycle Environment Paths":"生命周期环境路径","Lifecycle Environments":"生命周期环境","Limit":"限制","Limit:":"限制：","Limits":"限制","List/Remove":"列表 / 删除","Loading Environment Paths...":"正在载入环境路径......","Loading...":"载入中......","Manage Docker Images":"管理 Docker 映像","Manage Docker Images for Repository:":"管理库的Docker 映像：","Manage Manifest":"管理清单","Manage Packages":"管理软件包","Manage Packages for Repository:":"为库管理软件包：","Manage Puppet Modules":"管理 Puppet 模块","Manage Puppet Modules for Repository:":"管理库的 Puppet 模块","Manifest History":"清单历史记录","Manifest successfully deleted.":"成功删除清单。","Manifest successfully imported.":"成功导入清单。","Manifest successfully refreshed.":"成功刷新清单。","Message":"信息","Model":"型号","Moderate":"中等","Module %s removed from Content View.":"已从内容视图中删除模块 %s。","Module List":"模块列表","More Details":"详情","More...":"更多......","Multi-entitlement":"多授权","N/A":"无","Name":"名称","Name of the upstream repository you want to sync. Example: 'busybox' or 'fedora/ssh'.":"要同步的 upstream 库名称。例如：'busybox' 或者 'fedora/ssh'。","Networking":"联网","Never synced":"从不同步 ","New Activation Key":"新建激活码","New Content View":"新建内容视图","New Environment":"新建环境","New Environment Path":"新建环境路径","New Filter":"新建过滤器","New Filter for Content View:":"新建内容视图过滤器：","New GPG Key":"新建 GPG 密钥","New Gpg Key":"新建 Gpg 密钥","New Host Collection":"新建主机集合","New Name:":"新建名称：","New Product":"新建产品","New Repository":"新建库","New Sync Plan":"新建同步计划","New sync plan successfully created.":"成功创建新同步计划。","Next":"下一步","Next Sync":"下一次同步","No":"否","No \"Edit\" permissions on some of the Activation Keys in that Lifecycle Environment.":"在那个生命周期环境中的一些激活码没有“Edit”权限。","No \"Edit\" permissions on some of the Content Hosts in that Lifecycle Environment.":"在那个生命周期环境中的一些内容主机没有“Edit”权限。","No \"Promote/Remove\" permissions on the Lifecycle Environment.":"在该生命周期环境中没有“Promote/Remove”权限。","No (Default)":"无（默认）","No Content Hosts are affected.":"没有内容主机受到影响。","No Content Hosts match this Erratum.":"没有与此勘误匹配的内容主机。","No Content Views available, please select another environment.":"没有内容主机可用，请选择其他环境。","No Host Collections to show, you can add Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"无主机集合可显示，您可以在 ‘主机’ 主菜单中选择 ‘主机集合’ 后添加主机集合。","No Host Collections to show, you can add Host Collections after selecting the 'Add' tab.":"无主机集合可显示，您可以选择 ‘添加’ 标签后添加主机集合。","No Products to show.":"无产品显示。","No Repositories contain this Erratum.":"没有包含此勘误的库。","No activation keys are affected.":"没有激活码受到影响。","No alternate release version choices are available. The available releases are based upon what is available in \"{{ contentHost.content_view.name }}\", the selected <a href=\"/content_views\">content view</a> this content host is attached to for the given <a href=\"/lifecycle_environments\">lifecycle environment</a>, \"{{ contentHost.environment.name }}\".":"没有可用的备用发行本选项。可用发行本是基于 \"{{ contentHost.content_view.name }}\" 中可用的发行本，所选 <a href=\"/content_views\">内容视图</a> 是 <a href=\"/lifecycle_environments\">生命周期环境</a>, \"{{ contentHost.environment.name }}\" 附加的内容主机。","No content views exist for {{selected.environment.name}}":"{{selected.environment.name}} 不存在内容视图","No host collections":"无主机集合","No provisioning details are available for this content host.":"没有可用于这个内容主机的置备详情。","No repository content provided through subscriptions.":"没有通过订阅提供的库内容。","No repository content provided through this subscription.":"没有通过这个订阅提供的库内容。","No repository content.":"无库内容。","No subscription manifest imported":"未导入订阅清单","None":"无","Not Synced":"未同步","Not installed":"未安装","Not yet published":"尚未发布","Number of CPUs":"CPU 数","OS":"操作系统","Only show Errata that is Applicable to one or more Content Hosts":"只显示可用于一个或多个内容主机的勘误","Only show Errata that is Installable on one or more Content Hosts":"只显示可在一个或多个内容主机中安装的勘误","Operating System":"操作系统","Options:":"选项：","Organization":"机构","Original Sync Date":"最初同步日期","Override to No":"覆盖为 No","Override to Yes":"覆盖为 Yes","Package":"软件包","Package Actions":"软件包操作","Package Filter:":"软件包过滤器：","Package Group":"软件包组","Package Group Filter:":"软件包组过滤器：","Package Group successfully added.":"成功添加软件包组。","Package Group successfully removed.":"成功删除软件包组。","Package Groups":"软件包组","Package Install":"软件包安装","Package Installation, Removal, and Update":"软件包安装、删除及更新","Package Remove":"软件包删除","Package Update":"软件包更新","Package removal failed":"软件包删除失败","Package successfully added.":"成功添加软件包。","Package successfully removed.":"成功删除软件包。","Package successfully updated.":"成功更新软件包","Package/Group Name":"软件包/组名称","Packages":"软件包","Packages <div>{{ library.counts.packages || 0 }}</div>":"软件包 <div>{{ library.counts.packages || 0 }}</div>","Packages for:":"软件包：","Packages in {{ repository.name }}":"{{ repository.name }} 中的软件包","Parameters":"参数","Paste GPG Key Contents":"粘贴 GPG 密钥内容","Paste contents...":"正在粘贴内容......","Paused":"已暂停","Perform":"执行","Physical":"物理","Please select a Lifecycle Environment and Content View to move the affected Activation Keys to:":"请选择生命周期环境以及内容视图以便将受影响的激活码转移过去：","Please select a Lifecycle Environment and Content View to move these Content Hosts to:":"请选择生命周期环境以及内容视图以便将这些内容主机转移过去：","Please select a Lifecycle Environment.":"请选择生命周期环境。","Please select an environment.":"请选择环境。","Please select one from the list below and you will be redirected.":"请从下面的列表中选择一个环境，然后会将您重新指向该环境。","Press 'Sync Now' below to begin syncing the following product(s).":"按下面的“现在同步”按纽开始同步以下产品。","Previous Environment (%e/%cv)":"上一个环境 (%e/%cv)","Processing":"处理中","Product":"产品","Product Bulk Actions":"产品批量操作","Product Content":"产品内容","Product Count":"产品计数","Product Enhancement Advisory":"产品强化通知","Product ID":"产品 ID","Product Management":"产品管理","Product Management for Sync Plan:":"同步计划的产品管理：","Product Name":"产品名称","Product Saved":"已保存产品","Product Sync":"产品同步 ","Product delete operation has been initiated in the background.  Click <a href=\"{{ taskUrl(productDeletionTaskId) }}\">Here</a> to monitor the progress.":"已在后台启动产品删除操作。请点击  <a href=\"{{ taskUrl(productDeletionTaskId) }}\">这里</a> 监控该进程。","Product information for:":"产品信息：","Product {{ product.name }}":"产品 {{ product.name }}","Products":"产品","Products <div>{{ library.counts.products || 0 }}</div>":"产品 <div>{{ library.counts.products || 0 }}</div>","Promote":"推广","Promote Content View:":"推广内容视图：","Promote Version":"推广版本","Promote Version {{ version.version }}":"推广版本 {{ version.version }}","Promoted to %s":"已推广至 %s","Promoting to %count environments.":"正在推广至 %count 环境。","Promoting to 1 environment.":"推广至 1 环境。","Promotion History":"推广历史记录","Provided Products":"提供的产品","Provisioning Details":"置备详情","Provisioning Host Details":"置备主机详情","Publish Content View:":"发布内容视图：","Publish New Version":"发布新版本","Publish via HTTP":"使用 HTTP 发布","Publish via HTTPS":"使用 HTTPS 发布","Published":"已发布","Published At":"发布于","Published Repository Information":"已发布库信息","Published new version":"已发布新版本","Publishing and promoting to %count environments.":"正在发布并推广至 %count 环境。","Publishing and promoting to 1 environment.":"正在发布并推广至 1 环境。","Puppet Environment":"Puppet 环境","Puppet Modules":"Puppet 模块","Puppet Modules <div>{{ library.counts.puppet_modules || 0 }}</div>":"Puppet 模块 <div>{{ library.counts.puppet_modules || 0 }}</div>","Puppet Modules for Content View:":"内容视图的 Puppet 模块：","Puppet Modules in {{ repository.name }}":"{{ repository.name }} 中的 Puppet 模块","Puppet module added to Content View":"添加到内容视图中的 Puppet 模块","Puppet module metadata generation has been initiated in the background.\n        Click <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"已知后台启动 Puppet 模块元数据生成。\n        请点 <a ng-href=\"{{ taskUrl() }}\">这里</a> 监控此进程。","Quantity":"数量","RAM (GB)":"RAM (GB)","RAM: %s GB":"RAM: %s GB","Range":"范围","Reboot Suggested":"建议重启","Reboot Suggested?":"是否建议重启？","Red Hat CDN URL":"Red Hat CDN URL","Red Hat Provider Details":"Red Hat 供应商详情","Red Hat Repositories page":"Red Hat Repositories 页面启用。","Red Hat Repositories page.":"Red Hat Repositories 页面。","Refresh Manifest":"刷新清单","Refresh Table":"数显 表格","Refreshing Manifest...":"正在刷新清单......","Refreshing manifest not suppported by this manifest.":"这个清单不支持刷新。","Register Content Host":"注册内容主机","Register a Content Host":"注册内容主机","Registered":"已注册","Registered By":"注册人","Registry URL":"注册 URL","Release":"发行本","Release Version":"发行本版本","Remove":"删除","Remove Activation Key \"{{ activationKey.name }}\"?":"是否要删除激活码 \"{{ activationKey.name }}\"？","Remove Content View":"删除内容视图","Remove Content Views":"删除内容视图","Remove Docker Images":"删除 Docker 映像","Remove Errata":"删除勘误","Remove Filters":"删除过滤器","Remove From":"删除","Remove GPG Key":"删除 GPG 密钥","Remove GPG Key {{ gpgKey.name }}":"删除 GPG 密钥 {{ gpgKey.name }}","Remove Host Collection \"{{ hostCollection.name }}\"?":"是否要删除主机集合 \"{{ hostCollection.name }}\"？","Remove Module":"删除模块","Remove Package Group":"删除软件包组","Remove Packages":"删除软件包","Remove Product":"删除产品","Remove Product \"{{ product.name }}\"?":"是否要删除产品 \"{{ product.name }}\"？","Remove Products":"删除产品","Remove Puppet Modules":"删除 Puppet 模块","Remove Repositories":"删除库","Remove Repository":"删除库","Remove Repository \"{{ repository.name }}\"?":"是否要删除储存库 \"{{ repository.name }}\"？","Remove Selected":"删除所选","Remove Successful.":"成功删除。","Remove Sync Plan":"删除同步计划","Remove Sync Plan \"{{ syncPlan.name }}\"?":"是否要删除同步计划 \"{{ syncPlan.name }}\"？","Remove Version":"删除版本","Remove Version Confirmation":"删除版本确认","Remove View":"删除视图","Remove {{ productTable.getSelected().length }} Products?":"是否要删除 {{ productTable.getSelected().length }} 产品？","Removed %x host collections from activation key \"%y\".":"从激活码 \"%y\" 中删除 %x 主机集合。","Removed %x host collections from content host \"%y\".":"从内容主机 \"%y\" 中删除 %x 主机集合。","Removed %x products from sync plan \"%y\".":"从同步计划 \"%y\" 中删除 %x 产品。","Removing Content Hosts":"正在删除内容主机","Removing Products":"正在删除产品","Removing Repositories":"正在删除库","Repo Discovery":"库查找","Repo Type":"库类型","Repositories":"库","Repositories containing Errata {{ errata.errata_id }}":"包含勘误  {{ errata.errata_id }} 的库","Repositories for Content View:":"内容视图使用的库：","Repositories for Errata:":"勘误库：","Repositories for Filter:":"过滤器使用的库：","Repositories for Product:":"产品使用的库：","Repository":"库","Repository \"%s\" successfully deleted":"已成功删除储存库 \"%s\"。","Repository %s successfully created.":"已成功创建 %s 库。","Repository Count":"库计数","Repository Discovery":"查找库","Repository Name":"库名称","Repository Saved.":"已保存库。","Repository Selection":"库选择","Repository URL updated":"已更新库 URL","Repository:":"库：","Required Host":"所需主机","Result":"结果","Results":"结果","Rules Source":"规则源","Rules Version":"规则版本","Run Auto-Attach":"运行 Auto-Attach","Run subscription-manager in a console on the client host.  You may use an Activation Key to register:":"在客户端主机的控制台中运行 subscription-manager。您可以使用激活码注册：","Save Successful.":"成功保存。","Search...":"搜索中......","Security":"安全性","Security Advisory":"安全通知","Select A New Puppet Module To Add":"新选择一个 Puppet 模块添加","Select Content Hosts":"选择内容主机","Select Module":"选择模块","Select Version":"选择版本","Select a Content Source:":"选择一个内容源：","Select a Version":"选择一个版本","Select an Available Version of {{ $stateParams.moduleName }}":"选择一个 {{ $stateParams.moduleName }} 可用版本","Select an Organization":"选择机构","Select new version":"选择新版本","Select the Lifecycle Environments you would like to remove Version {{ version.version }} from:":"选择您要从其中删除版本 {{ version.version }} 的生命周期环境：","Selected URLs:":"所选 URL：","Selecting this option will result in the Version being completely deleted and no longer being available for promotion.  The version must be removed from all Lifecycle Environments in order to select this option.":"选择这个选项会完全删除该版本，使其不再能够推广。必须先从所有生命周期环境中删除该版本方可选择这个选项。","Serve via HTTP.":"通过 HTTP 提供。","Service Level":"服务等级","Severity":"严重性","Show affected Activation Keys":"显示受影响的激活码","Show affected Content Hosts":"显示受影响的内容主机","Show from:":"显示来自：","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"显示 {{ detailsTable.resource.subtotal }} (总计 {{ detailsTable.resource.total }}) 内容主机中的 {{ detailsTable.rows.length }}","Showing {{ addSubscriptionsTable.rows.length }} of {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} Total) Subscriptions":"显示 {{ addSubscriptionsTable.resource.subtotal }} (总计 {{ addSubscriptionsTable.resource.total }}) 订阅中的 {{ addSubscriptionsTable.rows.length }}","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"显示 {{ detailsTable.resource.subtotal }} (总计 {{ detailsTable.resource.total }}) 内容主机中的 {{ detailsTable.rows.length }}","Showing {{ currentPackagesTable.resource.total }} Packages":"显示 {{ currentPackagesTable.resource.total }} 软件包","Showing {{ subscriptionsTable.rows.length }} of {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} Total) Subscriptions":"显示 {{ subscriptionsTable.resource.subtotal }} (总计 {{ subscriptionsTable.resource.total }}) 订阅中的 {{ subscriptionsTable.rows.length }} ","Sockets":"插槽","Sockets: %s":"插槽：%s","Solution":"解决方案","Some of the Errata shown below may not be installable as they are not in this Content Host's\n          Content View and Lifecycle Environment.  In order to apply such Errata an Incremental Update is required.":"下面显示的一些勘误因为不在内容主机的内容视图和生命周期环境中，\n          可能无法安装。应用此类勘误需要进行增量更新。","Stacking ID":"堆栈 ID","Start Date":"起始日期","Start Time":"起始时间","Started At":"开始于：","Starting":"正在开始","Starts":"起始","State":"状态","Status":"状态","Subscription Details":"订阅详情","Subscription Info":"订阅信息","Subscription Manifest":"订阅清单","Subscription Server Details":"订阅服务器详情","Subscription Status":"订阅状态","Subscriptions":"订阅","Subscriptions for Activation Key:":"用于激活码的订阅：","Subscriptions for:":"订阅：","Succesfully scheduled package installation":"成功计划软件包安装","Succesfully scheduled package removal":"成功计划软件包删除","Succesfully scheduled package update":"成功计划软件包更新 ","Success":"成功","Successfully Scheduled Auto-attach.":"成功计划自动添加。","Successfully added %s subscriptions.":"成功添加 %s 订阅。","Successfully deleted %cv version %ver.":"成功删除 %cv 版本 %ver。","Successfully initiated deletion of %cv version %ver.":"成功启动 %cv 版本 %ver。","Successfully initiated promotion of %cv version %ver to %env.":"成功将 %cv 版本从 %ver 推广为 %env。","Successfully initiated removal of %cv version %ver.":"成功删除 %cv 版本 %ver。","Successfully promoted %cv version %ver to %env":"成功将 %cv 版本从 %ver 推广为 %env。","Successfully published %cv version %ver and promoted to Library":"成功发布 %cv 版本 %ver 并将其推广至库中。","Successfully removed %cv version %ver from environments: %env":"成功从环境删除 %cv 版本 %ver： %env","Successfully removed %s items.":"成功删除 %s 个项目。","Successfully removed %s subscriptions.":"成功删除 %s 订阅。","Successfully removed 1 item.":"成功删除 1 个项目。","Successfully updated subscriptions.":"成功更新订阅。","Summary":"小结","Support Level":"支持等级","Support Type":"支持类型","Sync Enabled":"启动同步","Sync Information":"同步信息","Sync Interval":"同步间隔","Sync Now":"现在同步","Sync Plan":"同步计划","Sync Plan %s has been deleted.":"已删除 %s 同步计划。","Sync Plan Saved":"已保存同步计划","Sync Plan {{ syncPlan.name }}":"同步计划 {{ syncPlan.name }}","Sync Plan:":"同步计划：","Sync Plans":"同步计划","Sync State":"同步状态","Sync Status":"同步状态","Synced manually, no interval set.":"手动同步，未设置间隔。","Syncing...":"正在同步......","Tags":"标签","Target":"目标","Task Details":"任务详情","Tasks":"任务","Tasks for Content View:":"内容视图任务：","Temporary":"临时","The Activation Keys listed below are currently using this Content View Version.  Before deleting the Version you must move these Activation Keys to a Lifecycle Environment where this Version is not in use.":"以下激活码目前正在使用这个内容视图版本。删除该版本前必须将这些激活码移动到不使用这个版本的生命周期环境中。","The Content Hosts listed below are currently using this Content View Version.  Before removing the version you must move these Content Hosts to an environment where this version is not in use.":"以下内容主机目前正在使用这个内容视图版本。删除该版本前必须将这些内容主机移动到不使用这个版本的生命周期环境中。","The Content View or Lifecycle Environment needs to be updated in order to make errata available to these hosts.":"要更新内容视图或生命周期环境以便在这些主机中可以使用勘误。","The archive of this Version will be deleted.  This Version will not be available once deletion is complete.":"将删除这个版本的归档。删除后将完全不能使用这个版本。","The archive of this Version will not be deleted.  This Version will still be available to be promoted to a Lifecycle Environment.":"不会删除这个版本的归档。仍可将这个版本推广到生命周期环境中。","The following actions can be performed on content hosts in this host collection:":"可在中国主机集合的内容主机中执行以下操作：","The host requires being attached to a content view and the environment you have chosen has no content views promoted to it.\n            See the <a href=\"/content_views\">content views page</a> to manage and promote a content view.":"主机要求被附加到您选择没有向其推广任何内容视图的内容视图和环境中。\n            有关管理和推广内容视图的详情，请查看 <a href=\"/content_views\">内容视图页面</a> 。","The katello-agent package is required to manage errata on this Host.\n      Please install the katello-agent package to manage errata on\n      {{ contentHost.name }}.":"要求 katello-agent 软件包在这台主机中管理勘误。\n    请安装 katello-agent 软件包，以便\n   在  {{ contentHost.name }} 中管理勘误。","The katello-agent package is required to manage packages on this Host.\n      Please install the katello-agent package to manage packages on\n      {{ contentHost.name }}. To install the package, connect directly to the\n      host from a terminal and use yum.":"要求 katello-agent 软件包在这台主机中管理软件包。\n      请安装 katello-agent 软件包，以便在 \n      {{ contentHost.name }} 中管理软件包。要安装该软件包，请在终端\n     使用 yum 直接连接到该主机。","The page you are attempting to access requires selecting a specific organization.":"需要选择具体机构方可访问您尝试访问的页面。","The selected environment contains no Content Views, please select a different environment.":"所选环境不包含任何内容视图，请选择另一个环境。","There are currently no repositories associated with this Content View, you can add some by clicking on the \"Add\" tab above.":"目前没有任何库与这个内容视图关联，您可以点击上面的“添加”标签添加一些库。","There are currently no repositories to add to this Content View,":"目前没有在这个内容视图中添加任何库。","There are no Content Views in this Environment.":"这个环境中没有任何内容视图。","There are no Content Views that match the criteria.":"没有符合该标准的内容视图。","There are no Docker Tags that match the criteria.":"没有符合该标准的 Docker 标签。","There are no Errata associated with this Content Host to display.":"没有与这个内容主机关联的勘误。","There are no Errata in this organization.  Create one or more Products with Errata to view Errata on this page.":"这个机构中没有勘误。创建一个或多有包含勘误的产品在这个页面中查看勘误。","There are no Errata that match the criteria.":"没有符合该标准的勘误。","There are no Errata to display.":"没有勘误可显示。","There are no Host Collections available. You can create new Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"无主机集合可用，您可以在 ‘主机’ 主菜单中选择 ‘主机集合’ 后添加主机集合。","There are no Packages that match the criteria.":"没有符合该标准的软件包。","There are no Puppet Modules that match the criteria.":"没有符合该标准的 Puppet 模块。","There are no Yum Repositories that match the criteria.":"没有符合该标准的 Yum 储存库。","There is currently an Incremental Update task in progress.  This update must finish before applying existing updates.":"目前没有进行任何增量更新任务。必须在应用现有更新前完成此更新。","There were errors while removing the following Repositories:":"删除以下储存库时出错：","This Content Host is not attached to any subscriptions that provide content":"没有将这个内容主机附加到任何提供内容的订阅中","This Content View does not have any versions, create your first Content View Version by using the \"Publish New Version\" button on the right.":"这个内容视图没有任何版本，请使用右侧的 \"发布新版本\" 按钮创建您的第一个内容视图版本。","This Version cannot be deleted from some of the Lifecycle Environments due to one of the reasons below.":"有以下原因之一即可将这个版本从一些生命周期环境中删除。","This Version is not associated with any Lifecycle Environments.":"这个版本没有与任何生命周期环境关联。","This activation key is not associated with any content hosts.":"这个激活码没有与任何内容主机关联。","This activation key may be used during system registration. For example:":"可在系统注册时使用了这个激活码。例如：","This filter applies only to a subset of repositories in the content view.":"这个过滤器只是用于内容主机中的库子网。","This filter applies to all repositories in the content view (current and future).":"这个过滤器适用于该内容视图中的所有库（现在和将来）。","This subscription is not associated with any activation keys.":"这个订阅没有与任何激活码关联。","This subscription is not attached to any content hosts.":"这个订阅没有附加到任何内容主机中。","This version will be removed from:":"将删除这个版本：","Time":"时间","Title":"标题","To register a content host to this server, follow these steps.":"请按照以下步骤在这台服务器中注册内容主机。","Topic":"标题","Type":"类型","URL":"URL","URL of the registry you want to sync. Example: https://registry.hub.docker.com":"要同步的注册 URL。例如：https://registry.hub.docker.com","URL to discover":"要查找的 URL","UUID":"UUID","Unattach Sync Plan":"取消附加同步计划","Unattaching...":"正在取消附加......","Unlimited Content Hosts:":"无限内容主机：","Unregister Content Host":"取消注册内容主机","Unregister Content Host \"{{contentHost.name}}\"?":"是否取消注册内容主机 \"{{contentHost.name}}\"？","Unregister Content Hosts":"取消注册内容主机","Unregister {{ table.numSelected  }} Content Hosts?":"是否取消注册 {{ table.numSelected  }} 内容主机？","Update":"更新","Update All":"更新全部","Update Repositories":"更新库","Update Sync Plan":"更新同步计划","Updated":"已更新","Updated errata filter -":"已更新","Updated override to \"%x\".":"已更新覆盖至 \"%x\"。","Updating...":"正在更新......","Upload":"上传","Upload Docker Image":"上传 Docket 映像","Upload GPG Key":"上传 GPG 密钥","Upload New Manifest":"上传新清单","Upload Package":"上传软件包","Upload Puppet Module":"上传 puppet 模块","Uploading...":"正在上传......","Upstream Repository Name":"Upstream 库名称","Upstream Subscription":"Upstream 订阅","Upstream Subscription Management Application":"Upstream 订阅管理应用程序","Upstream Type":"Upstream 类型","Upstream User":"Upstream 用户","Use Latest (currently %s)":"最新（当前 %s）","Use the cancel button on content view selection to revert your environment selection.":"在内容试图选择中使用取消按钮恢复环境选择。","User":"用户","Version":"版本","Version %s":"版本 %s","Version 1":"版本 1","Version Deletion":"版本删除","Version Details":"版本详情","Version {{ version.version }}":"版本 {{ version.version }}","Version:":"版本：","Versions":"版本","Versions for Content View:":"内容视图版本：","Versions for Puppet Module:":"Puppet 模块版本：","View Details":"视图详情","Virtual":"虚拟","Virtual Guest From":"虚拟客户机","Virtual Guests":"虚拟客户机","Virtual Host":"虚拟主机","Weekly on {{ product.sync_plan.sync_date | date:'EEEE' }} at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"每周于 {{ product.sync_plan.sync_date | date:'mediumTime' }} (服务器时间) 在 {{ product.sync_plan.sync_date | date:'EEEE' }}","When Auto Attach is disabled, registering systems will be attached to all associated subscriptions.":"禁用自动附加后，会在所有关联订阅中附加正在注册的系统。","Working":"工作中","Yes":"是","Yes (Default)":"是（默认）","You cannot remove these repositories because they belong to a Red Hat product.":"您无法删除这些库因为它们属于 Red Hat 产品。","You cannot remove these repositories because you do not have permission.":"您无法删除这些库因为您没有授权。","You cannot remove this product because it is a Red Hat product.":"您无法删除这个产品因为它是 Red Hat 产品。","You cannot remove this product because it was published to a content view.":"您无法删除这个产品因为是将其发布到内容视图中。","You cannot remove this product because you do not have permission.":"您无法删除这个产品因为您没有授权。","You cannot remove this repository because it is a Red Hat repository.":"您无法删除这个库因为它是 Red Hat 库。","You cannot remove this repository because it was published to a content view.":"您无法删除这个库因为是将其发布到内容视图中。","You cannot remove this repository because you do not have permission.":"您无法删除这个库因为您没有授权。","You currently do not have any Packages installed. You can use Package Actions above this message to install new Packages.":"您尚未安装任何软件包。您可以使用这个信息以前的软件包动作安装新软件包。","You currently don't have any Activation Keys, you can add Activation Keys using the button on the right.":"您目前没有激活码，可使用右边的按钮添加激活码。","You currently don't have any Content Hosts in this Host Group, you can add Content Hosts after selecting the 'Add' tab.":"您目前在这个主机组中没有任何内容主机，可在选择 ‘添加’标签后添加内容主机。","You currently don't have any Content Hosts, you can create new Content Hosts by selecting Contents Host from main menu and then clicking the button on the right.":"您目前没有任何内容主机，可从主菜单中选择内容主机，并点击右侧的按钮生成新的内容主机。","You currently don't have any Content Hosts, you can register one by clicking the button on the right and following the instructions.":"您目前没有任何内容主机，只要点击右侧的按钮并安装说明操作即可注册。","You currently don't have any Content Views.  A Content View can be added by using the button on the right.":"您目前没有任何内容视图，可使用右侧的按钮添加内容视图。","You currently don't have any Docker Tags.":"您目前没有任何 Docker 标签。","You currently don't have any Filters included in this Content View, you can add a new Filter by using the button on the right.":"目前您没有包含在这个内容视图中的过滤器，您可以使用右边的按钮创建新过滤器。","You currently don't have any Gpg keys, you can add Gpg keys using the button on the right.":"您目前没有 Gpg 密钥，可使用右边的按钮添加 Gpg 密钥。","You currently don't have any Host Collections, you can add Host Collections using the button on the right.":"您目前没有任何主机集合，可使用右边的按钮添加主机集合。","You currently don't have any Products associated with this Gpg Key.":"您目前没有任何与这个 Gpg 密钥关联的产品。","You currently don't have any Products to subscribe to, you can add Products after selecting 'Products' under 'Content' in the main menu.":"您目前没有任何要订阅的产品，在主菜单的 ‘内容’ 项下选择 ‘产品’ 后即可添加产品。","You currently don't have any Products<span bst-feature-flag=\"custom_products\">, you can add Products using the button on the right</span>.":"您目前没有任何产品 <span bst-feature-flag=\"custom_products\">，可使用右侧的按钮添加产品</span>。","You currently don't have any Puppet Modules included in this Content View, you can add Puppet Modules using the button on the right.":"目前您没有任何包含在这个内容视图中的 Puppet 模块，您可以使用右边的按钮创添加 Puppet 模块。","You currently don't have any Repositories associated with this Gpg Key.":"您目前没有任何与这个 Gpg 密钥关联的库。","You currently don't have any Repositories included in this Product, you can add Repositories using the button on the right.":"目前您没有任何包含在这个产品中的过库，您可以使用右边的按钮添加库。","You currently don't have any Subscriptions associated with this Content Host, you can add Subscriptions after selecting the 'Add' tab.":"您目前没有任何订阅，可使用右侧标记为 ‘管理清单’ 的按钮导入清单，从而添加订阅。","You currently don't have any Subscriptions, you can add Subscriptions by importing a manifest using the button on the right labeled 'Manage Manifest'.":"您目前没有任何订阅，可使用右侧标记为 ‘管理清单’ 的按钮导入清单，从而添加订阅。","You currently don't have any Sync Plan, you can add Sync Plans using the button on the right.":"您目前没有任何同步计划，可使用右边的按钮添加同步计划。","You currently don't have any Sync Plans.  A Sync Plan can be created by using the button on the right.":"您目前没有任何同步计划，可使用右侧的按钮创建同步计划。","You do not have any Installed Products":"您没有任何安装的产品","You must select a content view in order to save your environment.":"必须选择一个内容视图方可保存环境。","You must select a new content view before your change of environment can be saved.":"更改可以保存的环境前必须选择新内容视图。","You must select a new content view before your change of environment can be saved. Use the cancel button on content view selection to revert your environment selection.":"更改可以保存的环境前必须选择新内容视图。在内容试图选择中使用取消按钮恢复环境选择。","You must select at least one Content Host in order to apply Errata.":"您必须至少选择一个内容主机方可应用更新。","You must select at least one repository.":"您必须至少选择一个库。","Your selected content hosts have changed, it is highly recommended to refresh the calculated list of errata using the 'Refresh Errata' button below.":"已更改所选内容主机。强烈建议您使用下面的“刷新勘误”按钮刷新勘误计算表。","Yum Content":"Yum 内容","Yum Metadata Checksum":"Yum 元数据 Checksum","Yum Repo Discovery":"Yum 库查找","Yum Repositories":"Yum 库","Yum Repositories <div>{{ library.counts.yum_repositories || 0 }}</div>":"Yum 库 <div>{{ library.counts.yum_repositories || 0 }}</div>","Yum metadata generation has been initiated in the background.  Click <a href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"已在后台启动 Yum 元数据生成进程。点击 <a href=\"{{ taskUrl() }}\">这里</a> 监控该进程。","add some repositories.":"添加一些库。","daily":"每日","here":"这里","hourly":"每小时","or authenticate with a username and password:":"或使用用户名和密码认证：","subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"":"subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"","task details":"任务详情","to manage them individually.":"对其进行单独管理。","true":"true","weekly":"每周","{{ detailsTable.numSelected }} Selected":"选择 {{ detailsTable.numSelected }} ","{{ addSubscriptionsTable.numSelected }} Selected":"选择 {{ addSubscriptionsTable.numSelected }} ","{{ contentHost.registered_by }}":"{{ contentHost.registered_by }}","{{ detailsTable.numSelected }} Selected":"选择 {{ detailsTable.numSelected }} ","{{ contentView.name }} cannot be deleted as one or more Content View Versions are still promoted to a Lifecycle Environment.\n      Each Content View Version must be removed from their Lifecycle Environments before the Content View can be deleted.":"无法删除 {{ contentView.name }}，因为在生命周期环境中仍有一个或多个内容视图版本。\n      删除内容视图前，必须将所有内容视图版本从其生命周期中删除。","{{ contentView.puppet_modules.length || 0 }} Puppet Modules":"{{ contentView.puppet_modules.length || 0 }} Puppet 模块","{{ contentView.repositories.length || 0 }} Repositories":"{{ contentView.repositories.length || 0 }} 库","{{ contentViewVersion.puppet_module_count || 0 }} Puppet Modules":"{{ contentViewVersion.puppet_module_count || 0 }} Puppet 模块","{{ contentViewVersion.repositories.length || 0 }} Repositories":"{{ contentViewVersion.repositories.length || 0 }} 库","{{ discoveryTable.numSelected }} Selected":"选择 {{ discoveryTable.numSelected }}","{{ environment.name }} Environment":"{{ environment.name }} 环境","{{ errata.hosts_applicable_count || 0 }} Applicable,":"{{ errata.hosts_applicable_count || 0 }} 可应用，","{{ errata.hosts_available_count || 0 }} Installable":"{{ errata.hosts_available_count || 0 }} 可安装","{{ hostCollectionsTable.numSelected }} Selected":"选择 {{ hostCollectionsTable.numSelected }}","{{ package.removeTask.affected_units }} package removed":"删除 {{ package.removeTask.affected_units }} 软件包 ","{{ package.removeTask.affected_units }} packages removed":"删除 {{ package.removeTask.affected_units }}  软件包","{{ product.last_sync_words }} Ago ({{ product.sync_status.finish_time | date:'medium' }} Local Time)":"{{ product.last_sync_words }} 前 ({{ product.sync_status.finish_time | date:'medium' }} 本地时间)","{{ product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ product.sync_plan.next_sync | date:'medium' }} (服务器时间)","{{ productsTable.numSelected }} Selected":"选择 {{ productsTable.numSelected }} ","{{ repository.content_counts.docker_image || 0 }} Docker Images":"{{ repository.content_counts.docker_image || 0 }} Docker 映像","{{ repository.content_counts.docker_image }} Docker Images":"{{ repository.content_counts.docker_image }} Docker 映像","{{ repository.content_counts.docker_tag || 0 }} Docker Tags":"{{ repository.content_counts.docker_tag || 0 }} Docker 标签","{{ repository.content_counts.docker_tag }} Docker Tags":"{{ repository.content_counts.docker_tag }} Docker 标签","{{ repository.content_counts.erratum || 0 }} Errata":"{{ repository.content_counts.erratum || 0 }} 勘误","{{ repository.content_counts.erratum }} Errata":"{{ repository.content_counts.erratum }} 勘误","{{ repository.content_counts.puppet_module || 0 }} Puppet Modules":"{{ repository.content_counts.puppet_module || 0 }} Puppet 模块","{{ repository.content_counts.rpm || 0 }} Packages":"{{ repository.content_counts.rpm || 0 }} 软件包","{{ repository.content_counts.rpm }} Packages":"{{ repository.content_counts.rpm }} 软件包","{{ repository.last_sync_words }} Ago ({{ repository.last_sync.ended_at | date:'medium' }} Local Time)":"{{ repository.last_sync_words }} 前（{{ repository.last_sync.ended_at | date:'medium' }} 本地时间）","{{ repository.last_sync_words }} ago":"{{ repository.last_sync_words }} 前","{{ repository.product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ repository.product.sync_plan.next_sync | date:'medium' }} (服务器时间)","{{ subscription.host.name }}":"{{ subscription.host.name }}","{{ subscriptionsTable.numSelected }} Selected":"选择 {{ subscriptionsTable.numSelected }}","{{ totalActivationKeyCount() }} Activation Keys will be moved to {{ deleteOptions.activationKeys.contentView.name }} in {{ deleteOptions.activationKeys.environment.name }}":"会将 {{ totalActivationKeyCount() }} 激活码移动到 {{ deleteOptions.activationKeys.contentView.name }} in {{ deleteOptions.activationKeys.environment.name }}","{{ totalHostCount() }} Content Hosts will be moved to {{ deleteOptions.contentHosts.contentView.name }} in {{ deleteOptions.contentHosts.environment.name }}":"会将 {{ totalHostCount() }} 主机内容移动到 {{ deleteOptions.contentHosts.contentView.name }} in {{ deleteOptions.contentHosts.environment.name }}","{{ type.display }}":"{{ type.display }}","{{ version.docker_image_count }} Docker Images":"{{ version.docker_image_count }} Docker 映像","{{ version.docker_tag_count }} Docker Tags":"{{ version.docker_tag_count }} Docker 标签","{{ version.errata_counts.total }} Errata":"{{ version.errata_counts.total }} 勘误","{{ version.package_count }} Packages":"{{ version.package_count }} 软件包","{{ version.puppet_module_count }} Puppet Modules":"{{ version.puppet_module_count }} Puppet 模块"});
    gettextCatalog.setStrings('zh-TW', {"%(consumed)s out of %(quantity)s":"%(consumed)s，總數為 %(quantity)s","(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)":"(<a href=\"/foreman_tasks/tasks/{{ subscription.runningTask.task.id }}\">{{ \"task details\" | translate }}</a>)","(Not all Activation Keys editable )":"（並非所有的啟動金鑰都可以編輯）","(Not all Content Hosts editable )":"（並非所有的內容主機都可以編輯）","* These marked Content View Versions are from Composite Content Views.  Their components needing updating are listed underneath.":"* 這些被標記的內容視域來自於複合式內容視域。它們需要更新的元件列在下方。","+ New Sync Plan":"+ 新的同步計畫","- Date and Type":"- 日期與類型","-- select an interval --":"-- 選擇一個間隔 --","1 Content Host":"{{ contentHost.virtual_guests.length }} 部內容主機","1 repository sync has errors.":"1 個軟體庫同步發生錯誤。","1 repository sync in progress.":"1 個軟體庫同步正在進行中。","1 successfully synced repository.":"1 個軟體庫已成功同步。","<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">Version {{ version.version }}</a>":"<a ui-sref=\"content-views.details.version.details({versionId: version.id})\">版本 {{ version.version }}</a>","<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            Create":"<i class=\"fa fa-spinner fa-spin\" ng-show=\"creating()\"></i>\n            建立","<span translate=\"\">\n        If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">here</a>\n      <span translate=\"\">to manage them individually.</span>":"<span translate=\"\">\n        若您比較希望將部分的這些啟動金鑰移至不同的內容視域或生命週期環境中，請點選\n      </span>\n      <a ui-sref=\"activation-keys.index({})\" translate=\"\">這裡</a>\n      <span translate=\"\">以為它們進行個別管理。</span>","A composite view contains other content views.":"複合視域包含了其它內容視域。","A new version of {{ contentView.name }} will be created and promoted to the Library environment.\n    It can be promoted to other environments from the Versions tab of this Content View.":"一個新版本的 {{ contentView.name }} 將會被建立並推送至函式庫環境中。\n    它亦可透過此內容視域的「版本」分頁來被推送至其它環境中。","Account":"帳號","Account Number":"帳號號碼","Action":"動作","Action Type":"動作類型","Actions":"動作","Actions for:":"給以下的動作：","Activation Key":"啟動金鑰","Activation Key Content":"啟動金鑰的內容","Activation Key Type:":"啟動金鑰類型：","Activation Key removed.":"啟動金鑰已移除。","Activation Key updated":"啟動金鑰已更新","Activation Key:":"啟動金鑰：","Activation Key: {{ activationKey.name }}":"啟動金鑰：{{ activationKey.name }}","Activation Keys":"啟動金鑰","Activation Keys for Subscription:":"訂閱用的啟動金鑰：","Activation Keys using Version {{ version.version }}":"啟動金鑰使用版本 {{ version.version }}","Add":"新增","Add Content Hosts to:":"新增內容主機至：","Add Content Views":"新增內容視域","Add Errata":"新增勘誤","Add New Environment":"新增環境","Add New Filter":"新增篩選器","Add New Module":"新增模組","Add New Repository":"新增軟體庫","Add Package Group":"新增套件群組","Add Repositories":"新增軟體庫","Add Selected":"加入選擇的項目","Add Subscriptions for Activation Key:":"為啟動金鑰新增訂閱：","Add To":"加至","Added %x host collections to activation key \"%y\".":"已新增 %x 主機集項目至啟動金鑰「%y」。","Added %x host collections to content host \"%y\".":"已新增 %x 主機集項目至內容主機「%y」。","Added %x products to sync plan \"%y\".":"已新增 %x 產品至同步計畫「%y」。","Adding Lifecycle Environment to the end of \"{{ priorEnvironment.name }}\"":"新增生命週期環境至 \"{{ priorEnvironment.name }}\" 之後","Advanced Information":"進階資訊","Advisory":"諮詢","Affected Activation Keys":"受影響的啟動金鑰","Affected Content Hosts":"受影響的內容主機","Affected Packages":"受影響的套件","Affected Systems":"受影響的系統","Affected repositories":"受影響的軟體庫","Affected repositories have been updated.":"受影響的軟體庫已更新。","All Content Views":"所有內容視域","All Products":"所有產品","All Repositories":"所有軟體庫","All Versions":"所有版本","Alter Sync Plans":"修改同步計劃","An error occurred applying Subscriptions:":"套用訂閱時發生錯誤：","An error occurred removing the Activation Key:":"移除啟動金鑰時發生錯誤：","An error occurred removing the Host Collection:":"移除主機集項目時發生錯誤：","An error occurred removing the Products:":"移除產品時發生錯誤：","An error occurred removing the content hosts.":"移除內容主機時發生錯誤：","An error occurred removing the environment:":"移除環境時發生錯誤：","An error occurred removing the subscriptions.":"移除訂閱時發生錯誤：","An error occurred saving the Activation Key:":"儲存啟動金鑰時發生錯誤：","An error occurred saving the Content Host:":"儲存內容主機時發生錯誤：","An error occurred saving the Environment:":"儲存環境時發生了錯誤：","An error occurred saving the Filter:":"儲存篩選條件時發生錯誤：","An error occurred saving the Host Collection:":"儲存主機集項目時發生錯誤：","An error occurred saving the Product:":"儲存產品時發生錯誤：","An error occurred saving the Repository:":"儲存軟體庫時發生錯誤：","An error occurred saving the Sync Plan:":"儲存同步計畫時發生錯誤：","An error occurred saving the URL:":"儲存 URL 時發生錯誤：","An error occurred trying to auto-attach subscriptions.  Please check your log for further information.":"試圖自動連接訂閱服務時發生錯誤。請檢查日誌檔，以取得進一步訊息。","An error occurred updating the Content View:":"更新內容視域時發生錯誤：","Applicable":"可套用","Applicable Content Hosts for Errata":"給勘誤使用的可套用內容主機","Applicable Errata":"可套用勘誤","Applicable hosts for {{ erratum.errata_id }}":"給 {{ erratum.errata_id }} 使用的可套用主機","Apply":"套用","Apply Errata":"套用勘誤","Apply Errata to Content Host \"{{contentHost.name}}\"?":"是否要套用勘誤至內容主機 \"{{contentHost.name}}\"？","Apply Errata to Content Hosts":"套用勘誤至內容主機","Apply Errata to Content Hosts immediately after publishing.":"發佈後即刻套用勘誤至內容主機。","Apply Selected":"套用所選項目","Apply to Content Hosts":"套用內容主機","Apply to Hosts":"套用至主機","Apply {{ errata.errata_id }}":"套用 {{ errata.errata_id }}","Apply {{ errata.errata_id }} to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"是否要套用 {{ errata.errata_id }} 至 {{ numberOfContentHostsToUpdate }} 部內容主機？","Apply {{ errataIds.length }} Errata to {{ numberOfContentHostsToUpdate }} Content Host(s)?":"是否要套用 {{ errataIds.length }} 勘誤至 {{ numberOfContentHostsToUpdate }} 部內容主機？","Applying":"正在套用","Applying Bulk Errata Update":"正在套用大批勘誤更新","Arch":"架構","Architecture(s)":"架構","Archived Copy":"已備份的複本","Are you sure you want to add the {{ table.numSelected }} content host(s) selected to the host collection(s) chosen?":"確定要新增所選的 {{ table.numSelected }} 部內容主機至選擇的主機集項目中？","Are you sure you want to apply Errata to content host \"{{ contentHost.name }}\"?":"確定要套用勘誤至內容主機 \"{{ contentHost.name }}\"？","Are you sure you want to apply the {{ detailsTable.numSelected }} selected errata to the content hosts chosen?":"確定要套用所選的 {{ detailsTable.numSelected }} 個勘誤至所選的內容主機？","Are you sure you want to assign the {{ table.numSelected }} content host(s) selected to {{ selected.contentView.name }} in {{ selected.environment.name }}?":"確定要指定所選的 {{ table.numSelected }} 部內容主機至 {{ selected.environment.name }} 的 {{ selected.contentView.name }}？","Are you sure you want to auto-attach available subscriptions to all {{ table.resource.total }} content hosts?":"確定要自動連結所有可用的訂閱至所有的 {{ table.resource.total }} 部內容主機上？","Are you sure you want to delete Content View \"{{ contentView.name }}\"?":"確定要刪除「{{ contentView.name }}」？","Are you sure you want to install {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"確定要在所選的 {{ getSelectedSystemIds().length }} 部系統上安裝 {{ content.content }}？","Are you sure you want to remove Activation Key \"{{ activationKey.name }}\"?":"確定要移除啟動金鑰 \"{{ activationKey.name }}\"？","Are you sure you want to remove GPG key {{ gpgKey.name }}?":"確定要移除 GPG 金鑰 {{ gpgKey.name }}？","Are you sure you want to remove Host Collection \"{{ hostCollection.name }}\"?":"確定要移除主機集項目 \"{{ hostCollection.name }}\"？","Are you sure you want to remove Sync Plan \"{{ syncPlan.name }}\"?":"確定要移除同步計畫 \"{{ syncPlan.name }}\"？","Are you sure you want to remove product \"{{ product.name }}\"?":"確定要移除產品 \"{{ product.name }}\"？","Are you sure you want to remove repository \"{{ repository.name }}\"?":"確定要移除軟體庫 \"{{ repository.name }}\"？","Are you sure you want to remove the {{ productTable.getSelected().length }} products(s) selected?":"確定要移除 {{ productTable.getSelected().length }} 項選擇的產品？","Are you sure you want to remove the {{ table.numSelected }} content host(s) selected from the host collection(s) chosen?":"確定要從所選擇的主機集項目中移除所選擇的 {{ table.numSelected }} 部內容主機？","Are you sure you want to remove {{ content.content }} from the {{ getSelectedSystemIds().length }} system(s) selected?":"確定要從所選擇的 {{ getSelectedSystemIds().length }} 部系統中移除 {{ content.content }}？","Are you sure you want to unregister content host \"{{ contentHost.name }}\"?":"確定要取消內容主機 \"{{ contentHost.name }}\" 的註冊？","Are you sure you want to unregister the {{ table.numSelected }} content host(s) selected?":"確定要取消 {{ table.numSelected }} 部已選主機的註冊？","Are you sure you want to update {{ content.content }} on the {{ getSelectedSystemIds().length }} system(s) selected?":"確定要在所選的 {{ getSelectedSystemIds().length }} 部系統上更新 {{ content.content }}？","As part of this deletion, 1 Content View Version will be deleted.":"本次刪除中包含 {{ versions.length }} 內容視域版本將被移除。","Assign":"指定","Assign Environment and Content View":"指定環境與內容視域","Associations":"相聯性","At least one Errata needs to be selected to Apply.":"需要選擇至少一項勘誤來套用。","At least one content host needs to be selected in order to perform a bulk action.":"需要選擇至少一部內容主機，以進行大批作業。","At least one product needs to be selected in order to perform a bulk action.":"需要選擇至少一種產品，以進行大批作業。","Attached":"已連接","Attached to Content Hosts":"已連接至內容主機","Author":"作者","Auto-Attach":"自動連接","Auto-attach available subscriptions to all content hosts":"自動連接可用的訂閱至所有內容主機","Automatic":"自動","Available Content Views for Composite Content View:":"可供複合內容視域使用的內容視域：","Available Puppet Modules for Content View:":"可供內容視域使用的 Puppet 模組：","Back":"上一步","Back To Errata List":"回到勘誤清單","Back To Repository Selection":"回到選擇軟體庫","Back to Module List":"回到模組清單","Back to Repository Details":"回到軟體庫的詳細資料","Back to Repository List":"回到軟體庫清單","Back to Version List":"回到版本清單","Basic Information":"基本資訊","Below are the product's repository content sets currently available for this subscription. For Red Hat subscriptions, additional content can be made available through the":"以下是目前可以讓此訂閱使用的產品之軟體庫內容。Red Hat 訂閱的額外內容可透過這裡取得：","Below are the repository content sets currently available for this activation key through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"以下是目前可以讓此啟動金鑰透過其訂閱來使用的軟體庫內容。Red Hat 訂閱的額外內容可透過這裡取得：","Below are the repository content sets currently available for this content host through its subscriptions. For Red Hat subscriptions, additional content can be made available through the":"以下是目前可以讓此內容透過其訂閱來使用的軟體庫內容。Red Hat 訂閱的額外內容可透過這裡取得：","Bug Fix":"錯誤修正","Bug Fix Advisory":"錯誤修正諮詢","Bugfix":"錯誤修正","Bulk Actions":"大批動作","Bulk Alter Products Sync Plan":"大批修改產品的同步計畫","Bulk Sync Products":"大批同步產品","CVEs":"CVE","Cancel":"取消","Capacity":"容量","Change assigned Lifecycle Environment or Content View":"變更已指定的生命週期環境或內容視域","Changing default settings for content hosts that register with this activation key requires subscription-manager version 1.10 or newer to be installed on that host.":"為向此註冊金鑰註冊的內容主機改變預設設定，需要在那台主機上安裝 1.10 以上版本的 subscription-manager。","Changing default settings requires subscription-manager version 1.10 or newer to be installed on this host.":"改變預設設定需要在此主機上安裝 1.10 以上版本的 subscription-manager。","Checkin":"簽入","Checksum":"Checksum","Choose a lifecycle environment from the available promotion paths.":"從可用的推送路徑選擇一個生命週期環境。","Click here to check the status of the task.":"點選此處以檢查任務的狀態。","Click here to refresh.":"點選此處以更新。","Click here to see the task for the successful removals.":"點選此處以查看成功移除的任務。","Click here to select Errata for an Incremental Update.":"點選此處以選擇勘誤進行遞增更新。","Close":"關閉","Collection Actions":"蒐集項目的動作","Completely remove version?":"完全移除版本？","Components":"元件","Components:":"元件：","Composite Content View {{ contentView.name }}":"複合內容視域 {{ contentView.name }}","Composite View":"複合視域","Composite View?":"複合視域？","Composite?":"複合式？","Confirm":"確認","Confirm Remove":"確認移除","Confirm Version Removal: Version {{ version.version }}":"確認移除版本：版本 {{ version.version }}","Confirmation":"確認","Consumed":"已消耗","Content":"內容","Content Counts":"內容的數量","Content Download URL":"內容的下載網址","Content Host":"內容主機","Content Host %s has been deleted.":"已偵測到內容主機 %s。","Content Host Bulk Actions":"內容主機的大批動作","Content Host Bulk Content":"內容主機的大批內容","Content Host Bulk Errata":"內容主機的大批勘誤","Content Host Bulk Host Collections":"內容主機的大批主機集項目","Content Host Bulk Packages":"內容主機的大批套件","Content Host Bulk Subscriptions":"內容主機的大批訂閱","Content Host Content":"內容主機的內容","Content Host Counts":"內容主機計數","Content Host Errata Management":"內容主機的勘誤管理","Content Host Limit":"內容主機的限制","Content Host Package Management":"內容主機的套件管理","Content Host Properties":"內容主機的屬性","Content Host Registration":"內容主機的註冊","Content Host Status":"內容主機的狀態","Content Host Subscriptions":"內容主機的訂閱","Content Host {{ contentHost.name }}":"內容主機 {{ contentHost.name }}","Content Host:":"{{ contentHost.name }","Content Hosts":"內容主機","Content Hosts for Activation Key:":"給啟動金鑰的內容主機：","Content Hosts for Subscription:":"訂閱用的內容主機：","Content Hosts for:":"給以下使用的內容主機：","Content Hosts using Version {{ version.version }}":"內容主機使用版本 {{ version.version }}","Content Type":"內容類型","Content View":"內容視域","Content View Name":"內容視域名稱","Content View Version":"內容視域版本","Content View updated.":"內容視域已更新。","Content View:":"內容視域：","Content Views":"內容視域","Content Views <div>{{ library.counts.content_views || 0 }}</div>":"內容視域 <div>{{ library.counts.content_views || 0 }}</div>","Content Views for Composite Content View:":"供複合內容視域使用的內容視域：","Content successfully uploaded":"已成功上傳內容","Contents":"內容","Contract":"合約","Contract Number":"聯絡號碼","Copy":"複製","Copy Activation Key":"複製啟動金鑰","Copy Collection":"複製蒐集項目","Copy View":"複製檢視","Cores per Socket":"每個插槽的核心數","Cores: %s":"核心：%s","Create":"建立","Create New View":"建立新的檢視","Create Repositories":"建立軟體庫","Create Repositories Within:":"建立軟體庫於：","Create Repository":"建立軟體庫","Create Selected":"建立所選項目","Create Sync Plan":"建立同步計畫","Create new view":"建立新的檢視項目","Create view from existing views":"從現有檢視建立檢視項目","Critical":"重要","Current Environment (%e/%cv)":"目前的版本（%e/%cv）","Current Subscriptions for Activation Key:":"供啟動金鑰使用的現有訂閱：","Currently Selected Puppet Modules":"目前所選的 Puppet 模組","Daily at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"每日的 {{ product.sync_plan.sync_date | date:'mediumTime' }}（伺服器時間）","Date":"日期","Default":"預設值","Delete":"刪除","Delete Manifest":"刪除資訊清單","Delete Version {{ version.version }}":"刪除版本 {{ version.version }}","Delete {{ contentView.name }}":"刪除 {{ contentView.name }}","Deleted from %s":"從 %s 刪除","Deleting Manifest...":"刪除清單……","Deleting from %count environments.":"從 %count 個環境中刪除。","Deleting from 1 environment.":"從一個環境中刪除。","Deletion from %s":"由 %s 進行刪除程序","Description":"描述","Detail":"詳細資料","Details":"詳細資料","Details for Activation Key:":"啟動金鑰的詳細資料：","Details for Content View:":"內容視域的詳細資料：","Details for Product:":"產品的詳細資料：","Discover":"尋找","Discovered URLs":"找到的網址","Docker Content":"Docker 內容","Docker Images":"Docker 映像檔","Docker Images in {{ repository.name }}":"{{ repository.name }} 中的 Docker 映像檔","Docker Repositories":"Docker 軟體庫","Docker Repositories <div>{{ library.counts.docker_repositories || 0 }}</div>":"Docker 軟體庫 <div>{{ library.counts.docker_repositories || 0 }}</div>","Docker Tag:":"Docker 標籤：","Docker Tags":"Docker 標籤：","Docker metadata generation has been initiated in the background.  Click\n        <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Docker metadata 產生程序已在背景中進行。請點選\n        <a ng-href=\"{{ taskUrl() }}\">這裡</a>以監控進度。","Edit":"編輯","Enabled?":"啟用？","End Date":"結束日期","Ends":"結束","Enhancement":"增強","Enter Package Group Name(s)...":"輸入套件群組名稱……","Enter Package Name(s)...":"輸入套件名稱……","Environment":"環境","Environment saved":"環境已儲存","Environment: {{ environment.name }}":"環境：{{ environment.name }}","Environments":"環境","Environments List":"環境清單","Environments:":"環境：","Equal To":"相當於","Errata":"勘誤","Errata <div>{{ library.counts.errata.total || 0 }}</div>":"勘誤 <div>{{ library.counts.errata.total || 0 }}</div>","Errata Details":"勘誤的詳細資料","Errata Filter:":"勘誤篩選器：","Errata ID":"勘誤 ID","Errata Installation":"安裝勘誤","Errata List":"勘誤清單","Errata Task List":"勘誤任務清單","Errata Tasks":"勘誤任務","Errata Type":"勘誤類型","Errata for:":"給以下使用的勘誤：","Errata successfully added.":"已成功新增勘誤。","Errata successfully removed.":"已成功移除勘誤。","Errata:":"勘誤：","Erratum - by Date and Type":"勘誤—透過時間與類型","Erratum - by ID":"勘誤—透過 ID","Erratum Date Range":"勘誤的日期範圍","Error deleting manifest.":"刪除清單時發生錯誤。","Error during upload:":"上傳時發生錯誤。","Error importing manifest.":"匯入清單時發生錯誤。","Error refreshing manifest.":"更新清單時發生錯誤。","Event":"事件","Events":"事件","Exclude":"排除","Existing Product":"既有的產品","Expires":"有效期限","Failed":"失敗","File too large. Please use the CLI instead.":"檔案太大。請使用 CLI 來代替。","Filter":"篩選器","Filter Saved":"已儲存篩選器","Filter information for:":"給下列使用的篩選器資訊：","Filter...":"篩選器……","Filter:":"篩選器：","Filters":"篩選器","Filters List":"篩選器清單","Filters for Content View:":"給內容視域用的篩選器：","Filters successfully removed.":"已成功移除篩選器。","Finished At":"完成於","For older operating systems such as Red Hat Enterprise Linux 5 or CentOS 5 it is recommended to use sha1.":"建議在較舊的作業系統上（例如 Red Hat Enterprise Linux 5 或是 CentOS 5）使用 sha1。","GPG Key":"GPG 金鑰","GPG Key URL":"GPG 金鑰的網址","GPG Key successfully uploaded":"已成功上傳 GPG 金鑰","GPG Key {{ gpgKey.name }}":"GPG 金鑰 {{ gpgKey.name }}","GPG Keys":"GPG 金鑰","General Information":"一般資訊","Generated On":"產生於","Gpg Key updated":"已更新 GPG 金鑰","Gpg Keys":"GPG 金鑰","Greater Than":"大於","Group Install":"安裝群組","Group Remove":"移除群組","Guests of":"客座","Hide affected Activation Keys":"隱藏受影響的啟動金鑰","Hide affected Content Hosts":"隱藏受影響的內容主機","History":"紀錄","History for Content View:":"內容視域的歷史：","Host Collection Management":"管理主機集","Host Collection Membership":"主機集的成員","Host Collection Membership Management":"主機集的成員管理","Host Collection removed.":"已移除主機集。","Host Collection updated":"已更新主機集","Host Collection:":"主機集：","Host Collection: {{ hostCollection.name }}":"主機集：{{ hostCollection.name }}","Host Collections":"主機集","Host Collections for:":"給以下使用的主機集：","Host Count":"主機計數","Host Group":"主機群組","Hostname":"主機名稱","Hourly at {{ product.sync_plan.sync_date | date:'m' }} minutes and {{ product.sync_plan.sync_date | date:'s' }} seconds":"每小時於 {{ product.sync_plan.sync_date | date:'m' }} 分 {{ product.sync_plan.sync_date | date:'s' }} 秒","IPv4 Address":"IPv4 位址","IPv6 Address":"IPv6 位址","Id":"Id","If you would prefer to move portions of these Activation Keys to different Content Views or Lifecycle Environments click <a href=\"{{ activationKeyLink() }}\">here</a> to manage them individually.":"如果您想要移動部分啟動金鑰至不同的內容視域或生命週期檢視環境，請點選<a href=\"{{ activationKeyLink() }}\">此處</a>以進行個別管理。","If you would prefer to move portions of these Activation Keys to different content views or Lifecycle Environments click":"如果您想要移動部分啟動金鑰至不同的內容視域或生命週期檢視環境，請點選","If you would prefer to move portions of these Content Hosts to different content views or environments click <a href=\"{{ contentHostsLink() }}\">here</a>  to manage these Content Hosts in bulk.":"如果您想要移動部分內容主機至不同的內容視域或環境，請點選<a href=\"{{ contentHostsLink() }}\">此處</a>來大批管理這些內容主機。","Image":"映像檔","Image ID":"映像檔 ID","Import History":"匯入紀錄","Important":"重要","In order to add a content view to a composite view you must first publish an initial version of the content view.":"要新增內容視域至複合視域，您首先必須出版內容視域的初始版本。","Include":"包含","Included in Activation Keys":"包含於啟動金鑰中","Incremental Update":"遞增的更新","Install":"安裝","Install Selected":"安裝所選項目","Install katello-agent for remote actions and displaying errata information:":"安裝 katello-agent 以進行遠端動作並顯示勘誤資訊：","Install the pre-built bootstrap RPM:":"安裝預建立的 bootstrap RPM：","Installable":"可安裝","Installable Errata":"可安裝的勘誤","Installed":"已安裝","Installed Package":"已安裝的套件","Installed Packages":"已安裝的套件","Installed Products":"已安裝的產品","Instance-based":"以 Instance 為基礎","Interfaces":"介面","Interval":"間隔","Issued":"已簽發","Katello Agent":"Katello 代理程式","Label":"標籤","Last Checkin":"前一次簽入","Last Published":"前一次出版","Last Puppet Report":"前一次 Puppet 報告","Last Sync":"前一次同步","Last Updated":"前一次更新","Last Updated On":"最後更新於","Last synced {{ product.last_sync_words }} ago.":"最後同步於 {{ product.last_sync_words }} 前。","Latest (Currently %s)":"最新的（目前 %s）","Less Than":"小於","Library":"函示庫","Library Synced Content":"函式庫已同步內容","Lifecycle Environment \"{{ priorEnvironment.name }}\" exists within the middle of a path. New Lifecycle Environments can only be added to the end of a path.":"路徑的中央存在著一個生命週期環境 \"{{ priorEnvironment.name }}\"。新的生命週期環境只能新增至一個路徑的尾端。","Lifecycle Environment Paths":"生命週期環境的路徑","Lifecycle Environments":"生命週期環境","Limit":"限制","Limit:":"限制：","Limits":"限制","List/Remove":"列出/移除","Loading Environment Paths...":"正在載入環境路徑...","Loading...":"載入中……","Manage Docker Images":"管理 Docker 映像檔","Manage Docker Images for Repository:":"管理軟體庫的 Docker 映像檔：","Manage Manifest":"管理清單","Manage Packages":"管理套件","Manage Packages for Repository:":"管理軟體庫的套件：","Manage Puppet Modules":"管理 Puppet 模組","Manage Puppet Modules for Repository:":"管理軟體庫的 Puppet 模組：","Manifest History":"清單紀錄","Manifest successfully deleted.":"已成功刪除清單。","Manifest successfully imported.":"已成功匯入清單。","Manifest successfully refreshed.":"已成功更新清單。","Message":"訊息","Model":"型號","Moderate":"控管","Module %s removed from Content View.":"模組 %s 已從內容視域中移除。","Module List":"模組清單","More Details":"更多詳細資訊","More...":"更多…","Multi-entitlement":"多重權利","N/A":"N/A","Name":"名稱","Name of the upstream repository you want to sync. Example: 'busybox' or 'fedora/ssh'.":"您希望同步的上游軟體庫的名稱。例如：'busybox' 或是 'fedora/ssh'。","Networking":"網路","Never synced":"從未同步","New Activation Key":"新增啟動金鑰","New Content View":"新內容視域","New Environment":"新增環境","New Environment Path":"新環境路徑","New Filter":"新的篩選器","New Filter for Content View:":"內容視域的新篩選器：","New GPG Key":"新增 GPG 金鑰","New Gpg Key":"新增 GPG 金鑰","New Host Collection":"新主機集","New Name:":"新名稱：","New Product":"新產品","New Repository":"新軟體庫","New Sync Plan":"新增同步計劃","New sync plan successfully created.":"已建立新的同步計劃。","Next":"下一步","Next Sync":"下個同步","No":"否","No \"Edit\" permissions on some of the Activation Keys in that Lifecycle Environment.":"在該生命週期環境中的部分啟動金鑰上沒有「編輯（edit）」權限。","No \"Edit\" permissions on some of the Content Hosts in that Lifecycle Environment.":"在該生命週期環境中的部分內容主機上沒有「編輯」權限。","No \"Promote/Remove\" permissions on the Lifecycle Environment.":"在生命週期環境上沒有「升階（promote）/移除（Remove）」權限。","No (Default)":"無（預設值）","No Content Hosts are affected.":"沒有內容主機會受影響。","No Content Hosts match this Erratum.":"沒有符合此勘誤的內容主機。","No Content Views available, please select another environment.":"沒有可用的內容視域，請選擇另一個環境。","No Host Collections to show, you can add Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"沒有可顯示的主機集，您可在選擇了主選單中，「主機」下的「主機集」之後新增主機集。","No Host Collections to show, you can add Host Collections after selecting the 'Add' tab.":"沒有可顯示的主機集，您可在選擇了「新增」分頁後新增主機集。","No Products to show.":"沒有可顯示的產品。","No Repositories contain this Erratum.":"沒有包含了此勘誤的軟體庫。","No activation keys are affected.":"沒有會受到影響的啟動金鑰。","No alternate release version choices are available. The available releases are based upon what is available in \"{{ contentHost.content_view.name }}\", the selected <a href=\"/content_views\">content view</a> this content host is attached to for the given <a href=\"/lifecycle_environments\">lifecycle environment</a>, \"{{ contentHost.environment.name }}\".":"沒有其它可供選擇的發行版本。可用的發行版本取決於「{{ contentHost.content_view.name }}」中的內容，在 <a href=\"/lifecycle_environments\">生命週期環境</a> 下，此內容連接至的內容主機的選定 <a href=\"/content_views\">內容視域</a>（{{ contentHost.environment.name }}）。","No content views exist for {{selected.environment.name}}":"{{selected.environment.name}} 沒有存在的內容視域","No host collections":"沒有主機集","No provisioning details are available for this content host.":"此內容主機沒有佈建的詳細資訊。","No repository content provided through subscriptions.":"未透過訂閱提供軟體庫內容。","No repository content provided through this subscription.":"未透過此訂閱提供軟體庫內容。","No repository content.":"沒有軟體庫內容。","No subscription manifest imported":"未匯入訂閱清單","None":"無","Not Synced":"未同步","Not installed":"未安裝","Not yet published":"尚未發佈","Number of CPUs":"CPU 數量","OS":"作業系統","Only show Errata that is Applicable to one or more Content Hosts":"僅顯示適用於一或更多個內容主機的勘誤","Only show Errata that is Installable on one or more Content Hosts":"僅顯示可安裝在一或更多個內容主機上的勘誤","Operating System":"作業系統","Options:":"選項：","Organization":"組織","Original Sync Date":"原始同步日期","Override to No":"覆寫為「否」","Override to Yes":"覆寫為「是」","Package":"套件","Package Actions":"套件動作","Package Filter:":"套件篩選器：","Package Group":"套件群組","Package Group Filter:":"套件群組篩選器：","Package Group successfully added.":"已成功新增套件群組。","Package Group successfully removed.":"已成功移除套件群組。","Package Groups":"套件群組","Package Install":"套件安裝","Package Installation, Removal, and Update":"套件安裝、移除與更新","Package Remove":"套件移除","Package Update":"套件更新","Package removal failed":"套件移除失敗","Package successfully added.":"套件新增成功。","Package successfully removed.":"套件已成功移除。","Package successfully updated.":"套件已成功更新。","Package/Group Name":"套件/群組名稱","Packages":"套件","Packages <div>{{ library.counts.packages || 0 }}</div>":"套件 <div>{{ library.counts.packages || 0 }}</div>","Packages for:":"套件屬於：","Packages in {{ repository.name }}":"{{ repository.name }} 中的套件","Parameters":"參數","Paste GPG Key Contents":"貼上 GPG 金鑰內容","Paste contents...":"貼上內容...","Paused":"已暫停","Perform":"執行","Physical":"實體的","Please select a Lifecycle Environment and Content View to move the affected Activation Keys to:":"請選擇一個生命週期環境和內容視域以將受影響的啟動金鑰移至：","Please select a Lifecycle Environment and Content View to move these Content Hosts to:":"請選擇一個生命週期環境和內容視域以將這些內容主機移至：","Please select a Lifecycle Environment.":"請選擇生命週期環境。","Please select an environment.":"請選擇一個環境。","Please select one from the list below and you will be redirected.":"請從下方清單中選擇一項，而您將會被重新導向。","Press 'Sync Now' below to begin syncing the following product(s).":"請按下下方的「現在同步」以開始同步下列產品。","Previous Environment (%e/%cv)":"先前的環境（%e/%cv）","Processing":"正在處理","Product":"產品","Product Bulk Actions":"產品大批動作","Product Content":"產品內容","Product Count":"產品計數","Product Enhancement Advisory":"產品功能增強諮詢","Product ID":"產品 ID","Product Management":"產品管理","Product Management for Sync Plan:":"產品管理以進行同步計劃：","Product Name":"產品名稱","Product Saved":"產品已儲存","Product Sync":"產品同步","Product delete operation has been initiated in the background.  Click <a href=\"{{ taskUrl(productDeletionTaskId) }}\">Here</a> to monitor the progress.":"產品刪除作業已在背景中啟動。請按<a href=\"{{ taskUrl(productDeletionTaskId) }}\">此</a>來監控進度。","Product information for:":"產品資訊：","Product {{ product.name }}":"產品 {{ product.name }}","Products":"產品","Products <div>{{ library.counts.products || 0 }}</div>":"產品 <div>{{ library.counts.products || 0 }}</div>","Promote":"推送","Promote Content View:":"推送內容視域：","Promote Version":"推送版本","Promote Version {{ version.version }}":"推送版本 {{ version.version }}","Promoted to %s":"已推送至 %s","Promoting to %count environments.":"正在推送至 %count 個環境。","Promoting to 1 environment.":"正在推送至 1 個環境。","Promotion History":"正在推送歷史","Provided Products":"提供的產品","Provisioning Details":"佈建詳細資訊","Provisioning Host Details":"佈建主機詳細資訊","Publish Content View:":"發佈內容視域：","Publish New Version":"發佈新版本","Publish via HTTP":"透過 HTTP 發佈","Publish via HTTPS":"透過 HTTPS 發佈","Published":"已發佈","Published At":"已發佈於","Published Repository Information":"已發佈的軟體庫資訊","Published new version":"發佈新版本","Publishing and promoting to %count environments.":"正在發佈並推送至 %count 個環境。","Publishing and promoting to 1 environment.":"正在發佈並推送至 1 個環境。","Puppet Environment":"Puppet 環境","Puppet Modules":"Puppet 模組","Puppet Modules <div>{{ library.counts.puppet_modules || 0 }}</div>":"Puppet 模組 <div>{{ library.counts.puppet_modules || 0 }}</div>","Puppet Modules for Content View:":"供內容視域使用的 Puppet 模組：","Puppet Modules in {{ repository.name }}":"{{ repository.name }} 中的 Puppet 模組","Puppet module added to Content View":"Puppet 模組已新增至內容視域","Puppet module metadata generation has been initiated in the background.\n        Click <a ng-href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"Puppet 模組的 metadata 產生程序已在背景中進行。\n        請按 <a ng-href=\"{{ taskUrl() }}\">此</a> 來監控進度。","Quantity":"數量","RAM (GB)":"記憶體 (GB)","RAM: %s GB":"記憶體：%s GB","Range":"範圍","Reboot Suggested":"建議重新啟動","Reboot Suggested?":"建議重新啟動？","Red Hat CDN URL":"Red Hat CDN 網址","Red Hat Provider Details":"Red Hat 供應方詳細資訊","Red Hat Repositories page":"Red Hat 軟體庫頁面","Red Hat Repositories page.":"Red Hat 軟體庫頁面。","Refresh Manifest":"更新清單","Refresh Table":"更新表格","Refreshing Manifest...":"正在更新清單...","Refreshing manifest not suppported by this manifest.":"此清單不支援清單更新。","Register Content Host":"註冊內容主機","Register a Content Host":"註冊內容主機","Registered":"已註冊","Registered By":"由...註冊","Registry URL":"登錄檔網址","Release":"發行版","Release Version":"發行版本","Remove":"移除","Remove Activation Key \"{{ activationKey.name }}\"?":"確定要移除啟動金鑰 \"{{ activationKey.name }}\"？","Remove Content View":"移除內容視域","Remove Content Views":"移除內容視域","Remove Docker Images":"移除 Docker 映像檔","Remove Errata":"移除勘誤","Remove Filters":"移除篩選器","Remove From":"從...移除","Remove GPG Key":"移除 GPG 金鑰","Remove GPG Key {{ gpgKey.name }}":"移除 GPG 金鑰 {{ gpgKey.name }}","Remove Host Collection \"{{ hostCollection.name }}\"?":"確定要移除主機集 \"{{ hostCollection.name }}\"？","Remove Module":"移除模組","Remove Package Group":"移除套件群組","Remove Packages":"移除套件","Remove Product":"移除產品","Remove Product \"{{ product.name }}\"?":"確定要移除產品 \"{{ product.name }}\"？","Remove Products":"移除產品","Remove Puppet Modules":"移除 Puppet 模組","Remove Repositories":"移除軟體庫","Remove Repository":"移除軟體庫","Remove Repository \"{{ repository.name }}\"?":"確定要移除軟體庫 \"{{ repository.name }}\"？","Remove Selected":"移除已選擇的項目","Remove Successful.":"移除成功。","Remove Sync Plan":"移除同步計劃","Remove Sync Plan \"{{ syncPlan.name }}\"?":"確定要移除同步計畫 \"{{ syncPlan.name }}\"？","Remove Version":"移除版本","Remove Version Confirmation":"確認移除版本","Remove View":"移除視域","Remove {{ productTable.getSelected().length }} Products?":"確定要移除 {{ productTable.getSelected().length }} 項產品？","Removed %x host collections from activation key \"%y\".":"已從啟動金鑰 \"%y\" 移除了 %x 個主機集。","Removed %x host collections from content host \"%y\".":"已從內容主機 \"%y\" 移除了 %x 個主機集。","Removed %x products from sync plan \"%y\".":"已從同步計劃 \"%y\" 移除了 %x 個產品。","Removing Content Hosts":"移除內容主機","Removing Products":"移除產品","Removing Repositories":"移除軟體庫","Repo Discovery":"搜尋軟體庫","Repo Type":"軟體庫類型","Repositories":"軟體庫","Repositories containing Errata {{ errata.errata_id }}":"包含勘誤 {{ errata.errata_id }} 的軟體庫","Repositories for Content View:":"內容視域的軟體庫：","Repositories for Errata:":"勘誤的軟體庫：","Repositories for Filter:":"篩選器的軟體庫：","Repositories for Product:":"產品的軟體庫：","Repository":"軟體庫","Repository \"%s\" successfully deleted":"已成功刪除軟體庫 \"%s\"","Repository %s successfully created.":"已成功建立軟體庫 %s。","Repository Count":"軟體庫計數","Repository Discovery":"探索軟體庫","Repository Name":"軟體庫名稱","Repository Saved.":"已儲存軟體庫。","Repository Selection":"軟體庫選擇","Repository URL updated":"軟體庫 URL 已更新","Repository:":"軟體庫：","Required Host":"需要的主機","Result":"結果","Results":"結果","Rules Source":"規則來源","Rules Version":"規則版本","Run Auto-Attach":"執行自動連結","Run subscription-manager in a console on the client host.  You may use an Activation Key to register:":"在客戶端主機上的主控台中執行 subscription-manager。您可使用一組啟動金鑰來註冊：","Save Successful.":"儲存成功。","Search...":"搜尋……","Security":"安全性","Security Advisory":"安全性諮詢","Select A New Puppet Module To Add":"選擇欲新增的 Puppet 模組","Select Content Hosts":"選擇內容主機","Select Module":"選擇模組","Select Version":"選擇版本","Select a Content Source:":"選擇一項內容來源：","Select a Version":"選擇版本","Select an Available Version of {{ $stateParams.moduleName }}":"選擇可用的 {{ $stateParams.moduleName }} 版本","Select an Organization":"選擇組織","Select new version":"選擇新版本","Select the Lifecycle Environments you would like to remove Version {{ version.version }} from:":"選擇您希望由哪個生命週期環境移除版本 {{ version.version }}：","Selected URLs:":"已選擇的網址：","Selecting this option will result in the Version being completely deleted and no longer being available for promotion.  The version must be removed from all Lifecycle Environments in order to select this option.":"選擇此選項將會造成版本被完全刪除，並且不再能推送。您必須由所有生命週期環境中移除該版本，才能選擇此選項。","Serve via HTTP.":"透過 HTTP 提供服務。","Service Level":"服務等級","Severity":"嚴重性","Show affected Activation Keys":"顯示受影響的啟動金鑰","Show affected Content Hosts":"顯示受影響的內容主機","Show from:":"由...顯示：","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"顯示 {{ detailsTable.resource.subtotal }} 之 {{ detailsTable.rows.length }} 個（總數為 {{ detailsTable.resource.total }}）個內容主機","Showing {{ addSubscriptionsTable.rows.length }} of {{ addSubscriptionsTable.resource.subtotal }} ({{ addSubscriptionsTable.resource.total }} Total) Subscriptions":"顯示 {{ addSubscriptionsTable.resource.subtotal }} 之 {{ addSubscriptionsTable.rows.length }} 個（總數為 {{ addSubscriptionsTable.resource.total }}）訂閱","Showing {{ detailsTable.rows.length }} of {{ detailsTable.resource.subtotal }} ({{ detailsTable.resource.total }} Total) Content Hosts":"顯示 {{ detailsTable.resource.subtotal }} 之 {{ detailsTable.rows.length }} 個（總數為 {{ detailsTable.resource.total }}）內容主機","Showing {{ currentPackagesTable.resource.total }} Packages":"顯示 {{ currentPackagesTable.resource.total }} 個套件","Showing {{ subscriptionsTable.rows.length }} of {{ subscriptionsTable.resource.subtotal }} ({{ subscriptionsTable.resource.total }} Total) Subscriptions":"顯示 {{ subscriptionsTable.resource.subtotal }} 之 {{ subscriptionsTable.rows.length }} 個（總數為 {{ subscriptionsTable.resource.total }}）訂閱","Sockets":"插槽","Sockets: %s":"插槽：%s","Solution":"解決方案","Some of the Errata shown below may not be installable as they are not in this Content Host's\n          Content View and Lifecycle Environment.  In order to apply such Errata an Incremental Update is required.":"以下部分勘誤可能無法安裝，因為它們不在此內容主機的\n          內容視域以及生命週期環境中。若要套用此種勘誤，您需要進行一項遞增更新。","Stacking ID":"堆疊 ID","Start Date":"起始日期","Start Time":"起始時間","Started At":"起始於","Starting":"正在開始","Starts":"起始","State":"狀態","Status":"狀態","Subscription Details":"訂閱詳細資訊","Subscription Info":"訂閱服務資訊","Subscription Manifest":"訂閱資訊清單","Subscription Server Details":"訂閱伺服器詳細資訊","Subscription Status":"訂閱狀態","Subscriptions":"訂閱服務","Subscriptions for Activation Key:":"啟動金鑰的訂閱：","Subscriptions for:":"訂閱屬於：","Succesfully scheduled package installation":"已成功排程套件安裝","Succesfully scheduled package removal":"已成功排程套件移除","Succesfully scheduled package update":"已成功排程套件更新","Success":"成功","Successfully Scheduled Auto-attach.":"已成功排程自動連接。","Successfully added %s subscriptions.":"已成功新增了 %s 個訂閱。","Successfully deleted %cv version %ver.":"已成功刪除了 %cv 版本 %ver。","Successfully initiated deletion of %cv version %ver.":"已成功開始進行 %cv 版本 %ver 的刪除程序。","Successfully initiated promotion of %cv version %ver to %env.":"已成功開始將 %cv 版本 %ver 推送至 %env。","Successfully initiated removal of %cv version %ver.":"已成功開始 %cv 版本 %ver 的移除程序。","Successfully promoted %cv version %ver to %env":"已成功推送 %cv 版本 %ver 至 %env","Successfully published %cv version %ver and promoted to Library":"已成功發佈 %cv 版本 %ver 並推送至函式庫","Successfully removed %cv version %ver from environments: %env":"已成功由環境 %env 移除了 %cv 版本 %ver","Successfully removed %s items.":"成功移除了 %s 個項目。","Successfully removed %s subscriptions.":"已成功移除了 %s 項訂閱。","Successfully removed 1 item.":"已成功移除了 1 個項目。","Successfully updated subscriptions.":"已成功更新了訂閱。","Summary":"概要","Support Level":"支援等級","Support Type":"支援類型","Sync Enabled":"已啟用同步","Sync Information":"同步資訊","Sync Interval":"同步間隔","Sync Now":"現在同步","Sync Plan":"同步計劃","Sync Plan %s has been deleted.":"已刪除同步計劃 %s。","Sync Plan Saved":"已儲存同步計劃","Sync Plan {{ syncPlan.name }}":"同步計劃 {{ syncPlan.name }}","Sync Plan:":"同步計劃：","Sync Plans":"同步計劃","Sync State":"同步狀態","Sync Status":"同步狀態","Synced manually, no interval set.":"已手動同步，未設置間隔。","Syncing...":"正在同步...","Tags":"標籤","Target":"目標","Task Details":"任務的詳細資訊","Tasks":"任務","Tasks for Content View:":"內容視域的任務：","Temporary":"暫時性","The Activation Keys listed below are currently using this Content View Version.  Before deleting the Version you must move these Activation Keys to a Lifecycle Environment where this Version is not in use.":"列在以下的啟動金鑰目前正使用此內容視域版本。在刪除此版本之前，您必須將這些啟動金鑰移至一個未使用此版本的 Lifecycle Environment 中。","The Content Hosts listed below are currently using this Content View Version.  Before removing the version you must move these Content Hosts to an environment where this version is not in use.":"列在以下的內容主機目前正使用此內容視域版本。在刪除此版本之前，您必須將這些內容主機移至一個未使用此版本的環境中。","The Content View or Lifecycle Environment needs to be updated in order to make errata available to these hosts.":"內容視域或生命週期環境需要更新才能讓這些主機使用勘誤。","The archive of this Version will be deleted.  This Version will not be available once deletion is complete.":"此版本的封存檔案將會被刪除。當刪除程序完成後，此版本將無法使用。","The archive of this Version will not be deleted.  This Version will still be available to be promoted to a Lifecycle Environment.":"此版本的封存檔案將不會被刪除。此版本依然能被推送至生命週期環境中。","The following actions can be performed on content hosts in this host collection:":"下列動作能在此主機集中的內容主機上執行：","The host requires being attached to a content view and the environment you have chosen has no content views promoted to it.\n            See the <a href=\"/content_views\">content views page</a> to manage and promote a content view.":"主機需要連接至一個內容視域，而您所選擇的環境卻沒有推送給它的內容視域。\n            請參閱〈<a href=\"/content_views\">內容視域網頁</a>〉以管理和推送內容視域。","The katello-agent package is required to manage errata on this Host.\n      Please install the katello-agent package to manage errata on\n      {{ contentHost.name }}.":"您需要擁有 katello-agent 套件才能在這部主機上管理勘誤。\n      請安裝 katello-agent 套件以在\n      {{ contentHost.name }} 上管理勘誤。","The katello-agent package is required to manage packages on this Host.\n      Please install the katello-agent package to manage packages on\n      {{ contentHost.name }}. To install the package, connect directly to the\n      host from a terminal and use yum.":"欲管理這部主機上的套件需要使用到 katello-agent 套件。\n      請安裝 katello-agent 套件以管理\n      {{ contentHost.name }} 上的套件。若要安裝套件，請透過終端機直接\n      連上主機，並使用 yum 來進行安裝。","The page you are attempting to access requires selecting a specific organization.":"您嘗試存取的網頁需要選擇特定組織。","The selected environment contains no Content Views, please select a different environment.":"選擇的環境不包含內容視域，請選擇一個不同的環境。","There are currently no repositories associated with this Content View, you can add some by clicking on the \"Add\" tab above.":"目前沒有與此內容視域相聯的軟體庫，您可藉由點選上方的「新增」分頁來新增軟體庫。","There are currently no repositories to add to this Content View,":"目前沒有可加入此內容視域的軟體庫。","There are no Content Views in this Environment.":"此環境中沒有內容視域。","There are no Content Views that match the criteria.":"沒有符合條件的內容視域。","There are no Docker Tags that match the criteria.":"沒有符合條件的 Docker Tag。","There are no Errata associated with this Content Host to display.":"沒有與此內容主機相聯的勘誤可顯示。","There are no Errata in this organization.  Create one or more Products with Errata to view Errata on this page.":"這個組織中沒有勘誤。請建立一或更多項產品與勘誤以在此網站上檢視勘誤。","There are no Errata that match the criteria.":"沒有符合條件的勘誤。","There are no Errata to display.":"沒有可顯示的勘誤。","There are no Host Collections available. You can create new Host Collections after selecting 'Host Collections' under 'Hosts' in main menu.":"沒有可用的主機集。您可在選擇了主選單中，「主機」下的「主機集」之後建立新的主機集。","There are no Packages that match the criteria.":"沒有符合條件的套件。","There are no Puppet Modules that match the criteria.":"沒有符合條件的 Puppet 模組。","There are no Yum Repositories that match the criteria.":"沒有符合條件的 Yum 軟體庫。","There is currently an Incremental Update task in progress.  This update must finish before applying existing updates.":"目前正有一項遞增更新任務進行中。在套用既有更新之前，這項更新必須先完成。","There were errors while removing the following Repositories:":"移除下列軟體庫時發生了錯誤：","This Content Host is not attached to any subscriptions that provide content":"內容主機尚未與任何提供內容的訂閱相聯","This Content View does not have any versions, create your first Content View Version by using the \"Publish New Version\" button on the right.":"這個內容視域沒有任何版本，請藉由使用右方的「發佈新版本」按鈕來建立您的第一個內容視域版本。","This Version cannot be deleted from some of the Lifecycle Environments due to one of the reasons below.":"基於以下原因之一，此版本無法由部分生命週期環境中刪除。","This Version is not associated with any Lifecycle Environments.":"此版本未與任何生命週期環境相聯。","This activation key is not associated with any content hosts.":"此啟動金鑰不與任何內容主機相聯。","This activation key may be used during system registration. For example:":"此啟動金鑰能在進行系統註冊時使用。例如：","This filter applies only to a subset of repositories in the content view.":"篩選器僅適用於內容視域中的軟體庫子集。","This filter applies to all repositories in the content view (current and future).":"此篩選器適用於內容視域（目前及未來）中的所有軟體庫。","This subscription is not associated with any activation keys.":"此訂閱不與任何啟動金鑰相聯。","This subscription is not attached to any content hosts.":"此訂閱未連接至任何內容主機。","This version will be removed from:":"這版本將會由此移除：","Time":"時間","Title":"標題","To register a content host to this server, follow these steps.":"若要向這部伺服器註冊內容主機，請依照下列步驟進行：","Topic":"主題","Type":"類型","URL":"網址","URL of the registry you want to sync. Example: https://registry.hub.docker.com":"您希望與其進行同步的登錄檔之網址。例如 https://registry.hub.docker.com","URL to discover":"欲尋找的網址","UUID":"UUID","Unattach Sync Plan":"取消同步計劃的連接","Unattaching...":"正在取消連接...","Unlimited Content Hosts:":"無限的內容主機：","Unregister Content Host":"尚未註冊的內容主機","Unregister Content Host \"{{contentHost.name}}\"?":"確定要取消內容主機 \"{{contentHost.name}}\" 的註冊？","Unregister Content Hosts":"取消內容主機的註冊","Unregister {{ table.numSelected  }} Content Hosts?":"確定要取消 {{ table.numSelected  }} 部內容主機的註冊？","Update":"更新","Update All":"全部更新","Update Repositories":"更新軟體庫","Update Sync Plan":"更新同步計劃","Updated":"已更新","Updated errata filter -":"更新勘誤篩選器 -","Updated override to \"%x\".":"已更新置換為「%x」。","Updating...":"正在更新...","Upload":"上傳","Upload Docker Image":"上傳 Docker 映像檔","Upload GPG Key":"上傳 GPG 金鑰","Upload New Manifest":"上傳新清單","Upload Package":"上傳套件","Upload Puppet Module":"上傳 Puppet 模組","Uploading...":"正在上傳...","Upstream Repository Name":"上游軟體庫名稱","Upstream Subscription":"上游訂閱","Upstream Subscription Management Application":"上游訂閱管理應用程式","Upstream Type":"上游類型","Upstream User":"上游使用者","Use Latest (currently %s)":"使用最新的（目前為 %s）","Use the cancel button on content view selection to revert your environment selection.":"使用內容視域上的取消按鈕來取消您的環境選項。","User":"使用者","Version":"版本","Version %s":"版本 %s","Version 1":"版本 1","Version Deletion":"版本刪除","Version Details":"版本詳細資訊","Version {{ version.version }}":"版本 {{ version.version }}","Version:":"版本：","Versions":"版本","Versions for Content View:":"內容視域的版本：","Versions for Puppet Module:":"Puppet 模組的版本：","View Details":"檢視詳細資訊","Virtual":"虛擬","Virtual Guest From":"虛擬客座端來自於","Virtual Guests":"虛擬客座端","Virtual Host":"虛擬主機","Weekly on {{ product.sync_plan.sync_date | date:'EEEE' }} at {{ product.sync_plan.sync_date | date:'mediumTime' }} (Server Time)":"每週在 {{ product.sync_plan.sync_date | date:'EEEE' }} 上於 {{ product.sync_plan.sync_date | date:'mediumTime' }}（伺服器時間）","When Auto Attach is disabled, registering systems will be attached to all associated subscriptions.":"當停用了自動連接時，註冊的系統將會被連至所有相聯的訂閱。","Working":"處理","Yes":"是","Yes (Default)":"是（預設值）","You cannot remove these repositories because they belong to a Red Hat product.":"您無法移除這些軟體庫，因為它們屬於某項 Red Hat 產品。","You cannot remove these repositories because you do not have permission.":"您無法移除這些軟體庫，因為您沒有權限。","You cannot remove this product because it is a Red Hat product.":"您無法移除這項產品，因為這是一項 Red Hat 產品。","You cannot remove this product because it was published to a content view.":"您無法移除這項產品，因為它已發佈至某個內容視域。","You cannot remove this product because you do not have permission.":"您無法移除這項產品，因為您沒有權限。","You cannot remove this repository because it is a Red Hat repository.":"您無法移除此軟體庫，因為它是 Red Hat 的軟體庫。","You cannot remove this repository because it was published to a content view.":"您無法移除此軟體庫，因為它已發佈至某個內容視域。","You cannot remove this repository because you do not have permission.":"您無法移除此軟體庫，因為您沒有權限。","You currently do not have any Packages installed. You can use Package Actions above this message to install new Packages.":"您目前尚未安裝任何套件。您可使用此訊息上方的「套件動作」來安裝新套件。","You currently don't have any Activation Keys, you can add Activation Keys using the button on the right.":"您目前沒有任何啟動金鑰，您可使用右方的按鈕來新增啟動金鑰。","You currently don't have any Content Hosts in this Host Group, you can add Content Hosts after selecting the 'Add' tab.":"您目前在這個主機群組中沒有任何內容主機，您可在選擇了「新增」分頁後新增內容主機。","You currently don't have any Content Hosts, you can create new Content Hosts by selecting Contents Host from main menu and then clicking the button on the right.":"您目前沒有任何內容主機，您可藉由從主選單選擇內容主機然後按下右方的按鈕來建立新的內容主機。","You currently don't have any Content Hosts, you can register one by clicking the button on the right and following the instructions.":"您目前沒有任何內容主機，您可藉由點選右方的按鈕並依照指示進行，以註冊內容主機。","You currently don't have any Content Views.  A Content View can be added by using the button on the right.":"您目前沒有任何內容視域。內容視域能藉由使用右方的按鈕來新增。","You currently don't have any Docker Tags.":"您目前沒有任何 Docker Tag。","You currently don't have any Filters included in this Content View, you can add a new Filter by using the button on the right.":"您目前在此內容視域中沒有任何篩選器，您可藉由使用右方的按鈕來新增篩選器。","You currently don't have any Gpg keys, you can add Gpg keys using the button on the right.":"您目前沒有任何 GPG 金鑰，您可藉由使用右方的按鈕來新增 GPG 金鑰。","You currently don't have any Host Collections, you can add Host Collections using the button on the right.":"您目前沒有任何主機集，您可藉由使用右方的按鈕來新增主機集。","You currently don't have any Products associated with this Gpg Key.":"您目前沒有任何與這組 GPG 金鑰相聯的產品。","You currently don't have any Products to subscribe to, you can add Products after selecting 'Products' under 'Content' in the main menu.":"您目前沒有任何能訂閱的產品，您可在選擇了主選單中「內容」下的「產品」之後來新增產品。","You currently don't have any Products<span bst-feature-flag=\"custom_products\">, you can add Products using the button on the right</span>.":"您目前沒有任何產品<span bst-feature-flag=\"custom_products\">，您可藉由使用右方的的按鈕來新增產品。</span>。","You currently don't have any Puppet Modules included in this Content View, you can add Puppet Modules using the button on the right.":"您目前沒有任何 Puppet 模組包含在此內容視域中，您可藉由使用右方的按鈕來新增 Puppet 模組。","You currently don't have any Repositories associated with this Gpg Key.":"您目前沒有任何與這組 GPG 金鑰相聯的軟體庫。","You currently don't have any Repositories included in this Product, you can add Repositories using the button on the right.":"您目前沒有任何軟體庫包含在這項產品中，您可藉由使用右方的按鈕來新增軟體庫。","You currently don't have any Subscriptions associated with this Content Host, you can add Subscriptions after selecting the 'Add' tab.":"您目前沒有任何與此內容主機相聯的訂閱，您可在選擇了「新增」分頁後新增訂閱。","You currently don't have any Subscriptions, you can add Subscriptions by importing a manifest using the button on the right labeled 'Manage Manifest'.":"您目前沒有任何訂閱，您可藉由使用右方標記了「管理清單」的按鈕來匯入清單以新增訂閱。","You currently don't have any Sync Plan, you can add Sync Plans using the button on the right.":"您目前沒有任何同步計劃，您可使用右方的按鈕來新增同步計劃。","You currently don't have any Sync Plans.  A Sync Plan can be created by using the button on the right.":"您目前沒有任何同步計畫。同步計畫能藉由使用右方的按鈕來建立。","You do not have any Installed Products":"您尚未安裝任何產品","You must select a content view in order to save your environment.":"您必須選擇一項內容視域以儲存您的環境。","You must select a new content view before your change of environment can be saved.":"若要儲存您的環境變更，您必須先選擇新的內容視域。","You must select a new content view before your change of environment can be saved. Use the cancel button on content view selection to revert your environment selection.":"若要儲存您的環境變更，您必須先選擇新的內容視域。請使用內容視域上的取消按鈕來復原您的環境選擇。","You must select at least one Content Host in order to apply Errata.":"您必須選擇至少一個內容主機才能套用勘誤。","You must select at least one repository.":"您至少必須選擇一個軟體庫。","Your selected content hosts have changed, it is highly recommended to refresh the calculated list of errata using the 'Refresh Errata' button below.":"您所選擇的內容視域已遭到更改，強烈建議您透過以下的「更新勘誤」按鈕來更新勘誤清單的計數。","Yum Content":"Yum 內容","Yum Metadata Checksum":"Yum Metadata Checksum","Yum Repo Discovery":"尋找 Yum 軟體庫","Yum Repositories":"Yum 軟體庫","Yum Repositories <div>{{ library.counts.yum_repositories || 0 }}</div>":"Yum 軟體庫 <div>{{ library.counts.yum_repositories || 0 }}</div>","Yum metadata generation has been initiated in the background.  Click <a href=\"{{ taskUrl() }}\">Here</a> to monitor the progress.":"已在背景開始了 Yum metadata 的產生。請按<a href=\"{{ taskUrl() }}\">此</a>來監控進度。","add some repositories.":"新增一些軟體庫。","daily":"每日","here":"這裡","hourly":"每小時","or authenticate with a username and password:":"或透過一組使用者名稱和密碼來進行認證：","subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"":"subscription-manager register --org=\"{{ activationKey.organization.label }}\" --activationkey=\"{{ activationKey.name }}\"","task details":"任務詳細資料","to manage them individually.":"來為它們進行個別的管理。","true":"true","weekly":"每週","{{ detailsTable.numSelected }} Selected":"已選擇了 {{ detailsTable.numSelected }}","{{ addSubscriptionsTable.numSelected }} Selected":"已選擇了 {{ addSubscriptionsTable.numSelected }}","{{ contentHost.registered_by }}":"{{ contentHost.registered_by }}","{{ detailsTable.numSelected }} Selected":"已選擇了 {{ detailsTable.numSelected }}","{{ contentView.name }} cannot be deleted as one or more Content View Versions are still promoted to a Lifecycle Environment.\n      Each Content View Version must be removed from their Lifecycle Environments before the Content View can be deleted.":"{{ contentView.name }} 無法刪除，因為一或更多個內容視域版本依然正在推送至某個生命週期環境中。\n      每個內容視域版本皆必須由其生命週期環境中移除，之後內容視域才能被刪除。","{{ contentView.puppet_modules.length || 0 }} Puppet Modules":"{{ contentView.puppet_modules.length || 0 }} 個 Puppet 模組","{{ contentView.repositories.length || 0 }} Repositories":"{{ contentView.repositories.length || 0 }} 個軟體庫","{{ contentViewVersion.puppet_module_count || 0 }} Puppet Modules":"{{ contentViewVersion.puppet_module_count || 0 }} 個 Puppet 模組","{{ contentViewVersion.repositories.length || 0 }} Repositories":"{{ contentViewVersion.repositories.length || 0 }} 個軟體庫","{{ discoveryTable.numSelected }} Selected":"選擇了 {{ discoveryTable.numSelected }}","{{ environment.name }} Environment":"{{ environment.name }} 環境","{{ errata.hosts_applicable_count || 0 }} Applicable,":"{{ errata.hosts_applicable_count || 0 }} 個可套用，","{{ errata.hosts_available_count || 0 }} Installable":"{{ errata.hosts_available_count || 0 }} 個可安裝","{{ hostCollectionsTable.numSelected }} Selected":"選擇了 {{ hostCollectionsTable.numSelected }}","{{ package.removeTask.affected_units }} package removed":"移除了 {{ package.removeTask.affected_units }} 套件","{{ package.removeTask.affected_units }} packages removed":"移除了 {{ package.removeTask.affected_units }} 套件","{{ product.last_sync_words }} Ago ({{ product.sync_status.finish_time | date:'medium' }} Local Time)":"{{ product.last_sync_words }} 之前（{{ product.sync_status.finish_time | date:'medium' }} 本機時間）","{{ product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ product.sync_plan.next_sync | date:'medium' }}（伺服器時間）","{{ productsTable.numSelected }} Selected":"選擇了 {{ productsTable.numSelected }}","{{ repository.content_counts.docker_image || 0 }} Docker Images":"{{ repository.content_counts.docker_image || 0 }} 個 Docker 映像檔","{{ repository.content_counts.docker_image }} Docker Images":"{{ repository.content_counts.docker_image }} 個 Docker 映像檔","{{ repository.content_counts.docker_tag || 0 }} Docker Tags":"{{ repository.content_counts.docker_tag || 0 }} 個 Docker 標籤","{{ repository.content_counts.docker_tag }} Docker Tags":"{{ repository.content_counts.docker_tag }} 個 Docker 標籤","{{ repository.content_counts.erratum || 0 }} Errata":"{{ repository.content_counts.erratum || 0 }} 個勘誤","{{ repository.content_counts.erratum }} Errata":"{{ repository.content_counts.erratum }} 個勘誤","{{ repository.content_counts.puppet_module || 0 }} Puppet Modules":"{{ repository.content_counts.puppet_module || 0 }} 個 Puppet 模組","{{ repository.content_counts.rpm || 0 }} Packages":"{{ repository.content_counts.rpm || 0 }} 個套件","{{ repository.content_counts.rpm }} Packages":"{{ repository.content_counts.rpm }} 個套件","{{ repository.last_sync_words }} Ago ({{ repository.last_sync.ended_at | date:'medium' }} Local Time)":"{{ repository.last_sync_words }} 之前 ({{ repository.last_sync.ended_at | date:'medium' }} 本地時間)","{{ repository.last_sync_words }} ago":"{{ repository.last_sync_words }} 前","{{ repository.product.sync_plan.next_sync | date:'medium' }} (Server Time)":"{{ repository.product.sync_plan.next_sync | date:'medium' }}（伺服器時間）","{{ subscription.host.name }}":"{{ subscription.host.name }}","{{ subscriptionsTable.numSelected }} Selected":"選擇了 {{ subscriptionsTable.numSelected }}","{{ totalActivationKeyCount() }} Activation Keys will be moved to {{ deleteOptions.activationKeys.contentView.name }} in {{ deleteOptions.activationKeys.environment.name }}":"{{ totalActivationKeyCount() }} 組啟動金鑰將會被移至 {{ deleteOptions.activationKeys.contentView.name }} 於 {{ deleteOptions.activationKeys.environment.name }} 中","{{ totalHostCount() }} Content Hosts will be moved to {{ deleteOptions.contentHosts.contentView.name }} in {{ deleteOptions.contentHosts.environment.name }}":"{{ totalHostCount() }} 個內容主機將會被移至 {{ deleteOptions.contentHosts.contentView.name }} 於 {{ deleteOptions.contentHosts.environment.name }} 中","{{ type.display }}":"{{ type.display }}","{{ version.docker_image_count }} Docker Images":"{{ version.docker_image_count }} 個 Docker 映像檔","{{ version.docker_tag_count }} Docker Tags":"{{ version.docker_tag_count }} 個 Docker 標籤","{{ version.errata_counts.total }} Errata":"{{ version.errata_counts.total }} 個勘誤","{{ version.package_count }} Packages":"{{ version.package_count }} 個套件","{{ version.puppet_module_count }} Puppet Modules":"{{ version.puppet_module_count }} 個 Puppet 模組"});
/* jshint +W100 */
}]);
BASTION_MODULES.push(
    'Bastion.capsule-content',
    'Bastion.activation-keys',
    'Bastion.architectures' ,
    'Bastion.common',
    'Bastion.content-views',
    'Bastion.content-views.versions',
    'Bastion.debs',
    'Bastion.docker-tags',
    'Bastion.files',
    'Bastion.ostree-branches',
    'Bastion.hosts',
    'Bastion.puppet-modules',
    'Bastion.ostree-branches',
    'Bastion.environments',
    'Bastion.gpg-keys',
    'Bastion.hosts',
    'Bastion.capsules',
    'Bastion.organizations',
    'Bastion.products',
    'Bastion.repositories',
    'Bastion.subscriptions',
    'Bastion.sync-plans',
    'Bastion.host-collections',
    'Bastion.content-hosts',
    'Bastion.tasks',
    'Bastion.settings'
);
(function () {

    /**
     * @ngdoc run
     * @name Bastion.run:FeatureFlag
     *
     * @description
     *   Handles checking if a given feature is enabled. Provides this functionality on
     *   the root scope and also checks routes to send the user to a 404.
     */
    function KatelloFeatures(FeatureFlag) {
        var remoteActions = [
            'content-host.provisioning',
            'content-hosts.bulk-actions.packages',
            'content-hosts.bulk-actions.errata.list',
            'content-hosts.bulk-actions.erratum',
            'content-hosts.bulk-actions.errata.content-hosts',
            'content-hosts.bulk-actions.environment'
        ];

        var customProducts = [
            'products.new.form',
            'products.discovery',
            'products.discovery.scan',
            'products.discovery.create',
            'gpg-keys',
            'gpg-keys.new',
            'gpg-key.info',
            'gpg-key.products',
            'gpg-key.repositories'
        ];

        FeatureFlag.addStates('remote_actions', remoteActions);
        FeatureFlag.addStates('custom_products', customProducts);
    }

    angular
        .module('Bastion.features')
        .run(KatelloFeatures);

    KatelloFeatures.$inject = ['FeatureFlag'];

})();

































































































